#define SHORTVERSION "15.31"

// #define VERBOSE
// enable this for diagnostic output

// #define SHOWRUMBLE

// #define NOBCC
// enable this to change the AOF loader to expect files to be missing their
// BCC codes.

// #define BASICMODE
// whether to prefer memmap "E" for games which are compatible with either memmap
// but have extra features on a deluxe TVGC (otherwise uses memmap "F").

// #define DEBUGINTERRUPTS
// whether to print lots of information about which interrupts are generated or suppressed

// #define PROFILE
// whether or not to enable instruction profiling
//  (Slash (/) (AmiArcadia) or Ctrl+Shift+J (WinArcadia) to view profile report)
// #define SUPPRESSUNPROFILED
// if SUPPRESSUNPROFILED is defined, opcodes which have not been executed
//  at all will not be shown in the profile report.

// Defined by AmigaOS are: IMPORT, UBYTE, UWORD, SWORD, ULONG, LONG, TEXT, STRPTR.

#define TRANSIENT              auto   /* automatic variables */
#define MODULE                 static /* external static (file-scope) */
#define PERSIST                static /* internal static (function-scope) */
#define DISCARD                (void) /* discarded return values */
#define elif                   else if
#define acase                  break; case
#define adefault               break; default
#define EXPORT

#if !defined( __LCLINT__ ) && !defined(__amigaos4__)
    #define ASM    __asm
    #define REG(x) register __ ## x
#endif
#ifndef __amigaos4__
    #define UNUSED
#endif

typedef signed char        FLAG;    /*  8-bit signed quantity (replaces BOOL) */
typedef signed char        SBYTE;   /*  8-bit signed quantity (replaces Amiga BYTE) */
typedef signed short       SWORD;   /* 16-bit signed quantity (replaces Amiga WORD) */
typedef signed long        SLONG;   /* 32-bit signed quantity (same as LONG) */
typedef unsigned long      UHUGE;
typedef unsigned int       UINT;
typedef long               SHUGE;

typedef unsigned char      ASCREEN; // we use whatever is fastest!

#define EOS                0        // end of string (ie. ASCII NUL)

#define STICKCENTRE 0x70
// #define SPRINGLOADED

#define QUEUESIZE         2560 // 80*16*2=2560
#define REPEATRATE           5 // keys repeat every 5 frames (ie. 10 per second)

#ifdef AMIGA
    #define MAX_PATH       512 // you will have problems if this is too low!
#endif
// it's already built-in on IBM-PC

#define PROJECTSIZE (((64 + 2) * 32) + 1)

#define KILOBYTE          1024
#define ONE_MILLION    1000000
#define ONE_BILLION 1000000000

#define MINIWIDTH           48 // 1 + (8 * 6) - 1
                               // or, 1 + 4 + 2 + 4 + 2 + 4 + 2 + 4 + 2 + 4 + 2 + 4 + 2 + 4 + 2 + 4 + 1
                               //        1st     2nd     3rd     4th     5th     6th     7th     8th
#define MINIHEIGHT           7

#define NEWCONFIGVERSION     3 // V15.31+
#define OLDESTCONFIGVERSION  1 // V15.25
#define CONFIGMACHINE_AMIGA  0
#define CONFIGMACHINE_IBM    1

// true CPL (for Interton/Elektor) is about 56.75
#define PVI_CPL     56 // 56.75 or more doesn't work for Elektor monitor software (missing bottom text row)
#define PVI_VBLANK 271 // 271 (meaning 272 lines) is likely to be correct (eg. see p. 6 of PVI datasheet)
#define HIDDEN_X    15
/* Circledrive   (Elektor) prefers a value of >=  9 for this.
   Painting      (Elektor) prefers a value of >=  9 for this.
   Tiny Tim      (Elektor) prefers a value of >=  9 for this.
   Omega Landing (Elektor) prefers a value of >= 15 for this. */

#define KIND_ILBM  0
#define KIND_ACBM  1
#define KIND_BMP   2
#define KIND_PCX   3
// screenshot kinds must be first, savescreenshot() depends on it
#define KIND_BIN   4
#define KIND_PGM   5
#define KIND_TVC   6
#define KIND_COS   7
#define KIND_COR   8
#define KIND_EAP   9
#define KIND_EAL  10
#define KIND_ANIM 11
#define KIND_ZIP  12
#define KIND_RAW  13
#define KIND_AOF  14
#define KINDS     15 // counting from 1

#define NONPAGE 0x1fff          /* mask page offset */
#define PLEN    0x2000          /* page length */
#define PAGE    0x6000          /* mask page */
#define AMSK    0x7fff          /* mask address range */

// don't change these!
#define REGION_NTSC             0
#define REGION_PAL              1

#define SERIALIZE_READ          0
#define SERIALIZE_WRITE         1

// memflags[] bitfield
#define COVERAGE_WRITE        0x1
#define COVERAGE_READ         0x2
#define COVERAGE_EXECUTE      0x4
#define COVERAGE_ADDRESS      0x8
// view_coverage() depends on coverage bits being bits 0..3 and in this order!
#define COVERAGE_BITS    (COVERAGE_ADDRESS | COVERAGE_EXECUTE | COVERAGE_READ | COVERAGE_WRITE)
#define NOREAD               0x10
#define NOWRITE              0x20
#define READWRITE_BITS   (NOREAD | NOWRITE)
#define READONCE             0x40
#define AUDIBLE              0x80
#define MIRRORED_R          0x100
#define MIRRORED_W          0x200
#define EPROM               0x400
#define BREAKPOINT          0x800
#define WATCHPOINT         0x1000
#define STEPPOINT          0x2000
#define VISIBLE            0x4000
#define RANDOM             0x8000
#define TRAINABLE         0x10000
#define BANKU             0x20000
#define BANKQ             0x40000
#define BIOSRAM           0x80000
#define BANKED           0x100000 // for Zaccaria
#define FIELD            0x200000 // for Malzak

#define NET_OFF                 0
#define NET_CLIENT              1
#define NET_SERVER              2

#define NETPORT              6666

#define HOSTVOLUME_MIN          0
#define HOSTVOLUME_DEFAULT      4
#define HOSTVOLUME_MAX          8

#define A_OFFSET_UPPERSCREEN 0x1800
#define A_OFFSET_LOWERSCREEN 0x1A00
#define A_OFFSET_SPRITES     0x1980
#define A_OFFSET_UDCS        0x19A0

#define SENSITIVITY_VLOW    1
#define SENSITIVITY_LOW     2
#define SENSITIVITY_MEDIUM  3
#define SENSITIVITY_HIGH    4
#define SENSITIVITY_VHIGH   5
#define SENSITIVITY_MIN     1
#define SENSITIVITY_DEFAULT 3
#define SENSITIVITY_MAX     5

#define STYLE_SIGNETICS 0
#define STYLE_CALM      1

#define PH_VDUROWS     32
#define P_VDUCOLUMNS   32
#define H_VDUCOLUMNS   64
#define MAX_VDUROWS    PH_VDUROWS
#define MAX_VDUCOLUMNS H_VDUCOLUMNS

#define OUTOFRANGE     32768

#define GUESTCOLOURS   24

#define COMMASTRLEN    (13 + 1) // enough for "1,234,567,890" including EOS

#ifdef AMIGA
    #define T1         ;
    #define T2         ;
#endif
#ifdef WIN32
    #define T1         if (!consoleopen) openconsole();
    #define T2         reactivate();
#endif
#define CHECKINPUT     if (trace) checkinput();

// amiga|ibm.c
EXPORT void updatescreen(void);
EXPORT void redrawscreen(void);
EXPORT void readkybd(void);
EXPORT UBYTE KeyDown(UWORD scancode);
EXPORT void checkinput(void);
EXPORT void cleanexit(SBYTE rc);
EXPORT void playsound(FLAG force);
EXPORT void settitle(void);
EXPORT ULONG getsize(const STRPTR filename);
EXPORT void emu_pause(FLAG showcpu); // unistd.h has a pause() function!
EXPORT void emu_unpause(void);
EXPORT void openconsole(void);
EXPORT void cripple(void);
EXPORT void uncripple(void);
EXPORT void changemachine(int whichmachine, FLAG user);
EXPORT void updatebiggads(void);
EXPORT void updatesmlgad(int which, ULONG state, FLAG now);
EXPORT void updatesmlgads(void);
EXPORT void tick(int which, int ticked);
EXPORT void ghost(int which, int ghosted);
EXPORT void enable(int which, int enabled);
EXPORT void checkmx(int which, int first, int last);
EXPORT void uncheckmx(int which);
EXPORT void beep(void);
EXPORT void close_memory(void);
EXPORT void close_monitor(void);
EXPORT void close_controls(void);
EXPORT void close_gameinfo(void);
EXPORT void close_palette(void);
EXPORT void sound_off(FLAG force);
EXPORT void sound_on(void);
EXPORT void setbutton(int which, FLAG enabled, FLAG state);
EXPORT void say(STRPTR text, FLAG handlesound);
EXPORT void view_controls(void);
EXPORT void view_monitor(void);
EXPORT void view_memory(void);
EXPORT void edit_palette(void);
EXPORT void changefilepart(STRPTR newfilepart);
EXPORT void debug_set(int kind);
EXPORT void debug_setrastn(void);
EXPORT void debug_edit(void);
EXPORT void lockgui(void);
EXPORT void unlockgui(void);
EXPORT void clearmsgs(void);
EXPORT void updatefps(void);
EXPORT int asl(STRPTR hail, STRPTR pattern, FLAG save, STRPTR startdir,
 STRPTR startfile, STRPTR passedfn);
EXPORT void breakparts(void);
EXPORT void invert(int i);
EXPORT void hostnamewindow(void);
EXPORT void clearkybd(void);
EXPORT void listen_open(void);
EXPORT void listen_loop(void);
EXPORT void listen_close(void);
EXPORT void help_gameinfo(void);
EXPORT FLAG system_dopaddle(int source, int dest);
EXPORT void updatepointer(FLAG full);
EXPORT void audit(FLAG user);
EXPORT void TonePlay(int guestchan, float hertz, int volume);
EXPORT void NoisePlay(int guestchan, float hertz, int volume);
EXPORT void SoundClose(int guestchan);
EXPORT void SoundStop(int guestchan);
EXPORT void textpen(int colour);
EXPORT void updateprinter(void);
EXPORT void changecolours(void);
EXPORT void resize(int newsize);
EXPORT void clip_write(STRPTR data, int localsize, unsigned int kind);
EXPORT void clip_close(void);
EXPORT void edit_copytext(void);
EXPORT void edit_pastetext(void);
EXPORT void setselection(void);
EXPORT void make_stars(void);
EXPORT void play_sample(int sample);
EXPORT ULONG thetime(void);
EXPORT void drawpixel(int x, int y, int colour);
EXPORT void calcsize(void);
EXPORT void load_samples(void);
EXPORT void apply_rastlines(void);
EXPORT void apply_scale2x(void);
EXPORT void apply_scale3x(void);
EXPORT void apply_scale4x(void);
EXPORT void rectfill(int leftx, int topy);
EXPORT void edit_dips(void);
EXPORT void close_dips(void);
EXPORT void close_gameinfo(void);
EXPORT void s_updatedips(FLAG full);
EXPORT void g_updatedips(FLAG full);
EXPORT void m_updatedips(FLAG full);
EXPORT void ghost_dips(FLAG state);
EXPORT void buildlistbrowser(void);
EXPORT void view_hiscores(void);
EXPORT void datestamp(int which);
EXPORT void loadhiscores(void);
EXPORT void savehiscores(void);
EXPORT void closewindow(void);
EXPORT void openwindow(void);
EXPORT void setmemory(int whichgad, STRPTR contents);
EXPORT void audit_coinops(void);
EXPORT void dosidebar(void);
EXPORT void ReadJoystick(int joy);
EXPORT FLAG confirm(void);
EXPORT void help_manual(void);
EXPORT void updatespeech(void);
EXPORT void openurl(STRPTR command);
EXPORT void open_spriteviewer(void);
EXPORT void close_spriteviewer(void);
EXPORT void update_spriteviewer(FLAG force);
EXPORT void setsprview(int whichgad, int passedcolour);
EXPORT void system_loadedconfig(void);
EXPORT void resetcolour(int thecolourset, int i, FLAG update);
EXPORT void fixshadowcolours(void);
#ifdef AMIGA
    EXPORT void detachsmlgads(void);
    EXPORT void attachsmlgads(void);
    EXPORT void writeicon(int kind, STRPTR passedfn);
    EXPORT void setmonitor(struct Gadget* gadgetptr, STRPTR contents);
    EXPORT FLAG lockscreen(FLAG panic);
    EXPORT void unlockscreen(void);
    EXPORT void rq(STRPTR text);
    EXPORT LONG easyrequest(FLAG handlesound);
    EXPORT void lowerpri(void);
    EXPORT void raisepri(void);
    EXPORT void upgradeconfigs(void);
    EXPORT void edit_palette(void);
    EXPORT void amiga_setup(void);
    EXPORT void rexxmsgs(void);
    EXPORT void iconify(void);
    EXPORT void uniconify(void);
    EXPORT void stop_samples(void);
    EXPORT void free_samples(void);
    EXPORT void thesleep(int milliseconds);
    EXPORT void setjoyports(void);
    EXPORT void InitHook(struct Hook* hook, ULONG (*func)(), void* data);
    EXPORT void edit_s_dips(void);
    EXPORT void edit_m_dips(void);
    EXPORT void edit_g_dips(void);
    EXPORT void fullscreen_open(void);
    EXPORT void fullscreen_close(void);
    EXPORT void help_about(void);
    EXPORT void redefinekeys(void);
    EXPORT void slider(int kind);
    EXPORT void getpens(void);
    EXPORT void make_buffer(void);
    EXPORT void destroy_buffer(void);
    EXPORT void freepens(void);
    EXPORT void do_autopause(void);
    EXPORT void macro_arexx(FLAG wantasl);
    EXPORT FLAG Exists(STRPTR name);
    EXPORT void handlekybd(UWORD code);
    EXPORT void loadbigglyphs(void);
    EXPORT void loadmiscglyphs(void);
    EXPORT void memory_loop(void);
    EXPORT void palette_loop(void);
    EXPORT void update_memory_system(void);
    EXPORT void gameinfo_loop(void);
    EXPORT void rb_clearlist(struct List* ListPtr);
    EXPORT void flipbool(int* thevar, UNUSED int command);
    EXPORT void clip_open(UNUSED FLAG empty);
    EXPORT void loadsmlglyphs(void);
    EXPORT void enqueue(UWORD hostcode, TEXT guestcode);
    EXPORT void ch_clearlist(struct List* ListPtr);
    EXPORT void spriteviewer_loop(void);
    EXPORT void updatewheel(void);
    EXPORT void updatesliders(void);
    EXPORT void updatebrightness(void);
    #ifndef __amigaos4__
        EXPORT void updatefilter(void);
    #endif
#endif
#ifdef __PPC__
    EXPORT void InitModules(void);
    EXPORT void CleanupModules(void);
    EXPORT void InitModulesPPC(void);
    EXPORT void CleanupModulesPPC(void);
#endif
#ifdef WIN32
    EXPORT void mm_ReadJoystick(UWORD joynum);
    EXPORT void reactivate(void);
    EXPORT void setmonitor(HWND gadgetptr, STRPTR contents);
    EXPORT void OpenChannels(void);
    EXPORT void stop_sample(int sample);
    #ifdef FLASHSL
        EXPORT void leds(FLAG state);
    #endif
    EXPORT FLAG ctrl(void);
    EXPORT FLAG shift(void);
    EXPORT void iconify(void);
    EXPORT void uniconify(FLAG exiting);
    EXPORT void do_autopause(WPARAM wParam, LPARAM lParam);
    EXPORT void refreshkeys(HWND hwnd);
    EXPORT void handlekybd(UINT code);
    EXPORT void makedisplay(void);
    EXPORT void traymenu(void);
    EXPORT void handlealtkybd(UINT code);
    EXPORT void drawsky(int x, int y);
    EXPORT void clearscreen(void);
    EXPORT void load_catalog(void);
    EXPORT void parse_catalog(void);
    EXPORT void translate(void);
    EXPORT void getpointerpos(int* x, int* y);
    EXPORT void capslock_off(void);
    EXPORT void project_register(void);
    EXPORT void freelanguage(void);
    EXPORT void changelanguage(void);
    EXPORT void calibrate(void);
    EXPORT STRPTR GetCatalogStr(int catalog, LONG stringNum, STRPTR defaultString);
    BOOL CALLBACK    ControlsDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
    BOOL CALLBACK       AboutDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
    BOOL CALLBACK      MemoryDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
    BOOL CALLBACK     MonitorDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
    BOOL CALLBACK   RearrangeDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
    BOOL CALLBACK    RedefineDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
    BOOL CALLBACK RedefineSubDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
    BOOL CALLBACK    RegisterDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
    BOOL CALLBACK      s_DIPsDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
    BOOL CALLBACK      m_DIPsDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
    BOOL CALLBACK      g_DIPsDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
    BOOL CALLBACK   FrameSkipDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
    BOOL CALLBACK SensitivityDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
    BOOL CALLBACK      VolumeDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
    #ifdef USEDIRECTSOUND
        EXPORT void change_instrument(void);
    #endif
    EXPORT void setpri(void);
    EXPORT void flipbool(int* thevar, int command);
    EXPORT void clip_open(UNUSED FLAG empty);
    EXPORT void associate(int kind);
    EXPORT void enqueue(UINT hostcode, TEXT guestcode);
    EXPORT void drawpoint(int x, int y, COLORREF colour);
    EXPORT void doaxes(void);
#endif

// cpu.c
EXPORT void cpu(void);
EXPORT void cpu_setup(void);
EXPORT void disassemble(int address1, int address2);
EXPORT void view_next(void);

// uvi.c
EXPORT void uvi(void);
EXPORT void uviwrite(UWORD address, UBYTE data);
EXPORT void a_drawpixel(int x, int y, UBYTE colour);

// pvi.c
EXPORT void pvi(void);
EXPORT void pviwrite(signed int address, UBYTE data, FLAG ispvi);
EXPORT void the_drawline(int colour, int y);
EXPORT void ie_inputmirrors(int address);
EXPORT void ie_viewscreen(void);
EXPORT void mousedown(int mousex, int mousey);
EXPORT void mouseup(void);
EXPORT void domouse(void);
EXPORT void pvi_memmap(void);

// engine1|engine2.c
EXPORT void change_read(STRPTR thestring);
EXPORT void change_write(STRPTR thestring, FLAG quiet);
EXPORT void engine_setup(void);
EXPORT FLAG engine_load(FLAG silent);
EXPORT void engine_save(int kind, FLAG updatetitle);
EXPORT void engine_quit(void);
EXPORT void view_cram(void);
EXPORT void view_mram(void);
EXPORT void view_xvi(void);
EXPORT void view_udcs(void);
EXPORT void view_udcs_real(void);
EXPORT void view_cpu(FLAG linefeed);
EXPORT void view_psgs(void);
EXPORT void view_allmemory(void);
EXPORT void view_screen(void);
EXPORT void view_bram(void);
EXPORT void view_nvram(void);
EXPORT FLAG parse_bytes(FLAG loading, FLAG silent);
EXPORT void project_reset(FLAG keepalive);
EXPORT void project_open(void);
EXPORT void project_save(int kind);
EXPORT void runto(void);
EXPORT void updatemenu(int which);
EXPORT void updatemenus(void);
EXPORT int getfriendly(int number);
EXPORT int friendly_to_number(STRPTR thestring);
EXPORT int number_to_friendly(int number, STRPTR thestring, FLAG both);
EXPORT void docommand(int which);
EXPORT void fixextension(STRPTR extension, STRPTR passedfn, FLAG force);
EXPORT void macro_restartplayback(void);
EXPORT void macro_stop(void);
EXPORT void generate_crctable(void);
EXPORT void hex1(char* out, UBYTE value);
EXPORT void hex2(char* out, UWORD value);
EXPORT void decimal(char* out, unsigned long value);
EXPORT void savescreenshot(int kind, FLAG wantasl, FLAG clip);
EXPORT void debug_cli(FLAG showcpu);
EXPORT FLAG debug_command(void);
EXPORT void view_imagery(void);
EXPORT void view_coverage(void);
EXPORT void engine_dopaddle(int source, int dest);
EXPORT void update_controls(void);
EXPORT void view_controls_engine(void);
EXPORT void update_monitor(FLAG force);
EXPORT void update_memory(FLAG force);
EXPORT void initmonitor_engine(void);
EXPORT void closemonitors(FLAG both);
EXPORT void auditfile(STRPTR passedname, int passedsize, FLAG user);
EXPORT void sortaudit(void);
EXPORT FLAG NetExchange(char* In, const char* Out, int N);
EXPORT void NetPlay(int Switch);
EXPORT FLAG NetSendByte(UBYTE value);
EXPORT UBYTE NetReceiveByte(void);
EXPORT UBYTE ReadByte(void);
EXPORT void WriteByte(UBYTE what);
EXPORT void WriteWord(UWORD what);
EXPORT void free_iobuffer(void);
EXPORT void engine_set(int kind);
EXPORT void clearruntos(void);
EXPORT FLAG bp_add(int address);
EXPORT FLAG bp_clear(int address);
EXPORT FLAG bp_clearall(void);
EXPORT FLAG wp_add(int address);
EXPORT FLAG wp_clear(int address);
EXPORT FLAG wp_clearall(void);
EXPORT void s_updatecontrols(void);
EXPORT void edit_copytext_engine(void);
EXPORT int bin2int(STRPTR thestring);
EXPORT int oct2int(STRPTR thestring);
EXPORT void endofframe(void);
EXPORT void calc_wait(void);
EXPORT void do_wait(void);
EXPORT UBYTE ReadByteAt(int where);
EXPORT UWORD ReadWordAt(int where);
EXPORT UWORD ReadWord(void);
EXPORT ULONG ReadLong(void);
EXPORT void engine_reset(void);
EXPORT UBYTE thechin(void);
EXPORT void drawsegment(int position, int segment, int colour);
EXPORT void updatedips(void);
EXPORT void updatehiscores(void);
EXPORT void generate_autotext(void);
EXPORT int readbase(STRPTR input);
EXPORT void zprintf(const char* format, ...);
EXPORT void logfile_open(void);
EXPORT void logfile_close(void);
EXPORT void logport(int port, UBYTE data, FLAG handled, FLAG write);
EXPORT void emulate(void);
EXPORT void serialize_byte(UBYTE* var);
EXPORT void serialize_word(UWORD* var);
EXPORT void serialize_byte_int(int* var);
EXPORT void serialize_word_int(int* var);
EXPORT void serialize_byte_ulong(ULONG* var);
EXPORT void handle_menu(int command);
EXPORT void handle_menu2(int command);
EXPORT UBYTE readport(int port);
EXPORT void writeport(int port, UBYTE data);
EXPORT void makeaudittext(STRPTR outputstring);
EXPORT void calcspritecolour(void);
EXPORT void spritebutton(int x, int y);
EXPORT ULONG crc32(UBYTE* address, int thesize);
EXPORT void alloc_iobuffer(ULONG thesize);
EXPORT void serializeconfig(void);
EXPORT void loadconfig(void);
EXPORT void saveconfig(void);
EXPORT void fixupcolours(void);
EXPORT void dumprow(int i);
EXPORT void longcomma(ULONG value, STRPTR out);
#ifdef PROFILE
    EXPORT void printprofile(void);
#endif
#if (!defined(__SASC) && !defined(__VBCC__)) // eg. __GNUC__, __STORM__, WIN32
    EXPORT int stcd_l(const char* in, SLONG* lvalue);
    EXPORT int stch_l(const char* in, SLONG* lvalue);
    EXPORT int stcl_d(char* out, long lvalue);
    EXPORT int stcl_h(char* out, unsigned long lvalue);
    EXPORT int stcul_d(char* out, unsigned long lvalue);
    EXPORT char* stptok(const char *s, char *tok, int toklen, char *brk);
#endif

// arcadia.c
EXPORT void a_setmemmap(void);
EXPORT void a_viewscreen(void);
EXPORT void a_serializerom(void);
EXPORT void save_eap(int kind);
EXPORT void save_eal(void);

// interton.c
EXPORT void i_setmemmap(void);
EXPORT void id_serializerom(void);

// elektor.c
EXPORT void e_setmemmap(void);
EXPORT void monob(int y);
EXPORT UBYTE loadrand(void);
EXPORT void saverand(UBYTE value);

// pipbug.c
EXPORT void ph_cout(FLAG printroutine, UBYTE value);
EXPORT float pipbug_getpitch(UWORD value);
EXPORT void draw_pipbug(void);
EXPORT void p_emulate(void);
EXPORT void p_setmemmap(void);
EXPORT void p_drawdigit(int position, UBYTE value);
EXPORT UBYTE p_readport(int port);
EXPORT void p_writeport(int port, UBYTE data);

// instruct.c
EXPORT void s_drawcontrolspixel(int x, int y, int colour);
EXPORT void smiwrite(UWORD address, UBYTE data);
EXPORT void s_emulate(void);
EXPORT void s_setmemmap(void);
EXPORT void s_viewscreen(void);
EXPORT UBYTE s_readport(int port);
EXPORT void s_writeport(int port, UBYTE data);
EXPORT void s_drawglows(void);

// cd2650.c
EXPORT void draw_cd2650(void);
EXPORT void c_emulate(void);
EXPORT void c_setmemmap(void);
EXPORT void c_viewscreen(void);
EXPORT UBYTE c_readport(int port);
EXPORT void c_writeport(int port, UBYTE data);

// dolphin.c
EXPORT void d_emulate(void);
EXPORT void d_setmemmap(void);
EXPORT UBYTE d_readport(int port);
EXPORT void d_writeport(int port, UBYTE data);
EXPORT void d_drawglows(void);
EXPORT void dolphin_viewscreen(void);

// malzak.c
EXPORT void m_drawscreen(void);
EXPORT void switchbank(int whichbank);
EXPORT void m_setmemmap(void);
EXPORT void m_viewscreen(void);
EXPORT UBYTE m_readport(int port);
EXPORT void m_writeport(int port, UBYTE data);
EXPORT void m_emuinput(void);
EXPORT void m_savenvram(void);
EXPORT void mg_playerinput(int source, int dest);
EXPORT void m_viewplayfield(void);

// zaccaria.c
EXPORT void astrowars_setmemmap(void);
EXPORT void galaxia_setmemmap(void);
EXPORT void astrowars_drawscreen(void);
EXPORT void galaxia_drawscreen(void);
EXPORT UBYTE g_readport(int port);
EXPORT void g_writeport(int port, UBYTE data);
EXPORT void g_emuinput(void);
EXPORT void g_viewscreen(void);
EXPORT void g_viewcolours(void);

// chaos.c
EXPORT void h_setmemmap(void);
EXPORT UBYTE h_readport(int port);
EXPORT void h_writeport(int port, UBYTE data);
EXPORT void h_emulate(void);
EXPORT void read_rtc(void);

// phunsy.c
EXPORT void phunsy_emulate(void);
EXPORT UBYTE phunsy_readport(int port);
EXPORT void phunsy_writeport(int port, UBYTE data);
EXPORT void draw_phunsy(void);
EXPORT void phunsy_setmemmap(void);
EXPORT void ramtoubank(int bank);
EXPORT void ramtoqbank(int bank);
EXPORT void ubanktoram(int bank);
EXPORT void qbanktoram(int bank);
EXPORT void phunsy_viewscreen(void);
EXPORT float phunsy_getpianopitch(UWORD value);
EXPORT float phunsy_getfiftypitch(UWORD value);
EXPORT float phunsy_getbelmachpitch(UWORD value);

// unzip.c
EXPORT FLAG IsZip(STRPTR filename);

// amiga1.c|ff.c
EXPORT void ff_on(int joy);
EXPORT void ff_off(int joy);
EXPORT void ff_decay(int joy);

#ifdef WIN32
// ff.c
    EXPORT void ff_init(void);
    EXPORT void ff_initjoy(int joy);
    EXPORT void ff_die(void);
    EXPORT void ff_ReadJoystick(UWORD joynum);
// speech.c
    EXPORT void speech_init(void);
    EXPORT void speech_die(void);
    EXPORT void speak(void);
#endif

#ifdef AMIGA
    // saveanim.c
    EXPORT struct AnimInfo* OpenAnim(STRPTR Name, struct Screen* Screen);
    EXPORT BYTE             AddAnim(struct AnimInfo* Info);
    EXPORT VOID             CloseAnim(struct AnimInfo* Info, BYTE Success);
#endif
#ifdef WIN32
    // ibm.c
    EXPORT void             OpenAnim(STRPTR name);
    EXPORT void             AddAnim(void);
    EXPORT void             CloseAnim(void);
    EXPORT void             activateconsole(void);
#endif

#define PSU_S            128
#define PSU_F             64
#define PSU_II            32
// bits 4..3 are unused and unusable
#define PSU_SP             7 // %00000111
#define PSU_WRITABLE    0x67 // %01100111
#define PSU_READONLY    0x98 // %10011000

#define SNPSIZE_MAX    (301 * KILOBYTE) // 300K of RAM, 1K for header, CPU state, etc.

// these values are all in guest-pixels
#define PVI_XSIZE        256 // 227 is unsuitable for Galaxia

#define A_BOXWIDTH       160 // must be a multiple of 4!
#define A_BOXHEIGHT      262 // should be 222?
#define A_XMARGIN         16

#define IE_BOXWIDTH      168 // must be a multiple of 4! Was 180 until V11.4.
#define IE_BOXHEIGHT     269 // 252 is more likely to be correct

#define EFLAG            TRUE // default flagline for most Elektor games

#define P_BOXWIDTH       192 // must be a multiple of 4!
#define P_BOXHEIGHT      304

#define S_BOXWIDTH       180 // must be a multiple of 4!
#define S_BOXHEIGHT       74

#define C_BOXWIDTH       480 // must be a multiple of 4!
#define C_BOXHEIGHT      192 // 16*12=192

#define M_BOXWIDTH       240 // must be a multiple of 4!
#define M_BOXHEIGHT      240

#define G_BOXWIDTH       240 // must be a multiple of 4!
#define G_BOXHEIGHT      240

// Chaos 2
#define H_BOXWIDTH       384 // must be a multiple of 4!
#define H_BOXHEIGHT      260

// Dolphin
#define D_BOXWIDTH       100 // must be a multiple of 4!
#define D_BOXHEIGHT       74

#define PHUNSY_BOXWIDTH  384 // 64*6. must be a multiple of 4!
#define PHUNSY_BOXHEIGHT 256 // 32*8.

#define BOXWIDTH         C_BOXWIDTH  // widest box
#define BOXHEIGHT        P_BOXHEIGHT // tallest box
#define MAXBOXWIDTH      (BOXWIDTH  * 8)
#define MAXBOXHEIGHT     (BOXHEIGHT * 4)

// These are for Interton/Elektor, not Arcadia!
// Interton/Elektor colours are referred to in inverted RGB format!
#define BLACK              7 // 0,0,0 = 0
#define BLUE               6 // 0,0,1 = 1
#define GREEN              5 // 0,1,0 = 2 (different to Arcadia!)
#define CYAN               4 // 0,1,1 = 3 (different to Arcadia!)
#define RED                3 // 1,0,0 = 4 (different to Arcadia!)
#define PURPLE             2 // 1,0,1 = 5 (different to Arcadia!)
#define YELLOW             1 // 1,1,0 = 6
#define WHITE              0 // 1,1,1 = 7

#define GREY1              8
#define BROWN              9
#define DARKRED           11
#define DARKGREEN         13
#define DARKBLUE          14
#define DARKBLACK         15
#define GREY2             16
#define GREY3             17
#define GREY4             18
#define GREY5             19
#define GREY6             20
#define GREY7             21
#define ORANGE            22
#define PINK              23

// joysticks
#define JOYUP              1
#define JOYDOWN            2
#define JOYLEFT            4
#define JOYRIGHT           8
#define JOYFIRE1          16
#define JOYFIRE2          32
#define JOYFIRE3          64
#define JOYFIRE4         128
#define JOYSTART         256
#define JOYA             512
#define JOYB            1024
#define JOYRESET        2048
#define JOYPAUSE        4096
#define JOYAUTOFIRE     8192

// CPU
#define PSL_CC 192 // $11000000
#define PSL_IDC 32
#define PSL_RS  16
#define PSL_WC   8
#define PSL_OVF  4
#define PSL_COM  2
#define PSL_C    1

#define A_SPRITE0Y       6384 // $18F0 R/W
#define A_SPRITE0X       6385 // $18F1 R/W
#define A_SPRITE1Y       6386 // $18F2 R/W
#define A_SPRITE1X       6387 // $18F3 R/W
#define A_SPRITE2Y       6388 // $18F4 R/W
#define A_SPRITE2X       6389 // $18F5 R/W
#define A_SPRITE3Y       6390 // $18F6 R/W
#define A_SPRITE3X       6391 // $18F7 R/W
#define A_VSCROLL        6396 // $18FC -/W?
#define A_PITCH          6397 // $18FD -/W?
#define A_VOLUME         6398 // $18FE -/W?
#define A_CHARLINE       6399 // $18FF (dma)
#define A_P1LEFTKEYS     6400 // $1900
#define A_P1MIDDLEKEYS   6401 // $1901
#define A_P1RIGHTKEYS    6402 // $1902
#define A_P1PALLADIUM    6403 // $1903
#define A_P2LEFTKEYS     6404 // $1904
#define A_P2MIDDLEKEYS   6405 // $1905
#define A_P2RIGHTKEYS    6406 // $1906
#define A_P2PALLADIUM    6407 // $1907
#define A_CONSOLE        6408 // $1908
#define A_RESOLUTION     6648 // $19F8
#define A_BGCOLOUR       6649 // $19F9
#define A_SPRITES23CTRL  6650 // $19FA
#define A_SPRITES01CTRL  6651 // $19FB
#define A_BGCOLLIDE      6652 // $19FC
#define A_SPRITECOLLIDE  6653 // $19FD
#define A_P2PADDLE       6654 // $19FE
#define A_P1PADDLE       6655 // $19FF

#define IE_NOISE         0x1E80
#define IE_P1LEFTKEYS    0x1E88
#define IE_P1MIDDLEKEYS  0x1E89
#define IE_P1RIGHTKEYS   0x1E8A
#define IE_CONSOLE       0x1E8B
#define IE_P2LEFTKEYS    0x1E8C
#define IE_P2MIDDLEKEYS  0x1E8D
#define IE_P2RIGHTKEYS   0x1E8E

#define PVI_SPRITE0        0x00
#define PVI_SPRITE0AX      0x0A
#define PVI_SPRITE0BX      0x0B
#define PVI_SPRITE0AY      0x0C
#define PVI_SPRITE0BY      0x0D
#define PVI_SPRITE1        0x10
#define PVI_SPRITE1AX      0x1A
#define PVI_SPRITE1BX      0x1B
#define PVI_SPRITE1AY      0x1C
#define PVI_SPRITE1BY      0x1D
#define PVI_SPRITE2        0x20
#define PVI_SPRITE2AX      0x2A
#define PVI_SPRITE2BX      0x2B
#define PVI_SPRITE2AY      0x2C
#define PVI_SPRITE2BY      0x2D
#define PVI_SPRITE3        0x40
#define PVI_SPRITE3AX      0x4A
#define PVI_SPRITE3BX      0x4B
#define PVI_SPRITE3AY      0x4C
#define PVI_SPRITE3BY      0x4D
#define PVI_VERTGRID       0x80
#define PVI_HORIZ1         0xA8
#define PVI_HORIZ2         0xA9
#define PVI_HORIZ3         0xAA
#define PVI_HORIZ4         0xAB
#define PVI_HORIZ5         0xAC
#define PVI_SIZES          0xC0
#define PVI_SPR01COLOURS   0xC1
#define PVI_SPR23COLOURS   0xC2
#define PVI_SCORECTRL      0xC3
#define PVI_BGCOLOUR       0xC6
#define PVI_PITCH          0xC7
#define PVI_SCORELT        0xC8
#define PVI_SCORERT        0xC9
#define PVI_BGCOLLIDE      0xCA
#define PVI_SPRITECOLLIDE  0xCB
#define PVI_BGCOLLIDE2     0xDA
#define PVI_SPRITECOLLIDE2 0xDB
#define PVI_BGCOLLIDE3     0xEA
#define PVI_SPRITECOLLIDE3 0xEB
#define PVI_BGCOLLIDE4     0xFA
#define PVI_SPRITECOLLIDE4 0xFB
#define PVI_P1PADDLE       0xCC
#define PVI_P2PADDLE       0xCD

#define E_MONOB           0x800
#define E_MLINE           0x890
#define E_FSEQ            0x898
#define E_SILENC          0x898
#define E_ENTM            0x899
#define E_MFUNC           0x89A
#define E_MSCR            0x89B
#define E_RKBST           0x89D
#define E_LKBST           0x89E
#define E_MKBST           0x89F
#define E_FSCRM           0x8A0
#define E_REGM            0x8AC
#define E_BK1             0x8B5
#define E_BK2             0x8B7
#define E_INTADR          0x8B9
#define E_PC              0x8BE

#define E_PITCHA1_L      0x1D00
#define E_PITCHA1_H      0x1D01
#define E_PITCHB1_L      0x1D02
#define E_PITCHB1_H      0x1D03
#define E_PITCHC1_L      0x1D04
#define E_PITCHC1_H      0x1D05
#define E_PITCHD1        0x1D06
#define E_MIXER1         0x1D07
#define E_AMPLITUDEA1    0x1D08
#define E_AMPLITUDEB1    0x1D09
#define E_AMPLITUDEC1    0x1D0A
#define E_PERIOD1_L      0x1D0B
#define E_PERIOD1_H      0x1D0C
#define E_SHAPE1         0x1D0D
#define E_PORTA1         0x1D0E
#define E_PORTB1         0x1D0F

#define E_PITCHA2_L      0x1D10
#define E_PITCHA2_H      0x1D11
#define E_PITCHB2_L      0x1D12
#define E_PITCHB2_H      0x1D13
#define E_PITCHC2_L      0x1D14
#define E_PITCHC2_H      0x1D15
#define E_PITCHD2        0x1D16
#define E_MIXER2         0x1D17
#define E_AMPLITUDEA2    0x1D18
#define E_AMPLITUDEB2    0x1D19
#define E_AMPLITUDEC2    0x1D1A
#define E_PERIOD2_L      0x1D1B
#define E_PERIOD2_H      0x1D1C
#define E_SHAPE2         0x1D1D
#define E_PORTA2         0x1D1E
#define E_PORTB2         0x1D1F

#define E_RANDOM2        0x1D20
#define E_CASIN          0x1DBF
#define E_CASOUT         0x1DFF
#define E_RANDOM1G       0x1FAF
#define E_RANDOM1E       0x1FFF

#define T_COM             0x400
#define T_XGOT            0x409
#define T_TEMP            0x40D
#define T_TEMQ            0x40F
#define T_TEMR            0x411
#define T_TEMS            0x412
#define T_BUFF            0x413
#define T_BPTR            0x427
#define T_MCNT            0x428
#define T_CNT             0x429
#define T_CODE            0x42A
#define T_OKGO            0x42B
#define T_BCC             0x42C
#define T_MARK            0x42D
#define T_HDAT            0x42F
#define T_LDAT            0x431
#define T_HADR            0x433
#define T_LADR            0x435

#define S_IOPORT            0xFFF
#define S_DISLSD_SUB       0x1A76
#define S_GNPA_SUB         0x1B20
#define S_MOV_SUB          0x1DB6
#define S_DISPLY_SUB       0x1E13
#define S_USRDSP_SUB       0x1FD5
#define S_USRDSP           0x1FE6
#define S_DISPLY           0x1FEC
#define S_DISLSD           0x1FF4
#define S_GNPA             0x1FFC
#define S_MOV              0x1FFE

#define S_DIGITS           0x17D1 // $17D1..$17D8

// CD2650 RAM data
#define C_LENT             0x17EA
#define C_BKP1             0x17F4
#define C_BKP2             0x17F5
#define C_BPD1             0x17F6
#define C_BPD2             0x17F7
#define C_SUMC             0x17F9
#define C_ADD1             0x17FA
#define C_ADD2             0x17FB
#define C_CUR1             0x17FE
#define C_CUR2             0x17FF

#define MENUITEM_ANALOG            0
#define MENUITEM_AUTOFIRE          1
#define MENUITEM_AUTOSAVE          2
#define MENUITEM_EDIT              3
#define MENUITEM_CLEARBPS          4
#define MENUITEM_CLEARWPS          5
#define MENUITEM_CLIENT            6
#define MENUITEM_COLLISIONS        7
#define MENUITEM_CONTROLS          8
#define MENUITEM_CRAM              9
#define MENUITEM_CREATEICONS      10
#define MENUITEM_DEMULTIPLEX      11
#define MENUITEM_FLAGLINE         12
#define MENUITEM_FULLSCREEN       13
#define MENUITEM_ABOUT            14
#define MENUITEM_LEFT             15
#define MENUITEM_PASTETEXT        16
#define MENUITEM_INSTRUCTIONS     17
#define MENUITEM_READWRITES       18
#define MENUITEM_LOGEPROMS        19
#define MENUITEM_LOOP             20
#define MENUITEM_FORCEFEEDBACK    21
#define MENUITEM_MACHINE          22
#define MENUITEM_MACRO            23
#define MENUITEM_MANUAL           24
#define MENUITEM_MENUBAR          25
#define MENUITEM_VIEWMONITOR      26
#define MENUITEM_NARROW           27
#define MENUITEM_OPEN             28
#define MENUITEM_PAUSED           29
#define MENUITEM_POINTER          30
#define MENUITEM_PSGS             31
#define MENUITEM_QUICKLOAD        32
#define MENUITEM_QUICKSAVE        33
#define MENUITEM_REGION           34
#define MENUITEM_RESET            35
#define MENUITEM_RESETTOMONITOR   36
#define MENUITEM_REVERT           37
#define MENUITEM_RUNTODMA         38
#define MENUITEM_SAVESNP          39
#define MENUITEM_SAVEROM          40
#define MENUITEM_SAVETVC          41
#define MENUITEM_SENSEGAME        42
#define MENUITEM_SENSITIVITY      43
#define MENUITEM_SERVER           44
#define MENUITEM_CLEARBP          45
#define MENUITEM_CLEARWP          46
#define MENUITEM_SIZE             47
#define MENUITEM_SOUND            48
#define MENUITEM_STEP             49
#define MENUITEM_STRETCH          50
#define MENUITEM_SWAPPED          51
#define MENUITEM_TITLEBAR         52
#define MENUITEM_TOOLBAR          53
#define MENUITEM_TRACECPU         54
#define MENUITEM_USEHACKS         55
#define MENUITEM_TURBO            56
#define MENUITEM_WATCHREADS       57
#define MENUITEM_CLI              58
#define MENUITEM_REQUIREBUTTON    59
#define MENUITEM_RIGHT            60
#define MENUITEM_FILTERED         61
#define MENUITEM_IMAGERY          62
#define MENUITEM_JOY0             63
#define MENUITEM_LOGBIOS          64
#define MENUITEM_RUNTOINTERRUPT   65
#define MENUITEM_BRAM             66
#define MENUITEM_SUBROUTINES      67
#define MENUITEM_DOTSPACE         68
#define MENUITEM_GAMINGGUIDE      69
#define MENUITEM_AUDIT            70
#define MENUITEM_SIDEBAR          71
#define MENUITEM_SORTBY           72
#define MENUITEM_SAVEEAP          73
#define MENUITEM_GREYSCALE        74
#define MENUITEM_SAVEEAL          75
#define MENUITEM_RUNTOLOOPEND     76
#define MENUITEM_HIGHRES          77
#define MENUITEM_STARFIELDS       78
#define MENUITEM_SCREEN           79
#define MENUITEM_UDCS             80
#define MENUITEM_XVI              81
#define MENUITEM_RUNTORAST        82
#define MENUITEM_RUNTORASTN       83
#define MENUITEM_RUNTOFRAME       84
#define MENUITEM_REDEFINEKEYS     85
#define MENUITEM_FRAMESKIP        86
#define MENUITEM_VOLUME           87
#define MENUITEM_CONFINE          88
#define MENUITEM_TRACEVERBOSE     89
#define MENUITEM_ANSI             90
#define MENUITEM_INTERNAL         91
#define MENUITEM_AUTOPAUSE        92
#define MENUITEM_SPEECH           93
#define MENUITEM_STEPOVER         94
#define MENUITEM_PRINTER          95
#define MENUITEM_USEGUIDEPIXEL    96
#define MENUITEM_RUNTOSUBEND      97
#define MENUITEM_UNDITHER         98
#define MENUITEM_ANIMS            99
#define MENUITEM_RASTLINES       100
#define MENUITEM_GAMEINFO        101
#define MENUITEM_ICONIFY         102
#define MENUITEM_INTERRUPT       103
#define MENUITEM_CORNERS         104
#define MENUITEM_UNLIT           105
#define MENUITEM_RETUNE          106
#define MENUITEM_COPYTEXT        107
#define MENUITEM_LOWERCASE       108
#define MENUITEM_PALETTE         109
#define MENUITEM_CONSOLE         110
#define MENUITEM_SPRITESVISIBLE  111
#define MENUITEM_SHADOWS         112
#define MENUITEM_DIPSWITCHES     113
#define MENUITEM_INVERSE         114
#define MENUITEM_SCREENCOLOUR    115
#define MENUITEM_SCALE2X         116
#define MENUITEM_EAPSPRITES      117
#define MENUITEM_EAPUDCS         118
#define MENUITEM_VIEWHISCORES    119
#define MENUITEM_NVRAM           120
#define MENUITEM_LANGUAGE        121
#define MENUITEM_CODINGGUIDE     122
#define MENUITEM_LOGTOFILE       123
#define MENUITEM_IOPORTS         124
#define MENUITEM_VIEWMEMORY      125
#define MENUITEM_PRIORITY        126
#define MENUITEM_COLOURRAM       127
#define MENUITEM_SPRMAKER        128
#define MENUITEM_TXTMAKER        129
#define MENUITEM_CONFIRM         130
#define MENUITEM_REARRANGE       131
#define MENUITEM_CALIBRATE       132
#define MENUITEM_ROTATE          133
#define MENUITEM_IGNORE          134
#define MENUITEM_PLAYFIELD       135
#define MENUITEM_PAUSELOG        136
#define MENUITEM_VIEWALLMEMORY   137
#define MENUITEM_QUIT            138
#define MENUITEM_UNICONIFY       139
#define MENUITEM_REGISTER        140
#define MENUITEM_MRAM            141
#define MENUITEM_ADDBP           142
#define MENUITEM_ADDWP           143
#define MENUITEM_JUMP            144
#define MENUITEM_COPYSCRN        145
#define MENUITEM_SAVEACBM        146
#define MENUITEM_SAVEBMP         147
#define MENUITEM_SAVEILBM        148
#define MENUITEM_SAVEPCX         149
#define MENUITEM_VIEWCPU         150
#define MENUITEM_STARTRECORDING  151
#define MENUITEM_RESTARTPLAYBACK 152
#define MENUITEM_STOP            153
#define MENUITEM_BREAK           154
#define MENUITEM_AREXX           155
#define MENUITEM_CRYSTAL         156
#define MENUITEM_SPRITEVIEWER    157
#define MENUITEM_CPUTYPE         158
#define MENUITEM_MNEMONICS       159
#define MENUITEMS                160 // counting from 1 (ie. last entry is [MENUITEMS - 1])

// these numbers must not conflict with any MENUITEMs
#define COMMAND_ARCADIA          200
#define COMMAND_ELEKTOR          201
#define COMMAND_INTERTON         202
#define COMMAND_PIPBUG           203
#define COMMAND_INSTRUCTOR       204
#define COMMAND_CD2650           205
#define COMMAND_ASTROWARS        206
#define COMMAND_GALAXIA          207
#define COMMAND_MALZAK1          208
#define COMMAND_MALZAK2          209
#define COMMAND_CHAOS            210
#define COMMAND_DOLPHIN          211
#define COMMAND_PHUNSY           212
#define COMMAND_LOADER           213
#define COMMAND_SAVER            214

#define MENUOPT_1XSIZE             0
#define MENUOPT_2XSIZE             1
#define MENUOPT_3XSIZE             2
#define MENUOPT_SLOW               3
#define MENUOPT_FAST               4
#define MENUOPT_ARCADIA            5
#define MENUOPT_INTERTON           6
#define MENUOPT_ELEKTOR            7
#define MENUOPT_PIPBUG             8
#define MENUOPT_INSTRUCTOR         9
#define MENUOPT_CD2650            10
#define MENUOPT_ASTROWARS         11
#define MENUOPT_GALAXIA           12
#define MENUOPT_MALZAK1           13
#define MENUOPT_MALZAK2           14
#define MENUOPT_CHAOS             15
#define MENUOPT_DOLPHIN           16
#define MENUOPT_STARTRECORDING    17
#define MENUOPT_RESTARTPLAYBACK   18
#define MENUOPT_STOP              19
#define MENUMENU_LEFT             20
#define MENUOPT_LEFT_TRACKBALL    21
#define MENUOPT_LEFT_NONE         22
#define MENUOPT_LEFT_1STDJOY      23
#define MENUOPT_LEFT_2NDDJOY      24
#define MENUOPT_AMBER             25
#define MENUOPT_LEFT_2NDAJOY      26
#define MENUOPT_LEFT_1STDPAD      27
#define MENUOPT_LEFT_2NDDPAD      28
#define MENUOPT_LEFT_1STAPAD      29
#define MENUOPT_LEFT_2NDAPAD      30
#define MENUMENU_RIGHT            31
#define MENUOPT_RIGHT_TRACKBALL   32
#define MENUOPT_RIGHT_NONE        33
#define MENUOPT_RIGHT_1STDJOY     34
#define MENUOPT_RIGHT_2NDDJOY     35
#define MENUOPT_GREEN             36
#define MENUOPT_RIGHT_2NDAJOY     37
#define MENUOPT_RIGHT_1STDPAD     38
#define MENUOPT_RIGHT_2NDDPAD     39
#define MENUOPT_RIGHT_1STAPAD     40
#define MENUOPT_RIGHT_2NDAPAD     41
#define MENUOPT_NTSC              42
#define MENUOPT_PAL               43
#define MENUOPT_SORTBYNAME        44
#define MENUOPT_SORTBYMACHINE     45
#define MENUOPT_4XSIZE            46
#define MENUOPT_ENG               47
#define MENUOPT_HOL               48
#define MENUOPT_FRA               49
#define MENUOPT_GER               50
#define MENUOPT_ITA               51
#define MENUOPT_RUS               52
#define MENUOPT_HIGHPRI           53
#define MENUOPT_NORMALPRI         54
#define MENUOPT_LOWPRI            55
#define MENUOPT_SPA               56
#define MENUOPT_PHUNSY            57
#define MENUOPT_WHITE             58
#define MENUOPT_2650A             59
#define MENUOPT_2650B             60
#define MENUOPT_CALM              61
#define MENUOPT_SIGNETICS         62
#define MENUOPTS                  63 // counting from 1

// small gadget positions, left to right, on GID_SB2
// graphics
#define GADPOS_FLAGLINE     0
// input
#define GADPOS_AUTOFIRE     1
#define GADPOS_SWAPPED      2
// speed
#define GADPOS_PAUSED       3
#define GADPOS_TURBO        4
// sprites
#define GADPOS_COLLISIONS   5
// sound
#define GADPOS_SOUND        6
// macro (not really part of toolbar in Amiga version)
#define GADPOS_MACRO        7

#define NUMKEYS            28 // counting from 1 (last array index is NUMKEYS-1)

#define AUTOLINE_GAMENAME     0
#define AUTOLINE_CREDITS      1
#define AUTOLINE_REFERENCE    2
#define AUTOLINE_MACHINE      3
#define AUTOLINE_SIZE         4
#define AUTOLINE_COMPLANG     5
#define AUTOLINE_STARTADDR    6
#define AUTOLINE_AUTOFIRE     7
#define AUTOLINE_CONTROLLERS  8
#define AUTOLINE_DEMULTIPLEX  9
#define AUTOLINE_FLAGLINE    10
#define GAMEINFOLINES        11 // counting from 1 (last array index is GAMEINFOLINES-1)

#define KEYS               40 // counting from 1 (last array index is KEYS-1)

#define KIND_BP             0
#define KIND_WP             1
#define KIND_BC             2
#define KIND_WC             3

#define MONITORGADS       202 // counting from 1 (last array index is MONITORGADS-1)

#define ARCADIA         0
#define INTERTON        1
#define ELEKTOR         2
#define PIPBUG          3
#define INSTRUCTOR      4
#define CD2650          5
#define ASTROWARS       6
#define GALAXIA         7
#define MALZAK          8
#define CHAOS           9
#define DOLPHIN        10
#define PHUNSY         11
#define MACHINES       12 // counting from 1
#define FOLDERGLYPH    12
#define SYSFOLDERGLYPH 13

// these must start from 0 and be consecutive
#define MEMMAP_A          0 // Interton                    (2K ROM + 0K RAM)
#define MEMMAP_B          1 // Interton                    (4K ROM + 0K RAM)
#define MEMMAP_C          2 // Interton                    (4K ROM + 1K RAM)
#define MEMMAP_D          3 // Interton                    (6K ROM + 1K RAM)
#define MEMMAP_E          4 // Elektor                     (basic)
#define MEMMAP_F          5 // Elektor                     (expanded)
#define MEMMAP_G          6 // Arcadia    ( 512 bytes RAM) (Emerson)
#define MEMMAP_H          7 // Arcadia    (1024 bytes RAM) (Tele-Fever)
#define MEMMAP_I          8 // Arcadia    ( 512 bytes RAM) (Palladium)
#define MEMMAP_J          9 // PIPBUG     ( 256 bytes RAM) (EA 77up2 aka "Baby")
#define MEMMAP_K         10 // PIGBUG     ( 512 bytes RAM) (Adaptable Board Computer)
#define MEMMAP_L         11 // PIPBUG     (1024 bytes RAM) (1K EA 78up5 aka "Mini Computer")
#define MEMMAP_M         12 // PIPBUG     (        7K RAM) (7K EA 78up5 aka "Mini Computer")
#define MEMMAP_N         13 // Instructor ( 640 bytes RAM) (basic)
#define MEMMAP_O         14 // Instructor (  28K+128b RAM) (expanded)
#define MEMMAP_P         15 // PIPBUG     (       11K RAM) (11K EA 78up5 with EPROM)
#define MEMMAP_MALZAK1   16 // aka memmap "T"
#define MEMMAP_CD2650    17 // aka memmap "U"
#define MEMMAP_MALZAK2   18 // aka memmap "V"
#define MEMMAP_GALAXIA   19 // aka memmap "W"
#define MEMMAP_ASTROWARS 20 // aka memmap "X"
#define MEMMAP_CHAOS     21 // Chaos 2
#define MEMMAP_DOLPHIN   22 // Dolphin
#define MEMMAP_PHUNSY    23 // Phunsy
#define MEMMAP_ARCADIA   MEMMAP_G // default Arcadia memory map you prefer
#ifdef BASICMODE
    #define MEMMAP_ELEKTOR MEMMAP_E
#else
    #define MEMMAP_ELEKTOR MEMMAP_F
#endif
// memmaps "Q", "R" and "S" are PIPBUG and are not supported

#define EXIT_WARNING        5
#define EXIT_ERROR         10

#define FRAMESKIP_MIN       1
#define FRAMESKIP_MAX      10

EXPORT struct KeyTableStruct
{   UWORD a_address;
    UBYTE a_mask;
    UWORD ie_address;
    UBYTE ie_mask;
    STRPTR ai_name;
    STRPTR e_name;
};

EXPORT struct KeyNameStruct
{         STRPTR name;
    const TEXT   unshifted,
                 capslock,
                 shifted;
    const FLAG   reserved;
};

EXPORT struct KindStruct
{   const STRPTR extension,
                 pattern;
#ifdef WIN32
    const STRPTR extname,
                 desc;
#endif
#ifdef AMIGA
          STRPTR icon;
#endif
          STRPTR hail;
};

#define OVERLAYS       (183 + 1)
#define A_OVERLAY         0
#define I_OVERLAY        15
#define E_OVERLAY         1
#define S_OVERLAY         2
#define P_OVERLAY         3
#define C_OVERLAY         3
#define M1_OVERLAY      161
#define M2_OVERLAY      173
#define G_OVERLAY       160
#define PHUNSY_OVERLAY    3
#define KNOWNGAMES      882 // was (sizeof(known) / sizeof(known[0]))
#define FOUNDGAMES_MAX (KNOWNGAMES + 6) // this can be theoretically exceeded
EXPORT struct AuditStruct
{   TEXT  filename[MAX_PATH + 1],
          gamename[MAX_PATH + 1];
    SWORD whichgame;
    SBYTE glyph,
          machine,
          memmap;
};

#define SORTBYNAME    0
#define SORTBYMACHINE 1

#define DUMP_GOOD     0
#define DUMP_OK       1
#define DUMP_SUSPECT  2
#define DUMP_BAD      3

// One of these (only!) must be #defined.
#define EPROM_2716
// #define EPROM_2732

EXPORT struct KnownStruct
{   UBYTE  machine,
           memmap;
    ULONG  crc32;
    UWORD  size;
    SBYTE  key1,
           key2,
           key3,
           key4,
           up,
           down,
           left,
           right;
    FLAG   analog;
    UBYTE  sensitivity, // SI50: s_io
           demultiplex;
    FLAG   swapped,     // SI50: s_intselector
           flagline;    // SI50: s_intdir
    UBYTE  starfield,
           downframes,
           totalframes;
    SWORD  latestsprite,
           cpl,
           vblank;
    UBYTE  forceinterrupt,
           dump;
    SBYTE  p1sprcol[6],
           p1bgcol[4], // not UBYTE!
           p2sprcol[6],
           p2bgcol[4]; // not UBYTE!
    STRPTR ggnode,
           name,
           credits,
           reference;
    SWORD  whichoverlay;
    SBYTE  hiscore; // not UBYTE!
    SWORD  firstmnemonic,
           lastmnemonic;
    UBYTE  complang;
    UWORD  startaddr;
    UBYTE  banked;
    ULONG  the1staddr;
    UWORD  the1stsize;
    ULONG  the2ndaddr;
    UWORD  the2ndsize;
    UBYTE  kind;
// We could conceivably want to also do turbo, etc. in this manner.
};

EXPORT struct PaletteStruct
{   ULONG  red,
           green,
           blue;
};

EXPORT struct MachineStruct
{   STRPTR cli;
    int    width,
           height,
           memmap,
           region,
           wide,
           overlay;
    STRPTR codingguide;
    int    firstequiv,
           lastequiv,
           keys,
           maxsprites,
           monitor1;
    STRPTR monitor2;
    UBYTE  readonce,
           cosversion;
};
EXPORT struct HostMachineStruct
{   STRPTR titlebartext;
#ifdef WIN32
    int    icon;
#endif
#ifdef AMIGA
    STRPTR icon;
    STRPTR fn_controls;
#endif
};

EXPORT struct HiScoreStruct
{   UWORD hipos[4];
    UBYTE length;
    SBYTE nextdigit; // must be signed!
    ULONG score;
    TEXT  thetime[5 + 1],
          thedate[8 + 1];
};

EXPORT struct RTCStruct
{   int wkday,
        day,
        month,
        hour,
        min,
        sec;
};

#define CONTROLLER_NONE      0
#define CONTROLLER_TRACKBALL 1
#define CONTROLLER_1STDJOY   2
#define CONTROLLER_2NDDJOY   3
#define CONTROLLER_2NDAJOY   4
#define CONTROLLER_1STDPAD   5
#define CONTROLLER_2NDDPAD   6
#define CONTROLLER_1STAPAD   7
#define CONTROLLER_2NDAPAD   8

// don't change these!
#define COLOURFUL         0
#define GREYSCALE         1

#define PVI_LATESTSPRITE 271 // 251 is more likely to be correct

#define UVI_RASTLINES    262
#define PVI_RASTLINES    312

#define GOLFPOS            0
#define ESCAPEPOS         23
#define ROBOTKILLERPOS    40
#define TETRISPOS         61
#define SUPERINVPOS      116
#define LIFE1POS         480
#define LIFE2POS         481
#define MUSIC1POS        482
#define MUSIC2POS        483
#define PRINTERPOS       535
#define P_MORSE1POS      536
#define P_MORSE2POS      542
#define BELMACH1POS      668
#define FIFTYTHEMEPOS    797
#define PIANOPOS         804
#define BELMACH2POS      852
#define BELMACH0POS      863

#define HS_BASEBALL_POS    4
#define HS_ASTROWARS_POS  46
#define HS_GALAXIA_POS    47
#define HS_MALZAK1_POS    48
#define HS_MALZAK2_POS    49

#define LF                10
#define CR                13

#define UNLIT DARKBLUE
// BLACK, DARKRED, DARKBLUE, DARKGREEN are supported
// (but DARKGREEN is too light)

// if you don't want the cursor to flash (blink), enable this!
// #define SOLIDCURSOR

#define p_bgc BLACK
#define p_sgc GREY1

#define BLANK                0x60 // for PIPBUG

#define PORTC                256
#define PORTD                257

// these must correspond with the ordering of the relevant IDC_#? gadgets and constants
#define INTSEL_ACLINE        0
#define INTSEL_KYBD          1
#define INTSEL_DEFAULT       INTSEL_KYBD
#define PARALLEL_MEMMAPPED   0
#define PARALLEL_EXTENDED    1
#define PARALLEL_NONEXTENDED 2
#define PARALLEL_DEFAULT     PARALLEL_EXTENDED
#define INTDIR_DIRECT        0
#define INTDIR_INDIRECT      1
#define INTDIR_DEFAULT       INTDIR_DIRECT

#define A_HISCORES           46
#define HISCORES             50

EXPORT struct LangStruct
{
// initialized constants
    const int    mainmenu,
                 popupmenu,
                 traymenu,
                 minwidth1,
                 minwidth2,
                 minwidth3;
    const STRPTR filename,
                 tooltip[21];
// uninitialized (non-constants)
    UBYTE       *dosbuffer,
                *winbuffer;
    ULONG*       offset;
    int          size;
};

EXPORT struct Chunk
{   LONG          ckID;
    LONG          ckSize;
    LONG          ckType;
    void*         ckData;
    struct Chunk* ckNext;
};

EXPORT struct KeyRectStruct
{   UWORD left,
          top,
          width,
          height;
};

#ifdef __amigaos4__
    #define LLL (STRPTR)GetCatalogStr(CatalogPtr
#else
    #define LLL GetCatalogStr(CatalogPtr
#endif

#define CHARSET_DOS  0
#define CHARSET_WIN  1

#define BASE_DECIMAL 0
#define BASE_HEX     1
#define BASE_BINARY  2
#define BASE_OCTAL   3

// for unZIPping subsystem

typedef void* unzFile;
typedef struct tm_unz_s
{   UINT tm_sec;
    UINT tm_min;
    UINT tm_hour;
    UINT tm_mday;
    UINT tm_mon;
    UINT tm_year;
} tm_unz;
typedef struct unz_file_info_s
{   ULONG version;
    ULONG version_needed;
    ULONG flag;
    ULONG compression_method;
    ULONG dosDate;
    ULONG crc;
    ULONG compressed_size;
    ULONG uncompressed_size;
    ULONG size_filename;
    ULONG size_file_extra;
    ULONG size_file_comment;
    ULONG disk_num_start;
    ULONG internal_fa;
    ULONG external_fa;
    tm_unz tmu_date;
} unz_file_info;

EXPORT unzFile unzOpen(const char* path);
EXPORT int unzGoToFirstFile(unzFile file);
EXPORT int unzGoToNextFile(unzFile file);
EXPORT int unzClose(unzFile file);
EXPORT int unzGetCurrentFileInfo(unzFile file, unz_file_info* pfile_info, char* szFileName, ULONG fileNameBufferSize, void* extraField, ULONG extraFieldBufferSize, char* szComment, ULONG commentBufferSize);
EXPORT int unzOpenCurrentFile(unzFile file);
EXPORT int unzCloseCurrentFile(unzFile file);
EXPORT int unzReadCurrentFile(unzFile file, void* buf, unsigned len);

// computer languages
#define CL_MISC  0 // miscellaneous
#define CL_MC    1 // machine code
#define CL_TEXT  2 // text
#define CL_MWB   3 // MicroWorld BASIC (PHUNSY)
#define CL_LOTEL 4 // ?
#define CL_FORTH 5
#define CL_12KB  6 // 12K BASIC (CD2650)
#define CL_MB    7 // Micro BASIC (PIPBUG)

#define UBANKSIZE   0x800
#define UBANK0     0x8000
#define UBANK1     0x8800
#define UBANK2     0x9000
#define UBANK3     0x9800
#define UBANK4     0xA000
#define UBANK5     0xA800
#define UBANK6     0xB000
#define UBANK7     0xB800
#define UBANK8     0xC000
#define UBANK9     0xC800
#define UBANKA     0xD000
#define UBANKB     0xD800
#define UBANKC     0xE000
#define UBANKD     0xE800
#define UBANKE     0xF000
#define UBANKF     0xF800

#define QBANKSIZE  0x4000
#define QBANK0    0x10000
#define QBANK1    0x14000
#define QBANK2    0x18000
#define QBANK3    0x1C000
#define QBANK4    0x20000
#define QBANK5    0x24000
#define QBANK6    0x28000
#define QBANK7    0x2C000
#define QBANK8    0x30000
#define QBANK9    0x34000
#define QBANKA    0x38000
#define QBANKB    0x3C000
#define QBANKC    0x40000
#define QBANKD    0x44000
#define QBANKE    0x48000
#define QBANKF    0x4C000
