// INCLUDES---------------------------------------------------------------

#ifdef __VBCC__
    #define AMIGA
#endif
#ifdef AMIGA
    #include "amiga.h"
    #include <stdio.h>
#endif
#ifdef _lint
    #define WIN32
    #define _inline
#endif
#ifdef WIN32
    #include "ibm.h"
    #define EXEC_TYPES_H
#endif

#define CATCOMP_NUMBERS
#define CATCOMP_BLOCK
#include "aa_strings.h"
#ifdef AMIGA
    #include <proto/locale.h>  // GetCatalogStr()
#endif

#include "aa.h"
#include "cd2650.h"

#include <stdio.h>

// DEFINES----------------------------------------------------------------

// (none)

// IMPORTED VARIABLES-----------------------------------------------------

IMPORT FLAG                 changed;
IMPORT UBYTE                memory[32768];
IMPORT ASCREEN              screen[BOXWIDTH][BOXHEIGHT];
IMPORT UWORD                iar,
                            mirror_r[32768],
                            mirror_w[32768];
IMPORT ULONG                ea,
                            flagline,
                            memflags[32768];
IMPORT int                  cy_bgc,
                            cy_fgc,
                            consoleopen,
                            echoconsole,
                            echoprinter,
                            game,
                            inframe,
                            logmonitor,
                            machine,
                            overcalc,
                            shadowing,
                            slice;

#ifdef AMIGA
    IMPORT struct Catalog*  CatalogPtr;
    IMPORT FILE*            PrinterHandle;
#endif
#ifdef WIN32
    IMPORT int              CatalogPtr; // APTR doesn't work
    IMPORT HANDLE           PrinterHandle;
#endif

// EXPORTED VARIABLES

EXPORT UBYTE c_tone  = 0;
EXPORT int   fastcpu = FALSE;

// MODULE FUNCTIONS-------------------------------------------------------

MODULE __inline void local_drawpixel(int x, int y, int colour);

// CODE-------------------------------------------------------------------

EXPORT void draw_cd2650(void)
{   // all PERSISTent for speed
    PERSIST int   chessmen,
                  memptr,
                  x, xx, y, yy;
    PERSIST UBYTE tempvdu;

    if (machine != CD2650)
    {   return;
    }

    // generate screen[][] and display[] based on memory[]

    if (flagline) chessmen = 1; else chessmen = 0;

    if (shadowing)
    {   memptr = 0x1000;
        for (x = 0; x < 80; x++)
        {   for (y = 0; y < 16; y++)
            {   tempvdu = memory[memptr] & 0x7F;
                for (yy = 0; yy < 10; yy++)
                {   for (xx = 0; xx < 6; xx++)
                    {   if ((x == 79 && xx >= 4) || (y == 15 && yy == 9))
                        {   continue;
                        }

                        if (c_chars[chessmen][tempvdu][yy] & (0x10 >> xx))
                        {   local_drawpixel((x * 6) + xx + 2, (y * 12) + yy + 3, GREY1);
                        } else
                        {   local_drawpixel((x * 6) + xx + 2, (y * 12) + yy + 3, cy_bgc);
                }   }   }
                memptr++;
        }   }

        memptr = 0x1000;
        for (x = 0; x < 80; x++)
        {   for (y = 0; y < 16; y++)
            {   tempvdu = memory[memptr] & 0x7F;
                for (yy = 0; yy < 10; yy++)
                {   for (xx = 0; xx < 6; xx++)
                    {   if (x == 79 && xx == 5)
                        {   continue;
                        }

                        if (c_chars[chessmen][tempvdu][yy] & (0x10 >> xx))
                        {   local_drawpixel((x * 6) + xx + 1, (y * 12) + yy + 2, cy_fgc);
                        } elif (screen[(x * 6) + xx + 1][(y * 12) + yy + 2] != (UBYTE) GREY1)
                        {   local_drawpixel((x * 6) + xx + 1, (y * 12) + yy + 2, cy_bgc);
                }   }   }
                memptr++;
    }   }   }
    else
    {   memptr = 0x1000;
        for (x = 0; x < 80; x++)
        {   for (y = 0; y < 16; y++)
            {   tempvdu = memory[memptr] & 0x7F;
                for (yy = 0; yy < 10; yy++)
                {   for (xx = 0; xx < 6; xx++)
                    {   if (x == 79 && xx == 5)
                        {   continue;
                        }

                        if (c_chars[chessmen][tempvdu][yy] & (0x10 >> xx))
                        {   local_drawpixel((x * 6) + xx + 1, (y * 12) + yy + 2, cy_fgc);
                        } else
                        {   local_drawpixel((x * 6) + xx + 1, (y * 12) + yy + 2, cy_bgc);
                }   }   }
                memptr++;
    }   }   }

    changed = FALSE;
}

EXPORT void c_viewscreen(void)
{   SLONG x, y;
    TEXT  rowchars[80 + 1];

    rowchars[80] = 0;

    T1; textpen(TEXTPEN_VIEW);
    for (x = 0; x < 80; x++)
    {   for (y = 0; y < 16; y++)
        {   rowchars[y] = cd2650_chars[memory[0x1000 + y + (x * 16)] & 0x7F];
        }
        zprintf
        (   "$%04X: %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %s\n",
            (unsigned int) (0x1000 + (x * 16)),
            memory[0x1000 + (x * 16)],
            memory[0x1001 + (x * 16)],
            memory[0x1002 + (x * 16)],
            memory[0x1003 + (x * 16)],
            memory[0x1004 + (x * 16)],
            memory[0x1005 + (x * 16)],
            memory[0x1006 + (x * 16)],
            memory[0x1007 + (x * 16)],
            memory[0x1008 + (x * 16)],
            memory[0x1009 + (x * 16)],
            memory[0x100A + (x * 16)],
            memory[0x100B + (x * 16)],
            memory[0x100C + (x * 16)],
            memory[0x100D + (x * 16)],
            memory[0x100E + (x * 16)],
            memory[0x100F + (x * 16)],
            rowchars
        );
    }
    zprintf("\n");
    T2;
}

EXPORT void c_setmemmap(void)
{   int i, address;

    // assert(memmap == MEMMAP_U);

    game = FALSE;

    for (i = 0; i <= 0x3FE; i++)
    {   memory[i] = c_bios[i];
    }
    for (i = 0x3FF; i <= 0x7FFF; i++)
    {   memory[i] = 0;
    }
    for (i = 0x1000; i <= 0x14FF; i++)
    {   memory[i] = 0x20;
    }

    for (address = 0; address < 32 * KILOBYTE; address++)
    {   memflags[address] = 0;

        if     (address <= 0xFFF)
        {   memflags[address] |= NOWRITE;
        } elif (address <= 0x14FF)
        {   memflags[address] |= VISIBLE;
        } elif (address >= 0x1800 && address <= 0x1FFF)
        {   memflags[address] |= NOREAD | NOWRITE;
        }

        mirror_r[address] =
        mirror_w[address] = (UWORD) address;
}   }

EXPORT void c_emulate(void)
{   inframe = TRUE;

    if (fastcpu)
    {   // 4,732KHz = 4,732,000 cycles per second
        // 4,732,000 / 60 = 78866.6'
        slice = 78867 - overcalc; // really 78666.6'
    } else
    {   // 1,183KHz = 1,183,000 cycles per second
        // 1,183,000 / 60 = 19716.6'
        slice = 19717 - overcalc; // really 19716.6'
    }

    cpu();

    endofframe();
}

EXPORT void c_writeport(int port, UBYTE data)
{   FLAG handled = TRUE;

    switch (port)
    {
    case PORTD:
        // BIOS writes:
        //  %00000000 for ?
        //  %00001000 for ?
        // Morse Code writes:
        //  %00001111 for tone on
        //  %00000111 for tone off
        // so bit 3 is the important bit
        if (data == 0x0F)
        {   c_tone = 1;
        } else
        {   c_tone = 0;
        }
        playsound(FALSE);
    adefault:
        handled = FALSE;
    }

    logport(port, data, handled, TRUE);
}

EXPORT UBYTE c_readport(int port)
{   UBYTE t;
    FLAG  handled = TRUE;

    switch (port)
    {
    case PORTD:
        t = thechin();
    adefault:
        t = 0;
        handled = FALSE;
    }

    logport(port, t, handled, FALSE);
    return t;
}

MODULE __inline void local_drawpixel(int x, int y, int colour)
{   if (screen[x][y] != (UBYTE) colour)
    {   screen[x][y] = (UBYTE) colour;
        drawpixel(x, y, colour);
}   }
