// INCLUDES---------------------------------------------------------------

#ifdef __VBCC__
    #define AMIGA
#endif
#ifdef AMIGA
    #include "amiga.h"
#endif
#ifdef _lint
    #define WIN32
    #define _inline
#endif
#ifdef WIN32
    #include "ibm.h"
    #define EXEC_TYPES_H
#endif

#define CATCOMP_NUMBERS
#define CATCOMP_BLOCK
#include "aa_strings.h"
#ifdef AMIGA
    #include <proto/locale.h>  // GetCatalogStr()
#endif

#include <stdio.h>

#include "aa.h"
#include "chaos.h"

// DEFINES----------------------------------------------------------------

// (none)

// IMPORTED VARIABLES-----------------------------------------------------

IMPORT FLAG                 changed;
IMPORT UBYTE                memory[32768];
IMPORT UWORD                iar,
                            mirror_r[32768],
                            mirror_w[32768];
IMPORT ULONG                ea,
                            flagline,
                            frames,
                            memflags[32768];
IMPORT int                  consoleopen,
                            echoprinter,
                            game,
                            inframe,
                            logmonitor,
                            machine,
                            overcalc,
                            queuepos,
                            shadowing,
                            slice,
                            vducols;
IMPORT const STRPTR         asciiname[32];

#ifdef AMIGA
    IMPORT struct Catalog*  CatalogPtr;
    IMPORT FILE*            PrinterHandle;
#endif
#ifdef WIN32
    IMPORT int              CatalogPtr; // APTR doesn't work
    IMPORT HANDLE           PrinterHandle;
#endif

// EXPORTED VARIABLES-----------------------------------------------------

EXPORT struct RTCStruct     rtc;

// MODULE FUNCTIONS-------------------------------------------------------

// (none)

// CODE-------------------------------------------------------------------

EXPORT void h_setmemmap(void)
{   int i;

    // assert(memmap == MEMMAP_CHAOS);

    game    = FALSE;
    vducols = H_VDUCOLUMNS;

    for (i = 0;      i <= 0x2FFF; i++)
    {   memory[i]          = h_bios[i]; // $0000..$2FFF -> $0000..$2FFF
    }
    for (i = 0x3000; i <= 0x6FFF; i++)
    {   memory[i]          = 0;         // 0            -> $3000..$6FFF
    }
    for (i = 0x3000; i <= 0x3FFF; i++)
    {   memory[0x4000 + i] = h_bios[i]; // $3000..$3FFF -> $7000..$7FFF
    }

    for (i =      0; i <=  0x7FF; i++)
    {   memflags[i] = NOWRITE;
    }
    for (i =  0x800; i <=  0xFFF; i++)
    {   memflags[i] = 0;
    }
    for (i = 0x1000; i <= 0x1FFF; i++)
    {   memflags[i] = NOWRITE;
    }
    for (i = 0x2000; i <= 0x7FFF; i++)
    {   memflags[i] = 0;
    }

    for (i = 0; i <= 0x7FFF; i++)
    {   mirror_r[i] =
        mirror_w[i] = (UWORD) i;
    }

    memory[0x6000] =
    memory[0x617F] =
    memory[0x622B] =
    memory[0x631F] =
    memory[0x6509] =
    memory[0x65FC] =
    memory[0x69DB] = 0x17; // RETC,un
}

EXPORT UBYTE h_readport(int port)
{   UBYTE t;
    FLAG  handled = TRUE;

    switch (port)
    {
    case PORTD:
        // $03 (ETX) and $11 (DC1) seem to be important values.
        t = 0;      
    acase 0x1B:
        // Only bit 3 (%00001000) is ever actually read.
        t = 0;
    acase 0x1E:
        // Only bits 1..0 (%00000011) are ever actually read.
        t = 1;
    acase 0x1F: // this is both read and written
        // This is compared against $0D (CR).
        t = thechin();
        /* if (t != 0)
        {   T1;
            if (t >= 32)
            {   zprintf("READPORT: $%02X ('%c').\n", t, t);
            } else
            {   zprintf("READPORT: $%02X (%s).\n", t, asciiname[t]);
            }
            T2;
        } */
    acase 0x23:
        read_rtc();
        t = (UBYTE) (rtc.month  / 10);
    acase 0x24:
        read_rtc();
        t = (UBYTE) (rtc.month  % 10);
    acase 0x25:
        read_rtc();
        t = (UBYTE)  rtc.wkday;
        if (t == 0)
        {   t = 7;
        }
    acase 0x26:
        read_rtc();
        t = (UBYTE) (rtc.day    / 10);
    acase 0x27:
        read_rtc();
        t = (UBYTE) (rtc.day    % 10);
    acase 0x28:
        read_rtc();
        t = (UBYTE) (rtc.hour   / 10);
    acase 0x29:
        read_rtc();
        t = (UBYTE) (rtc.hour   % 10);
    acase 0x2A:
        read_rtc();
        t = (UBYTE) (rtc.min    / 10);
    acase 0x2B:
        read_rtc();
        t = (UBYTE) (rtc.min    % 10);
    acase 0x2C:
        read_rtc();
        t = (UBYTE) (rtc.sec    / 10);
    acase 0x2D:
        read_rtc();
        t = (UBYTE) (rtc.sec    % 10);
    acase 0x90:
        // This is a flag (it is always just compared to zero).
        if (queuepos)
        {   t = 1;
        } else
        {   t = 0;
        }
    acase 0x91:
        t = 0x80; // value of $81 provides the prompt but causes problems
        if (queuepos)
        {   t |= 0x08;
        }
    adefault:
        t = 0;
        handled = FALSE;
    }

    logport(port, t, handled, FALSE);
    return t;
}

EXPORT void h_writeport(int port, UBYTE data)
{   FLAG handled = TRUE;

    switch (port)
    {
    case 0x1F: // this is both read and written
        ph_cout(FALSE, data);
        if (data == CR)
        {   ph_cout(FALSE, LF);
        }
    /* acase 0x92:
        ph_cout(FALSE, data);
        if (data == CR)
        {   ph_cout(FALSE, LF);
        } */
    adefault: // eg. $18, $1A, $92
        handled = FALSE;
    }

    logport(port, data, handled, TRUE);
}

EXPORT void h_emulate(void)
{   inframe = TRUE;
    changed = TRUE;

    // 800KHz = 800,000 cycles per second
    // 800,000 / 50 = 16,000
    slice = 16000 - overcalc;
    cpu();

    endofframe();
}

