#include <windows.h>
#include "dinput.h"

#include "ibm.h"
#include "aa.h"

#include <stdio.h>
#include <math.h>

// #define FFVERBOSE

EXPORT struct IDirectInput*        lpdi         = NULL;
EXPORT struct IDirectInputDevice2* ffjoy2[2]    = { NULL,  NULL };
EXPORT TEXT                        joyname[2][MAX_PATH];
EXPORT int                         rumbling[2]  = {    0,     0 };

IMPORT UBYTE                       button[2][8],
                                   jx[2], jy[2];
IMPORT UWORD                       jf[2];
IMPORT ULONG                       swapped;
IMPORT int                         joys,
                                   machine,
                                   p1rumble,
                                   p2rumble,
                                   unit[2],
                                   useff;
IMPORT HINSTANCE                   InstancePtr;
IMPORT HWND                        MainWindowPtr;
#ifndef _lint
    IMPORT int                     consoleopen;
#endif
IMPORT const DWORD                 joyfires[8];

MODULE FLAG                        rumblable[2] = { FALSE, FALSE };
MODULE struct IDirectInputEffect*  ffeffect[2]  = { NULL,  NULL  };

BOOL CALLBACK DIEnumDevicesProc(LPCDIDEVICEINSTANCE lpddi, LPVOID pvRef);
MODULE void joySetForcesXY(int joy, int x, int y);
MODULE void acquire(int joy);

EXPORT void ff_init(void)
{   HRESULT hr;

    hr = DirectInputCreate(InstancePtr, DIRECTINPUT_VERSION, &lpdi, NULL); 
    switch (hr)
    {
    case DI_OK:
        ;
#ifdef FFVERBOSE
        T1; zprintf("Opened DirectInput.\n"); T2;
    acase DIERR_BETADIRECTINPUTVERSION:
        T1; zprintf("DirectInputCreate(): Beta version!\n"); T2;
        return;
    acase DIERR_INVALIDPARAM:
        T1; zprintf("DirectInputCreate(): Invalid parameter!\n"); T2;
        return;
    acase DIERR_OLDDIRECTINPUTVERSION:
        T1; zprintf("DirectInputCreate(): Old version!\n"); T2;
        return;
    acase DIERR_OUTOFMEMORY:
        T1; zprintf("DirectInputCreate(): Out of memory!\n"); T2;
        return;
#endif
    adefault:
#ifdef FFVERBOSE
        T1; zprintf("DirectInputCreate(): Unknown error!\n"); T2;
#endif
        return;
    }

    joys = 0;
    lpdi->lpVtbl->EnumDevices
    (   lpdi,
        DIDEVTYPE_JOYSTICK, 
        DIEnumDevicesProc,
        NULL, 
        DIEDFL_ATTACHEDONLY
    );

    if (joys == 0)
    {
#ifdef FFVERBOSE
        T1; zprintf("No devices found!\n"); T2;
#endif
        return;
    } // implied else
#ifdef FFVERBOSE
    T1; zprintf("%d device(s) found.\n", joys); T2;
#endif
}

BOOL CALLBACK DIEnumDevicesProc(LPCDIDEVICEINSTANCE lpddi, LPVOID pvRef)
{   HRESULT             hr;
    LPDIRECTINPUTDEVICE ffjoy1;

#ifdef FFVERBOSE
    T1; zprintf("Found a device.\n"); T2;
#endif

    // create game device (ie. create V1 interface)
    hr = lpdi->lpVtbl->CreateDevice(lpdi, &lpddi->guidInstance, &ffjoy1, NULL); 
    switch (hr)
    {
    case DI_OK:
        ;
#ifdef FFVERBOSE
        T1; zprintf("Created device.\n"); T2;
    acase DIERR_DEVICENOTREG:
        T1; zprintf("CreateDevice(): Device not registered!\n"); T2;
        return DIENUM_CONTINUE;
    acase DIERR_INVALIDPARAM:
        T1; zprintf("CreateDevice(): Invalid parameter!\n"); T2;
        ffjoy1->lpVtbl->Release(ffjoy1);
        return DIENUM_CONTINUE;
    acase DIERR_NOINTERFACE:
        T1; zprintf("CreateDevice(): No interface!\n"); T2;
        ffjoy1->lpVtbl->Release(ffjoy1);
        return DIENUM_CONTINUE;
    acase DIERR_NOTINITIALIZED:
        T1; zprintf("CreateDevice(): Not initialized!\n"); T2;
        ffjoy1->lpVtbl->Release(ffjoy1);
        return DIENUM_CONTINUE;
    acase DIERR_OUTOFMEMORY:
        T1; zprintf("CreateDevice(): Out of memory!\n"); T2;
        ffjoy1->lpVtbl->Release(ffjoy1);
        return DIENUM_CONTINUE;
#endif
    adefault:
#ifdef FFVERBOSE
        T1; zprintf("CreateDevice(): Unknown error!\n"); T2;
#endif
        ffjoy1->lpVtbl->Release(ffjoy1);
        return DIENUM_CONTINUE;
    } 

    // get V2 interface
    hr = ffjoy1->lpVtbl->QueryInterface(ffjoy1, &IID_IDirectInputDevice2, &ffjoy2[joys]);
    // release V1 interface
    ffjoy1->lpVtbl->Release(ffjoy1);
    if FAILED(hr)
    {
#ifdef FFVERBOSE
        T1; zprintf("Failed to obtain interface!\n"); T2;
#endif
        return DIENUM_CONTINUE;
    }

    strcpy(joyname[joys], lpddi->tszProductName);
    if (joys == 0)
    {   strcat(joyname[joys], " (&1)");
    } else
    {   // assert(joys == 1);
        strcat(joyname[joys], " (&2)");
    }

    unit[joys] = 0;
    joys++;
    if (joys == 2)
    {   return DIENUM_STOP; // Two is enough.
    } else
    {   // assert(joys == 1);
        return DIENUM_CONTINUE;
}   }

EXPORT void ff_die(void)
{   int i;

    if (!lpdi || !joys)
    {   return;
    }

    for (i = 0; i < joys; i++)
    {   if (ffeffect[i])
        {   ffeffect[i]->lpVtbl->Release(ffeffect[i]);
            ffeffect[i] = NULL;
        }
        if (ffjoy2[i])
        {   if (ffjoy2[i]->lpVtbl->Unacquire(ffjoy2[i]) == DI_OK)
            {   ;
#ifdef FFVERBOSE
                T1; zprintf("Unacquired device.\n"); T2;
#endif
            } else
            {   ;
#ifdef FFVERBOSE
                T1; zprintf("Can't unacquire device!\n"); T2;
#endif
            }
            ffjoy2[i]->lpVtbl->Release(ffjoy2[i]);
            ffjoy2[i] = NULL;
    }   }

    if (lpdi)
    {   lpdi->lpVtbl->Release(lpdi); 
        lpdi = NULL; 
}   }

EXPORT void ff_ReadJoystick(UWORD joynum)
{   HRESULT    hr;
    DIJOYSTATE js; 
    FLAG       ok;

    // joynum will be '0' if they want the PRIMARY port (the FIRST detected joystick)
    // joynum will be '1' if they want the SECONDARY port (the SECOND detected joystick)

    if (joynum >= joys || !ffjoy2[joynum])
    {   jx[joynum] =
        jy[joynum] = 128;
        jf[joynum] = 0;
        return;
    }

    ok = FALSE;
    do
    {   hr = ffjoy2[joynum]->lpVtbl->Poll(ffjoy2[joynum]);
        switch (hr)
        {
        case DI_OK:
            ok = TRUE;
#ifdef FFVERBOSE
            T1; zprintf("Polled data.\n"); T2;
#endif
        acase DIERR_INPUTLOST:
#ifdef FFVERBOSE
            T1; zprintf("Poll(): Input lost!\n"); T2;
#endif
            ; // acquire(joynum);
#ifdef FFVERBOSE
        acase DIERR_NOTACQUIRED:
            T1; zprintf("Poll(): Not acquired!\n"); T2;
#endif
            ; // acquire(joynum);
#ifdef FFVERBOSE
            return;
        acase DIERR_NOTINITIALIZED:
            T1; zprintf("Poll(): Not initialized!\n"); T2;
            return; 
#endif
        adefault:
#ifdef FFVERBOSE
            T1; zprintf("Poll(): Unknown error!\n"); T2;
#endif
            return; 
    }   }
    while (!ok);
 
    ok = FALSE;
    do
    {   hr = ffjoy2[joynum]->lpVtbl->GetDeviceState(ffjoy2[joynum], sizeof(DIJOYSTATE), &js); 
        switch (hr)
        {
        case DI_OK:
            ok = TRUE;
#ifdef FFVERBOSE
            T1; zprintf("Read data.\n"); T2;
#endif
        acase DIERR_INPUTLOST:
#ifdef FFVERBOSE
            T1; zprintf("GetDeviceState(): Input lost!\n"); T2;
#endif
            ; // acquire(joynum);
#ifdef FFVERBOSE
        acase E_PENDING:
            T1; zprintf("GetDeviceState(): Data is not yet available.\n"); T2;
            return; 
        acase DIERR_INVALIDPARAM:
            T1; zprintf("GetDeviceState(): Invalid parameter!\n"); T2;
            return; 
        acase DIERR_NOTACQUIRED:
            T1; zprintf("GetDeviceState(): Not acquired!\n"); T2;
            return; 
        acase DIERR_NOTINITIALIZED:
            T1; zprintf("GetDeviceState(): Not initialized!\n"); T2;
            return; 
#endif
        adefault:
#ifdef FFVERBOSE
            T1; zprintf("GetDeviceState(): Unknown error!\n"); T2;
#endif
            return; 
    }   }
    while (!ok);

    jx[joynum] = (UBYTE) (js.lX >> 8);
    jy[joynum] = (UBYTE) (js.lY >> 8);
    if (machine == ARCADIA && jx[joynum] >= 0x7E && jx[joynum] <= 0x81) // && game == ALIENINV
    {   jx[joynum] = 0x7D;
    }

    jf[joynum] = 0;
    if (js.rgbButtons[ 0] & 0x80) jf[joynum] |= joyfires[button[joynum][0] - 1];
    if (js.rgbButtons[ 1] & 0x80) jf[joynum] |= joyfires[button[joynum][1] - 1];
    if (js.rgbButtons[ 2] & 0x80) jf[joynum] |= joyfires[button[joynum][2] - 1];
    if (js.rgbButtons[ 3] & 0x80) jf[joynum] |= joyfires[button[joynum][3] - 1];
    if (js.rgbButtons[ 4] & 0x80) jf[joynum] |= joyfires[button[joynum][4] - 1];
    if (js.rgbButtons[ 5] & 0x80) jf[joynum] |= joyfires[button[joynum][5] - 1];
    if (js.rgbButtons[ 6] & 0x80) jf[joynum] |= joyfires[button[joynum][6] - 1];
    if (js.rgbButtons[ 7] & 0x80) jf[joynum] |= joyfires[button[joynum][7] - 1];
    if (js.rgbButtons[ 8] & 0x80) jf[joynum] |= JOYRESET;
    if (js.rgbButtons[ 9] & 0x80) jf[joynum] |= JOYSTART;
    if (js.rgbButtons[10] & 0x80) jf[joynum] |= JOYFIRE1;
    if (js.rgbButtons[11] & 0x80) jf[joynum] |= JOYFIRE1;
}

EXPORT void ff_initjoy(int joy)
{   HRESULT         hr;
    DWORD           rgdwAxes[2] = { DIJOFS_X, DIJOFS_Y }; 
    LONG            rglDirection[2] = { 0, 0 }; 
    DICONSTANTFORCE cf = { 0 }; 
    DIDEVCAPS       devcaps;
    DIEFFECT        eff; 
    DIPROPDWORD     DIPropAutoCenter;

    // set game data format
    hr = ffjoy2[joy]->lpVtbl->SetDataFormat(ffjoy2[joy], &c_dfDIJoystick);
    switch (hr)
    {
    case DI_OK:
        ;
#ifdef FFVERBOSE
        T1; zprintf("Set data format.\n"); T2;
    acase DIERR_ACQUIRED:
        T1; zprintf("SetDataFormat(): Already acquired!\n"); T2;
        return;
    acase DIERR_INVALIDPARAM:
        T1; zprintf("SetDataFormat(): Invalid parameter!\n"); T2;
        return;
    acase DIERR_NOTINITIALIZED:
        T1; zprintf("SetDataFormat(): Not initialized!\n"); T2;
        return;
#endif
    adefault:
#ifdef FFVERBOSE
        T1; zprintf("SetDataFormat(): Unknown error!\n"); T2;
#endif
        return;
    }

    devcaps.dwSize = sizeof(DIDEVCAPS);
    hr = ffjoy2[joy]->lpVtbl->GetCapabilities(ffjoy2[joy], &devcaps);
    switch (hr)
    {
    case DI_OK:
        ;
#ifdef FFVERBOSE
        T1; zprintf("Ascertained device capabilities.\n"); T2;
    acase DIERR_INVALIDPARAM:
        T1; zprintf("GetCapabilities(): Invalid parameter!\n"); T2;
        return;
    acase DIERR_NOTINITIALIZED:
        T1; zprintf("GetCapabilities(): Not initialized!\n"); T2;
        return;
#endif
    adefault:
#ifdef FFVERBOSE
        T1; zprintf("GetCapabilities(): Unknown error!\n"); T2;
#endif
        return;
    }
    if (devcaps.dwFlags & DIDC_FORCEFEEDBACK)
    {   rumblable[joy] = TRUE;
#ifdef FFVERBOSE
        T1; zprintf("Unit %d supports force feedback.\n", joy); T2;
#endif
    } else
    {   rumblable[joy] = FALSE;
#ifdef FFVERBOSE
        T1; zprintf("Unit %d doesn't support force feedback.\n", joy); T2;
#endif
    }

    // set cooperative level
    hr = ffjoy2[joy]->lpVtbl->SetCooperativeLevel(ffjoy2[joy],
        MainWindowPtr,
        rumblable[joy] ?
        DISCL_BACKGROUND | DISCL_EXCLUSIVE    // DISCL_FOREGROUND causes input loss issues
        :
        DISCL_BACKGROUND | DISCL_NONEXCLUSIVE // DISCL_FOREGROUND causes input loss issues
    );
    switch (hr)
    {
    case DI_OK:
        ;
#ifdef FFVERBOSE
        T1; zprintf("Set cooperative level.\n"); T2;
    acase DIERR_INVALIDPARAM:
        T1; zprintf("SetCooperativeLevel(): Invalid parameter!\n"); T2;
        return;
    acase DIERR_NOTINITIALIZED:
        T1; zprintf("SetCooperativeLevel(): Not initialized!\n"); T2;
        return;
#endif
    adefault:
#ifdef FFVERBOSE
        T1; zprintf("SetCooperativeLevel(): Unknown error!\n"); T2;
#endif
        return;
    }

    if (!rumblable[joy])
    {   acquire(joy);
        return;
    }

    DIPropAutoCenter.diph.dwSize       = sizeof(DIPropAutoCenter);
    DIPropAutoCenter.diph.dwHeaderSize = sizeof(DIPROPHEADER);
    DIPropAutoCenter.diph.dwObj        = 0;
    DIPropAutoCenter.diph.dwHow        = DIPH_DEVICE;
    DIPropAutoCenter.dwData            = 0;
    hr = ffjoy2[joy]->lpVtbl->SetProperty(ffjoy2[joy], DIPROP_AUTOCENTER, &DIPropAutoCenter.diph);
    switch (hr)
    {
    case DI_OK:
        ;
#ifdef FFVERBOSE
        T1; zprintf("Changed device property.\n"); T2;
    acase DI_PROPNOEFFECT:
        T1; zprintf("SetProperty(): No effect.\n"); T2;
    acase DIERR_INVALIDPARAM:
        T1; zprintf("SetProperty(): Invalid parameter!\n"); T2;
        return;
    acase DIERR_NOTINITIALIZED:
        T1; zprintf("SetProperty(): Not initialized!\n"); T2;
        return;
    acase DIERR_OBJECTNOTFOUND:
        T1; zprintf("SetProperty(): Object not found!\n"); T2;
        return;
    acase DIERR_UNSUPPORTED:
        T1; zprintf("SetProperty(): Unsupported property!\n"); T2;
#endif
    adefault:
#ifdef FFVERBOSE
        T1; zprintf("SetProperty(): Unknown error!\n"); T2;
#endif
        return;
    }
 
    eff.dwSize                  = sizeof(DIEFFECT); 
    eff.dwFlags                 = DIEFF_CARTESIAN | DIEFF_OBJECTOFFSETS; 
    eff.dwDuration              = INFINITE; 
    eff.dwSamplePeriod          = 0; 
    eff.dwGain                  = DI_FFNOMINALMAX; 
    eff.dwTriggerButton         = DIEB_NOTRIGGER; 
    eff.dwTriggerRepeatInterval = 0; 
    eff.cAxes                   = 2; 
    eff.rgdwAxes                = rgdwAxes; 
    eff.rglDirection            = rglDirection; 
    eff.lpEnvelope              = 0; 
    eff.cbTypeSpecificParams    = sizeof(DICONSTANTFORCE); 
    eff.lpvTypeSpecificParams   = &cf; 
    hr = ffjoy2[joy]->lpVtbl->CreateEffect(ffjoy2[joy], &GUID_ConstantForce, &eff, &ffeffect[joy], NULL);
    switch (hr)
    {
    case DI_OK:
        ;
#ifdef FFVERBOSE
        T1; zprintf("Created effect.\n"); T2;
#endif
    acase S_FALSE:
        ;
#ifdef FFVERBOSE
        T1; zprintf("CreateEffect(): Already created effect.\n"); T2;
    acase DIERR_INVALIDPARAM:
        T1; zprintf("CreateEffect(): Invalid parameter!\n"); T2;
        return;
    acase DIERR_NOTINITIALIZED:
        T1; zprintf("CreateEffect(): Not initialized!\n"); T2;
        return;
    acase DIERR_DEVICENOTREG:
        T1; zprintf("CreateEffect(): Device not registered!\n"); T2;
        return;
    acase DIERR_DEVICEFULL:
        T1; zprintf("CreateEffect(): Device is full!\n"); T2;
        return;
#endif
    adefault:
#ifdef FFVERBOSE
        T1; zprintf("CreateEffect(): Unknown error!\n"); T2;
#endif
        return;
    }

    acquire(joy);
    joySetForcesXY(joy, 5000, 5000); // half power
}

EXPORT void ff_on(int joy)
{   HRESULT hr;

    if (!ffeffect[joy] || !useff || rumbling[joy])
    {   return;
    }

    hr = ffeffect[joy]->lpVtbl->Start(ffeffect[joy], 1, 0);
    switch (hr)
    {
    case DI_OK:
        ;
#ifdef FFVERBOSE
        T1; zprintf("Started effect.\n"); T2;
    acase DIERR_INVALIDPARAM:
        T1; zprintf("Start(): Invalid parameter!\n"); T2;
        return;
    acase DIERR_INCOMPLETEEFFECT:
        T1; zprintf("Start(): Incomplete effect!\n"); T2;
        return;
    acase DIERR_NOTEXCLUSIVEACQUIRED:
        T1; zprintf("Start(): Not exclusively acquired!\n"); T2;
        return;
    acase DIERR_NOTINITIALIZED:
        T1; zprintf("Start(): Not initialized!\n"); T2;
        return;
    acase DIERR_UNSUPPORTED:
        T1; zprintf("Start(): Unsupported!\n"); T2;
        return;
#endif
    adefault:
#ifdef FFVERBOSE
        T1; zprintf("Start(): Unknown error!\n"); T2;
#endif
        return;
    }

    if (joy == 0)
    {   if (swapped)
        {   rumbling[0] = p2rumble;
        } else
        {   rumbling[0] = p1rumble;
    }   }
    else
    {   // assert(joy == 1);
        if (swapped)
        {   rumbling[1] = p1rumble;
        } else
        {   rumbling[1] = p2rumble;
}   }   }

EXPORT void ff_off(int joy)
{   if (rumbling[joy])
    {   rumbling[joy] = 0;
        DISCARD ffeffect[joy]->lpVtbl->Stop(ffeffect[joy]);
}   }

EXPORT void ff_decay(int joy)
{   HRESULT hr;

    if (!ffeffect[joy] || !rumblable[joy] || !rumbling[joy])
    {   return;
    }

    rumbling[joy]--;
    if (rumbling[joy])
    {   return;
    }

    hr = ffeffect[joy]->lpVtbl->Stop(ffeffect[joy]);
    switch (hr)
    {
    case DI_OK:
        ;
#ifdef FFVERBOSE
        T1; zprintf("Stopped effect.\n"); T2;
    acase DIERR_NOTEXCLUSIVEACQUIRED:
        T1; zprintf("Stop(): Not exclusively acquired!\n"); T2;
        return;
    acase DIERR_NOTINITIALIZED:
        T1; zprintf("Stop(): Not initialized!\n"); T2;
        return;
#endif
    adefault:
#ifdef FFVERBOSE
        T1; zprintf("Stop(): Unknown error!\n"); T2;
#endif
        return;
}   }

MODULE void joySetForcesXY(int joy, int x, int y) 
{   HRESULT         hr;
    LONG            rglDirection[2] = { x, y }; 
    DICONSTANTFORCE cf; 
    DIEFFECT        eff; 

    cf.lMagnitude = (DWORD) sqrt(  ((double) x * (double) x)
                                 + ((double) y * (double) y)
                                ); 
    eff.dwSize                = sizeof(DIEFFECT); 
    eff.dwFlags               = DIEFF_CARTESIAN | DIEFF_OBJECTOFFSETS; 
    eff.cAxes                 = 2; 
    eff.rglDirection          = rglDirection; 
    eff.lpEnvelope            = 0; 
    eff.cbTypeSpecificParams  = sizeof(DICONSTANTFORCE); 
    eff.lpvTypeSpecificParams = &cf; 
    hr = ffeffect[joy]->lpVtbl->SetParameters(ffeffect[joy],
        &eff,
        DIEP_DIRECTION | DIEP_TYPESPECIFICPARAMS
    );
    switch (hr)
    {
    case DI_OK:
        ;
#ifdef FFVERBOSE
        T1; zprintf("Set forces.\n"); T2;
#endif
    acase DI_EFFECTRESTARTED:
        ;
#ifdef FFVERBOSE
        T1; zprintf("SetParameters(): Effect restarted.\n"); T2;
#endif
    acase DI_DOWNLOADSKIPPED:
        ;
#ifdef FFVERBOSE
        T1; zprintf("SetParameters(): Download skipped.\n"); T2;
#endif
    acase DI_TRUNCATED:
        ;
#ifdef FFVERBOSE
        T1; zprintf("SetParameters(): Truncated.\n"); T2;
#endif
    acase DI_TRUNCATEDANDRESTARTED:
        ;
#ifdef FFVERBOSE
        T1; zprintf("SetParameters(): Truncated and restarted.\n"); T2;
    acase DIERR_INCOMPLETEEFFECT:
        T1; zprintf("SetParameters(): Incomplete effect!\n"); T2;
        return;
    acase DIERR_NOTINITIALIZED:
        T1; zprintf("SetParameters(): Not initialized!\n"); T2;
        return;
    acase DIERR_INVALIDPARAM:
        T1; zprintf("SetParameters(): Invalid parameter!\n"); T2;
        return;
    acase DIERR_EFFECTPLAYING:
        T1; zprintf("SetParameters(): Effect is already playing!\n"); T2;
        return;
#endif
    acase DIERR_INPUTLOST:
#ifdef FFVERBOSE
        T1; zprintf("SetParameters(): Input lost!\n"); T2;
#endif
        ; // acquire(joy);
        return;
    adefault:
#ifdef FFVERBOSE
        T1; zprintf("SetParameters(): Unknown error!\n"); T2;
#endif
        return; 
}   }

MODULE void acquire(int joy)
{   HRESULT hr;

    hr = ffjoy2[joy]->lpVtbl->Acquire(ffjoy2[joy]); 
    switch (hr)
    {
    case DI_OK:
        ;
#ifdef FFVERBOSE
        T1; zprintf("Acquired device.\n"); T2;
#endif
    acase S_FALSE:
        ;
#ifdef FFVERBOSE
        T1; zprintf("Acquire(): Already acquired device!\n"); T2;
    acase DIERR_INVALIDPARAM:
        T1; zprintf("Acquire(): Invalid parameter!\n"); T2;
        return;
    acase DIERR_NOTINITIALIZED:
        T1; zprintf("Acquire(): Not initialized!\n"); T2;
        return;
    acase DIERR_OTHERAPPHASPRIO:
        T1; zprintf("Acquire(): Another application has priority!\n"); T2;
        return;
#endif
    adefault:
#ifdef FFVERBOSE
        T1; zprintf("Acquire(): Unknown error!\n"); T2;
#endif
        return;
}   }

/* Some notes about the suitability of the current force feedback subsystem for
   certain games:
Arcadia:
 Auto Race would be good for -1 cars vs. background rumbling, except it causes
  rumbles on the title screen :-( .
 Crazy Gobbler would be better with background rumbles (feel eating of dots), except
  that the title screen would have continuous rumble (because there is in fact a dot
  at the player's starting position).
 Space Squadron (and therefore, presumably, Space Raiders) is unsuitable for this
  reason:
  (*) there are constant sprite collisions on the title screen (because it is
      putting all sprites at the same coordinate).
Interton:
 Hunting/Shooting Gallery has self-collision issues.
 Pinball: the non-playing player's gamepad is affected
  by what the playing player is doing.
 Olympics is unsuitable for background rumbling on some levels because of player
  paddle positioning.
*/
