#include <windows.h>
#include <mmreg.h>

typedef unsigned char       UBYTE;
typedef unsigned char       TEXT;
typedef unsigned short      UWORD;
typedef unsigned char*      STRPTR;
typedef void*               APTR;
#define IMPORT              extern
#define UNUSED
#define DebugPrintF         if (0) printf

#include <assert.h>

#define TITLEBARTEXT        "WinArcadia " SHORTVERSION
#define LOGFILENAME         "WA.LOG"

// One of these (only!) must be #defined.
// NEWWAY causes "CreateDIBSection() failed!" errors when eg.
//  depth-arranging.
#define OLDWAY
// #define NEWWAY

/* Enable this to have the Scroll Lock light mirror the state of player 1's primary firebutton
    (can cause emulator to miss keyboard input though).
#define FLASHSL

Enable this to replace the square waves with instruments.
It has a bug: sounds are terminating prematurely.
The problem is made worse at high speeds (eg. with turbo on) and better at
 low speeds (eg. with Ctrl+; held).
Also, it doesn't fall back properly into normal (ie. "multimedia" waveform)
 audio if it fails in its allocations (although it does at least clean up
 its resources properly).
Instruments are assumed to be 22050Hz, mono, 8-bit.
#define USEDIRECTSOUND
*/

#define SCAN_ESCAPE       1
#define SCAN_A1           2
#define SCAN_A2           3
#define SCAN_A3           4
#define SCAN_A4           5
#define SCAN_A5           6
#define SCAN_A6           7
#define SCAN_A7           8
#define SCAN_A8           9
#define SCAN_A9          10
#define SCAN_A0          11
#define SCAN_HYPHEN      12
#define SCAN_EQUALS      13
#define SCAN_BACKSPACE   14
#define SCAN_TAB         15
#define SCAN_Q           16
#define SCAN_W           17
#define SCAN_E           18
#define SCAN_R           19
#define SCAN_T           20
#define SCAN_Y           21
#define SCAN_U           22
#define SCAN_I           23
#define SCAN_O           24
#define SCAN_P           25
#define SCAN_OS          26
#define SCAN_CS          27
#define SCAN_RETURN      28
#define SCAN_AE          SCAN_RETURN
#define SCAN_LCTRL       29
#define SCAN_A           30
#define SCAN_S           31
#define SCAN_D           32
#define SCAN_F           33
#define SCAN_G           34
#define SCAN_H           35
#define SCAN_J           36
#define SCAN_K           37
#define SCAN_L           38
#define SCAN_SEMICOLON   39
#define SCAN_APOSTROPHE  40
#define SCAN_BACKTICK    41
#define SCAN_LSHIFT      42
#define SCAN_BACKSLASH   43
#define SCAN_Z           44
#define SCAN_X           45
#define SCAN_C           46
#define SCAN_V           47
#define SCAN_B           48
#define SCAN_N           49
#define SCAN_M           50
#define SCAN_COMMA       51
#define SCAN_FULLSTOP    52
#define SCAN_SLASH       53
#define SCAN_RSHIFT      54
#define SCAN_AS          55 // numeric *
#define SCAN_LALT        56
#define SCAN_SPACEBAR    57
#define SCAN_CAPSLOCK    58
#define SCAN_F1          59
#define SCAN_F2          60
#define SCAN_F3          61
#define SCAN_F4          62
#define SCAN_F5          63
#define SCAN_F6          64
#define SCAN_F7          65
#define SCAN_F8          66
#define SCAN_F9          67
#define SCAN_F10         68
#define SCAN_PAUSE       69
#define SCAN_SCRLK       70
#define SCAN_N7          71
#define SCAN_N8          72
#define SCAN_N9          73
#define SCAN_MI          74 // numeric -
#define SCAN_N4          75
#define SCAN_N5          76
#define SCAN_N6          77
#define SCAN_PL          78 // numeric +
#define SCAN_N1          79
#define SCAN_N2          80
#define SCAN_N3          81
#define SCAN_N0          82
#define SCAN_ND          83 // numeric .
#define SCAN_F11         87
#define SCAN_F12         88
#define SCAN_PREV      0x90
#define SCAN_NEXT      0x99
#define SCAN_MUTE      0xA0
#define SCAN_CALC      0xA1
#define SCAN_PLAY      0xA2
#define SCAN_STOP      0xA4
#define SCAN_VOLDOWN   0xAE
#define SCAN_VOLUP     0xB0
#define SCAN_WEB       0xB2
#define SCAN_FFWD      0xE9
#define SCAN_NE         284 // numeric ENTER
#define SCAN_RCTRL      285
#define SCAN_SL         309 // numeric /
#define SCAN_RALT       312
#define SCAN_NL         325 // NumLock
#define SCAN_UP         328
#define SCAN_LEFT       331
#define SCAN_RIGHT      333
#define SCAN_END        335
#define SCAN_DOWN       336
#define SCAN_PGDN       337
#define SCAN_DEL        339
#define UNASSIGNED      511 // unused scan code
#define SCANCODES       512

#define PRINTERDEVICE   "LPT1" // NB. no colon!
#define DIR_SCREENSHOTS ".\\Scrnshot\\"
#define DIR_PROJECTS    ".\\Projects\\"
#define PATTERN_OPEN    "All Signetics Files (AOF,BIN,COR,COS,PGM,TVC,ZIP)\0*.AOF;*.BIN;*.COR;*.COS;*.PGM;*.TVC;*.ZIP\0" \
                        "Recordings (COR,ZIP)\0*.COR;*.ZIP\0" \
                        "Snapshots (COS,ZIP)\0*.COS;*.ZIP\0" \
                        "Arcadia/Interton ROMs and PHUNSY Programs (BIN,ZIP)\0*.BIN;*.ZIP\0" \
                        "Programs (AOF,BIN,PGM,TVC,ZIP)\0*.AOF;*.BIN;*.PGM;*.TVC;*.ZIP\0" \
                        "Signetics Absolute Object Format (AOF,ZIP)\0*.AOF;*.ZIP\0" \
                        "ZIP Archives (ZIP)\0*.ZIP\0" \
                        "All Files (*.*)\0*.*\0"

                        // 0    IBM: black (invisible)
#define TEXTPEN_LOG        1 // IBM: red
#define TEXTPEN_NET        2 // IBM: green
#define TEXTPEN_VERBOSE    2 // IBM: green
#define TEXTPEN_DEBUG      2 // IBM: green
#define TEXTPEN_ERROR      2 // IBM: green
#define TEXTPEN_SWAPEPROMS 2 // IBM: green
#define TEXTPEN_VIEW       2 // IBM: green
#define TEXTPEN_TRACE      3 // IBM: orange
#define TEXTPEN_PRINTER    4 // IBM: (dark) blue
#define TEXTPEN_PIPBUG     5 // IBM: purple
#define TEXTPEN_CLIOUTPUT  6 // IBM: cyan
#define TEXTPEN_CLIINPUT   7 // IBM: grey
#define TEXTPEN_DEFAULT    8 // IBM: grey

                            // BBGGRR
#define EMUPEN_RED         0x000000FF
#define EMUPEN_BLUE        0x00FF8080
#define EMUPEN_GREEN       0x0000FF00
#define EMUPEN_YELLOW      0x0000FFFF
#define EMUPEN_ORANGE      0x000080FF
#define EMUPEN_PURPLE      0x00FF00FF
#define EMUPEN_GREY        0x00808080

#define SAMPLE_BELL        0
#define SAMPLES            1 // counting from 1

#ifdef USEDIRECTSOUND
    #include "dsound.h"

    #define INSTRUMENT_FLUTE    0
    #define INSTRUMENT_GUITAR   1
    #define INSTRUMENT_OBOE     2
    #define INSTRUMENT_ORGAN    3
    #define INSTRUMENT_PIANO    4
    #define INSTRUMENT_SAX      5
    #define INSTRUMENT_TROMBONE 6
    #define INSTRUMENT_VIOLIN   7
    #define INSTRUMENTS         8 // counting from 1
    #define INSTRUMENT_DEFAULT  INSTRUMENT_GUITAR
#endif

typedef struct bmiData
{
  BITMAPINFOHEADER	bmiHeader;
  RGBQUAD			bmiColors[256];
} bmiData;

const struct RearrangeStruct
{   UBYTE rearr[4];
    DWORD hGad[2];
};

#define USR_NOTIFYICON (WM_USER + 1)

// #define DEBUGSOUND

// always define one (only!) of these (normally GUI)
// #define CLI
#define GUI

#define TITLEBARTEXTLENGTH 31 // enough for eg. "WinInstructor xx.xxx beta x: "
#define PENS               64

#define LANGUAGE_ENG        0
#define LANGUAGE_HOL        1
#define LANGUAGE_FRA        2
#define LANGUAGE_GER        3
#define LANGUAGE_ITA        4
#define LANGUAGE_RUS        5
#define LANGUAGE_SPA        6
#define LANGUAGES           7 // counting from 1

#define MAKE_ID(a,b,c,d)	\
	((ULONG) (d)<<24 | (ULONG) (c)<<16 | (ULONG) (b)<<8 | (ULONG) (a))
// note that MAKE_ID() is defined differently for IBM-PC because of endianness issues!

#define ID_FORM             MAKE_ID('F', 'O', 'R', 'M')
#define ID_CTLG             MAKE_ID('C', 'T', 'L', 'G')
#define ID_STRS             MAKE_ID('S', 'T', 'R', 'S')

#define BIGGADGETS          7
#define MEMGADGETS        128

#define RECMODE_NORMAL   1000000000
#define RECMODE_PLAY     1000000001
#define RECMODE_RECORD   1000000002

#define PRI_HIGH            0
#define PRI_NORMAL          1
#define PRI_LOW             2

#define ICONS            (MACHINES + 2) // sidebar icons
