// INCLUDES---------------------------------------------------------------

#ifdef __VBCC__
    #define AMIGA
#endif
#ifdef AMIGA
    #include "amiga.h"
#endif
#ifdef _lint
    #define WIN32
    #define _inline
#endif
#ifdef WIN32
    #include "ibm.h"
    #define EXEC_TYPES_H
#endif

#include <stdio.h>

#include "aa.h"
#define CATCOMP_NUMBERS
#define CATCOMP_BLOCK
#include "aa_strings.h"
#ifdef AMIGA
    #include <proto/locale.h>  // GetCatalogStr()
#endif

#include "instruct.h"

// DEFINES----------------------------------------------------------------

// (none)

// EXPORTED VARIABLES-----------------------------------------------------

EXPORT UBYTE                sd_column[7],
                            s_glow        = 0,
                            s_toggles     = 0;
EXPORT int                  s_intdir      = INTDIR_DEFAULT,
                            s_intselector = INTSEL_DEFAULT,
                            s_io          = PARALLEL_DEFAULT;

// IMPORTED VARIABLES-----------------------------------------------------

IMPORT UBYTE                memory[32768],
                            OutputBuffer[7];
IMPORT ASCREEN              ctrlscreen[MINIWIDTH][MINIHEIGHT],
                            screen[BOXWIDTH][BOXHEIGHT];
IMPORT UWORD                mirror_r[32768],
                            mirror_w[32768];
IMPORT ULONG                frames,
                            memflags[32768];
IMPORT int                  console_start,
                            console_a,
                            console_b,
                            console_reset,
                            consoleopen,
                            drawcorners,
                            drawunlit,
                            game,
                            machine,
                            memmap,
                            inframe,
                            offset,
                            overcalc,
                            recmode,
                            shadowing,
                            slice,
                            zinterrupt;
IMPORT UBYTE                psu;
IMPORT UBYTE*               IOBuffer;
IMPORT UWORD                console[4],
                            iar,
                            keypads[2][NUMKEYS];
IMPORT ULONG                cycles,
                            elapsed;
IMPORT ULONG*               ctrldisplay;
IMPORT FILE*                MacroHandle;
IMPORT TEXT                 friendly[21 + 1];

#ifdef AMIGA
    IMPORT struct Catalog*  CatalogPtr;
    IMPORT LONG             pens[PENS];
    IMPORT UBYTE*           minibyteptr[MINIHEIGHT];
    IMPORT struct Window*   ControlsWindowPtr;
#endif
#ifdef WIN32
    IMPORT int              CatalogPtr; // APTR doesn't work
    IMPORT ULONG            pens[2][PENS];
#endif

// MODULE VARIABLES-------------------------------------------------------

MODULE UBYTE                s_position = 0,
                            s_segments = 0;

// MODULE FUNCTIONS-------------------------------------------------------

MODULE void s_emuinput(void);
MODULE __inline void local_drawpixel(int x, int y, int colour);

// CODE-------------------------------------------------------------------

MODULE void s_emuinput(void)
{   PERSIST UBYTE t; // PERSISTent for speed

    readkybd();

    if (recmode == RECMODE_PLAY)
    {   t = IOBuffer[offset++];
        sd_column[0] = (UBYTE) ((t & 0xF0) >> 4);
        sd_column[1] = (UBYTE)  (t & 0x0F)      ;
        t = IOBuffer[offset++];
        sd_column[2] = (UBYTE) ((t & 0xF0) >> 4);
        sd_column[3] = (UBYTE)  (t & 0x0F)      ;
        t = IOBuffer[offset++];
        sd_column[4] = (UBYTE) ((t & 0xF0) >> 4);
        sd_column[5] = (UBYTE)  (t & 0x0F)      ;
        t = IOBuffer[offset++];
        sd_column[6] = (UBYTE) ((t & 0xF0) >> 4);
        s_intdir      = (t & 8) >> 3; // 0/8 -> 0/1
        s_intselector = (t & 4) >> 2; // 0/4 -> 0/1
        s_io          = (t & 3);      // 0..2
        s_toggles = IOBuffer[offset++];
    } else
    {   // assert(recmode == RECMODE_NORMAL || recmode == RECMODE_RECORD);

        t =  0;
        t |= (UBYTE) (0x1 * KeyDown(console[0]));
        if (console_start)
        {   t |= 1;
        }
        t |= (UBYTE) (0x2 * KeyDown(console[1]));
        if (console_a)
        {   t |= 2;
        }
        t |= (UBYTE) (0x4 * KeyDown(console[2]));
        if (console_b)
        {   t |= 4;
        }
        t |= (UBYTE) (0x8 * KeyDown(console[3]));
        if (console_reset)
        {   t |= 8;
        }
        sd_column[0] = t;

        t =  0;
        t |= (UBYTE) (0x1 * KeyDown(keypads[0][1])); // "WCAS" key
        t |= (UBYTE) (0x2 * KeyDown(keypads[0][4])); // "RCAS" key
        t |= (UBYTE) (0x4 * KeyDown(keypads[0][7])); // "STEP" key
        t |= (UBYTE) (0x8 * KeyDown(keypads[0][10])); // "RUN" key
        sd_column[1] = t;

        t =  0;
        t |= (UBYTE) (0x1 * KeyDown(keypads[0][2])); // "BKPT" key
        t |= (UBYTE) (0x2 * KeyDown(keypads[0][5])); // "REG" key
        t |= (UBYTE) (0x4 * KeyDown(keypads[0][8])); // "MEM" key
        t |= (UBYTE) (0x8 * KeyDown(keypads[0][11])); // "ENT NXT" key
        sd_column[2] = t;

        t =  0;
        t |= (UBYTE) (0x8 * KeyDown(keypads[0][3])); // "C" key
        t |= (UBYTE) (0x4 * KeyDown(keypads[0][6])); // "8" key
        t |= (UBYTE) (0x2 * KeyDown(keypads[0][9])); // "4" key
        t |= (UBYTE) (0x1 * KeyDown(keypads[0][12])); // "0" key
        sd_column[3] = t;

        t =  0;
        t |= (UBYTE) (0x8 * KeyDown(keypads[1][1])); // "D" key
        t |= (UBYTE) (0x4 * KeyDown(keypads[1][4])); // "9" key
        t |= (UBYTE) (0x2 * KeyDown(keypads[1][7])); // "5" key
        t |= (UBYTE) (0x1 * KeyDown(keypads[1][10])); // "1" key
        sd_column[4] = t;

        t =  0;
        t |= (UBYTE) (0x8 * KeyDown(keypads[1][2])); // "E" key
        t |= (UBYTE) (0x4 * KeyDown(keypads[1][5])); // "A" key
        t |= (UBYTE) (0x2 * KeyDown(keypads[1][8])); // "6" key
        t |= (UBYTE) (0x1 * KeyDown(keypads[1][11])); // "2" key
        sd_column[5] = t;

        t =  0;
        t |= (UBYTE) (0x8 * KeyDown(keypads[1][3])); // "F" key
        t |= (UBYTE) (0x4 * KeyDown(keypads[1][6])); // "B" key
        t |= (UBYTE) (0x2 * KeyDown(keypads[1][9])); // "7" key
        t |= (UBYTE) (0x1 * KeyDown(keypads[1][12])); // "3" key
        sd_column[6] = t;

        domouse();
    }

    if (recmode == RECMODE_RECORD)
    {   // assert(MacroHandle);

        OutputBuffer[0] = ((sd_column[0] & 0x0F) << 4)
                        |  (sd_column[1] & 0x0F);
        OutputBuffer[1] = ((sd_column[2] & 0x0F) << 4)
                        |  (sd_column[3] & 0x0F);
        OutputBuffer[2] = ((sd_column[4] & 0x0F) << 4)
                        |  (sd_column[5] & 0x0F);
        OutputBuffer[3] = (UBYTE) (((sd_column[6] & 0x0F) << 4)
                        |  ((ULONG) s_intdir      << 3) // 0/1 -> 0/8
                        |  ((ULONG) s_intselector << 2) // 0/1 -> 0/4
                        |   s_io);                      // 0..2
        OutputBuffer[4] =   s_toggles;

/* We only support recording/playing back s_intdir/intselector/io/toggles
   once per frame, but in theory they could get changed by the user, and then
   read by the game, many times per frame. */

        DISCARD fwrite(OutputBuffer, 5, 1, MacroHandle);
        // should really check return code of fwrite()
        // 5 bytes per frame * 60 frames per second = 300 bytes/sec.
    }

    if (console_start) console_start--;
    if (console_a    ) console_a--;
    if (console_b    ) console_b--;
    if (console_reset) console_reset--;
}

EXPORT void s_drawcontrolspixel(int x, int y, int colour)
{   // assert(x >= 0);
    // assert(x < S_BOXWIDTH);
    // assert(y >= 0);
    // assert(y < S_BOXHEIGHT);
    // assert(colour >= 0);
    // assert(colour <= GUESTCOLOURS);

    if (machine != INSTRUCTOR)
    {   return; // important!
    }

    if (ctrlscreen[x][y] == colour)
    {   return;
    }
    ctrlscreen[x][y] = (ASCREEN) colour;

#ifdef WIN32
    ctrldisplay[x + (y * MINIWIDTH)] = pens[COLOURFUL][colour];
#endif
#ifdef AMIGA
    if (ControlsWindowPtr)
    {   *(minibyteptr[y] + x) = pens[colour];
    }
#endif
}

EXPORT void s_viewscreen(void)
{   // assert(machine == INSTRUCTOR);

    T1; textpen(TEXTPEN_VIEW);
    zprintf
    (   "$17D1..$17D8: %02X %02X %02X %02X %02X %02X %02X %02X %c%c%c%c%c%c%c%c\n\n",
        memory[0x17D1],
        memory[0x17D2],
        memory[0x17D3],
        memory[0x17D4],
        memory[0x17D5],
        memory[0x17D6],
        memory[0x17D7],
        memory[0x17D8],
        instructor_chars[memory[0x17D1]],
        instructor_chars[memory[0x17D2]],
        instructor_chars[memory[0x17D3]],
        instructor_chars[memory[0x17D4]],
        instructor_chars[memory[0x17D5]],
        instructor_chars[memory[0x17D6]],
        instructor_chars[memory[0x17D7]],
        instructor_chars[memory[0x17D8]]
    );
    T2;
}

EXPORT void s_setmemmap(void)
{   int i, address, mirror;

    game = FALSE;

    for (i = 0x0; i <= 0x7FFF; i++)
    {   memory[i] = 0;
    }
    for (i = 0x0; i <= 0x7FF; i++) // 2K
    {   memory[0x1800 + i] = s_bios[i];
    }

    for (address = 0; address < 32 * KILOBYTE; address++)
    {   memflags[address] = 0;
        mirror            = address;

        if (memmap == MEMMAP_N)
        {   if
            (             //    $0..               $1FF:    user RAM
                (address >=  0x200 && address <=  0xFFE) // unused
                          //  $FFF:                         I/O port
             || (address >= 0x1000 && address <= 0x177F) // unused
                          // $1780..              $17FF:    user RAM, monitor RAM
             || (address >= 0x1800 && address <= 0x7FFF) // monitor ROM, unused
            )
            {   memflags[address] |= NOWRITE;
            }
            if
            (             //    $0..               $1FF:    user RAM
                (address >=  0x200 && address <=  0xFFE) // unused
                          //  $FFF:                         I/O port
             || (address >= 0x1000 && address <= 0x177F) // unused
                          // $1780..              $1FFF:    user RAM, monitor RAM, monitor ROM
             || (address >= 0x2000 && address <= 0x7FFF) // unused
            )
            {   memflags[address] |= NOREAD;
        }   }
        else
        {   // assert(memmap == MEMMAP_O);
            if
            (             //    $0..               $FFF:    user RAM, expansion RAM, I/O port
                (address >= 0x1000 && address <= 0x177F) // unused
                          // $1780..              $17FF:    user RAM, monitor RAM
             || (address >= 0x1800 && address <= 0x1FFF) // monitor ROM
                          // $2000..              $7FFF:    expansion RAM
            )
            {   memflags[address] |= NOWRITE;
            }
            if
            (             //    $0..               $FFF:    user RAM, expansion RAM, I/O port
                (address >= 0x1000 && address <= 0x177F) // unused
                          // $1780..              $7FFF:    user RAM, monitor RAM, monitor ROM, expansion RAM
            )
            {   memflags[address] |= NOREAD;
        }   }

        if (address >= 0x17D1 && address <= 0x17D8)
        {   memflags[address] |= VISIBLE;
        }

        if (mirror != address)
        {   memflags[address] |= MIRRORED_R | MIRRORED_W;
        }
        mirror_r[address] =
        mirror_w[address] = (UWORD) mirror;
}   }

EXPORT void s_emulate(void)
{   PERSIST UBYTE old_column0; // PERSISTent for speed
    PERSIST int   i, j;        // PERSISTent for speed

    inframe = TRUE;

    // this is to emulate the non-persistence of the segment display
    if (drawunlit)
    {   for (i = 0; i < 8; i++)
        {   for (j = 0; j < 7; j++)
            {   drawsegment(i, j, UNLIT);
    }   }   }
    else
    {   for (i = 0; i < 8; i++)
        {   for (j = 0; j < 7; j++)
            {   drawsegment(i, j, BLACK);
    }   }   }
    if (drawunlit && drawcorners)
    {   for (i = 0; i < 8; i++)
        {   for (j = 7; j <= 12; j++)
            {   drawsegment(i, j, UNLIT);
    }   }   }
    else
    {   for (i = 0; i < 8; i++)
        {   for (j = 7; j <= 12; j++)
            {   drawsegment(i, j, BLACK);
    }   }   }

    // 895KHz = 895,000 cycles per second?
    // 895,000 / 60 = 14,916.6'
    // slice = 14917 - overcalc; // really 14,916.6'
    // 895,000 / 50 = 17,900
    slice = 17900 - overcalc;
    cpu();

    old_column0 = sd_column[0];

    s_emuinput();

    if ((sd_column[0] & 1) && !(old_column0 & 1))
    {   if (psu & PSU_S)
        {   psu &= ~(PSU_S);
        } else
        {   psu |= PSU_S;
    }   }
    if ((sd_column[0] & 2) && !(old_column0 & 2))
    {   if (s_intselector == INTSEL_KYBD && !(psu & PSU_II))
        {   zinterrupt = TRUE;
    }   }
    if ((sd_column[0] & 4) && !(old_column0 & 4))
    {   iar  = 0x1800;
        // "jump to BIOS" command also calls macro_stop(), but we (deliberately) don't.
    }
    if ((sd_column[0] & 8) && !(old_column0 & 8))
    {   iar  = 0;
        // "reset to game" command also calls macro_stop(), but we (deliberately) don't.
    }

    if (s_intselector == INTSEL_ACLINE && !(psu & PSU_II))
    {   zinterrupt = TRUE;
    }

    if (recmode == RECMODE_PLAY)
    {   updatedips();
    }
    endofframe();
}

EXPORT void smiwrite(UWORD address, UBYTE data)
{   // assert(address < 32768);

    /* Writes to memory from the SMI always go through this
    routine (so that watchpoints work properly).
      No mirroring is done here, the SMI module is expected to use the
    resolved address. |*/

    if ((memflags[address] & WATCHPOINT) && data != memory[address])
    {   DISCARD getfriendly(address);
        T1; textpen(TEXTPEN_DEBUG);
        zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_SMI_HITWP,
                "SMI is writing $%X to %s at clock %ld!\n\n"
            ),
                       data,
            (char *)   friendly,
            (long int) (elapsed + cycles)
        );
        T2;
        runto();
    }

    memory[address] = data;
}

EXPORT UBYTE s_readport(int port)
{   TRANSIENT FLAG  handled = TRUE;
    PERSIST   UBYTE t; // PERSISTent for speed

    // assert(machine == INSTRUCTOR);

    switch (port)
    {
    case 0x07:
        if (s_io == PARALLEL_EXTENDED)
        {   t = s_toggles;
        }
    acase 0xFE:
        if (frames % 2)
        {   switch (s_position)
            {
            case 0: // they want the 4th column
                t = (UBYTE) (sd_column[3] ^ 0xFF);
            acase 1: // they want the 5th column
                t = (UBYTE) (sd_column[4] ^ 0xFF);
            acase 2: // they want the 6th column
                t = (UBYTE) (sd_column[5] ^ 0xFF);
            acase 3: // they want the 7th column
                t = (UBYTE) (sd_column[6] ^ 0xFF);
            acase 4: // they want the 2nd column
                t = (UBYTE) (sd_column[1] ^ 0xFF);
            acase 5: // they want the 3rd column
                t = (UBYTE) (sd_column[2] ^ 0xFF);
            adefault: // don't know
                t = 0xFF; // t = (UBYTE) (sd_column[0] ^ 0xFF);
        }   }
        else
        {   t = 0xFF;
        }
    acase PORTD:
        if (s_io == PARALLEL_NONEXTENDED)
        {   t = s_toggles;
        }
    adefault:
        handled = FALSE;
        t = 0;
    }

    logport(port, t, handled, FALSE);
    return t;
}

EXPORT void s_writeport(int port, UBYTE data)
{   TRANSIENT FLAG handled   = TRUE;
    TRANSIENT int  i;
    PERSIST   int  zeroed[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };

    // ioport[port] = data;
    switch (port)
    {
    case 0x07:
        if (s_io == PARALLEL_EXTENDED)
        {   s_glow = data;
        }
    acase 0xF9:
        if (s_position == 8)
        {   return; // this is maybe supposed to clear the entire display or something? (which we haven't implemented)
        }

        s_segments = data;
        if
        (   s_segments == 0
         && (   (s_position == 0 && zeroed[s_position] < 2) // for some reason the USE BIOS likes to write zeroes *twice* to this position
             || zeroed[s_position] < 1 // and only *once* to all the other positions
        )   )
        {   zeroed[s_position]++;
            return;
        }
        zeroed[s_position] = 0;

        for (i = 0; i < 13; i++)
        {   drawsegment(s_position, i, BLACK);
        }

        if (drawcorners)
        {   if (s_segments & 0x21) // top left
            {   drawsegment(s_position, 7, RED);
            } elif (drawunlit)
            {   drawsegment(s_position, 7, UNLIT);
            }
            if (s_segments & 0x03) // top right
            {   drawsegment(s_position, 8, RED);
            } elif (drawunlit)
            {   drawsegment(s_position, 8, UNLIT);
            }
            if (s_segments & 0x70) // middle left
            {   drawsegment(s_position, 9, RED);
            } elif (drawunlit)
            {   drawsegment(s_position, 9, UNLIT);
            }
            if (s_segments & 0x46) // middle right
            {   drawsegment(s_position, 10, RED);
            } elif (drawunlit)
            {   drawsegment(s_position, 10, UNLIT);
            }
            if (s_segments & 0x18) // bottom left
            {   drawsegment(s_position, 11, RED);
            } elif (drawunlit)
            {   drawsegment(s_position, 11, UNLIT);
            }
            if (s_segments & 0x0C) // bottom right
            {   drawsegment(s_position, 12, RED);
            } elif (drawunlit)
            {   drawsegment(s_position, 12, UNLIT);
        }   }

        for (i = 0; i < 7; i++)
        {   if (s_segments & (1 << i))
            {   drawsegment(s_position, i, RED);
            } elif (drawunlit)
            {   drawsegment(s_position, i, UNLIT);
        }   }
    acase 0xFA:
        switch (data)
        {
        case 0x80:
            s_position = 7;
        acase 0x40:
            s_position = 6;
        acase 0x20:
            s_position = 5;
        acase 0x10:
            s_position = 4;
        acase 0x08:
            s_position = 3;
        acase 0x04:
            s_position = 2;
        acase 0x02:
            s_position = 1;
        acase 0x01:
            s_position = 0;
        adefault: // eg. $00 (which does happen)
            s_position = 8;
        }
    acase PORTD:
        if (s_io == PARALLEL_NONEXTENDED)
        {   s_glow = data;
        }
    adefault: // eg. $F8
        handled = FALSE;
    }

    logport(port, data, handled, TRUE);
}

EXPORT void s_drawglows(void)
{   int colour,
        i,
        shadowcolour;

    for (i = 0; i < 8; i++)
    {   if (s_glow & (128 >> i))
        {   colour       = RED;
        } elif (drawunlit)
        {   colour       = DARKBLUE;
        } else
        {   colour       = BLACK;
        }
        if (shadowing && colour != BLACK)
        {   shadowcolour = GREY1;
        } else
        {   shadowcolour = BLACK;
        }

        local_drawpixel(20 + (i * 21), 61, shadowcolour);
        local_drawpixel(20 + (i * 21), 62, shadowcolour);
        local_drawpixel(20 + (i * 21), 63, shadowcolour);
        local_drawpixel(19 + (i * 21), 64, shadowcolour);
        local_drawpixel(20 + (i * 21), 64, shadowcolour);
        local_drawpixel(16 + (i * 21), 65, shadowcolour);
        local_drawpixel(17 + (i * 21), 65, shadowcolour);
        local_drawpixel(18 + (i * 21), 65, shadowcolour);
        local_drawpixel(19 + (i * 21), 65, shadowcolour);

        local_drawpixel(15 + (i * 21), 59, colour);
        local_drawpixel(16 + (i * 21), 59, colour);
        local_drawpixel(17 + (i * 21), 59, colour);
        local_drawpixel(18 + (i * 21), 59, colour);
        local_drawpixel(14 + (i * 21), 60, colour);
        local_drawpixel(15 + (i * 21), 60, colour);
        local_drawpixel(16 + (i * 21), 60, colour);
        local_drawpixel(17 + (i * 21), 60, colour);
        local_drawpixel(18 + (i * 21), 60, colour);
        local_drawpixel(19 + (i * 21), 60, colour);
        local_drawpixel(14 + (i * 21), 61, colour);
        local_drawpixel(15 + (i * 21), 61, colour);
        local_drawpixel(16 + (i * 21), 61, colour);
        local_drawpixel(17 + (i * 21), 61, colour);
        local_drawpixel(18 + (i * 21), 61, colour);
        local_drawpixel(19 + (i * 21), 61, colour);
        local_drawpixel(14 + (i * 21), 62, colour);
        local_drawpixel(15 + (i * 21), 62, colour);
        local_drawpixel(16 + (i * 21), 62, colour);
        local_drawpixel(17 + (i * 21), 62, colour);
        local_drawpixel(18 + (i * 21), 62, colour);
        local_drawpixel(19 + (i * 21), 62, colour);
        local_drawpixel(14 + (i * 21), 63, colour);
        local_drawpixel(15 + (i * 21), 63, colour);
        local_drawpixel(16 + (i * 21), 63, colour);
        local_drawpixel(17 + (i * 21), 63, colour);
        local_drawpixel(18 + (i * 21), 63, colour);
        local_drawpixel(19 + (i * 21), 63, colour);
        local_drawpixel(15 + (i * 21), 64, colour);
        local_drawpixel(16 + (i * 21), 64, colour);
        local_drawpixel(17 + (i * 21), 64, colour);
        local_drawpixel(18 + (i * 21), 64, colour);
}   }

MODULE __inline void local_drawpixel(int x, int y, int colour)
{   if (screen[x][y] != (UBYTE) colour)
    {   screen[x][y] = (UBYTE) colour;
        drawpixel(x, y, colour);
}   }
