/* To write a tile:
 we do a WRTE to I/O port $60 to set malzak_x
  (the X-coordinate of where to put the tile)
 we do a STRA $aa to $16y0
  where y is the Y-coordinate of where to put the tile
  (ie. the Y-coordinate is chosen by the high nybble of the low byte of the address)
  and $aa is the tile contents to write (modulo 32)

INCLUDES--------------------------------------------------------------- */

#ifdef __VBCC__
    #define AMIGA
#endif
#ifdef AMIGA
    #include "amiga.h"
#endif
#ifdef _lint
    #define WIN32
    #define _inline
#endif
#ifdef WIN32
    #include "ibm.h"
#endif

#include <stdio.h> // for FILE*

#include "aa.h"
#include "malzak.h"

// DEFINES----------------------------------------------------------------

// (none)

// EXPORTED VARIABLES-----------------------------------------------------

EXPORT int    offsetsprites = 0,
              malzak_x,
              malzak_y,
              m_bank1, // text data bank
              m_bank2; // terrain data bank
EXPORT UBYTE  m_field[16][16],
              mg_console;

// IMPORTED VARIABLES-----------------------------------------------------

IMPORT FLAG                 lmb,
                            mmb;
IMPORT UBYTE                bgcoll,
                            g_paddles,
                            memory[32768],
                            OutputBuffer[7],
                            sx[2],
                            sy[2];
IMPORT UWORD                colltable[PVI_RASTLINES][PVI_XSIZE],
                            console[4],
                            jf[2],
                            keypads[2][NUMKEYS],
                            mirror_r[32768],
                            mirror_w[32768];
IMPORT ULONG                analog,
                            autofire,
                            downframes,
                            frames,
                            memflags[32768],
                            swapped,
                            totalframes;
IMPORT int                  ax[2],
                            ay[2],
                            collx,
                            colly,
                            console_start,
                            console_a,
                            console_b,
                            console_reset,
                            consoleopen,
                            cpl,
                            firstmnemonic,
                            latestsprite,
                            game,
                            hostcontroller[2],
                            key1,
                            key2,
                            key3,
                            key4,
                            lastmnemonic,
                            machine,
                            memmap,
                            offset,
                            p1bgcol[4],
                            p2bgcol[4],
                            p1rumble,
                            p2rumble,
                            p1sprcol[6],
                            p2sprcol[6],
                            pvibase,
                            recmode,
                            requirebutton,
                            starfield,
                            vblank,
                            whichoverlay;
IMPORT struct MachineStruct machines[MACHINES];
IMPORT UBYTE*               IOBuffer;
IMPORT FILE*                MacroHandle;
IMPORT ASCREEN              screen[BOXWIDTH][BOXHEIGHT];

// MODULE VARIABLES-------------------------------------------------------

MODULE UBYTE                saa5050_forecol,
                            saa5050_backcol;

// MODULE FUNCTIONS-------------------------------------------------------

MODULE void m_drawtext(UBYTE code, int x, int y, UWORD flags);
MODULE void m_drawtile(UBYTE code, int x, int y);
MODULE __inline void local_drawpixel(int x, int y, int colour);

// CODE-------------------------------------------------------------------

MODULE void m_drawtext(UBYTE code, int x, int y, UWORD flags)
{   PERSIST int xx, yy; // PERSISTent for speed

    if (flags)
    {   for (yy = 0; yy < 10; yy++)
        {   for (xx = 0; xx < 6; xx++)
            {   if (m_chars[((code - 32) * 10) + yy] & (0x20 >> xx))
                {   local_drawpixel(x + xx, y + (yy * 2)    , from_m[saa5050_forecol]);
                    local_drawpixel(x + xx, y + (yy * 2) + 1, from_m[saa5050_forecol]);
                    /* if (saa5050_forecol != BLACK)
                    {   colltable[y + (yy * 2)    ][x + xx] |= 8;
                        colltable[y + (yy * 2) + 1][x + xx] |= 8;
                    } */
                } else
                {   local_drawpixel(x + xx, y + (yy * 2)    , from_m[saa5050_backcol]);
                    local_drawpixel(x + xx, y + (yy * 2) + 1, from_m[saa5050_backcol]);
                    /* if (saa5050_backcol != BLACK)
                    {   colltable[y + (yy * 2)    ][x + xx] |= 8;
                        colltable[y + (yy * 2) + 1][x + xx] |= 8;
                    } */
    }   }   }   }
    else
    {   for (yy = 0; yy < 10; yy++)
        {   for (xx = 0; xx < 6; xx++)
            {   if (m_chars[((code - 32) * 10) + yy] & (0x20 >> xx))
                {   local_drawpixel(x + xx, y + yy, from_m[saa5050_forecol]);
                    /* if (saa5050_forecol != BLACK)
                    {   colltable[y + yy][x + xx] |= 8;
                    } */
                } else
                {   local_drawpixel(x + xx, y + yy, from_m[saa5050_backcol]);
                    /* if (saa5050_backcol != BLACK)
                    {   colltable[y + yy][x + xx] |= 8;
                    } */
}   }   }   }   }

MODULE void m_drawtile(UBYTE code, int x, int y)
{   PERSIST int   inverted,
                  xx, yy; // all PERSISTent for speed
    PERSIST UBYTE colour;

/* The imagery is laid out in memory as follows for each 16*16 tile:
(0,0), (1,0),
(0,1), (1,1).

These were the (arbitrary) colours under V14.03 and earlier:
$00..$1F: YELLOW
$20..$3F: RED
$40..$5F: BLUE
$60..$7F: GREEN
$80..$FF: CYAN
At slow speed: Red  ($20..$3F) enemies and green ($60..$7F) land.
At fast speed: Cyan ($80..$FF) enemies and red   ($20..$3F) land.

Under V14.04+ (still arbitrary but works better):
$00..$1F: YELLOW
$20..$2F: RED
$30..$3B: GREEN
$3C..$3F: RED
$40..$5F: BLUE
$60..$7F: GREEN
$80..$FF: RED */

    switch (code)
    {
    case 0x00:
    case 0x40:
    case 0x60:
    case 0x80:
    case 0xA0:
    case 0xC0:
    case 0xE0:
        return; // for speed
    case 0x01: // -lint fallthrough
    case 0x02:
    case 0x03:
    case 0x04:
    case 0x05:
    case 0x06:
    case 0x07:
    case 0x08:
    case 0x09:
    case 0x0A:
    case 0x0B:
    case 0x0C:
    case 0x0D:
    case 0x0E:
    case 0x0F:
    case 0x10:
    case 0x11:
    case 0x12:
    case 0x13:
    case 0x14:
    case 0x15:
    case 0x16:
    case 0x17:
    case 0x18:
    case 0x19:
    case 0x1A:
    case 0x1B:
    case 0x1C:
    case 0x1D:
    case 0x1E:
    case 0x1F:
        colour = YELLOW;
        inverted = FALSE;
    acase 0x20:
    case 0x25:
        colour = GREEN;
        inverted = TRUE;
        code &= 0x1F;
    acase 0x21:
    case 0x22:
    case 0x23:
    case 0x24:
    case 0x26:
    case 0x27:
    case 0x28:
    case 0x29:
    case 0x2A:
    case 0x2B:
    case 0x2C:
    case 0x2D:
    case 0x2E:
    case 0x2F:
    case 0x30:
    case 0x31:
    case 0x32:
    case 0x33:
    case 0x34:
    case 0x35:
    case 0x36:
    case 0x37:
    case 0x38:
    case 0x39:
    case 0x3A:
    case 0x3B:
        colour = GREEN;
        inverted = FALSE;
        code &= 0x1F;
    acase 0x3C:
    case 0x3D:
    case 0x3E:
    case 0x3F:
        colour = RED; // 2nd colour
        inverted = FALSE;
        code &= 0x1F;
    acase 0x41:
    case 0x42:
    case 0x43:
    case 0x44:
    case 0x45:
    case 0x46:
    case 0x47:
    case 0x48:
    case 0x49:
    case 0x4A:
    case 0x4B:
    case 0x4C:
    case 0x4D:
    case 0x4E:
    case 0x4F:
    case 0x50:
    case 0x51:
    case 0x52:
    case 0x53:
    case 0x54:
    case 0x55:
    case 0x56:
    case 0x57:
    case 0x58:
    case 0x59:
    case 0x5A:
    case 0x5B:
    case 0x5C:
    case 0x5D:
    case 0x5E:
    case 0x5F:
        colour = BLUE;
        inverted = FALSE;
        code &= 0x1F;
    acase 0x61:
    case 0x62:
    case 0x63:
    case 0x64:
    case 0x65:
    case 0x66:
    case 0x67:
    case 0x68:
    case 0x69:
    case 0x6A:
    case 0x6B:
    case 0x6C:
    case 0x6D:
    case 0x6E:
    case 0x6F:
    case 0x70:
    case 0x71:
    case 0x72:
    case 0x73:
    case 0x74:
    case 0x75:
    case 0x76:
    case 0x77:
    case 0x78:
    case 0x79:
    case 0x7A:
    case 0x7B:
    case 0x7C:
    case 0x7D:
    case 0x7E:
    case 0x7F:
        colour = GREEN; // 1st colour
        inverted = FALSE;
        code &= 0x1F;
    adefault:
        // assert(code >= 0x81);
        // assert(code <= 0xFF);
        colour = RED;
        if (code >= 0xEC && code <= 0xEE)
        {   // also $EF, but that causes problems with
            // Malzak glyph on title screen
            code &= 0x3F;
            inverted = FALSE;
        } else
        {   if (code == 0xC5)
            {   inverted = TRUE;
            } else
            {   inverted = FALSE;
            }
            code &= 0x1F;
    }   }

    if (inverted)
    {   for (yy = 0; yy < 16; yy++)
        {   for (xx = 0; xx < 8; xx++)
            {   if
                (   x + xx >= 0
                 && x + xx <  machines[machine].width
                 && y + yy >= 0
                 && y + yy <  machines[machine].height
                )
                {   if (!(m_tiles[(code * 32) + (yy * 2)] & (1 << xx)))
                    {   local_drawpixel(x + xx, y + yy, colour);
                        colltable[y + yy][x + xx + HIDDEN_X] |= 8;
            }   }   }

            for (xx = 0; xx < 8; xx++)
            {   if
                (   x + xx + 8 >= 0
                 && x + xx + 8 <  machines[machine].width
                 && y + yy     >= 0
                 && y + yy     <  machines[machine].height
                )
                {   if (!(m_tiles[(code * 32) + (yy * 2) + 1] & (1 << xx)))
                    {   local_drawpixel(x + xx + 8, y + yy, colour);
                        colltable[y + yy][x + xx + 8 + HIDDEN_X] |= 8;
    }   }   }   }   }
    else
    {   for (yy = 0; yy < 16; yy++)
        {   for (xx = 0; xx < 8; xx++)
            {   if
                (   x + xx >= 0
                 && x + xx <  machines[machine].width
                 && y + yy >= 0
                 && y + yy <  machines[machine].height
                )
                {   if (m_tiles[(code * 32) + (yy * 2)] & (1 << xx))
                    {   local_drawpixel(x + xx, y + yy, colour);
                        colltable[y + yy][x + xx + HIDDEN_X] |= 8;
            }   }   }

            for (xx = 0; xx < 8; xx++)
            {   if
                (   x + xx + 8 >= 0
                 && x + xx + 8 <  machines[machine].width
                 && y + yy     >= 0
                 && y + yy     <  machines[machine].height
                )
                {   if (m_tiles[(code * 32) + (yy * 2) + 1] & (1 << xx))
                    {   local_drawpixel(x + xx + 8, y + yy, colour);
                        colltable[y + yy][x + xx + 8 + HIDDEN_X] |= 8;
}   }   }   }   }   }

EXPORT void m_viewscreen(void)
{   int where;

    // Ideally, we would show the text in the correct colours

    T1; textpen(TEXTPEN_VIEW);
    for (where = 0x1800; where < 0x1E80; where += 16)
    {   dumprow(where);
        if (where % 64 == 48)
        {   zprintf("\n");
    }   }
    zprintf("\n");
    T2;
}

EXPORT void switchbank(int whichbank)
{   int i;

    switch (whichbank)
    {
    case 1:
        for (i = 0; i <= 0x3FF; i++)
        {   memory[0xC00 + i]  = bank1[m_bank1][i];
        }
    acase 2:
        for (i = 0; i <= 0x5FF; i++) // $4400..$49FF
        {   memory[0x4400 + i] = bank2[m_bank2][i];
}   }   }

EXPORT void m_setmemmap(void)
{   int   i, address, mirror;
    FILE* TheLocalHandle; // LocalHandle is a variable of winbase.h

    game = FALSE;
    // assert(memmap == MEMMAP_MALZAK1 || memmap == MEMMAP_MALZAK2);

    for (i = 0x1000; i <= 0x7FFF; i++)
    {   memory[i] = 0; // important
    }

    if (memmap == MEMMAP_MALZAK1)
    {   for (i = 0; i <= 0xFFF; i++)
        {   memory[i] = m1_bios[i];                   // $0000..$0FFF -> $0000..$0FFF (4K)
        }
        for (i = 0; i <= 0x7FF; i++)
        {   memory[0x2000 + i] = m1_bios[0x1000 + i]; // $1000..$17FF -> $2000..$27FF (2K)
        }
        for (i = 0; i <= 0x9FF; i++)
        {   memory[0x4000 + i] = m1_bios[0x1800 + i]; // $1800..$21FF -> $4000..$49FF (2.5K)
        }

        whichoverlay  = M1_OVERLAY;
    } elif (memmap == MEMMAP_MALZAK2)
    {   for (i = 0; i <= 0xBFF; i++)
        {   memory[         i] = m2_bios[         i]; // $0000..$0BFF -> $0000..$0BFF (3K)
        }
        // $C00..$FFF (1K) is banked (so 1K*2)
        for (i = 0; i <= 0xBFF; i++)
        {   memory[0x2000 + i] = m2_bios[0xC00  + i]; // $0C00..$17FF -> $2000..$2BFF (3K)
        }
        for (i = 0; i <= 0x3FF; i++)
        {   memory[0x4000 + i] = m2_bios[0x1800 + i]; // $1800..$1BFF -> $4000..$43FF (1K)
        }
        // $4400..$49FF (1.5K) is probably banked (so 1.5K*2)
        for (i = 0; i <= 0x3FF; i++)
        {   memory[0x6000 + i] = m2_bios[0x1C00 + i]; // $1C00..$1FFF -> $6000..$63FF (1K)
        }
        m_bank1 = 0;
        m_bank2 = 1;
        switchbank(1);
        switchbank(2);

#ifdef AMIGA
        if ((TheLocalHandle = fopen("PROGDIR:MALZAK2.NVR", "rb")))
#endif
#ifdef WIN32
        if ((TheLocalHandle = fopen("MALZAK2.NVR", "rb")))
#endif

        {   DISCARD fread(&memory[0x1700], 256, 1, TheLocalHandle);
            DISCARD fread(&memory[0x14CC],   1, 1, TheLocalHandle); // potentiometer
            DISCARD fclose(TheLocalHandle);
            // TheLocalHandle = NULL;
#ifdef VERBOSE
        T1; textpen(TEXTPEN_VERBOSE);
        zprintf("Loaded NVRAM.\n"); T2;
#endif
        } else
        {   for (i = 0; i <= 255; i++)
            {   memory[0x1700 + i] = default_nvram[i];
            }
            memory[0x14CC] = default_nvram[256];
        }

        whichoverlay  = M2_OVERLAY;
    }

    for (address = 0; address < 32 * KILOBYTE; address++)
    {   memflags[address] = 0;
        mirror            = address;

        if
        (   (address <= 0x0FFF)                      // ROM
         || (address >= 0x2000 && address <= 0x2FFF) // ROM
         || (address >= 0x4000 && address <= 0x4FFF) // ROM
         || (address >= 0x6000 && address <= 0x6FFF) // ROM
        )
        {   memflags[address] |= NOWRITE;
        }

        if
        (
        // legal:        $1400..$142D                   hardware registers, RAM, h/w regs, RAM, h/w regs
            (address >= 0x142E && address <= 0x143F) // unmapped
        // legal:        $1440..$146D                   hardware registers, RAM
         || (address >= 0x146E && address <= 0x147F) // unmapped
        // legal:        $1480..$14AD                   grid, RAM
         || (address >= 0x14AE && address <= 0x14BF) // unmapped
        // legal:        $14C0..$14C3                   hardware registers
         || (address >= 0x14C4 && address <= 0x14C5) // unused
        // legal:        $14C6..$14C9                   hardware registers
         || (address >= 0x14CA && address <= 0x14FF) // R/- hardware registers, unused, mirrors (not occurring), semi-mirrors
        // legal:        $1500..$152D                   hardware registers, RAM, h/w regs, RAM, h/w regs
         || (address >= 0x152E && address <= 0x153F) // unmapped
        // legal:        $1540..$156D                   hardware registers, RAM
         || (address >= 0x156E && address <= 0x157F) // unmapped
        // legal:        $1580..$15AD                   grid, RAM
         || (address >= 0x15AE && address <= 0x15BF) // unmapped
        // legal:        $15C0..$15C3                   hardware registers
         || (address >= 0x15C4 && address <= 0x15C5) // unused
        // legal:        $15C6..$15C9                   hardware registers
         || (address >= 0x15CA && address <= 0x15FF) // R/- hardware registers, unused, mirrors (not occurring), semi-mirrors
        )
        {   memflags[address] |= NOWRITE;
        }

        if
        (
    // legal:            $1400              ..$142D     hardware registers, RAM, h/w regs, RAM, h/w regs
            (address >= 0x142E && address <= 0x143F) // unmapped
    // legal:            $1440              ..$146D     hardware registers, RAM
         || (address >= 0x146E && address <= 0x147F) // unmapped
    // legal:            $1480              ..$14AD     grid, RAM
         || (address >= 0x14AE && address <= 0x14C9) // unmapped, -/W (write-only) and -/- (unusable) hardware registers
    // legal:            $14CA              ..$14CD     R/- (read-only) hardware registers
         || (address >= 0x14CE && address <= 0x14D9) // unused, mirrors (not occurring)
    // legal:            $14DA              ..$14DB     semi-mirrors
         || (address >= 0x14DC && address <= 0x14E9) // mirrors (not occurring)
    // legal:            $14EA              ..$14EB     semi-mirrors
         || (address >= 0x14EC && address <= 0x14F9) // mirrors (not occurring)
    // legal:            $14FA              ..$14FB     semi-mirrors
         || (address >= 0x14FC && address <= 0x14FF) // mirrors (not occurring)
    // legal:            $1500              ..$152D     hardware registers, RAM, h/w regs, RAM, h/w regs
         || (address >= 0x152E && address <= 0x153F) // unmapped
    // legal:            $1540              ..$156D     hardware registers, RAM
         || (address >= 0x156E && address <= 0x157F) // unmapped
    // legal:            $1580              ..$15AD     grid, RAM
         || (address >= 0x15AE && address <= 0x15C9) // unmapped, -/W (write-only) and -/- (unusable) hardware registers
    // legal:            $15CA              ..$15CD     R/- (read-only) hardware registers
         || (address >= 0x15CE && address <= 0x15D9) // unused, mirrors (not occurring)
    // legal:            $15DA              ..$15DB     semi-mirrors
         || (address >= 0x15DC && address <= 0x15E9) // mirrors (not occurring)
    // legal:            $15EA              ..$15EB     semi-mirrors
         || (address >= 0x15EC && address <= 0x15F9) // mirrors (not occurring)
    // legal:            $15FA              ..$15FB     semi-mirrors
         || (address >= 0x15FC && address <= 0x15FF) // mirrors (not occurring)
        )
        {   memflags[address] |= NOREAD;
        }

        if
        (   (address >= 0x14CA && address <= 0x14CB)
         || (address >= 0x14DA && address <= 0x14DB)
         || (address >= 0x14EA && address <= 0x14EB)
         || (address >= 0x14FA && address <= 0x14FB)
         || (address >= 0x15CA && address <= 0x15CB)
         || (address >= 0x15DA && address <= 0x15DB)
         || (address >= 0x15EA && address <= 0x15EB)
         || (address >= 0x15FA && address <= 0x15FB)
        )
        {   memflags[address] |= READONCE; // semi-mirrors
        }

        if (address >= 0x1600 && address <= 0x16FF)
        {   memflags[address] |= FIELD;
        }

        if
        (   address == 0x1400 + PVI_PITCH
         || address == 0x1500 + PVI_PITCH
        )
        {   memflags[address] |= AUDIBLE;
        }

        if (address & 0x1000) // $3xxx -> $1xxx, $5xxx -> $1xxx, $7xxx -> $1xxx
        {   mirror = (address & 0x0FFF) | 0x1000;
        }

        if (mirror != address)
        {   memflags[address] |= MIRRORED_R | MIRRORED_W;
        }
        mirror_r[address] =
        mirror_w[address] = (UWORD) mirror;
    }

    firstmnemonic =
    lastmnemonic  = -1;
    pvibase       = 0x1400;
    starfield     = BLACK;
    vblank        = PVI_VBLANK;
    latestsprite  = PVI_LATESTSPRITE;
    key1          =
    key2          =
    key3          =
    key4          = 2;
    bgcoll        = 0;
    analog        =
    swapped       = FALSE;
    for (i = 0; i < 4; i++)
    {   p1bgcol[i] =
        p2bgcol[i] = 0;
    }
    for (i = 0; i < 6; i++)
    {   p1sprcol[i] =
        p2sprcol[i] = 0;
    }
    p1bgcol[0] = 1;

    // Really, using 100 instead of 61 is a compatibility hack, and thus
    // should be adjustable by using the "use compatibility hacks?" setting.
    cpl = 100; // 61 causes problems
    // 950KHz = 950,000 cycles per second?
    // 950,000 / 50 = 19,000
    // For Malzak: 19,000 / 312 = 60.89744
}

#define SAA5050_DBLHI	0x0001
#define SAA5050_SEPGR	0x0002
#define SAA5050_FLASH	0x0004
#define SAA5050_BOX		0x0008
#define SAA5050_GRAPH	0x0010
#define SAA5050_CONCEAL	0x0020
#define SAA5050_HOLDGR	0x0040

#define SAA5050_BLACK   0
#define SAA5050_WHITE   7

EXPORT void m_drawscreen(void)
{   UWORD saa5050_flags;
    UBYTE // saa_colour,
          saa5050_prvcol,
          saa5050_prvchr;
    PERSIST int   blank,
                  ssx, ssy,
                  x, y;  // all PERSISTent for speed
    PERSIST UBYTE code;

    for (ssy = 0; ssy < machines[machine].height; ssy++)
    {   for (x = 0; x < PVI_XSIZE; x++)
        {   colltable[ssy][x] = 0;
        }
        the_drawline(BLACK, ssy);
    }

    // SAA 5050 - Teletext character generator
    for (ssy = 24; ssy >= 0; ssy--)
    {   // Set start of line state
        saa5050_flags   = 0;
        saa5050_prvchr  = 32;
        saa5050_forecol = SAA5050_WHITE;
        saa5050_prvcol  = SAA5050_WHITE;
        saa5050_backcol = SAA5050_BLACK;

        /* Malzak 2 does this: $00, $03, $1D, $01, $0D */

        for (ssx = 0; ssx < 42; ssx++)
        {   blank = 0;
			code = memory[0x1800 + (ssy * 64) + ssx];
			if (code < 32)
            {   switch (code)
                {
				case 0x00:
                    blank = 1;  // code 0x00 should not display anything unless HOLDGR is set
                acase 0x01: // alpha red
                case 0x02:  // alpha green
                case 0x03:  // alpha yellow
                case 0x04:  // alpha blue
				case 0x05:  // alpha purple
                case 0x06:  // alpha cyan
                case 0x07:  // alpha white
                    saa5050_prvcol = saa5050_forecol = code;
                    saa5050_flags &= ~(SAA5050_GRAPH | SAA5050_CONCEAL);
                acase 0x11: // graphics red
                case 0x12:  // graphics green
                case 0x13:  // graphics yellow
                case 0x14:  // graphics blue
				case 0x15:  // graphics purple
                case 0x16:  // graphics cyan
                case 0x17:  // graphics white
                    saa5050_prvcol = (saa5050_forecol = (code & 0x07));
                    saa5050_flags &= ~SAA5050_CONCEAL;
                    saa5050_flags |= SAA5050_GRAPH;
                acase 0x08: // flash on
                    saa5050_flags |= SAA5050_FLASH;
                acase 0x09: // flash off ("steady")
                    saa5050_flags &= ~SAA5050_FLASH;
                acase 0x0a: // end box
                    saa5050_flags &= ~SAA5050_BOX;
                acase 0x0b: // start box
                    saa5050_flags |= SAA5050_BOX;
                acase 0x0c: // normal height
                    saa5050_flags &= ~SAA5050_DBLHI;
                acase 0x0d: // double height
                    saa5050_flags |= SAA5050_DBLHI;
                acase 0x18:
                    saa5050_flags |= SAA5050_CONCEAL;
                acase 0x19:
                    saa5050_flags |= SAA5050_SEPGR;
                acase 0x1a:
                    saa5050_flags &= ~SAA5050_SEPGR;
                acase 0x1c:
                    saa5050_backcol = SAA5050_BLACK;
                acase 0x1d:
                    saa5050_backcol = saa5050_prvcol;
                acase 0x1e:
                    saa5050_flags |= SAA5050_HOLDGR;
                acase 0x1f:
                    saa5050_flags &= ~SAA5050_HOLDGR;
				}
                if (saa5050_flags & SAA5050_HOLDGR)
                {   code = saa5050_prvchr;
                } else
                {   code = 32;
            }   }

         /* if (code & 0x80)
            {   saa_colour = (saa5050_forecol << 3) | saa5050_backcol;
            } else
            {   saa_colour = saa5050_forecol | (saa5050_backcol << 3);
            } */

            if (saa5050_flags & SAA5050_CONCEAL)
            {   code = 32;
            } elif ((saa5050_flags & SAA5050_FLASH) && ((frames % 50) > 38))
            {   code = 32;
            } else
            {   saa5050_prvchr = code;
                if ((saa5050_flags & SAA5050_GRAPH) && (code & 0x20))
                {   code += (code & 0x40) ? 64 : 96;
                    if (saa5050_flags & SAA5050_SEPGR)
                    {   code += 64;
            }   }   }

            if ((blank == 0) || (saa5050_flags & SAA5050_HOLDGR))
            {   m_drawtext(code, ssx * 6, ssy * 10, (UWORD) (saa5050_flags & SAA5050_DBLHI));
        }   }
        if (saa5050_flags & SAA5050_DBLHI)
        {   ssy--;
            // saa5050_flags &= ~SAA5050_DBLHI;
    }   }

/* malzak_y is the inverted size of the top margin. Ie.
 low values of malzak_y mean the display starts towards the bottom (  0 would mean start at row       0)
high values of malzak_y mean the display starts towards the top    (255 would mean start at row    -255)

    During play, malzak_y is always 0 at the time that the screen is redrawn. */

    for (y = 0; y < 16; y++)
    {   for (x = 0; x < 16; x++)
        {   // m_field[][] is a grid of bytes, containing the codes for each tile
            ssx = ((479 + (x * 16) - malzak_x) % 256) - 15 + offsetsprites;
            ssy =         (y * 16) - malzak_y;
            m_drawtile(m_field[x][y], ssx, ssy);
}   }   }

EXPORT void m_viewplayfield(void)
{   int y;

    T1;
    for (y = 0; y < 16; y++)
    {   zprintf
        (   "%02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X\n",
            m_field[0][y],
            m_field[1][y],
            m_field[2][y],
            m_field[3][y],
            m_field[4][y],
            m_field[5][y],
            m_field[6][y],
            m_field[7][y],
            m_field[8][y],
            m_field[9][y],
            m_field[10][y],
            m_field[11][y],
            m_field[12][y],
            m_field[13][y],
            m_field[14][y],
            m_field[15][y]
        );
    }
    zprintf("\n"); T2;
}

EXPORT UBYTE m_readport(int port)
{   TRANSIENT FLAG  handled = TRUE;
 // TRANSIENT int   i;
    PERSIST   UBYTE t; // PERSISTent for speed

    // assert(machine == MALZAK);

    switch (port)
    {
    case 0:
        if (bgcoll)
        {   /* for (i = 0; i < 8; i++)
            {   if (bgcoll & (0x80 >> i))
                {   T1; zprintf("Sprite #%d hit background on frame %d!\n", i, frames);
            }   } */

            t = ((colly / 16) * 16) | (((malzak_x + 32 + collx) % 256) / 16);
        } else
        {   t = 0;
        }
    acase 0x80:
    case 0xA0:
        t = mg_console;
    adefault: // eg. $20, $65
        handled = FALSE;
        t = 0;
    }

    logport(port, t, handled, FALSE);
    return t;
}

EXPORT void m_writeport(int port, UBYTE data)
{   FLAG handled = TRUE;

    // ioport[port] = data;
    switch (port)
    {
    case 0x40: /* possibly used for SN76477 sound?
    Bit  7:    unknown
    Bit  6:    set high after checking the selected version
                (Malzak 2 only)
    Bit  5:    unused?
    Bit  4:    set high, then low, upon death
    Bits 3..1: are all set high upon death, until the game continues
    Bit  1:    set high on boot and on the title screens and during demo
    Bit  0:    constantly set high during paid gameplay */
        if (data & 0x40)
        {   if (m_bank1 == 0)
            {   m_bank1 = 1; // means Malzak 2
                switchbank(1);
        }   }
        else
        {   if (m_bank1 == 1)
            {   m_bank1 = 0; // means Malzak 1
                switchbank(1);
        }   }
    acase 0x60:
        malzak_x = data;
 /* acase 0xA0:
        game writes an echo of mg_console here */
    acase 0xC0:
        malzak_y = data;
    adefault:
        handled = FALSE;
    }

    logport(port, data, handled, TRUE);
}

EXPORT void m_emuinput(void)
{   PERSIST FLAG holding = FALSE;

    readkybd();
    ReadJoystick(0);
    ReadJoystick(1);

    if (recmode == RECMODE_PLAY)
    {   mg_console     = IOBuffer[offset++];
        memory[0x14CC] = IOBuffer[offset++]; // test switch

        if (((mg_console & 0x40) >> 1) == (mg_console & 0x20))
        {   ax[0] = sx[0] = 128;
        } elif (mg_console & 0x40) // bit 6 is right
        {   ax[0] = sx[0] = 255;
        } else
        {   // assert(mg_paddles & 0x20); // bit 5 is left
            ax[0] = sx[0] =   0;
        }
        ay[0] =
        ax[1] =
        ay[1] = 128;
        sy[0] =
        sx[1] =
        sy[1] = 128;
    } else
    {   mg_console = 0;

        if (swapped)
        {   mg_playerinput(1, 0);
        } else
        {   mg_playerinput(0, 0);
        }
        p1rumble =
        p2rumble = 0;

        if (  KeyDown(console[0]) || (jf[0] & JOYSTART) || (jf[1] & JOYSTART) || console_start)
        {   mg_console |= 4; // 1P
        }
        if (  KeyDown(console[1]) || (jf[0] & JOYA    ) || (jf[1] & JOYA    ) || console_a    )
        {   mg_console |= 8; // 2P
        }
        if (!(KeyDown(console[2]) || (jf[0] & JOYB    ) || (jf[1] & JOYB    ) || console_b   )) // active low
        {   mg_console |= 1; // coin slot
        }
        if (  KeyDown(console[3])                                             || console_reset)
        {   holding = TRUE;
        } elif (holding)
        {   holding = FALSE;

            if (memmap == MEMMAP_MALZAK2)
            {   switch (memory[0x1400 + PVI_P1PADDLE])
                {
                case 0xF0: // state "1", change to "2"
                    memory[0x1400 + PVI_P1PADDLE] = 0x90;
                acase 0x90: // state "2", change to "3"
                    memory[0x1400 + PVI_P1PADDLE] = 0x70;
                acase 0x70: // state "3", change to "4"
                    memory[0x1400 + PVI_P1PADDLE] = 0x00;
                adefault: // state "4" (eg. $00), change to "1"
                    memory[0x1400 + PVI_P1PADDLE] = 0xF0;
                }
                updatedips();
        }   }

        domouse();

        if (recmode == RECMODE_RECORD)
        {   OutputBuffer[0] = mg_console;
            OutputBuffer[1] = memory[0x14CC]; // test switch

            DISCARD fwrite(OutputBuffer, 2, 1, MacroHandle);
            // should really check return code of fwrite()
            // 2 bytes per frame * 50 frames per second = 100 bytes/sec.
    }   }

    if (console_start) console_start--;
    if (console_a    ) console_a--;
    if (console_b    ) console_b--;
    if (console_reset) console_reset--;
}
        
EXPORT void mg_playerinput(int source, int dest)
{   PERSIST UWORD jg; // PERSISTent for speed

    if
    (   hostcontroller[source] == CONTROLLER_1STDJOY
     || hostcontroller[source] == CONTROLLER_1STDPAD
  // || hostcontroller[source] == CONTROLLER_1STAJOY
     || hostcontroller[source] == CONTROLLER_1STAPAD
    )
    {   jg = jf[0];
    } elif
    (   hostcontroller[source] == CONTROLLER_2NDDJOY
     || hostcontroller[source] == CONTROLLER_2NDDPAD
     || hostcontroller[source] == CONTROLLER_2NDAJOY
     || hostcontroller[source] == CONTROLLER_2NDAPAD
    )
    {   jg = jf[1];
    } else
    {   jg = 0;
    }

    if (autofire)
    {   if
        (   (   !requirebutton
             || (jg & JOYFIRE1)
             || KeyDown(keypads[source][0])
             || (   hostcontroller[source] == CONTROLLER_TRACKBALL
                 && lmb
            )   )
         && ((frames % totalframes) < downframes)
        )
        {   // 1st firebutton
            if (machine == MALZAK)
            {   mg_console |= 0x10;
            } else
            {   // assert(machine == ASTROWARS || machine == GALAXIA);
                if (dest == 0)
                {   mg_console |= 0x20;
                } else
                {   // assert(dest == 1);
                    mg_console |= 0x10;
    }   }   }   }
    else
    {   if
        (   (jg & JOYFIRE1)
         || KeyDown(keypads[source][0])
         || KeyDown(keypads[source][2])
         || (hostcontroller[source] == CONTROLLER_TRACKBALL && lmb)
        )
        {   // 1st firebutton
            if (machine == MALZAK)
            {   mg_console |= 0x10;
            } else
            {   // assert(machine == ASTROWARS || machine == GALAXIA);
                if (dest == 0)
                {   mg_console |= 0x20;
                } else
                {   // assert(dest == 1);
                    mg_console |= 0x10;
    }   }   }   }

    if
    (   (jg & JOYFIRE2)
     || (jg & JOYFIRE3)
     || (jg & JOYFIRE4)
     || (KeyDown(keypads[source][21]))
     || (KeyDown(keypads[source][22]))
     || (KeyDown(keypads[source][23]))
     || (hostcontroller[source] == CONTROLLER_TRACKBALL && mmb)
    )
    {   // 2nd..4th firebuttons
        if (machine == MALZAK)
        {   mg_console |= 0x10;
        } else
        {   // assert(machine == ASTROWARS || machine == GALAXIA);
            if (dest == 0)
            {   mg_console |= 0x20;
            } else
            {   // assert(dest == 1);
                mg_console |= 0x10;
    }   }   }

    engine_dopaddle(source, 0);

    if (ax[0] < 64)
    {   if (machine == MALZAK)
        {   mg_console |= 0x40; // left
        } else
        {   // assert(machine == ASTROWARS || machine == GALAXIA);
            if (dest == 0)
            {   g_paddles |= 0x10;
            } else
            {   // assert(dest == 1);
                g_paddles |= 0x04;
    }   }   }
    elif (ax[0] > 192)
    {   if (machine == MALZAK)
        {   mg_console |= 0x20; // right
        } else
        {   // assert(machine == ASTROWARS || machine == GALAXIA);
            if (dest == 0)
            {   g_paddles |= 0x20;
            } else
            {   // assert(dest == 1);
                g_paddles |= 0x08;
    }   }   }

    if (machine == MALZAK)
    {   if (ay[0] < 64)
        {   mg_console |= 0x80; // up
        } elif (ay[0] > 192)
        {   mg_console |= 0x02; // down
    }   }

    sx[0] = (UBYTE) ax[0];
    sy[0] = (UBYTE) ay[0];
}

EXPORT void m_savenvram(void)
{   FILE* TheLocalHandle; // LocalHandle is a variable of winbase.h

#ifdef AMIGA
    if ((TheLocalHandle = fopen("PROGDIR:MALZAK2.NVR", "wb")))
#endif
#ifdef WIN32
    if ((TheLocalHandle = fopen("MALZAK2.NVR", "wb")))
#endif
    {   DISCARD fwrite(&memory[0x1700], (size_t) 256, 1, TheLocalHandle);
        DISCARD fwrite(&memory[0x14CC],            1, 1, TheLocalHandle); // test switch
        DISCARD fclose(TheLocalHandle);
        // TheLocalHandle = NULL;
#ifdef VERBOSE
        T1; textpen(TEXTPEN_VERBOSE);
        zprintf("Saved NVRAM.\n"); T2;
#endif
}   }

MODULE __inline void local_drawpixel(int x, int y, int colour)
{   if (screen[x][y] != colour)
    {   screen[x][y] = (UBYTE) colour;
        drawpixel(x, y, colour);
}   }
