// INCLUDES---------------------------------------------------------------

#ifdef __VBCC__
    #define AMIGA
#endif
#ifdef AMIGA
    #include "amiga.h"
    #include <stdio.h>
#endif
#ifdef _lint
    #define WIN32
    #define _inline
#endif
#ifdef WIN32
    #include "ibm.h"
    #define EXEC_TYPES_H
#endif

#define CATCOMP_NUMBERS
#define CATCOMP_BLOCK
#include "aa_strings.h"
#ifdef AMIGA
    #include <proto/locale.h>  // GetCatalogStr()
#endif

#include <stdio.h>

#include "aa.h"
#include "phunsy.h"

// DEFINES----------------------------------------------------------------

// (none)

// IMPORTED VARIABLES-----------------------------------------------------

IMPORT FLAG                 changed,
                            fiftytheme,
                            piano;
IMPORT UBYTE                memory[32768];
IMPORT UBYTE*               IOBuffer;
IMPORT ASCREEN              screen[BOXWIDTH][BOXHEIGHT];
IMPORT UWORD                iar,
                            mirror_r[32768],
                            mirror_w[32768];
IMPORT ULONG                memflags[32768];
IMPORT int                  consoleopen,
                            cy_fgc,
                            cy_bgc,
                            game,
                            inframe,
                            inverse,
                            logmonitor,
                            machine,
                            queuepos,
                            offset,
                            overcalc,
                            recmode,
                            shadowing,
                            slice,
                            whichgame;
IMPORT FILE*                MacroHandle;
IMPORT struct KnownStruct   known[KNOWNGAMES];

#ifdef AMIGA
    IMPORT struct Catalog*  CatalogPtr;
    IMPORT FILE*            PrinterHandle;
#endif
#ifdef WIN32
    IMPORT int              CatalogPtr; // APTR doesn't work
    IMPORT HANDLE           PrinterHandle;
#endif

// EXPORTED VARIABLES-----------------------------------------------------

EXPORT UBYTE                banked,
                            q_bank[16][16384],
                            u_bank[16][2048];

// MODULE VARIABLES-------------------------------------------------------

// (none)

// MODULE FUNCTIONS-------------------------------------------------------

MODULE __inline void local_drawpixel(int x, int y, int colour);

// CODE-------------------------------------------------------------------

EXPORT void draw_phunsy(void)
{   // all PERSISTent for speed
    PERSIST int   x, xx, y, yy;
    PERSIST UBYTE fgc,
                  imagery;

    if (machine != PHUNSY)
    {   return;
    }

    // generate screen[][] and display[] based on memory[]

    if (shadowing)
    {   for (x = 0; x < 64; x++)
        {   for (y = 0; y < 32; y++)
            {   imagery = memory[0x1000 + (y * 64) + x];
                if (imagery >= 128)
                {   imagery &= 0x0F;

                    for (yy = 0; yy < 8; yy++)
                    {   for (xx = 0; xx < 6; xx++)
                        {   if ((x == 63 && xx == 5) || (y == 31 && yy == 7))
                            {   continue;
                            }
                            if (phunsy_gfx[imagery][yy] & (0x80 >> xx))
                            {   local_drawpixel((x * 6) + xx + 1, (y * 8) + yy + 1, GREY1);
                            } else
                            {   local_drawpixel((x * 6) + xx + 1, (y * 8) + yy + 1, cy_bgc);
                }   }   }   }
                else
                {   if (imagery == 0)
                    {   imagery = 0x20; // maybe we shouldn't be doing this?
                    } elif (imagery <= 0x1F)
                    {   imagery += 0x40;
                    } elif (imagery >= 0x40 && imagery <= 0x5F)
                    {   imagery -= 0x40;
                    }
                    for (yy = 0; yy < 8; yy++)
                    {   for (xx = 0; xx < 6; xx++)
                        {   if ((x == 63 && xx == 5) || (y == 31 && yy == 7))
                            {   continue;
                            }
                            if (phunsy_chars[imagery][yy] & (0x20 >> xx))
                            {   local_drawpixel((x * 6) + xx + 1, (y * 8) + yy + 1, GREY1);
                            } else
                            {   local_drawpixel((x * 6) + xx + 1, (y * 8) + yy + 1, cy_bgc);
    }   }   }   }   }   }   }

    for (x = 0; x < 64; x++)
    {   for (y = 0; y < 32; y++)
        {   imagery = memory[0x1000 + (y * 64) + x];
            if (imagery >= 128)
            {   if ((imagery & 0x70) == 0)
                {   fgc = cy_bgc;
                } elif ((imagery & 0x70) == 0x70)
                {   fgc = cy_fgc;
                } else
                {   if (inverse)
                    {   fgc = 15 + ((imagery & 0x70) >> 4); // 1..6 -> 16..21
                    } else
                    {   fgc = 22 - ((imagery & 0x70) >> 4); // 6..1 -> 21..16
                }   }
                imagery &= 0x0F;

                for (yy = 0; yy < 8; yy++)
                {   for (xx = 0; xx < 6; xx++)
                    {   if (phunsy_gfx[imagery][yy] & (0x80 >> xx))
                        {   local_drawpixel((x * 6) + xx, (y * 8) + yy, fgc);
                        } elif (!shadowing)
                        {   local_drawpixel((x * 6) + xx, (y * 8) + yy, cy_bgc);
            }   }   }   }
            else
            {   if (imagery == 0)
                {   imagery = 0x20; // maybe we shouldn't be doing this?
                } elif (imagery <= 0x1F)
                {   imagery += 0x40;
                } elif (imagery >= 0x40 && imagery <= 0x5F)
                {   imagery -= 0x40;
                }
                for (yy = 0; yy < 8; yy++)
                {   for (xx = 0; xx < 6; xx++)
                    {   if (phunsy_chars[imagery][yy] & (0x20 >> xx))
                        {   local_drawpixel((x * 6) + xx, (y * 8) + yy, cy_fgc);
                        } elif (!shadowing)
                        {   local_drawpixel((x * 6) + xx, (y * 8) + yy, cy_bgc);
    }   }   }   }   }   }

    changed = FALSE;
}

EXPORT void phunsy_setmemmap(void)
{   int i, j;

    // assert(memmap == MEMMAP_PHUNSY);

    game = FALSE;

    for (i = 0; i < 16; i++)
    {   for (j = 0; j <=  0x7FF; j++)
        {   u_bank[i][j] = 0;
        }
        for (j = 0; j <= 0x3FFF; j++)
        {   q_bank[i][j] = 0;
    }   }
    for (i = 0x800; i <= 0x7FFF; i++)
    {   memory[i]    = 0;
    }
    for (i = 0; i <=  0x7FF; i++)
    {   memory[i]    = phunsy_bios[i];
        u_bank[1][i] = pdcr[i];
        u_bank[2][i] = dass[i];
        u_bank[3][i] = labhnd[i];
    }

    if (known[whichgame].complang == CL_MWB)
    {   for (i = 0; i <= 0x1983; i++)
        {   memory[0x4000 + i] = mwb[i];
        }
#ifdef VERBOSE
        T1; zprintf("Loading MWB...\n"); T2;
#endif
    }

    for (i = 0; i < 32 * KILOBYTE; i++)
    {   memflags[i] = 0;

        if (i <= 0x7FF)
        {   memflags[i] |= NOWRITE;
        } elif (i >= 0xEE0 && i <= 0xFFF)
        {   memflags[i] |= BIOSRAM;
        } elif (i >= 0x1000 && i <= 0x17FF)
        {   memflags[i] |= VISIBLE;
        } elif (i >= 0x1800 && i <= 0x1FFF)
        {   memflags[i] |= BANKU;
        } elif (i >= 0x4000)
        {   memflags[i] |= BANKQ;
        }

        mirror_r[i] =
        mirror_w[i] = (UWORD) i;
    }

    banked = 0; // U0, Q0
    ramtoubank(0);
    ramtoqbank(0);
}

EXPORT void phunsy_emulate(void)
{   inframe = TRUE;

    // 1,000KHz = 1,000,000 cycles per second
    // 1,000,000 / 50 = 20000.0
    slice = 20000 - overcalc;

    cpu();

    endofframe();
}

EXPORT void phunsy_writeport(int port, UBYTE data)
{   FLAG handled = TRUE;

    switch (port)
    {
    case PORTC:
        if ((data & 0xF0) != (banked & 0xF0))
        {   ramtoubank((banked & 0xF0) >> 4);
            ubanktoram((data   & 0xF0) >> 4);
        }
        if ((data & 0x0F) != (banked & 0x0F))
        {   ramtoqbank( banked & 0x0F      );
            qbanktoram( data   & 0x0F      );
        }
        banked = data;
    acase PORTD:
        if ((piano || fiftytheme) && iar == 0x1827)
        {   playsound(FALSE);
        } else
        {   handled = FALSE;
        }
    adefault:
        handled = FALSE;
    }

    logport(port, data, handled, TRUE);
}

EXPORT UBYTE phunsy_readport(int port)
{   TRANSIENT UBYTE t = 0; // inititialized to avoid spurious SAS/C compiler warnings
    TRANSIENT FLAG  handled = TRUE;
    PERSIST   UBYTE lastkey;
    PERSIST   int   localfresh = 0;

    switch (port)
    {
    case PORTC:
        t = 2;
    acase PORTD:
        if (recmode == RECMODE_PLAY)
        {   t = IOBuffer[offset++];
        } else
        {   // assert(recmode == RECMODE_NORMAL || recmode == RECMODE_RECORD);

            switch (localfresh)
            {
            case 0:
                if (queuepos)
                {   lastkey = t = thechin();
                    localfresh = 1;
                } else
                {   t = 0x80;
                }
            acase 1:
                t = lastkey;
                localfresh = 2;
            acase 2:
                t = 0x80;
                localfresh = 0;
            }

            if (recmode == RECMODE_RECORD)
            {   DISCARD fwrite(&t, 1, 1, MacroHandle);
        }   }
    adefault:
        t = 0;
        handled = FALSE;
    }

    logport(port, t, handled, FALSE);
    return t;
}

MODULE __inline void local_drawpixel(int x, int y, int colour)
{   if (screen[x][y] != (UBYTE) colour)
    {   screen[x][y] = (UBYTE) colour;
        drawpixel(x, y, colour);
}   }

EXPORT void ramtoubank(int bank)
{   int i;

#ifdef VERBOSE
    T1; zprintf("Memory -> U bank #%d.\n", bank); T2;
#endif
    for (i = 0; i <= 0x7FF; i++)
    {   u_bank[bank][i] = memory[0x1800 + i];
}   }

EXPORT void ramtoqbank(int bank)
{   int i;

#ifdef VERBOSE
    T1; zprintf("Memory -> Q bank #%d.\n", bank); T2;
#endif
    for (i = 0; i <= 0x3FFF; i++)
    {   q_bank[bank][i] = memory[0x4000 + i];
}   }

EXPORT void ubanktoram(int bank)
{   int i;

#ifdef VERBOSE
    T1; zprintf("U bank #%d -> memory.\n", bank); T2;
#endif
    for (i = 0; i <= 0x7FF; i++)
    {   memory[0x1800 + i] = u_bank[bank][i];
    }

    if (bank == 0 || bank >= 4) // banks U0, U4..U15 (RAM)
    {   ;
#ifdef VERBOSE
        T1; zprintf("Bank is U%X (RAM).\n", bank); T2;
#endif
        for (i = 0x1800; i <= 0x1FFF; i++)
        {   memflags[i] &= ~(NOWRITE | EPROM); // RAM
        }
    } else // banks U1..U3 (EPROM)
    {   ;
#ifdef VERBOSE
        T1; zprintf("Bank is U%X (EPROM).\n", bank); T2;
#endif
        for (i = 0x1800; i <= 0x1FFF; i++)
        {   memflags[i] |= NOWRITE | EPROM; // EPROM
}   }   }

EXPORT void qbanktoram(int bank)
{   int i;

#ifdef VERBOSE
    T1; zprintf("Q bank #%d -> memory.\n", bank); T2;
#endif
    for (i = 0; i <= 0x3FFF; i++)
    {   memory[0x4000 + i] = q_bank[bank][i];
    }

    // banks Q0..Q15 (RAM)
    // Q0 does need to be RAM!
#ifdef VERBOSE
    T1; zprintf("Bank is Q%X (RAM).\n", bank); T2;
#endif
}

EXPORT void phunsy_viewscreen(void)
{   int i,
        y;

    T1; textpen(TEXTPEN_VIEW);
    for (y = 0; y < 32; y++)
    {   for (i = 0; i < 4; i++)
        {   dumprow(0x1000 + (y * 64) + (i * 16));
        }
        zprintf("\n");
    }
    T2;
}

EXPORT float phunsy_getpianopitch(UWORD value)
{   PERSIST const float phunsy_pitch[13 + 1] = {
    (const float)    0.00,
    (const float)  261.63, //  1: middle C
    (const float)  277.18, //  2: middle C#
    (const float)  293.66, //  3: middle D
    (const float)  311.13, //  4: middle D#
    (const float)  329.63, //  5: middle E
    (const float)  349.23, //  6: middle F
    (const float)  369.99, //  7: middle F#
    (const float)  392.00, //  8: middle G
    (const float)  415.30, //  9: middle G#
    (const float)  440.00, // 10: middle A
    (const float)  466.16, // 11: middle A#
    (const float)  493.88, // 12: middle B
    (const float)  523.25  // 13:   high C
};
    
    if (value > 13)
    {   return 0.0;
    } // implied else
    return phunsy_pitch[value];
}

EXPORT float phunsy_getfiftypitch(UWORD value)
{   return (const float) (7874.0 / value);
}

EXPORT float phunsy_getbelmachpitch(UWORD value)
{   PERSIST const float phunsy_pitch[62] = {
    (const float) ( 233.08 / 2.0), // $00: vlow   A#
    (const float) ( 246.94 / 2.0), // $01: vlow   B
    (const float) ( 261.63 / 2.0), // $02: low    C
    (const float) ( 277.18 / 2.0), // $03: low    C#
    (const float) ( 293.66 / 2.0), // $04: low    D
    (const float) ( 311.13 / 2.0), // $05: low    D#
    (const float) ( 329.63 / 2.0), // $06: low    E
    (const float) ( 349.23 / 2.0), // $07: low    F
    (const float) ( 369.99 / 2.0), // $08: low    F#
    (const float) ( 392.00 / 2.0), // $09: low    G
    (const float) ( 415.30 / 2.0), // $0A: low    G#
    (const float)   220.00,        // $0B: low    A
    (const float)   233.08,        // $0C: low    A#
    (const float)   246.94,        // $0D: low    B
    (const float)   261.63,        // $0E: middle C
    (const float)   277.18,        // $0F: middle C#
    (const float)   293.66,        // $10: middle D
    (const float)   311.13,        // $11: middle D#
    (const float)   329.63,        // $12: middle E
    (const float)   349.23,        // $13: middle F
    (const float)   369.99,        // $14: middle F#
    (const float)   392.00,        // $15: middle G
    (const float)   415.30,        // $16: middle G#
    (const float)   440.00,        // $17: middle A
    (const float)   466.16,        // $18: middle A#
    (const float)   493.88,        // $19: middle B
    (const float) ( 523.25 * 1.0), // $1A:  high   C
    (const float) ( 554.37 * 1.0), // $1B:  high   C#
    (const float) ( 587.32 * 1.0), // $1C:  high   D
    (const float) ( 622.25 * 1.0), // $1D:  high   D#
    (const float) ( 659.26 * 1.0), // $1E:  high   E
    (const float) ( 698.46 * 1.0), // $1F:  high   F
    (const float) ( 739.98 * 1.0), // $20:  high   F#
    (const float) ( 783.99 * 1.0), // $21:  high   G
    (const float) ( 830.60 * 1.0), // $22:  high   G#
    (const float) ( 880.00 * 1.0), // $23:  high   A
    (const float) ( 932.32 * 1.0), // $24:  high   A#
    (const float) ( 987.80 * 1.0), // $25:  high   B
    (const float) ( 523.25 * 2.0), // $26: vhigh   C
    (const float) ( 554.37 * 2.0), // $27: vhigh   C#
    (const float) ( 587.32 * 2.0), // $28: vhigh   D
    (const float) ( 622.25 * 2.0), // $29: vhigh   D#
    (const float) ( 659.26 * 2.0), // $2A: vhigh   E
    (const float) ( 698.46 * 2.0), // $2B: vhigh   F
    (const float) ( 739.98 * 2.0), // $2C: vhigh   F#
    (const float) ( 783.99 * 2.0), // $2D: vhigh   G
    (const float) ( 830.60 * 2.0), // $2E: vhigh   G#
    (const float) ( 880.00 * 2.0), // $2F: vhigh   A
    (const float) ( 932.32 * 2.0), // $30: vhigh   A#
    (const float) ( 987.80 * 2.0), // $31: vhigh   B
    (const float) ( 523.25 * 3.0), // $32: xhigh   C
    (const float) ( 554.37 * 3.0), // $33: xhigh   C#
    (const float) ( 587.32 * 3.0), // $34: xhigh   D
    (const float) ( 622.25 * 3.0), // $35: xhigh   D#
    (const float) ( 659.26 * 3.0), // $36: xhigh   E
    (const float) ( 698.46 * 3.0), // $37: xhigh   F
    (const float) ( 739.98 * 3.0), // $38: xhigh   F#
    (const float) ( 783.99 * 3.0), // $39: xhigh   G
    (const float) ( 830.60 * 3.0), // $3A: xhigh   G#
    (const float) ( 880.00 * 3.0), // $3B: xhigh   A
    (const float) ( 932.32 * 3.0), // $3C: xhigh   A#
    (const float) ( 987.80 * 3.0), // $3D: xhigh   B
};

    return (float) (phunsy_pitch[value] / 2.0);
}
