// INCLUDES---------------------------------------------------------------

#ifdef __VBCC__
    #define AMIGA
#endif
#ifdef AMIGA
    #include "amiga.h"
    #include <string.h>
    #include <stdlib.h> // eg. for rand()
#endif
#ifdef _lint
    #define WIN32
    #define _inline
#endif
#ifdef WIN32
    #include "ibm.h"
#endif

#include <stdio.h>

#include "aa.h"
#include "pipbug.h"

// DEFINES----------------------------------------------------------------

// (none)

// EXPORTED VARIABLES-----------------------------------------------------

EXPORT UBYTE                SpeakBuffer[80 + 1];

// IMPORTED VARIABLES-----------------------------------------------------

IMPORT FLAG                 changed,
                            printer;
IMPORT UBYTE                opcode,
                            memory[32768],
                            p_digit[4],
                            psu,
                            r[7],
                            vdu[MAX_VDUCOLUMNS][MAX_VDUROWS];
IMPORT ASCREEN              screen[BOXWIDTH][BOXHEIGHT];
IMPORT UWORD                iar,
                            mirror_r[32768],
                            mirror_w[32768],
                            ras[8];
IMPORT ULONG                frames,
                            memflags[32768],
                            sound;
IMPORT int                  colourset,
                            consoleopen,
                            drawcorners,
                            drawunlit,
                            game,
                            hostvolume,
                            machine,
                            memmap,
                            echoconsole,
                            echoprinter,
                            inframe,
                            highres,
                            overcalc,
                            pbx,
                            shadowing,
                            slice,
                            speakoffset,
                            usespeech,
                            vducols;
IMPORT struct MachineStruct machines[MACHINES];
IMPORT const STRPTR         asciiname[32];
#ifdef AMIGA
    IMPORT FILE*            PrinterHandle;
    IMPORT BPTR             SpeakHandle;
#endif
#ifdef WIN32
    IMPORT HANDLE           PrinterHandle;
#endif

// MODULE STRUCTURES -----------------------------------------------------

// (none)

// MODULE FUNCTIONS-------------------------------------------------------

MODULE void p_scroll(void);
MODULE __inline void local_drawpixel(int x, int y, int colour);

// CODE-------------------------------------------------------------------

EXPORT void ph_cout(FLAG printroutine, UBYTE value)
{   PERSIST   TEXT  printstring[2] = "#";
#ifdef WIN32
    TRANSIENT DWORD numwritten;
#endif

    /* T1;
    if (value >= 32)
    {   zprintf("COUT: $%02X ('%c').\n", value, value);
    } else
    {   zprintf("COUT: $%02X (%s).\n",   value, asciiname[value]);
    }
    T2; */

    // emit value as an ASCII character
    if
    (   value !=    0 // NULl
     && value != 0x1B // ESCape
#ifdef AMIGA
     && value != '\t' // workaround for a strange reset-to-game bug
#endif
    )
    {   if (printroutine)
        {   T1; textpen(TEXTPEN_PRINTER);
            zprintf("%c", value & 0x3F);
            T2;
        } else
        {   if (echoconsole)
            {   T1; textpen(TEXTPEN_PIPBUG);
                zprintf("%c", value);
                if (value == 8) // backspace
                {   zprintf(" \b");
                }
                T2;
#ifdef AMIGA
                DISCARD fflush(NULL); // it would be better to only flush stdout?
#endif
            }

            changed = TRUE;

            switch (value)
            {
            case 7: // Ctrl+G, ie. bell (\a)
                play_sample(SAMPLE_BELL); // unfortunately we also get the default Windows bell (for WinArcadia)
                goto HERE;
            acase 8: // backspace (\b)
                if (pbx)
                {   vdu[pbx--][PH_VDUROWS - 1] = BLANK;
                    vdu[pbx  ][PH_VDUROWS - 1] = BLANK;
                    goto HERE;
                }
            /* This doesn't work properly.
            acase 9: // horizontal tab (\t)
                pbx = (pbx / 8) + 8;
                goto HERE; */
            acase LF: // 10
                p_scroll();
                goto HERE;
            acase CR: // 13
                vdu[pbx][PH_VDUROWS - 1] = BLANK;
                pbx = 0;
                goto HERE;
            adefault:
                if (value < 32 || value > 127)
                {   return;
            }   }

            vdu[pbx++][PH_VDUROWS - 1] = value - 32;

            if (pbx == vducols)
            {   pbx = 0;
                p_scroll();
            }

HERE:
            if (usespeech)
            {   if (value == 10)
                {   SpeakBuffer[speakoffset] = 0;
                    // assert(speakoffset == strlen(SpeakBuffer));
                    if (
#ifdef AMIGA
                    SpeakHandle &&
#endif
                    sound && hostvolume && speakoffset)
                    {
#ifdef AMIGA
                        DISCARD Write(SpeakHandle, SpeakBuffer, speakoffset + 1);
#endif
#ifdef WIN32
                        speak();
#endif
#ifdef VERBOSE
                        zprintf("SPEAK: %s\n", SpeakBuffer);
#endif
                    }
                    speakoffset = 0;
                } elif
                (   value != 13
                 && speakoffset < 80
                 && (ras[(psu & PSU_SP) - 1] != 0x002C) // better than (value != '*')
                )
                {   if
                    (   value == '='
                     && !strncmp((const char *) SpeakBuffer, "NO LEFT", 7) // special-case code for Nim.pgm
                    )
                    {   strcpy((char *) SpeakBuffer, "NUMBER LEFT="); // NO. instead of NUMBER doesn't work because of end-of-sentence (.) detection
                        speakoffset = 12;
                    } elif
                    (   value == '-'
                     && !strncmp((const char *) SpeakBuffer, "MOVE-",   5) // special-case code for Nim.pgm
                    )
                    {   strcpy((char *) SpeakBuffer, "MOVE");
                        speakoffset = 4;
                    } else
                    {   SpeakBuffer[speakoffset++] = value;

                        if
                        (   value == '!'
                         || value == '?'
                         || value == '.'
                        )
                        {   SpeakBuffer[speakoffset] = 0;
                            // assert(speakoffset == strlen(SpeakBuffer));
                            if (
#ifdef AMIGA
                            SpeakHandle &&
#endif
                            sound && hostvolume)
                            {   // assert(speakoffset);
#ifdef AMIGA
                                DISCARD Write(SpeakHandle, SpeakBuffer, speakoffset + 1);
#endif
#ifdef WIN32
                                speak();
#endif
#ifdef VERBOSE
                                zprintf("SPEAK: %s\n", SpeakBuffer);
#endif
                            }
                            speakoffset = 0;
            }   }   }   }
            if (echoprinter && PrinterHandle)
            {   printstring[0] = (TEXT) value;
                // assert(printstring[1] == 0);
#ifdef AMIGA
                DISCARD fwrite(printstring, 1, 1, PrinterHandle); // was 2, 1 until V9.32
#endif
#ifdef WIN32
                DISCARD WriteFile(PrinterHandle, printstring, 1, &numwritten, NULL);
                // assert(numwritten == 1);
#endif

                // would it be better to buffer it and print a
                // line at a time, as is done for speech?
}   }   }   }

MODULE void p_scroll(void)
{   int x, y;

    for (y = 0; y < PH_VDUROWS - 1; y++)
    {   for (x = 0; x < vducols; x++)
        {   vdu[x][y] = vdu[x][y + 1];
    }   }
    for (x = 0; x < vducols; x++)
    {   vdu[x][PH_VDUROWS - 1] = BLANK;
}   }

EXPORT void draw_pipbug(void)
{   // all PERSISTent for speed
    PERSIST int   incy,
                  starty,
                  x, xx, y, yy, yyy;
    PERSIST UBYTE p_cc,
                  tempfgc,
                  tempvdu;

    // assert(machine == PIPBUG || machine == CHAOS);

    // generate screen[][] and display[] based on vdu[][]

    if (colourset == GREYSCALE)
    {   p_cc = WHITE;
    } else
    {   p_cc = CYAN;
    }

    if (highres)
    {   starty = 0;
        incy = 1;
    } else
    {   starty = PH_VDUROWS / 2;
        incy = 2;
    }

    for (x = 0; x < machines[machine].width; x++)
    {   for (y = 0; y <= 259; y++)
        {   if (screen[x][y] != p_bgc)
            {   local_drawpixel(x, y, p_bgc);
    }   }   }

    if (shadowing)
    {   for (x = 0; x < vducols; x++)
        {   yyy = 3;
            for (y = starty; y < PH_VDUROWS; y++)
            {   tempvdu = vdu[x][y];
                for (yy = 0; yy < 8; yy++)
                {   for (xx = 0; xx < 6; xx++)
                    {   if ((x == vducols - 1 && xx == 4) || (y == PH_VDUROWS - 1 && yy == 8))
                        {   continue;
                        }

                        if (p_chars[tempvdu][yy] & (0x10 >> xx))
                        {   local_drawpixel((x * 6) + xx + 2, yyy + 1, p_sgc);
                            if (!highres)
                            {   local_drawpixel((x * 6) + xx + 2, yyy + 2, p_sgc);
                    }   }   }
                    yyy += incy;
    }   }   }   }

    for (x = 0; x < vducols; x++)
    {   yyy = 3;
        for (y = starty; y < PH_VDUROWS; y++)
        {   tempvdu = vdu[x][y];
            if (colourset == GREYSCALE)
            {   tempfgc = WHITE;
            } else
            {   tempfgc = p_fgc[tempvdu];
            }
            for (yy = 0; yy < 8; yy++)
            {   for (xx = 0; xx < 6; xx++)
                {   if (p_chars[tempvdu][yy] & (0x10 >> xx))
                    {   // assert(x < vducols - 1 || xx < 5);
                        local_drawpixel((x * 6) + xx + 1, yyy, tempfgc);
                        if (!highres)
                        {   local_drawpixel((x * 6) + xx + 1, yyy + 1, tempfgc);
                }   }   }
                yyy += incy;
    }   }   }

#ifdef SOLIDCURSOR
    if (1)
#else
    if ((frames % 50) < 25)
#endif
    {   if (highres)
        {   for (xx = 0; xx < 5; xx++)
            {   local_drawpixel((pbx * 6) + xx + 1, ((PH_VDUROWS - 1) * 8) + 9, p_cc);
        }   }
        else
        {   for (xx = 0; xx < 5; xx++)
            {   local_drawpixel((pbx * 6) + xx + 1, ((PH_VDUROWS - 1) * 8) + 7, p_cc);
                local_drawpixel((pbx * 6) + xx + 1, ((PH_VDUROWS - 1) * 8) + 8, p_cc);
    }   }   }

    if (machine == PIPBUG)
    {   p_drawdigit(0, (UBYTE) (0x80 | p_digit[0]));
        p_drawdigit(0, (UBYTE) (0x40 | p_digit[1]));
        p_drawdigit(0, (UBYTE) (0x20 | p_digit[2]));
        p_drawdigit(0, (UBYTE) (0x10 | p_digit[3]));
    }

    changed = FALSE;
}

EXPORT void p_setmemmap(void)
{   int i, address, mirror;

    game    = FALSE;
    vducols = P_VDUCOLUMNS;

    for (i = 0; i <= 0x3FF; i++)
    {   memory[i] = p_bios[i];
    }
    for (i = 0x400; i <= 0x7FFF; i++)
    {   memory[i] = 0;
    }
    if (memmap == MEMMAP_P)
    {   for (i = 0; i < (int) P_EPROMSIZE; i++)
        {   memory[0x3C00 + i] = p_eprom[i];
    }   }

    for (address = 0; address < 32 * KILOBYTE; address++)
    {   memflags[address] = 0;
        mirror            = address;

        if (address <= 0x3FF)
        {   memflags[address] |= NOWRITE;
        }

        if (memmap == MEMMAP_J)
        {   if (mirror >= 0x800 && mirror <= 0x7FFF)
            {   mirror &= 0x7FF;
            }
            if (mirror >= 0x500 && mirror <= 0x7FF)
            {   mirror &= 0x4FF; // %00000x00 %xxxxxxxx
        }   }
        elif (memmap == MEMMAP_K)
        {   if (mirror >= 0x800 && mirror <= 0x7FFF)
            {   mirror &= 0x7FF;
            }
            if (mirror >= 0x600 && mirror <= 0x7FF)
            {   mirror &= 0x5FF; // %00000x0x %xxxxxxxx
        }   }
        elif (memmap == MEMMAP_L)
        {   if (address >= 0x7FF)
            {   memflags[address] |= NOREAD | NOWRITE;
        }   }
        elif (memmap == MEMMAP_M)
        {   if (address >= 0x1FFF)
            {   memflags[address] |= NOREAD | NOWRITE;
        }   }
        else
        {   // assert(memmap == MEMMAP_P);
            if
            (   (address >= 0x3000 && address <= 0x3BFF)
             || (address >= 0x4000 && address <= 0x6CFF)
            )
            {   memflags[address] |= NOREAD | NOWRITE;
            } elif (address >= 0x3C00 && address <= 0x3FFF)
            {   memflags[address] |= NOWRITE | EPROM;
        }   }

        if (mirror != address)
        {   memflags[address] |= MIRRORED_R | MIRRORED_W;
        }
        mirror_r[address] =
        mirror_w[address] = (UWORD) mirror;
}   }

EXPORT void p_emulate(void)
{   PERSIST int i, j;

    inframe = TRUE;

    // this is to emulate the non-persistence of the segment display
    if (drawunlit)
    {   for (i = 0; i < 4; i++)
        {   for (j = 0; j < 7; j++)
            {   drawsegment(i, j, UNLIT);
    }   }   }
    else
    {   for (i = 0; i < 4; i++)
        {   for (j = 0; j < 7; j++)
            {   drawsegment(i, j, BLACK);
    }   }   }
    if (drawunlit && drawcorners)
    {   for (i = 0; i < 4; i++)
        {   for (j = 7; j <= 12; j++)
            {   drawsegment(i, j, UNLIT);
    }   }   }
    else
    {   for (i = 0; i < 4; i++)
        {   for (j = 7; j <= 12; j++)
            {   drawsegment(i, j, BLACK);
    }   }   }
    changed = TRUE;

    // 1MHz = 1,000,000 cycles per second
    // 1,000,000 / 50 = 20,000
    slice = 20000 - overcalc;
    cpu();

    endofframe();
}

EXPORT float pipbug_getpitch(UWORD value)
{   PERSIST const float p_pitch[] = {
    (const float)   82.40, // $30:   vlow E
    (const float)   87.30, // $31:   vlow F
    (const float)   92.49, // $32:   vlow F#
    (const float)   98.00, // $33:   vlow G
    (const float)  103.82, // $34:   vlow G#
    (const float)  110.00, // $35:   vlow A
    (const float)  116.54, // $36:   vlow A#
    (const float)  123.47, // $37:   vlow B
    (const float)  130.81, // $38:    low C
    (const float)  138.59, // $39:    low C#
    (const float)  146.83, // $3A:    low D
    (const float)  155.56, // $3B:    low D#
    (const float)    0.00, // $3C
    (const float)    0.00, // $3D
    (const float)    0.00, // $3E
    (const float)    0.00, // $3F
    (const float)  164.81, // $40:    low E
    (const float)  174.61, // $41:    low F
    (const float)  184.99, // $42:    low F#
    (const float)  196.00, // $43:    low G
    (const float)  207.65, // $44:    low G#
    (const float)  220.00, // $45:    low A
    (const float)  233.08, // $46:    low A#
    (const float)  246.94, // $47:    low B
    (const float)  261.63, // $48: middle C
    (const float)  277.18, // $49: middle C#
    (const float)  293.66, // $4A: middle D
    (const float)  311.13, // $4B: middle D#
    (const float)    0.00, // $4C
    (const float)    0.00, // $4D
    (const float)    0.00, // $4E
    (const float)    0.00, // $4F
    (const float)  329.63, // $50: middle E
    (const float)  349.23, // $51: middle F
    (const float)  369.99, // $52: middle F#
    (const float)  392.00, // $53: middle G
    (const float)  415.30, // $54: middle G#
    (const float)  440.00, // $55: middle A
    (const float)  466.16, // $56: middle A#
    (const float)  493.88, // $57: middle B
    (const float)  523.25, // $58:   high C
    (const float)  554.37, // $59:   high C#
    (const float)  587.33, // $5A:   high D
    (const float)  622.25, // $5B:   high D#
    (const float)    0.00, // $5C
    (const float)    0.00, // $5D
    (const float)    0.00, // $5E
    (const float)    0.00, // $5F
    (const float)  659.26, // $60:   high E
    (const float)  698.46, // $61:   high F
    (const float)  739.99, // $62:   high F#
    (const float)  783.99, // $63:   high G
    (const float)  830.61, // $64:   high G#
    (const float)  880.00, // $65:   high A
    (const float)  932.30, // $66:   high A#
    (const float)  987.80, // $67:   high B
    (const float) 1046.50, // $68:  vhigh C
    (const float) 1108.70, // $69:  vhigh C#
    (const float) 1174.70, // $6A:  vhigh D
    (const float) 1244.50, // $6B:  vhigh D#
    (const float)    0.00, // $6C
    (const float)    0.00, // $6D
    (const float)    0.00, // $6E
    (const float)    0.00, // $6F
    (const float) 1318.50, // $70:  vhigh E
    (const float) 1396.90, // $71:  vhigh F
    (const float) 1480.00, // $72:  vhigh F#
    (const float) 1568.00, // $73:  vhigh G
    (const float) 1661.22, // $74:  vhigh G#
    (const float) 1760.00, // $75:  vhigh A
    (const float) 1864.60, // $76:  vhigh A#
    (const float) 1975.60, // $77:  vhigh B
    (const float) 2093.00, // $78:  xhigh C
    (const float) 2217.40, // $79:  xhigh C#
    (const float) 2349.40, // $7A:  xhigh D
    (const float) 2489.00  // $7B:  xhigh D#
};

    if (value < 0x30 || value > 0x7B)
    {   return 0.0;
    } else
    {   return(p_pitch[value - 0x30]);
}   }

EXPORT void p_drawdigit(int position, UBYTE value)
{   int   i;
    FLAG  drawable;
    UBYTE glyph = 0; // initialized to prevent spurious SAS/C compiler warnings

    // assert(machine == PIPBUG);

    // derive position from value
    if   (value & 0x80) position = 0;
    elif (value & 0x40) position = 1;
    elif (value & 0x20) position = 2;
    elif (value & 0x10) position = 3;
    else return; // how are we intended to handle the case where multiple bits of the high nybble are set?
    value &= 0xF;
    p_digit[position] = value;
    if (value <= 9)
    {   drawable = TRUE;
        glyph = p_digitglyphs[value];
    } else
    {   drawable = FALSE;
    }

    for (i = 0; i < 13; i++)
    {   drawsegment(position, i, BLACK);
    }

    if (drawcorners)
    {   if (drawable && (glyph & 0x21)) // top left
        {   drawsegment(position, 7, RED);
        } elif (drawunlit)
        {   drawsegment(position, 7, UNLIT);
        }
        if (drawable && (glyph & 0x03)) // top right
        {   drawsegment(position, 8, RED);
        } elif (drawunlit)
        {   drawsegment(position, 8, UNLIT);
        }
        if (drawable && (glyph & 0x70)) // middle left
        {   drawsegment(position, 9, RED);
        } elif (drawunlit)
        {   drawsegment(position, 9, UNLIT);
        }
        if (drawable && (glyph & 0x46)) // middle right
        {   drawsegment(position, 10, RED);
        } elif (drawunlit)
        {   drawsegment(position, 10, UNLIT);
        }
        if (drawable && (glyph & 0x18)) // bottom left
        {   drawsegment(position, 11, RED);
        } elif (drawunlit)
        {   drawsegment(position, 11, UNLIT);
        }
        if (drawable && (glyph & 0x0C)) // bottom right
        {   drawsegment(position, 12, RED);
        } elif (drawunlit)
        {   drawsegment(position, 12, UNLIT);
    }   }

    for (i = 0; i < 7; i++)
    {   if (drawable && (glyph & (1 << i)))
        {   drawsegment(position, i, RED);
	    } elif (drawunlit)
        {   drawsegment(position, i, UNLIT);
}   }   }

EXPORT UBYTE p_readport(int port)
{   TRANSIENT FLAG  handled = TRUE;
    PERSIST   UBYTE t; // PERSISTent for speed

    // assert(machine == PIPBUG);

    switch (port)
    {
    case 5: // Linearisatie reads from ports 5 and 6
    case 6:
        t = (UBYTE) (rand() % 256);
    acase PORTD:
        if
        (   opcode  == 112
         && printer
         && iar     == 0x1424
        )
        {   return 0x80;
        }
    adefault:
        handled = FALSE;
        t = 0;
    }

    logport(port, t, handled, FALSE);
    return t;
}

EXPORT void p_writeport(int port, UBYTE data)
{   TRANSIENT FLAG handled = TRUE;

    switch (port)
    {
    case PORTD:
        if (printer && iar == 0x141A)
        {   ph_cout(TRUE, r[0]);
        } elif (!printer || (iar != 0x1404 && iar != 0x140B))
        {   changed = TRUE;
            p_drawdigit(0, data);
        }
    adefault: // eg, PORTC is written to, and Linearisatie writes to ports 4..6
        handled = FALSE;
    }

    logport(port, data, handled, TRUE);
}

MODULE __inline void local_drawpixel(int x, int y, int colour)
{   if (screen[x][y] != (UBYTE) colour)
    {   screen[x][y] = (UBYTE) colour;
        drawpixel(x, y, colour);
}   }
