// INCLUDES---------------------------------------------------------------

#ifdef __VBCC__
    #define AMIGA
#endif
#ifdef AMIGA
    #include "amiga.h"
#endif
#ifdef _lint
    #define WIN32
    #define _inline
#endif
#ifdef WIN32
    #include "ibm.h"
#endif

#include <stdio.h> // for FILE*

#include "aa.h"
#include "zaccaria.h"

// DEFINES----------------------------------------------------------------

// #define SHOWCHARSET
// if you want to examine the character generator PROM contents (Astro Wars only)
// Character $00 is at top right, $01 is beneath it, etc.

// #define GRIDLINES
// enable the above #define if you want to see character cell outlines
// On WinArcadia, the titlebar will indicate the memory address, contents and colour of the
//  cell (tile) under the pointer.
#ifdef GRIDLINES
    #define G_GRIDCOLOUR DARKBLUE
#endif

// EXPORTED VARIABLES-----------------------------------------------------

EXPORT UBYTE                g_bank[1024],
                            g_dips   = 0x8F,
                            g_paddles;

// IMPORTED VARIABLES-----------------------------------------------------

IMPORT UBYTE                memory[32768],
                            mg_console,
                            OutputBuffer[7],
                            sx[2],
                            sy[2];
IMPORT ASCREEN              screen[BOXWIDTH][BOXHEIGHT];
IMPORT UWORD                console[4],
                            jf[2],
                            mirror_r[32768],
                            mirror_w[32768];
IMPORT ULONG                analog,
                            memflags[32768],
                            swapped;
IMPORT int                  ax[2],
                            ay[2],
                            console_start,
                            console_a,
                            console_b,
                            console_reset,
                            consoleopen,
                            cpl,
                            firstmnemonic,
                            latestsprite,
                            game,
                            key1,
                            key2,
                            key3,
                            key4,
                            lastmnemonic,
                            machine,
                            memmap,
                            offset,
                            p1bgcol[4],
                            p2bgcol[4],
                            p1rumble,
                            p2rumble,
                            p1sprcol[6],
                            p2sprcol[6],
                            pvibase,
                            recmode,
                            rotate,
                            starfield,
                            usehacks,
                            vblank,
                            whichoverlay;
IMPORT UBYTE*               IOBuffer;
IMPORT FILE*                MacroHandle;
#ifdef WIN32
    #ifdef GRIDLINES
        IMPORT HWND         MainWindowPtr;
    #endif
#endif

// MODULE VARIABLES-------------------------------------------------------

// (none)

// MODULE ARRAYS----------------------------------------------------------

MODULE const UBYTE g_colours[4][4] = {
{ BLACK,
  RED,    // 3rd line of status area (exterior), heads of 1st enemy row, wings of 2nd enemy row
  BLUE,   // 3rd line of status area (interior), bodies of 1st..2nd enemy rows
  PURPLE  // 3rd line of status area (sides), wings of 1st enemy row, eyes of 2nd enemy row
}, // $00
{ BLACK,  // unused?
  YELLOW, // wings of 3rd enemy row
  BLUE,   // bodies of 3rd enemy row
  GREEN   // 2nd line of status area, eyes of 3rd enemy row
}, // $01
{ WHITE,  // unused?
  WHITE,  // unused?
  WHITE,  // unused?
  WHITE   // unused?
}, // $02 (unused?)
{ BLACK,  // unused?
  YELLOW, // wings of 4th..6th enemy rows, explosions
  RED,    // bodies of 4th..6th enemy rows, explosions
  WHITE   // 1st line of status area, eyes of 4th..6th enemy rows, explosions
}  // $03
};

/* For Galaxia:
 Explosions use the 14th..16th colours.
 So do the 4th..6th enemy rows.
 Explosion colours are probably wrong.

MODULE FUNCTIONS------------------------------------------------------- */

MODULE __inline void local_drawpixel(int x, int y, int colour);

// CODE-------------------------------------------------------------------

EXPORT void astrowars_setmemmap(void)
{   int i, address, mirror;

    game = FALSE;
    // assert(memmap == MEMMAP_ASTROWARS);

    for (i = 0x1400; i <= 0x7FFF; i++)
    {   memory[i] = 0; // important
    }
    for (i = 0; i <= 0x13FF; i++)
    {   memory[         i] = g_bios[1][         i]; // $0000..$13FF -> $0000..$13FF (  5K)
        memory[0x2000 + i] = g_bios[1][0x1400 + i]; // $1400..$27FF -> $2000..$33FF (  5K)
    }
    for (i = 0; i <= 0x9FF; i++)
    {   memory[0x3400 + i] = g_bios[1][0x2800 + i]; // $2800..$31FF -> $3400..$3DFF (2.5K)
    }
   
    // This is necessary for the boss level to work properly (bad dump?).
    memory[0x2A00] = 0xA4; // $FB -> $A4
    memory[0x2A01] =    1; // $F8 -> $01

    for (address = 0; address < 32 * KILOBYTE; address++)
    {   memflags[address] = 0;
        mirror            = address;

        if
        (   (address >= 0x1600 && address <= 0x17FF)
         || (address >= 0x1D00 && address <= 0x1FFF)
        )
        {   memflags[address] |= NOREAD | NOWRITE;
        }

        if
        (   (address <= 0x13FF)                      // ROM
         || (address >= 0x2000 && address <= 0x33FF) // ROM
        )
        {   memflags[address] |= NOWRITE;
        }

        if  (address >= 0x1800 && address <= 0x1BFF) // ROM
        {   memflags[address] |= VISIBLE | BANKED;
        }

        if (address >= 0x3400 && address <= 0x3FFF)
        {   mirror = address & 0x1FFF;
        }

        if (mirror != address)
        {   memflags[address] |= MIRRORED_R | MIRRORED_W;
        }
        mirror_r[address] =
        mirror_w[address] = (UWORD) mirror;
    }

    pvibase = 0x1500;
    pvi_memmap();

    whichoverlay  = G_OVERLAY;
    firstmnemonic =
    lastmnemonic  = -1;
    starfield     = BLACK;
    key1          =
    key2          =
    key3          =
    key4          = 2;
    analog        =
    swapped       = FALSE;

    for (i = 0; i < 4; i++)
    {   p1bgcol[i] =
        p2bgcol[i] = 0;
    }
    for (i = 0; i < 6; i++)
    {   p1sprcol[i] =
        p2sprcol[i] = 0;
    }

    vblank        = PVI_VBLANK; // 271..273 are no good, with cpl of 107!
    latestsprite  = PVI_LATESTSPRITE;
    cpl           = 107; /* 105..108 are no good, with vblank of 271!
       2,000KHz = 2,000,000 cycles per second (2MHz is just a guess)
       2,000,000 / 60 = 33,333'
       For Galaxia: 33,333 / 312 = 106.8365

       What was happening was that we were generating a sprite-complete
       interrupt at rastline 271. So, when the vblank occurred, we were
       still in the sprite-complete interrupt handler; therefore, the
       vblank interrupt did not occur because interrupts were inhibited.
       And therefore, we could not escape from the tight loop at $27C..
       $280 and thus the game would hang. Adjusting the vblank or CPL
       would fix this. We decided instead not to generate interrupts for
       sprite-complete events when emulating Galaxia. */
}

EXPORT void galaxia_setmemmap(void)
{   int i, address, mirror;

    game = FALSE;
    // assert(memmap == MEMMAP_GALAXIA);
    
    for (i = 0x1400; i <= 0x7FFF; i++)
    {   memory[i] = 0; // important
    }
    for (i = 0; i <= 0x13FF; i++)
    {   memory[         i] = g_bios[0][         i]; // $0000..$13FF -> $0000..$13FF (  5K)
        memory[0x2000 + i] = g_bios[0][0x1400 + i]; // $1400..$27FF -> $2000..$33FF (  5K)
    }
    for (i = 0; i <= 0x9FF; i++)
    {   memory[0x3400 + i] = g_bios[0][0x2800 + i]; // $2800..$31FF -> $3400..$3DFF (2.5K)
    }

    if (usehacks)
    {   memory[0xA26] =
        memory[0xA27] =
        memory[0xAAF] =
        memory[0xAB0] = 0xC0; // NOP
    }

    for (address = 0; address < 32 * KILOBYTE; address++)
    {   memflags[address] = 0;
        mirror            = address;

        if
        (   (address <= 0x13FF)                      // ROM
         || (address >= 0x2000 && address <= 0x33FF) // ROM
        )
        {   memflags[address] |= NOWRITE;
        }

        if  (address >= 0x1800 && address <= 0x1BFF) // ROM
        {   memflags[address] |= VISIBLE | BANKED;
        }

        if (address >= 0x3400 && address <= 0x3FFF)
        {   mirror = address & 0x1FFF;
        }

        if (mirror != address)
        {   memflags[address] |= MIRRORED_R | MIRRORED_W;
        }
        mirror_r[address] =
        mirror_w[address] = (UWORD) mirror;
    }

    pvibase = 0x1500;
    pvi_memmap();
    pvibase = 0x1600;
    pvi_memmap();
    pvibase = 0x1700;
    pvi_memmap();
    pvibase = 0x1500;

    whichoverlay  = G_OVERLAY;
    firstmnemonic =
    lastmnemonic  = -1;
    starfield     = BLACK;
    key1          =
    key2          =
    key3          =
    key4          = 2;
    analog        =
    swapped       = FALSE;

    for (i = 0; i < 4; i++)
    {   p1bgcol[i] =
        p2bgcol[i] = 0;
    }
    for (i = 0; i < 6; i++)
    {   p1sprcol[i] =
        p2sprcol[i] = 0;
    }

    vblank        = PVI_VBLANK; // 271..273 are no good, with cpl of 107!
    latestsprite  = PVI_LATESTSPRITE;
    cpl           = 107; /* 105..108 are no good, with vblank of 271!
       2,000KHz = 2,000,000 cycles per second (2MHz is just a guess)
       2,000,000 / 60 = 33,333'
       For Galaxia: 33,333 / 312 = 106.8365

       What was happening was that we were generating a sprite-complete
       interrupt at rastline 271. So, when the vblank occurred, we were
       still in the sprite-complete interrupt handler; therefore, the
       vblank interrupt did not occur because interrupts were inhibited.
       And therefore, we could not escape from the tight loop at $27C..
       $280 and thus the game would hang. Adjusting the vblank or CPL
       would fix this. We decided instead not to generate interrupts for
       sprite-complete events when emulating Galaxia. */
}

EXPORT void astrowars_drawscreen(void)
{   PERSIST   int a, aa, // X-axis from guest's viewpoint, inverted Y-axis from user's viewpoint
                  b, bb, // Y-axis from guest's viewpoint,          X-axis from user's viewpoint
                  contents,
                  localoffset,
                  starta,
                  startb;
#ifdef SHOWCHARSET
    TRANSIENT int whichcontents = 0;
#endif

    // assert(machine == ASTROWARS);

    // Each character is 8*8 pixels (8 bytes).
    // Character PROM is 2K (with 256 characters, each of 8 bytes).

#ifdef SHOWCHARSET
    contents = 0;
    for (b = 2; b < 32; b++)
    {   for (a = 0; a < 30; a++)
        {   starta = a << 3;
            startb = (b - 2) << 3;

            if (a >= 8 && a <= 23 && b >= 8 && b <= 23)
            {   contents = whichcontents++;
            } else
            {   contents = 239;
            }
            localoffset = contents * 8;
            for (aa = 7; aa >= 0; aa--)
            {   for (bb = 0; bb < 8; bb++)
                {   if (g_x_tiles[localoffset + bb] & (128 >> aa))
                    {   local_drawpixel(starta + aa, startb + bb, WHITE);
                    } else
                    {   local_drawpixel(starta + aa, startb + bb, BLACK);
    }   }   }   }   }
#else
    for (b = 2; b < 32; b++)        // Y-axis from guest's viewpoint,          X-axis from user's viewpoint
    {   for (a = 0; a < 30; a++)    // X-axis from guest's viewpoint, inverted Y-axis from user's viewpoint
        {   starta = a * 8;         // from guest's viewpoint, X-start of this character (ie. left edge)
            startb = (b - 2) * 8;   // from guest's viewpoint, Y-start of this character (ie. top  edge)
            contents = memory[0x1800 + (b * 32) + a];

            localoffset = contents * 8;
            for (aa = 7; aa >= 0; aa--)
            {   for (bb = 0; bb < 8; bb++)
                {   if (g_x_tiles[localoffset + bb] & (128 >> aa))
                    {   if (screen[starta + aa][startb + bb] != (g_bank[(b * 32) + a] & 7))
                        {   local_drawpixel(starta + aa, startb + bb, g_bank[(b * 32) + a] & 7);
                    }   }
                    else
                    {
#ifdef GRIDLINES
                        if (aa == 7 || aa == 0 || bb == 0 || bb == 7)
                        {   if (screen[starta + aa][startb + bb] != G_GRIDCOLOUR)
                            {   local_drawpixel(starta + aa, startb + bb, G_GRIDCOLOUR);
                        }   }
                        else
#endif
                        if (screen[starta + aa][startb + bb] != BLACK)
                        {   local_drawpixel(starta + aa, startb + bb, BLACK);
    }   }   }   }   }   }
#endif
}

EXPORT void galaxia_drawscreen(void)
{   PERSIST int a, aa, // X-axis from guest's viewpoint, inverted Y-axis from user's viewpoint
                b, bb, // Y-axis from guest's viewpoint,          X-axis from user's viewpoint
                contents,
                starta,
                startb;

    // assert(machine == GALAXIA);
    for (b = 2; b < 32; b++)
    {   for (a = 0; a < 30; a++)
        {   starta = (int) (a << 3);
            startb = (int) ((b - 2) << 3);
            contents = (memory[0x1800 + (b * 32) + a] & 0x7F) << 3;
            for (aa = 7; aa >= 0; aa--)
            {   for (bb = 0; bb < 8; bb++)
                {   if (g_w_tiles1[contents + bb] & (128 >> aa))
                    {   if (g_w_tiles2[contents + bb] & (128 >> aa))
                        {   if (screen[starta + aa][startb + bb] != g_colours[g_bank[(b * 32) + a] & 3][3])
                            {   local_drawpixel(starta + aa, startb + bb, g_colours[g_bank[(b * 32) + a] & 3][3]);
                        }   }
                        else
                        {   if (screen[starta + aa][startb + bb] != g_colours[g_bank[(b * 32) + a] & 3][2])
                            {   local_drawpixel(starta + aa, startb + bb, g_colours[g_bank[(b * 32) + a] & 3][2]);
                    }   }   }
                    else
                    {   if (g_w_tiles2[contents + bb] & (128 >> aa))
                        {   if (screen[starta + aa][startb + bb] != g_colours[g_bank[(b * 32) + a] & 3][1])
                            {   local_drawpixel(starta + aa, startb + bb, g_colours[g_bank[(b * 32) + a] & 3][1]);
                        }   }
                        else
                        {
#ifdef GRIDLINES
                            if (aa == 7 || aa == 0 || bb == 0 || bb == 7)
                            {   if (screen[starta + aa][startb + bb] != G_GRIDCOLOUR)
                                {   local_drawpixel(starta + aa, startb + bb, G_GRIDCOLOUR);
                            }   }
                            else
#endif
                            if (screen[starta + aa][startb + bb] != BLACK)
                            {   local_drawpixel(starta + aa, startb + bb, BLACK); // g_colours[g_bank[(b * 32) + a] & 3][0]);
}   }   }   }   }   }   }   }

EXPORT UBYTE g_readport(int port)
{   TRANSIENT FLAG  handled = TRUE;
    PERSIST   UBYTE t; // PERSISTent for speed

    // assert(machine == GALAXIA);

    switch (port)
    {
    case 0: // both read this port
        if (machine == GALAXIA)
        {   t = 0;
        } else
        {   // assert(machine == ASTROWARS);
            t = mg_console;
        }
    acase 1:
    case 3:
        t = 0xFF;
    acase 2: // both read this port
        // bit  7    "freeze" (active low)
        // bit  6    doesn't matter?
        // bit  5    is p1 right
        // bit  4    is p1 left
        // bit  3    is p2 right
        // bit  2    is p2 left
        // bit  1    doesn't matter?
        // bit  0    0=multi-firebutton cabinet, 1=single-firebutton cabinet
        //  (currently we only emulate the multi-firebutton cabinet)
        t = (UBYTE) ((g_dips & 0x80) | g_paddles);
    acase 5:
        t = 1;
        // for Astro Wars:
        // bits 7..1 are unused
        // bit  0    enables collisions
    acase 6:
        // bits 7..6 are unknown
        // bits 5..4 are used but unknown (difficulty?)
        // bit  3    is  no. of lives (clear=3, set=5)
        // bit  2    is  B coinage ( %1 is best)
        // bits 1..0 are A coinage (%11 is best)
        t = (UBYTE) (g_dips & 0x7F);
    acase 7:
        // for Galaxia:
        //  bits 7..5 are unknown
        //  bits 4..3 are used but unknown
        //  bits 2..1 are used for hiscore randomization (at least)
        //  bit  0    is whether a hiscore is randomly generated
        // for Astro Wars: all bits are unknown, but it does hiscore
        //  randomization in an apparently similar manner.
        t = 0;
    acase 0xAC: // Galaxia reads this port, Astro Wars doesn't seem to do so
        // bits 7..6 don't matter?
        // bit  5    is 1P fire
        // bit  4    is 2P fire
        // bit  3    is coin slot A
        // bit  2    is coin slot B
        // bit  1    is 2P start
        // bit  0    is 1P start
        // assert(memmap == MEMMAP_W);
        t = mg_console;
    acase PORTC: // used for collision detection
        if (machine == GALAXIA)
        {   // bit 7: whether you want to die all the time
            // bit 4: would you like to die at the end of each level?
            t = 0x6F; // probably isn't correct but seems to work
        } else
        {   // assert(machine == ASTROWARS);
            // setting bit 1 is causing problems
            // bit 0 enables enemy deaths
            t = 1;
        }
    acase PORTD:
        t = 0xFF; // important for Galaxia
    adefault: // eg. port $05 is read by both games
        handled = FALSE;
        t = 0;
    }

    logport(port, t, handled, FALSE);
    return t;
}

EXPORT void g_writeport(int port, UBYTE data)
{   // Galaxia and Astro Wars both write to both the control and data ports (PORTC and PORTD).

    // ioport[port] = data;
    logport(port, data, FALSE, TRUE);
}

EXPORT void g_emuinput(void)
{   PERSIST int stale[4] = { 0, 0, 0, 0 };
#ifdef GRIDLINES
    TRANSIENT int   x, y;
    TRANSIENT TEXT  tempstring[80 + 1];
#endif

    readkybd();
    ReadJoystick(0);
    ReadJoystick(1);

    if (recmode == RECMODE_PLAY)
    {   mg_console     = IOBuffer[offset++];
        g_paddles      = IOBuffer[offset++];

        if (((g_paddles & 0x20) >> 1) == (g_paddles & 0x10))
        {   ax[0] = sx[0] = 128;
        } elif (g_paddles & 0x20) // bit 5 is p1 right
        {   ax[0] = sx[0] = 255;
        } else
        {   // assert(g_paddles & 0x10); // bit 4 is p1 left
            ax[0] = sx[0] =   0;
        }
        if (((g_paddles & 8) >> 1) == (g_paddles & 4))
        {   ax[1] = sx[1] = 128;
        } elif (g_paddles & 8) // bit 3 is p2 right
        {   ax[1] = sx[1] = 255;
        } else
        {   // assert(g_paddles & 4); // bit 2 is p2 left
            ax[1] = sx[1] =   0;
        }
        ay[0] = ay[1] = 128;
        sy[0] = sy[1] = 128;
    } else
    {   mg_console =
        g_paddles  = 0;

        if (swapped)
        {   mg_playerinput(1, 0);
            mg_playerinput(0, 1);
        } else
        {   mg_playerinput(0, 0);
            mg_playerinput(1, 1);
        }
        p1rumble =
        p2rumble = 0;

        if (KeyDown(console[0]) || (jf[0] & JOYSTART) || (jf[1] & JOYSTART) || console_start)
        {   if (!stale[0])
            {   mg_console |= 1; // 1P
            }
            stale[0] = 1;
        } else
        {   stale[0] = 0;
        }

        if (KeyDown(console[1]) || (jf[0] & JOYA    ) || (jf[1] & JOYA    ) || console_a    )
        {   if (!stale[1])
            {   mg_console |= 2; // 2P
            }
            stale[1] = 1;
        } else
        {   stale[1] = 0;
        }

        if (KeyDown(console[2]) || (jf[0] & JOYB    ) || (jf[1] & JOYB    ) || console_b    )
        {   if (!stale[2])
            {   mg_console |= 8; // generous coin slot A (1C, 2C, 3C or 5C)
            }
            stale[2] = 1;
        } else
        {   stale[2] = 0;
        }

        if (KeyDown(console[3]) || console_reset)
        {   if (!stale[3])
            {   mg_console |= 4; // jewish coin slot B (C or 1C)
            }
            stale[3] = 1;
        } else
        {   stale[3] = 0;
        }

        domouse();
        // order-dependent re. the pviwrite(0x7214) below!

        if (recmode == RECMODE_RECORD)
        {   OutputBuffer[0] = mg_console;
            OutputBuffer[1] = g_paddles;

            DISCARD fwrite(OutputBuffer, 2, 1, MacroHandle);
            // should really check return code of fwrite()
            // 2 bytes per frame * 60 frames per second = 120 bytes/sec.
    }   }

    pviwrite(0x7214, mg_console, FALSE); // definitely needed for Galaxia (at least)

#ifdef WIN32
    #ifdef GRIDLINES
        // Unfortunately, this doesn't update while paused.

        getpointerpos(&x, &y);
        x /= 2184; // scale from 0..65535 to 0..29
        y /= 2184; // scale from 0..65535 to 0..29
        if (x >= 0 && x < 30 && y >= 0 && y < 30)
        {    sprintf
             (   tempstring, "$%04X: contents $%02X, colour $%02X",
                 0x1840 + ((29 - x) * 32) + y,
                 memory[0x1840 + ((29 - x) * 32) + y],
                 g_bank[0x40 + ((29 - x) * 32) + y]
             );
        } else
        {    strcpy(tempstring, "-,-");
        }
        DISCARD SetWindowText(MainWindowPtr, tempstring);
    #endif
#endif

    if (console_start) console_start--;
    if (console_a    ) console_a--;
    if (console_b    ) console_b--;
    if (console_reset) console_reset--;
}

EXPORT void g_viewscreen(void)
{   TEXT rowchars[16 + 1];
    int  where,
         x, y;

    rowchars[16] = 0;
    T1; textpen(TEXTPEN_VIEW);
    // Ideally, we would show the text in the correct colours

    if (!rotate)
    {   for (where = 0x1800; where < 0x1C00; where += 16)
        {   dumprow(where);
            if (where % 32)
            {   zprintf("\n");
    }   }   }
    else
    {   zprintf("---Left Half---\n");
        for (y = 0; y < 32; y++)
        {   where = 0x1BE0 + y;
            for (x = 0; x < 16; x++)
            {   if (machine == GALAXIA)
                {   rowchars[x] = galaxia_chars[memory[where - (x * 0x20)] & 0x7F];
                } else
                {   // assert(machine == ASTROWARS);
                    rowchars[x] = astrowars_chars[memory[where - (x * 0x20)]];
            }   }
            zprintf
            (   "$%04X: %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %s\n",
                where,
                memory[where        ],
                memory[where -  0x20],
                memory[where -  0x40],
                memory[where -  0x60],
                memory[where -  0x80],
                memory[where -  0xA0],
                memory[where -  0xC0],
                memory[where -  0xE0],
                memory[where - 0x100],
                memory[where - 0x120],
                memory[where - 0x140],
                memory[where - 0x160],
                memory[where - 0x180],
                memory[where - 0x1A0],
                memory[where - 0x1C0],
                memory[where - 0x1E0],
                rowchars
            );
        }
        zprintf("---Right Half---\n");
        for (y = 0; y < 32; y++)
        {   where = 0x19E0 + y;
            for (x = 0; x < 16; x++)
            {   if (machine == GALAXIA)
                {   rowchars[x] = galaxia_chars[memory[where - (x * 0x20)] & 0x7F];
                } else
                {   // assert(machine == ASTROWARS);
                    rowchars[x] = astrowars_chars[memory[where - (x * 0x20)]];
            }   }
            zprintf
            (   "$%04X: %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %s\n",
                where,
                memory[where        ],
                memory[where -  0x20],
                memory[where -  0x40],
                memory[where -  0x60],
                memory[where -  0x80],
                memory[where -  0xA0],
                memory[where -  0xC0],
                memory[where -  0xE0],
                memory[where - 0x100],
                memory[where - 0x120],
                memory[where - 0x140],
                memory[where - 0x160],
                memory[where - 0x180],
                memory[where - 0x1A0],
                memory[where - 0x1C0],
                memory[where - 0x1E0],
                rowchars
            );
        }
        zprintf("\n");
    }

    T2;
}

EXPORT void g_viewcolours(void)
{   int where,
        y;

    /* We should show this according to the rotation flag
       (currently it's always rotated). */

    T1; textpen(TEXTPEN_VIEW);

    if (rotate)
    {   zprintf("---Left Half---\n");
        for (y = 0; y < 32; y++)
        {   where = 0x3E0 + y;
            zprintf
            (   "$%04X: %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X\n",
                where + 0x1800,
                g_bank[where        ],
                g_bank[where -  0x20],
                g_bank[where -  0x40],
                g_bank[where -  0x60],
                g_bank[where -  0x80],
                g_bank[where -  0xA0],
                g_bank[where -  0xC0],
                g_bank[where -  0xE0],
                g_bank[where - 0x100],
                g_bank[where - 0x120],
                g_bank[where - 0x140],
                g_bank[where - 0x160],
                g_bank[where - 0x180],
                g_bank[where - 0x1A0],
                g_bank[where - 0x1C0],
                g_bank[where - 0x1E0]
            );
        }
        zprintf("---Right Half---\n");
        for (y = 0; y < 32; y++)
        {   where = 0x1E0 + y;
            zprintf
            (   "$%04X: %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X\n",
                where + 0x1800,
                g_bank[where        ],
                g_bank[where -  0x20],
                g_bank[where -  0x40],
                g_bank[where -  0x60],
                g_bank[where -  0x80],
                g_bank[where -  0xA0],
                g_bank[where -  0xC0],
                g_bank[where -  0xE0],
                g_bank[where - 0x100],
                g_bank[where - 0x120],
                g_bank[where - 0x140],
                g_bank[where - 0x160],
                g_bank[where - 0x180],
                g_bank[where - 0x1A0],
                g_bank[where - 0x1C0],
                g_bank[where - 0x1E0]
            );
        }
        zprintf("\n");
    } else
    {   for (y = 0; y < 64; y++)
        {   where = y * 16;
            zprintf
            (   "$%04X: %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %s\n",
                where + 0x1800,
                g_bank[where     ],
                g_bank[where +  1],
                g_bank[where +  2],
                g_bank[where +  3],
                g_bank[where +  4],
                g_bank[where +  5],
                g_bank[where +  6],
                g_bank[where +  7],
                g_bank[where +  8],
                g_bank[where +  9],
                g_bank[where + 10],
                g_bank[where + 11],
                g_bank[where + 12],
                g_bank[where + 13],
                g_bank[where + 14],
                g_bank[where + 15]
            );
            if (y % 2)
            {   zprintf("\n");
    }   }   }

    T2;
}

MODULE __inline void local_drawpixel(int x, int y, int colour)
{   if (screen[x][y] != colour)
    {   screen[x][y] = (UBYTE) colour;
        drawpixel(x, y, colour);
}   }
