/* engine1.c: Miscellaneous platform-independent code

INCLUDES--------------------------------------------------------------- */

#ifdef __VBCC__
    #define AMIGA
#endif
#ifdef AMIGA
    #include "amiga.h"
#endif
#ifdef _lint
    #define WIN32
    #define _inline
#endif
#ifdef WIN32
    #include "ibm.h"
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include "aa.h"
#define CATCOMP_NUMBERS
#define CATCOMP_BLOCK
#ifdef WIN32
    #define EXEC_TYPES_H
    #include "aa_strings.h"
    #undef EXEC_TYPES_H
#endif
#ifdef AMIGA
    #include "aa_strings.h"
    #include <intuition/intuition.h>
    #define ALL_REACTION_CLASSES
    #define ALL_REACTION_MACROS
    #include <reaction/reaction.h>
    #include <proto/gadtools.h>
    #include <proto/intuition.h>
    #include <proto/locale.h>        // GetCatalogStr()

    #if defined(__SASC) || defined(__STORM__) || defined(__STORMGCC__) || defined(__VBCC__)
        #define BASE_PAR_DECL

        struct hostent* gethostbyname(const UBYTE* name);
        LONG            recv(LONG s, UBYTE* buf, LONG len, LONG flags); /* V3 */
        LONG            closesocket(LONG d);
        LONG            connect(LONG s, const struct sockaddr* name, LONG namelen);
        LONG            send(LONG s, const UBYTE* msg, LONG len, LONG flags);
        LONG            socket(LONG domain, LONG type, LONG protocol);
        LONG            bind(LONG s, const struct sockaddr* name, LONG namelen);
        LONG            listen(LONG s, LONG backlog);
        LONG            accept(LONG s, struct sockaddr* addr, LONG* addrlen);
        LONG            ioctlsocket(BASE_PAR_DECL LONG d, ULONG request, char* argp);
        LONG            gethostname(STRPTR hostname, LONG size); /* V3 */

        #ifndef __PPC__
        #pragma libcall SocketBase socket         1E 21003
        #pragma libcall SocketBase bind           24 18003
        #pragma libcall SocketBase listen         2A 1002
        #pragma libcall SocketBase accept         30 98003
        #pragma libcall SocketBase connect        36 18003
        #pragma libcall SocketBase send           42 218004
        #pragma libcall SocketBase recv           4E 218004
        #pragma libcall SocketBase ioctlsocket    72 81003
        #pragma libcall SocketBase closesocket    78 001
        #pragma libcall SocketBase gethostbyname  D2 801
        #pragma libcall SocketBase gethostname   11A 0802
        #endif

        #define htonl(x)   (x)
        #define htons(x)   (x)

        struct in_addr
        {   ULONG          s_addr;
        };
        struct sockaddr_in
        {   UBYTE          sin_len;
            UBYTE          sin_family;
            USHORT         sin_port;
            struct in_addr sin_addr;
            char           sin_zero[8];
        };
    #else
        #include <netinet/in.h>
        #ifndef __MORPHOS__
            #include <proto/bsdsocket.h>
        #else
            #include <proto/socket.h>
        #endif
        #define closesocket(d) CloseSocket(d)
        #define ioctlsocket(d, request, argp) IoctlSocket(d, request, argp)
    #endif

    #ifndef __amigaos4__
        struct hostent
        {   char*          h_name;
            char**         h_aliases;
            int            h_addrtype;
            int            h_length;
            char**         h_addr_list;
    #define h_addr         h_addr_list[0] /* address, for backward compatiblity */
        };
    #endif

    #define SOCK_STREAM               1 /* stream socket */
    #define AF_INET                   2 /* internetwork: UDP, TCP, etc. */
    #define _IOC(inout,group,num,len) (inout | ((len & IOCPARM_MASK) << 16) | ((group) << 8) | (num))
    #define _IOW(g,n,t)               _IOC(IOC_IN, (g), (n), sizeof(t))
    #define FIONBIO                   _IOW('f', 126, long) /* set/clear non-blocking i/o */
//  #define EAGAIN                    35        /* Resource temporarily unavailable */
    #define EWOULDBLOCK               EAGAIN        /* Operation would block */
    #define WSAEWOULDBLOCK            EWOULDBLOCK
    #if defined(__SASC) || defined(__VBCC__)
        typedef unsigned long         u_long;
    #endif
    typedef u_long                    n_long; /* long as received from the net */
    #ifndef INADDR_ANY
        #define INADDR_ANY            (u_long)0x00000000
    #endif
    #define IOCPARM_MASK              0x1fff /* parameter length, at most 13 bits */
    #define IOC_IN                    0x80000000 /* copy in parameters */
#endif

#ifdef WIN32
    #include <commctrl.h>
    #include "resource.h"
#endif

#include "engine1.h"
#include "games.h"

// DEFINES----------------------------------------------------------------

// #define THERULER
// whether you want guide pixels every 10th column
// RULER is a reserved word (already #defined by something), at least for wa!

// #define QUICKSORT
// whether to use quicksort instead of bubble sort (but this can mix up the order of "equal" items)

#define LOGHANDLEDREADS
#define LOGHANDLEDWRITES
// whether to log all I/O port accesses (otherwise it only logs unhandled accesses)

#define HISCORESLENGTH (1 + (18 * HISCORES))
#ifdef AMIGA
    #define FN_HISCORES "PROGDIR:Configs/AmiArcadia.hgh"
#endif
#ifdef WIN32
    #define FN_HISCORES "CONFIGS\\WA.HGH"
#endif

// MODULE VARIABLES-------------------------------------------------------

MODULE       TEXT   ggstring[80 + 1],
                    fn_anim[MAX_PATH + 1] = "";
MODULE       int    Socket                = -1;
MODULE       ULONG  crcTable[256];

// EXPORTED VARIABLES-----------------------------------------------------

// variables which correspond to ReAction gadgets must be ULONGs
EXPORT ULONG  analog           = FALSE,
              autofire         = FALSE,
              collisions       = TRUE,
              swapped          = FALSE,
              demultiplex      = FALSE,
              flagline         = TRUE,
              fps              = 0,
              paused           = FALSE,
              region           = REGION_NTSC,
              rhythm           = 0,
              sound            = TRUE,
              turbo            = FALSE;
// other ULONGs go here
EXPORT ULONG  downframes       = 4,
              frames,
              totalframes      = 16;
EXPORT TEXT   friendly[21 + 1],
           // "$19FD (SPRITECOLLIDE)"
           //  0123456789012345678901
              fromstr[5 + 1 + 1],
              tostr[5 + 1 + 1],
              lastchange[32 + 1 + 1],
              filepart[MAX_PATH + 1],
              pathpart[MAX_PATH + 1],
              fn_game[MAX_PATH + 1] = "", // complete pathname of the game
              thequeue[QUEUESIZE];
EXPORT FILE  *MacroHandle      = NULL;
EXPORT int    anims            = FALSE,
              autopause        = FALSE,
              autosave         = TRUE,
              ax[2],
              ay[2],
              cy_bgc,
              cy_fgc,
              cli              = FALSE,
              colourset        = COLOURFUL,
              configmachine,
              configversion    = NEWCONFIGVERSION,
              confine          = FALSE,
              confirmable      = FALSE,
              connected        = NET_OFF,
              console_start    = FALSE,
              console_a        = FALSE,
              console_b        = FALSE,
              console_reset    = FALSE,
              consoleopen      = FALSE,
              cpl              = PVI_CPL,
              crippled         = FALSE,
              dotspace         = FALSE,
              echoconsole      = FALSE,
              echoprinter      = FALSE,
              enhancestars     = TRUE,
              filesize,
              firstmnemonic    = -1, // important!
              frameskip        = 1,
              fromnum,
              foundgames       = 0,
              fullscreen       = FALSE,
              game             = FALSE,
              wsc              = 2, // white
              icons            = FALSE,
              inframe          = FALSE,
              inverse          = FALSE,
              lastmnemonic     = -1, // important!
              latestsprite     = PVI_LATESTSPRITE,
              loader           = TRUE,
              logbios          = FALSE,
              logeprom         = FALSE,
              logfile          = FALSE,
              loginstructions  = FALSE,
              logioports       = FALSE,
              logreadwrites    = FALSE,
              logsubroutines   = FALSE,
              logwrites        = FALSE,
              loop             = TRUE,
              lowercase        = FALSE,
              machine          = ARCADIA,
              memmap           = MEMMAP_ARCADIA,
              offset,
              p1rumble         = 0,
              p2rumble         = 0,
              pauselog         = FALSE,
              rast,
              rastlines        = FALSE,
              recmode          = RECMODE_NORMAL,
              recsize,
              requirebutton    = FALSE,
              retune           = FALSE,
              rotate           = TRUE,
              scale2x          = FALSE,
              sensegame        = TRUE,
              sensitivity      = SENSITIVITY_DEFAULT,
              shadowing        = FALSE,
              showmenubar      = TRUE,
              showsidebar      = TRUE,
              showtitlebar     = TRUE,
              showtoolbar      = TRUE,
              showpointer      = TRUE,
              size             = 1,
              sortby           = SORTBYNAME,
              spritesvisible   = TRUE,
              step             = FALSE,
              stretch          = TRUE,
              style            = STYLE_SIGNETICS,
              supercpu         = 0,
              tonum,
              tolimit,
              trace            = FALSE,
              traceverbose     = TRUE,
              undither         = FALSE,
              useansi          = TRUE,
              useguidepixel    = TRUE,
              usehacks         = TRUE,
              usespeech        = FALSE,
              vblank           = PVI_VBLANK,
              watchreads       = FALSE,
              wheremouse       = -1,
              whichgame        = -1,
              whichpen         = 0,
              wide             = 2;
EXPORT UBYTE* IOBuffer         = NULL;
EXPORT UBYTE  oldkeys[KEYS],
              oldflag,
              oldglow,
              oldrun,
              OutputBuffer[7],
              newkeys[KEYS],
              startaddr_h,
              startaddr_l,
              sx[2],
              sy[2];
EXPORT UWORD  console[4],
              joystate,
              keypads[2][NUMKEYS];
EXPORT FLAG   belmach          = FALSE,
              belmach0         = FALSE,
              bps              = FALSE,
              fiftytheme       = FALSE,
              guestplaying[14] = {FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE,
                                  FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE},
              p_morse          = FALSE, // may not require initialization
              piano            = FALSE,
              music            = FALSE, // may not require initialization
              wps              = FALSE;
EXPORT int                  sidebar[FOUNDGAMES_MAX];
EXPORT struct AuditStruct   auditlist[FOUNDGAMES_MAX];
EXPORT STRPTR               ggnode = ""; // must be initialized
EXPORT signed int           dx, dy;

#ifdef AMIGA
    EXPORT FILE*  PrinterHandle = NULL;
#endif
#ifdef WIN32
    EXPORT HANDLE PrinterHandle = NULL;
#endif

EXPORT const struct MachineStruct machines[MACHINES] =
{ { "ARCADIA",     A_BOXWIDTH,     A_BOXHEIGHT,      MEMMAP_ARCADIA,   REGION_NTSC, 2, A_OVERLAY,      "a",   0,  28, 40,  8, MSG_HAIL_MONITOR_A,      "Emerson Arcadia 2001 Monitor"                , 0xFF, 1 },
  { "INTERTON",   IE_BOXWIDTH,    IE_BOXHEIGHT,      MEMMAP_D,         REGION_PAL , 2, I_OVERLAY,      "i",  29,  69, 40,  4, MSG_HAIL_MONITOR_I,      "Interton VC 4000 Monitor"                    ,    0, 1 },
  { "ELEKTOR",    IE_BOXWIDTH,    IE_BOXHEIGHT,      MEMMAP_F,         REGION_PAL , 2, E_OVERLAY,      "e",  29, 324, 36,  4, MSG_HAIL_MONITOR_E_F,    "Elektor TV Games Computer (Expanded) Monitor",    0, 1 },
  { "PIPBUG",      P_BOXWIDTH,     P_BOXHEIGHT,      MEMMAP_P,         REGION_PAL , 2, P_OVERLAY,      "p", 553, 655,  0,  0, MSG_HAIL_MONITOR_P,      "PIPBUG-Based Machine Monitor"                ,    0, 1 },
  { "INSTRUCTOR",  S_BOXWIDTH,     S_BOXHEIGHT,      MEMMAP_O,         REGION_PAL , 1, S_OVERLAY,      "s", 656, 666, 28,  0, MSG_HAIL_MONITOR_S,      "Signetics Instructor 50 Monitor"             ,    0, 1 },
  { "CD2650",      C_BOXWIDTH,     C_BOXHEIGHT,      MEMMAP_CD2650,    REGION_NTSC, 1, C_OVERLAY,      "c", 667, 729,  0,  0, MSG_HAIL_MONITOR_C,      "Central Data 2650 Monitor"                   ,    0, 1 },
  { "ASTROWARS",   G_BOXWIDTH,     G_BOXHEIGHT,      MEMMAP_ASTROWARS, REGION_NTSC, 1, G_OVERLAY,      "s", 325, 356,  5,  4, MSG_HAIL_MONITOR_G_X,    "Astro Wars Monitor"                          ,    0, 2 },
  { "GALAXIA",     G_BOXWIDTH,     G_BOXHEIGHT,      MEMMAP_GALAXIA,   REGION_NTSC, 1, G_OVERLAY,      "s", 325, 423,  5, 12, MSG_HAIL_MONITOR_G,      "Galaxia Monitor"                             ,    0, 2 },
  { "MALZAK",      M_BOXWIDTH,     M_BOXHEIGHT,      MEMMAP_MALZAK1,   REGION_PAL , 1, A_OVERLAY,      "s", 325, 390, 11,  8, MSG_HAIL_MONITOR_M,      "Malzak Monitor"                              ,    0, 1 },
  { "CHAOS",       H_BOXWIDTH,     H_BOXHEIGHT,      MEMMAP_CHAOS,     REGION_PAL,  1, A_OVERLAY,      "s", 424, 538,  0,  0, MSG_HAIL_MONITOR_H,      "Chaos 2 Monitor"                             ,    0, 2 },
  { "DOLPHIN",     D_BOXWIDTH,     D_BOXHEIGHT,      MEMMAP_DOLPHIN,   REGION_PAL,  1, A_OVERLAY,      "s", 539, 552, 18,  0, MSG_HAIL_MONITOR_D,      "Dolphin Monitor"                             ,    0, 2 },
  { "PHUNSY",     PHUNSY_BOXWIDTH, PHUNSY_BOXHEIGHT, MEMMAP_PHUNSY,    REGION_PAL,  1, PHUNSY_OVERLAY, "s", 730, 961,  0,  0, MSG_HAIL_MONITOR_PHUNSY, "PHUNSY Monitor"                              ,    0, 3 }
};

EXPORT const ULONG defpencolours[2][3][GUESTCOLOURS] = {
{ { 0x00FFFFFF, // white  (UVI #0, PVI #0)
    0x00FFFF00, // yellow (UVI #1, PVI #1)
    0x00FF00FF, // purple (UVI #4, PVI #2)
    0x00FF0000, // red    (UVI #5, PVI #3)
    0x0000FFFF, // cyan   (UVI #2, PVI #4)
    0x0000DD00, // green  (UVI #3, PVI #5)
    0x004444FF, // blue   (UVI #6, PVI #6)
    0x00000000, // black  (UVI #7, PVI #7)
    0x00BBBBBB, // grey #1
    0x00BBBB00, // dark yellow
    0x00BB00BB, // dark purple
    0x00990000, // dark red
    0x0000AAAA, // dark cyan
    0x0000BB00, // dark green
    0x000000AA, // dark blue
    0x00000000, // dark black
    0x00251300, // grey #2 (amber)
    0x00492500, // grey #3 (amber)
    0x006D3700, // grey #4 (amber)
    0x00924900, // grey #5 (amber)
    0x00B65B00, // grey #6 (amber)
    0x00DB6E00, // grey #7 (amber)
    0x00FF8000, // orange
    0x00FF8080  // pink
  },
  { 0x00FFFFFF, // white  (UVI #0, PVI #0)
    0x00FFFF00, // yellow (UVI #1, PVI #1)
    0x00FF00FF, // purple (UVI #4, PVI #2)
    0x00FF0000, // red    (UVI #5, PVI #3)
    0x0000FFFF, // cyan   (UVI #2, PVI #4)
    0x0000DD00, // green  (UVI #3, PVI #5)
    0x004444FF, // blue   (UVI #6, PVI #6)
    0x00000000, // black  (UVI #7, PVI #7)
    0x00BBBBBB, // grey #1
    0x00BBBB00, // dark yellow
    0x00BB00BB, // dark purple
    0x00990000, // dark red
    0x0000AAAA, // dark cyan
    0x0000BB00, // dark green
    0x000000AA, // dark blue
    0x00000000, // dark black
    0x00002000, // grey #2 (green)
    0x00003F00, // grey #3 (green)
    0x00005F00, // grey #4 (green)
    0x00007E00, // grey #5 (green)
    0x00009E00, // grey #6 (green)
    0x0000BD00, // grey #7 (green)
    0x00FF8000, // orange
    0x00FF8080  // pink
  },
  { 0x00FFFFFF, // white  (UVI #0, PVI #0)
    0x00FFFF00, // yellow (UVI #1, PVI #1)
    0x00FF00FF, // purple (UVI #4, PVI #2)
    0x00FF0000, // red    (UVI #5, PVI #3)
    0x0000FFFF, // cyan   (UVI #2, PVI #4)
    0x0000DD00, // green  (UVI #3, PVI #5)
    0x004444FF, // blue   (UVI #6, PVI #6)
    0x00000000, // black  (UVI #7, PVI #7)
    0x00BBBBBB, // grey #1
    0x00BBBB00, // dark yellow
    0x00BB00BB, // dark purple
    0x00990000, // dark red
    0x0000AAAA, // dark cyan
    0x0000BB00, // dark green
    0x000000AA, // dark blue
    0x00000000, // dark black
    0x00252525, // grey #2 (grey)
    0x00494949, // grey #3 (grey)
    0x006D6D6D, // grey #4 (grey)
    0x00929292, // grey #5 (grey)
    0x00B6B6B6, // grey #6 (grey)
    0x00DBDBDB, // grey #7 (grey)
    0x00FF8000, // orange
    0x00FF8080  // pink
} },
{ { 0x00FFFFFF, // white       $FF = 255
    0x00E2E2E2, // yellow      $E2 = 226
    0x00686868, // purple      $68 = 104
    0x004C4C4C, // red         $4C =  76 (same as dark purple!)
    0x00B2B2B2, // cyan        $B2 = 178
    0x00848484, // green       $84 = 132
    0x003A3A3A, // blue        $4A =  74
    0x00000000, // black       $00 =   0
    0x00BBBBBB, // dark white  $BB = 187 (grey #1)
    0x00A6A6A6, // dark yellow $A6 = 166 (brown)
    0x004C4C4C, // dark purple $4C =  76 (same as red!)
    0x002D2D2D, // dark red    $2D =  45
    0x00777777, // dark cyan   $77 = 119
    0x006E6E6E, // dark green  $6E = 110
    0x001D1D1D, // dark blue   $1A =  29
    0x00000000, // dark black  $00 =   0 (black)
    0x00252525, // grey #2
    0x00494949, // grey #3
    0x006D6D6D, // grey #4
    0x00929292, // grey #5
    0x00B6B6B6, // grey #6
    0x00DBDBDB, // grey #7
    0x009D9D9D, // orange
    0x009D9D9D  // pink
  },
  { 0x00FFFFFF, // white       $FF = 255
    0x00E2E2E2, // yellow      $E2 = 226
    0x00686868, // purple      $68 = 104
    0x004C4C4C, // red         $4C =  76 (same as dark purple!)
    0x00B2B2B2, // cyan        $B2 = 178
    0x00848484, // green       $84 = 132
    0x003A3A3A, // blue        $4A =  74
    0x00000000, // black       $00 =   0
    0x00BBBBBB, // dark white  $BB = 187 (grey #1)
    0x00A6A6A6, // dark yellow $A6 = 166 (brown)
    0x004C4C4C, // dark purple $4C =  76 (same as red!)
    0x002D2D2D, // dark red    $2D =  45
    0x00777777, // dark cyan   $77 = 119
    0x006E6E6E, // dark green  $6E = 110
    0x001D1D1D, // dark blue   $1A =  29
    0x00000000, // dark black  $00 =   0 (black)
    0x00252525, // grey #2
    0x00494949, // grey #3
    0x006D6D6D, // grey #4
    0x00929292, // grey #5
    0x00B6B6B6, // grey #6
    0x00DBDBDB, // grey #7
    0x009D9D9D, // orange
    0x009D9D9D  // pink
  },
  { 0x00FFFFFF, // white       $FF = 255
    0x00E2E2E2, // yellow      $E2 = 226
    0x00686868, // purple      $68 = 104
    0x004C4C4C, // red         $4C =  76 (same as dark purple!)
    0x00B2B2B2, // cyan        $B2 = 178
    0x00848484, // green       $84 = 132
    0x003A3A3A, // blue        $4A =  74
    0x00000000, // black       $00 =   0
    0x00BBBBBB, // dark white  $BB = 187 (grey #1)
    0x00A6A6A6, // dark yellow $A6 = 166 (brown)
    0x004C4C4C, // dark purple $4C =  76 (same as red!)
    0x002D2D2D, // dark red    $2D =  45
    0x00777777, // dark cyan   $77 = 119
    0x006E6E6E, // dark green  $6E = 110
    0x001D1D1D, // dark blue   $1A =  29
    0x00000000, // dark black  $00 =   0 (black)
    0x00252525, // grey #2
    0x00494949, // grey #3
    0x006D6D6D, // grey #4
    0x00929292, // grey #5
    0x00B6B6B6, // grey #6
    0x00DBDBDB, // grey #7
    0x009D9D9D, // orange
    0x009D9D9D  // pink
} } };

// IMPORTED VARIABLES-----------------------------------------------------

IMPORT       FLAG    aborting,
                     changed,
                     freshly,
                     hadpaused,
                     jump,
                     lame,
                     undithernow,
                     upgrade;
IMPORT       ULONG   keyframes[SCANCODES],
                     memflags[32768],
                     waittill;
IMPORT       UBYTE   c_chars[2][128][10],
                     c_tone,
                     g_dips,
                     g_paddles,
                     p_chars[97][8],
                     phunsy_chars[128][8],
                     opcode,
                     mg_console,
                     psu,
                     psl,
                     r[7],
                     memory[32768],
                     sd_column[7],
                     s_glow,
                     s_toggles,
                     spritecontents[10],
                     table_sz[256];
IMPORT       ASCREEN screen[BOXWIDTH][BOXHEIGHT];
IMPORT       TEXT    autotext[GAMEINFOLINES][80 + 1],
                     pgmtext[6 * KILOBYTE],
                     hostname[256 + 1],
                     OneStr[10][1 + 1],
                     TwoStr[8][2 + 1],
                     FourStr[17][4 + 1],
                     FiveStr[24][5 + 1],
                     MemoryStr[MEMGADGETS][2 + 1],
                     MonitorStr[MONITORGADS][2 + 1],
                     TimeStr[11 + 1]; // "hh:mm:ss.uu"
IMPORT       UWORD   iar,
                     ras[8];
IMPORT       int     assoc[7],
                     base,
                     consoleopen,
                     drawcorners,
                     drawunlit,
                     fastcpu,
                     forceinterrupt,
                     highres,
                     hostcontroller[2],
                     hostvolume,
                     pvibase,
                     queuepos,
                     rastn,
                     reassociate,
                     regionstart,
                     runtodma,
                     runtointerrupt,
                     runtoframe,
                     runtoloopend,
                     serializemode,
                     sidebarwidth,
                     showlegend,
                     s_io,
                     s_intdir,
                     s_intselector,
                     useff,
                     viewingsprite,
                     winleftx,
                     wintopy,
                     zinterrupt; // interrupt confuses Lint

IMPORT const UBYTE                   e_loader[256],
                                     e_saver[256],
                                     from_a[16];
IMPORT       TEXT                    ClockStr[COMMASTRLEN],
                                     FrameStr[COMMASTRLEN];
IMPORT const TEXT                    dolphin_chars[128];
IMPORT       struct KindStruct       filekind[KINDS];
IMPORT       struct IDirectInput*    lpdi;
IMPORT const UWORD                   pvi_spritedata[4];

#ifdef AMIGA
    IMPORT int                  addedserver,
                                currentgame,
                                led,
                                morphos;
    IMPORT FLAG                 gotallpens,
                                guideexists;
    IMPORT UWORD                InvisiblePointerData[6],
                                pending;
    IMPORT BPTR                 SpeakHandle;
    IMPORT struct Gadget*       gadgets[GIDS + 1];
    IMPORT struct Menu*         MenuPtr;
    IMPORT struct MenuItem*     ItemPtr;
    IMPORT struct Screen*       ScreenPtr;
    IMPORT struct Catalog*      CatalogPtr;
    IMPORT struct AnimInfo*     AnimInfo;
    IMPORT struct Node*         SmallSpeedBarNodePtr[SMALLGADGETS + 1];
    IMPORT struct Window       *MainWindowPtr,
                               *ControlsWindowPtr,
                               *DIPsWindowPtr,
                               *GameInfoWindowPtr,
                               *MemoryWindowPtr,
                               *MonitorWindowPtr,
                               *PaletteWindowPtr,
                               *SpriteWindowPtr;
    IMPORT struct Library      *AmigaGuideBase,
                               *OpenURLBase,
                               *SocketBase;
    IMPORT Object*              smalltbimage[SMALLTBIMAGES];
    IMPORT struct PaletteStruct pencolours[2][PENS];
#endif
#ifdef WIN32
    IMPORT HWND           ControlsWindowPtr,
                          DIPsWindowPtr,
                          GameInfoWindowPtr,
                          MainWindowPtr,
                          MemoryWindowPtr,
                          MonitorWindowPtr,
                          PaletteWindowPtr,
                          SpriteWindowPtr,
                          hToolbar;
    IMPORT HDC            ControlsRastPtr;
    IMPORT FLAG           opening;
    IMPORT int            language,
                          priority,
                          realwide,
                          speakable,
                          sprmakerexists,
                          txtmakerexists,
                          unit[2],
                          useinternal,
                          CatalogPtr; // APTR doesn't work
    IMPORT HMENU          MenuPtr,
                          PopUpMenuPtr;
    IMPORT UBYTE          button[2][8];
    IMPORT ULONG          pens[2][PENS];
    IMPORT const struct RearrangeStruct rearrange[48];
#endif
#ifdef PROFILE
    IMPORT ULONG          profile[256];
#endif

// MODULE STRUCTURES------------------------------------------------------

// see engine1.h

// MODULE ARRAYS----------------------------------------------------------

// (none)

// MODULE FUNCTIONS-------------------------------------------------------

MODULE void NetConnect(const char* Server, unsigned short Port);
MODULE void NetClose(void);
MODULE const char* NetMyName(char* Buffer, int MaxChars);
MODULE void sortbyname(void);
MODULE void sortbymachine(void);
MODULE void flipguidepixel(void);
MODULE __inline void local_drawpixel(int x, int y, int colour);
MODULE __inline void swapgames(int game1, int game2);
#ifdef QUICKSORT
    MODULE void quicksort_bymachine(int m, int n);
    MODULE void quicksort_byname(int m, int n);
#endif

// CODE-------------------------------------------------------------------

EXPORT void change_read(STRPTR thestring)
{   int number;

    /* thestring: contents of the "address" gadget

       This looks at thestring and sets fromnum, tonum, fromstr, tostr,
       tolimit. It is used when the "address" field in the "change"
       requester is modified (and also used by the debugger CLI
       interface). */

    if (!stricmp(thestring, "PSU"))
    {   fromnum = (int) psu;
        tolimit = 255;
    } elif (!stricmp(thestring, "PSL"))
    {   fromnum = (int) psl;
        tolimit = 255;
    } elif (!stricmp(thestring, "IAR"))
    {   fromnum = (int) iar;
        tolimit = 32767;
    } elif
    (   strlen(thestring) == 4
     && !strnicmp(thestring, "RAS", 3)
     && thestring[3] >= '0'
     && thestring[3] <= '7'
    )
    {   fromnum = (int) ras[thestring[3] - '0'];
        tolimit = 32767;
    } elif
    (   strlen(thestring) == 2
     && (thestring[0] == 'R' || thestring[0] == 'r')
     && thestring[1] >= '0'
     && thestring[1] <= '6'
    )
    {   fromnum = (int) r[thestring[1] - '0'];
        tolimit = 255;
    } else
    {   tolimit = 255;

        number = friendly_to_number(thestring);
        if (number < OUTOFRANGE)
        {   fromnum = (int) memory[number];
        } else
        {   fromnum = OUTOFRANGE;
            fromstr[0] = 0;
    }   }

    if (fromnum < OUTOFRANGE)
    {   fromstr[0] = '$';
        DISCARD stcl_h(&fromstr[1], (unsigned long) fromnum);
    }

    tonum = fromnum;
    strcpy(tostr, fromstr);
}

EXPORT void change_write(STRPTR thestring, FLAG quiet)
{   FLAG found = TRUE;
    int  number;

    /* thestring: contents of the "address" gadget

       This looks at thestring and tonum and actually changes the value of
       the register that the user has specified. It is used at the end of
       use of the "change" requester, and also by the CLI-based debugger
       interface. */

    // assert(strlen(lastchange) <= 14);
    DISCARD strupr(thestring);
    if (!jump)
    {   strcpy(lastchange, thestring); // for next time that we reopen this requester
    }

    if (tonum > tolimit)
    {   if (!quiet)
        {   T1; textpen(TEXTPEN_DEBUG);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_OUTOFBOUNDS,
                    "Value is out of bounds (limit is %d)!\n\n"
                ),
                tolimit
            );
            T2;
        }
        return;
    }

    if (!strcmp(thestring, "PSU"))
    {   psu = (UBYTE) tonum;
    } elif (!strcmp(thestring, "PSL"))
    {   psl = (UBYTE) tonum;
    } elif (!strcmp(thestring, "IAR"))
    {   iar = (UWORD) tonum;
    } elif
    (   strlen(thestring) == 4
     && !strncmp(thestring, "RAS", 3)
     && thestring[3] >= '0'
     && thestring[3] <= '7'
    )
    {   ras[thestring[3] - '0'] = (UWORD) tonum;
    } elif
    (   strlen(thestring) == 2
     && (thestring[0] == 'R' || thestring[0] == 'r')
     && thestring[1] >= '0'
     && thestring[1] <= '6'
    )
    {   r[thestring[1] - '0'] = (UBYTE) tonum;
    } else
    {   found = FALSE;

        number = friendly_to_number(thestring);
        if (number < OUTOFRANGE)
        {   memory[number] = (UBYTE) tonum;
            if (memflags[number] & VISIBLE)
            {   changed = TRUE;
            }
            found = TRUE;
            DISCARD number_to_friendly(number, thestring, TRUE);
    }   }

    if (found && fromnum != tonum)
    {   if (!quiet)
        {   T1; textpen(TEXTPEN_DEBUG);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_CHANGEDCONTENTS,
                    "Changed contents of %s from $%02X to $%02X!\n\n"
                ),
                thestring,
                (ULONG) fromnum,
                (ULONG) tonum
            );
            T2;
        }
        update_monitor(FALSE);
    } /* else
    {   if (!quiet)
        {   T1; textpen(TEXTPEN_DEBUG);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_NOCHANGE,
                    "No effect (no change to contents of address).\n\n"
            )   );
            T2;
    }   } */
}

EXPORT int getfriendly(int number)
{   return number_to_friendly(number, friendly, TRUE);
}

EXPORT int number_to_friendly(int number, STRPTR thestring, FLAG both)
{   int i;

/*  number:    input number
    thestring: pointer to output string buffer
    both:      do we want friendly and hex names (TRUE) (eg. "BGCOLOUR
               ($19FD)") or just friendly names (eg. "BGCOLOUR")? */

    switch (machine)
    {
    case ASTROWARS:
    case GALAXIA:
        for (i = machines[machine].firstequiv; i <= machines[machine].lastequiv; i++)
        {
/* #ifdef VERBOSE
            T1; textpen(TEXTPEN_VERBOSE);
            zprintf("Comparing %s against %s...\n", thestring, equivalents[i].text);
            T2;
#endif */
            if (number == (int) (0x1500 + equivalents[i].number))
            {   strcpy(thestring, equivalents[i].text);
                if (both)
                {   strcat(thestring, " ($");
                    DISCARD stcl_h(&thestring[strlen(thestring)], (unsigned long) number); // integer -> hexstring
                    strcat(thestring, ")");
                }
                return TRUE;
        }   }
    acase MALZAK:
        for (i = machines[machine].firstequiv; i <= machines[machine].lastequiv; i++)
        {
/* #ifdef VERBOSE
            T1; textpen(TEXTPEN_VERBOSE);
            zprintf("Comparing %s against %s...\n", thestring, equivalents[i].text);
            T2;
#endif */
            if (number == (int) (0x1400 + equivalents[i].number))
            {   strcpy(thestring, equivalents[i].text);
                if (both)
                {   strcat(thestring, " ($");
                    DISCARD stcl_h(&thestring[strlen(thestring)], (unsigned long) number); // integer -> hexstring
                    strcat(thestring, ")");
                }
                return TRUE;
        }   }
    adefault:
        for (i = machines[machine].firstequiv; i <= machines[machine].lastequiv; i++)
        {
/* #ifdef VERBOSE
            T1; textpen(TEXTPEN_VERBOSE);
            zprintf("Comparing %s against %s...\n", thestring, equivalents[i].text);
            T2;
#endif */
            if (number == (int) equivalents[i].number)
            {   strcpy(thestring, equivalents[i].text);
                if (both)
                {   strcat(thestring, " ($");
                    DISCARD stcl_h(&thestring[strlen(thestring)], (unsigned long) number); // integer -> hexstring
                    strcat(thestring, ")");
                }
                return TRUE;
    }   }   }

    if (firstmnemonic != -1)
    {   for (i = firstmnemonic; i <= lastmnemonic; i++)
        {
/* #ifdef VERBOSE
            T1; textpen(TEXTPEN_VERBOSE);
            zprintf("Comparing %s against %s...\n", thestring, game_equivalents[i].name);
            T2;
#endif */
            if (number == (int) game_equivalents[i].address)
            {   strcpy(thestring, game_equivalents[i].name);
                if (both)
                {   strcat(thestring, " ($");
                    DISCARD stcl_h(&thestring[strlen(thestring)], (unsigned long) number); // integer -> hexstring
                    strcat(thestring, ")");
                }
                return TRUE;
    }   }   }

    strcpy(thestring, "$");
    DISCARD stcl_h(&thestring[1], (unsigned long) number); // integer -> hexstring
    return FALSE;
}

EXPORT int friendly_to_number(STRPTR thestring)
{   /* This never returns a negative number. */

    int  i;
    LONG number;

    if (strlen(thestring) < 1 || strlen(thestring) > 32)
    {   thestring[0] = 0;
        return OUTOFRANGE;
    }

    for (i = machines[machine].firstequiv; i <= machines[machine].lastequiv; i++)
    {
/* #ifdef VERBOSE
        T1; textpen(TEXTPEN_VERBOSE);
        zprintf("Comparing %s against %s...\n", thestring, equivalents[i].text);
        T2;
#endif */
        if (!stricmp(thestring, equivalents[i].text))
        {  if (machine == ASTROWARS || machine == GALAXIA)
           {   return (int) (0x1500 + equivalents[i].number);
           } elif (machine == MALZAK)
           {   return (int) (0x1400 + equivalents[i].number);
           } else
           {   return (int)           equivalents[i].number ;
    }   }  }

    if (firstmnemonic != -1)
    {   for (i = firstmnemonic; i <= lastmnemonic; i++)
        {
/* #ifdef VERBOSE
            T1; textpen(TEXTPEN_VERBOSE);
            zprintf("Comparing %s against %s...\n", thestring, game_equivalents[i].name);
            T2;
#endif */
            if (!stricmp(thestring, game_equivalents[i].name))
            {   return (int) game_equivalents[i].address;
    }   }   }

    number = readbase(thestring);

    if (number < 0 || number >= OUTOFRANGE)
    {   thestring[0] = 0;
        return OUTOFRANGE;
    }

    return (int) number;
}

EXPORT void macro_stop(void)
{   if (recmode == RECMODE_PLAY)
    {   free_iobuffer();
        ghost_dips(TRUE);
    } elif (recmode == RECMODE_RECORD)
    {   // assert(MacroHandle);
        DISCARD fclose(MacroHandle);
        MacroHandle = NULL;

#ifdef AMIGA
        if (AnimInfo)
        {   CloseAnim(AnimInfo, TRUE);
            AnimInfo = NULL;
        }
#endif
#ifdef WIN32
        CloseAnim();
#endif

    }

    if (recmode != RECMODE_NORMAL) // avoids useless refreshing
    {   recmode = RECMODE_NORMAL;
        updatesmlgad(GADPOS_MACRO, (ULONG) recmode, TRUE);
        updatemenus();
}   }

EXPORT void macro_restartplayback(void)
{   offset = 0;
    DISCARD parse_bytes(FALSE, FALSE); // this sets recmode = RECMODE_PLAY for us
}

EXPORT void emu_pause(FLAG showcpu)
{   if (!paused)
    {   paused = TRUE;
        clearruntos();
        sound_off(FALSE);
        updatesmlgad(GADPOS_PAUSED, paused, TRUE);
        updatemenu(MENUITEM_PAUSED);
        fps = 0;
        updatefps();
        settitle();

        if (useguidepixel)
        {   flipguidepixel();
        }
        redrawscreen(); // helpful in frame skipping mode (ensures most
        // recent executed frame is shown, not a frame from a while ago).
        // also needed for guide pixel
    }
    debug_cli(showcpu);
}

EXPORT void emu_unpause(void)
{   if (paused) // important!
    {   paused = FALSE;
        sound_on();
        updatesmlgad(GADPOS_PAUSED, paused, TRUE);
        updatemenu(MENUITEM_PAUSED);
        settitle();
        hadpaused = TRUE;

        if (useguidepixel)
        {   flipguidepixel();
        }
        redrawscreen();
}   }

EXPORT void cripple(void)
{   crippled = TRUE;
    updatemenus();
    updatebiggads();
    updatesmlgads();
}
EXPORT void uncripple(void)
{   crippled = FALSE;
    updatemenus();
    updatebiggads();
    updatesmlgads();
}

EXPORT void updatemenu(int which)
{   if (!MainWindowPtr)
    {   return; // for speed
    }

#ifdef AMIGA
    if (!MenuPtr)
    {   return; // for speed
    }
    // can't do this for IBM-PC because pop-up menu still needs updating
#endif

    switch (which)
    {
    case MENUITEM_ANALOG:
        ghost(which,
            machine != ARCADIA
         && machine != INTERTON
         && machine != ELEKTOR
             ); // eg. PIPBUG, INSTRUCTOR, CD2650, coin-ops, CHAOS, DOLPHIN, PHUNSY
        tick(which, (int) analog);
    acase MENUITEM_AUDIT:
        ghost(which, !showsidebar);
    acase MENUITEM_AUTOFIRE:
        ghost(which,
            machine != ARCADIA
         && machine != INTERTON
         && machine != ELEKTOR
         && machine != MALZAK
         && machine != ASTROWARS
         && machine != GALAXIA
             ); // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS, PHUNSY
        tick(which, (int) autofire);
    acase MENUITEM_AUTOPAUSE:
        tick(which, autopause);
    acase MENUITEM_AUTOSAVE:
        tick(which, autosave);
    acase MENUITEM_BRAM:
        ghost(which,
            machine != ELEKTOR
         && machine != PIPBUG
         && machine != INSTRUCTOR
         && machine != CD2650
         && machine != PHUNSY
             ); // eg. ARCADIA, INTERTON, ASTROWARS, GALAXIA, MALZAK, CHAOS, DOLPHIN
    acase MENUITEM_COLOURRAM:
        ghost(which, machine != ASTROWARS && machine != GALAXIA);
    acase MENUITEM_SENSITIVITY:
        ghost(which,
            (   machine != ARCADIA
             && machine != INTERTON
             && machine != ELEKTOR
            )
         || !analog
        );
    acase MENUITEM_CLEARBP:
    case MENUITEM_CLEARBPS:
        ghost(which, bps == FALSE);
    acase MENUITEM_CLEARWP:
    case MENUITEM_CLEARWPS:
        ghost(which, wps == FALSE);
    acase MENUITEM_CLI:
        tick(which, cli);
    acase MENUITEM_CLIENT:
        ghost(which,
            (   machine != ARCADIA
             && machine != INTERTON
             && machine != ELEKTOR
            )
         || crippled
#ifdef AMIGA
         || !SocketBase
#endif
        );
        tick(which, connected == NET_CLIENT);
    acase MENUITEM_CONFINE:
        ghost(which,
            (   hostcontroller[0] != CONTROLLER_TRACKBALL
             && hostcontroller[1] != CONTROLLER_TRACKBALL
        )   );
        tick(which, confine);
    acase MENUITEM_COLLISIONS:
        ghost(which,
                machine != ARCADIA
             && machine != INTERTON
             && machine != ELEKTOR
             && machine != ASTROWARS
             && machine != GALAXIA
             && machine != MALZAK
            ); // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS, DOLHPIN, PHUNSY
        tick(which, (int) collisions);
    acase MENUITEM_CONFIRM:
        tick(which, confirmable);
    acase MENUITEM_CONTROLS:
        ghost(which,
            (   machine != ARCADIA
             && machine != INTERTON
             && machine != ELEKTOR
             && machine != INSTRUCTOR
             && machine != ASTROWARS
             && machine != GALAXIA
             && machine != MALZAK
             && machine != DOLPHIN
            ) // eg. PIPBUG, CD2650, CHAOS, PHUNSY
#ifdef AMIGA
         || ControlsWindowPtr
#endif
        );
#ifdef WIN32
        tick(which, ControlsWindowPtr != NULL);
#endif
    acase MENUITEM_COPYTEXT:
        ghost(which,
            machine != ARCADIA
         && machine != ELEKTOR
         && machine != PIPBUG
         && machine != INSTRUCTOR
         && machine != CD2650
         && machine != CHAOS
         && machine != PHUNSY
        );
    acase MENUITEM_CRAM:
        ghost(which, memmap != MEMMAP_C && memmap != MEMMAP_D);
    acase MENUITEM_DEMULTIPLEX:
        ghost(which, machine != ARCADIA);
        tick(which, (int) demultiplex);
    acase MENUITEM_DIPSWITCHES:
        ghost(which,
            (   memmap  != MEMMAP_MALZAK2
             && machine != INSTRUCTOR
             && machine != ASTROWARS
             && machine != GALAXIA
            ) // eg. ARCADIA, INTERTON, ELEKTOR, PIPBUG, CD2650, Malzak 1, CHAOS
#ifdef AMIGA
         || DIPsWindowPtr
#endif
        );
#ifdef WIN32
        tick(which, DIPsWindowPtr != NULL);
#endif
    acase MENUITEM_GAMEINFO:
        ghost(which,
            (   (   machine != MALZAK
                 && machine != ASTROWARS
                 && machine != GALAXIA
                 && (   !game
                    || (!pgmtext[0] && !strcmp(autotext[AUTOLINE_GAMENAME], "?"))
                )   )
#ifdef AMIGA
             || GameInfoWindowPtr
#endif
            ));
#ifdef WIN32
            tick(which, GameInfoWindowPtr != NULL);
#endif
    acase MENUITEM_DOTSPACE:
        ghost(which,
            machine != PIPBUG
         && machine != CD2650
         && machine != CHAOS
         && machine != PHUNSY
        ); // eg. ARCADIA, INTERTON, ELEKTOR, INSTRUCTOR, coin-ops, DOLPHIN
        tick(which, dotspace);
    acase MENUITEM_GREYSCALE:
        ghost(which, recmode == RECMODE_RECORD && anims);
        tick(which, colourset == GREYSCALE);
    acase MENUITEM_HIGHRES:
        ghost(which, machine != PIPBUG && machine != CHAOS);
        tick(which, highres);
    acase MENUITEM_INTERRUPT:
        ghost(which,
            machine != INTERTON
         && machine != ELEKTOR
         && machine != INSTRUCTOR
         && machine != MALZAK
         && machine != ASTROWARS
         && machine != GALAXIA
             ); // eg. ARCADIA, PIPBUG, CD2650, CHAOS, DOLPHIN, PHUNSY
    acase MENUITEM_CORNERS:
        ghost(which, machine != PIPBUG && machine != INSTRUCTOR && machine != DOLPHIN);
        tick(which, drawcorners);
    acase MENUITEM_UNLIT:
        ghost(which, machine != PIPBUG && machine != INSTRUCTOR && machine != DOLPHIN);
        tick(which, drawunlit);
    acase MENUITEM_FLAGLINE:
        ghost(which, machine != ARCADIA && machine != INTERTON && machine != ELEKTOR && machine != CD2650);
        tick(which, (int) flagline);
    acase MENUITEM_IMAGERY:
        ghost(which,
            machine != ASTROWARS
         && machine != GALAXIA
         && machine != MALZAK
         && machine != PHUNSY
         && (   !game
             || (   machine != ARCADIA
                 && machine != INTERTON
                 && machine != ELEKTOR
        )   )   );
    acase MENUITEM_LOWERCASE:
        ghost(which,
            machine != PIPBUG
         && machine != CD2650
         && machine != CHAOS
         && machine != PHUNSY
        );
        tick(which, lowercase);
    acase MENUITEM_INSTRUCTIONS:
        tick(which, loginstructions);
    acase MENUITEM_LOGEPROMS:
        ghost(which,
            memmap != MEMMAP_F
         && memmap != MEMMAP_P
         && machine != PHUNSY
        );
        tick(which, logeprom);
    acase MENUITEM_IOPORTS:
        tick(which, logioports);
    acase MENUITEM_LOGTOFILE:
        tick(which, logfile);
    acase MENUITEM_SHADOWS:
        ghost(which,
            machine != ARCADIA
         && machine != INTERTON
         && machine != ELEKTOR
         && machine != PIPBUG
         && machine != INSTRUCTOR
         && machine != CD2650
         && machine != CHAOS
         && machine != DOLPHIN
         && machine != PHUNSY
             ); // eg. MALZAK, ASTROWARS, GALAXIA
        tick(which, shadowing);
    acase MENUITEM_SPRITESVISIBLE:
        ghost(which,
            machine != ARCADIA
         && machine != INTERTON
         && machine != ELEKTOR
         && machine != MALZAK
         && machine != ASTROWARS
         && machine != GALAXIA
             ); // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS, DOLPHIN
        tick(which, spritesvisible);
    acase MENUITEM_LOGBIOS:
        ghost(which,
            machine != ELEKTOR
         && machine != PIPBUG
         && machine != INSTRUCTOR
         && machine != CD2650
         && machine != PHUNSY
        );
        tick(which, logbios);
    acase MENUITEM_INVERSE:
        tick(which, inverse);
        ghost(which, machine != CD2650 && machine != PHUNSY);
    acase MENUITEM_RASTLINES:
        tick(which, rastlines);
    acase MENUITEM_SCALE2X:
        tick(which, scale2x);
    acase MENUITEM_READWRITES:
        tick(which, logreadwrites);
    acase MENUITEM_SUBROUTINES:
        tick(which, logsubroutines);
    acase MENUITEM_PAUSELOG:
        tick(which, pauselog);
    acase MENUITEM_ANSI:
        tick(which, useansi);
    acase MENUITEM_LOOP:
        tick(which, loop);
    acase MENUITEM_PASTETEXT:
        ghost(which,
            (machine != PIPBUG
          && machine != CD2650
          && machine != CHAOS
          && machine != PHUNSY)
         || recmode == RECMODE_PLAY
        );
    acase MENUITEM_PLAYFIELD:
        ghost(which, machine != MALZAK);
    acase MENUITEM_ANIMS:
        ghost(which, recmode == RECMODE_RECORD);
        tick(which, anims);
    acase MENUITEM_REDEFINEKEYS:
        ghost(which,
            machine != ARCADIA
         && machine != INTERTON
         && machine != ELEKTOR
         && machine != INSTRUCTOR
         && machine != MALZAK
         && machine != ASTROWARS
         && machine != GALAXIA
         && machine != DOLPHIN
        ); // eg. PIPBUG, CD2650, CHAOS
    acase MENUITEM_REARRANGE:
        ghost(which,
            machine != ARCADIA
         && machine != INTERTON
         && machine != ELEKTOR
         && machine != MALZAK
         && machine != ASTROWARS
         && machine != GALAXIA
        ); // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS, DOLPHIN
    acase MENUITEM_RETUNE:
        ghost(which,
            machine != ARCADIA
         && machine != INTERTON
         && machine != ELEKTOR
         && machine != MALZAK
         && machine != ASTROWARS
         && machine != GALAXIA
        ); // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS
        tick(which, retune);
    acase MENUITEM_SCREEN: // view screen contents
        ghost(which,
            machine != ARCADIA
         && machine != INTERTON
         && machine != ELEKTOR
         && machine != INSTRUCTOR
         && machine != CD2650
         && machine != MALZAK
         && machine != ASTROWARS
         && machine != GALAXIA
         && machine != DOLPHIN
         && machine != PHUNSY
        ); // eg. PIPBUG, CHAOS
    acase MENUITEM_NVRAM:
        ghost(which, memmap != MEMMAP_MALZAK2);
    acase MENUITEM_UDCS:
    case MENUITEM_XVI:
    case MENUITEM_RUNTORAST:
    case MENUITEM_RUNTORASTN:
        ghost(which,
            machine != ARCADIA
         && machine != INTERTON
         && machine != ELEKTOR
         && machine != MALZAK
         && machine != ASTROWARS
         && machine != GALAXIA
        ); // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS
    acase MENUITEM_MACHINE:
        enable(MENUOPT_ARCADIA,    !crippled);
        enable(MENUOPT_INTERTON,   !crippled);
        enable(MENUOPT_ELEKTOR,    !crippled);
        enable(MENUOPT_PIPBUG,     !crippled);
        enable(MENUOPT_INSTRUCTOR, !crippled);
        enable(MENUOPT_CD2650,     !crippled);
        enable(MENUOPT_ASTROWARS,  !crippled);
        enable(MENUOPT_GALAXIA,    !crippled);
        enable(MENUOPT_MALZAK1,    !crippled);
        enable(MENUOPT_MALZAK2,    !crippled);
        enable(MENUOPT_CHAOS,      !crippled);
        enable(MENUOPT_DOLPHIN,    !crippled);
        enable(MENUOPT_PHUNSY,     !crippled);
        if (machine == ARCADIA   )
            checkmx(MENUOPT_ARCADIA,    MENUOPT_ARCADIA, MENUOPT_PHUNSY);
        else
            uncheckmx(MENUOPT_ARCADIA   );
        if (machine == INTERTON  )
            checkmx(MENUOPT_INTERTON,   MENUOPT_ARCADIA, MENUOPT_PHUNSY);
        else
            uncheckmx(MENUOPT_INTERTON  );
        if (machine == ELEKTOR   )
            checkmx(MENUOPT_ELEKTOR,    MENUOPT_ARCADIA, MENUOPT_PHUNSY);
        else
            uncheckmx(MENUOPT_ELEKTOR   );
        if (machine == INSTRUCTOR)
            checkmx(MENUOPT_INSTRUCTOR, MENUOPT_ARCADIA, MENUOPT_PHUNSY);
        else
            uncheckmx(MENUOPT_INSTRUCTOR);
        if (machine == PIPBUG    )
            checkmx(MENUOPT_PIPBUG,     MENUOPT_ARCADIA, MENUOPT_PHUNSY);
        else
            uncheckmx(MENUOPT_PIPBUG    );
        if (machine == CD2650    )
            checkmx(MENUOPT_CD2650,     MENUOPT_ARCADIA, MENUOPT_PHUNSY);
        else
            uncheckmx(MENUOPT_CD2650    );
        if (memmap == MEMMAP_ASTROWARS)
            checkmx(MENUOPT_ASTROWARS,  MENUOPT_ARCADIA, MENUOPT_PHUNSY);
        else
            uncheckmx(MENUOPT_ASTROWARS );
        if (memmap == MEMMAP_GALAXIA    )
            checkmx(MENUOPT_GALAXIA,    MENUOPT_ARCADIA, MENUOPT_PHUNSY);
        else
            uncheckmx(MENUOPT_GALAXIA   );
        if (memmap == MEMMAP_MALZAK1)
            checkmx(MENUOPT_MALZAK1,    MENUOPT_ARCADIA, MENUOPT_PHUNSY);
        else
            uncheckmx(MENUOPT_MALZAK1   );
        if (memmap == MEMMAP_MALZAK2)
            checkmx(MENUOPT_MALZAK2,    MENUOPT_ARCADIA, MENUOPT_PHUNSY);
        else
            uncheckmx(MENUOPT_MALZAK2   );
        if (machine == CHAOS     )
            checkmx(MENUOPT_CHAOS,      MENUOPT_ARCADIA, MENUOPT_PHUNSY);
        else
            uncheckmx(MENUOPT_CHAOS     );
        if (machine == DOLPHIN   )
            checkmx(MENUOPT_DOLPHIN,    MENUOPT_ARCADIA, MENUOPT_PHUNSY);
        else
            uncheckmx(MENUOPT_DOLPHIN   );
        if (machine == PHUNSY    )
            checkmx(MENUOPT_PHUNSY,     MENUOPT_ARCADIA, MENUOPT_PHUNSY);
        else
            uncheckmx(MENUOPT_PHUNSY    );
    acase MENUITEM_MACRO:
        if (crippled)
        {   enable(MENUOPT_STARTRECORDING,  FALSE);
            enable(MENUOPT_RESTARTPLAYBACK, FALSE);
            enable(MENUOPT_STOP,            FALSE);
        } elif (recmode == RECMODE_PLAY)
        {   enable(MENUOPT_STARTRECORDING,  FALSE);
            enable(MENUOPT_RESTARTPLAYBACK, TRUE );
            enable(MENUOPT_STOP,            TRUE );
        } elif (recmode == RECMODE_RECORD)
        {   enable(MENUOPT_STARTRECORDING,  FALSE);
            enable(MENUOPT_RESTARTPLAYBACK, FALSE);
            enable(MENUOPT_STOP,            TRUE );
        } elif (recmode == RECMODE_NORMAL)
        {   enable(MENUOPT_STARTRECORDING,  TRUE );
            enable(MENUOPT_RESTARTPLAYBACK, FALSE);
            enable(MENUOPT_STOP,            FALSE);
        } // else assert(0);
    acase MENUITEM_VIEWMONITOR:
#ifdef AMIGA
        ghost(which, MonitorWindowPtr != NULL);
#endif
#ifdef WIN32
        tick(which, MonitorWindowPtr != NULL);
#endif
    acase MENUITEM_VIEWMEMORY:
#ifdef AMIGA
        ghost(which, MemoryWindowPtr != NULL);
#endif
#ifdef WIN32
        tick(which, MemoryWindowPtr != NULL);
#endif
    acase MENUITEM_VIEWHISCORES:
        ghost(which,
            machine != ARCADIA
         && machine != ASTROWARS
         && machine != GALAXIA
         && machine != MALZAK
             ); // eg. INTERTON, ELEKTOR, PIPBUG, INSTRUCTOR, CD2650, CHAOS, DOLPHIN
    acase MENUITEM_NARROW:
        ghost(which,
            fullscreen
         || (recmode == RECMODE_RECORD && anims)
        );
        tick(which, wide == 1);
    acase MENUITEM_OPEN:
    case MENUITEM_RESET:
        ghost(which, crippled);
    acase MENUITEM_QUICKLOAD:
        ghost(which, crippled);
    acase MENUITEM_PAUSED:
        tick(which, (int) paused);
    acase MENUITEM_POINTER:
        tick(which, showpointer);
    acase MENUITEM_CONSOLE:
        ghost(which,
             machine != PIPBUG
          && machine != CHAOS
             );
        tick(which, echoconsole);
    acase MENUITEM_PRINTER:
        ghost(which,
             machine != PIPBUG
          && machine != CHAOS
             );
        tick(which, echoprinter);
    acase MENUITEM_PSGS:
        ghost(which, memmap != MEMMAP_F);
    acase MENUITEM_LEFT:
        enable(MENUMENU_LEFT,          machine == ARCADIA || machine == INTERTON || machine == ELEKTOR || machine == ASTROWARS || machine == GALAXIA || machine == MALZAK);
        enable(MENUOPT_LEFT_TRACKBALL, machine == ARCADIA || machine == INTERTON || machine == ELEKTOR || machine == ASTROWARS || machine == GALAXIA || machine == MALZAK);
        enable(MENUOPT_LEFT_NONE,      machine == ARCADIA || machine == INTERTON || machine == ELEKTOR || machine == ASTROWARS || machine == GALAXIA || machine == MALZAK);
        enable(MENUOPT_LEFT_1STDJOY,   machine == ARCADIA || machine == INTERTON || machine == ELEKTOR || machine == ASTROWARS || machine == GALAXIA || machine == MALZAK
#ifdef WIN32
        && unit[0] != -1
#endif
        );
        enable(MENUOPT_LEFT_2NDDJOY,   machine == ARCADIA || machine == INTERTON || machine == ELEKTOR || machine == ASTROWARS || machine == GALAXIA || machine == MALZAK
#ifdef WIN32
        && unit[1] != -1
#endif
        );
#ifdef AMIGA
        enable(MENUOPT_LEFT_2NDAJOY, addedserver && (machine == ARCADIA || machine == INTERTON || machine == ELEKTOR || machine == ASTROWARS || machine == GALAXIA || machine == MALZAK));
        enable(MENUOPT_LEFT_1STAPAD, morphos     && (machine == ARCADIA || machine == INTERTON || machine == ELEKTOR || machine == ASTROWARS || machine == GALAXIA || machine == MALZAK));
        enable(MENUOPT_LEFT_2NDAPAD, morphos     && (machine == ARCADIA || machine == INTERTON || machine == ELEKTOR || machine == ASTROWARS || machine == GALAXIA || machine == MALZAK));
#endif
        if (hostcontroller[0] == CONTROLLER_TRACKBALL)
            checkmx(   MENUOPT_LEFT_TRACKBALL, MENUOPT_LEFT_TRACKBALL, MENUOPT_LEFT_NONE);
        else uncheckmx(MENUOPT_LEFT_TRACKBALL);
        if (hostcontroller[0] == CONTROLLER_1STDJOY)
            checkmx(   MENUOPT_LEFT_1STDJOY, MENUOPT_LEFT_TRACKBALL, MENUOPT_LEFT_NONE);
        else uncheckmx(MENUOPT_LEFT_1STDJOY);
        if (hostcontroller[0] == CONTROLLER_2NDDJOY)
            checkmx(   MENUOPT_LEFT_2NDDJOY, MENUOPT_LEFT_TRACKBALL, MENUOPT_LEFT_NONE);
        else uncheckmx(MENUOPT_LEFT_2NDDJOY);
        if (hostcontroller[0] == CONTROLLER_1STDPAD)
            checkmx(   MENUOPT_LEFT_1STDPAD, MENUOPT_LEFT_TRACKBALL, MENUOPT_LEFT_NONE);
        else uncheckmx(MENUOPT_LEFT_1STDPAD);
        if (hostcontroller[0] == CONTROLLER_2NDDPAD)
            checkmx(   MENUOPT_LEFT_2NDDPAD, MENUOPT_LEFT_TRACKBALL, MENUOPT_LEFT_NONE);
        else uncheckmx(MENUOPT_LEFT_2NDDPAD);
        if (hostcontroller[0] == CONTROLLER_2NDAJOY)
            checkmx(   MENUOPT_LEFT_2NDAJOY, MENUOPT_LEFT_TRACKBALL, MENUOPT_LEFT_NONE);
        else uncheckmx(MENUOPT_LEFT_2NDAJOY);
        if (hostcontroller[0] == CONTROLLER_1STAPAD)
            checkmx(   MENUOPT_LEFT_1STAPAD, MENUOPT_LEFT_TRACKBALL, MENUOPT_LEFT_NONE);
        else uncheckmx(MENUOPT_LEFT_1STAPAD);
        if (hostcontroller[0] == CONTROLLER_2NDAPAD)
            checkmx(   MENUOPT_LEFT_2NDAPAD, MENUOPT_LEFT_TRACKBALL, MENUOPT_LEFT_NONE);
        else uncheckmx(MENUOPT_LEFT_2NDAPAD);
        if (hostcontroller[0] == CONTROLLER_NONE)
            checkmx(   MENUOPT_LEFT_NONE,    MENUOPT_LEFT_TRACKBALL, MENUOPT_LEFT_NONE);
        else uncheckmx(MENUOPT_LEFT_NONE);
    acase MENUITEM_RIGHT:
        enable(MENUMENU_RIGHT,          machine == ARCADIA || machine == INTERTON || machine == ELEKTOR || machine == ASTROWARS || machine == GALAXIA || machine == MALZAK);
        enable(MENUOPT_RIGHT_TRACKBALL, machine == ARCADIA || machine == INTERTON || machine == ELEKTOR || machine == ASTROWARS || machine == GALAXIA || machine == MALZAK);
        enable(MENUOPT_RIGHT_NONE,      machine == ARCADIA || machine == INTERTON || machine == ELEKTOR || machine == ASTROWARS || machine == GALAXIA || machine == MALZAK);
        enable(MENUOPT_RIGHT_1STDJOY,   machine == ARCADIA || machine == INTERTON || machine == ELEKTOR || machine == ASTROWARS || machine == GALAXIA || machine == MALZAK
#ifdef WIN32
        && unit[0] != -1
#endif
        );
        enable(MENUOPT_RIGHT_2NDDJOY,   machine == ARCADIA || machine == INTERTON || machine == ELEKTOR || machine == ASTROWARS || machine == GALAXIA || machine == MALZAK
#ifdef WIN32
        && unit[1] != -1
#endif
        );
#ifdef AMIGA
        enable(MENUOPT_RIGHT_2NDAJOY, addedserver && (machine == ARCADIA || machine == INTERTON || machine == ELEKTOR || machine == ASTROWARS || machine == GALAXIA || machine == MALZAK));
        enable(MENUOPT_RIGHT_1STAPAD, morphos     && (machine == ARCADIA || machine == INTERTON || machine == ELEKTOR || machine == ASTROWARS || machine == GALAXIA || machine == MALZAK));
        enable(MENUOPT_RIGHT_2NDAPAD, morphos     && (machine == ARCADIA || machine == INTERTON || machine == ELEKTOR || machine == ASTROWARS || machine == GALAXIA || machine == MALZAK));
#endif
        if (hostcontroller[1] == CONTROLLER_TRACKBALL)
            checkmx(   MENUOPT_RIGHT_TRACKBALL, MENUOPT_RIGHT_TRACKBALL, MENUOPT_RIGHT_NONE);
        else uncheckmx(MENUOPT_RIGHT_TRACKBALL);
        if (hostcontroller[1] == CONTROLLER_1STDJOY)
            checkmx(   MENUOPT_RIGHT_1STDJOY, MENUOPT_RIGHT_TRACKBALL, MENUOPT_RIGHT_NONE);
        else uncheckmx(MENUOPT_RIGHT_1STDJOY);
        if (hostcontroller[1] == CONTROLLER_2NDDJOY)
            checkmx(   MENUOPT_RIGHT_2NDDJOY, MENUOPT_RIGHT_TRACKBALL, MENUOPT_RIGHT_NONE);
        else uncheckmx(MENUOPT_RIGHT_2NDDJOY);
        if (hostcontroller[1] == CONTROLLER_1STDPAD)
            checkmx(   MENUOPT_RIGHT_1STDPAD, MENUOPT_RIGHT_TRACKBALL, MENUOPT_RIGHT_NONE);
        else uncheckmx(MENUOPT_RIGHT_1STDPAD);
        if (hostcontroller[1] == CONTROLLER_2NDDPAD)
            checkmx(   MENUOPT_RIGHT_2NDDPAD, MENUOPT_RIGHT_TRACKBALL, MENUOPT_RIGHT_NONE);
        else uncheckmx(MENUOPT_RIGHT_2NDDPAD);
        if (hostcontroller[1] == CONTROLLER_2NDAJOY)
            checkmx(   MENUOPT_RIGHT_2NDAJOY, MENUOPT_RIGHT_TRACKBALL, MENUOPT_RIGHT_NONE);
        else uncheckmx(MENUOPT_RIGHT_2NDAJOY);
        if (hostcontroller[1] == CONTROLLER_1STAPAD)
            checkmx(   MENUOPT_RIGHT_1STAPAD, MENUOPT_RIGHT_TRACKBALL, MENUOPT_RIGHT_NONE);
        else uncheckmx(MENUOPT_RIGHT_1STAPAD);
        if (hostcontroller[1] == CONTROLLER_2NDAPAD)
            checkmx(   MENUOPT_RIGHT_2NDAPAD, MENUOPT_RIGHT_TRACKBALL, MENUOPT_RIGHT_NONE);
        else uncheckmx(MENUOPT_RIGHT_2NDAPAD);
        if (hostcontroller[1] == CONTROLLER_NONE)
            checkmx(   MENUOPT_RIGHT_NONE,    MENUOPT_RIGHT_TRACKBALL, MENUOPT_RIGHT_NONE);
        else uncheckmx(MENUOPT_RIGHT_NONE);
    acase MENUITEM_CRYSTAL:
        enable(MENUOPT_SLOW, machine == CD2650);
        enable(MENUOPT_FAST, machine == CD2650);
        if (!fastcpu)
        {     checkmx(MENUOPT_SLOW, MENUOPT_SLOW, MENUOPT_FAST);
            uncheckmx(MENUOPT_FAST);
        } else
        {   uncheckmx(MENUOPT_SLOW);
              checkmx(MENUOPT_FAST, MENUOPT_SLOW, MENUOPT_FAST);
        }
    acase MENUITEM_REGION:
        if (region == REGION_NTSC)
        {     checkmx(MENUOPT_NTSC, MENUOPT_NTSC, MENUOPT_PAL);
            uncheckmx(MENUOPT_PAL);
        } else
        {   // assert(region == REGION_PAL);
            uncheckmx(MENUOPT_NTSC);
              checkmx(MENUOPT_PAL,  MENUOPT_NTSC, MENUOPT_PAL);
        }
    acase MENUITEM_CPUTYPE:
        if (supercpu == 1)
        {     checkmx(MENUOPT_2650B, MENUOPT_2650A, MENUOPT_2650B);
            uncheckmx(MENUOPT_2650A);
        } else
        {     checkmx(MENUOPT_2650A, MENUOPT_2650A, MENUOPT_2650B);
            uncheckmx(MENUOPT_2650B);
        }
    acase MENUITEM_MNEMONICS:
        if (style == STYLE_CALM)
        {     checkmx(MENUOPT_CALM,      MENUOPT_CALM, MENUOPT_SIGNETICS);
            uncheckmx(MENUOPT_SIGNETICS);
        } else
        {   // assert(style == STYLE_SIGNETICS);
              checkmx(MENUOPT_SIGNETICS, MENUOPT_CALM, MENUOPT_SIGNETICS);
            uncheckmx(MENUOPT_CALM);
        }
#ifdef WIN32
	acase MENUITEM_PRIORITY:
        switch (priority)
        {
        case PRI_HIGH:
            checkmx(MENUOPT_HIGHPRI,   MENUOPT_HIGHPRI, MENUOPT_LOWPRI);
            uncheckmx(MENUOPT_NORMALPRI);
            uncheckmx(MENUOPT_LOWPRI);
        acase PRI_NORMAL:
            uncheckmx(MENUOPT_HIGHPRI);
            checkmx(MENUOPT_NORMALPRI, MENUOPT_HIGHPRI, MENUOPT_LOWPRI);
            uncheckmx(MENUOPT_LOWPRI);
        acase PRI_LOW:
            uncheckmx(MENUOPT_HIGHPRI);
            uncheckmx(MENUOPT_NORMALPRI);
            checkmx(MENUOPT_LOWPRI,    MENUOPT_HIGHPRI, MENUOPT_LOWPRI);
        }
#endif
	acase MENUITEM_SCREENCOLOUR:
        enable(MENUOPT_AMBER, colourset == COLOURFUL && (machine == CD2650 || machine == PHUNSY));
        enable(MENUOPT_GREEN, colourset == COLOURFUL && (machine == CD2650 || machine == PHUNSY));
        enable(MENUOPT_WHITE, colourset == COLOURFUL && (machine == CD2650 || machine == PHUNSY));
        switch (wsc)
        {
        case 0:
            checkmx(MENUOPT_AMBER,     MENUOPT_AMBER, MENUOPT_WHITE);
            uncheckmx(MENUOPT_GREEN);
            uncheckmx(MENUOPT_WHITE);
        acase 1:
            uncheckmx(MENUOPT_AMBER);
            checkmx(MENUOPT_GREEN,     MENUOPT_AMBER, MENUOPT_WHITE);
            uncheckmx(MENUOPT_WHITE);
        acase 2:
            uncheckmx(MENUOPT_AMBER);
            uncheckmx(MENUOPT_GREEN);
            checkmx(MENUOPT_WHITE,     MENUOPT_AMBER, MENUOPT_WHITE);
        }
    acase MENUITEM_REQUIREBUTTON:
        ghost(which,
            machine != ARCADIA
         && machine != INTERTON
         && machine != ELEKTOR
         && machine != MALZAK
         && machine != ASTROWARS
         && machine != GALAXIA
        ); // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS
        tick(which, requirebutton);
    acase MENUITEM_RESETTOMONITOR:
        ghost(which,
            machine != ELEKTOR
         && machine != PIPBUG
         && machine != INSTRUCTOR
         && machine != CD2650
         && machine != PHUNSY
        ); // eg. ARCADIA, INTERTON, MALZAK, ASTROWARS, GALAXIA, CHAOS
    acase MENUITEM_ROTATE:
        ghost(which, machine != ASTROWARS && machine != GALAXIA);
        // eg. ARCADIA, INTERTON, ELEKTOR, INSTRUCTOR, PIPBUG, MALZAK, CD2650, CHAOS
        tick(which, rotate);
    acase MENUITEM_RUNTODMA:
    case MENUITEM_EAPSPRITES:
    case MENUITEM_EAPUDCS:
    case MENUITEM_SAVEEAL:
        ghost(which, machine != ARCADIA);
    acase MENUITEM_RUNTOINTERRUPT:
        ghost(which,
            machine != INTERTON
         && machine != ELEKTOR
         && machine != INSTRUCTOR
         && machine != MALZAK
         && machine != ASTROWARS
         && machine != GALAXIA
        ); // eg. ARCADIA, PIPBUG, CD2650, CHAOS, PHUNSY
    acase MENUITEM_REVERT:
        ghost(which, crippled || !game);
    acase MENUITEM_SAVEROM:
        ghost(which, crippled || !game || machine == DOLPHIN || machine == PHUNSY);
    acase MENUITEM_QUICKSAVE:
    case MENUITEM_SAVESNP:
        ghost(which, crippled);
    acase MENUITEM_SAVETVC:
        ghost(which, crippled || !game || machine != ELEKTOR);
    acase MENUITEM_SORTBY:
        if (sortby == SORTBYNAME)
        {     checkmx(MENUOPT_SORTBYNAME,    MENUOPT_SORTBYNAME, MENUOPT_SORTBYMACHINE);
            uncheckmx(MENUOPT_SORTBYMACHINE);
        } else
        {   // assert(sortby == SORTBYMACHINE);
            uncheckmx(MENUOPT_SORTBYNAME);
              checkmx(MENUOPT_SORTBYMACHINE, MENUOPT_SORTBYNAME, MENUOPT_SORTBYMACHINE);
        }
    acase MENUITEM_SENSEGAME:
        tick(which, sensegame);
    acase MENUITEM_SERVER:
        ghost(which,
            (   machine != ARCADIA
             && machine != INTERTON
             && machine != ELEKTOR
            )
         || crippled
#ifdef AMIGA
         || !SocketBase
#endif
        );
        tick(which, connected == NET_SERVER);
    acase MENUITEM_SIZE:
        enable(MENUOPT_1XSIZE, !fullscreen && (recmode != RECMODE_RECORD || !anims));
        enable(MENUOPT_2XSIZE, !fullscreen && (recmode != RECMODE_RECORD || !anims));
        enable(MENUOPT_3XSIZE, !fullscreen && (recmode != RECMODE_RECORD || !anims));
        enable(MENUOPT_4XSIZE, !fullscreen && (recmode != RECMODE_RECORD || !anims));
        if (size == 1)
            checkmx(   MENUOPT_1XSIZE, MENUOPT_1XSIZE, MENUOPT_4XSIZE);
        else uncheckmx(MENUOPT_1XSIZE);
        if (size == 2)
            checkmx(   MENUOPT_2XSIZE, MENUOPT_1XSIZE, MENUOPT_4XSIZE);
        else uncheckmx(MENUOPT_2XSIZE);
        if (size == 3)
            checkmx(   MENUOPT_3XSIZE, MENUOPT_1XSIZE, MENUOPT_4XSIZE);
        else uncheckmx(MENUOPT_3XSIZE);
        if (size == 4)
            checkmx(   MENUOPT_4XSIZE, MENUOPT_1XSIZE, MENUOPT_4XSIZE);
        else uncheckmx(MENUOPT_4XSIZE);
    acase MENUITEM_SPEECH:
        ghost(which,
            (machine != PIPBUG && machine != CHAOS)
#ifdef WIN32
         || !speakable
#endif
        );
        tick(which, usespeech);
    acase MENUITEM_SOUND:
        tick(which, (int) sound);
    acase MENUITEM_SWAPPED:
        ghost(which,
            machine != ARCADIA
         && machine != INTERTON
         && machine != ELEKTOR
         && machine != MALZAK
         && machine != ASTROWARS
         && machine != GALAXIA
        ); // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS
        tick(which, (int) swapped);
    acase MENUITEM_TITLEBAR:
        tick(which, showtitlebar);
    acase MENUITEM_TRACECPU:
        tick(which, trace);
    acase MENUITEM_TRACEVERBOSE:
        tick(which, traceverbose);
    acase MENUITEM_UNDITHER:
        ghost(which, machine != ARCADIA);
        tick(which, undither);
    acase MENUITEM_STARFIELDS:
        ghost(which,
            machine != ARCADIA
         && machine != INTERTON
         && machine != ELEKTOR
         && machine != MALZAK
         && machine != ASTROWARS
         && machine != GALAXIA
             ); // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS
        tick(which, enhancestars);
    acase MENUITEM_USEHACKS:
        ghost(which,
            machine != GALAXIA
         && machine != MALZAK
         && (   (    machine != INTERTON
                  && machine != ELEKTOR
                )
             || !sensegame
        )   );
        tick(which, usehacks);
    acase MENUITEM_TURBO:
        tick(which, (int) turbo);
    acase MENUITEM_WATCHREADS:
        tick(which, watchreads);
    acase MENUITEM_USEGUIDEPIXEL:
        ghost(which,
            machine != ARCADIA
         && machine != INTERTON
         && machine != ELEKTOR
         && machine != MALZAK
         && machine != ASTROWARS
         && machine != GALAXIA
        ); // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS, DOLPHIN
        tick(which, useguidepixel);
    acase MENUITEM_TOOLBAR:
        tick(which, showtoolbar);
    acase MENUITEM_SIDEBAR:
        tick(which, showsidebar);
    acase MENUITEM_SPRITEVIEWER:
#ifdef AMIGA
        ghost(which,
            (   machine != ARCADIA
             && machine != INTERTON
             && machine != ELEKTOR
             && machine != MALZAK
             && machine != ASTROWARS
             && machine != GALAXIA
            ) // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS, DOLPHIN
         || SpriteWindowPtr != NULL
        );
#endif
#ifdef WIN32
        ghost(which,
            (   machine != ARCADIA
             && machine != INTERTON
             && machine != ELEKTOR
             && machine != MALZAK
             && machine != ASTROWARS
             && machine != GALAXIA
        )   ); // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS, DOLPHIN
        tick(which, SpriteWindowPtr != NULL);
#endif
    acase MENUITEM_PALETTE:
#ifdef AMIGA
        ghost(which, !gotallpens || PaletteWindowPtr != NULL);
#endif
#ifdef WIN32
        tick(which, PaletteWindowPtr != NULL);
#endif
#ifdef AMIGA
    acase MENUITEM_CREATEICONS:
        tick(which, icons);
    acase MENUITEM_FILTERED:
        ghost(which, (int) lame);
        tick(which, led);
    acase MENUITEM_CODINGGUIDE:
    case MENUITEM_GAMINGGUIDE:
        ghost(which, !OpenURLBase);
    acase MENUITEM_MANUAL:
        ghost(which, !AmigaGuideBase || !guideexists);
    acase MENUITEM_FORCEFEEDBACK:
        ghost(which,
            !morphos
         || (   machine != ARCADIA
             && machine != INTERTON
             && machine != ELEKTOR
             && machine != ASTROWARS
             && machine != GALAXIA
             && machine != MALZAK
        )   ); // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS
        tick(which, useff);
#endif
#ifdef WIN32
    acase MENUITEM_FORCEFEEDBACK:
        ghost(which,
            !lpdi
         || (   machine != ARCADIA
             && machine != INTERTON
             && machine != ELEKTOR
             && machine != ASTROWARS
             && machine != GALAXIA
             && machine != MALZAK
        )   ); // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS
        tick(which, useff);
    acase MENUITEM_STRETCH:
        ghost(which, !fullscreen || (recmode == RECMODE_RECORD && anims));
        tick(which, stretch);
    acase MENUITEM_MENUBAR:
        tick(which, showmenubar);
    acase MENUITEM_INTERNAL:
        ghost(which, lame);
        tick(which, useinternal);
    acase MENUITEM_FULLSCREEN:
        ghost(which, recmode == RECMODE_RECORD && anims);
        tick(which, fullscreen);
    acase MENUITEM_SPRMAKER:
        ghost(which, machine != ARCADIA || !sprmakerexists);
    acase MENUITEM_TXTMAKER:
        ghost(which, machine != ARCADIA || !txtmakerexists);
    acase MENUITEM_LANGUAGE:
        if (language == LANGUAGE_ENG)
            checkmx(   MENUOPT_ENG, MENUOPT_ENG, MENUOPT_SPA);
        else uncheckmx(MENUOPT_ENG);
        if (language == LANGUAGE_HOL)
            checkmx(   MENUOPT_HOL, MENUOPT_ENG, MENUOPT_SPA);
        else uncheckmx(MENUOPT_HOL);
        if (language == LANGUAGE_FRA)
            checkmx(   MENUOPT_FRA, MENUOPT_ENG, MENUOPT_SPA);
        else uncheckmx(MENUOPT_FRA);
        if (language == LANGUAGE_GER)
            checkmx(   MENUOPT_GER, MENUOPT_ENG, MENUOPT_SPA);
        else uncheckmx(MENUOPT_GER);
        if (language == LANGUAGE_ITA)
            checkmx(   MENUOPT_ITA, MENUOPT_ENG, MENUOPT_SPA);
        else uncheckmx(MENUOPT_ITA);
        if (language == LANGUAGE_RUS)
            checkmx(   MENUOPT_RUS, MENUOPT_ENG, MENUOPT_SPA);
        else uncheckmx(MENUOPT_RUS);
        if (language == LANGUAGE_SPA)
            checkmx(   MENUOPT_SPA, MENUOPT_ENG, MENUOPT_SPA);
        else uncheckmx(MENUOPT_SPA);
    acase MENUITEM_CALIBRATE:
        ghost(which,
            !lpdi
         || (   machine != ARCADIA
             && machine != INTERTON
             && machine != ELEKTOR
             && machine != MALZAK
             && machine != ASTROWARS
             && machine != GALAXIA
        )   ); // eg. INSTRUCTOR, PIPBUG, CD2650, CHAOS
#endif
}   }

EXPORT void updatesmlgad(int which, ULONG state, FLAG now)
{   if (!MainWindowPtr || !showtoolbar)
    {   return; // important!
    }

#ifdef AMIGA
    if (now)
    {   detachsmlgads();
    }
#endif

    switch (which)
    {
    case GADPOS_AUTOFIRE:
        if
        (   machine == ARCADIA
         || machine == INTERTON
         || machine == ELEKTOR
         || machine == MALZAK
         || machine == ASTROWARS
         || machine == GALAXIA
        )
        {   setbutton(which, TRUE,  (FLAG) state);
        } else
        {   setbutton(which, FALSE, (FLAG) state);
        }
    acase GADPOS_COLLISIONS:
        if
        (   machine == ARCADIA
         || machine == INTERTON
         || machine == ELEKTOR
         || machine == MALZAK
         || machine == ASTROWARS
         || machine == GALAXIA
        )
        {   setbutton(which, TRUE,  (FLAG) state);
        } else
        {   setbutton(which, FALSE, (FLAG) state);
        }
	acase GADPOS_FLAGLINE:
        if
        (   machine == ARCADIA
         || machine == INTERTON
         || machine == ELEKTOR
         || machine == CD2650
        )
        {   setbutton(which, TRUE,  (FLAG) state);
        } else
        {   setbutton(which, FALSE, (FLAG) state);
        }
    acase GADPOS_SWAPPED:
        if
        (   machine == ARCADIA
         || machine == INTERTON
         || machine == ELEKTOR
         || machine == MALZAK
         || machine == ASTROWARS
         || machine == GALAXIA
        )
        {   setbutton(which, TRUE,  (FLAG) state);
        } else
        {   setbutton(which, FALSE, (FLAG) state);
        }
    acase GADPOS_PAUSED:
    case GADPOS_SOUND:
    case GADPOS_TURBO:
        setbutton(which, TRUE, (FLAG) state);
    acase GADPOS_MACRO:
#ifdef AMIGA
        DISCARD SetGadgetAttrs
        (   gadgets[GID_BU1], MainWindowPtr, NULL,
            GA_Image,    smalltbimage[recmode],
            GA_Disabled, crippled,
        TAG_DONE);
        // this autorefreshes
#endif
#ifdef WIN32
        if (crippled)
        {   DISCARD SendMessage(hToolbar, TB_SETSTATE, ID_MACRO_TOGGLEPLAYING,   0);
            DISCARD SendMessage(hToolbar, TB_SETSTATE, ID_MACRO_TOGGLERECORDING, 0);
        } elif (recmode == RECMODE_PLAY)
        {   DISCARD SendMessage(hToolbar, TB_SETSTATE, ID_MACRO_TOGGLEPLAYING,   TBSTATE_ENABLED | TBSTATE_CHECKED);
            DISCARD SendMessage(hToolbar, TB_SETSTATE, ID_MACRO_TOGGLERECORDING, 0);
        } elif (recmode == RECMODE_RECORD)
        {   DISCARD SendMessage(hToolbar, TB_SETSTATE, ID_MACRO_TOGGLEPLAYING,   0);
            DISCARD SendMessage(hToolbar, TB_SETSTATE, ID_MACRO_TOGGLERECORDING, TBSTATE_ENABLED | TBSTATE_CHECKED);
        } else
        {   // assert(recmode == RECMODE_NORMAL);
            DISCARD SendMessage(hToolbar, TB_SETSTATE, ID_MACRO_TOGGLEPLAYING,   0);
            DISCARD SendMessage(hToolbar, TB_SETSTATE, ID_MACRO_TOGGLERECORDING, TBSTATE_ENABLED);
        }
#endif
    }

#ifdef AMIGA
    if (now)
    {   attachsmlgads();
    }
#endif
}

EXPORT void updatesmlgads(void)
{   if (!MainWindowPtr)
    {   return; // important!
    }

#ifdef AMIGA
    if (!showtoolbar)
    {   return;
    }
    detachsmlgads();
#endif

    // controllers
    updatesmlgad(GADPOS_AUTOFIRE   , autofire   , FALSE);
    updatesmlgad(GADPOS_SWAPPED    , swapped    , FALSE);
    // speed
    updatesmlgad(GADPOS_PAUSED     , paused     , FALSE);
    updatesmlgad(GADPOS_TURBO      , turbo      , FALSE);
    // sprites
    updatesmlgad(GADPOS_COLLISIONS , collisions , FALSE);
    // graphics
    updatesmlgad(GADPOS_FLAGLINE   , flagline   , FALSE);
    // sound
    updatesmlgad(GADPOS_SOUND      , sound      , FALSE);

    updatesmlgad(GADPOS_MACRO      , (ULONG) recmode, TRUE );

#ifdef WIN32
    if
    (   machine == ARCADIA
     || machine == INTERTON
     || machine == ELEKTOR
     || machine == INSTRUCTOR
     || machine == ASTROWARS
     || machine == GALAXIA
     || machine == MALZAK
    )
    {   DISCARD SendMessage(hToolbar, TB_SETSTATE, (UINT) ID_CONSOLE_START, TBSTATE_ENABLED);
        DISCARD SendMessage(hToolbar, TB_SETSTATE, (UINT) ID_CONSOLE_A,     TBSTATE_ENABLED);
        if (machine == INTERTON)
        {   DISCARD SendMessage(hToolbar, TB_SETSTATE, (UINT) ID_CONSOLE_B, 0              );
        } else // eg. ARCADIA, ELEKTOR, INSTRUCTOR, coin-ops
        {   DISCARD SendMessage(hToolbar, TB_SETSTATE, (UINT) ID_CONSOLE_B, TBSTATE_ENABLED);
    }   }
    else // eg. PIPBUG, CD2650, CHAOS, DOLPHIN
    {   DISCARD SendMessage(hToolbar, TB_SETSTATE, (UINT) ID_CONSOLE_START, 0              );
        DISCARD SendMessage(hToolbar, TB_SETSTATE, (UINT) ID_CONSOLE_A,     0              );
        DISCARD SendMessage(hToolbar, TB_SETSTATE, (UINT) ID_CONSOLE_B,     0              );
    }
    if
    (   machine == ELEKTOR
     || machine == INSTRUCTOR
     || machine == ASTROWARS
     || machine == GALAXIA
     || memmap  == MEMMAP_MALZAK2
    )
    {   DISCARD SendMessage(hToolbar, TB_SETSTATE, (UINT) ID_CONSOLE_RESET, TBSTATE_ENABLED);
    } else // eg. ARCADIA, INTERTON, PIPBUG, CD2650, CHAOS, DOLPHIN, Malzak 1
    {   DISCARD SendMessage(hToolbar, TB_SETSTATE, (UINT) ID_CONSOLE_RESET, 0              );
    }
#endif
#ifdef AMIGA
    if
    (   machine == ARCADIA
     || machine == INTERTON
     || machine == ELEKTOR
     || machine == INSTRUCTOR
     || machine == ASTROWARS
     || machine == GALAXIA
     || machine == MALZAK
    )
    {   SetSpeedButtonNodeAttrs
        (   SmallSpeedBarNodePtr[7],
            SBNA_Disabled, FALSE,
        TAG_DONE);
        SetSpeedButtonNodeAttrs
        (   SmallSpeedBarNodePtr[8],
            SBNA_Disabled, FALSE,
        TAG_DONE);
        if (machine == INTERTON)
        {   SetSpeedButtonNodeAttrs
            (   SmallSpeedBarNodePtr[9],
                SBNA_Disabled, TRUE,
            TAG_DONE);
        } else // eg. ARCADIA, ELEKTOR, INSTRUCTOR, coin-ops
        {   SetSpeedButtonNodeAttrs
            (   SmallSpeedBarNodePtr[9],
                SBNA_Disabled, FALSE,
            TAG_DONE);
    }   }
    else // eg. PIPBUG, CD2650, CHAOS, DOLPHIN
    {   SetSpeedButtonNodeAttrs
        (   SmallSpeedBarNodePtr[7],
            SBNA_Disabled, TRUE,
        TAG_DONE);
        SetSpeedButtonNodeAttrs
        (   SmallSpeedBarNodePtr[8],
            SBNA_Disabled, TRUE,
        TAG_DONE);
        SetSpeedButtonNodeAttrs
        (   SmallSpeedBarNodePtr[9],
            SBNA_Disabled, TRUE,
        TAG_DONE);
    }
    if
    (   machine == ELEKTOR
     || machine == INSTRUCTOR
     || machine == ASTROWARS
     || machine == GALAXIA
     || machine == MALZAK
    )
    {   SetSpeedButtonNodeAttrs
        (   SmallSpeedBarNodePtr[10],
            SBNA_Disabled, FALSE,
        TAG_DONE);
    } else // eg. ARCADIA, INTERTON, PIPBUG, CD2650, CHAOS, DOLPHIN
    {   SetSpeedButtonNodeAttrs
        (   SmallSpeedBarNodePtr[10],
            SBNA_Disabled, TRUE,
        TAG_DONE);
    }

    attachsmlgads();
#endif
}

EXPORT void updatemenus(void)
{   int i;
#ifdef WIN32
    HMENU MenuColumnPtr;
#endif

    if (!MainWindowPtr)
    {   return; // important!
    }

#ifdef AMIGA
    if (!MenuPtr)
    {   return; // not for IBM-PC, because we have to update popup menu
    }

    if (machine == ARCADIA)
    {   DISCARD  OnMenu(MainWindowPtr, FULLMENUNUM(MN_PROJECT,  IN_SAVEEAP, NOSUB));
    } else
    {   DISCARD OffMenu(MainWindowPtr, FULLMENUNUM(MN_PROJECT,  IN_SAVEEAP, NOSUB));
    }

    switch (machine)
    {
    case ARCADIA:
    case INTERTON:
    case ELEKTOR:
    case MALZAK:
    case ASTROWARS:
    case GALAXIA:
        DISCARD  OnMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_LEFT,  NOSUB));
        DISCARD  OnMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_RIGHT, NOSUB));
    adefault: // PIPBUG, INSTRUCTOR, CD2650, CHAOS, DOLPHIN, PHUNSY
        DISCARD OffMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_LEFT,  NOSUB));
        DISCARD OffMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_RIGHT, NOSUB));
    }

    switch (machine)
    {
    case ARCADIA:
    case INTERTON:
    case ELEKTOR:
    case INSTRUCTOR:
    case MALZAK:
    case ASTROWARS:
    case GALAXIA:
        DISCARD  OnMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_SPRITES, NOSUB));
    adefault: // PIPBUG, CD2650, CHAOS, DOLPHIN, PHUNSY
        DISCARD OffMenu(MainWindowPtr, FULLMENUNUM(MN_SETTINGS, IN_SPRITES, NOSUB));
    }
#endif
#ifdef WIN32
    MenuColumnPtr = GetSubMenu(MenuPtr, 0); // File menu
    if (machine == ARCADIA)
    {   EnableMenuItem(MenuColumnPtr, 11, MF_BYPOSITION | MF_ENABLED);
        MenuColumnPtr = GetSubMenu(PopUpMenuPtr, 0); // File menu
        EnableMenuItem(MenuColumnPtr, 11, MF_BYPOSITION | MF_ENABLED);
    } else
    {   EnableMenuItem(MenuColumnPtr, 11, MF_BYPOSITION | MF_GRAYED);
        MenuColumnPtr = GetSubMenu(PopUpMenuPtr, 0); // File menu
        EnableMenuItem(MenuColumnPtr, 11, MF_BYPOSITION | MF_GRAYED);
    }

    MenuColumnPtr = GetSubMenu(MenuPtr, 7); // Options menu
    switch (machine)
    {
    case ARCADIA:
    case INTERTON:
    case ELEKTOR:
    case MALZAK:
    case ASTROWARS:
    case GALAXIA:
        EnableMenuItem(MenuColumnPtr, 0, MF_BYPOSITION | MF_ENABLED);
        EnableMenuItem(MenuColumnPtr, 1, MF_BYPOSITION | MF_ENABLED);
        MenuColumnPtr = GetSubMenu(PopUpMenuPtr, 7); // Options menu
        EnableMenuItem(MenuColumnPtr, 0, MF_BYPOSITION | MF_ENABLED);
        EnableMenuItem(MenuColumnPtr, 1, MF_BYPOSITION | MF_ENABLED);
    adefault: // PIPBUG, INSTRUCTOR, CD2650, CHAOS, DOLPHIN, PHUNSY
        EnableMenuItem(MenuColumnPtr, 0, MF_BYPOSITION | MF_GRAYED);
        EnableMenuItem(MenuColumnPtr, 1, MF_BYPOSITION | MF_GRAYED);
        MenuColumnPtr = GetSubMenu(PopUpMenuPtr, 7); // Options menu
        EnableMenuItem(MenuColumnPtr, 0, MF_BYPOSITION | MF_GRAYED);
        EnableMenuItem(MenuColumnPtr, 1, MF_BYPOSITION | MF_GRAYED);
    }

    MenuColumnPtr = GetSubMenu(MenuPtr, 7); // Options menu
    switch (machine)
    {
    case ARCADIA:
    case INTERTON:
    case ELEKTOR:
    case INSTRUCTOR:
    case MALZAK:
    case ASTROWARS:
    case GALAXIA:
        EnableMenuItem(MenuColumnPtr, 8, MF_BYPOSITION | MF_ENABLED);
        MenuColumnPtr = GetSubMenu(PopUpMenuPtr, 7); // Options menu
        EnableMenuItem(MenuColumnPtr, 8, MF_BYPOSITION | MF_ENABLED);
    adefault: // PIPBUG, CD2650, CHAOS, DOLPHIN, PHUNSY
        EnableMenuItem(MenuColumnPtr, 8, MF_BYPOSITION | MF_GRAYED);
        MenuColumnPtr = GetSubMenu(PopUpMenuPtr, 7); // Options menu
        EnableMenuItem(MenuColumnPtr, 8, MF_BYPOSITION | MF_GRAYED);
    }
#endif

    for (i = 0; i < MENUITEMS; i++)
    {   updatemenu(i);
}   }

EXPORT void free_iobuffer(void)
{   // assert(IOBuffer);
    free(IOBuffer);
    IOBuffer = NULL;
#ifdef VERBOSE
    T1; textpen(TEXTPEN_VERBOSE); zprintf("Freed I/O buffer.\n"); T2;
#endif
}

EXPORT UBYTE ReadByte(void)
{   UBYTE result;

    result = IOBuffer[offset++];
    return result;
}
EXPORT void WriteByte(UBYTE what)
{   IOBuffer[offset++]   = what;
}

EXPORT void docommand(int which)
{   int i;

    switch (which)
    {
    case MENUITEM_CLIENT:
        if (connected == NET_OFF)
        {   NetPlay(NET_CLIENT);
            if (connected != NET_OFF)
            {    project_reset(TRUE); /* reset machine when netplay connected */
        }   }
        else
        {   NetPlay(NET_OFF);
        }
        updatemenu(MENUITEM_CLIENT);
    acase MENUITEM_COLLISIONS:
        updatesmlgad(GADPOS_COLLISIONS, collisions, TRUE);
        updatemenu(MENUITEM_COLLISIONS);
    acase MENUITEM_CONFINE:
        updatemenu(MENUITEM_CONFINE);
        updatepointer(TRUE);
    acase MENUITEM_DEMULTIPLEX:
        updatemenu(MENUITEM_DEMULTIPLEX);
        if (paused)
        {   redrawscreen();
        }
    acase MENUITEM_FLAGLINE:
        switch (machine)
        {
        case CD2650:
            changed = TRUE;
#ifdef WIN32
        acase INTERTON:
        case ELEKTOR:
            make_stars();
#endif
        }
        updatesmlgad(GADPOS_FLAGLINE, flagline, TRUE);
        updatemenu(MENUITEM_FLAGLINE);
        if (paused || changed)
        {   redrawscreen();
        }
    acase MENUITEM_CODINGGUIDE:
        strcpy(ggstring, "http://amigan.yatho.com/");
        strcat(ggstring, machines[machine].codingguide);
        strcat(ggstring, "-coding.txt");
        openurl(ggstring);
    acase MENUITEM_GAMINGGUIDE:
        switch (machine)
        {
        case ARCADIA:
        case INTERTON:
        case ELEKTOR:
            if (machine == ARCADIA)
            {   strcpy(ggstring, "http://amigan.1emu.net/agg/");
            } elif (machine == INTERTON)
            {   strcpy(ggstring, "http://amigan.1emu.net/igg/");
            } elif (machine == ELEKTOR)
            {   strcpy(ggstring, "http://amigan.1emu.net/egg/");
            }
            if (ggnode[0])
            {   strcat(ggstring, "#");
                strcat(ggstring, ggnode);
            }
        acase PIPBUG:
            strcat(ggstring, "http://amigan.yatho.com/p-coding.txt");
        acase INSTRUCTOR:
        case MALZAK:
        case ASTROWARS:
        case GALAXIA:
        case CHAOS:
            strcpy(ggstring, "http://amigan.yatho.com/s-coding.txt");
        acase CD2650:
            strcpy(ggstring, "http://amigan.yatho.com/c-coding.txt");
        }
        openurl(ggstring);
    acase MENUITEM_USEGUIDEPIXEL:
        updatemenu(MENUITEM_USEGUIDEPIXEL);
        if (paused)
        {   flipguidepixel();
            redrawscreen();
        }
    acase MENUITEM_INTERRUPT:
        updatemenu(MENUITEM_INTERRUPT);
        T1; textpen(TEXTPEN_DEBUG);
        if (!(psu & PSU_II) || forceinterrupt >= 1)
        {   zinterrupt = TRUE;
            zprintf
            (   GetCatalogStr
                (   CatalogPtr,
                    MSG_ENGINE_GENERATED,
                    "Generated user-requested interrupt.\n\n"
            )   );
        } else
        {   zprintf
            (   GetCatalogStr
                (   CatalogPtr,
                    MSG_ENGINE_SUPPRESSED,
                    "Suppressed user-requested interrupt.\n\n"
            )   );
        }
        T2;
    acase MENUITEM_LEFT:
        updatemenu(MENUITEM_LEFT);
        updatemenu(MENUITEM_CONFINE);
        updatepointer(TRUE);
#ifdef AMIGA
        setjoyports();
#endif
        // overkill, we only need to turn off one
        ff_off(0);
        ff_off(1);
    acase MENUITEM_RIGHT:
        updatemenu(MENUITEM_RIGHT);
        updatemenu(MENUITEM_CONFINE);
        updatepointer(TRUE);
#ifdef AMIGA
        setjoyports();
#endif
        // overkill, we only need to turn off one
        ff_off(0);
        ff_off(1);
    acase MENUITEM_POINTER:
        updatemenu(MENUITEM_POINTER);
        updatepointer(TRUE);
    acase MENUITEM_ANSI:
        updatemenu(MENUITEM_ANSI);
        if (!useansi)
        {   useansi = TRUE;
            textpen(TEXTPEN_DEFAULT);
            useansi = FALSE;
        }
    acase MENUITEM_RESETTOMONITOR:
        switch (machine)
        {
        case ELEKTOR:
            iar = 0x23; // gives blue background with yellow "IIII" at bottom
        acase PIPBUG:
        case CD2650:
        case PHUNSY:
            iar = 0;
        acase INSTRUCTOR:
            iar = 0x1800;
        }
        NetPlay(NET_OFF);
        updatemenu(MENUITEM_SERVER);
        updatemenu(MENUITEM_CLIENT);
    acase MENUITEM_REVERT:
        if (crippled || !game) return;
        macro_stop();
        DISCARD engine_load(FALSE); // arguably TRUE might be better
    acase MENUITEM_QUICKLOAD:
        if (crippled) return;
        macro_stop();
        changefilepart("QUICKSAV.COS");
#ifdef VERBOSE
        T1; zprintf("Attempting to quickload %s...\n", fn_game); T2;
#endif
        if (!(engine_load(TRUE)))
        {   beep();
        }
    acase MENUITEM_QUICKSAVE:
        if (crippled) return;
        changefilepart("QUICKSAV.COS");
#ifdef VERBOSE
        T1; zprintf("Attempting to quicksave %s...\n", fn_game); T2;
#endif
        engine_save(KIND_COS, TRUE);
    acase MENUITEM_SAVEROM:
        switch (machine)
        {
        case ARCADIA:
        case INTERTON:
            project_save(KIND_BIN);
        acase ELEKTOR:
        case PIPBUG:
        case INSTRUCTOR:
        case CD2650:
        case PHUNSY:
            project_save(KIND_PGM);
        }
    acase MENUITEM_SWAPPED:
        updatesmlgad(GADPOS_SWAPPED, swapped, TRUE);
        updatemenu(MENUITEM_SWAPPED);
    acase MENUITEM_AUTOFIRE:
        updatesmlgad(GADPOS_AUTOFIRE, autofire, TRUE);
        updatemenu(MENUITEM_AUTOFIRE);
    acase MENUITEM_TOOLBAR:
        closewindow();
#ifdef WIN32
        if (showtoolbar && !fullscreen)
        {   winleftx -= 6;
        }
#endif
        openwindow();
#ifdef WIN32
        redrawscreen();
        updatepointer(TRUE);
#endif
    acase MENUITEM_TURBO:
        if (!turbo)
        {   waittill = thetime(); // makes it effectively IMMEDIATELY
        }
        updatesmlgad(GADPOS_TURBO, turbo, TRUE);
        updatemenu(MENUITEM_TURBO);
    acase MENUITEM_UNDITHER:
        updatemenu(MENUITEM_UNDITHER);
        if (undither && (whichgame == ESCAPEPOS || whichgame == ROBOTKILLERPOS))
        {   undithernow = TRUE;
        } else
        {   undithernow = FALSE;
        }
    acase MENUITEM_HIGHRES:
        updatemenu(MENUITEM_HIGHRES);
        changed = TRUE;
        if ((machine == PIPBUG || machine == CHAOS) && paused)
        {   draw_pipbug();  // both of these are needed
            redrawscreen(); // (order-dependent)
        }
    acase MENUITEM_CORNERS:
        updatemenu(MENUITEM_CORNERS);
        if (machine == PIPBUG)
        {   draw_pipbug();
        }
        redrawscreen();
    acase MENUITEM_UNLIT:
        updatemenu(MENUITEM_UNLIT);
        if (machine == PIPBUG)
        {   draw_pipbug();
        }
        redrawscreen();
#ifdef WIN32
    acase MENUITEM_FULLSCREEN:
        // no point doing updatemenu(MENUITEM_FULLSCREEN),
        // as we're about to close the window
        closewindow();
        winleftx =
        wintopy  = -1;
        
        showtitlebar =
        showtoolbar  =
        showmenubar  =
        showsidebar  =
        showpointer  = fullscreen ? FALSE : TRUE;
        stretch      = fullscreen;

        openwindow();
        redrawscreen();
        updatepointer(TRUE);
    acase MENUITEM_MENUBAR:
        // no point doing updatemenu(MENUITEM_MENUBAR),
        // as we're about to close the window
        closewindow();
        if (!fullscreen)
        {   winleftx -= 3;
        }
        openwindow();
        redrawscreen();
        updatepointer(TRUE);
    acase MENUITEM_STRETCH:
        updatemenu(MENUITEM_STRETCH);
        if (fullscreen)
        {   calcsize();
#ifdef NEWWAY
            makedisplay();
#endif
            clearscreen();
            redrawscreen();
            updatepointer(TRUE);
        }
    acase MENUITEM_PRIORITY:
        updatemenu(MENUITEM_PRIORITY);
        setpri();
#endif
    acase COMMAND_LOADER:
        // assert(machine == ELEKTOR);
        // assert(memmap == MEMMAP_F);
        T1; textpen(TEXTPEN_SWAPEPROMS);
        for (i = 0; i < 128; i++)
        {   memory[0x1C00 + i] = e_loader[i];
            memory[0x1E00 + i] = e_loader[128 + i];
        }
        zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_ENGINE_INSERTLOADER,
                "Saver EPROM removed.\nLoader EPROM inserted.\n\n"
        )   );
        T2;
        loader = TRUE;
    acase COMMAND_SAVER:
        // assert(machine == ELEKTOR);
        // assert(memmap == MEMMAP_F);
        T1; textpen(TEXTPEN_SWAPEPROMS);
        for (i = 0; i < 128; i++)
        {   memory[0x1C00 + i] = e_saver[i];
            memory[0x1E00 + i] = e_saver[128 + i];
        }
        zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_ENGINE_INSERTSAVER,
                "Loader EPROM removed.\nSaver EPROM inserted.\n\n"
        )   );
        T2;
        loader = FALSE;
    acase MENUITEM_SERVER:
        if (connected == NET_OFF)
        {   NetPlay(NET_SERVER);
            if (connected != NET_OFF)
            {   project_reset(TRUE); /* reset machine when netplay connected */
        }   }
        else
        {   NetPlay(NET_OFF);
        }
        updatemenu(MENUITEM_SERVER);
    acase MENUITEM_SOUND:
        if (sound)
        {   sound_on();
        } else
        {   sound_off(TRUE);
            // We have to force it, because sound variable is already FALSE
        }
        updatesmlgad(GADPOS_SOUND, sound, TRUE);
        updatemenu(MENUITEM_SOUND);
    acase MENUITEM_STEP:
        step = TRUE;
        emu_unpause();
        // updatemenu(MENUITEM_STEP); not required here
    acase MENUITEM_STEPOVER:
        // shouldn't we resolve mirrors first?
        if ((memory[iar] & 0x78) == 0x38) // %x0111xxx
        // we don't use op because we want the NEXT instruction, not the LAST one
        {   T1; textpen(TEXTPEN_TRACE);
            DISCARD getfriendly((int) iar);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_STEPPINGOVER,
                    "Stepping over subroutine call (%s) at %s.\n"
                ),
                opcodename[memory[iar]],
                friendly
            );
            if (!traceverbose)
            {   zprintf("\n");
            }
            T2;
            memflags[(iar & PAGE) + ((iar + table_sz[memory[iar]]) & NONPAGE)] |= STEPPOINT;
            emu_unpause();
            // updatemenu(MENUITEM_STEPOVER); not required here
        } else
        {   docommand(MENUITEM_STEP);
        }
    acase MENUITEM_TITLEBAR:
        // no point in doing updatemenu(MENUITEM_TITLEBAR);
        closewindow();
#ifdef WIN32
        if (showtitlebar && !fullscreen)
        {   winleftx -= 4;
        }
#endif
        openwindow();
#ifdef WIN32
        redrawscreen();
        updatepointer(TRUE);
#endif
    acase COMMAND_ARCADIA:
        // assert(!crippled);
        if (machine != ARCADIA)
        {   strcpy(fn_game, pathpart);
            changemachine(ARCADIA, TRUE);
            updatemenus();
            updatebiggads();
            updatesmlgads();
            settitle();
            redrawscreen();
        }
    acase COMMAND_INTERTON:
        // assert(!crippled);
        if (machine != INTERTON)
        {   strcpy(fn_game, pathpart);
            changemachine(INTERTON, TRUE);
            updatemenus();
            updatebiggads();
            updatesmlgads();
            settitle();
            redrawscreen();
        }
    acase COMMAND_ELEKTOR:
        // assert(!crippled);
        if (machine != ELEKTOR)
        {   strcpy(fn_game, pathpart);
            changemachine(ELEKTOR, TRUE);
            updatemenus();
            updatebiggads();
            updatesmlgads();
            settitle();
            redrawscreen();
        }
    acase COMMAND_PIPBUG:
        // assert(!crippled);
        if (machine != PIPBUG)
        {   strcpy(fn_game, pathpart);
            changemachine(PIPBUG, TRUE);
            updatemenus();
            updatebiggads();
            updatesmlgads();
            settitle();
            redrawscreen();
        }
    acase COMMAND_INSTRUCTOR:
        // assert(!crippled);
        if (machine != INSTRUCTOR)
        {   strcpy(fn_game, pathpart);
            changemachine(INSTRUCTOR, TRUE);
            updatemenus();
            updatebiggads();
            updatesmlgads();
            settitle();
            redrawscreen();
        }
    acase COMMAND_CD2650:
        // assert(!crippled);
        if (machine != CD2650)
        {   strcpy(fn_game, pathpart);
            changemachine(CD2650, TRUE);
            updatemenus();
            updatebiggads();
            updatesmlgads();
            settitle();
            redrawscreen();
        }
    acase COMMAND_ASTROWARS:
        // assert(!crippled);
        if (machine != ASTROWARS)
        {   changemachine(ASTROWARS, TRUE);
            updatemenus();
            updatebiggads();
            updatesmlgads();
            settitle();
            redrawscreen();
        }
    acase COMMAND_GALAXIA:
        // assert(!crippled);
        if (machine != GALAXIA)
        {   changemachine(GALAXIA, TRUE);
            updatemenus();
            updatebiggads();
            updatesmlgads();
            settitle();
            redrawscreen();
        }
    acase COMMAND_MALZAK1:
        // assert(!crippled);
        if (memmap != MEMMAP_MALZAK1)
        {   memmap = MEMMAP_MALZAK1;
            changemachine(MALZAK, TRUE);
            updatemenus();
            updatebiggads();
            updatesmlgads();
            settitle();
            redrawscreen();
        }
    acase COMMAND_MALZAK2:
        // assert(!crippled);
        if (memmap != MEMMAP_MALZAK2)
        {   memmap = MEMMAP_MALZAK2;
            changemachine(MALZAK, TRUE);
            updatemenus();
            updatebiggads();
            updatesmlgads();
            settitle();
            redrawscreen();
        }
    acase COMMAND_CHAOS:
        // assert(!crippled);
        if (memmap != MEMMAP_CHAOS)
        {   memmap = MEMMAP_CHAOS;
            changemachine(CHAOS, TRUE);
            updatemenus();
            updatebiggads();
            updatesmlgads();
            settitle();
            redrawscreen();
        }
    acase COMMAND_DOLPHIN:
        // assert(!crippled);
        if (memmap != MEMMAP_DOLPHIN)
        {   memmap = MEMMAP_DOLPHIN;
            changemachine(DOLPHIN, TRUE);
            updatemenus();
            updatebiggads();
            updatesmlgads();
            settitle();
            redrawscreen();
        }
    acase COMMAND_PHUNSY:
        // assert(!crippled);
        if (memmap != MEMMAP_PHUNSY)
        {   memmap = MEMMAP_PHUNSY;
            changemachine(PHUNSY, TRUE);
            updatemenus();
            updatebiggads();
            updatesmlgads();
            settitle();
            redrawscreen();
        }
    acase MENUITEM_CLI:
        T1; textpen(TEXTPEN_CLIOUTPUT);
        if (cli)
        {   zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_CLIENABLED,
                    "Debugger CLI enabled (\"H\" for help).\n\n"
            )   );
        } else
        {   zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_CLIDISABLED,
                    "Debugger CLI disabled.\n\n"
            )   );
        }
        T2;
        updatemenu(MENUITEM_CLI);
        if (cli)
        {   emu_pause(TRUE);
        }
    acase MENUITEM_IGNORE:
        T1; textpen(TEXTPEN_TRACE);
        zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_ENGINE_IGNORING,
                "Ignoring %s instruction at $%X.\n"
            ),
            opcodename[memory[iar]],
            (unsigned int) iar
        );
        iar = (iar & PAGE) + ((iar + table_sz[memory[iar]]) & NONPAGE);
        if (traceverbose)
        {   DISCARD getfriendly((int) iar);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_NEXTINSTRUCTION,
                    "Next instruction is %s at %s.\n\n"
                ),
                opcodename[memory[iar]],
                friendly
            );
        } else
        {   zprintf("\n");
        }
        T2;
    acase MENUITEM_RUNTORAST:
        // assert(machine != PIPBUG && machine != INSTRUCTOR && machine != CD2650);
        clearruntos();
        rastn          = rast + 1;
        if (machine == ARCADIA)
        {   if (rastn >= UVI_RASTLINES)
            {   rastn = 0;
        }   }
        else
        {   // assert(machine == INTERTON || machine == ELEKTOR);
            if (rastn >= PVI_RASTLINES)
            {   rastn = 0;
        }   }
        emu_unpause();
    acase MENUITEM_RUNTORASTN:
        // assert(machine != PIPBUG && machine != INSTRUCTOR);
        clearruntos();
        debug_setrastn(); // if they cancel this requester, should we still clear all runtos?
        emu_unpause();
    acase MENUITEM_RUNTODMA:
        // assert(machine == ARCADIA);
        clearruntos();
        runtodma       = TRUE;
        emu_unpause();
    acase MENUITEM_RUNTOINTERRUPT:
        // assert(machine == INTERTON || machine == ELEKTOR);
        clearruntos();
        runtointerrupt = TRUE;
        emu_unpause();
    acase MENUITEM_RUNTOFRAME:
        clearruntos();
        runtoframe     = TRUE;
        emu_unpause();
    acase MENUITEM_RUNTOSUBEND:
        T1; textpen(TEXTPEN_DEBUG);
        if ((psu & PSU_SP) == 0)
        {   zprintf
            (   (STRPTR) GetCatalogStr(CatalogPtr,
                    MSG_NOT_IN_SUBROUTINE,
                    "Not currently inside a subroutine.\n\n"
            )   ); // but what if SP has just overflowed, we might actually be inside a subroutine!
        } else
        {   zprintf
            (   (STRPTR) GetCatalogStr(CatalogPtr,
                    MSG_RUNNING_TO,
                    "Running to $%X.\n\n"
                ),
                ras[(psu & PSU_SP) - 1]
            );
            clearruntos();
            memflags[ras[(psu & PSU_SP) - 1]] |= STEPPOINT;
            emu_unpause();
        }
        T2;
    acase MENUITEM_RUNTOLOOPEND:
        clearruntos();
        runtoloopend = TRUE;
        emu_unpause();
    acase MENUITEM_DOTSPACE:
        if (dotspace)
        {   p_chars[0][3]         =
            c_chars[0][0x20][3]   =
            c_chars[0][0x60][3]   =
            c_chars[1][0x20][3]   =
            c_chars[1][0x60][3]   =
            phunsy_chars[0x20][4] = 0x04; // ..#..
        } else
        {   p_chars[0][3]         =
            c_chars[0][0x20][3]   =
            c_chars[0][0x60][3]   =
            c_chars[1][0x20][3]   =
            c_chars[1][0x60][3]   =
            phunsy_chars[0x20][4] = 0x00; // .....
        }
        changed = TRUE;
}   }

EXPORT void fixextension(STRPTR extension, STRPTR passedfn, FLAG force)
{   FLAG   found = FALSE;
    int    i;
    size_t length;

    // This routine should really only be used on the file component of the pathname.

    // force means we will replace an existing extension
    //  (typically used before showing an ASL save requester)
    // otherwise, we just add an extension if there is no recognised extension
    //  (typically used after showing an ASL load requester)

    length = strlen(passedfn);
    if (force) // saving
    {   if (length >= 1)
        {   // find extension
            for (i = (int) (length - 1); i >= 0; i--)
            {   /* if (passedfn[i] == '/' || passedfn[i] == '\\' || passedfn[i] == ':')
                {   break;
                } else */ if (passedfn[i] == '.')
                {   found = TRUE;
                    break;
            }   }
            if (found)
            {   passedfn[i] = 0;
        }   }
        strcat(passedfn, extension);
    } else // loading
    {   for (i = 0; i < KINDS; i++) // this is why filekind[].extension needs to be fully populated
        {   if
            (   length > strlen(filekind[i].extension)
             && !stricmp(&passedfn[length - strlen(filekind[i].extension)], filekind[i].extension)
            )
            {   found = TRUE;
                break; // for speed
        }   }
        if (!found)
        {   // if there is no recognized extension, add an appropriate extension
            strcat(passedfn, extension); // add extension
}   }   }

EXPORT ULONG crc32(UBYTE* address, int thesize)
{   register ULONG crc = 0xFFFFFFFF;
             int   i;

    for (i = 0; i < thesize; i++)
    {   crc = ((crc >> 8) & 0x00FFFFFF) ^ crcTable[(crc ^ *address++) & 0xFF];
    }
    return crc ^ 0xFFFFFFFF;
}

EXPORT void generate_crctable(void)
{   ULONG crc;
    int   i, j;

    for (i = 0; i < 256; i++)
    {   crc = (ULONG) i;
        for (j = 8; j > 0; j--)
        {   if (crc & 1)
            {   crc = (crc >> 1) ^ 0xEDB88320L;
            } else
            {   crc >>= 1;
        }   }
        crcTable[i] = crc;
}   }

EXPORT void runto(void)
{   if (inframe && !crippled)
    {   cripple();
    }
    emu_pause(TRUE);

    update_monitor(FALSE);

    while (paused && !cli)
    {   checkinput();
}   }

EXPORT void hex1(char* out, UBYTE value)
{   ULONG digit;

    // This routine deliberately doesn't append an EOS, for speed.

    digit = value / 16;
    if (digit <= 9)
    {   *(out) = (char) ('0' + digit);
    } else
    {   *(out) = (char) ('A' + digit - 10);
    }

    digit = value % 16;
    if (digit <= 9)
    {   *(out + 1) = (char) ('0' + digit);
    } else
    {   *(out + 1) = (char) ('A' + digit - 10);
}   }

EXPORT void hex2(char* out, UWORD value)
{   ULONG digit;

    // This routine deliberately doesn't append an EOS, for speed.

    digit = value / 4096;
    if (digit <= 9)
    {   *(out) = (char) ('0' + digit);
    } else
    {   *(out) = (char) ('A' + digit - 10);
    }

    digit = (value % 4096) / 256;
    if (digit <= 9)
    {   *(out + 1) = (char) ('0' + digit);
    } else
    {   *(out + 1) = (char) ('A' + digit - 10);
    }

    digit = (value % 256) / 16;
    if (digit <= 9)
    {   *(out + 2) = (char) ('0' + digit);
    } else
    {   *(out + 2) = (char) ('A' + digit - 10);
    }

    digit = value % 16;
    if (digit <= 9)
    {   *(out + 3) = (char) ('0' + digit);
    } else
    {   *(out + 3) = (char) ('A' + digit - 10);
}   }

EXPORT void decimal(char* out, unsigned long value)
{   ULONG calc,
          i,
          where   = 0;
    FLAG  started = FALSE;

    for (i = ONE_BILLION; i >= 1; i /= 10)
    {   calc = value / i;
        // assert(calc < 10);
        if (calc || started || i == 1)
        {   *(out + where) = (char) ('0' + calc);
            started = TRUE;
        } else
        {   *(out + where) = ' ';
        }
        where++;
        value %= i;
    }

    *(out + where) = 0;
}

EXPORT void engine_setup(void)
{   int i, j;

    srand((unsigned int) time(NULL));

    generate_crctable();

    TimeStr[2]  =
    TimeStr[5]  = ':';
    TimeStr[8]  = '.';
    TimeStr[11] = 0;
    ClockStr[COMMASTRLEN] =
    FrameStr[COMMASTRLEN] = EOS;
    for (i = 0; i < 10; i++)
    {   OneStr[i][1] = EOS;
    }
    for (i = 0; i < 8; i++)
    {   TwoStr[i][2] = EOS;
    }
    for (i = 0; i < 17; i++)
    {   FourStr[i][4] = EOS;
    }
    for (i = 0; i < 24; i++)
    {   FiveStr[i][2] = ',';;
        FiveStr[i][5] = EOS;
    }
    for (i = 0; i < MONITORGADS; i++)
    {   MonitorStr[i][2] = EOS;
    }
    for (i = 0; i < MEMGADGETS; i++)
    {   MemoryStr[i][2] = EOS;
    }

    lastchange[0]  = 0;

    for (i = 0; i < 2; i++)
    {   for (j = 0; j < NUMKEYS; j++)
        {   keypads[i][j] = default_keypads[i][j];
    }   }
    for (i = 0; i < 4; i++)
    {   console[i] = default_console[i];
    }

    pgmtext[0] = 0;
    for (i = 0; i < GAMEINFOLINES; i++)
    {   strcpy(autotext[i], "?");
    }

    for (i = 0; i < HISCORES; i++)
    {   hiscore[i].score      = 0;
        hiscore[i].thetime[0] =
        hiscore[i].thedate[0] = '-';
        hiscore[i].thetime[1] =
        hiscore[i].thedate[1] = 0;
    }

    for (i = 0; i < SCANCODES; i++)
    {   keyframes[i] = 0;
    }

    cpu_setup();
}

#ifdef PROFILE
EXPORT void printprofile(void)
{   int i;

    /* Possible enhancements:
    print percentage of total instructions executed, for each instruction.
    print percentage of total cycles spent, for each instruction.
    take any indirection penalties into account, instead of just saying
     "at least".
    option of whether to view as "fast" or "slow" cycles (currently only
     shows as "fast" - ie. short - cycles). Such an option should also
     affect the clock shown in the monitor, tracing, etc. (ie. everywhere
     "cycles" and/or "elapsed" are shown to the user) for consistency. */

    T1; textpen(TEXTPEN_VERBOSE);
    for (i = 0; i <= 255; i++)
    {
#ifdef SUPPRESSUNPROFILED
        if (profile[i] != 0)
#endif
            if (profile[i] == 1)
            {   if (i & 8) // interestingly, this algorithm gives all indirectable opcodes
                {   zprintf("%-7s ($%02lX):          1 time  (at least %10u cycles)\n", opcodename[i], i, table_clockperiods[i]);
                } else
                {   zprintf("%-7s ($%02lX):          1 time  (exactly  %10u cycles)\n", opcodename[i], i, table_clockperiods[i]);
            }   }
            else
            {   if (i & 8) // interestingly, this algorithm gives all indirectable opcodes
                {   zprintf("%-7s ($%02lX): %10u times (at least %10u cycles)\n", opcodename[i], i, profile[i], profile[i] * table_clockperiods[i]);
                } else
                {   zprintf("%-7s ($%02lX): %10u times (exactly  %10u cycles)\n", opcodename[i], i, profile[i], profile[i] * table_clockperiods[i]);
            }   }
    }
    zprintf("\n");
    T2;
}
#endif

EXPORT void view_coverage(void)
{   TRANSIENT     int    i = 0,
                         localregionstart;
    TRANSIENT     ULONG  flags;
    PERSIST const STRPTR coveragename[16] =
    {   "----",
        "---W",
        "--R-",
        "--RW",
        "-E--",
        "-E-W",
        "-ER-",
        "-ERW",
        "A---",
        "A--W",
        "A-R-",
        "A-RW",
        "AE--",
        "AE-W",
        "AER-",
        "AERW"
    };

    /* Possible enhancements:
    keep a count of how many times they were each accessed.
    does coverage work properly with mirroring, especially for the "A"
     flag? perhaps not.
    read/writes by the UVI/PVI don't set any coverage bits; perhaps they
     should?
    make some distinction between relative/absolute/zero page accesses
     (but why?). */

    T1; textpen(TEXTPEN_VIEW);
    do
    {   localregionstart = i;
        flags = (memflags[i] & COVERAGE_BITS);
        do
        {   i++;
        } while
        (   (memflags[i] & COVERAGE_BITS) == flags
         && i < 32 * KILOBYTE
        );
        if (flags != 0) // to show all areas (including uncovered), change this line to "if (1)"
        {   if (localregionstart != i - 1)
            {   zprintf("%s: $%04X..$%04X\n", coveragename[flags], (unsigned int) localregionstart, (unsigned int) (i - 1));
            } else
            {   zprintf("%s: $%04X\n"       , coveragename[flags], (unsigned int) localregionstart);
        }   }
    } while (i < 32 * KILOBYTE);
    zprintf("\n");
    T2;
}

EXPORT void project_open(void)
{   TRANSIENT FLAG found;
    TRANSIENT int  i,
                   length,
                   rc;
    PERSIST   TEXT oldpathpart[MAX_PATH + 1]; // PERSISTent for speed (and so as not to overflow the stack)

    if (crippled)
    {   return;
    }

    macro_stop();
    // asl() turns sound off and on for us
    strcpy(oldpathpart, pathpart);
    rc = asl
    (   (STRPTR)GetCatalogStr(CatalogPtr,
            MSG_ENGINE_OPEN,
            "Open Game/Snapshot/Recording"
        ),
        PATTERN_OPEN,
        FALSE,
        pathpart,
        filepart,
        fn_game
    );
    if (rc == EXIT_WARNING)
    {   audit(FALSE);
    }
    if (rc != EXIT_SUCCESS)
    {   return;
    }

    breakparts(); // to set pathpart

    length = (int) strlen(fn_game);
    found = FALSE;
    if (length)
    {   for (i = 0; i < length; i++)
        {   if (fn_game[i] == '.')
            {   found = TRUE;
                break;
    }   }   }

    if (found)
    {   if (!engine_load(FALSE))
        {   beep();
    }   }
    else
    {   if
        (   machine == ELEKTOR
         || machine == PIPBUG
         || machine == INSTRUCTOR
         || machine == CD2650
        )
        {   fixextension(".PGM", fn_game, FALSE);
            if (!engine_load(FALSE))
            {   fixextension(".BIN", fn_game, TRUE);
                if (!engine_load(FALSE))
                {   fixextension(".ZIP", fn_game, TRUE);
                    if (!engine_load(FALSE))
                    {   beep();
        }   }   }   }
        else
        {   // ARCADIA, INTERTON, coin-ops, (CHAOS), DOLPHIN, PHUNSY
            fixextension(".BIN", fn_game, FALSE);
            if (!engine_load(FALSE))
            {   fixextension(".PGM", fn_game, TRUE);
                if (!engine_load(FALSE))
                {   fixextension(".ZIP", fn_game, TRUE);
                    if (!engine_load(FALSE))
                    {   beep();
    }   }   }   }   }

    if (stricmp(oldpathpart, pathpart)) // if directory path has changed
    {   audit(FALSE);
    }

#ifdef AMIGA
    else
    {   DISCARD SetGadgetAttrs
        (   gadgets[GID_LB1],
            MainWindowPtr,
            NULL,
            LISTBROWSER_Selected, (ULONG) ~0,
        TAG_END);
        // this autorefreshes

        currentgame = -1;
    }
#endif
// how to do this for IBM-PC?

    setselection();

    sound_off(FALSE);
    sound_on();
}

EXPORT void project_save(int kind)
{   if (crippled)
    {   return;
    }

    macro_stop();
    fixextension(filekind[kind].extension, fn_game, TRUE);
    breakparts();
    if (asl
    (   filekind[kind].hail,
        filekind[kind].pattern,
        TRUE,
        pathpart,
        filepart,
        fn_game
    ) != EXIT_SUCCESS)
    {   updatesmlgad(GADPOS_MACRO, (ULONG) recmode, TRUE);
        // on IBM-PC version, the above line is needed to "pop" the
        // "record" button back out
        return;
    }
    fixextension(filekind[kind].extension, fn_game, FALSE);
    breakparts();
    engine_save(kind, TRUE);

    if (kind == KIND_COR)
    {   if (anims)
        {   fixextension(filekind[KIND_ANIM].extension, fn_game, TRUE);
            breakparts();
            strcpy(fn_anim, fn_game);
            fixextension(filekind[kind].extension, fn_game, TRUE);
            // not very elegant

            /* if (asl
            (   filekind[KIND_ANIM].hail,
                filekind[KIND_ANIM].pattern,
                TRUE,
                pathpart,
                filepart,
                fn_anim
            ) == EXIT_SUCCESS) */
            {
#ifdef AMIGA
                lockscreen(TRUE);
                AnimInfo = OpenAnim((STRPTR) fn_anim, ScreenPtr);
                unlockscreen();

                if (AnimInfo)
                {   writeicon(KIND_ANIM, fn_anim);
                }
#endif
#ifdef WIN32
                OpenAnim(fn_anim);
#endif
        }   }

        updatebiggads();
        updatesmlgads();
        updatemenus();
}   }

/* This routine returns its results in ax[dest] and ay[dest]. */
EXPORT void engine_dopaddle(int source, int dest)
{   TRANSIENT     FLAG  kybdonly;
    PERSIST const UBYTE digipos[MACHINES][3] =
    { { 1, 112, 254 }, // ARCADIA
      { 1, 112, 225 }, // INTERTON
// (2636 PVI datasheet recommends 20 and 225 as min and max values)
      { 1, 111, 225 }, // ELEKTOR
      { 0,   0,   0 }, // PIPBUG
      { 0,   0,   0 }, // INSTRUCTOR
      { 0,   0,   0 }, // CD2650
      { 0, 128, 255 }, // ASTROWARS
      { 0, 128, 255 }, // GALAXIA
      { 0, 128, 255 }, // MALZAK
      { 0,   0,   0 }  // CHAOS
    };
    // centred value of >= $70 as needed for Interton SUPERINV
    // centred value of <= $6F is needed for Elektor JOYSTICK
    // very low left X-value is needed for Interton INVADERS

    dx = dy = 0;
    kybdonly = system_dopaddle(source, dest);
    if
    (    KeyDown(keypads[source][24])
     && !KeyDown(keypads[source][25])
     && !KeyDown(keypads[source][26])
     && !KeyDown(keypads[source][27])
    )
    {   dx =
        dy = -1;
    } elif
    (   !KeyDown(keypads[source][24])
     &&  KeyDown(keypads[source][25])
     && !KeyDown(keypads[source][26])
     && !KeyDown(keypads[source][27])
    )
    {   dx =  1;
        dy = -1;
    } elif
    (   !KeyDown(keypads[source][24])
     && !KeyDown(keypads[source][25])
     &&  KeyDown(keypads[source][26])
     && !KeyDown(keypads[source][27])
    )
    {   dx = -1;
        dy =  1;
    } elif
    (   !KeyDown(keypads[source][24])
     && !KeyDown(keypads[source][25])
     && !KeyDown(keypads[source][26])
     &&  KeyDown(keypads[source][27])
    )
    {   dx =
        dy =  1;
    } else
    {   if   ( KeyDown(keypads[source][17]) && !KeyDown(keypads[source][18])) dy = -1;
        elif (!KeyDown(keypads[source][17]) &&  KeyDown(keypads[source][18])) dy =  1;
        if   ( KeyDown(keypads[source][19]) && !KeyDown(keypads[source][20])) dx = -1;
        elif (!KeyDown(keypads[source][19]) &&  KeyDown(keypads[source][20])) dx =  1;
    }

    if
    (   (dest == 0 && p1rumble)
     || (dest == 1 && p2rumble)
    )
    {   if (
#ifdef WIN32
            hostcontroller[source] == CONTROLLER_1STDJOY
#endif
#ifdef AMIGA
            hostcontroller[source] == CONTROLLER_1STAPAD
#endif
        )
        {   ff_on(0);
        } elif (
#ifdef WIN32
            hostcontroller[source] == CONTROLLER_2NDDJOY
#endif
#ifdef AMIGA
            hostcontroller[source] == CONTROLLER_2NDAPAD
#endif
        )
        {   ff_on(1);
    }   }
    else
    {   if (
#ifdef WIN32
            hostcontroller[source] == CONTROLLER_1STDJOY
#endif
#ifdef AMIGA
            hostcontroller[source] == CONTROLLER_1STAPAD
#endif
        )
        {   ff_decay(0);
        } elif (
#ifdef WIN32
            hostcontroller[source] == CONTROLLER_2NDDJOY
#endif
#ifdef AMIGA
            hostcontroller[source] == CONTROLLER_2NDAPAD
#endif
        )
        {   ff_decay(1);
    }   }

    if (dx == 0 && dy == 0 && !kybdonly)
    {   return;
    }

    if (analog)
    {   // host is digital or keyboard, guest is analog
        if   (dx == -1) { ax[dest] -= sensitivity; if (ax[dest] <   0) ax[dest] =   0; }
        elif (dx ==  1) { ax[dest] += sensitivity; if (ax[dest] > 255) ax[dest] = 255; }
#ifdef SPRINGLOADED
        else
        {   // assert(dx == 0);
            if (ax[dest] < STICKCENTRE)
            {   ax[dest] += sensitivity * 2;
                if (ax[dest] > STICKCENTRE)
                {   ax[dest] = STICKCENTRE;
            }   }
            elif (ax[dest] > STICKCENTRE)
            {   ax[dest] -= sensitivity * 2;
                if (ax[dest] < STICKCENTRE)
                {   ax[dest] = STICKCENTRE;
        }   }   }
#endif

        if   (dy == -1) { ay[dest] -= sensitivity; if (ay[dest] <   0) ay[dest] =   0; }
        elif (dy ==  1) { ay[dest] += sensitivity; if (ay[dest] > 255) ay[dest] = 255; }
#ifdef SPRINGLOADED
        else
        {   // assert(dy == 0);
            if (ay[dest] < STICKCENTRE)
            {   ay[dest] += sensitivity * 2;
                if (ay[dest] < STICKCENTRE)
                {   ay[dest] = STICKCENTRE;
            }   }
            elif (ay[dest] > STICKCENTRE)
            {   ay[dest] -= sensitivity * 2;
                if (ay[dest] > STICKCENTRE)
                {   ay[dest] = STICKCENTRE;
        }   }   }
#endif

    } else
    {   // host is digital or keyboard, guest is digital
        ax[dest] = digipos[machine][dx + 1];
        ay[dest] = digipos[machine][dy + 1];
}   }

EXPORT void update_controls(void)
{   int i;

#ifdef FLASHSL
    if
    (   (   machine == ARCADIA
         && (memory[A_P1MIDDLEKEYS] & 0x08)
        )
     || (   (machine == INTERTON || machine == ELEKTOR)
         && (memory[I_P1MIDDLEKEYS] & 0x80)
    )   )
    {   leds(TRUE);
    } else
    {   leds(FALSE);
    }
    // not designed to handle other guests
#endif

    s_updatecontrols();

    if (!ControlsWindowPtr)
    {   return;
    }

#ifdef WIN32
    ControlsRastPtr = GetDC(ControlsWindowPtr);
#endif

    switch (machine)
    {
    case ARCADIA:
        newkeys[0]  = memory[A_CONSOLE]      & 0x01;
        newkeys[1]  = memory[A_CONSOLE]      & 0x02;
        newkeys[2]  = memory[A_CONSOLE]      & 0x04;
        newkeys[3]  = 0;
        newkeys[4]  = memory[A_P1LEFTKEYS]   & 0x08;
        newkeys[5]  = memory[A_P1LEFTKEYS]   & 0x04;
        newkeys[6]  = memory[A_P1LEFTKEYS]   & 0x02;
        newkeys[7]  = memory[A_P1LEFTKEYS]   & 0x01;
        newkeys[8]  =
        newkeys[36] =
        newkeys[37] = memory[A_P1MIDDLEKEYS] & 0x08;
        newkeys[9]  = memory[A_P1MIDDLEKEYS] & 0x04;
        newkeys[10] = memory[A_P1MIDDLEKEYS] & 0x02;
        newkeys[11] = memory[A_P1MIDDLEKEYS] & 0x01;
        newkeys[12] = memory[A_P1RIGHTKEYS]  & 0x08;
        newkeys[13] = memory[A_P1RIGHTKEYS]  & 0x04;
        newkeys[14] = memory[A_P1RIGHTKEYS]  & 0x02;
        newkeys[15] = memory[A_P1RIGHTKEYS]  & 0x01;
        newkeys[16] = memory[A_P2LEFTKEYS]   & 0x08;
        newkeys[17] = memory[A_P2LEFTKEYS]   & 0x04;
        newkeys[18] = memory[A_P2LEFTKEYS]   & 0x02;
        newkeys[19] = memory[A_P2LEFTKEYS]   & 0x01;
        newkeys[20] =
        newkeys[38] =
        newkeys[39] = memory[A_P2MIDDLEKEYS] & 0x08;
        newkeys[21] = memory[A_P2MIDDLEKEYS] & 0x04;
        newkeys[22] = memory[A_P2MIDDLEKEYS] & 0x02;
        newkeys[23] = memory[A_P2MIDDLEKEYS] & 0x01;
        newkeys[24] = memory[A_P2RIGHTKEYS]  & 0x08;
        newkeys[25] = memory[A_P2RIGHTKEYS]  & 0x04;
        newkeys[26] = memory[A_P2RIGHTKEYS]  & 0x02;
        newkeys[27] = memory[A_P2RIGHTKEYS]  & 0x01;
    acase INTERTON:
    case ELEKTOR:
        newkeys[0]  = memory[IE_CONSOLE]      & 0x40;
        newkeys[1]  = memory[IE_CONSOLE]      & 0x80;
        newkeys[2]  = memory[IE_CONSOLE]      & 0x20;
        newkeys[3]  = memory[IE_CONSOLE]      & 0x10;
        newkeys[4]  = memory[IE_P1LEFTKEYS]   & 0x80;
        newkeys[5]  = memory[IE_P1LEFTKEYS]   & 0x40;
        newkeys[6]  = memory[IE_P1LEFTKEYS]   & 0x20;
        newkeys[7]  = memory[IE_P1LEFTKEYS]   & 0x10;
        newkeys[8]  =
        newkeys[36] =
        newkeys[37] = memory[IE_P1MIDDLEKEYS] & 0x80;
        newkeys[9]  = memory[IE_P1MIDDLEKEYS] & 0x40;
        newkeys[10] = memory[IE_P1MIDDLEKEYS] & 0x20;
        newkeys[11] = memory[IE_P1MIDDLEKEYS] & 0x10;
        newkeys[12] = memory[IE_P1RIGHTKEYS]  & 0x80;
        newkeys[13] = memory[IE_P1RIGHTKEYS]  & 0x40;
        newkeys[14] = memory[IE_P1RIGHTKEYS]  & 0x20;
        newkeys[15] = memory[IE_P1RIGHTKEYS]  & 0x10;
        newkeys[16] = memory[IE_P2LEFTKEYS]   & 0x80;
        newkeys[17] = memory[IE_P2LEFTKEYS]   & 0x40;
        newkeys[18] = memory[IE_P2LEFTKEYS]   & 0x20;
        newkeys[19] = memory[IE_P2LEFTKEYS]   & 0x10;
        newkeys[20] =
        newkeys[38] =
        newkeys[39] = memory[IE_P2MIDDLEKEYS] & 0x80;
        newkeys[21] = memory[IE_P2MIDDLEKEYS] & 0x40;
        newkeys[22] = memory[IE_P2MIDDLEKEYS] & 0x20;
        newkeys[23] = memory[IE_P2MIDDLEKEYS] & 0x10;
        newkeys[24] = memory[IE_P2RIGHTKEYS]  & 0x80;
        newkeys[25] = memory[IE_P2RIGHTKEYS]  & 0x40;
        newkeys[26] = memory[IE_P2RIGHTKEYS]  & 0x20;
        newkeys[27] = memory[IE_P2RIGHTKEYS]  & 0x10;
    acase INSTRUCTOR:
        newkeys[0]  = sd_column[0] & 0x1;
        newkeys[1]  = sd_column[0] & 0x2;
        newkeys[2]  = sd_column[0] & 0x4;
        newkeys[3]  = sd_column[0] & 0x8;
        newkeys[4]  = sd_column[1] & 0x1;
        newkeys[5]  = sd_column[1] & 0x2;
        newkeys[6]  = sd_column[1] & 0x4;
        newkeys[7]  = sd_column[1] & 0x8;
        newkeys[8]  =
        newkeys[36] =
        newkeys[37] = sd_column[2] & 0x1;
        newkeys[9]  = sd_column[2] & 0x2;
        newkeys[10] = sd_column[2] & 0x4;
        newkeys[11] = sd_column[2] & 0x8;
        newkeys[12] = sd_column[3] & 0x8;
        newkeys[13] = sd_column[3] & 0x4;
        newkeys[14] = sd_column[3] & 0x2;
        newkeys[15] = sd_column[3] & 0x1;
        newkeys[16] = sd_column[4] & 0x8;
        newkeys[17] = sd_column[4] & 0x4;
        newkeys[18] = sd_column[4] & 0x2;
        newkeys[19] = sd_column[4] & 0x1;
        newkeys[20] =
        newkeys[38] =
        newkeys[39] = sd_column[5] & 0x8;
        newkeys[21] = sd_column[5] & 0x4;
        newkeys[22] = sd_column[5] & 0x2;
        newkeys[23] = sd_column[5] & 0x1;
        newkeys[24] = sd_column[6] & 0x8;
        newkeys[25] = sd_column[6] & 0x4;
        newkeys[26] = sd_column[6] & 0x2;
        newkeys[27] = sd_column[6] & 0x1;
    acase MALZAK:
        newkeys[0]  = mg_console & 0x10; // fire
        newkeys[1]  = mg_console & 0x04; // 1P
        newkeys[2]  = mg_console & 0x08; // 2P

        newkeys[3]  = // 10 03 04
        newkeys[4]  = // 09    05
        newkeys[5]  = // 08 07 06
        newkeys[6]  =
        newkeys[7]  =
        newkeys[8]  =
        newkeys[9]  =
        newkeys[10] = 0;

        if ((mg_console & 0xE0) == 0xC0) newkeys[10] = 1; // l/u/r
        if ((mg_console & 0xE0) == 0xA0) newkeys[ 4] = 1; // l/u/r
        if ((mg_console & 0xE0) == 0x80) newkeys[ 3] = 1; // l/u/r
        if ((mg_console & 0x62) == 0x42) newkeys[ 8] = 1; // l/d/r
        if ((mg_console & 0x62) == 0x22) newkeys[ 6] = 1; // l/d/r
        if ((mg_console & 0x62) == 0x02) newkeys[ 7] = 1; // l/d/r
        if ((mg_console & 0xC2) == 0x40) newkeys[ 9] = 1; // u/l/d
        if ((mg_console & 0xA2) == 0x20) newkeys[ 5] = 1; // u/r/d
    acase ASTROWARS:
    case GALAXIA:
        newkeys[0]  = mg_console & 0x30; // fire
        newkeys[1]  = mg_console & 0x01; // 1P
        newkeys[2]  = mg_console & 0x02; // 2P
        newkeys[3]  = g_paddles  & 0x14; // left
        newkeys[4]  = g_paddles  & 0x28; // right
    acase DOLPHIN:
        newkeys[0]  = sd_column[1] & 1;
        newkeys[1]  = sd_column[2] & 1;
        newkeys[2]  = sd_column[3] & 8; // C
        newkeys[3]  = sd_column[3] & 4; // 8
        newkeys[4]  = sd_column[3] & 2; // 4
        newkeys[5]  = sd_column[3] & 1; // 0
        newkeys[6]  = sd_column[4] & 8; // D
        newkeys[7]  = sd_column[4] & 4; // 9
        newkeys[8]  = sd_column[4] & 2; // 5
        newkeys[9]  = sd_column[4] & 1; // 1
        newkeys[10] = sd_column[5] & 8; // E
        newkeys[11] = sd_column[5] & 4; // A
        newkeys[12] = sd_column[5] & 2; // 6
        newkeys[13] = sd_column[5] & 1; // 2
        newkeys[14] = sd_column[6] & 8; // F
        newkeys[15] = sd_column[6] & 4; // B
        newkeys[16] = sd_column[6] & 2; // 7
        newkeys[17] = sd_column[6] & 1; // 3
    }

    if (machine == ARCADIA || machine == INTERTON || machine == ELEKTOR)
    {   if (sx[0] < 32)
        {   newkeys[29] = 1;
            newkeys[31] = 0;
        } elif (sx[0] > 224)
        {   newkeys[29] = 0;
            newkeys[31] = 1;
        } else
        {   newkeys[29] =
            newkeys[31] = 0;
        }
        if (sx[1] < 32)
        {   newkeys[33] = 1;
            newkeys[35] = 0;
        } elif (sx[1] > 224)
        {   newkeys[33] = 0;
            newkeys[35] = 1;
        } else
        {   newkeys[33] =
            newkeys[35] = 0;
        }
        if (sy[0] < 32)
        {   newkeys[28] = 1;
            newkeys[30] = 0;
        } elif (sy[0] > 224)
        {   newkeys[28] = 0;
            newkeys[30] = 1;
        } else
        {   newkeys[28] =
            newkeys[30] = 0;
        }
        if (sy[1] < 32)
        {   newkeys[32] = 1;
            newkeys[34] = 0;
        } elif (sy[1] > 224)
        {   newkeys[32] = 0;
            newkeys[34] = 1;
        } else
        {   newkeys[32] =
            newkeys[34] = 0;
    }   }

    for (i = 0; i < machines[machine].keys; i++)
    {   if
        (   oldkeys[i] != newkeys[i]
        )
        {   invert(i);
            oldkeys[i] = newkeys[i];
    }   }

#ifdef WIN32
    ReleaseDC(ControlsWindowPtr, ControlsRastPtr);
#endif
}

EXPORT void mousedown(int mousex, int mousey)
{   PERSIST int i; // PERSISTent for speed

    if (recmode == RECMODE_PLAY)
    {   return;
    }

    // assert(ControlsWindowPtr);

    wheremouse = -1;
    for (i = 0; i < machines[machine].keys; i++)
    {   if
        (   mousex >= keyrect[machine][i].left
         && mousex <  keyrect[machine][i].left + keyrect[machine][i].width
         && mousey >= keyrect[machine][i].top
         && mousey <  keyrect[machine][i].top  + keyrect[machine][i].height
        )
        {   switch (machine)
            {
            case ARCADIA:
                if (i == 3) // RESET key
                {   project_reset(TRUE);
                } elif (a_keyemul[i].address != 0)
                {   wheremouse = i;
                    newkeys[i] |= memory[a_keyemul[i].mask];
                }
            acase INTERTON:
                if (i == 3) // RESET key
                {   project_reset(TRUE);
                } elif (i_keyemul[i].address != 0)
                {   wheremouse = i;
                    newkeys[i] |= memory[i_keyemul[i].mask];
                }
            acase ELEKTOR:
                if (e_keyemul[i].address != 0)
                {   wheremouse = i;
                    newkeys[i] |= memory[e_keyemul[i].mask];
                }
            acase INSTRUCTOR:
                wheremouse = i;
                newkeys[i] |= sd_column[s_keyemul[i].mask];
            acase ASTROWARS:
                wheremouse = i;
                if (g_keyemul[i].address == 0)
                {   newkeys[i] |= (mg_console & g_keyemul[i].mask);
                } else
                {   // assert(g_keyemul[i].address == 1);
                    newkeys[i] |= (g_paddles & g_keyemul[i].mask);
                }
            acase GALAXIA:
                wheremouse = i;
                if (g_keyemul[i].address == 0)
                {   newkeys[i] |= (mg_console & g_keyemul[i].mask);
                } else
                {   // assert(g_keyemul[i].address == 1);
                    newkeys[i] |= (g_paddles & g_keyemul[i].mask);
                }
            acase MALZAK:
                wheremouse = i;
                newkeys[i] |= (mg_console & m_keyemul[i].mask);
            acase DOLPHIN:
                wheremouse = i;
                newkeys[i] |= sd_column[dolphin_keyemul[i].mask];
            }

            return; // for speed
}   }   }

EXPORT void mouseup(void)
{   wheremouse = -1;
}

EXPORT void domouse(void)
{   PERSIST UBYTE t; // PERSISTent for speed

    if (wheremouse != -1)
    {   // assert(ControlsWindowPtr);

        switch (machine)
        {
        case ARCADIA:
            t = memory[a_keyemul[wheremouse].address] | a_keyemul[wheremouse].mask;
            uviwrite(a_keyemul[wheremouse].address, t);
        acase INTERTON:
            t = memory[i_keyemul[wheremouse].address] | i_keyemul[wheremouse].mask;
            pviwrite((int) i_keyemul[wheremouse].address, t, FALSE);
        acase ELEKTOR:
            t = memory[e_keyemul[wheremouse].address] | e_keyemul[wheremouse].mask;
            pviwrite((int) e_keyemul[wheremouse].address, t, FALSE);
        acase INSTRUCTOR:
            sd_column[s_keyemul[wheremouse].address] |= s_keyemul[wheremouse].mask;
        acase MALZAK:
            mg_console |= m_keyemul[wheremouse].mask;
        acase ASTROWARS:
        case GALAXIA:
            if (g_keyemul[wheremouse].address == 0)
            {   mg_console |= g_keyemul[wheremouse].mask;
            } else
            {   // assert(g_keyemul[wheremouse].address == 1);
                g_paddles |= g_keyemul[wheremouse].mask;
            }
        acase DOLPHIN:
            sd_column[dolphin_keyemul[wheremouse].address] |= dolphin_keyemul[wheremouse].mask;
}   }   }

/* Connect/disconnect netplay interface. */
EXPORT void NetPlay(int Switch)
{   int i;

#ifdef AMIGA
    if (!SocketBase)
    {   return;
    }
#endif

    switch (Switch)
    {
    case NET_OFF:
        // Disconnect netplay
        NetClose();
    acase NET_SERVER:
        // If already connected, drop out
        if (connected != NET_OFF)
        {   break;
        }

        if (!NetMyName(hostname, (int) sizeof(hostname)))
        {   say
            (   (STRPTR) GetCatalogStr(CatalogPtr,
                    MSG_NET_NORESOLVEHOSTNAME,
                    "Can't resolve hostname!"
                ),
                TRUE
            );
            return;
        }

        T1; textpen(TEXTPEN_NET);
        zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_NET_ACTINGASSERVER,
               "Acting as server at %s...\n"
            ),
            hostname
        );
        T2;
        // Connect second player, report problems if present
        NetConnect(0, NETPORT);
        /* if (connected == NET_OFF)
        {   T1; textpen(TEXTPEN_NET); zprintf("Can't act as server at %s!\n", hostname); T2;
        } */
    acase NET_CLIENT:
        // If already connected, drop out
        if (connected != NET_OFF)
        {   break;
        }

        aborting = FALSE;
        sound_off(FALSE);
        hostnamewindow();
        sound_on();
        if (!hostname[0] || aborting)
        {   aborting = FALSE;

            T1; textpen(TEXTPEN_NET);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_CANCELLED,
                    "Cancelled.\n\n"
            )   );
            T2;
            return;
        }

        // convert to lower case
        for (i = 0; i < (int) strlen(hostname); i++)
        {   if (hostname[i] >= 'A' && hostname[i] <= 'Z')
            {   hostname[i] += 32;
        }   }

        T1; textpen(TEXTPEN_NET);
        zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_NET_CONNECTINGASCLIENT,
                "Connecting as client to server at %s...please wait...\n"
            ),
            hostname
        );
        T2;
        // Connect second player, report problems if present
        lockgui();
        NetConnect(hostname, NETPORT);
        unlockgui();
        /* if (connected == NET_OFF)
        {   T1; textpen(TEXTPEN_NET); zprintf("Can't connect as client to server at %s!\n", hostname); T2;
        } */
    }

    settitle();
}

MODULE void NetConnect(const char* Server, unsigned short Port)
{   struct sockaddr_in Addr;
    struct hostent*    Host;
    int                LSocket;
#ifdef WIN32
    int                AddrLength;
    unsigned long      J;
#endif
#ifdef AMIGA
    LONG               AddrLength;
    char               J;
#endif

#ifdef AMIGA
    if (!SocketBase)
    {   return;
    }
#endif

    /* Close existing network connection */
    NetClose();

    /* Clear the address structure */
    memset(&Addr, 0, sizeof(Addr));

    /* If the server name is given, we first try to */
    /* connect as a client. */
    if (Server)
    {   /* Look up server address */

        if (!(Host = gethostbyname((STRPTR) Server))) // OS4 needs this cast!
        {   T1; textpen(TEXTPEN_NET);
            zprintf
            (   (STRPTR)GetCatalogStr(CatalogPtr,
                    MSG_NET_NOGETHOSTNAME,
                    "Can't get hostname!\n"
            )   );
            T2;
            connected = NET_OFF;
            return;
        }

        /* Set fields of the address structure */
        memcpy(&Addr.sin_addr, Host->h_addr, (size_t) (Host->h_length));
        Addr.sin_family = AF_INET;
        Addr.sin_port   = htons(Port);

        /* Create a socket */
        if ((Socket = (int) socket(AF_INET, SOCK_STREAM, 0)) < 0)
        {   T1; textpen(TEXTPEN_NET);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_NET_NOCREATECLIENT,
                    "Can't create client socket!\n"
            )   );
            T2;
            connected = NET_OFF;
            return;
        }

        /* Connecting... */
        if (connect((unsigned int) Socket, (struct sockaddr *) &Addr, (LONG) sizeof(Addr)) >= 0)
        {   /* Make communication socket blocking */
            J = 0;
            if (ioctlsocket((unsigned int) Socket, (long) FIONBIO, &J) < 0)
            {   T1; textpen(TEXTPEN_NET);
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_NET_NOBLOCKCLIENT,
                        "Can't block client socket!\n"
                )   );
                T2;
                DISCARD closesocket((unsigned int) Socket);
                Socket = -1;
                connected = NET_OFF;
                return;
            }

            connected = NET_CLIENT;
            T1; textpen(TEXTPEN_NET);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_NET_CLIENTCONNECTED,
                    "Successfully connected as client!\n"
            )   );
            T2;
            return;
        }

        DISCARD closesocket((unsigned int) Socket);
        T1; textpen(TEXTPEN_NET);
        zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_NET_NOCONNECTCLIENT,
                "Can't connect as client!\n"
        )   );
        T2;
    } else
    {   /* Becoming a server and waiting for connection request. */

        /* Set fields of the address structure */
        Addr.sin_addr.s_addr = htonl(INADDR_ANY);
        Addr.sin_family      = AF_INET;
        Addr.sin_port        = htons(Port);

        /* Create a listening socket */
        if ((LSocket = (int) socket(AF_INET, SOCK_STREAM, 0)) < 0)
        {   T1; textpen(TEXTPEN_NET);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_NET_NOCREATESERVER,
                    "Can't create server socket!\n"
            )   );
            T2;
            connected = NET_OFF;
            return;
        }

        /* Bind listening socket */
        if (bind((unsigned int) LSocket, (struct sockaddr *) &Addr, (LONG) sizeof(Addr)) < 0)
        {   T1; textpen(TEXTPEN_NET);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_NET_NOBINDSOCKET,
                    "Can't bind server socket!\n"
            )   );
            T2;
            DISCARD closesocket((unsigned int) LSocket);
            connected = NET_OFF;
            return;
        }

        /* Make listening socket non-blocking */
        J = 1;
        if (ioctlsocket((unsigned int) LSocket, (long) FIONBIO, &J) < 0)
        {   T1; textpen(TEXTPEN_NET);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_NET_NOUNBLOCKSOCKET,
                    "Can't unblock server socket!\n"
            )   );
            T2;
            DISCARD closesocket((unsigned int) LSocket);
            connected = NET_OFF;
            return;
        }

        /* Listen for one client */
        if (listen((unsigned int) LSocket, 1) < 0)
        {   T1; textpen(TEXTPEN_NET);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_NET_NOLISTEN,
                    "Can't listen!\n"
            )   );
            T2;
            DISCARD closesocket((unsigned int) LSocket);
            connected = NET_OFF;
            return;
        }

        T1; textpen(TEXTPEN_NET);
        zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_NET_LISTENING,
                "Listening...\n"
        )   );
        T2;

        lockgui();
        aborting = FALSE;
        listen_open();

        /* Accepting calls... */

        AddrLength = (LONG) sizeof(Addr);
        do
        {   /* Process all messages */
#ifdef WIN32
            clearmsgs();
#endif
#ifdef AMIGA
            listen_loop();
#endif
            Socket = (int) accept((unsigned int) LSocket, (struct sockaddr *) &Addr, &AddrLength);
#ifdef WIN32
            if (WSAGetLastError() && (WSAGetLastError() != WSAEWOULDBLOCK))
            {   break;
            }
#endif
        }
        while (Socket < 0 && !aborting);

#ifdef AMIGA
        raisepri();
#endif

        listen_close();
        DISCARD closesocket((unsigned int) LSocket);
        clearkybd();
        unlockgui();

        /* Client failed to connect */
        if (Socket < 0)
        {   T1; textpen(TEXTPEN_NET);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_NET_ABORTED,
                    "Aborted by user!\n"
            )   );
            T2;
            connected = NET_OFF;
            return;
        }

        /* Make communication socket blocking */
        J = 0;
        if (ioctlsocket((unsigned int) Socket, (long) FIONBIO, &J) < 0)
        {   T1; textpen(TEXTPEN_NET);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_NET_NOBLOCKSERVER,
                    "Can't block server socket!\n"
            )   );
            T2;
            DISCARD closesocket((unsigned int) Socket);
            Socket = -1;
            connected = NET_OFF;
            return;
        }

        T1; textpen(TEXTPEN_NET);
        zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_NET_SERVERCONNECTED,
                "The client has successfully connected to your server!\n"
        )   );
        T2;
        connected = NET_SERVER;
        // return;
}   }

/* Closes connection opened with NetConnect(). */
MODULE void NetClose(void)
{
#ifdef AMIGA
    if (!SocketBase)
    {   return;
    }
#endif

    if (Socket >= 0)
    {   // assert(connected != NET_OFF);

        DISCARD closesocket((unsigned int) Socket);
        Socket = -1;
        connected = NET_OFF;
        T1; textpen(TEXTPEN_NET);
        zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_NET_DISCONNECTED,
                "Disconnected.\n"
        )   );
        T2;
}   }

/* Returns local address or 0 on failure. */
MODULE const char* NetMyName(char* Buffer, int MaxChars)
{   struct hostent* Host;

#ifdef AMIGA
    if (!SocketBase)
    {   return 0;
    }
#endif

    /* Have to have enough characters */
    if (MaxChars < 16)
    {   return 0;
    }
    /* Get local hostname */
    DISCARD gethostname(Buffer, MaxChars);
    /* Look up address */
    if (!(Host = gethostbyname((STRPTR) Buffer))) // OS4 needs this cast!
    {   return 0;
    }
    /* Must have an address */
    if (!Host->h_addr_list || !Host->h_addr_list[0])
    {   return 0;
    }
    /* Copy address */
    sprintf
    (   Buffer,
        "%d.%d .%d.%d",
        (UBYTE) Host->h_addr_list[0][0],
        (UBYTE) Host->h_addr_list[0][1],
        (UBYTE) Host->h_addr_list[0][2],
        (UBYTE) Host->h_addr_list[0][3]
    );

    return Buffer;
}

/* Exchanges given number of bytes with the other side and
   returns TRUE on success, FALSE on failure. */
EXPORT FLAG NetExchange(char* In, const char* Out, int N)
{   int I, J;

#ifdef AMIGA
    if (!SocketBase)
    {   return FALSE;
    }
#endif

    /* Have to have a socket */
    if (Socket < 0)
    {   return FALSE;
    }

    /* Send data */
    for (I = J = N; (J >= 0) && I;)
    {   J = (int) send((LONG) Socket, (APTR) Out, (LONG) I, 0); // OS4 needs the APTR cast!
        if (J > 0)
        {   Out += J;
            I   -= J;
    }   }
    if (I)
    {   return FALSE;
    }

    /* Receive data */
    for (I = J = N; (J >= 0) && I;)
    {   J = (int) recv((unsigned int) Socket, (UBYTE*) In, I, 0);
        if (J > 0)
        {   In += J;
            I  -= J;
    }   }
    if (I)
    {   return FALSE;
    }

    return TRUE;
}

EXPORT FLAG NetSendByte(UBYTE value)
{
#ifdef AMIGA
    if (!SocketBase)
    {   return FALSE;
    }
#endif

    /* Have to have a socket */
    if (Socket < 0)
    {   return FALSE;
    }

    /* Send data */
    if (send((unsigned int) Socket, &value, 1, 0) > 0)
    {   return TRUE;
    } // implied else
    return FALSE;
}

EXPORT UBYTE NetReceiveByte(void)
{   PERSIST UBYTE value; // PERSISTent for speed

#ifdef AMIGA
    if (!SocketBase)
    {   return 0;
    }
#endif

    /* Have to have a socket */
    if (Socket < 0)
    {   return 0;
    }

    /* Send data */
    if (recv((unsigned int) Socket, &value, 1, 0) > 0)
    {   return value;
    } // implied else
    return 0;

    /* When successful, returns 0..255.
       When failed, returns 0.
       Therefore, 0 can mean either success or failure :-( . */
}

EXPORT void closemonitors(FLAG both)
{   /* If both is TRUE,  controls are     closed/reopened.
       If both is FALSE, controls are not closed/reopened. */

    if (DIPsWindowPtr)
    {   close_dips();
        edit_dips();
        // if switching between Astro Wars and Galaxia, that was unnecessary
    }
    if (MonitorWindowPtr)
    {   close_monitor();
        view_monitor();
    }
    if (SpriteWindowPtr)
    {   close_spriteviewer();
        open_spriteviewer();
    }
    if (both && ControlsWindowPtr)
    {   close_controls();
        view_controls();
}   }

EXPORT void playsound(FLAG force)
{   PERSIST UWORD  guestpitch[14],
                   oldguestpitch[14]  = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    PERSIST UBYTE  guestvolume[14],
                   oldguestvolume[14] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    PERSIST int    i;     // PERSISTent for speed
    PERSIST float  hertz; // PERSISTent for speed

    /* guestvolume[0..13]: indexed by guest channel.
                           contents are stored in guest format.
       guestpitch[0..13]:  indexed by guest channel.
                           contents are stored in guest format.

       Although Arcadia/PIPBUG/CD2650 and Interton/Elektor volumes do not match,
       this is unimportant as playsound(TRUE) is called when changing
       machines. Thus, Arcadia/PIPBUG/CD2650 volumes are never compared against
       Interton/Elektor volumes.

    Channels are used as follows:
     Tones:
       0: Expanded Elektor:          1st PSG, tone  channel A
          MALZAK/Galaxia:          2nd PVI tone generator
       1: Expanded Elektor:          1st PSG, tone  channel B
          Galaxia:                   3rd PVI tone generator
       2: Expanded Elektor:          1st PSG, tone  channel C
       3: Expanded Elektor:          2nd PSG, tone  channel A
       4: Expanded Elektor:          2nd PSG, tone  channel B
       5: Expanded Elektor:          2nd PSG, tone  channel C
       6: Arcadia:                   UVI tone generator
          PIPBUG/CD2650/PHUNSY:      Tone generator
          Interton/Elektor/coin-ops: (1st) PVI tone generator
     Noises:
       7: Expanded Elektor:          1st PSG, noise channel A
       8: Expanded Elektor:          1st PSG, noise channel B
       9: Expanded Elektor:          1st PSG, noise channel C
      10: Expanded Elektor:          2nd PSG, noise channel A
      11: Expanded Elektor:          2nd PSG, noise channel B
      12: Expanded Elektor:          2nd PSG, noise channel C
      13: Arcadia:                   UVI noise generator
          Interton/Expanded Elektor: IE_NOISE generator
*/

    if (!sound)
    {   return;
    }

    switch (machine)
    {
    case ARCADIA:
        if ((memory[A_VOLUME]) & 0x08)
        {   guestpitch[6]   = memory[A_PITCH]  & 0x7F;
            guestvolume[6]  = memory[A_VOLUME] & 0x07; // 0..7
        } else
        {   guestpitch[6]   =
            guestvolume[6]  = 0;
        }

        if ((memory[A_VOLUME]) & 0x10)
        {   guestpitch[13]  = memory[A_PITCH]  & 0x7F;
            guestvolume[13] = memory[A_VOLUME] & 0x07; // 0..7
        } else
        {   guestpitch[13]  =
            guestvolume[13] = 0;
        }
    acase INTERTON:
    case ELEKTOR:
        // PVI tones
        guestpitch[6] = memory[0x1F00 + PVI_PITCH];
        if
        (   memmap == MEMMAP_F
         && (memory[E_MIXER1] & 0x40)
        )
        {   guestvolume[6] = memory[E_PORTA1] & 0xF;
        } else
        {   guestvolume[6] = 15;
        }
        // if ((memory[IE_NOISE]) & 0x04) guestvolume[6] = 0;

        if (memmap == MEMMAP_E)
        {   guestpitch[13]  =  1; // must be nonzero!
            guestvolume[13] =  0;
        } else
        {

/* For Interton COMBATA:

Firing:                NOISE is $04 (%00,0,00100).
Moving and not firing: NOISE is $88 (%10,0,01000).
Not moving nor firing: NOISE is $C8 (%11,0,01000).
Killing:               NOISE is $10 (%00,0,10000), then
                       NOISE is $00 (%00,0,00000).
Firing:                PITCH starts at about $09 and increases.
                       PITCH becomes $87 during impact with boundary.
Moving and not firing: PITCH is $59.
Not moving nor firing: PITCH is $5E.
Killing:               PITCH holds at whatever it was from the firing.

For Interton LASERATT: PITCH is always $08 (%00,0,01000), even on title screen.

For Elektor 10-1-Invaders:
                       NOISE is $00 on title screen
                       NOISE is $04 during play
                       NOISE is $10 briefly while dying */

            if (memory[IE_NOISE] & 0xD3)
            {   guestpitch[13]  = memory[IE_NOISE] & 0xD3; // mask of %11,0,10011
                guestvolume[13] = 15;
            } else
            {   guestpitch[13]  = 1;
                guestvolume[13] = 0;
        }   }
    acase PIPBUG:
        if (music)
        {   if (iar == 0x4A5)
            {   guestpitch[6] = r[0];
                if (pipbug_getpitch(guestpitch[6]) != 0.0)
                {   guestvolume[6] = 7;
                } else
                {   guestvolume[6] = 0;
            }   }
            // else don't change from previous value
        } elif (p_morse && iar == 0x49B)
        {   guestpitch[6]  = 0x55; // middle A (440 Hz)
            guestvolume[6] = 7;
        } else
        {   guestpitch[6]  =
            guestvolume[6] = 0;
        }
        guestpitch[13]   =
        guestvolume[13]  = 0;
    acase CD2650:
        if (c_tone)
        {   guestpitch[6]  = 0x55; // middle A (440 Hz)
            guestvolume[6] = 7;
        } else
        {   guestpitch[6]  =
            guestvolume[6] = 0;
        }
        guestpitch[13]   =
        guestvolume[13]  = 0;
    acase MALZAK:
        guestpitch[0]    = memory[0x1500 + PVI_PITCH];
        guestvolume[0]   = 15;
        guestpitch[6]    = memory[0x1400 + PVI_PITCH];
        guestvolume[6]   = 15;
        guestpitch[13]   =
        guestvolume[13]  = 0;
    acase ASTROWARS:
        guestpitch[6]    = memory[0x1500 + PVI_PITCH];
        guestvolume[6]   = 15;
        guestpitch[13]   =
        guestvolume[13]  = 0;
    acase GALAXIA:
        guestpitch[6]    = memory[0x1500 + PVI_PITCH];
        guestvolume[6]   = 15;
        guestpitch[13]   =
        guestvolume[13]  = 0;
        guestpitch[0]    = memory[0x1600 + PVI_PITCH];
        guestvolume[0]   = 15;
        guestpitch[1]    = memory[0x1700 + PVI_PITCH];
        guestvolume[1]   = 15;
    acase PHUNSY:
        if (piano)
        {   if (iar == 0x1816)
            {   guestpitch[6] = r[3] + 1;
                if (phunsy_getpianopitch(guestpitch[6]) != 0.0)
                {   guestvolume[6] = 7;
                } else
                {   guestvolume[6] = 0;
            }   }
            elif (iar == 0x1827)
            {   guestpitch[6]  =
                guestvolume[6] = 0;
            } // else don't change from previous value
        } elif (fiftytheme)
        {   // assert(iar == 0x1827);
            guestpitch[6]  = r[3] + 1;
            // assert(phunsy_getfiftypitch(guestpitch[6]) != 0.0);
            guestvolume[6] = 7;
        } elif (belmach || belmach0)
        {   if ((r[0] & 0x3E) == 0x3E) // rest
            {   guestpitch[6]  =
                guestvolume[6] = 0;
            } else
            {   guestpitch[6]  = r[0] & 0x3F;
                guestvolume[6] = 7;
        }   }
        else
        {   guestpitch[6]  =
            guestvolume[6] = 0;
        }
        guestpitch[13]   =
        guestvolume[13]  = 0;
    adefault: // eg. INSTRUCTOR, CHAOS, DOLPHIN
        guestpitch[6]    =
        guestvolume[6]   = 0;
        guestpitch[13]   =
        guestvolume[13]  = 0;
    }

    if (memmap == MEMMAP_F)
    {   // 1st PSG, output channel A
        if (!(memory[E_AMPLITUDEA1] & 0x10))
        {   guestvolume[0] =
            guestvolume[7] = memory[E_AMPLITUDEA1] & 0x0F;
        } else
        {   guestvolume[0] =
            guestvolume[7] = 15;
        }
        if (!(memory[E_MIXER1] &    1)) // 1st tone A enable
        {   guestpitch[0] = ((memory[E_PITCHA1_H] & 0x0F) * 256) + memory[E_PITCHA1_L];
            if (guestpitch[0] == 0)
            {   guestpitch[0] = 1;
        }   }
        else
        {   guestpitch[0] = 0;
        }
        if (!(memory[E_MIXER1] &    8)) // 1st noise A enable
        {   guestpitch[7] = memory[E_PITCHD1] & 0x1F;
            if (guestpitch[7] == 0)
            {   guestpitch[7] = 1;
        }   }
        else
        {   guestpitch[7] = 0;
        }

        // 1st PSG, output channel B
        if (!(memory[E_AMPLITUDEB1] & 0x10))
        {   guestvolume[1] =
            guestvolume[8] = memory[E_AMPLITUDEB1] & 0x0F;
        } else
        {   guestvolume[1] =
            guestvolume[8] = 15;
        }
        if (!(memory[E_MIXER1] &    2)) // 1st tone B enable
        {   guestpitch[1] = ((memory[E_PITCHB1_H] & 0x0F) * 256) + memory[E_PITCHB1_L];
            if (guestpitch[1] == 0)
            {   guestpitch[1] = 1;
        }   }
        else
        {   guestpitch[1] = 0;
        }
        if (!(memory[E_MIXER1] & 0x10)) // 1st noise B enable
        {   guestpitch[8] = memory[E_PITCHD1] & 0x1F;
            if (guestpitch[8] == 0)
            {   guestpitch[8] = 1;
        }   }
        else
        {   guestpitch[8] = 0;
        }

        // 1st PSG, output channel C
        if (!(memory[E_AMPLITUDEC1] & 0x10))
        {   guestvolume[2] =
            guestvolume[9] = memory[E_AMPLITUDEC1] & 0x0F;
        } else
        {   guestvolume[2] =
            guestvolume[9] = 15;
        }
        if (!(memory[E_MIXER1] &    4)) // 1st tone C enable
        {   guestpitch[2] = ((memory[E_PITCHC1_H] & 0x0F) * 256) + memory[E_PITCHC1_L];
            if (guestpitch[2] == 0)
            {   guestpitch[2] = 1;
        }   }
        else
        {   guestpitch[2] = 0;
        }
        if (!(memory[E_MIXER1] & 0x20)) // 1st noise C enable
        {   guestpitch[9] = memory[E_PITCHD1] & 0x1F;
            if (guestpitch[9] == 0)
            {   guestpitch[9] = 1;
        }   }
        else
        {   guestpitch[9] = 0;
        }

        // 2nd PSG, output channel A
        if (!(memory[E_AMPLITUDEA2] & 0x10))
        {   guestvolume[3]  =
            guestvolume[10] = memory[E_AMPLITUDEA2] & 0x0F;
        } else
        {   guestvolume[3]  =
            guestvolume[10] = 15;
        }
        if (!(memory[E_MIXER2] &    1)) // 2nd tone A enable
        {   guestpitch[3] = ((memory[E_PITCHA2_H] & 0x0F) * 256) + memory[E_PITCHA2_L];
            if (guestpitch[3] == 0)
            {   guestpitch[3] = 1;
        }   }
        else
        {   guestpitch[3] = 0;
        }
        if (!(memory[E_MIXER2] &    8)) // 2nd noise A enable
        {   guestpitch[10] = memory[E_PITCHD2] & 0x1F;
            if (guestpitch[10] == 0)
            {   guestpitch[10] = 1;
        }   }
        else
        {   guestpitch[10] = 0;
        }

        // 2nd PSG, output channel B
        if (!(memory[E_AMPLITUDEB2] & 0x10))
        {   guestvolume[4]  =
            guestvolume[11] = memory[E_AMPLITUDEB2] & 0x0F;
        } else
        {   guestvolume[4]  =
            guestvolume[11] = 15;
        }
        if (!(memory[E_MIXER2] &    2)) // 2nd tone B enable
        {   guestpitch[4] = ((memory[E_PITCHB2_H] & 0x0F) * 256) + memory[E_PITCHB2_L];
            if (guestpitch[4] == 0)
            {   guestpitch[4] = 1;
        }   }
        else
        {   guestpitch[4] = 0;
        }
        if (!(memory[E_MIXER2] & 0x10)) // 2nd noise B enable
        {   guestpitch[11] = memory[E_PITCHD2] & 0x1F;
            if (guestpitch[11] == 0)
            {   guestpitch[11] = 1;
        }   }
        else
        {   guestpitch[11] = 0;
        }

        // 2nd PSG, output channel C
        if (!(memory[E_AMPLITUDEC2] & 0x10))
        {   guestvolume[5]  =
            guestvolume[12] = memory[E_AMPLITUDEC2] & 0x0F;
        } else
        {   guestvolume[5]  =
            guestvolume[12] = 15;
        }
        if (!(memory[E_MIXER2] &    4)) // 2nd tone C enable
        {   guestpitch[5] = ((memory[E_PITCHC2_H] & 0x0F) * 256) + memory[E_PITCHC2_L];
            if (guestpitch[5] == 0)
            {   guestpitch[5] = 1;
        }   }
        else
        {   guestpitch[5] = 0;
        }
        if (!(memory[E_MIXER2] & 0x20)) // 2nd noise C enable
        {   guestpitch[12] = memory[E_PITCHD2] & 0x1F;
            if (guestpitch[12] == 0)
            {   guestpitch[12] = 1;
        }   }
        else
        {   guestpitch[12] = 0;
    }   }
    else
    {   if (machine != ASTROWARS)
        {   guestpitch[1]   =
            guestvolume[1]  = 0;
            if (machine != MALZAK)
            {   guestpitch[0]   =
                guestvolume[0]  = 0;
        }   }

        guestpitch[2]   =
        guestpitch[3]   =
        guestpitch[4]   =
        guestpitch[5]   =
        guestpitch[7]   =
        guestpitch[8]   =
        guestpitch[9]   =
        guestpitch[10]  =
        guestpitch[11]  =
        guestpitch[12]  = 0;

        guestvolume[3]  =
        guestvolume[4]  =
        guestvolume[5]  =
        guestvolume[7]  =
        guestvolume[8]  =
        guestvolume[9]  =
        guestvolume[10] =
        guestvolume[11] =
        guestvolume[12] = 0;
    }

    for (i = 0; i <= 6; i++)
    {   if (guestpitch[i] && guestvolume[i])
        {   if
            (   force
             || !guestplaying[i]
             || guestpitch[i]  != oldguestpitch[i]
             || guestvolume[i] != oldguestvolume[i]
            )
            {   if
                (   i == 6
                 || (    i == 0 && (machine == MALZAK || machine == GALAXIA))
                 || (    i == 1 &&                       machine == GALAXIA )
                )
                {   if (machine == PIPBUG || machine == CD2650)
                    {   hertz = pipbug_getpitch(guestpitch[i]);
                    } elif (machine == PHUNSY)
                    {   if (piano)
                        {   hertz = phunsy_getpianopitch(guestpitch[i]);
                        } elif (fiftytheme)
                        {   hertz = phunsy_getfiftypitch(guestpitch[i]);
                        } else
                        {   // assert(belmach || belmach0);
                            hertz = phunsy_getbelmachpitch(guestpitch[i]);
                    }   }
                    elif (retune)
                    {   if (guestpitch[i] == 28)
                        {   if (machine == ARCADIA) // eg, GUNDAM (271.52) // needs to be C# not C for (Arcadiaz) GUNDAM
                            {   hertz = (float) 277.18; // middle C#
                            } else // eg. MUSICALG (271.52) // needs to be C not C# for (Interton) MUSICALG
                            {   hertz = (float) 261.63; // middle C
                        }   }
                        else
                        {   hertz = idealfreq[guestpitch[i]];
                    }   }
                    else
                    {   hertz = (float) (  7874.0 / (guestpitch[i] + 1)); // desired frequency of the waveform
                }   }
                else
                {   hertz     = (float) (110837.0 / (guestpitch[i]    )); // desired frequency of the waveform
                }
                TonePlay(i, hertz, guestvolume[i]);
        }   }
        else
        {   SoundStop(i);
    }   }

    for (i = 7; i <= 13; i++)
    {   if (guestpitch[i] && guestvolume[i])
        {   if
            (   force
             || !guestplaying[i]
             || guestvolume[i] != oldguestvolume[i]
             || guestpitch[i] != oldguestpitch[i]
            )
            {   if (i == 13)
                {   hertz     = (float)   7874.0 / (guestpitch[i] + 1);
                } else
                {   // assert(i >= 7 && i <= 12);
                    hertz     = (float) 110837.0 / (guestpitch[i] + 1);
                }
                NoisePlay(i, hertz, guestvolume[i]);
        }   }
        else
        {   SoundStop(i);
    }   }

    for (i = 0; i < 14; i++)
    {   oldguestvolume[i] = guestvolume[i];
        oldguestpitch[i]  = guestpitch[i];
}   }

EXPORT void sortaudit(void)
{   if (sortby == SORTBYNAME)
    {   sortbymachine();
        sortbyname();
    } else
    {   // assert(sortby == SORTBYMACHINE);
        sortbyname();
        sortbymachine();
}   }

MODULE void sortbyname(void)
{
#ifdef QUICKSORT
    quicksort_byname(0, foundgames - 1);
#else
    int i, j;

    for (i = 0; i < foundgames - 1; i++)
    {   for (j = 2; j < foundgames - i - 1; j++)
        {   if
            (   stricmp
                (   auditlist[sidebar[j    ]].gamename,
                    auditlist[sidebar[j + 1]].gamename
                ) > 0
            )
            {   swapgames(j, j + 1);
    }   }   }
#endif
}

MODULE void sortbymachine(void)
{
#ifdef QUICKSORT
    quicksort_bymachine(0, foundgames - 1);
#else
    int i, j;

    for (i = 0; i < foundgames - 1; i++)
    {   for (j = 2; j < foundgames - i - 1; j++)
        {   if
            (   auditlist[sidebar[j    ]].machine
              > auditlist[sidebar[j + 1]].machine
            )
            {   swapgames(j, j + 1);
    }   }   }
#endif
}

#ifdef QUICKSORT
MODULE void quicksort_bymachine(int m, int n)
{   int i, j, k,
        key;

    if (m < n)
    {   k = (m + n) / 2; // k is pivot
        swapgames(m, k);
        key = sidebar[m].machine;
        i = m + 1;
        j = n;
        while (i <= j)
        {   while ((i <= n) && (sidebar[i].machine <= key))
            {   i++;
            }
            while ((j >= m) && (sidebar[j].machine >  key))
            {   j--;
            }
            if (i < j)
            {   swapgames(i, j);
        }   }
        // swap two elements
        swapgames(m, j);
        // recursively sort the lesser list
        quicksort_bymachine(m,     j - 1);
        quicksort_bymachine(j + 1, n    );
}   }

MODULE void quicksort_byname(int m, int n)
{   int    i, j, k;
    STRPTR key;

    if (m < n)
    {   k = (m + n) / 2; // k is pivot
        swapgames(m, k);
        key = sidebar2[sidebar[m].gamenum].gamename;
        i = m + 1;
        j = n;
        while (i <= j)
        {   while ((i <= n) && (stricmp(sidebar2[sidebar[i].gamenum].gamename, key) <= 0))
            {   i++;
            }
            while ((j >= m) && (stricmp(sidebar2[sidebar[j].gamenum].gamename, key) >  0))
            {   j--;
            }
            if (i < j)
            {   swapgames(i, j);
        }   }
        // swap two elements
        swapgames(m, j);
        // recursively sort the lesser list
        quicksort_byname(m,     j - 1);
        quicksort_byname(j + 1, n    );
}   }
#endif

#if (!defined(__SASC)) && (!defined(__VBCC__))
    EXPORT int stcd_l(const char* in, SLONG* lvalue)
    {   return sscanf(in, "%ld", lvalue);
    } /* stcd_l */

    EXPORT int stch_l(const char* in, SLONG* lvalue)
    {   char *endptr;

        *lvalue = strtoul(in, &endptr, 16);
        return endptr - in;
    } /* stch_l */

    EXPORT int stcl_d(char* out, long lvalue)
    {   return sprintf(out, "%ld", lvalue);
    } /* stcl_d */

    EXPORT int stcl_h(char* out, unsigned long lvalue)
    {   return sprintf(out, "%lX", lvalue);
    }

    EXPORT int stcul_d(char* out, unsigned long val)
    {   // Convert (32-bit) unsigned integer to ASCII decimal string.

        return sprintf(out, "%lu", val);
    }

    EXPORT char* stptok(const char* s, char* tok, int toklen, char* brk)
    {   char *lim,
             *b;

        if (!*s)
        {   return NULL;
        }

        lim = tok + toklen - 1;
        while (*s && tok < lim)
        {   for (b = brk; *b; b++)
            {   if (*s == *b)
                {   *tok = 0;
                    return (char *)s;
            }   }
            *tok++ = *s++;
        }
        *tok = 0;
        return (char *) s;
    }
#endif

EXPORT void updateprinter(void)
{   if (echoprinter && !PrinterHandle)
    {

#ifdef AMIGA
        if (!(PrinterHandle = fopen(PRINTERDEVICE, "wb"))) // just cast for lint
        {
#endif
#ifdef WIN32
        PrinterHandle = CreateFile(PRINTERDEVICE, GENERIC_WRITE | GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL); // NB. no colon after LPT1
// would write-only access be enough?
        if (PrinterHandle == INVALID_HANDLE_VALUE)
        {   PrinterHandle = NULL;
#endif
            echoprinter = FALSE;
            updatemenu(MENUITEM_PRINTER);
    }   }
    elif (!echoprinter && PrinterHandle)
    {
#ifdef AMIGA
        fclose(PrinterHandle);
#endif
#ifdef WIN32
        DISCARD CloseHandle(PrinterHandle);
#endif
        PrinterHandle = NULL;
}   }

MODULE void flipguidepixel(void)
{
#ifdef THERULER
    int y;
#endif

    switch (machine)
    {
    case ARCADIA:
        if (rast < machines[machine].height)
        {   screen[machines[machine].width - 1][rast] = from_a[7 - from_a[screen[machines[machine].width - 1][rast]]];
        } // we can never have BROWN at that coordinate, so we are safe
    acase INTERTON:
    case ELEKTOR:
    case MALZAK:
    case ASTROWARS:
    case GALAXIA:
        if (rast < machines[machine].height)
        {   screen[machines[machine].width - 1][rast] = (UBYTE) (15 - screen[machines[machine].width - 1][rast]);
    }   }

#ifdef THERULER
    for (y = 0; y < machines[machine].height; y += 10)
    {   local_drawpixel(0, y, (UBYTE) (15 - screen[0][y]));
        if (y % 50 == 0)
        {   local_drawpixel(1, y, (UBYTE) (15 - screen[1][y]));
            if (y % 100 == 0)
            {   local_drawpixel(2, y, (UBYTE) (15 - screen[2][y]));
    }   }   }
#endif
}

EXPORT UBYTE thechin(void)
{   int   i;
    UBYTE inchar;

    if (recmode == RECMODE_PLAY)
    {   // assert(machine != PHUNSY);
        if (freshly)
        {   rhythm = ReadLong();
            freshly = FALSE;
        }
        if (rhythm)
        {   rhythm--;
        } else
        {   thequeue[0] = (TEXT) IOBuffer[offset++];
            queuepos = 1;
            freshly = TRUE;
    }   }

    if (queuepos)
    {   inchar = (UBYTE) thequeue[0];
        queuepos--;

        if (recmode == RECMODE_RECORD && machine != PHUNSY)
        {   // big endian 32-bit longword
            OutputBuffer[0] = (UBYTE)  (rhythm / 16777216);
            OutputBuffer[1] = (UBYTE) ((rhythm % 16777216) / 65536);
            OutputBuffer[2] = (UBYTE) ((rhythm %    65536) /   256);
            OutputBuffer[3] = (UBYTE)  (rhythm %      256);
            OutputBuffer[4] = inchar;
            DISCARD fwrite(OutputBuffer, 5, 1, MacroHandle);
            rhythm = 0;
        }
        if (machine != CHAOS && machine != PHUNSY && (trace || logbios))
        {   T1; textpen(TEXTPEN_TRACE);
            switch (machine)
            {
            case PIPBUG:
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_CHIN,
                        "Returning from %s with input of '%c' (%d).\n\n"
                    ),
                    "CHIN",
                    inchar,
                    (int) inchar
                );
            acase CD2650:
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_CHIN,
                        "Returning from %s with input of '%c' (%d).\n\n"
                    ),
                    "KBIN",
                    inchar,
                    (int) inchar
                ); // but they may not actually be using KBIN
            }
            T2;
        }

        if (queuepos)
        {   for (i = 0; i < queuepos; i++)
            {   thequeue[i] = thequeue[i + 1];
        }   }

        return inchar;
    } else
    {   if (recmode == RECMODE_RECORD && machine != PHUNSY)
        {   if (rhythm == (ULONG) -1)
            {   // big endian 32-bit longword
                OutputBuffer[0] = (UBYTE)  (rhythm / 16777216);
                OutputBuffer[1] = (UBYTE) ((rhythm % 16777216) / 65536);
                OutputBuffer[2] = (UBYTE) ((rhythm %    65536) /   256);
                OutputBuffer[3] = (UBYTE)  (rhythm %      256);
                switch (machine)
                {
                case PIPBUG:
                case CHAOS:
                    OutputBuffer[4] = 0;
                acase CD2650:
                    OutputBuffer[4] = 0xFF;
                acase PHUNSY:
                    OutputBuffer[4] = 0x80;
                }
                DISCARD fwrite(OutputBuffer, 5, 1, MacroHandle);
            }
            rhythm++;
        }

        switch (machine)
        {
        case PIPBUG:
        case CHAOS:
            return 0;
        acase CD2650:
            return 0xFF;
        acase PHUNSY:
            return 0x80;
        adefault: // should never happen
            // assert(0);
            return 0; // to avoid spurious GCC compiler warnings
}   }   }

EXPORT void drawsegment(int position, int segment, int colour)
{   PERSIST int controlsx,
                x, xx, y, yy; // PERSISTent for speed

    x = 10 + (21 * position);
    controlsx = 6 * position;
    y = 10;

    if (machine == PIPBUG)
    {   x +=  47;
        y += 258;
    }

    switch (segment)
    {
    case 0: // top side
        if (!drawcorners)
        {   local_drawpixel(x +  2,  y +  1, colour);
            local_drawpixel(x + 11,  y +  1, colour);
        }
        for (xx =  3; xx <= 10; xx++)
        {   local_drawpixel(x + xx,  y     , colour);
            local_drawpixel(x + xx,  y +  1, colour);
            local_drawpixel(x + xx,  y +  2, colour);
        }

        s_drawcontrolspixel(controlsx + 1, 0, colour);
        s_drawcontrolspixel(controlsx + 2, 0, colour);
    acase 1: // top right side
        if (!drawcorners)
        {   local_drawpixel(x + 12,  y +  2, colour);
            local_drawpixel(x + 12,  y + 13, colour);
	    }
        for (yy =  3; yy <= 12; yy++)
        {   local_drawpixel(x + 11,  y + yy, colour);
            local_drawpixel(x + 12,  y + yy, colour);
            local_drawpixel(x + 13,  y + yy, colour);
        }
        s_drawcontrolspixel(controlsx + 3, 1, colour);
        s_drawcontrolspixel(controlsx + 3, 2, colour);
    acase 2: // bottom right side
        if (!drawcorners)
        {   local_drawpixel(x + 12, y + 15, colour);
            local_drawpixel(x + 12, y + 26, colour);
	    }
        for (yy = 16; yy <= 25; yy++)
        {   local_drawpixel(x + 11, y + yy, colour);
            local_drawpixel(x + 12, y + yy, colour);
            local_drawpixel(x + 13, y + yy, colour);
        }
        s_drawcontrolspixel(controlsx + 3, 4, colour);
        s_drawcontrolspixel(controlsx + 3, 5, colour);
    acase 3: // bottom side
        if (!drawcorners)
        {   local_drawpixel(x +  2, y + 27, colour);
            local_drawpixel(x + 11, y + 27, colour);
	    }
        for (xx =  3; xx <= 10; xx++)
        {   local_drawpixel(x + xx, y + 26, colour);
            local_drawpixel(x + xx, y + 27, colour);
            local_drawpixel(x + xx, y + 28, colour);
        }
        s_drawcontrolspixel(controlsx + 1, 6, colour);
        s_drawcontrolspixel(controlsx + 2, 6, colour);
    acase 4: // bottom left side
        if (!drawcorners)
        {   local_drawpixel(x + 1, y + 15, colour);
            local_drawpixel(x + 1, y + 26, colour);
        }
        for (yy = 16; yy <= 25; yy++)
        {   local_drawpixel(x     , y + yy, colour);
            local_drawpixel(x +  1, y + yy, colour);
            local_drawpixel(x +  2, y + yy, colour);
        }
        s_drawcontrolspixel(controlsx    , 4, colour);
        s_drawcontrolspixel(controlsx    , 5, colour);
    acase 5: // top left side
        if (!drawcorners)
        {   local_drawpixel(x +  1,  y +  2, colour);
            local_drawpixel(x +  1,  y + 13, colour);
        }
        for (yy =  3; yy <= 12; yy++)
        {   local_drawpixel(x     ,  y + yy, colour);
            local_drawpixel(x +  1,  y + yy, colour);
            local_drawpixel(x +  2,  y + yy, colour);
        }
        s_drawcontrolspixel(controlsx    , 1, colour);
        s_drawcontrolspixel(controlsx    , 2, colour);
    acase 6: // middle "side"
        if (!drawcorners)
        {   local_drawpixel(x +  2,  y + 14, colour);
            local_drawpixel(x + 11,  y + 14, colour);
        }
        for (xx =  3; xx <= 10; xx++)
        {   local_drawpixel(x + xx,  y + 13, colour);
            local_drawpixel(x + xx,  y + 14, colour);
            local_drawpixel(x + xx,  y + 15, colour);
        }
        s_drawcontrolspixel(controlsx + 1, 3, colour);
        s_drawcontrolspixel(controlsx + 2, 3, colour);
    acase 7: // top left corner
        local_drawpixel(x     , y     , colour);
        local_drawpixel(x +  1, y     , colour);
        local_drawpixel(x +  2, y     , colour);
        local_drawpixel(x     , y +  1, colour);
        local_drawpixel(x +  1, y +  1, colour);
        local_drawpixel(x +  2, y +  1, colour);
        local_drawpixel(x     , y +  2, colour);
        local_drawpixel(x +  1, y +  2, colour);
        local_drawpixel(x +  2, y +  2, colour);
        s_drawcontrolspixel(controlsx, 0, colour);
    acase 8: // top right corner
        local_drawpixel(x + 11, y     , colour);
        local_drawpixel(x + 12, y     , colour);
        local_drawpixel(x + 13, y     , colour);
        local_drawpixel(x + 11, y +  1, colour);
        local_drawpixel(x + 12, y +  1, colour);
        local_drawpixel(x + 13, y +  1, colour);
        local_drawpixel(x + 11, y +  2, colour);
        local_drawpixel(x + 12, y +  2, colour);
        local_drawpixel(x + 13, y +  2, colour);
        s_drawcontrolspixel(controlsx + 3, 0, colour);
    acase 9: // middle left "corner"
        local_drawpixel(x     , y + 13, colour);
        local_drawpixel(x +  1, y + 13, colour);
        local_drawpixel(x +  2, y + 13, colour);
        local_drawpixel(x     , y + 14, colour);
        local_drawpixel(x +  1, y + 14, colour);
        local_drawpixel(x +  2, y + 14, colour);
        local_drawpixel(x     , y + 15, colour);
        local_drawpixel(x +  1, y + 15, colour);
        local_drawpixel(x +  2, y + 15, colour);
        s_drawcontrolspixel(controlsx    , 3, colour);
    acase 10: // middle right "corner"
        local_drawpixel(x + 11, y + 13, colour);
        local_drawpixel(x + 12, y + 13, colour);
        local_drawpixel(x + 13, y + 13, colour);
        local_drawpixel(x + 11, y + 14, colour);
        local_drawpixel(x + 12, y + 14, colour);
        local_drawpixel(x + 13, y + 14, colour);
        local_drawpixel(x + 11, y + 15, colour);
        local_drawpixel(x + 12, y + 15, colour);
        local_drawpixel(x + 13, y + 15, colour);
        s_drawcontrolspixel(controlsx + 3, 3, colour);
    acase 11: // bottom left corner
        local_drawpixel(x     , y + 26, colour);
        local_drawpixel(x +  1, y + 26, colour);
        local_drawpixel(x +  2, y + 26, colour);
        local_drawpixel(x     , y + 27, colour);
        local_drawpixel(x +  1, y + 27, colour);
        local_drawpixel(x +  2, y + 27, colour);
        local_drawpixel(x     , y + 28, colour);
        local_drawpixel(x +  1, y + 28, colour);
        local_drawpixel(x +  2, y + 28, colour);
        s_drawcontrolspixel(controlsx    , 6, colour);
    acase 12: // bottom right corner
        local_drawpixel(x + 11, y + 26, colour);
        local_drawpixel(x + 12, y + 26, colour);
        local_drawpixel(x + 13, y + 26, colour);
        local_drawpixel(x + 11, y + 27, colour);
        local_drawpixel(x + 12, y + 27, colour);
        local_drawpixel(x + 13, y + 27, colour);
        local_drawpixel(x + 11, y + 28, colour);
        local_drawpixel(x + 12, y + 28, colour);
        local_drawpixel(x + 13, y + 28, colour);
        s_drawcontrolspixel(controlsx + 3, 6, colour);
    acase 13: // decimal point (Dolphin only)
        local_drawpixel(x + 16, y + 26, colour);
        local_drawpixel(x + 17, y + 26, colour);
        local_drawpixel(x + 18, y + 26, colour);
        local_drawpixel(x + 16, y + 27, colour);
        local_drawpixel(x + 17, y + 27, colour);
        local_drawpixel(x + 18, y + 27, colour);
        local_drawpixel(x + 16, y + 28, colour);
        local_drawpixel(x + 17, y + 28, colour);
        local_drawpixel(x + 18, y + 28, colour);
        // s_drawcontrolspixel(controlsx + 5, 6, colour);
}   }

EXPORT void view_screen(void)
{   switch (machine)
    {
    case ARCADIA:
        a_viewscreen();
    acase INTERTON:
    case ELEKTOR:
        ie_viewscreen();
    acase INSTRUCTOR:
        s_viewscreen();
    acase CD2650:
        c_viewscreen();
    acase ASTROWARS:
    case GALAXIA:
        g_viewscreen();
    acase MALZAK:
        m_viewscreen();
    acase DOLPHIN:
        dolphin_viewscreen();
    acase PHUNSY:
        phunsy_viewscreen();
}   }

EXPORT void s_updatecontrols(void)
{   int i;

    if (machine != INSTRUCTOR || !ControlsWindowPtr)
    {   return;
    }

    for (i = 0; i < 8; i++)
    {   if ((s_glow & (128 >> i)) != (oldglow & (128 >> i)))
        {   rectfill(78 + (6 * i), 178);
    }   }

    if
    (   (opcode == 0x40 && oldrun != 0x40)
     || (opcode != 0x40 && oldrun == 0x40)
    )
    {   rectfill(143, 178);
    }

    if ((psu & PSU_F) != oldflag)
    {   rectfill(143, 188);
    }

    oldglow = s_glow;
    oldrun  = opcode;
    oldflag = psu & PSU_F;
}

EXPORT void updatedips(void)
{   if (DIPsWindowPtr)
    {   switch (machine)
        {
        case INSTRUCTOR:
            s_updatedips(TRUE);
        acase MALZAK:
            // assert(memmap == MEMMAP_V);
            m_updatedips(TRUE);
        acase ASTROWARS:
        case GALAXIA:
            g_updatedips(TRUE);
}   }   }

MODULE __inline void local_drawpixel(int x, int y, int colour)
{   if (screen[x][y] != colour)
    {   screen[x][y] = (UBYTE) colour;
        drawpixel(x, y, colour);
}   }

EXPORT void updatehiscores(void)
{   int bestscore  = 0,
        i, j,
        multiplier = 0, // to avoid spurious SAS/C compiler warnings
        newscore,
        resolved;

    switch (machine)
    {
    case ARCADIA:
        if (whichgame == -1)
        {   return;
        } // implied else
        resolved = known[whichgame].hiscore;
    acase ASTROWARS:
        resolved = HS_ASTROWARS_POS;
    acase GALAXIA:
        resolved = HS_GALAXIA_POS;
    acase MALZAK:
        if (memmap == MEMMAP_MALZAK1)
        {   resolved = HS_MALZAK1_POS;
        } else
        {   // assert(memmap == MEMMAP_MALZAK2);
            resolved = HS_MALZAK2_POS;
        }
    adefault:
        return; // important!
    }

    for (i = 0; i < 4; i++)
    {   if (hiscore[resolved].hipos[i] == 0)
        {   break;
        } // implied else

        if (resolved == HS_BASEBALL_POS)
        {   // assert(machine == ARCADIA);

            if
            (   (memory[hiscore[resolved].hipos[i] + 1] & 0x3F) >= 0x10 // '0'
             && (memory[hiscore[resolved].hipos[i] + 1] & 0x3F) <= 0x19 // '9'
            )
            {   newscore = (((memory[hiscore[resolved].hipos[i]    ] & 0x3F) - 0x10) * 10)
                         +  ((memory[hiscore[resolved].hipos[i] + 1] & 0x3F) - 0x10);
            } else
            {   newscore =   (memory[hiscore[resolved].hipos[i]    ] & 0x3F) - 0x10;
        }   }
        else
        {   newscore = 0;

            switch (hiscore[resolved].length)
            {
            case 1:
                multiplier = 1;
            acase 2:
                multiplier = 10;
            acase 3:
                multiplier = 100;
            acase 4:
                multiplier = 1000;
            acase 5:
                multiplier = 10000;
            acase 6:
                multiplier = 100000;
            }

            switch (machine)
            {
            case ARCADIA:
                for (j = 0; j < hiscore[resolved].length; j++)
                {   if
                    (   (memory[hiscore[resolved].hipos[i] + (j * hiscore[resolved].nextdigit)] & 0x3F) >= 0x10 // '0'
                     && (memory[hiscore[resolved].hipos[i] + (j * hiscore[resolved].nextdigit)] & 0x3F) <= 0x19 // '9'
                    )
                    {   newscore += ((memory[hiscore[resolved].hipos[i] + (j * hiscore[resolved].nextdigit)] & 0x3F) - 0x10) * multiplier;
                    }
                    multiplier /= 10;
                }
            acase ASTROWARS:
                for (j = 0; j < hiscore[resolved].length; j++)
                {   if
                    (   (memory[hiscore[resolved].hipos[i] + (j * hiscore[resolved].nextdigit)]) >= 0xB0 // '0'
                     && (memory[hiscore[resolved].hipos[i] + (j * hiscore[resolved].nextdigit)]) <= 0xB9 // '9'
                    )
                    {   newscore += ((memory[hiscore[resolved].hipos[i] + (j * hiscore[resolved].nextdigit)]) - 0xB0) * multiplier;
                    }
                    multiplier /= 10;
                }
            acase GALAXIA:
                for (j = 0; j < hiscore[resolved].length; j++)
                {   if
                    (   (memory[hiscore[resolved].hipos[i] + (j * hiscore[resolved].nextdigit)] & 0x7F) >= 0x60 // '0'
                     && (memory[hiscore[resolved].hipos[i] + (j * hiscore[resolved].nextdigit)] & 0x7F) <= 0x69 // '9'
                    )
                    {   newscore += ((memory[hiscore[resolved].hipos[i] + (j * hiscore[resolved].nextdigit)] & 0x7F) - 0x60) * multiplier;
                    }
                    multiplier /= 10;
                }
            acase MALZAK:
                for (j = 0; j < hiscore[resolved].length; j++)
                {   if
                    (   (memory[hiscore[resolved].hipos[i] + (j * hiscore[resolved].nextdigit)]) >= 0x30 // '0'
                     && (memory[hiscore[resolved].hipos[i] + (j * hiscore[resolved].nextdigit)]) <= 0x39 // '9'
                    )
                    {   newscore += ((memory[hiscore[resolved].hipos[i] + (j * hiscore[resolved].nextdigit)]) - 0x30) * multiplier;
                    }
                    multiplier /= 10;
        }   }   }

        if (newscore > bestscore)
        {   bestscore = newscore;
    }   }

    if (bestscore > (int) hiscore[resolved].score)
    {   hiscore[resolved].score = (ULONG) bestscore;
        datestamp(resolved);
}   }

EXPORT void loadhiscores(void)
{   UBYTE HiScoresBuffer[HISCORESLENGTH];
    ULONG thesize;
    FILE* TheLocalHandle /* = NULL */ ; // LocalHandle is a variable of winbase.h
    int   i, j;

    thesize = getsize(FN_HISCORES);
    if (thesize == (ULONG) HISCORESLENGTH)
    {   if ((TheLocalHandle = fopen(FN_HISCORES, "rb")))
        {   if (fread(HiScoresBuffer, (size_t) HISCORESLENGTH, 1, TheLocalHandle) == 1)
            {   switch (HiScoresBuffer[0])
                {
                case 3:
                    for (i = 0; i <= 40; i++)
                    {   hiscore[i].score = (ULONG) (
                                           (65536 * HiScoresBuffer[ 1 + (i * 18)])
                                         + (  256 * HiScoresBuffer[ 2 + (i * 18)])
                                         +          HiScoresBuffer[ 3 + (i * 18)]
                                           );
                        for (j = 0; j < 6; j++)
                        {   hiscore[i].thetime[j] = (TEXT) HiScoresBuffer[ 4 + j + (i * 18)];
                        }
                        for (j = 0; j < 9; j++)
                        {   hiscore[i].thedate[j] = (TEXT) HiScoresBuffer[10 + j + (i * 18)];
                    }   }
                    // 45 -> 41
                    hiscore[41].score = (ULONG) (
                                        (65536 * HiScoresBuffer[ 1 + (45 * 18)])
                                      + (  256 * HiScoresBuffer[ 2 + (45 * 18)])
                                      +          HiScoresBuffer[ 3 + (45 * 18)]
                                        );
                    for (j = 0; j < 6; j++)
                    {   hiscore[i].thetime[j] = (TEXT) HiScoresBuffer[ 4 + j + (45 * 18)];
                    }
                    for (j = 0; j < 9; j++)
                    {   hiscore[i].thedate[j] = (TEXT) HiScoresBuffer[10 + j + (45 * 18)];
                    }
                    // 41..44 -> 42..45
                    for (i = 41; i <= 44; i++)
                    {   hiscore[i + 1].score = (ULONG) (
                                               (65536 * HiScoresBuffer[ 1 + (i * 18)])
                                             + (  256 * HiScoresBuffer[ 2 + (i * 18)])
                                             +          HiScoresBuffer[ 3 + (i * 18)]
                                               );
                        for (j = 0; j < 6; j++)
                        {   hiscore[i + 1].thetime[j] = (TEXT) HiScoresBuffer[ 4 + j + (i * 18)];
                        }
                        for (j = 0; j < 9; j++)
                        {   hiscore[i + 1].thedate[j] = (TEXT) HiScoresBuffer[10 + j + (i * 18)];
                    }   }
                    for (i = 46; i <= 49; i++)
                    {   hiscore[i].score = (ULONG) (
                                           (65536 * HiScoresBuffer[ 1 + (i * 18)])
                                         + (  256 * HiScoresBuffer[ 2 + (i * 18)])
                                         +          HiScoresBuffer[ 3 + (i * 18)]
                                           );
                        for (j = 0; j < 6; j++)
                        {   hiscore[i].thetime[j] = (TEXT) HiScoresBuffer[ 4 + j + (i * 18)];
                        }
                        for (j = 0; j < 9; j++)
                        {   hiscore[i].thedate[j] = (TEXT) HiScoresBuffer[10 + j + (i * 18)];
                    }   }
                acase 4:
                    for (i = 0; i <= 49; i++)
                    {   hiscore[i].score = (ULONG) (
                                           (65536 * HiScoresBuffer[ 1 + (i * 18)])
                                         + (  256 * HiScoresBuffer[ 2 + (i * 18)])
                                         +          HiScoresBuffer[ 3 + (i * 18)]
                                           );
                        for (j = 0; j < 6; j++)
                        {   hiscore[i].thetime[j] = (TEXT) HiScoresBuffer[ 4 + j + (i * 18)];
                        }
                        for (j = 0; j < 9; j++)
                        {   hiscore[i].thedate[j] = (TEXT) HiScoresBuffer[10 + j + (i * 18)];
            }   }   }   }
            DISCARD fclose(TheLocalHandle);
            // TheLocalHandle = NULL;
}   }   }

EXPORT void savehiscores(void)
{   UBYTE HiScoresBuffer[HISCORESLENGTH];
    FILE* TheLocalHandle /* = NULL */ ; // LocalHandle is a variable of winbase.h
    int   i, j;

    if ((TheLocalHandle = fopen(FN_HISCORES, "wb")))
    {   HiScoresBuffer[0] = 4; // 1 means V11.7 -14.09 (no longer supported)
                               // 2 means V14.1 -14.2  (no longer supported)
                               // 3 means V14.21-14.22 (obsolete but importable)
                               // 4 means V14.23+      (current)
        for (i = 0; i < HISCORES; i++)
        {   HiScoresBuffer[1 + (i * 18)] = (UBYTE)  (hiscore[i].score / 65536);
            HiScoresBuffer[2 + (i * 18)] = (UBYTE) ((hiscore[i].score % 65536) / 256);
            HiScoresBuffer[3 + (i * 18)] = (UBYTE)  (hiscore[i].score %   256);
            for (j = 0; j < 6; j++)
            {   HiScoresBuffer[ 4 + j + (i * 18)] = (UBYTE) hiscore[i].thetime[j];
            }
            for (j = 0; j < 9; j++)
            {   HiScoresBuffer[10 + j + (i * 18)] = (UBYTE) hiscore[i].thedate[j];
        }   }

        DISCARD fwrite(HiScoresBuffer, (size_t) HISCORESLENGTH, 1, TheLocalHandle);
        DISCARD fclose(TheLocalHandle);
        // TheLocalHandle = NULL;
}   }

EXPORT void logport(int port, UBYTE data, FLAG handled, FLAG write)
{   PERSIST TEXT asciival[3 + 1],
                 portstr[80 + 1];

    if (logioports)
    {   switch (port)
        {
        case PORTC:
            strcpy
            (   portstr,
                GetCatalogStr(CatalogPtr,
                    MSG_IOCONTROLPORT,
                    "I/O control port"
            )   );
        acase PORTD:
            strcpy
            (   portstr,
                GetCatalogStr(CatalogPtr,
                    MSG_IODATAPORT,
                    "I/O data port"
            )   );
        adefault:
            sprintf
            (   portstr,
                GetCatalogStr(CatalogPtr,
                    MSG_IOEXTENDEDPORT,
                    "extended I/O port $%02X"
                ),
                (unsigned int) port
            );
        }
        DISCARD getfriendly((int) iar);

        T1; textpen(TEXTPEN_VERBOSE);
        if (machine == DOLPHIN)
        {   asciival[0] = '\'';
            asciival[1] = dolphin_chars[data & 0x7F];
            asciival[2] = '\'';
            asciival[3] = EOS;
        } else
        {   if (data >= 32)
            {   asciival[0] = '\'';
                asciival[1] = data;
                asciival[2] = '\'';
                asciival[3] = EOS;
            } else
            {   strcpy(asciival, asciiname[data]);
        }   }
        if (write)
        {   if (handled)
            {   ;
#ifdef LOGHANDLEDWRITES
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_HANDLEDWRITE,
                        "Handled write of $%02X (%s) to %s at address %s!\n\n"
                    ),
                    data,
                    asciival,
                    portstr,
                    friendly
                );
#endif
            } else
            {   zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_UNHANDLEDWRITE,
                        "Unhandled write of $%02X (%s) to %s at address %s!\n\n"
                    ),
                    data,
                    asciival,
                    portstr,
                    friendly
                );
        }   }
        else
        {   if (handled)
            {   ;
#ifdef LOGHANDLEDREADS
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_HANDLEDREAD,
                        "Handled read of $%02X (%s) from %s at address %s!\n\n"
                    ),
                    data,
                    asciival,
                    portstr,
                    friendly
                );
#endif
            } else
            {   zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_UNHANDLEDREAD,
                        "Unhandled read of $%02X (%s) from %s at address %s!\n\n"
                    ),
                    data,
                    asciival,
                    portstr,
                    friendly
                );
        }   }
        T2;
        if (pauselog)
        {   runto();
}   }   }

EXPORT void audit_coinops(void)
{
#ifdef WIN32
    strcpy(auditlist[0].gamename, "\\ (");
#endif
#ifdef AMIGA
    strcpy(auditlist[0].gamename, ": (");
#endif
    strcat
    (  auditlist[0].gamename,
       GetCatalogStr(CatalogPtr,
           MSG_ROOT,
           "root"
       )
    );
    strcat(auditlist[0].gamename, ")");

    auditlist[0].filename[0] = 0;
    auditlist[0].whichgame   = -1;
    auditlist[0].machine     = -1;
    auditlist[0].glyph       = SYSFOLDERGLYPH;
    auditlist[0].memmap      = -2;
    sidebar[0]               = 0;

#ifdef WIN32
    strcpy(auditlist[1].gamename, ".. (");
#endif
#ifdef AMIGA
    strcpy(auditlist[1].gamename, "/ (");
#endif
    strcat
    (  auditlist[1].gamename,
       GetCatalogStr(CatalogPtr,
           MSG_PARENT,
           "parent"
       )
    );
    strcat(auditlist[1].gamename, ")");

    auditlist[1].filename[0] = 0;
    auditlist[1].whichgame   = -1;
    auditlist[1].machine     = -1;
    auditlist[1].glyph       = SYSFOLDERGLYPH;
    auditlist[1].memmap      = -2;
    sidebar[1]               = 1;

    strcpy(auditlist[2].gamename, "Astro Wars");
    auditlist[2].filename[0] = 0;
    auditlist[2].whichgame   = -1;
    auditlist[2].machine     =
    auditlist[2].glyph       = ASTROWARS;
    auditlist[2].memmap      = MEMMAP_ASTROWARS;
    sidebar[2]               = 2;

    strcpy(auditlist[3].gamename, "Galaxia");
    auditlist[3].filename[0] = 0;
    auditlist[3].whichgame   = -1;
    auditlist[3].machine     =
    auditlist[3].glyph       = GALAXIA;
    auditlist[3].memmap      = MEMMAP_GALAXIA;
    sidebar[3]               = 3;

    strcpy(auditlist[4].gamename, "Malzak 1");
    auditlist[4].filename[0] = 0;
    auditlist[4].whichgame   = -1;
    auditlist[4].machine     =
    auditlist[4].glyph       = MALZAK;
    auditlist[4].memmap      = MEMMAP_MALZAK1;
    sidebar[4]               = 4;

    strcpy(auditlist[5].gamename, "Malzak 2");
    auditlist[5].filename[0] = 0;
    auditlist[5].whichgame   = -1;
    auditlist[5].machine     =
    auditlist[5].glyph       = MALZAK;
    auditlist[5].memmap      = MEMMAP_MALZAK2;
    sidebar[5]               = 5;

    foundgames = 6;
}

MODULE __inline void swapgames(int game1, int game2)
{   PERSIST int temp; // PERSISTent for speed

    temp           = sidebar[game1];
    sidebar[game1] = sidebar[game2];
    sidebar[game2] = temp;
}

EXPORT void engine_quit(void)
{   // engine_save() requires the window to be open (on Amiga at least)
    if
    (   !upgrade
     && autosave
     && !crippled
     && (game || machine == CHAOS || machine == DOLPHIN || machine == PHUNSY)
    )
    {   changefilepart("AUTOSAVE.COS");
        engine_save(KIND_COS, FALSE);
    }
    if (!upgrade && memmap == MEMMAP_MALZAK2)
    {   m_savenvram();
    }

    logfile_close();
}

EXPORT void emulate(void)
{   switch (machine)
    {
    case ARCADIA:
        uvi();
    acase INTERTON:
    case ELEKTOR:
    case ASTROWARS:
    case GALAXIA:
    case MALZAK:
        pvi();
    acase PIPBUG:
        p_emulate();
    acase INSTRUCTOR:
        s_emulate();
    acase CD2650:
        c_emulate();
    acase CHAOS:
        h_emulate();
    acase DOLPHIN:
        d_emulate();
    acase PHUNSY:
        phunsy_emulate();
}   }

EXPORT void handle_menu(int command)
{   switch (command)
    {
    case MENUITEM_RESETTOMONITOR:
    case MENUITEM_REVERT:
    case MENUITEM_QUICKLOAD:
    case MENUITEM_QUICKSAVE:
    case MENUITEM_SAVEROM:
    case MENUITEM_SERVER:
    case MENUITEM_CLIENT:
    case MENUITEM_CODINGGUIDE:
    case MENUITEM_GAMINGGUIDE:
    case MENUITEM_STEP:
    case MENUITEM_STEPOVER:
    case MENUITEM_IGNORE:
    case MENUITEM_RUNTORAST:
    case MENUITEM_RUNTORASTN:
    case MENUITEM_RUNTODMA:
    case MENUITEM_RUNTOINTERRUPT:
    case MENUITEM_RUNTOFRAME:
    case MENUITEM_RUNTOSUBEND:
    case MENUITEM_RUNTOLOOPEND:
    case MENUITEM_INTERRUPT:
        docommand(command);
    acase MENUITEM_RESET:
        project_reset(FALSE);
    acase MENUITEM_OPEN:
        project_open();
    acase MENUITEM_SAVESNP:
        project_save(KIND_COS);
    acase MENUITEM_SAVETVC:
        project_save(KIND_TVC);
    acase MENUITEM_EAPSPRITES:
        save_eap(0);
    acase MENUITEM_EAPUDCS:
        save_eap(1);
    acase MENUITEM_SAVEEAL:
        save_eal();
    acase MENUITEM_AUDIT:
        audit(TRUE);
    acase MENUITEM_SAVEILBM:
        savescreenshot(KIND_ILBM, TRUE, FALSE);
    acase MENUITEM_SAVEACBM:
        savescreenshot(KIND_ACBM, TRUE, FALSE);
    acase MENUITEM_SAVEBMP:
        savescreenshot(KIND_BMP, TRUE, FALSE);
    acase MENUITEM_SAVEPCX:
        savescreenshot(KIND_PCX, TRUE, FALSE);
    acase MENUITEM_COPYSCRN:
        savescreenshot(KIND_BMP, FALSE, TRUE);
    acase MENUITEM_ICONIFY:
        iconify();
    acase MENUITEM_COPYTEXT:
        edit_copytext();
    acase MENUITEM_PASTETEXT:
        edit_pastetext();
    acase MENUITEM_PSGS:
        view_psgs();
    acase MENUITEM_VIEWALLMEMORY:
        view_allmemory();
    acase MENUITEM_CRAM:
        view_cram();
    acase MENUITEM_MRAM:
        view_mram();
    acase MENUITEM_NVRAM:
        view_nvram();
    acase MENUITEM_SCREEN:
        view_screen();
    acase MENUITEM_UDCS:
        view_udcs();
    acase MENUITEM_XVI:
        view_xvi();
    acase MENUITEM_VIEWCPU:
        view_cpu(TRUE);
    acase MENUITEM_IMAGERY:
        view_imagery();
    acase MENUITEM_BRAM:
        view_bram();
    acase MENUITEM_COLOURRAM:
        g_viewcolours();
    acase MENUITEM_PLAYFIELD:
        m_viewplayfield();
    acase MENUITEM_STARTRECORDING:
        project_save(KIND_COR);
    acase MENUITEM_RESTARTPLAYBACK:
        macro_restartplayback();
    acase MENUITEM_STOP:
        macro_stop();
    acase MENUITEM_ADDBP:
        debug_set(KIND_BP);
    acase MENUITEM_CLEARBP:
        debug_set(KIND_BC);
    acase MENUITEM_CLEARBPS:
        DISCARD bp_clearall();
    acase MENUITEM_ADDWP:
        debug_set(KIND_WP);
    acase MENUITEM_CLEARWP:
        debug_set(KIND_WC);
    acase MENUITEM_CLEARWPS:
        DISCARD wp_clearall();
    acase MENUITEM_SHADOWS:
        flipbool(&shadowing, command);
        changed = TRUE;
    acase MENUITEM_INVERSE:
        flipbool(&inverse, command);
        fixupcolours();
        draw_cd2650();
        draw_phunsy();
        redrawscreen();
    acase MENUITEM_ANALOG:
        flipbool((int*) &analog, command);
        updatemenu(MENUITEM_SENSITIVITY); // this is no mistake
    acase MENUITEM_DIPSWITCHES:
#ifdef WIN32
        if (DIPsWindowPtr)
        {   close_dips();
        } else
        {   edit_dips();
        }
#endif
#ifdef AMIGA
        edit_dips();
#endif
    acase MENUITEM_CONTROLS:
#ifdef WIN32
        if (ControlsWindowPtr)
        {   close_controls();
        } else
        {   view_controls();
        }
#endif
#ifdef AMIGA
        view_controls();
#endif
    acase MENUITEM_GAMEINFO:
#ifdef WIN32
        if (GameInfoWindowPtr)
        {   close_gameinfo();
        } else
        {   help_gameinfo();
        }
#endif
#ifdef AMIGA
        help_gameinfo();
#endif
    acase MENUITEM_VIEWMONITOR:
#ifdef WIN32
        if (MonitorWindowPtr)
        {   close_monitor();
        } else
        {   view_monitor();
        }
#endif
#ifdef AMIGA
        view_monitor();
#endif
    acase MENUITEM_VIEWMEMORY:
#ifdef WIN32
        if (MemoryWindowPtr)
        {   close_memory();
        } else
        {   view_memory();
        }
#endif
#ifdef AMIGA
        view_memory();
#endif
    acase MENUITEM_PALETTE:
#ifdef WIN32
        if (PaletteWindowPtr)
        {   close_palette();
        } else
        {   edit_palette();
        }
#endif
#ifdef AMIGA
        edit_palette();
#endif
    acase MENUITEM_SPRITEVIEWER:
#ifdef WIN32
        if (SpriteWindowPtr)
        {   close_spriteviewer();
        } else
        {   open_spriteviewer();
        }
#endif
#ifdef AMIGA
        open_spriteviewer();
#endif
    acase MENUITEM_SCALE2X:
        flipbool(&scale2x, command);
        redrawscreen();
    acase MENUITEM_ROTATE:
        flipbool(&rotate, command);
        redrawscreen();
    acase MENUITEM_RASTLINES:
        flipbool(&rastlines, command);
        redrawscreen();
    acase MENUITEM_VIEWHISCORES:
        updatehiscores(); // this is just for the user's convenience
        view_hiscores();
    acase MENUITEM_SENSEGAME:
        flipbool(&sensegame, command);
        updatemenu(MENUITEM_USEHACKS); // this is no mistake
    acase MENUITEM_PRINTER:
        flipbool(&echoprinter, command);
        updateprinter();
        updatemenu(MENUITEM_PRINTER); // maybe not needed?
    acase MENUITEM_FORCEFEEDBACK:
        flipbool(&useff, command);
        if (!useff)
        {   ff_off(0);
            ff_off(1);
        }
    acase MENUITEM_SPEECH:
        flipbool(&usespeech, command);
        updatespeech();
    acase MENUITEM_SIDEBAR:
#ifdef WIN32
        flipbool(&showsidebar, command);
        // updatemenu(MENUITEM_AUDIT); is not needed
        closewindow();
        openwindow();
#endif
#ifdef AMIGA
        if (ItemPtr->Flags & CHECKED)
        {   pending |= PENDING_SIDEBARON;
        } else
        {   pending |= PENDING_SIDEBAROFF;
        }
        // updatemenu(MENUITEM_AUDIT); is not needed
#endif
    acase MENUITEM_NARROW:
#ifdef WIN32
        if (wide == 1)
        {   // it was checked, so now we uncheck it
            wide = realwide = 2;
        } else
        {   // assert(wide == 2);
            // it was unchecked, so now we check it
            wide = realwide = 1;
        }
        updatemenu(MENUITEM_NARROW);
        resize(size);
#endif
#ifdef AMIGA
        if (ItemPtr->Flags & CHECKED)
        {   if (wide == 2)
            {   pending |= PENDING_NARROW;
        }   }
        else
        {   if (wide == 1)
            {   pending |= PENDING_WIDE;
        }   }
#endif
    acase MENUITEM_PAUSED:
#ifdef WIN32
        if (paused)
        {   emu_unpause();
        } else
        {   emu_pause(TRUE);
        }
#endif
#ifdef AMIGA
        if (ItemPtr->Flags & CHECKED)
        {   emu_pause(TRUE);
        } else
        {   emu_unpause();
        }
#endif
    acase MENUITEM_BREAK:
        if (paused)
        {   emu_unpause();
        } else
        {   emu_pause(TRUE);
        }
    acase MENUITEM_GREYSCALE:
#ifdef WIN32
        if (colourset == COLOURFUL)
        {   colourset = GREYSCALE;
        } else
        {   // assert(colourset == GREYSCALE);
            colourset = COLOURFUL;
        }
        updatemenu(MENUITEM_GREYSCALE);
#endif
#ifdef AMIGA
        if (ItemPtr->Flags & CHECKED)
        {   colourset = GREYSCALE;
        } else
        {   colourset = COLOURFUL;
        }
#endif
        updatemenu(MENUITEM_SCREENCOLOUR);
        fixupcolours();
        draw_cd2650();
        draw_phunsy();
        redrawscreen();
    acase MENUITEM_ANIMS:
        flipbool(&anims, command);
    acase MENUITEM_PAUSELOG:
        flipbool(&pauselog, command);
    acase MENUITEM_TRACEVERBOSE:
        flipbool(&traceverbose, command);
    acase MENUITEM_INSTRUCTIONS:
        flipbool(&loginstructions, command);
    acase MENUITEM_READWRITES:
        flipbool(&logreadwrites, command);
    acase MENUITEM_LOGEPROMS:
        flipbool(&logeprom, command);
    acase MENUITEM_LOGBIOS:
        flipbool(&logbios, command);
    acase MENUITEM_IOPORTS:
        flipbool(&logioports, command);
    acase MENUITEM_SUBROUTINES:
        flipbool(&logsubroutines, command);
    acase MENUITEM_AUTOSAVE:
        flipbool(&autosave, command);
    acase MENUITEM_CONFIRM:
        flipbool(&confirmable, command);
    acase MENUITEM_AUTOPAUSE:
        flipbool(&autopause, command);
    acase MENUITEM_CONSOLE:
        flipbool(&echoconsole, command);
    acase MENUITEM_HIGHRES:
        flipbool(&highres, command);
    acase MENUITEM_LOWERCASE:
        flipbool(&lowercase, command);
    acase MENUITEM_RETUNE:
        flipbool(&retune, command); // it doesn't affect the tone(s) currently playing, only future tones
    acase MENUITEM_WATCHREADS:
        flipbool(&watchreads, command);
    acase MENUITEM_USEHACKS:
        flipbool(&usehacks, command);
    acase MENUITEM_REQUIREBUTTON:
        flipbool(&requirebutton, command);
    acase MENUITEM_TRACECPU:
        flipbool(&trace, command);
    acase MENUITEM_SPRITESVISIBLE:
        flipbool(&spritesvisible, command);
    acase MENUITEM_CORNERS:
        flipbool(&drawcorners, command);
        docommand(command);
    acase MENUITEM_ANSI:
        flipbool(&useansi, command);
        docommand(command);
    acase MENUITEM_TURBO:
        flipbool((int*) &turbo, command);
        docommand(command);
    acase MENUITEM_UNLIT:
        flipbool(&drawunlit, command);
        docommand(command);
    acase MENUITEM_UNDITHER:
        flipbool(&undither, command);
        docommand(command);
    acase MENUITEM_COLLISIONS:
        flipbool((int*) &collisions, command);
        docommand(command);
    acase MENUITEM_DEMULTIPLEX:
        flipbool((int*) &demultiplex, command);
        docommand(command);
    acase MENUITEM_LOOP:
        flipbool(&loop, command);
        docommand(command);
    acase MENUITEM_TITLEBAR:
        flipbool(&showtitlebar, command);
        docommand(command);
    acase MENUITEM_TOOLBAR:
        flipbool(&showtoolbar, command);
        docommand(command);
    acase MENUITEM_USEGUIDEPIXEL:
        flipbool(&useguidepixel, command);
        docommand(command);
    acase MENUITEM_POINTER:
        flipbool(&showpointer, command);
        docommand(command);
    acase MENUITEM_AUTOFIRE:
        flipbool((int*) &autofire, command);
        docommand(command);
    acase MENUITEM_SWAPPED:
        flipbool((int*) &swapped, command);
        docommand(command);
    acase MENUITEM_CONFINE:
        flipbool(&confine, command);
        docommand(command);
    acase MENUITEM_SOUND:
        flipbool((int*) &sound, command);
        docommand(command);
    acase MENUITEM_CLI:
        flipbool(&cli, command);
        docommand(command);
    acase MENUITEM_FLAGLINE:
        flipbool((int*) &flagline, command);
        docommand(command);
    acase MENUITEM_DOTSPACE:
        flipbool(&dotspace, command);
        docommand(command);
    acase MENUITEM_EDIT:
        jump = FALSE;
        debug_edit();
    acase MENUITEM_JUMP:
        jump = TRUE;
        debug_edit();
    acase MENUITEM_MANUAL:
        help_manual();
    acase MENUITEM_QUIT:
        if (confirm())
        {   cleanexit(EXIT_SUCCESS);
        }
    acase MENUITEM_ABOUT:
#ifdef WIN32
        sound_off(FALSE);
        opening = TRUE;
        DISCARD DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_ABOUT), MainWindowPtr, AboutDlgProc);
        opening = FALSE;
        sound_on();
#endif
#ifdef AMIGA
        help_about();
#endif
    acase MENUITEM_REDEFINEKEYS:
#ifdef WIN32
        sound_off(FALSE);
        opening = TRUE;
        DISCARD DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINE), MainWindowPtr, RedefineDlgProc);
        opening = FALSE;
        sound_on();
#endif
#ifdef AMIGA
        redefinekeys();
#endif
    acase MENUITEM_LOGTOFILE:
        flipbool(&logfile, command);
        if (logfile)
        {   ;
#ifdef AMIGA
            logfile_open();
#endif
        } else
        {   logfile_close();
        }
    acase MENUITEM_FRAMESKIP:
#ifdef WIN32
        sound_off(FALSE);
        opening = TRUE;
        DISCARD DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_FRAMESKIP), MainWindowPtr, FrameSkipDlgProc);
        opening = FALSE;
        sound_on();
#endif
#ifdef AMIGA
        slider(0);
#endif
    acase MENUITEM_SENSITIVITY:
#ifdef WIN32
        sound_off(FALSE);
        opening = TRUE;
        DISCARD DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_SENSITIVITY), MainWindowPtr, SensitivityDlgProc);
        opening = FALSE;
        sound_on();
#endif
#ifdef AMIGA
        slider(1);
#endif
    acase MENUITEM_VOLUME:
#ifdef WIN32
        sound_off(TRUE); // we want to force a refresh of the sound
        opening = TRUE;
        DISCARD DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_VOLUME), MainWindowPtr, VolumeDlgProc);
        opening = FALSE;
        sound_on();
#endif
#ifdef AMIGA
        slider(2);
#endif
#ifdef WIN32
    acase MENUITEM_STARFIELDS:
        flipbool(&enhancestars, command);
        redrawscreen();
    acase MENUITEM_STRETCH:
        flipbool(&stretch, command);
        docommand(command);
    acase MENUITEM_MENUBAR:
        flipbool(&showmenubar, command);
        docommand(command);
    acase MENUITEM_SPRMAKER:
        openurl("SpriteMakerEA6.exe");
    acase MENUITEM_TXTMAKER:
        openurl("TextMaker1.exe");
    acase MENUITEM_INTERNAL:
        sound_off(TRUE); // do this BEFORE changing useinternal
        flipbool(&useinternal, command);
        sound_on();
    acase MENUITEM_REGISTER:
        project_register();
    acase MENUITEM_UNICONIFY:
        uniconify(FALSE);
    acase MENUITEM_FULLSCREEN:
        flipbool(&fullscreen, command);
        docommand(MENUITEM_FULLSCREEN);
    acase MENUITEM_CALIBRATE:
        calibrate();
    acase MENUITEM_REARRANGE:
        sound_off(FALSE);
        opening = TRUE;
        DISCARD DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REARRANGE), MainWindowPtr, RearrangeDlgProc);
        opening = FALSE;
        sound_on();
#endif
#ifdef AMIGA
    acase MENUITEM_FILTERED:
        flipbool(&led, command);
#ifndef __amigaos4__
        if (!lame) updatefilter();
#endif
    acase MENUITEM_CREATEICONS:
        flipbool(&icons, command);
    acase MENUITEM_AREXX:
        macro_arexx(TRUE);
#endif
}   }

EXPORT void handle_menu2(int command)
{
#ifdef AMIGA
    if (!(ItemPtr->Flags & CHECKED))
    {   return;
    }
#endif

    switch (command)
    {
    case MENUOPT_1XSIZE:
#ifdef WIN32
        resize(1);
#endif
#ifdef AMIGA
        if (size != 1)
        {   pending |= PENDING_1XSIZE;
        }
#endif
    acase MENUOPT_2XSIZE:
#ifdef WIN32
        resize(2);
#endif
#ifdef AMIGA
        if (size != 2)
        {   pending |= PENDING_2XSIZE;
        }
#endif
    acase MENUOPT_3XSIZE:
#ifdef WIN32
        resize(3);
#endif
#ifdef AMIGA
        if (size != 3)
        {   pending |= PENDING_3XSIZE;
        }
#endif
    acase MENUOPT_4XSIZE:
#ifdef WIN32
        resize(4);
#endif
#ifdef AMIGA
        if (size != 4)
        {   pending |= PENDING_4XSIZE;
        }
#endif
    acase MENUOPT_SLOW:
        fastcpu = FALSE;
        updatemenu(MENUITEM_CRYSTAL);
    acase MENUOPT_FAST:
        fastcpu = TRUE;
        updatemenu(MENUITEM_CRYSTAL);
    acase MENUOPT_NTSC:
        region = REGION_NTSC;
        updatemenu(MENUITEM_REGION);
    acase MENUOPT_PAL:
        region = REGION_PAL;
        updatemenu(MENUITEM_REGION);
    acase MENUOPT_2650A:
        supercpu = 0;
        updatemenu(MENUITEM_CPUTYPE);
    acase MENUOPT_2650B:
        supercpu = 1;
        updatemenu(MENUITEM_CPUTYPE);
    acase MENUOPT_CALM:
        style = STYLE_CALM;
        updatemenu(MENUITEM_MNEMONICS);
    acase MENUOPT_SIGNETICS:
        style = STYLE_SIGNETICS;
        updatemenu(MENUITEM_MNEMONICS);
    acase MENUOPT_ARCADIA:
        docommand(COMMAND_ARCADIA);
    acase MENUOPT_INTERTON:
        docommand(COMMAND_INTERTON);
    acase MENUOPT_ELEKTOR:
        docommand(COMMAND_ELEKTOR);
    acase MENUOPT_PIPBUG:
        docommand(COMMAND_PIPBUG);
    acase MENUOPT_INSTRUCTOR:
        docommand(COMMAND_INSTRUCTOR);
    acase MENUOPT_CD2650:
        docommand(COMMAND_CD2650);
    acase MENUOPT_ASTROWARS:
        docommand(COMMAND_ASTROWARS);
    acase MENUOPT_GALAXIA:
        docommand(COMMAND_GALAXIA);
    acase MENUOPT_MALZAK1:
        docommand(COMMAND_MALZAK1);
    acase MENUOPT_MALZAK2:
        docommand(COMMAND_MALZAK2);
    acase MENUOPT_CHAOS:
        docommand(COMMAND_CHAOS);
    acase MENUOPT_DOLPHIN:
        docommand(COMMAND_DOLPHIN);
    acase MENUOPT_PHUNSY:
        docommand(COMMAND_PHUNSY);
    acase MENUOPT_SORTBYNAME:
        sortby = SORTBYNAME;
        updatemenu(MENUITEM_SORTBY);
        if (showsidebar)
        {   buildlistbrowser();
        }
    acase MENUOPT_SORTBYMACHINE:
        sortby = SORTBYMACHINE;
        updatemenu(MENUITEM_SORTBY);
        if (showsidebar)
        {   buildlistbrowser();
        }
    acase MENUOPT_AMBER:
        wsc = 0;
        updatemenu(MENUITEM_SCREENCOLOUR);
        fixupcolours();
        draw_cd2650();
        draw_phunsy();
        redrawscreen();
    acase MENUOPT_GREEN:
        wsc = 1;
        updatemenu(MENUITEM_SCREENCOLOUR);
        fixupcolours();
        draw_cd2650();
        draw_phunsy();
        redrawscreen();
    acase MENUOPT_WHITE:
        wsc = 2;
        updatemenu(MENUITEM_SCREENCOLOUR);
        fixupcolours();
        draw_cd2650();
        draw_phunsy();
        redrawscreen();
    acase MENUOPT_LEFT_TRACKBALL:
        hostcontroller[0] = CONTROLLER_TRACKBALL;
        docommand(MENUITEM_LEFT);
    acase MENUOPT_LEFT_1STDJOY:
        hostcontroller[0] = CONTROLLER_1STDJOY;
        docommand(MENUITEM_LEFT);
    acase MENUOPT_LEFT_2NDDJOY:
        hostcontroller[0] = CONTROLLER_2NDDJOY;
        docommand(MENUITEM_LEFT);
    acase MENUOPT_LEFT_NONE:
        hostcontroller[0] = CONTROLLER_NONE;
        docommand(MENUITEM_LEFT);
    acase MENUOPT_RIGHT_TRACKBALL:
        hostcontroller[1] = CONTROLLER_TRACKBALL;
        docommand(MENUITEM_RIGHT);
    acase MENUOPT_RIGHT_1STDJOY:
        hostcontroller[1] = CONTROLLER_1STDJOY;
        docommand(MENUITEM_RIGHT);
    acase MENUOPT_RIGHT_2NDDJOY:
        hostcontroller[1] = CONTROLLER_2NDDJOY;
        docommand(MENUITEM_RIGHT);
    acase MENUOPT_RIGHT_NONE:
        hostcontroller[1] = CONTROLLER_NONE;
        docommand(MENUITEM_RIGHT);
#ifdef WIN32
    acase MENUOPT_ENG:
        freelanguage();
        language = LANGUAGE_ENG;
        changelanguage();
    acase MENUOPT_HOL:
        freelanguage();
        language = LANGUAGE_HOL;
        changelanguage();
    acase MENUOPT_FRA:
        freelanguage();
        language = LANGUAGE_FRA;
        changelanguage();
    acase MENUOPT_GER:
        freelanguage();
        language = LANGUAGE_GER;
        changelanguage();
    acase MENUOPT_ITA:
        freelanguage();
        language = LANGUAGE_ITA;
        changelanguage();
    acase MENUOPT_RUS:
        freelanguage();
        language = LANGUAGE_RUS;
        changelanguage();
    acase MENUOPT_SPA:
        freelanguage();
        language = LANGUAGE_SPA;
        changelanguage();
    acase MENUOPT_HIGHPRI:
        priority = PRI_HIGH;
        docommand(MENUITEM_PRIORITY);
    acase MENUOPT_NORMALPRI:
        priority = PRI_NORMAL;
        docommand(MENUITEM_PRIORITY);
    acase MENUOPT_LOWPRI:
        priority = PRI_LOW;
        docommand(MENUITEM_PRIORITY);
#endif
#ifdef AMIGA
    acase MENUOPT_LEFT_1STDPAD:
        hostcontroller[0] = CONTROLLER_1STDPAD;
        docommand(MENUITEM_LEFT);
    acase MENUOPT_LEFT_2NDDPAD:
        hostcontroller[0] = CONTROLLER_2NDDPAD;
        docommand(MENUITEM_LEFT);
    acase MENUOPT_LEFT_2NDAJOY:
        hostcontroller[0] = CONTROLLER_2NDAJOY;
        docommand(MENUITEM_LEFT);
    acase MENUOPT_LEFT_1STAPAD:
        hostcontroller[0] = CONTROLLER_1STAPAD;
        docommand(MENUITEM_LEFT);
    acase MENUOPT_LEFT_2NDAPAD:
        hostcontroller[0] = CONTROLLER_2NDAPAD;
        docommand(MENUITEM_LEFT);
    acase MENUOPT_RIGHT_1STDPAD:
        hostcontroller[1] = CONTROLLER_1STDPAD;
        docommand(MENUITEM_RIGHT);
    acase MENUOPT_RIGHT_2NDDPAD:
        hostcontroller[1] = CONTROLLER_2NDDPAD;
        docommand(MENUITEM_RIGHT);
    acase MENUOPT_RIGHT_2NDAJOY:
        hostcontroller[1] = CONTROLLER_2NDAJOY;
        docommand(MENUITEM_RIGHT);
    acase MENUOPT_RIGHT_1STAPAD:
        hostcontroller[1] = CONTROLLER_1STAPAD;
        docommand(MENUITEM_RIGHT);
    acase MENUOPT_RIGHT_2NDAPAD:
        hostcontroller[1] = CONTROLLER_2NDAPAD;
        docommand(MENUITEM_RIGHT);
#endif
}   }

EXPORT UBYTE readport(int port)
{   switch (machine)
    {
    case PIPBUG:
        return p_readport(port);
    acase INSTRUCTOR:
        return s_readport(port);
    acase CD2650:
        return c_readport(port);
    acase MALZAK:
        return m_readport(port);
    acase ASTROWARS:
    case GALAXIA:
        return g_readport(port);
    acase CHAOS:
        return h_readport(port);
    acase DOLPHIN:
        return d_readport(port);
    acase PHUNSY:
        return phunsy_readport(port);
    adefault: // eg. ARCADIA, INTERTON, ELEKTOR
        logport(port, 0, FALSE, FALSE);
        return 0; // arbitrary
}   }

EXPORT void writeport(int port, UBYTE data)
{   switch (machine)
    {
    case PIPBUG:
        p_writeport(port, data);
    acase INSTRUCTOR:
        s_writeport(port, data);
    acase CD2650:
        c_writeport(port, data);
    acase MALZAK:
        m_writeport(port, data);
    acase ASTROWARS:
    case GALAXIA:
        g_writeport(port, data);
    acase CHAOS:
        h_writeport(port, data);
    acase DOLPHIN:
        d_writeport(port, data);
    acase PHUNSY:
        phunsy_writeport(port, data);
    adefault: // eg. ARCADIA, INTERTON, ELEKTOR
        logport(port, data, FALSE, TRUE);
}   }

EXPORT void spritebutton(int x, int y)
{   switch (machine)
    {
    case ARCADIA:
        if (x >= 0 && x <= 7 && y >= 0 && y <= 7)
        {   if (spritecontents[y] & (128 >> x))
            {   memory[0x1980 + (viewingsprite * 8) + y] &= ~(128 >> x); // turn off
            } else
            {   memory[0x1980 + (viewingsprite * 8) + y] |=   128 >> x ; // turn on
            }
            update_spriteviewer(FALSE);
        }
    acase INTERTON:
    case ELEKTOR:
    case ASTROWARS:
    case GALAXIA:
    case MALZAK:
        if (x >= 0 && x <= 7 && y >= 0 && y <= 9)
        {   if (spritecontents[y] & (128 >> x))
            {   memory[pvibase + pvi_spritedata[viewingsprite % 4] + y] &= ~(128 >> x); // turn off
            } else
            {   memory[pvibase + pvi_spritedata[viewingsprite % 4] + y] |=   128 >> x ; // turn on
            }
            update_spriteviewer(FALSE);
}   }   }

EXPORT void serializeconfig(void)
{   UBYTE t;
    int   i, j;

    if (serializemode == SERIALIZE_READ)
    {   offset = 2;
    } else
    {   // assert(serializemode == SERIALIZEMODE_WRITE);
        offset = 0;
        WriteByte(0xC6);                      //   0
        WriteByte(0x02);                      //   1
    }
    serialize_byte_int(  &configversion);     //   2
    serialize_byte_int(  &configmachine);     //   3
    serialize_byte_int(  &size             ); //   4
    serialize_byte_int(  &wide             ); //   5
    serialize_byte_int(  &requirebutton    ); //   6
    serialize_byte_ulong(&analog           ); //   7 could change analog to int
    serialize_byte_ulong(&autofire         ); //   8
    serialize_byte_ulong(&swapped          ); //   9
    serialize_byte_ulong(&paused           ); //  10
    serialize_byte_ulong(&turbo            ); //  11
    serialize_byte_ulong(&collisions       ); //  12
    serialize_byte_ulong(&sound            ); //  13
    serialize_byte_ulong(&demultiplex      ); //  14 could change demultiplex to int
    serialize_byte_int(  &autosave         ); //  15
    serialize_byte_int(  &showtitlebar     ); //  16
    serialize_byte_int(  &showtoolbar      ); //  17
    serialize_byte_ulong(&flagline         ); //  18
    serialize_byte_int(  &showpointer      ); //  19
    serialize_byte_int(  &loop             ); //  20
    serialize_byte_int(  &logreadwrites    ); //  21
    serialize_byte_int(  &logeprom         ); //  22
    serialize_byte_int(  &memmap           ); //  23
    serialize_byte_int(  &machine          ); //  24
    serialize_byte_int(  &loginstructions  ); //  25
    serialize_byte_int(  &watchreads       ); //  26
    serialize_byte_int(  &sensegame        ); //  27
    serialize_byte_int(  &usehacks         ); //  28
    serialize_byte_int(  &dotspace         ); //  29
    serialize_byte_ulong(&region           ); //  30 could change region to int
    serialize_byte_int(  &logbios          ); //  31
    serialize_byte_int(  &logsubroutines   ); //  32
    serialize_byte_int(  &sensitivity      ); //  33
    serialize_byte_int(  &hostvolume       ); //  34
    serialize_byte_int(  &frameskip        ); //  35
    serialize_byte_int(  &showsidebar      ); //  36
    serialize_byte_int(  &sortby           ); //  37
    serialize_byte_int(  &colourset        ); //  38
    serialize_byte_int(  &confine          ); //  39
    serialize_byte_int(  &traceverbose     ); //  40
    serialize_byte_int(  &useansi          ); //  41
    serialize_byte_int(  &autopause        ); //  42
    serialize_byte_int(  &echoprinter      ); //  43
    serialize_byte_int(  &useguidepixel    ); //  44
    serialize_byte_int(  &undither         ); //  45
    serialize_byte_int(  &rastlines        ); //  46
    serialize_byte_int(  &anims            ); //  47
    serialize_byte_int(  &drawcorners      ); //  48
    serialize_byte_int(  &drawunlit        ); //  49
    serialize_byte_int(  &lowercase        ); //  50
    serialize_byte_int(  &echoconsole      ); //  51
    serialize_byte_int(  &highres          ); //  52
    serialize_byte_int(  &retune           ); //  53
    serialize_byte_int(  &shadowing        ); //  54
    serialize_byte_int(  &inverse          ); //  55
    serialize_byte_int(  &wsc              ); //  56
    serialize_byte_int(  &scale2x          ); //  57
    serialize_byte_int(  &s_io             ); //  58
    serialize_byte_int(  &s_intselector    ); //  59
    serialize_byte_int(  &s_intdir         ); //  60
    serialize_byte_int(  &base             ); //  61
    serialize_byte_int(  &logfile          ); //  62
    serialize_byte_int(  &logioports       ); //  63
    serialize_byte_int(  &useff            ); //  64
    serialize_byte_int(  &spritesvisible   ); //  65
    serialize_byte_int(  &confirmable      ); //  66
    serialize_byte_int(  &usespeech        ); //  67
    serialize_byte_int(  &rotate           ); //  68
    serialize_byte_int(  &pauselog         ); //  69
    serialize_byte_int(  &fastcpu          ); //  70
    serialize_byte(      &g_dips           ); //  71
    serialize_byte(      &s_toggles        ); //  72
    serialize_word_int(  &winleftx         ); //  73.. 74
    serialize_word_int(  &wintopy          ); //  75.. 76
    serialize_word_int(  &sidebarwidth     ); //  77.. 78
    serialize_word_int(  &regionstart      ); //  79.. 80

    // assert(offset == 81);

    for (i = 0; i < 24; i++)                  //  81..224 (144 bytes)
    {
#ifdef AMIGA
        t = (UBYTE) ((pencolours[COLOURFUL][i].red   & 0xFF000000) >> 24);
#endif
#ifdef WIN32
        t = (UBYTE) ((pens[COLOURFUL][i]             & 0x00FF0000) >> 16);
#endif
        serialize_byte(&t);
        if (serializemode == SERIALIZE_READ)
        {
#ifdef AMIGA
            pencolours[COLOURFUL][i].red   = (t << 24) | (t << 16) | (t << 8) | t;
#endif
#ifdef WIN32
            pens[COLOURFUL][i]             = t << 16;
#endif

        }
#ifdef AMIGA
        t = (UBYTE) ((pencolours[COLOURFUL][i].green & 0xFF000000) >> 24);
#endif
#ifdef WIN32
        t = (UBYTE) ((pens[COLOURFUL][i]             & 0x0000FF00) >>  8);
#endif
        serialize_byte(&t);
        if (serializemode == SERIALIZE_READ)
        {
#ifdef AMIGA
            pencolours[COLOURFUL][i].green = (t << 24) | (t << 16) | (t << 8) | t;
#endif
#ifdef WIN32
            pens[COLOURFUL][i]             |= t << 8;
#endif
        }

#ifdef AMIGA
        t = (UBYTE) ((pencolours[COLOURFUL][i].blue  & 0xFF000000) >> 24);
#endif
#ifdef WIN32
        t = (UBYTE) ( pens[COLOURFUL][i]             & 0x000000FF       );
#endif
        serialize_byte(&t);
        if (serializemode == SERIALIZE_READ)
        {
#ifdef AMIGA
            pencolours[COLOURFUL][i].blue  = (t << 24) | (t << 16) | (t << 8) | t;
#endif
#ifdef WIN32
            pens[COLOURFUL][i]             |= t;
#endif
        }

#ifdef AMIGA
        t = (UBYTE) ((pencolours[GREYSCALE][i].red   & 0xFF000000) >> 24);
#endif
#ifdef WIN32
        t = (UBYTE) ((pens[GREYSCALE][i]             & 0x00FF0000) >> 16);
#endif
        serialize_byte(&t);
        if (serializemode == SERIALIZE_READ)
        {
#ifdef AMIGA
            pencolours[GREYSCALE][i].red   = (t << 24) | (t << 16) | (t << 8) | t;
#endif
#ifdef WIN32
            pens[GREYSCALE][i]             = t << 16;
#endif
        }

#ifdef AMIGA
        t = (UBYTE) ((pencolours[GREYSCALE][i].green & 0xFF000000) >> 24);
#endif
#ifdef WIN32
        t = (UBYTE) ((pens[GREYSCALE][i]             & 0x0000FF00) >>  8);
#endif
        serialize_byte(&t);
        if (serializemode == SERIALIZE_READ)
        {
#ifdef AMIGA
            pencolours[GREYSCALE][i].green = (t << 24) | (t << 16) | (t << 8) | t;
#endif
#ifdef WIN32
            pens[GREYSCALE][i]             |= t << 8;
#endif
        }

#ifdef AMIGA
        t = (UBYTE) ((pencolours[GREYSCALE][i].blue  & 0xFF000000) >> 24);
#endif
#ifdef WIN32
        t = (UBYTE) ( pens[GREYSCALE][i]             & 0x000000FF       );
#endif
        serialize_byte(&t);
        if (serializemode == SERIALIZE_READ)
        {
#ifdef AMIGA
            pencolours[GREYSCALE][i].blue  = (t << 24) | (t << 16) | (t << 8) | t;
#endif
#ifdef WIN32
            pens[GREYSCALE][i]             |= t;
#endif
    }   }
    fixshadowcolours();

    // assert(offset == 225);

    if (configversion >= 2)
    {   serialize_byte_int(  &style            );
        serialize_byte_int(  &supercpu         );
    }
    if (configversion <= 2)
    {   if (wsc == 22)
        {   wsc = 0; // amber
        } elif (wsc == 5)
        {   wsc = 1; // green
        } else
        {   wsc = 2; // grey
    }   }

    // all offsets shown from this point onwards are assuming V15.25 (ie. V1) format
    // add 2 to offsets to get the V15.3 (ie. V2) offsets

#ifdef AMIGA
    if (configmachine == CONFIGMACHINE_AMIGA) // native
#endif
#ifdef WIN32
    if (configmachine == CONFIGMACHINE_IBM) // native
#endif
    {   serialize_byte_int(  &hostcontroller[0]); // 225
        serialize_byte_int(  &hostcontroller[1]); // 226
        for (i = 0; i < 4; i++)
        {   serialize_word(&console[i]);          // 227..234 (  8 bytes)
        }
        for (i = 0; i < 2; i++)
        {   for (j = 0; j < 28; j++)
            {   serialize_word(&keypads[i][j]);   // 235..346 (112 bytes)
        }   }

        // assert(offset == 347);

#ifdef AMIGA
        serialize_byte_int(  &icons            ); // 347
        serialize_byte_int(  &led              ); // 348
        if (serializemode == SERIALIZE_READ)
        {   offset += 30;                         // 349..378 ( 30 bytes)
        } else
        {   // assert(serializemode == SERIALIZE_WRITE);
            for (i = 0; i < 30; i++)
            {   WriteByte(0);                     // 349..378 ( 30 bytes)
        }   }
#endif
#ifdef WIN32
        serialize_byte_int(  &fullscreen       ); // 347
        serialize_byte_int(  &showmenubar      ); // 348
        serialize_byte_int(  &useinternal      ); // 349
        serialize_byte_int(  &stretch          ); // 350
        serialize_byte_int(  &enhancestars     ); // 351
        serialize_byte_int(  &showlegend       ); // 352
        serialize_byte_int(  &language         ); // 353
        serialize_byte_int(  &priority         ); // 354
        serialize_byte_int(  &reassociate      ); // 355
        for (i = 0; i < 7; i++)
        {   serialize_byte_int(&assoc[i]       ); // 356..362 (  7 bytes)
        }
        for (i = 0; i < 2; i++)
        {   for (j = 0; j < 8; j++)
            {   serialize_byte(&button[i][j]   ); // 363..378 ( 16 bytes)
        }   }
#endif
        // assert(offset == 379);

        if (serializemode == SERIALIZE_READ)
        {   strcpy(pathpart, &IOBuffer[offset]);  // 379+
        } else
        {   // assert(serializemode == SERIALIZE_WRITE);
            strcpy(&IOBuffer[offset], pathpart);  // 379+
        }
        offset += strlen(pathpart) + 1;
    } else // non-native
    {   // assert(serializemode == SERIALIZE_READ);
        // assert(offset == 225);
        offset += 124;                            // 225..378 (124 bytes)
        // assert(offset == 379);
        offset += strlen(&IOBuffer[offset]) + 1;  // 379+
}   }

EXPORT void fixupcolours(void)
{   TRANSIENT       int x, y;
    PERSIST   const int screencolour[3] = { ORANGE, GREEN, WHITE };

    if (inverse)
    {   if (colourset == COLOURFUL)
        {   cy_bgc = screencolour[wsc];
        } else
        {   cy_bgc = WHITE;
        }
        cy_fgc = BLACK;
    } else
    {   if (colourset == COLOURFUL)
        {   cy_fgc = screencolour[wsc];
        } else
        {   cy_fgc = WHITE;
        }
        cy_bgc = BLACK;
    }

    for (x = 0; x < machines[machine].width; x++)
    {   for (y = 0; y < machines[machine].height; y++)
        {   screen[x][y] = (UBYTE) cy_bgc;
    }   }

    resetcolour(COLOURFUL, GREY2, FALSE);
    resetcolour(COLOURFUL, GREY3, FALSE);
    resetcolour(COLOURFUL, GREY4, FALSE);
    resetcolour(COLOURFUL, GREY5, FALSE);
    resetcolour(COLOURFUL, GREY6, FALSE);
    resetcolour(COLOURFUL, GREY7, FALSE);

#ifdef AMIGA
    changecolours();
    if (PaletteWindowPtr && whichpen >= GREY2 && whichpen <= GREY7)
    {   updatesliders();
        updatewheel();
        updatebrightness();
    }
#endif
#ifdef WIN32
    if (PaletteWindowPtr && whichpen >= GREY2 && whichpen <= GREY7)
    {   DISCARD SendMessage
        (   GetDlgItem(PaletteWindowPtr, IDC_RED),
            TBM_SETPOS,
            TRUE,
            255 - ((pens[colourset][whichpen] & 0x00FF0000) >> 16)
        );
        DISCARD SendMessage
        (   GetDlgItem(PaletteWindowPtr, IDC_GREEN),
            TBM_SETPOS,
            TRUE,
            255 - ((pens[colourset][whichpen] & 0x0000FF00) >>  8)
        );
        DISCARD SendMessage
        (   GetDlgItem(PaletteWindowPtr, IDC_BLUE),
            TBM_SETPOS,
            TRUE,
            255 -  (pens[colourset][whichpen] & 0x000000FF)
        );
    }
#endif
}
