/* engine2.c: Miscellaneous host-independent code

INCLUDES--------------------------------------------------------------- */

#ifdef __VBCC__
    #define AMIGA
#endif
#ifdef AMIGA
    #include "amiga.h"
#endif
#ifdef _lint
    #define WIN32
    #define _inline
#endif
#ifdef WIN32
    #include "ibm.h"
    #define EXEC_TYPES_H
    #include "resource.h"
#endif

#include <ctype.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#ifdef WIN32
    #include <conio.h>
#endif

#include "aa.h"
#define CATCOMP_NUMBERS
#define CATCOMP_BLOCK
#include "aa_strings.h"
#ifdef AMIGA
    #include <proto/locale.h>  // GetCatalogStr()
#endif

#include "engine2.h"

// DEFINES----------------------------------------------------------------

// (none)

// EXPORTED VARIABLES-----------------------------------------------------

EXPORT ASCREEN a_screen[A_BOXWIDTH][A_BOXHEIGHT];
EXPORT UBYTE  p_digit[4],
              vdu[MAX_VDUCOLUMNS][MAX_VDUROWS];
EXPORT FLAG   changed                  = TRUE,
              freshly                  = TRUE,
              golf                     = FALSE, // may not require initialization
              hadpaused                = TRUE,
              life                     = FALSE,
              printer                  = FALSE, // may not require initialization
              undithernow              = FALSE; // may not require initialization
EXPORT UWORD  mirror_r[32768],
              mirror_w[32768];
EXPORT ULONG  memflags[32768],
              waittill                 = 0;  // must be initialized!
EXPORT TEXT   arg1[MAX_PATH + 80 + 1],
              arg2[MAX_PATH + 80 + 1],
              arg3[MAX_PATH + 80 + 1],
              arg4[MAX_PATH + 80 + 1],
              arg5[MAX_PATH + 80 + 1],
              autotext[GAMEINFOLINES][80 + 1],
              fn_config[MAX_PATH + 1],
              fn_screenshot[MAX_PATH + 1],   // complete pathname of the screenshot
              pgmtext[6 * KILOBYTE],
              OneStr[10][1 + 1],
              TwoStr[8][2 + 1],
              FourStr[17][4 + 1],
              FiveStr[24][5 + 1],
              MemoryStr[MEMGADGETS][2 + 1],
              MonitorStr[MONITORGADS][2 + 1],
              ClockStr[COMMASTRLEN]    = "",
              FrameStr[COMMASTRLEN]    = "",
              RasterStr[3 + 1]         = "",
              TimeStr[11 + 1]          = "",
              ProjectBuffer[PROJECTSIZE];
// max. size of EAP project is 315 (+ 1).
// max. size of EAL project is 954 (+ 1).
// max. size of PHUNSY screen clipboard text is ((64+2)*32)+1
EXPORT int    base                     = BASE_HEX,
              drawcorners              = FALSE,
              drawunlit                = TRUE,
              forceinterrupt           = 0,
              foundaofs,
              founddirs,
              foundgood,
              foundbad,
              foundsnaps,
              foundicons,
              foundunknown,
              highres                  = FALSE,
              key1                     = 2,
              key2                     = 1,
              key3                     = 3,
              key4                     = 0,
              p1bgcol[4],
              p2bgcol[4],
              p1sprcol[6],
              p2sprcol[6],
              paddleup                 = -1,
              paddledown               = -1,
              paddleleft               = -1,
              paddleright              = -1,
              pbx                      = 0,
              pvibase,
              queuepos                 = 0,
              rastn                    = -1,
              regionstart              = 0x1800,
              romsize,
              runtorastline            = FALSE,
              runtodma                 = FALSE,
              runtoframe               = FALSE,
              runtointerrupt           = FALSE,
              runtoloopend             = FALSE,
              serializemode,
              speakoffset              = 0,
              starfield                = 16,
              vducols,
              whichoverlay;
EXPORT FILE*  LogfileHandle            = NULL;
EXPORT STRPTR colourname[8];

EXPORT struct KeyTableStruct keytable[16] =
{   { A_P1MIDDLEKEYS, 0x01, IE_P1MIDDLEKEYS, 0x10, "0",  "+/2"    }, //  0
    { A_P1LEFTKEYS,   0x08, IE_P1LEFTKEYS,   0x80, "1",  "RCAS/D" }, //  1
    { A_P1MIDDLEKEYS, 0x08, IE_P1MIDDLEKEYS, 0x80, "2",  "WCAS/E" }, //  2
    { A_P1RIGHTKEYS,  0x08, IE_P1RIGHTKEYS,  0x80, "3",  "C/F"    }, //  3
    { A_P1LEFTKEYS,   0x04, IE_P1LEFTKEYS,   0x40, "4",  "BP/9"   }, //  4
    { A_P1MIDDLEKEYS, 0x04, IE_P1MIDDLEKEYS, 0x40, "5",  "REG/A"  }, //  5
    { A_P1RIGHTKEYS,  0x04, IE_P1RIGHTKEYS,  0x40, "6",  "8/B"    }, //  6
    { A_P1LEFTKEYS,   0x02, IE_P1LEFTKEYS,   0x20, "7",  "PC/5"   }, //  7
    { A_P1MIDDLEKEYS, 0x02, IE_P1MIDDLEKEYS, 0x20, "8",  "MEM/6"  }, //  8
    { A_P1RIGHTKEYS,  0x02, IE_P1RIGHTKEYS,  0x20, "9",  "4/7"    }, //  9
    { A_P1LEFTKEYS,   0x01, IE_P1LEFTKEYS,   0x10, "Cl", "-/1"    }, // 10
    { A_P1RIGHTKEYS,  0x01, IE_P1RIGHTKEYS,  0x10, "En", "0/3"    }, // 11
    { A_P1PALLADIUM,  0x08, 0,                  0, "x1", "?"      }, // 12
    { A_P1PALLADIUM,  0x04, 0,                  0, "x2", "?"      }, // 13
    { A_P1PALLADIUM,  0x02, 0,                  0, "x3", "?"      }, // 14
    { A_P1PALLADIUM,  0x01, 0,                  0, "x4", "?"      }  // 15
};

// IMPORTED VARIABLES-----------------------------------------------------

IMPORT ASCREEN screen[BOXWIDTH][BOXHEIGHT],
               ctrlscreen[MINIWIDTH][MINIHEIGHT];
IMPORT FLAG    belmach,
               belmach0,
               bps,
               fiftytheme,
               jump,
               piano,
               p_morse,
               music,
               wps;
IMPORT int     anims,
               ax[2], // analog paddle X-coords
               ay[2], // analog paddle Y-coords
               charset,
               cli,
               colourset,
               configmachine,
               configversion,
               consoleopen,
               cpl,
               crippled,
               currentgame,
               filesize,
               firstmnemonic,
               flip,
               flips,
               foundgames,
               frameskip,
               fullscreen,
               game,
               inframe,
               lastmnemonic,
               latestsprite,
               loader,
               logeprom,
               logfile,
               loginstructions,
               logioports,
               logbios,
               logreadwrites,
               logsubroutines,
               loop,
               machine,
               memmap,
               malzak_x,
               malzak_y,
               m_bank1,
               m_bank2,
               offset,
               overcalc,
               pauselog,
               rast,
               recmode,
               recsize,
               s_io,
               s_intselector,
               s_intdir,
               sensegame,
               sensitivity,
               size,
               sprviewcolour,
               tonum,
               trace,
               traceverbose,
               undither,
               usehacks,
               vblank,
               viewingsprite,
               watchreads,
               whichgame,
               wide,
               zinterrupt;
IMPORT ULONG   analog,
               autofire,
               collisions,
               cycles,
               demultiplex,
               downframes,
               elapsed,
               flagline,
               fps,
               frames,
               paused,
               region,
               totalframes,
               swapped,
               turbo;
IMPORT UBYTE   banked,
               dolphin_leds[4],
               g_bank[0x400],
               g_dips,
               g_paddles,
               m_field[16][16],
               memory[32768],
               mg_console,
               oldkeys[KEYS],
               newkeys[KEYS],
               protect[4],
               psl,
               psu,
               q_bank[16][16384],
               r[7],
               s_glow,
               s_toggles,
               startaddr_h,
               startaddr_l,
               sx[2],
               sy[2],
               u_bank[16][2048];
IMPORT UWORD   iar,
               ras[8];
IMPORT TEXT    filepart[MAX_PATH + 1],
               fn_game[MAX_PATH + 1],
               friendly[21 + 1],
               letters[8 + 1],
               pathpart[MAX_PATH + 1];
IMPORT STRPTR  ggnode;
IMPORT UBYTE*  IOBuffer;
IMPORT FILE*   MacroHandle;

IMPORT const UBYTE                    from_a[16];
IMPORT const UWORD                    pvi_spritedata[4];
IMPORT const TEXT                     arcadia_chars[64 + 1],
                                      cd2650_chars[128 + 1],
                                      elektor_chars1[256 + 1],
                                      instructor_chars[256 + 1];
IMPORT       struct KindStruct        filekind[KINDS];
IMPORT const struct KnownStruct       known[KNOWNGAMES];
IMPORT       int                      sidebar[FOUNDGAMES_MAX];
IMPORT       struct AuditStruct       auditlist[FOUNDGAMES_MAX];
IMPORT const struct MachineStruct     machines[MACHINES];
IMPORT const struct HostMachineStruct hostmachines[MACHINES];
IMPORT const int                      from_m_spr[8],
                                      from_g_w_spr[8],
                                      from_g_x_spr[8];
IMPORT       char                     psubits[3 + 1],
                                      pslbits[6 + 1];
IMPORT const STRPTR                   astrowars_chars,
                                      ccstring[4],
                                      galaxia_chars;

#ifdef PROFILE
    IMPORT ULONG                      profile[256];
#endif

#ifdef AMIGA
    IMPORT struct Catalog*      CatalogPtr;
    IMPORT struct Window       *GameInfoWindowPtr,
                               *MainWindowPtr,
                               *MemoryWindowPtr,
                               *MonitorWindowPtr;
    IMPORT LONG                 pens[PENS];
    IMPORT struct PaletteStruct pencolours[2][PENS];
    IMPORT struct Gadget   // CPU chip...
                          *RGad[7],
                          *CCGad,
                          *OneGad[10],
                          *RASGad[8],
                          *IARGad,
                          *ClockGad,
                          *FrameGad,
                          *RasterGad,
                          *TimeGad,
                          // XVI chips...
                          *SPRITEnAXYGad[12],
                          *SPRITEnBXYGad[12],
                          *monitorgad[MONITORGADS],
                          // PSGs
                          *PITCHAGad[2],
                          *PITCHBGad[2],
                          *PITCHCGad[2],
                          *PERIODGad[2];
    IMPORT UWORD           pending;
    IMPORT int             blockarexx,
                           quiet;
#endif
#ifdef WIN32
    IMPORT int             CatalogPtr, // APTR doesn't work
                           language;
    IMPORT HWND            hRegionListBox,
                           GameInfoWindowPtr,
                           MainWindowPtr,
                           MemoryWindowPtr,
                           MonitorWindowPtr,
                           // CPU chip...
                           RGad[7],
                           CCGad,
                           OneGad[10],
                           RASGad[8],
                           IARGad,
                           ClockGad,
                           FrameGad,
                           RasterGad,
                           TimeGad,
                           // XVI chips...
                           SPRITEnAXYGad[12],
                           SPRITEnBXYGad[12],
                           monitorgad[MONITORGADS],
                           // PSGs
                           PITCHAGad[2],
                           PITCHBGad[2],
                           PITCHCGad[2],
                           PERIODGad[2];
    IMPORT ULONG           pens[2][PENS];
#endif

// MODULE VARIABLES-------------------------------------------------------

MODULE       TEXT   bits[4][10][8 + 1];
MODULE       UWORD  monitorras[8],
                    monitoriar,
                    // PSGs
                    monitorPITCHA[2],
                    monitorPITCHB[2],
                    monitorPITCHC[2],
                    monitorPERIOD[2];
MODULE       UBYTE  memorycontents[MEMGADGETS],
                    monitorr[7],
                    monitorpsu,
                    monitorpsl,
                    monitorSPRITEnAX[12],
                    monitorSPRITEnAY[12],
                    monitorSPRITEnBX[12],
                    monitorSPRITEnBY[12],
                    monitorcontents[MONITORGADS];
MODULE       ULONG  lateness,
                    oldframes,
                    oldtime,
                    newtime,
                    // XVI...
                    monitorelapsed,
                    monitorframes;
MODULE       int    cursor,
                    monitorrast,
                    oldmachine    = -1, // important!
                    oldmemmap     = -1, // important!
                    ZIPBufferSize;
MODULE       STRPTR ZIPBufferPtr;

// MODULE STRUCTURES------------------------------------------------------

// (none)

// MODULE ARRAYS----------------------------------------------------------

MODULE const UWORD monitoraddr[MONITORGADS] = {
    PVI_P2PADDLE,               //   0 -IEMG
    PVI_PITCH,                  //   1 -IEMG
    PVI_SIZES,                  //   2 -IEMG
    PVI_SCORECTRL,              //   3 -IEMG
    PVI_SCORELT,                //   4 -IEMG
    PVI_SCORERT,                //   5 -IEMG
    PVI_BGCOLOUR,               //   6 -IEMG
    PVI_BGCOLLIDE,              //   7 -IEMG
    PVI_SPRITECOLLIDE,          //   8 -IEMG
    PVI_SPR01COLOURS,           //   9 -IEMG
    PVI_SPR23COLOURS,           //  10 -IEMG
    PVI_P1PADDLE,               //  11 -IEMG
    IE_P1LEFTKEYS,              //  12 -IE--
    IE_P1MIDDLEKEYS,            //  13 -IE--
    IE_P1RIGHTKEYS,             //  14 -IE--
    IE_CONSOLE,                 //  15 -IE--
    IE_P2LEFTKEYS,              //  16 -IE--
    IE_P2MIDDLEKEYS,            //  17 -IE--
    IE_P2RIGHTKEYS,             //  18 -IE--
    IE_NOISE,                   //  19 -IE--
    E_AMPLITUDEA1,              //  20 --E--
    E_AMPLITUDEB1,              //  21 --E--
    E_AMPLITUDEC1,              //  22 --E--
    E_PITCHD1,                  //  23 --E--
    E_MIXER1,                   //  24 --E--
    E_SHAPE1,                   //  25 --E--
    E_PORTA1,                   //  26 --E--
    E_PORTB1,                   //  27 --E--
    E_AMPLITUDEA2,              //  28 --E--
    E_AMPLITUDEB2,              //  29 --E--
    E_AMPLITUDEC2,              //  30 --E--
    E_PITCHD2,                  //  31 --E--
    E_MIXER2,                   //  32 --E--
    E_SHAPE2,                   //  33 --E--
    E_PORTA2,                   //  34 --E--
    E_PORTB2,                   //  35 --E--
    0x1F0E,                     //  36 -I---
    0x1F0F,                     //  37 -I---
    0x1F1E,                     //  38 -I---
    0x1F1F,                     //  39 -I---
    0x1F4E,                     //  40 -I---
    0x1F4F,                     //  41 -I---
    0x1F50,                     //  42 -I---
    0x1F51,                     //  43 -I---
    0x1F52,                     //  44 -I---
    0x1F53,                     //  45 -I---
    0x1F54,                     //  46 -I---
    0x1F55,                     //  47 -I---
    0x1F56,                     //  48 -I---
    0x1F57,                     //  49 -I---
    0x1F58,                     //  50 -I---
    0x1F59,                     //  51 -I---
    0x1F5A,                     //  52 -I---
    0x1F5B,                     //  53 -I---
    0x1F5C,                     //  54 -I---
    0x1F5D,                     //  55 -I---
    0x1F5E,                     //  56 -I---
    0x1F5F,                     //  57 -I---
    0x1F60,                     //  58 -I---
    0x1F61,                     //  59 -I---
    0x1F62,                     //  60 -I---
    0x1F63,                     //  61 -I---
    0x1F64,                     //  62 -I---
    0x1F65,                     //  63 -I---
    0x1F66,                     //  64 -I---
    0x1F67,                     //  65 -I---
    0x1F68,                     //  66 -I---
    0x1F69,                     //  67 -I---
    0x1F6A,                     //  68 -I---
    0x1F6B,                     //  69 -I---
    0x1F6C,                     //  70 -I---
    0x1F6D,                     //  71 -I---
    0x1FAD,                     //  72 -I---
    A_CONSOLE,                  //  73 A----
    A_RESOLUTION,               //  74 A----
    A_BGCOLLIDE,                //  75 A----
    A_P1PADDLE,                 //  76 A----
    A_P2PADDLE,                 //  77 A----
    A_PITCH,                    //  78 A----
    A_BGCOLOUR,                 //  79 A----
    A_SPRITECOLLIDE,            //  80 A----
    A_P1LEFTKEYS,               //  81 A----
    A_P2LEFTKEYS,               //  82 A----
    A_VOLUME,                   //  83 A----
    A_VSCROLL,                  //  84 A----
    A_SPRITES01CTRL,            //  85 A----
    A_P1MIDDLEKEYS,             //  86 A----
    A_P2MIDDLEKEYS,             //  87 A----
    A_CHARLINE,                 //  88 A----
    A_SPRITES23CTRL,            //  89 A----
    A_P1RIGHTKEYS,              //  90 A----
    A_P2RIGHTKEYS,              //  91 A----
    A_P1PALLADIUM,              //  92 A----
    A_P2PALLADIUM,              //  93 A----
    0x18D0,                     //  94 A----
    0x18D1,                     //  95 A----
    0x18D2,                     //  96 A----
    0x18D3,                     //  97 A----
    0x18D4,                     //  98 A----
    0x18D5,                     //  99 A----
    0x18D6,                     // 100 A----
    0x18D7,                     // 101 A----
    0x18D8,                     // 102 A----
    0x18D9,                     // 103 A----
    0x18DA,                     // 104 A----
    0x18DB,                     // 105 A----
    0x18DC,                     // 106 A----
    0x18DD,                     // 107 A----
    0x18DE,                     // 108 A----
    0x18DF,                     // 109 A----
    0x18E0,                     // 110 A----
    0x18E1,                     // 111 A----
    0x18E2,                     // 112 A----
    0x18E3,                     // 113 A----
    0x18E4,                     // 114 A----
    0x18E5,                     // 115 A----
    0x18E6,                     // 116 A----
    0x18E7,                     // 117 A----
    0x18E8,                     // 118 A----
    0x18E9,                     // 119 A----
    0x18EA,                     // 120 A----
    0x18EB,                     // 121 A----
    0x18EC,                     // 122 A----
    0x18ED,                     // 123 A----
    0x18EE,                     // 124 A----
    0x18EF,                     // 125 A----
    0x18F8,                     // 126 A----
    0x18F9,                     // 127 A----
    0x18FA,                     // 128 A----
    0x18FB,                     // 129 A----
    0x1AD0,                     // 130 A----
    0x1AD1,                     // 131 A----
    0x1AD2,                     // 132 A----
    0x1AD3,                     // 133 A----
    0x1AD4,                     // 134 A----
    0x1AD5,                     // 135 A----
    0x1AD6,                     // 136 A----
    0x1AD7,                     // 137 A----
    0x1AD8,                     // 138 A----
    0x1AD9,                     // 139 A----
    0x1ADA,                     // 140 A----
    0x1ADB,                     // 141 A----
    0x1ADC,                     // 142 A----
    0x1ADD,                     // 143 A----
    0x1ADE,                     // 144 A----
    0x1ADF,                     // 145 A----
    0x1AE0,                     // 146 A----
    0x1AE1,                     // 147 A----
    0x1AE2,                     // 148 A----
    0x1AE3,                     // 149 A----
    0x1AE4,                     // 150 A----
    0x1AE5,                     // 151 A----
    0x1AE6,                     // 152 A----
    0x1AE7,                     // 153 A----
    0x1AE8,                     // 154 A----
    0x1AE9,                     // 155 A----
    0x1AEA,                     // 156 A----
    0x1AEB,                     // 157 A----
    0x1AEC,                     // 158 A----
    0x1AED,                     // 159 A----
    0x1AEE,                     // 160 A----
    0x1AEF,                     // 161 A----
    0x1AF0,                     // 162 A----
    0x1AF1,                     // 163 A----
    0x1AF2,                     // 164 A----
    0x1AF3,                     // 165 A----
    0x1AF4,                     // 166 A----
    0x1AF5,                     // 167 A----
    0x1AF6,                     // 168 A----
    0x1AF7,                     // 169 A----
    0x1AF8,                     // 170 A----
    0x1AF9,                     // 171 A----
    0x1AFA,                     // 172 A----
    0x1AFB,                     // 173 A----
    0x1AFC,                     // 174 A----
    0x1AFD,                     // 175 A----
    0x1AFE,                     // 176 A----
    0x1AFF,                     // 177 A----
    PVI_PITCH,                  // 178 ---MG
    PVI_SIZES,                  // 179 ---MG
    PVI_SCORECTRL,              // 180 ---MG
    PVI_SCORELT,                // 181 ---MG
    PVI_SCORERT,                // 182 ---MG
    PVI_BGCOLOUR,               // 183 ---MG
    PVI_BGCOLLIDE,              // 184 ---MG
    PVI_SPRITECOLLIDE,          // 185 ---MG
    PVI_SPR01COLOURS,           // 186 ---MG
    PVI_SPR23COLOURS,           // 187 ---MG
    PVI_P1PADDLE,               // 188 ---MG
    PVI_P2PADDLE,               // 189 ---MG
    0x1700 + PVI_PITCH,         // 190 ----G
    0x1700 + PVI_SIZES,         // 191 ----G
    0x1700 + PVI_SCORECTRL,     // 192 ----G
    0x1700 + PVI_SCORELT,       // 193 ----G
    0x1700 + PVI_SCORERT,       // 194 ----G
    0x1700 + PVI_BGCOLOUR,      // 195 ----G
    0x1700 + PVI_BGCOLLIDE,     // 196 ----G
    0x1700 + PVI_SPRITECOLLIDE, // 197 ----G
    0x1700 + PVI_SPR01COLOURS,  // 198 ----G
    0x1700 + PVI_SPR23COLOURS,  // 199 ----G
    0x1700 + PVI_P1PADDLE,      // 200 ----G
    0x1700 + PVI_P2PADDLE       // 201 ----G
};

// MODULE FUNCTIONS-------------------------------------------------------

MODULE void unautosense(void);
MODULE FLAG autosense(int thesize, FLAG silent);
MODULE void configure(FLAG silent);
MODULE void calcrunningtime(void);
MODULE void help_commands(int whichpage);
MODULE FLAG loadzip(STRPTR passedname, FLAG auditing, FLAG quiet);
MODULE void generate_overlaytext(void);
MODULE void view_mram_pvi(void);
MODULE int load_aof(int localsize);
MODULE int readblock(int localsize);
MODULE FLAG isok(int start, int end);
MODULE UBYTE getdigit(UBYTE data);
MODULE UBYTE makebcc(int start, int end);
MODULE UBYTE readaofbyte(int start);
MODULE UWORD readaofword(int start);
MODULE void serializecos(int version);
MODULE FLAG save_aof(STRPTR filename, int startaddr, int endaddr, int progstart);
MODULE void writeblock(FILE* aofhandle, int startaddr, int endaddr);
MODULE void write_aof_byte(UBYTE* output, UBYTE value);

// CODE-------------------------------------------------------------------

EXPORT void changemachine(int whichmachine, FLAG user)
{   int x, y;

    if (whichgame == 65535)
    {   whichgame = -1; // important!
    }

    if (oldmemmap == MEMMAP_MALZAK2)
    {   m_savenvram();
    }

    if
    (   machine != whichmachine
     || (!user && oldmachine != whichmachine)
    )
    {   machine = whichmachine;
        closemonitors(TRUE);
    } elif
    (   (oldmemmap == MEMMAP_E       && memmap == MEMMAP_F      )
     || (oldmemmap == MEMMAP_F       && memmap == MEMMAP_E      )
     || (oldmemmap == MEMMAP_MALZAK2 && memmap == MEMMAP_MALZAK1) // so that DIP switch editor subwindow gets closed
    )
    {   closemonitors(FALSE);
    }
    if (user)
    {   unautosense();
    }

    macro_stop();
    sound_off(FALSE);
    ff_off(0);
    ff_off(1);

    for (x = 0; x < BOXWIDTH; x++)
    {   for (y = 0; y < BOXHEIGHT; y++)
        {   screen[x][y] = GREY1;
    }   }

    switch (machine)
    {
    case ARCADIA:
        a_setmemmap();
    acase INTERTON:
        i_setmemmap();
    acase ELEKTOR:
        e_setmemmap();
    acase PIPBUG:
        p_setmemmap();
    acase INSTRUCTOR:
        s_setmemmap();
    acase CD2650:
        c_setmemmap();
    acase ASTROWARS:
        astrowars_setmemmap();
        generate_autotext();
        if (user)
        {   changefilepart("ASTROWARS.COS");
        }
    acase GALAXIA:
        galaxia_setmemmap();
        generate_autotext();
        if (user)
        {   changefilepart("GALAXIA.COS");
        }
    acase MALZAK:
        m_setmemmap();
        generate_autotext();
        if (user)
        {   if (memmap == MEMMAP_MALZAK1)
            {   changefilepart("MALZAK1.COS");
            } else
            {   // assert(memmap == MEMMAP_MALZAK2);
                changefilepart("MALZAK2.COS");
        }   }
    acase CHAOS:
        h_setmemmap();
        if (user)
        {   changefilepart("CHAOS2.COS");
        }
    acase DOLPHIN:
        d_setmemmap();
    acase PHUNSY:
        phunsy_setmemmap();
    }
    bps = wps = FALSE;

    if (machine != oldmachine)
    {   region = (ULONG) machines[machine].region;
        wide   = machines[machine].wide;
    }

    sound_on();
    updatepointer(TRUE);
    setselection();

#ifdef WIN32
    if (fullscreen && MainWindowPtr)
    {   clearscreen();
    }
#endif

    switch (machine)
    {
    case ARCADIA:
        for (x = 0; x < A_BOXWIDTH; x++)
        {   for (y = 0; y < A_BOXHEIGHT; y++)
            {   a_screen[x][y] = 15; // NOT anything we actually use!
        }   }
    acase PIPBUG:
        for (x = 0; x < P_BOXWIDTH; x++)
        {   for (y = 0; y < P_BOXHEIGHT; y++)
            {   screen[x][y] = p_bgc; // NOT local_drawpixel()!
            }
            for (y = 260; y <= 261; y++)
            {   screen[x][y] = GREY1; // NOT local_drawpixel()!
        }   }

        for (x = 0; x < P_VDUCOLUMNS; x++)
        {   for (y = 0; y < PH_VDUROWS; y++)
            {   vdu[x][y] = BLANK;
        }   }
        p_digit[0] =
        p_digit[1] =
        p_digit[2] =
        p_digit[3] = 0xF;
        docommand(MENUITEM_DOTSPACE);
    acase INSTRUCTOR:
        for (x = 0; x < S_BOXWIDTH; x++)
        {   for (y = 0; y < S_BOXHEIGHT; y++)
            {   screen[x][y] = BLACK; // NOT local_drawpixel()!
            }
            for (y = 49; y <= 50; y++)
            {   screen[x][y] = GREY1; // NOT local_drawpixel()!
        }   }
        for (x = 0; x < MINIWIDTH; x++)
        {   for (y = 0; y < MINIHEIGHT; y++)
            {   ctrlscreen[x][y] = BLACK;
        }   }
        s_glow = 0;
    acase CD2650:
    case PHUNSY:
        fixupcolours();
        // don't call draw_foo() nor redrawscreen()!
        docommand(MENUITEM_DOTSPACE);
    acase CHAOS:
        for (x = 0; x < H_BOXWIDTH; x++)
        {   for (y = 0; y < H_BOXHEIGHT; y++)
            {   screen[x][y] = p_bgc; // NOT local_drawpixel()!
        }   }

        for (x = 0; x < H_VDUCOLUMNS; x++)
        {   for (y = 0; y < PH_VDUROWS; y++)
            {   vdu[x][y] = BLANK;
        }   }
    acase DOLPHIN:
        for (x = 0; x < D_BOXWIDTH; x++)
        {   for (y = 0; y < D_BOXHEIGHT; y++)
            {   screen[x][y] = BLACK; // NOT local_drawpixel()!
            }
            for (y = 49; y <= 50; y++)
            {   screen[x][y] = GREY1; // NOT local_drawpixel()!
    }   }   }

    if (MainWindowPtr)
    {   resize(size);
        // there are order-dependent!
        redrawscreen();
    }
    if (user)
    {   engine_reset();
    }

    update_memory(TRUE);

    if (GameInfoWindowPtr)
    {   close_gameinfo();
        help_gameinfo();
}   }

EXPORT void view_cram(void)
{   int i,
        where;

    if (memmap != MEMMAP_C && memmap != MEMMAP_D)
    {   return;
    }

    T1; textpen(TEXTPEN_VIEW);

    if (memmap == MEMMAP_C)
    {   where = 0x1000; // 1K ($1000..$13FF)
    } else
    {   // assert(memmap == MEMMAP_D);
        where = 0x1800; // 1K ($1800..$1BFF)
    }
    for (i = where; i <= where + 1023; i += 16)
    {   dumprow(i);
    }
    zprintf("\n");

    T2;
}

EXPORT void view_allmemory(void)
{   int  i, j, k,
         max;
    TEXT asciiversion[16 + 1];

    switch (memmap)
    {
    case MEMMAP_A:
    case MEMMAP_B:
    case MEMMAP_C:
    case MEMMAP_D:
    case MEMMAP_E:
    case MEMMAP_N:
        max = 0x1FFF;
    acase MEMMAP_J:
        max = 0x4FF;
    acase MEMMAP_K:
        max = 0x5FF;
    acase MEMMAP_L:
        max = 0x7FF;
    acase MEMMAP_M:
        max = 0x1FFF;
    acase MEMMAP_MALZAK1:
        max = 0x53FF;
    acase MEMMAP_MALZAK2:
        max = 0x63FF;
    acase MEMMAP_DOLPHIN:
        max = 0xFFF;
    adefault: // eg. MEMMAP_F,G,H,I,O,P,CD2650,ASTROWARS,GALAXIA,CHAOS,PHUNSY
    // perhaps for Arcadia "G" and "H" we should suppress showing
    // $3000..$3FFF, $5000..$5FFF and $7000..$7FFF.
        max = 0x7FFF;
    }

    T1; textpen(TEXTPEN_VIEW);

    for (i = 0; i <= max; i += 16)
    {   dumprow(i);
    }

    if (machine == PHUNSY)
    {   asciiversion[16] = 0;
        for (i = 0; i <= 15; i++)
        {   zprintf("\nBank U%X ($1800..$1FFF) (currently using U%X):\n", i, (banked & 0xF0) >> 4);
            for (j = 0; j < 0x800; j += 16)
            {   for (k = 0; k < 16; k++)
                {   if (u_bank[i][j + k] <= 31) asciiversion[k] = '.'; else asciiversion[k] = u_bank[i][j + k];
                }
                zprintf
                (   "$%04X: %02X %02X %02X %02X %02X %02X %02X %02X  %02X %02X %02X %02X %02X %02X %02X %02X %s\n",
                    0x1800 + j,
                    u_bank[i][j     ],
                    u_bank[i][j +  1],
                    u_bank[i][j +  2],
                    u_bank[i][j +  3],
                    u_bank[i][j +  4],
                    u_bank[i][j +  5],
                    u_bank[i][j +  6],
                    u_bank[i][j +  7],
                    u_bank[i][j +  8],
                    u_bank[i][j +  9],
                    u_bank[i][j + 10],
                    u_bank[i][j + 11],
                    u_bank[i][j + 12],
                    u_bank[i][j + 13],
                    u_bank[i][j + 14],
                    u_bank[i][j + 15],
                    asciiversion
                );
        }   }
        for (i = 0; i <= 15; i++)
        {   zprintf("\nBank Q%X ($4000..$7FFF) (currently using Q%X):\n", i, banked & 0x0F);
            for (j = 0; j < 0x4000; j += 16)
            {   for (k = 0; k < 16; k++)
                {   if (q_bank[i][j + k] <= 31) asciiversion[k] = '.'; else asciiversion[k] = q_bank[i][j + k];
                }
                zprintf
                (   "$%04X: %02X %02X %02X %02X %02X %02X %02X %02X  %02X %02X %02X %02X %02X %02X %02X %02X %s\n",
                    0x4000 + j,
                    q_bank[i][j     ],
                    q_bank[i][j +  1],
                    q_bank[i][j +  2],
                    q_bank[i][j +  3],
                    q_bank[i][j +  4],
                    q_bank[i][j +  5],
                    q_bank[i][j +  6],
                    q_bank[i][j +  7],
                    q_bank[i][j +  8],
                    q_bank[i][j +  9],
                    q_bank[i][j + 10],
                    q_bank[i][j + 11],
                    q_bank[i][j + 12],
                    q_bank[i][j + 13],
                    q_bank[i][j + 14],
                    q_bank[i][j + 15],
                    asciiversion
                );
    }   }   }

    zprintf("\n");
    T2;
}

EXPORT void view_mram(void)
{   int i,
        limit;

    T1; textpen(TEXTPEN_VIEW);

    switch (machine)
    {
    case ARCADIA:
        // Only non-display RAM is shown by this.
        if (memmap == MEMMAP_H)
        {   for (i = 0x1000; i <= 0x10FF; i += 16) dumprow(i);
            for (i = 0x1200; i <= 0x12FF; i += 16) dumprow(i);
            zprintf("\n");
        }

        dumprow(0x18D0);
        dumprow(0x18E0);
        zprintf
        (   "$18F8:                          %02X %02X %02X %02X\n",
            memory[0x18F8],
            memory[0x18F9],
            memory[0x18FA],
            memory[0x18FB]
        );

        dumprow(0x1AD0);
        dumprow(0x1AE0);
        dumprow(0x1AF0);
    acase INTERTON:
    case ELEKTOR:
        if (machine == ELEKTOR)
        {   // monitor RAM
            for (i = 0x800; i <= 0x8BF; i += 16)
            {   dumprow(i);
            }
            zprintf("\n");

            if (memmap == MEMMAP_E)
            {   limit = 0x0FFF;
            } else
            {   // assert(memmap == MEMMAP_F);
                limit = 0x15FF;
            }

            // user RAM
            for (i = 0x8C0; i <= limit; i += 16)
            {   dumprow(i);
            }
            zprintf("\n");

            if (memmap == MEMMAP_F)
            {   // user RAM
                // 8 bytes in this range can be actually obscured RAM, so
                // shouldn't really be shown (if old monitor ROM is in
                // use).
                for (i = 0x1800; i <= 0x1BFF; i += 16)
                {   dumprow(i);
                }
                zprintf("\n");
        }   }
        view_mram_pvi();
    acase PIPBUG:
        if (memmap == MEMMAP_J)
        {   for (i = 0x400; i <=  0x4FF; i += 16) dumprow(i);
        } elif (memmap == MEMMAP_K)
        {   for (i = 0x400; i <=  0x5FF; i += 16) dumprow(i);
        } elif (memmap == MEMMAP_L)
        {   for (i = 0x400; i <=  0x7FF; i += 16) dumprow(i);
        } elif (memmap == MEMMAP_M)
        {   for (i = 0x400; i <= 0x1FFF; i += 16) dumprow(i);
        } else
        {   // assert(memmap == MEMMAP_P);
            for (i = 0x400; i <= 0x7FFF; i += 16) dumprow(i);
        }
    acase INSTRUCTOR:
        // assert(memmap == MEMMAP_J);
        for (i =    0x0; i <=  0x1FF; i += 16) dumprow(i);
        for (i = 0x1780; i <= 0x17FF; i += 16) dumprow(i);
    acase CD2650:
        // assert(memmap == MEMMAP_U);
        for (i =  0x400; i <= 0x17FF; i += 16) dumprow(i);
        for (i = 0x2000; i <= 0x7FFF; i += 16) dumprow(i);
    acase MALZAK:
        for (i = 0x1000; i <= 0x13FF; i += 16) dumprow(i);
        view_mram_pvi();
        pvibase = 0x1500;
        view_mram_pvi();
        pvibase = 0x1400;
        for (i = 0x1600; i <= 0x1FFF; i += 16) dumprow(i);
    acase ASTROWARS:
        for (i = 0x1400; i <= 0x14FF; i += 16) dumprow(i);
        view_mram_pvi();
    acase GALAXIA:
        for (i = 0x1400; i <= 0x14FF; i += 16) dumprow(i);
        view_mram_pvi();
        pvibase = 0x1600;
        view_mram_pvi();
        pvibase = 0x1700;
        view_mram_pvi();
        pvibase = 0x1500;
        for (i = 0x1C00; i <= 0x1FFF; i += 16) dumprow(i);
    acase CHAOS:
        // assert(memmap == MEMMAP_CHAOS);
        for (i =      0; i <= 0x7FFF; i += 16) dumprow(i);
    acase DOLPHIN:
        // assert(memmap == MEMMAP_DOLPHIN);
        for (i =      0; i <=  0xFFF; i += 16) dumprow(i);
    acase PHUNSY:
        // assert(memmap == MEMMAP_PHUNSY);
        // Only non-display, non-banked RAM is shown by this.
        for (i =  0x800; i <=  0xFFF; i += 16) dumprow(i);
        for (i = 0x2000; i <= 0x3FFF; i += 16) dumprow(i);
    }

    zprintf("\n");
    T2;
}

EXPORT void engine_reset(void)
{   int i;

    // We probably don't need to do these, as Arcadia/Interton games
    // (at least) have no right to assume they are in any particular
    // state.

    oldmemmap  = memmap;
    oldmachine = machine;

    switch (machine)
    {
    case ARCADIA:
        flip = 0;
        for (i = 0x1909; i <= 0x190F; i++)
        {   memory[i] = 0xC0; // some of them (randomly) are $40 though
        }
        for (i = 0x19C0; i <= 0x19F7; i++)
        {   memory[i] = 0xFF;
        }
    //lint -fallthrough
    case INTERTON:
        /* This clearing helps avoid these problems:
           (a) Resetting from Dr. Slump game screen to title screen turns
               on constant rumbling (due to sprite imagery being retained
               in RAM).
           (b) For netplay, both machines should be in a consistent state. */
        for (i = 0x1800; i <= 0x1FFF; i++)
        {   memory[i] = 0;
        }
    //lint -fallthrough
    case ASTROWARS:
    case GALAXIA:
    case MALZAK:
        changed = FALSE;

        for (i = 0; i <= 6; i++)
        {   r[i] = 0;
        }
        for (i = 0; i <= 7; i++)
        {   ras[i] = 0;
        }
        psu  =
        psl  = 0;
        iar  = 0;

        // don't touch m_bank!
    acase ELEKTOR:
        changed = FALSE;

        // $8A2/$8A3          = start address of game
        // $8B9/$8BA (INTADR) = address of game interrupt vector
        // $8BE/$8BF (PC)     = start address of game

        /* There are 3 feasible possibilities for starting the monitor:
        iar =     0; // gives black background with white "00 00" at top
        iar =  0x23; // gives blue  background with yellow "IIII" at bottom
        iar = 0x540; // gives blue  background
        */
        if (game)
        {   iar = 0x540;
            memory[0x8A2] = startaddr_h;
            memory[0x8A3] = startaddr_l;
        } else
        {   iar =  0x23; // gives blue  background with yellow "IIII" at bottom
        }

        for (i = 0x1D00; i <= 0x1D1F; i++)
        {   memory[i] = 0; // PSGs are cleared at reset
        }

        playsound(TRUE);
    acase PIPBUG:
    case INSTRUCTOR:
    case CD2650:
    case CHAOS:
    case DOLPHIN:
    case PHUNSY:
        changed = TRUE;

        switch (machine)
        {
        case PIPBUG:
        case CHAOS:
            pbx         = 0;
            freshly     = TRUE;
            queuepos    = 0;
            speakoffset = 0;
        acase CD2650:
            memory[0x17FE] = 0x10; // $1000
            memory[0x17FF] = 0x00;
            memory[0x1000] = 0x5C; // cursor
            freshly     = TRUE;
            queuepos    = 0;
        acase PHUNSY:
            freshly     = TRUE;
            queuepos    = 0;
        }

        if (game)
        {   iar = (startaddr_h * 256) + startaddr_l;
        } else
        {   switch (machine)
            {
            case PIPBUG:
            case CD2650:
            case CHAOS:
            case DOLPHIN:
            case PHUNSY:
                iar = 0;
            acase INSTRUCTOR:
                iar = 0x1800;
        }   }

        for (i = 0; i <= 6; i++)
        {   r[i] = 0;
        }
        for (i = 0; i <= 7; i++)
        {   ras[i] = 0;
        }
        psu =
        psl = 0;
    }

    if (machine == PHUNSY)
    {   memory[0xEE8] = 0x05;
        memory[0xEE9] = 0xFA;
        memory[0xEEA] = 0x05;
        memory[0xEEB] = 0xE6;
        memory[0xEEC] = 0x05;
        memory[0xEED] = 0xE0;
        memory[0xEEE] = 0xFE;
        memory[0xEEF] = 0x20;
        memory[0xEF2] = 0x1F;
        memory[0xEF3] = 0x3F;
        memory[0xEF4] = 0x10;
        memory[0xEF5] = 0x41;
        memory[0xEF7] = banked;
        memory[0xF01] = 0x0D; // CR
        // we call games as subroutines so they can return to the PHUNSY CLI
        // afterwards.
        ras[0]        = 0x23E;
        psu           = 1; // ie. SP = 1;
    }

    ax[0] = ax[1] = ay[0] = ay[1] = STICKCENTRE;

    zinterrupt     = FALSE;
    elapsed        = 0;
    frames         = 0;
    oldframes      = 0;
    overcalc       = 0;
    clearruntos();

#ifdef PROFILE
    for (i = 0; i <= 255; i++)
    {   profile[i] = 0;
    }
#endif

    for (i = 0; i < 32 * KILOBYTE; i++)
    {   memflags[i] &= ~(COVERAGE_BITS);
    }

    /* We should really clear all memory (important for netplay, to */
    /* keep states at both sides consistent). */
}

EXPORT void savescreenshot(int kind, FLAG wantasl, FLAG clip)
{   TRANSIENT UBYTE*  TheBody;
    TRANSIENT FILE*   TheLocalHandle = NULL; // initialized to avoid spurious SAS/C compiler warnings
// LocalHandle is a variable of winbase.h
    TRANSIENT int     bpl, // different variable from global bpl
                      count,
                      iffsize,
                      i,
                      x, y;
    TRANSIENT UBYTE   last,
                      t;
    TRANSIENT TEXT    annochunk[8];
    PERSIST   ASCREEN tempscreen[G_BOXWIDTH][G_BOXHEIGHT];

PERSIST UBYTE iff_header[104 + 48] =
{   0x46, 0x4F, 0x52, 0x4D, 0xFF, 0xFF, 0xFF, 0xFF,   // "FORMxxxx"
    0x49, 0x4C, 0x42, 0x4D, 0x42, 0x4D, 0x48, 0x44,   // "ILBMBMHD"
    0x00, 0x00, 0x00, 0x14, 0xFF, 0xFF, 0xFF, 0xFF,   // "xxxx1234"
    0x00, 0x00, 0x00, 0x00, 0x05, 0x00, 0x00, 0x00,   // "56789012"
    0x00, 0x00, 0x02, 0x01, 0xFF, 0xFF, 0xFF, 0xFF,   // "34567890"
    0x43, 0x4D, 0x41, 0x50, 0x00, 0x00, 0x00, 0x60,   // "CMAPxxxx"
    // R     G     B
    0x00, 0x00, 0x00, // colour  0
    0x00, 0x00, 0x00, // colour  1
    0x00, 0x00, 0x00, // colour  2
    0x00, 0x00, 0x00, // colour  3
    0x00, 0x00, 0x00, // colour  4
    0x00, 0x00, 0x00, // colour  5
    0x00, 0x00, 0x00, // colour  6
    0x00, 0x00, 0x00, // colour  7
    0x00, 0x00, 0x00, // colour  8
    0x00, 0x00, 0x00, // colour  9
    0x00, 0x00, 0x00, // colour 10
    0x00, 0x00, 0x00, // colour 11
    0x00, 0x00, 0x00, // colour 12
    0x00, 0x00, 0x00, // colour 13
    0x00, 0x00, 0x00, // colour 14
    0x00, 0x00, 0x00, // colour 15
    0x00, 0x00, 0x00, // colour 16
    0x00, 0x00, 0x00, // colour 17
    0x00, 0x00, 0x00, // colour 18
    0x00, 0x00, 0x00, // colour 19
    0x00, 0x00, 0x00, // colour 20
    0x00, 0x00, 0x00, // colour 21
    0x00, 0x00, 0x00, // colour 22
    0x00, 0x00, 0x00, // colour 23
    0x00, 0x00, 0x00, // colour 24
    0x00, 0x00, 0x00, // colour 25
    0x00, 0x00, 0x00, // colour 26
    0x00, 0x00, 0x00, // colour 27
    0x00, 0x00, 0x00, // colour 28
    0x00, 0x00, 0x00, // colour 29
    0x00, 0x00, 0x00, // colour 30
    0x00, 0x00, 0x00, // colour 31
    0x42, 0x4F, 0x44, 0x59, 0x00, 0xFF, 0xFF, 0xFF  // "BODYxxxx"
}, bmp_header[150] =
{   // bytes  0..1  are "BM"
    // bytes  2..5  is the size of the entire file (little-endian)
    // bytes  6..9  are reserved
    // bytes 10..13 is the offset to the bits (little-endian)
    // byte  14..17 is a constant (structure size) (little-endian)
    // bytes 18..21 is the width of the bitmap (little-endian)
    // bytes 22..25 is the height of the bitmap (little-endian)
    // bytes 26..27 is a constant
    // bytes 28..29 is the bits per pixel
    // bytes 30..33 is the compression
    // bytes 34..37 is the size (zero is allowed)
    // bytes 38..41 is X-pixels per metre
    // bytes 42..45 is Y-pixels per metre
    // bytes 46..49 is the number of colours used (ie. 24)
    // bytes 50..53 is the number of important colours (ie. 24)
    0x42, 0x4D, 0x62, 0x61, 0x00, 0x00, 0x00, 0x00, // "BMzzzz00"
    0x00, 0x00, 0xA6, 0x00, 0x00, 0x00, 0x28, 0x00, // "00yyyy*0"
    0x00, 0x00, 0xFF, 0xFF, 0x00, 0x00, 0xFF, 0xFF, // "00wwwwhh"
    0x00, 0x00, 0x01, 0x00, 0x08, 0x00, 0x00, 0x00, // "hh1080cc"
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x13, 0x0B, // "ccssssxx"
    0x00, 0x00, 0x13, 0x0B, 0x00, 0x00, 0x18, 0x00, // "xxyyyycc"
    0x00, 0x00, 0x10, 0x00, 0x00, 0x00,             // "cciiii"
    // B     G     R
    0x00, 0x00, 0x00, 0x00, // colour  0
    0x00, 0x00, 0x00, 0x00, // colour  1
    0x00, 0x00, 0x00, 0x00, // colour  2
    0x00, 0x00, 0x00, 0x00, // colour  3
    0x00, 0x00, 0x00, 0x00, // colour  4
    0x00, 0x00, 0x00, 0x00, // colour  5
    0x00, 0x00, 0x00, 0x00, // colour  6
    0x00, 0x00, 0x00, 0x00, // colour  7
    0x00, 0x00, 0x00, 0x00, // colour  8
    0x00, 0x00, 0x00, 0x00, // colour  9
    0x00, 0x00, 0x00, 0x00, // colour 10
    0x00, 0x00, 0x00, 0x00, // colour 11
    0x00, 0x00, 0x00, 0x00, // colour 12
    0x00, 0x00, 0x00, 0x00, // colour 13
    0x00, 0x00, 0x00, 0x00, // colour 14
    0x00, 0x00, 0x00, 0x00, // colour 15
    0x00, 0x00, 0x00, 0x00, // colour 16
    0x00, 0x00, 0x00, 0x00, // colour 17
    0x00, 0x00, 0x00, 0x00, // colour 18
    0x00, 0x00, 0x00, 0x00, // colour 19
    0x00, 0x00, 0x00, 0x00, // colour 20
    0x00, 0x00, 0x00, 0x00, // colour 21
    0x00, 0x00, 0x00, 0x00, // colour 22
    0x00, 0x00, 0x00, 0x00, // colour 23
}, pcx_header[129] =
{   0x0A,                    // manufacturer
    0x05,                    // version
    0x01,                    // encoding
    0x08,                    // bits per pixel
       0,    0,              // X-min
       0,    0,              // Y-min
     255,  255,              // X-max (done later)
     255,  255,              // Y-max (done later)
     255,  255,              // X-res (done later)
     255,  255,              // Y-res (done later)
    // R     G     B
    0x00, 0x00, 0x00,
    0x00, 0x00, 0x00,
    0x00, 0x00, 0x00,
    0x00, 0x00, 0x00,
    0x00, 0x00, 0x00,
    0x00, 0x00, 0x00,
    0x00, 0x00, 0x00,
    0x00, 0x00, 0x00,
    0x00, 0x00, 0x00,
    0x00, 0x00, 0x00,
    0x00, 0x00, 0x00,
    0x00, 0x00, 0x00,
    0x00, 0x00, 0x00,
    0x00, 0x00, 0x00,
    0x00, 0x00, 0x00,
    0x00, 0x00, 0x00,
       0,                   // reserved
       1,                   // planes
    0xFF, 0xFF,             // bytes per line per plane (ie. width)
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00,             // 58 zero bytes
      12                    // 256-colour palette is present
};

    // assert(GUESTCOLOURS == 24);

    if (wantasl)
    {   if (game)
        {   strcpy(fn_screenshot, filepart);
        } else
        {   fn_screenshot[0] = 0;
        }
        fixextension(filekind[kind].extension, fn_screenshot, TRUE); // so we have extension in ASL requester

        if (asl
        (   filekind[kind].hail,
            filekind[kind].pattern,
            TRUE,
            DIR_SCREENSHOTS,
            fn_screenshot, // this fortunately doesn't contain the path right now
            fn_screenshot
        ) != EXIT_SUCCESS)
        {   return;
        }
        fixextension(filekind[kind].extension, fn_screenshot, TRUE); // so we have extension even if user overtypes it in ASL requester
    }

    /* Possible enhancements:
     (a) could use compression (for IFF and BMP, because we already
         support this for PCX).
     (b) could optionally save them in "wide" (doubled horizontally)
         format.
     (c) more detailed/specific error messages.
     (d) check ALL possible errors (eg. calloc(), Close(), etc.).
     (e) perhaps omit left/right/bottom "overscan" margins (eg. from
         Arcadia screenshots).
     (f) Most files could be saved using less than 5 planes.

    ACBM files *do* have pad bits at the end of each row (based on Personal
    Paint 6.4 behaviour). */

    if (machine == ASTROWARS || machine == GALAXIA)
    {   // assert(G_BOXWIDTH == G_BOXHEIGHT);
        for (x = 0; x < G_BOXWIDTH; x++)
        {   for (y = 0; y < G_BOXHEIGHT; y++)
            {   tempscreen[x][y] = screen[x][y];
        }   }
        for (x = 0; x < G_BOXWIDTH; x++)
        {   for (y = 0; y < G_BOXHEIGHT; y++)
            {   screen[x][y] = tempscreen[y][G_BOXWIDTH - 1 - x];
    }   }   }

    if (clip)
    {   clip_open(TRUE);
    } else
    {   if (!(TheLocalHandle = fopen(fn_screenshot, "wb")))
        {   beep();
            return;
    }   }

    switch (kind)
    {
    case KIND_PCX:
        pcx_header[ 8] = (UBYTE) ((machines[machine].width  - 1) % 256);
        pcx_header[ 9] = (UBYTE) ((machines[machine].width  - 1) / 256);
        pcx_header[10] = (UBYTE) ((machines[machine].height - 1) % 256);
        pcx_header[11] = (UBYTE) ((machines[machine].height - 1) / 256);
        pcx_header[12] =
        pcx_header[66] = (UBYTE) ( machines[machine].width       % 256);
        pcx_header[13] =
        pcx_header[67] = (UBYTE) ( machines[machine].width       / 256);
        pcx_header[14] = (UBYTE) ( machines[machine].height      % 256);
        pcx_header[15] = (UBYTE) ( machines[machine].height      / 256);

        for (i = 0; i < 16; i++)
        {
#ifdef AMIGA
            pcx_header[16 + (i * 3)] = (UBYTE) ((pencolours[colourset][i].red   & 0xFF000000) >> 24);
            pcx_header[17 + (i * 3)] = (UBYTE) ((pencolours[colourset][i].green & 0xFF000000) >> 24);
            pcx_header[18 + (i * 3)] = (UBYTE) ((pencolours[colourset][i].blue  & 0xFF000000) >> 24);
#endif
#ifdef WIN32
            pcx_header[16 + (i * 3)] = (UBYTE) ((pens[colourset][i] & 0x00FF0000) >> 16);
            pcx_header[17 + (i * 3)] = (UBYTE) ((pens[colourset][i] & 0x0000FF00) >>  8);
            pcx_header[18 + (i * 3)] = (UBYTE)  (pens[colourset][i] & 0x000000FF)       ;
#endif
        }

#ifdef _lint
        if (!TheLocalHandle)
        {   return;
        }
#endif
        DISCARD fwrite(pcx_header, (size_t) 129, 1, TheLocalHandle);

        TheBody = calloc((unsigned int) (machines[machine].width * machines[machine].height), 1); // should check return code!
#ifdef _lint
        if (!TheBody)
        {   return;
        }
#endif

        i = 0;
        for (y = 0; y < machines[machine].height; y++)
        {   count = 1;
            last = (UBYTE) screen[0][y];

            /* Encode one line */
            for (x = 1; x < machines[machine].width; x++)
            {   t = (UBYTE) screen[x][y];

                if (t == last)
                {   count++;
                    if (count == 63)
                    {   TheBody[i++] = (UBYTE) (count | 0xC0);
                        TheBody[i++] = last;
                        count = 0;
                }   }
                else
                {   if (count)
                    {   if ((count == 1) && ((last & 0xC0) != 0xC0))
                        {   TheBody[i++] = last;
                        } else
                        {   TheBody[i++] = (UBYTE) (count | 0xC0);
                            TheBody[i++] = last;
                    }   }
                    last = t;
                    count = 1;
            }   }

            if (count)
            {   if ((count == 1) && ((last & 0xC0) != 0xC0))
                {   TheBody[i++] = last;
                } else
                {   TheBody[i++] = (UBYTE) (count | 0xC0);
                    TheBody[i++] = last;
        }   }   }

        DISCARD fwrite(TheBody, (size_t) i, 1, TheLocalHandle);
        free(TheBody);

        TheBody = calloc(768, 1);
#ifdef _lint
        if (!TheBody)
        {   return;
        }
#endif
        for (i = 0; i < 24; i++)
        {
#ifdef AMIGA
            TheBody[     i * 3 ] = (UBYTE) ((pencolours[colourset][i].red   & 0xFF000000) >> 24);
            TheBody[1 + (i * 3)] = (UBYTE) ((pencolours[colourset][i].green & 0xFF000000) >> 24);
            TheBody[2 + (i * 3)] = (UBYTE) ((pencolours[colourset][i].blue  & 0xFF000000) >> 24);
#endif
#ifdef WIN32
            TheBody[     i * 3 ] = (UBYTE) ((pens[colourset][i] & 0x00FF0000) >> 16);
            TheBody[1 + (i * 3)] = (UBYTE) ((pens[colourset][i] & 0x0000FF00) >>  8);
            TheBody[2 + (i * 3)] = (UBYTE)  (pens[colourset][i] & 0x000000FF)       ;
#endif
        }
        DISCARD fwrite(TheBody, (size_t) 768, 1, TheLocalHandle);
    acase KIND_BMP:
        bmp_header[18] = (UBYTE) (machines[machine].width  % 256);
        bmp_header[19] = (UBYTE) (machines[machine].width  / 256);
        bmp_header[22] = (UBYTE) (machines[machine].height % 256);
        bmp_header[23] = (UBYTE) (machines[machine].height / 256);

        for (i = 0; i < 24; i++)
        {
#ifdef AMIGA
            bmp_header[54 + (i * 4)] = (UBYTE) ((pencolours[colourset][i].blue  & 0xFF000000) >> 24);
            bmp_header[55 + (i * 4)] = (UBYTE) ((pencolours[colourset][i].green & 0xFF000000) >> 24);
            bmp_header[56 + (i * 4)] = (UBYTE) ((pencolours[colourset][i].red   & 0xFF000000) >> 24);
#endif
#ifdef WIN32
            bmp_header[54 + (i * 4)] = (UBYTE)  (pens[colourset][i] & 0x000000FF)       ;
            bmp_header[55 + (i * 4)] = (UBYTE) ((pens[colourset][i] & 0x0000FF00) >>  8);
            bmp_header[56 + (i * 4)] = (UBYTE) ((pens[colourset][i] & 0x00FF0000) >> 16);
#endif
        }

        TheBody = calloc((unsigned int) (150 + (machines[machine].width * machines[machine].height)), 1); // should check return code!
#ifdef _lint
        if (!TheBody)
        {   return;
        }
#endif
        for (i = 0; i < 150; i++)
        {   TheBody[i] = bmp_header[i];
        }
        i = 150;
        for (y = machines[machine].height - 1; y >= 0; y--)
        {   for (x = 0; x < machines[machine].width; x++)
            {   TheBody[i++] = (UBYTE) screen[x][y];
        }   }

        if (clip)
        {   clip_write((STRPTR) &TheBody[14], 150 - 14 + (machines[machine].width * machines[machine].height), CF_DIB);
        } else
        {
#ifdef _lint
            if (!TheBody)
            {   return;
            }
            if (!TheLocalHandle)
            {   // assert(TheBody);
                free(TheBody);
                return;
            }
#endif
            DISCARD fwrite(TheBody, (size_t) (150 + (machines[machine].width * machines[machine].height)), 1, TheLocalHandle);
        }
    acase KIND_ILBM:
    case KIND_ACBM:
        if (kind == KIND_ILBM)
        {   iff_header[ 8] = 'I';
            iff_header[ 9] = 'L';
         // iff_header[10] = 'B';
         // iff_header[11] = 'M';
            iff_header[96] = 'B';
            iff_header[97] = 'O';
            iff_header[98] = 'D';
            iff_header[99] = 'Y';
        } else
        {   // assert(kind == KIND_ACBM);
            iff_header[ 8] = 'A';
            iff_header[ 9] = 'C';
         // iff_header[10] = 'B';
         // iff_header[11] = 'M';
            iff_header[96] = 'A';
            iff_header[97] = 'B';
            iff_header[98] = 'I';
            iff_header[99] = 'T';
        }

        iff_header[20] =
        iff_header[36] = (UBYTE) (machines[machine].width  / 256);
        iff_header[21] =
        iff_header[37] = (UBYTE) (machines[machine].width  % 256);
        iff_header[22] =
        iff_header[38] = (UBYTE) (machines[machine].height / 256);
        iff_header[23] =
        iff_header[39] = (UBYTE) (machines[machine].height % 256);

        bpl = machines[machine].width / 8;
        if (machines[machine].width % 8)
        {   bpl++;
        }
        if (bpl % 2)
        {   bpl++;
        }
        iffsize = 104 - 8 + (bpl * 5 * machines[machine].height); // for 5 bitplanes
        if (!clip)
        {   iffsize += 8 + (int) strlen("Generated by ") + (int) strlen(hostmachines[machine].titlebartext) + 1;
            if (iffsize & 1)
            {   iffsize++;
        }   }
        iff_header[  5] = (UBYTE) ( iffsize / 65536       );
        iff_header[  6] = (UBYTE) ((iffsize % 65536) / 256);
        iff_header[  7] = (UBYTE) ( iffsize % 256         );
        iffsize = bpl * 5 * machines[machine].height; // for 5 bitplanes
        iff_header[100 + 48] = (UBYTE) ( iffsize / 16777216         );
        iff_header[101 + 48] = (UBYTE) ((iffsize % 16777216) / 65536);
        iff_header[102 + 48] = (UBYTE) ((iffsize % 65536)    /   256);
        iff_header[103 + 48] = (UBYTE) ( iffsize % 256              );

        for (i = 0; i < 24; i++)
        {
#ifdef AMIGA
            iff_header[48 + (i * 3)] = (UBYTE) ((pencolours[colourset][i].red   & 0xFF000000) >> 24);
            iff_header[49 + (i * 3)] = (UBYTE) ((pencolours[colourset][i].green & 0xFF000000) >> 24);
            iff_header[50 + (i * 3)] = (UBYTE) ((pencolours[colourset][i].blue  & 0xFF000000) >> 24);
#endif
#ifdef WIN32
            iff_header[48 + (i * 3)] = (UBYTE) ((pens[colourset][i] & 0x00FF0000) >> 16);
            iff_header[49 + (i * 3)] = (UBYTE) ((pens[colourset][i] & 0x0000FF00) >>  8);
            iff_header[50 + (i * 3)] = (UBYTE)  (pens[colourset][i] & 0x000000FF)       ;
#endif
        }
        if (clip)
        {   clip_write((STRPTR) iff_header, 104 + 48, 0);
        } else
        {
#ifdef _lint
            if (!TheLocalHandle)
            {   return;
            }
#endif
            DISCARD fwrite(iff_header, (size_t) 104 + 48, 1, TheLocalHandle);
        }

        TheBody = calloc((size_t) (5 * bpl * machines[machine].height), 1); // for 5 bitplanes
        // should check return code!
#ifdef _lint
        if (!TheBody)
        {   return;
        }
#endif
        if (kind == KIND_ILBM)
        {   for (y = 0; y < machines[machine].height; y++)
            {   for (i = 0; i < 5; i++) // for 5 bitplanes
                {   for (x = 0; x < machines[machine].width; x++)
                    {   if (screen[x][y] & (1 << i))
                        {   TheBody[(y * 5 * bpl) // for 5 bitplanes
                                   +    (i * bpl)
                                   +    (x /  8)] |= 0x80 >> (x % 8);
        }   }   }   }   }
        else
        {   // assert(kind == KIND_ACBM);
            for (i = 0; i < 5; i++) // for 5 bitplanes
            {   for (y = 0; y < machines[machine].height; y++)
                {   for (x = 0; x < machines[machine].width; x++)
                    {   if (screen[x][y] & (1 << i))
                        {   TheBody[(i * machines[machine].height * bpl)
                                  + (y * bpl)
                                  + (x / 8)] |= 0x80 >> (x % 8);
        }   }   }   }   }

        if (clip)
        {   clip_write((STRPTR) TheBody, bpl * 5 * machines[machine].height, 0); // for 5 bitplanes
        } else
        {
#ifdef _lint
            if (!TheBody)
            {   return;
            }
            if (!TheLocalHandle)
            {   // assert(TheBody);
                free(TheBody);
                return;
            }
#endif
            DISCARD fwrite(TheBody, (size_t) bpl * 5 * machines[machine].height, 1, TheLocalHandle); // for 5 bitplanes

            annochunk[0] = 'A';
            annochunk[1] =
            annochunk[2] = 'N';
            annochunk[3] = 'O';
            annochunk[4] =
            annochunk[5] =
            annochunk[6] = 0;
            annochunk[7] = (UBYTE) (strlen("Generated by ") + strlen(hostmachines[machine].titlebartext) + 1);
            DISCARD fwrite(annochunk                     ,                       8, 1, TheLocalHandle);
            DISCARD fwrite("Generated by "               , strlen("Generated by "), 1, TheLocalHandle);
            DISCARD fwrite(hostmachines[machine].titlebartext, strlen(hostmachines[machine].titlebartext) + 1, 1, TheLocalHandle); // we deliberately include the NULL-terminator
            // ANNOtation strings *are* NULL-terminated (based on behaviour of OctaMED Pro 5.0 and SoundBox 2.9 beta)
            if (annochunk[7] & 1)
            {   DISCARD fwrite("\0", 1, 1, TheLocalHandle); // pad byte
        }   }
    adefault:
        // assert(0);
        TheBody = NULL; // to avoid spurious SAS/C optimizer warnings
    }

    free(TheBody);
    // TheBody = NULL;
    if (clip)
    {   clip_close();
    } else
    {
#ifdef _lint
        if (!TheLocalHandle)
        {   return;
        }
#endif
        DISCARD fclose(TheLocalHandle);
        // TheLocalHandle = NULL;
    }

    if (machine == ASTROWARS || machine == GALAXIA)
    {   // assert(G_BOXWIDTH == G_BOXHEIGHT);
        for (x = 0; x < G_BOXWIDTH; x++)
        {   for (y = 0; y < G_BOXHEIGHT; y++)
            {   screen[x][y] = tempscreen[x][y];
    }   }   }

#ifdef AMIGA
    writeicon(kind, fn_screenshot);
#endif
}

EXPORT FLAG engine_load(FLAG silent)
{   FILE* TheLocalHandle /* = NULL */ ; // LocalHandle is a variable of winbase.h

    if (IsZip(fn_game))
    {   if (!loadzip(fn_game, FALSE, silent))
        {   return FALSE;
        }
        filesize = ZIPBufferSize;
        IOBuffer = (UBYTE *) ZIPBufferPtr;
    } else
    {   filesize = (int) getsize(fn_game);
        alloc_iobuffer((ULONG) filesize);

        if (!(TheLocalHandle = fopen((char *) fn_game, "rb"))) // just cast for lint
        {
#ifdef VERBOSE
            T1; textpen(TEXTPEN_VERBOSE); zprintf("Can't open %s for reading!\n", fn_game); T2;
#endif
            return FALSE;
        }
        if (fread(IOBuffer, (size_t) filesize, 1, TheLocalHandle) != 1)
        {   fclose(TheLocalHandle);
            // TheLocalHandle = NULL;
            return FALSE;
        }
        fclose(TheLocalHandle);
        // TheLocalHandle = NULL;
    }

    breakparts();

    if (!parse_bytes(TRUE, silent))
    {   return FALSE;
    }

    // updatemenus(); is already done by parse_bytes()
    // updatesmlgads(); is already done by parse_bytes()

    updatebiggads();
    updatedips();
    settitle();
    redrawscreen();

    return TRUE;
}

// we should localize the error messages in this routine
EXPORT FLAG parse_bytes(FLAG loading, FLAG silent)
{   TRANSIENT UBYTE kind;
    TRANSIENT UWORD loadaddress;
    TRANSIENT int   banktouse,
                    bankoffset,
                    datastart = 0, // initialized to avoid spurious SAS/C compiler warnings
                    i,
                    oldromsize,
                    tempsize,
                    version = 0; // initialized to avoid spurious SAS/C compiler warnings
    TRANSIENT TEXT  now;
    TRANSIENT ULONG crc;
    PERSIST   FLAG  found   = FALSE;

    // assert(!crippled);
    // assert(!inframe);

    oldmemmap  = memmap;
    oldmachine = machine;
    oldromsize = romsize;

    serializemode = SERIALIZE_READ;

/* TVC files are:
        0:  magic byte    (always $02)
     1..2:  load address  (where to load program to)
     3..4:  start address (where to point IAR to)
        5+: the program itself

    Magic bytes are:
    $00 = LODZ r0      = Elektor PGM
    $02 = LODZ r2      = TVC
    $3A = BSTR,lt      = AOF
    $40 = HALT         = CD2650 PGM
    $BF = BSXA         = COS/COR
    $C4 = ------------ = PIPBUG PGM
    $C5 = ------------ = Instructor PGM
    $C6 then $01 = --- = PHUNSY PGM
    all others   =       Arcadia/Interton/Dolphin/PHUNSY BIN */

    switch (IOBuffer[0])
    {
    case 0x00:
        // Some valid known PHUNSY BINs begin with $00 and thus can be misidentified as Elektor PGMs.
        // That is the purpose of this code.
#ifdef VERBOSE
        T1; textpen(TEXTPEN_VERBOSE); zprintf("%s might be a TVC.\n", fn_game); T2;
#endif
        offset = 0;
        crc = crc32(IOBuffer, filesize);
        for (i = 0; i < KNOWNGAMES; i++)
        {   if (filesize == known[i].size && crc == known[i].crc32 && known[i].machine != ELEKTOR)
            {   offset = 0;
                goto HERE;
        }   }

#ifdef VERBOSE
        T1; textpen(TEXTPEN_VERBOSE); zprintf("Loading %s as Elektor PGM.\n", fn_game); T2;
#endif
        kind = KIND_PGM;
        machine = ELEKTOR;
        startaddr_h = ReadByteAt(1);
        startaddr_l = ReadByteAt(2);
        romsize = filesize;
        offset = 0; // for autosense()
    acase 0x02:
        // Some valid known PHUNSY BINs begin with $02 and thus can be misidentified as TVCs.
        // That is the purpose of this code.
#ifdef VERBOSE
        T1; textpen(TEXTPEN_VERBOSE); zprintf("%s might be a TVC.\n", fn_game); T2;
#endif
        offset = 0;
        crc = crc32(IOBuffer, filesize);
        for (i = 0; i < KNOWNGAMES; i++)
        {   if (filesize == known[i].size && crc == known[i].crc32 && known[i].machine != ELEKTOR)
            {   offset = 0;
                goto HERE;
        }   }

#ifdef VERBOSE
        T1; textpen(TEXTPEN_VERBOSE); zprintf("Loading %s as TVC.\n", fn_game); T2;
#endif
        kind = KIND_TVC;
        machine = ELEKTOR;
        startaddr_h = ReadByteAt(3);
        startaddr_l = ReadByteAt(4);
        romsize = filesize - 5 + (startaddr_h * 256) + startaddr_l;
        offset = 0; // for autosense()
    acase 0x3A: // BSTR,lt
#ifdef VERBOSE
        T1; textpen(TEXTPEN_VERBOSE); zprintf("Loading %s as AOF.\n", fn_game); T2;
#endif
        if
        (   machine != ELEKTOR
         && machine != PIPBUG
         && machine != INSTRUCTOR
         && machine != CD2650
         && machine != CHAOS
         && machine != DOLPHIN
         && machine != PHUNSY
        ) // eg. ARCADIA, INTERTON, ASTROWARS, GALAXIA, MALZAK
        {   say("Wrong guest machine!", FALSE);
            return FALSE;
        }
        kind = KIND_AOF;
        if (load_aof(filesize) == 0)
        {   say("Corrupt AOF file!", FALSE);
            return FALSE;
        }
        romsize = 32768;
        offset = 0; // probably not needed
    acase 0x40: // HALT
#ifdef VERBOSE
        T1; textpen(TEXTPEN_VERBOSE); zprintf("Loading %s as CD2650 PGM.\n", fn_game); T2;
#endif
        if (filesize > 32 * KILOBYTE)
        {   say((STRPTR) GetCatalogStr(CatalogPtr, MSG_ENGINE_TOOLARGE, "File is too large or corrupt!"), FALSE);
            return FALSE;
        }
        kind = KIND_PGM;
        machine = CD2650;
        startaddr_h = ReadByteAt(1);
        startaddr_l = ReadByteAt(2);
        romsize = filesize;
        offset = 0; // for autosense()
    acase 0xBF:
        version = ReadByteAt(1);
        switch (ReadByteAt(2))
        {
        case 0:
            kind = KIND_COS;
#ifdef VERBOSE
            T1; textpen(TEXTPEN_VERBOSE); zprintf("Loading %s as COS.\n", fn_game); T2;
#endif
        acase 1:
            kind = KIND_COR;
#ifdef VERBOSE
            T1; textpen(TEXTPEN_VERBOSE); zprintf("Loading %s as COR.\n", fn_game); T2;
#endif
        adefault:
            say("Corrupt COS/COR file!", FALSE);
            return FALSE;
        }
        switch (ReadByteAt(3))
        {
        case 0:
            machine   = MALZAK;
            memmap    = MEMMAP_MALZAK1;
        acase 1:
            machine   = MALZAK;
            memmap    = MEMMAP_MALZAK2;
        acase 2:
            machine   = GALAXIA;
            memmap    = MEMMAP_GALAXIA;
        acase 3:
            machine   = ASTROWARS;
            memmap    = MEMMAP_ASTROWARS;
        acase 4:
            machine   = CHAOS;
            memmap    = MEMMAP_CHAOS;
        acase 5:
            machine   = DOLPHIN;
            memmap    = MEMMAP_DOLPHIN;
        acase 6:
            machine   = ARCADIA;
        acase 7:
            machine   = INTERTON;
        acase 8:
            machine   = ELEKTOR;
        acase 9:
            machine   = PIPBUG;
        acase 10:
            machine   = INSTRUCTOR;
        acase 11:
            machine   = CD2650;
        acase 12:
            machine   = PHUNSY;
        adefault:
            say("Unsupported COS/COR version!", FALSE);
            return FALSE;
        }
        if
        (   (machine == MALZAK     &&  version != 1)
         || (machine == ASTROWARS  && (version <  1 || version > 2))
         || (machine == GALAXIA    && (version <  1 || version > 2))
         || (machine == CHAOS      && (version <  1 || version > 2))
         || (machine == DOLPHIN    && (version <  1 || version > 2))
         || (machine == ARCADIA    &&  version != 1)
         || (machine == INTERTON   &&  version != 1)
         || (machine == ELEKTOR    &&  version != 1)
         || (machine == PIPBUG     &&  version != 1)
         || (machine == INSTRUCTOR &&  version != 1)
         || (machine == CD2650     &&  version != 1)
         || (machine == PHUNSY     && (version <  1 || version > 3))
        )
        {   say("Unsupported COS/COR version!", FALSE);
            return FALSE;
        }
        romsize = 0;
        // startaddr_h|l aren't used for coin-ops
        // offset doesn't need to be set here
        if (loading) // important!
        {   changemachine(machine, FALSE);
        }
    acase 0xC4:
#ifdef VERBOSE
        T1; textpen(TEXTPEN_VERBOSE); zprintf("Loading %s as PIPBUG PGM.\n", fn_game); T2;
#endif
        if (filesize > 32 * KILOBYTE)
        {   say((STRPTR) GetCatalogStr(CatalogPtr, MSG_ENGINE_TOOLARGE, "File is too large or corrupt!"), FALSE);
            return FALSE;
        }
        kind = KIND_PGM;
        machine = PIPBUG;
        startaddr_h = ReadByteAt(1);
        startaddr_l = ReadByteAt(2);
        romsize = filesize;
        offset = 0; // for autosense()
    acase 0xC5:
#ifdef VERBOSE
        T1; textpen(TEXTPEN_VERBOSE); zprintf("Loading %s as Instructor PGM.\n", fn_game); T2;
#endif
        kind = KIND_PGM;
        machine = INSTRUCTOR;
        startaddr_h = ReadByteAt(1);
        startaddr_l = ReadByteAt(2);
        romsize = filesize;
        offset = 0; // for autosense()
    acase 0xC6:
#ifdef VERBOSE
        T1; textpen(TEXTPEN_VERBOSE); zprintf("Loading %s as PHUNSY PGM.\n", fn_game); T2;
#endif
        if (filesize > 32 * KILOBYTE || ReadByteAt(1) != 1)
        {   say((STRPTR) GetCatalogStr(CatalogPtr, MSG_ENGINE_TOOLARGE, "File is too large or corrupt!"), FALSE);
            return FALSE;
        }
        kind = KIND_PGM;
        machine = PHUNSY;
        startaddr_h = ReadByteAt(2);
        startaddr_l = ReadByteAt(3);
        romsize = filesize;
        offset = 0; // for autosense()
    adefault:
HERE:
#ifdef VERBOSE
        T1; textpen(TEXTPEN_VERBOSE); zprintf("Loading %s as BIN.\n", fn_game); T2;
#endif
        kind = KIND_BIN;
        romsize = filesize;
        if (romsize > 32 * KILOBYTE)
        {   romsize = oldromsize;
            say((STRPTR) GetCatalogStr(CatalogPtr, MSG_ENGINE_TOOLARGE, "File is too large or corrupt!"), FALSE);
            return FALSE;
        }
        offset = 0; // for autosense()
    }

    if (romsize > 32 * KILOBYTE)
    {   romsize = oldromsize;
        machine = oldmachine;
        say((STRPTR) GetCatalogStr(CatalogPtr, MSG_ENGINE_TOOLARGE, "File is too large or corrupt!"), FALSE);
        return FALSE;
    }

    pgmtext[0] = 0;

    if (loading)
    {   unautosense();
    }
    game = TRUE;

    if (loading && kind != KIND_AOF)
    {   if (sensegame)
        {   switch (kind)
            {
            case KIND_BIN:
            case KIND_PGM:
                found = autosense(romsize, silent);
            acase KIND_TVC:
                found = autosense(filesize, silent);
            acase KIND_COS:
            case KIND_COR:
                if
                (   machine == ARCADIA
                 || machine == INTERTON
                 || machine == ELEKTOR
                 || machine == PIPBUG
                 || machine == INSTRUCTOR
                 || machine == CD2650
                 || machine == PHUNSY
                )
                {   if (machine == PHUNSY && version == 1)
                    {   // found = FALSE;
                        whichgame = -1;
                    } else
                    {   if (machine == PHUNSY)
                        {   // assert(version >= 2);
                            whichgame = (int) ReadWordAt(284706); // $45822..$45823
                        } else
                        {   whichgame = (int) ReadWordAt(33);
                        }
                        if (whichgame >= 0 && whichgame < KNOWNGAMES)
                        {   found = TRUE;
                            configure(silent);
                        } else
                        {   // found = FALSE;
#ifdef VERBOSE
                            T1; textpen(TEXTPEN_VERBOSE); zprintf("Unknown game!\n"); T2;
#endif
                            whichgame = -1;
        }   }   }   }   }

        if (!found)
        {   switch (machine)
            {
            case ARCADIA:
                if (romsize <= 8 * KILOBYTE)
                {   memmap = MEMMAP_ARCADIA;
                } else
                {   memmap = MEMMAP_I;
                }
            acase INTERTON:
                if (romsize <= 2048)
                {   memmap = MEMMAP_A;
                } elif (romsize <= 4096)
                {   memmap = MEMMAP_C; // could be B or C, C is generally safer
                } else
                {   memmap = MEMMAP_D;
                }
            acase ELEKTOR:
                memmap = MEMMAP_F;
            acase PIPBUG:
                if     (romsize <= 1280) // 1024 + 256
                {   memmap = MEMMAP_J;
                } elif (romsize <= 1536) // 1024 + 512
                {   memmap = MEMMAP_K;
                } elif (romsize <= 2048) // 1024 + 1024
                {   memmap = MEMMAP_L;
                } elif (romsize <= 8192) // 1024 + 7K
                {   memmap = MEMMAP_M;
                } else                   // 1024 + 11K
                {   memmap = MEMMAP_P;
                }
            acase INSTRUCTOR:
                if (romsize <= 512)
                {   memmap = MEMMAP_N;
                } else
                {   memmap = MEMMAP_O;
                }
            acase CD2650:
                memmap = MEMMAP_CD2650;
            acase DOLPHIN:
                memmap = MEMMAP_DOLPHIN;
            acase PHUNSY:
                memmap = MEMMAP_PHUNSY;
        }   }

        changemachine(machine, FALSE);
    }

    if
    (   kind == KIND_COS
     || kind == KIND_COR
    )
    {   serializecos(version);
        if (machine == PHUNSY)
        {   ubanktoram((banked & 0xF0) >> 4);
            qbanktoram( banked & 0x0F      );
        }

        speakoffset = 0;
        freshly = changed = TRUE;

        switch (machine)
        {
        case ELEKTOR:
            if (memory[0x1C04] == 0x06)
            {   loader = TRUE;
            } elif (memory[0x1C04] == 0x3F)
            {   loader = FALSE;
            } // else do nothing
        acase GALAXIA:
            memory[0x7214] = mg_console;
    }   }

    switch (kind)
    {
    case KIND_BIN:
    case KIND_COS:
    case KIND_COR:
        switch (machine)
        {
        case ARCADIA:
            a_serializerom();
        acase INTERTON:
        case DOLPHIN:
            id_serializerom();
        acase PHUNSY:
            if (kind == KIND_BIN)
            {   if (whichgame == -1)
                {   startaddr_h = 0x20;
                    startaddr_l = 0x00;
                    banked = 0x00;

                    if (romsize > 8192)
                    {   tempsize = 8192;
                    } else
                    {   tempsize = romsize;
                    }
                    for (i = 0; i < tempsize; i++)
                    {   serialize_byte(&memory[0x2000 + i]); // $0000..$1FFF -> $2000..$3FFF
                }   }
                else
                {   banked = known[whichgame].banked;
                    startaddr_h = (UBYTE) (known[whichgame].startaddr / 256); // needed!
                    startaddr_l = (UBYTE) (known[whichgame].startaddr % 256); // needed!
                    if (known[whichgame].the1stsize == 0)
                    {   tempsize = romsize;
                    } else
                    {   tempsize = known[whichgame].the1stsize;
                    }
                    if (known[whichgame].the1staddr >= UBANK0 && known[whichgame].the1staddr < UBANKF + UBANKSIZE)
                    {   // assert(tempsize <= 2048);
                        banktouse  = (known[whichgame].the1staddr - UBANK0) / UBANKSIZE;
                        bankoffset = (known[whichgame].the1staddr - UBANK0) % UBANKSIZE;
#ifdef VERBOSE
                        T1; zprintf("1st chunk: $%04X..$%04X of U%X.\n", 0x1800 + bankoffset, 0x1800 + bankoffset + tempsize, banktouse); T2;
#endif
                        for (i = 0; i < tempsize; i++)
                        {   serialize_byte(&u_bank[banktouse][bankoffset + i]);
                    }   }
                    elif (known[whichgame].the1staddr >= QBANK0 && known[whichgame].the1staddr < QBANKF + QBANKSIZE)
                    {   // assert(tempsize <= 16384);
                        banktouse  = (known[whichgame].the1staddr - QBANK0) / QBANKSIZE;
                        bankoffset = (known[whichgame].the1staddr - QBANK0) % QBANKSIZE;
#ifdef VERBOSE
                        T1; zprintf("1st chunk: $%04X..$%04X of Q%X.\n", 0x4000 + bankoffset, 0x4000 + bankoffset + tempsize, banktouse); T2;
#endif
                        for (i = 0; i < tempsize; i++)
                        {   serialize_byte(&q_bank[banktouse][bankoffset + i]);
                    }   }
                    else
                    {
#ifdef VERBOSE
                        T1; zprintf("1st chunk: $%04X..$%04X.\n", known[whichgame].the1staddr, known[whichgame].the1staddr + tempsize); T2;
#endif
                        for (i = 0; i < tempsize; i++)
                        {   serialize_byte(&memory[known[whichgame].the1staddr + i]);
                    }   }

                    if (known[whichgame].the2ndsize)
                    {   if (known[whichgame].the2ndaddr >= UBANK0 && known[whichgame].the2ndaddr < UBANKF + UBANKSIZE)
                        {   // assert(known[whichgame].the2ndsize <= 2048);
                            banktouse  = (known[whichgame].the2ndaddr - UBANK0) / UBANKSIZE;
                            bankoffset = (known[whichgame].the2ndaddr - UBANK0) % UBANKSIZE;
#ifdef VERBOSE
                            T1; zprintf("2nd chunk: $%04X..$%04X of U%X.\n", 0x1800 + bankoffset, 0x1800 + bankoffset + known[whichgame].the2ndsize, banktouse); T2;
#endif
                            for (i = 0; i < known[whichgame].the2ndsize; i++)
                            {   serialize_byte(&u_bank[banktouse][bankoffset + i]);
                        }   }
                        elif (known[whichgame].the2ndaddr >= QBANK0 && known[whichgame].the2ndaddr < QBANKF + QBANKSIZE)
                        {   // assert(known[whichgame].the2ndsize <= 16384);
                            banktouse  = (known[whichgame].the2ndaddr - QBANK0) / QBANKSIZE;
                            bankoffset = (known[whichgame].the2ndaddr - QBANK0) % QBANKSIZE;
#ifdef VERBOSE
                            T1; zprintf("2nd chunk: $%04X..$%04X of Q%X.\n", 0x4000 + bankoffset, 0x4000 + bankoffset + known[whichgame].the2ndsize, banktouse); T2;
#endif
                            for (i = 0; i < known[whichgame].the2ndsize; i++)
                            {   serialize_byte(&q_bank[banktouse][bankoffset + i]);
                        }   }
                        else
                        {
#ifdef VERBOSE
                            T1; zprintf("2nd chunk: $%04X..$%04X.\n", known[whichgame].the2ndaddr, known[whichgame].the2ndaddr + known[whichgame].the2ndsize); T2;
#endif
                            for (i = 0; i < known[whichgame].the2ndsize; i++)
                            {   serialize_byte(&memory[known[whichgame].the2ndaddr + i]);
                }   }   }   }

                memory[0xEE8] = 0x05;
                memory[0xEE9] = 0xFA;
                memory[0xEEA] = 0x05;
                memory[0xEEB] = 0xE6;
                memory[0xEEC] = 0x05;
                memory[0xEED] = 0xE0;
                memory[0xEEE] = 0xFE;
                memory[0xEEF] = 0x20;
                memory[0xEF2] = 0x1F;
                memory[0xEF3] = 0x3F;
                memory[0xEF4] = 0x10;
                memory[0xEF5] = 0x41;
                memory[0xEF7] = banked;
                memory[0xF01] = 0x0D; // CR
                // we call games as subroutines so they can return to the PHUNSY CLI
                // afterwards (eg. for COPY, although it still doesn't really work).
                ras[0]        = 0x23E;
                psu           = 1; // ie. SP = 1;

                ubanktoram((banked & 0xF0) >> 4);
                qbanktoram( banked & 0x0F      );
        }   }
    acase KIND_PGM:
        switch (machine)
        {
        case ELEKTOR:
            offset = datastart = 0x8C0;
            for (i = 0x8C0; i < romsize; i++)
            {   if (!(memflags[i] & NOWRITE))
                {   memory[i] = ReadByte();
                } else
                {   DISCARD ReadByte();
            }   }
        acase PIPBUG:
            // why do all known programs start at $440 instead of $437?
            offset = datastart = 0x437;
            for (i = 0x437; i < romsize; i++)
            {   memory[i] = ReadByte();
            }
        acase INSTRUCTOR:
            offset = datastart = 0x3;
            for (i = 0x3; i < romsize; i++)
            {   if (!(memflags[i] & NOWRITE))
                {   memory[i] = ReadByte();
                } else
                {   DISCARD ReadByte();
            }   }
        acase CD2650:
            offset = datastart = 0x200;
            for (i = 0x200; i < romsize; i++)
            {   if (!(memflags[i] & NOWRITE))
                {   if (memflags[i] & VISIBLE)
                    {   DISCARD ReadByte();
                        memory[i] = 0x20;
                    } else
                    {   memory[i] = ReadByte();
                }   }
                else
                {   DISCARD ReadByte();
            }   }
        acase PHUNSY:
            offset = datastart = 0x4;
            for (i = 0x4; i < romsize; i++)
            {   if (!(memflags[i] & BIOSRAM) && !(memflags[i] & NOWRITE))
                {   memory[i] = ReadByte();
                } else
                {   DISCARD ReadByte();
            }   }

            ramtoubank(0);
            ramtoqbank(0);
        }

        if (machine != INSTRUCTOR && !pgmtext[0])
        {   i = 0;
            if (machine == PHUNSY)
            {   offset = 4;
            } else
            {   // assert(machine == ELEKTOR || machine == PIPBUG || machine == CD2650);
                offset = 3;
            }

            do
            {   now = ReadByte();
#ifdef AMIGA
                if (now != CR)
                {   pgmtext[i++] = now;
                }
#endif
#ifdef WIN32
                pgmtext[i++] = now;
#endif
            } while (now != 0 && offset < filesize && offset < datastart);
        }
    acase KIND_TVC:
        // assert(machine == ELEKTOR);
        loadaddress = ReadWordAt(1);
        offset = 5;
        for (i = 0; i < filesize - 5; i++)
        {   if (!(memflags[loadaddress + i] & NOWRITE))
            {   memory[loadaddress + i] = ReadByte();
            } else
            {   DISCARD ReadByte();
    }   }   }

    game = TRUE; // must be done before engine_reset()!

    if
    (   kind == KIND_BIN
     || kind == KIND_PGM
     || kind == KIND_TVC
     || kind == KIND_AOF
    )
    {   engine_reset();
    } else
    {
#ifdef PROFILE
        for (i = 0; i <= 255; i++)
        {   profile[i] = 0;
        }
#endif

        for (i = 0; i < 32 * KILOBYTE; i++)
        {   memflags[i] &= ~(COVERAGE_BITS);
    }   }

    if
    (   machine == ELEKTOR
     && (kind == KIND_PGM || kind == KIND_TVC)
    )
    {   // assert(machine == ELEKTOR);
        // memory[$1FC0..$1FC9] = memory[$177..$180]
        // ie.:
        //  SIZES        = $AA
        //  SPR01COLOURS = $09
        //  SPR23COLOURS = $09
        //  SCORECTRL    = $00
        //  $1FC4        = $00
        //  $1FC5        = $00
        //  BGCOLOUR     = $19
        //  PITCH        = $00
        //  SCORELT      = $AA
        //  SCORERT      = $AA
        for (i = 0; i < 10; i++)
        {   memory[0x1FC0 + i] = memory[0x177 + i];
        }

        memory[0x1F00 + PVI_SPRITE0BX] =
        memory[0x1F00 + PVI_SPRITE0BY] =
        memory[0x1F00 + PVI_SPRITE1BX] =
        memory[0x1F00 + PVI_SPRITE1BY] =
        memory[0x1F00 + PVI_SPRITE2BX] =
        memory[0x1F00 + PVI_SPRITE2BY] =
        memory[0x1F00 + PVI_SPRITE3BX] =
        memory[0x1F00 + PVI_SPRITE3BY] = 0xFE; // helps with Figure27.pgm

        // make $8B9..$8BA = $0903 (important!)
        memory[0x8B9] = 0x09;
        memory[0x8BA] = 0x03;
        memory[0x8BD] = 0x1F; // bcta,un
        memory[0x8BE] = startaddr_h;
        memory[0x8BF] = startaddr_l;
        iar  = 0x8BD;
#ifdef VERBOSE
        T1; textpen(TEXTPEN_VERBOSE); zprintf("Setting IAR to $%X.\n", iar); T2;
#endif
    }

    switch (kind)
    {
    case KIND_AOF:
    case KIND_BIN:
    case KIND_COS:
    case KIND_PGM:
    case KIND_TVC:
        free_iobuffer();
        recmode   = RECMODE_NORMAL;
    acase KIND_COR:
        recmode   = RECMODE_PLAY;
        recsize   = filesize;
        ghost_dips(FALSE);
    }

    zinterrupt = FALSE;
    frames    =
    oldframes =
    cycles    =
    elapsed   = 0; // unsigned
    flip      = 0;

    updatemenus();
    updatesmlgads();

    return TRUE;
}

EXPORT void engine_save(int kind, FLAG updatetitle)
{   FILE* TheLocalHandle /* = NULL */ ;
    int   i,
          version;
    UWORD loadaddress,
          endaddress;

    macro_stop(); // in case we are eg. playing back a recording and do a snapshot save

    if (kind == KIND_BIN || kind == KIND_PGM || kind == KIND_TVC)
    {   alloc_iobuffer(32 * KILOBYTE);
    } else
    {   alloc_iobuffer((ULONG) SNPSIZE_MAX);
    }

    offset = 0;
    serializemode = SERIALIZE_WRITE;

    // magic byte (1 byte)
    switch (kind)
    {
    case KIND_PGM:
        switch (machine)
        {
        case ELEKTOR:
            WriteByte( 0x0); // indeterminate opcode (LODZ,r0)
        acase PIPBUG:
            WriteByte(0xC4); // unused opcode
        acase INSTRUCTOR:
            WriteByte(0xC5); // unused opcode
        acase CD2650:
            WriteByte(0x40); // HALT
        }
    acase KIND_TVC:
        WriteByte( 0x2); // valid opcode (LODZ,r2) :-(
        // (although admittedly useless as a first instruction)
    acase KIND_COS:
    case KIND_COR:
        WriteByte(0xBF);  // 0. BSXA,r3
        version = (int) machines[machine].cosversion;
        WriteByte((UBYTE) version); // 1.
        if (kind == KIND_COS)
        {   WriteByte(0); // 2.
        } else
        {   // assert(kind == KIND_COR);
            WriteByte(1); // 2.
        }
        switch (machine)
        {
        case ARCADIA:
            WriteByte(6);
        acase INTERTON:
            WriteByte(7);
        acase ELEKTOR:
            WriteByte(8);
        acase PIPBUG:
            WriteByte(9);
        acase INSTRUCTOR:
            WriteByte(10);
        acase CD2650:
            WriteByte(11);
        acase PHUNSY:
            WriteByte(12);
        acase ASTROWARS:
            WriteByte(3); // 3.
        acase GALAXIA:
            WriteByte(2); // 3.
        acase MALZAK:
            if (memmap == MEMMAP_MALZAK1)
            {   WriteByte(0); // 3.
            } else
            {   // assert(memmap == MEMMAP_MALZAK2);
                WriteByte(1); // 3.
            }
        acase CHAOS:
            WriteByte(4); // 3.
        acase DOLPHIN:
            WriteByte(5); // 3.
        }

        if (machine == PHUNSY)
        {   ramtoubank((banked & 0xF0) >> 4);
            ramtoqbank( banked & 0x0F      );
        }
        serializecos(version);
    }

    switch (kind)
    {
    case KIND_BIN:
    case KIND_COS:
    case KIND_COR:
        switch (machine)
        {
        case ARCADIA:
            a_serializerom();
        acase INTERTON:
            id_serializerom();
        acase DOLPHIN:
            if (version >= 2)
            {   id_serializerom();
        }   }
    acase KIND_PGM:
        switch (machine)
        {
        case ELEKTOR:
            // assert(offset == 1);
            WriteByte(memory[0x8BE]);
            WriteByte(memory[0x8BF]);
            for (i = 3; i <= 0x7FF; i++)
            {   WriteByte(0);
            }
            for (i = 0x800; i < romsize; i++)
            {   WriteByte(memory[i]);
            }
        acase PIPBUG:
        case CD2650:
            // assert(offset == 1);
            WriteByte(startaddr_h);
            WriteByte(startaddr_l);
            for (i = 3; i <= 0x3FF; i++)
            {   WriteByte(0);
            }
            for (i = 0x400; i < romsize; i++)
            {   WriteByte(memory[i]);
            }
        acase INSTRUCTOR:
            // assert(offset == 1);
            WriteByte(startaddr_h);
            WriteByte(startaddr_l);
            for (i = 3; i < romsize; i++)
            {   // currently, it writes the monitor RAM, but should it really?
                if (i >= 0x1800 && i <= 0x1FFF)
                {   WriteByte(0);
                } else
                {   WriteByte(memory[i]);
        }   }   }
    acase KIND_TVC:
        // assert(machine == ELEKTOR);
        // find the first nonzero byte
        loadaddress = // fallback for the rather theoretical case where it is completely empty from $8C0 onwards
        endaddress  = (UWORD) 0x8C0;
        offset      = 0x8C0;
        for (i = 0x8C0; i < romsize; i++)
        {   if (ReadByte() != 0)
            {   loadaddress = (UWORD) i;
                break;
        }   }
        // find the last nonzero byte
        for (i = (int) loadaddress; i < romsize; i++)
        {   if (ReadByte() != 0)
            {   endaddress = (UWORD) i;
        }   }
        // write out the header
        offset = 1;
        WriteWord(loadaddress);   // bytes 1 and 2
        WriteByte(memory[0x8BE]); // byte  3
        WriteByte(memory[0x8BF]); // byte  4
        // write out the body
        for (i = (int) loadaddress; i <= (int) endaddress; i++)
        {   WriteByte(memory[i]);
        }
#ifdef VERBOSE
        T1; textpen(TEXTPEN_VERBOSE);
        zprintf("Load  address is $%X.\n", loadaddress);
        zprintf("Start address is $%X.\n", (memory[0x8BE] * 256) + memory[0x8BF]);
        T2;
#endif
    }

    if ((TheLocalHandle = fopen(fn_game, "wb")))
    {   DISCARD fwrite(IOBuffer, (size_t) offset, 1, TheLocalHandle);
        DISCARD fclose(TheLocalHandle);
        // TheLocalHandle = NULL;
    } else
    {   beep();
        free_iobuffer();
        return;
    }
    free_iobuffer();

#ifdef AMIGA
    writeicon(kind, fn_game);
#endif

    if (kind != KIND_COR)
    {   if (updatetitle)
        {   settitle();
        }
        return;
    }

    if (!(MacroHandle = fopen(fn_game, "a+b")))
    {   beep();
        return;
    }

    // assert(updatetitle == TRUE);
    settitle();
    recmode = RECMODE_RECORD;

    if (MainWindowPtr)
    {   updatesmlgad(GADPOS_MACRO, (ULONG) recmode, TRUE);
}   }

EXPORT void view_xvi(void)
{   int i;

    if
    (   machine != ARCADIA
     && machine != INTERTON
     && machine != ELEKTOR
     && machine != MALZAK
     && machine != ASTROWARS
     && machine != GALAXIA
    ) // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS, DOLPHIN, PHUNSY
    {   return;
    }

    T1; textpen(TEXTPEN_VIEW);

    calcrunningtime();
    zprintf
    (   "%-7s %10d %-7s %10d %-6s %s\n",
        GetCatalogStr(CatalogPtr, MSG_RASTER_SHORT, "Raster:"),
        rast,
        GetCatalogStr(CatalogPtr, MSG_FRAME_SHORT,  "Frame:" ),
        (int) frames,
        GetCatalogStr(CatalogPtr, MSG_TIME_SHORT,   "Time:"  ),
        TimeStr
    );
    zprintf
    (   "%-11s ",
        GetCatalogStr(CatalogPtr, MSG_FLAG,  "Flag:" )
    );
    if (psu & PSU_F)
    {   zprintf
        (   "%6s",
            GetCatalogStr(CatalogPtr, MSG_ENGINE_ON2,  "On" )
        );
    } else
    {   zprintf
        (   "%6s",
            GetCatalogStr(CatalogPtr, MSG_ENGINE_OFF2, "Off")
        );
    }
    zprintf
    (   " %-11s ",
        GetCatalogStr(CatalogPtr, MSG_SENSE, "Sense:")
    );
    if (psu & PSU_S)
    {   zprintf
        (   "%6s",
            GetCatalogStr(CatalogPtr, MSG_ENGINE_ON2,  "On" )
        );
    } else
    {   zprintf
        (   "%6s",
            GetCatalogStr(CatalogPtr, MSG_ENGINE_OFF2, "Off")
        );
    }
    zprintf("\n");

    switch (machine)
    {
    case ARCADIA:
        zprintf
        (   "CONSOLE:        %02X PITCH:          %02X VOLUME:         %02X\n",
            memory[A_CONSOLE],
            memory[A_PITCH],
            memory[A_VOLUME]
        );
        zprintf
        (   "RESOLUTION:     %02X BGCOLOUR:       %02X VSCROLL:        %02X CHARLINE:       %02X\n",
            memory[A_RESOLUTION],
            memory[A_BGCOLOUR],
            memory[A_VSCROLL],
            memory[A_CHARLINE]
        );
        zprintf
        (   "BGCOLLIDE:      %02X SPRITECOLLIDE:  %02X SPRITES01CTRL:  %02X SPRITES23CTRL:  %02X\n",
            memory[A_BGCOLLIDE    ],
            memory[A_SPRITECOLLIDE],
            memory[A_SPRITES01CTRL],
            memory[A_SPRITES23CTRL]
        );
        zprintf
        (   "SPRITE0X,Y:  %02X,%02X SPRITE1X,Y:  %02X,%02X SPRITE2X,Y:  %02X,%02X SPRITE3X,Y:  %02X,%02X\n",
            memory[A_SPRITE0X],
            memory[A_SPRITE0Y],
            memory[A_SPRITE1X],
            memory[A_SPRITE1Y],
            memory[A_SPRITE2X],
            memory[A_SPRITE2Y],
            memory[A_SPRITE3X],
            memory[A_SPRITE3Y]
        );
        zprintf
        (   "P1PADDLE:       %02X P1LEFTKEYS:     %02X P1MIDDLEKEYS:   %02X P1RIGHTKEYS:    %02X\n",
            memory[A_P1PADDLE],
            memory[A_P1LEFTKEYS],
            memory[A_P1MIDDLEKEYS],
            memory[A_P1RIGHTKEYS]
        );
        zprintf
        (   "P2PADDLE:       %02X P2LEFTKEYS:     %02X P2MIDDLEKEYS:   %02X P2RIGHTKEYS:    %02X\n",
            memory[A_P2PADDLE],
            memory[A_P2LEFTKEYS],
            memory[A_P2MIDDLEKEYS],
            memory[A_P2RIGHTKEYS]
        );
        zprintf
        (   "P1PALLADIUM:    %02X P2PALLADIUM:    %02X\n\n",
            memory[A_P1PALLADIUM],
            memory[A_P2PALLADIUM]
        );

        if (traceverbose)
        {   zprintf
            (   "Board mode:                 %s\n",
                (memory[A_PITCH     ] & 0x80) ? "yes" : "no"
            );
            zprintf
            (   "Block mode:                 %s\n",
                (memory[A_RESOLUTION] & 0x80) ? "yes" : "no"
            );
            zprintf
            (   "Pitch:                      %4.3f Hz\n",
                (float) (7874.0 / ((memory[A_PITCH] & 0x7F) + 1))
            );
            zprintf
            (   "Horizontal scrolling:       %d pixels\n",
                (memory[A_VOLUME] & 0xE0) >> 5
            );
            zprintf
            (   "Tone:                       %s\n",
                (memory[A_VOLUME] & 0x08) ? "on" : "off"
            );
            zprintf
            (   "Noise:                      %s\n",
                (memory[A_VOLUME] & 0x10) ? "on" : "off"
            );
            zprintf
            (   "Volume:                     %d of 7\n",
                memory[A_VOLUME] & 0x07
            );
            zprintf
            (   "Character rows:             %d\n",
                (memory[A_RESOLUTION] & 0x40) ? 26 : 13
            );
            zprintf
            (   "Vertical resolution:        %s\n",
                (memory[A_BGCOLOUR] & 0x80) ? "high" : "low"
            );
            zprintf
            (   "Paddle interpolation:       %c-axis\n",
                (memory[A_BGCOLOUR] & 0x40) ? 'X' : 'Y'
            );
            zprintf
            (   "Colours:\n Foreground:                %s, %s, %s, %s\n",
                colourname[from_a[  (memory[A_BGCOLOUR] & 0x38) >> 3          ]],
                colourname[from_a[(((memory[A_BGCOLOUR] & 0x38) >> 3) + 2) % 8]],
                colourname[from_a[(((memory[A_BGCOLOUR] & 0x38) >> 3) + 4) % 8]],
                colourname[from_a[(((memory[A_BGCOLOUR] & 0x38) >> 3) + 6) % 8]]
            );
            zprintf
            (   " Board mode:                %s\n",
                colourname[from_a[(memory[A_RESOLUTION] & 0x38) >> 3]]
            );
            zprintf
            (   " Inner background:          %s\n",
                colourname[from_a[memory[A_BGCOLOUR  ] & 0x07]]
            );
            zprintf
            (   " Outer background:          %s\n",
                colourname[from_a[memory[A_RESOLUTION] & 0x07]]
            );

            zprintf
            (   "Collisions:\n Sprites vs. background:    %s, %s, %s, %s\n",
                (memory[A_BGCOLLIDE] & 0x01) ? "no " : "yes",
                (memory[A_BGCOLLIDE] & 0x02) ? "no " : "yes",
                (memory[A_BGCOLLIDE] & 0x04) ? "no " : "yes",
                (memory[A_BGCOLLIDE] & 0x08) ? "no " : "yes"
            );

            zprintf
            (   " Sprite #0 vs. #0/#1/#2/#3: n/a, %s, %s, %s\n",
                (memory[A_SPRITECOLLIDE] & 0x01) ? "no " : "yes",
                (memory[A_SPRITECOLLIDE] & 0x02) ? "no " : "yes",
                (memory[A_SPRITECOLLIDE] & 0x04) ? "no " : "yes"
            );
            zprintf
            (   " Sprite #1 vs. #0/#1/#2/#3: %s, n/a, %s, %s\n",
                (memory[A_SPRITECOLLIDE] & 0x01) ? "no " : "yes",
                (memory[A_SPRITECOLLIDE] & 0x08) ? "no " : "yes",
                (memory[A_SPRITECOLLIDE] & 0x10) ? "no " : "yes"
            );
            zprintf
            (   " Sprite #2 vs. #0/#1/#2/#3: %s, %s, n/a, %s\n",
                (memory[A_SPRITECOLLIDE] & 0x02) ? "no " : "yes",
                (memory[A_SPRITECOLLIDE] & 0x08) ? "no " : "yes",
                (memory[A_SPRITECOLLIDE] & 0x20) ? "no " : "yes"
            );
            zprintf
            (   " Sprite #3 vs. #0/#1/#2/#3: %s, %s, %s, n/a\n",
                (memory[A_SPRITECOLLIDE] & 0x04) ? "no " : "yes",
                (memory[A_SPRITECOLLIDE] & 0x10) ? "no " : "yes",
                (memory[A_SPRITECOLLIDE] & 0x20) ? "no " : "yes"
            );

/*          if (memory[A_SPRITES01CTRL] & 0x80)
            {   zprintf("Sprite #0 is double height\n");
            } else
            {   zprintf("Sprite #0 is normal height\n");
            }
            if (memory[A_SPRITES01CTRL] & 0x40)
            {   zprintf("Sprite #1 is double height\n");
            } else
            {   zprintf("Sprite #1 is normal height\n");
            }
            zprintf("Sprite #0 is %s\n", colourname[from_a[(memory[A_SPRITES01CTRL] & 0x38) >> 3]]);
            // 0 means yellow for $00..$3F, GREEN,RED,BLACK
            zprintf("Sprite #1 is %s\n", colourname[from_a[ memory[A_SPRITES01CTRL] & 0x08      ]]);

            if (memory[A_SPRITES23CTRL] & 0x80)
            {   zprintf("Sprite #2 is double height\n");
            } else
            {   zprintf("Sprite #2 is normal height\n");
            }
            if (memory[A_SPRITES23CTRL] & 0x40)
            {   zprintf("Sprite #3 is double height\n");
            } else
            {   zprintf("Sprite #3 is normal height\n");
            }
            zprintf("Sprite #2 is %s\n", colourname[from_a[(memory[A_SPRITES23CTRL] & 0x38) >> 3]]);
            zprintf("Sprite #3 is %s\n", colourname[from_a[ memory[A_SPRITES23CTRL] & 0x08      ]]);
*/      }
    acase INTERTON:
    case ELEKTOR:
    case MALZAK:
    case ASTROWARS:
    case GALAXIA:
        if (memmap == MEMMAP_E)
        {   zprintf
            (   "CONSOLE:        %02X\n",
                memory[IE_CONSOLE]
            );
        } elif (machine == INTERTON || memmap == MEMMAP_F)
        {   zprintf
            (   "CONSOLE:        %02X NOISE:          %02X\n",
                memory[IE_CONSOLE],
                memory[IE_NOISE]
            );
        }

        if (machine == MALZAK || machine == GALAXIA)
        {   zprintf("---1st PVI---\n");
        } elif (machine == INTERTON || machine == ELEKTOR)
        {   zprintf
            (   "P1LEFTKEYS:     %02X P1MIDDLEKEYS:   %02X P1RIGHTKEYS:    %02X\n",
                memory[IE_P1LEFTKEYS],
                memory[IE_P1MIDDLEKEYS],
                memory[IE_P1RIGHTKEYS]
            );
            zprintf
            (   "P2LEFTKEYS:     %02X P2MIDDLEKEYS:   %02X P2RIGHTKEYS:    %02X\n",
                memory[IE_P2LEFTKEYS],
                memory[IE_P2MIDDLEKEYS],
                memory[IE_P2RIGHTKEYS]
            );
        }

        for (i = 0; i <= 2; i++)
        {   if (i == 1)
            {   if (machine == MALZAK || machine == GALAXIA)
                {   pvibase += 0x100;
                    zprintf("---2nd PVI---\n");
                } else
                {   break;
            }   }
            elif (i == 2)
            {   if (machine == MALZAK)
                {   break;
                } else
                {   // assert(memmap == MEMMAP_W);
                    pvibase += 0x100;
                    zprintf("---3rd PVI---\n");
            }   }

            zprintf
            (   "PITCH:          %02X SIZES:          %02X P1PADDLE:       %02X P2PADDLE:       %02X\n",
                memory[pvibase + PVI_PITCH],
                memory[pvibase + PVI_SIZES],
                memory[pvibase + PVI_P1PADDLE],
                memory[pvibase + PVI_P2PADDLE]
            );
            zprintf
            (   "SCORECTRL:      %02X SCORELT:        %02X SCORERT:        %02X BGCOLOUR:       %02X\n",
                memory[pvibase + PVI_SCORECTRL],
                memory[pvibase + PVI_SCORELT],
                memory[pvibase + PVI_SCORERT],
                memory[pvibase + PVI_BGCOLOUR]
            );
            zprintf
            (   "BGCOLLIDE:      %02X SPRITECOLLIDE:  %02X SPR01COLOURS:   %02X SPR23COLOURS:   %02X\n",
                memory[pvibase + PVI_BGCOLLIDE],
                memory[pvibase + PVI_SPRITECOLLIDE],
                memory[pvibase + PVI_SPR01COLOURS],
                memory[pvibase + PVI_SPR23COLOURS]
            );
            zprintf
            (   "SPRITE0AX,Y: %02X,%02X SPRITE1AX,Y: %02X,%02X SPRITE2AX,Y: %02X,%02X SPRITE3AX,Y: %02X,%02X\n",
                memory[pvibase + PVI_SPRITE0AX],
                memory[pvibase + PVI_SPRITE0AY],
                memory[pvibase + PVI_SPRITE1AX],
                memory[pvibase + PVI_SPRITE1AY],
                memory[pvibase + PVI_SPRITE2AX],
                memory[pvibase + PVI_SPRITE2AY],
                memory[pvibase + PVI_SPRITE3AX],
                memory[pvibase + PVI_SPRITE3AY]
            );
            zprintf
            (   "SPRITE0BX,Y: %02X,%02X SPRITE1BX,Y: %02X,%02X SPRITE2BX,Y: %02X,%02X SPRITE3BX,Y: %02X,%02X\n",
                memory[pvibase + PVI_SPRITE0BX],
                memory[pvibase + PVI_SPRITE0BY],
                memory[pvibase + PVI_SPRITE1BX],
                memory[pvibase + PVI_SPRITE1BY],
                memory[pvibase + PVI_SPRITE2BX],
                memory[pvibase + PVI_SPRITE2BY],
                memory[pvibase + PVI_SPRITE3BX],
                memory[pvibase + PVI_SPRITE3BY]
            );

            if (traceverbose)
            {   if (machine == INTERTON || memmap == MEMMAP_F)
                {   zprintf
                    (   "\nNoise pitch:       %4.3f Hz\n",
                        (float) (7874.0 / (memory[IE_NOISE] + 1))
                    );
                }

                if (memory[pvibase + PVI_SCORECTRL] & 0x02)
                {   zprintf("Score format:      1 group of 4 digits\n");
                } else
                {   zprintf("Score format:      2 groups of 2 digits\n");
                }
                if (memory[pvibase + PVI_SCORECTRL] & 0x01)
                {   zprintf("Score position:    bottom of screen\n");
                } else
                {   zprintf("Score position:    top of screen\n");
                }
                zprintf
                (   "Grid/background:   %s\n",
                    (memory[pvibase + PVI_BGCOLOUR] & 0x08) ? "enabled" : "disabled"
                );
                zprintf
                (   "Grid colour:       %s\n",
                    colourname[(memory[pvibase + PVI_BGCOLOUR] & 0x70) >> 4]
                );
                zprintf
                (   "Background colour: %s\n",
                    colourname[memory[pvibase + PVI_BGCOLOUR] & 0x07]
                );
        }   }

        if (machine == MALZAK)
        {   pvibase = 0x1400;
        } elif (machine == ASTROWARS || machine == GALAXIA)
        {   pvibase = 0x1500;
    }   }

    zprintf("\n");
    T2;
}

EXPORT void view_udcs(void)
{   switch (machine)
    {
    case ARCADIA:
    case INTERTON:
    case ELEKTOR:
        view_udcs_real();
    acase MALZAK:
        // assert(pvibase = 0x1400);
        view_udcs_real();
        pvibase = 0x1500;
        view_udcs_real();
        pvibase = 0x1400;
    acase ASTROWARS:
        // assert(pvibase = 0x1500);
        view_udcs_real();
    acase GALAXIA:
        // assert(pvibase = 0x1500);
        view_udcs_real();
        pvibase = 0x1600;
        view_udcs_real();
        pvibase = 0x1700;
        view_udcs_real();
        pvibase = 0x1500;
}   }

EXPORT void view_udcs_real(void)
{   TRANSIENT       int    i, j, k;
    PERSIST   const STRPTR a_sizename[2] =
    {  "(8*16",
       " (8*8",
    }, ie_sizename[4] =
    { " (8*10",
      "(16*20",
      "(32*40",
      "(64*80"
    };

    // Ideally, we would show the UDCs/sprites in the correct colours

    T1; textpen(TEXTPEN_VIEW);
    switch (machine)
    {
    case ARCADIA:
        for (i = 0; i < 4; i++)
        {   for (j = 0; j < 8; j++)
            {   for (k = 0; k < 8; k++)
                {   if (memory[A_OFFSET_SPRITES + (i * 8) + j] & (0x80 >> k))
                    {   bits[i][j][k] = '#';
                    } else
                    {   bits[i][j][k] = '.';
                }   }
                bits[i][j][8] = 0;
        }   }
        zprintf
        (   "SPRITE0: %s SPRITE1: %s SPRITE2: %s SPRITE3: %s\n",
            bits[0][0], bits[1][0], bits[2][0], bits[3][0]
        );
        if (flagline && (psu & PSU_F))
        {   zprintf
            (   "%8s %s %8s %s %8s %s %8s %s",
                colourname[from_a[7 - ((memory[A_SPRITES01CTRL] & 0x38) >> 3)]],
                bits[0][1],
                colourname[from_a[7 - ( memory[A_SPRITES01CTRL] & 0x07      )]],
                bits[1][1],
                colourname[from_a[7 - ((memory[A_SPRITES23CTRL] & 0x38) >> 3)]],
                bits[2][1],
                colourname[from_a[7 - ( memory[A_SPRITES23CTRL] & 0x07      )]],
                bits[3][1]
            );
        } else
        {   zprintf
            (   "%8s %s %8s %s %8s %s %8s %s",
                colourname[from_a[(memory[A_SPRITES01CTRL] & 0x38) >> 3]],
                bits[0][1],
                colourname[from_a[ memory[A_SPRITES01CTRL] & 0x07      ]],
                bits[1][1],
                colourname[from_a[(memory[A_SPRITES23CTRL] & 0x38) >> 3]],
                bits[2][1],
                colourname[from_a[memory[A_SPRITES23CTRL] & 0x07       ]],
                bits[3][1]
            );
        }

        zprintf
        (   "\n  %s) %s   %s) %s   %s) %s   %s) %s\n",
            a_sizename[(memory[A_SPRITES01CTRL] & 0x80) >> 7],
            bits[0][2],
            a_sizename[(memory[A_SPRITES01CTRL] & 0x40) >> 6],
            bits[1][2],
            a_sizename[(memory[A_SPRITES23CTRL] & 0x80) >> 7],
            bits[2][2],
            a_sizename[(memory[A_SPRITES23CTRL] & 0x40) >> 6],
            bits[3][2]
        );

        for (i = 3; i < 8; i++)
        {   zprintf
            (   "         %s          %s          %s          %s\n",
                bits[0][i], bits[1][i], bits[2][i], bits[3][i]
            );
        }

        for (i = 4; i < 8; i++)
        {   for (j = 0; j < 8; j++)
            {   for (k = 0; k < 8; k++)
                {   if (memory[A_OFFSET_SPRITES + (i * 8) + j] & (0x80 >> k))
                    {   bits[i - 4][j][k] = (TEXT) '#';
                    } else
                    {   bits[i - 4][j][k] = (TEXT) '.';
                }   }
                bits[i - 4][j][8] = 0;
        }   }

        zprintf
        (   "\n   UDC0: %s    UDC1: %s    UDC2: %s    UDC3: %s\n",
            bits[0][0], bits[1][0], bits[2][0], bits[3][0]
        );
        if (memory[A_BGCOLOUR] & 0x80)
        {   zprintf
            (   "   (8*8) %s    (8*8) %s    (8*8) %s    (8*8) %s\n",
                bits[0][1],
                bits[1][1],
                bits[2][1],
                bits[3][1]
            );
        } else
        {   zprintf
            (   "  (8*16) %s   (8*16) %s   (8*16) %s   (8*16) %s\n",
                bits[0][1],
                bits[1][1],
                bits[2][1],
                bits[3][1]
            );
        }

        for (i = 2; i < 8; i++)
        {   zprintf
            (   "         %s          %s          %s          %s\n",
                bits[0][i], bits[1][i], bits[2][i], bits[3][i]
            );
        }
    acase INTERTON:
    case ELEKTOR:
    case ASTROWARS:
    case GALAXIA:
    case MALZAK:
        for (i = 0; i < 4; i++)
        {   for (j = 0; j < 10; j++)
            {   for (k = 0; k < 8; k++)
                {   if (memory[pvibase + pvi_spritedata[i] + j] & (0x80 >> k))
                    {   bits[i][j][k] = '#';
                    } else
                    {   bits[i][j][k] = '.';
                }   }
                bits[i][j][8] = 0;
        }   }

        if ( machine == GALAXIA && pvibase == 0x1700)
        {   zprintf
            (   "SPRITE8: %s SPRITE9: %s SPRITE10:%s SPRITE11:%s\n",
                bits[0][0], bits[1][0], bits[2][0], bits[3][0]
            );
        } elif
        (   (machine == MALZAK  && pvibase == 0x1500)
         || (memmap  == GALAXIA && pvibase == 0x1600)
        )
        {   zprintf
            (   "SPRITE4: %s SPRITE5: %s SPRITE6: %s SPRITE7: %s\n",
                bits[0][0], bits[1][0], bits[2][0], bits[3][0]
            );
        } else
        {   zprintf
            (   "SPRITE0: %s SPRITE1: %s SPRITE2: %s SPRITE3: %s\n",
                bits[0][0], bits[1][0], bits[2][0], bits[3][0]
            );
        }

        switch (machine)
        {
        case INTERTON:
        case ELEKTOR:
            zprintf
            (   "%8s %s %8s %s %8s %s %8s %s\n",
                colourname[(memory[pvibase + PVI_SPR01COLOURS] >> 3) & 7],
                bits[0][1],
                colourname[ memory[pvibase + PVI_SPR01COLOURS]       & 7],
                bits[1][1],
                colourname[(memory[pvibase + PVI_SPR23COLOURS] >> 3) & 7],
                bits[2][1],
                colourname[ memory[pvibase + PVI_SPR23COLOURS]       & 7],
                bits[3][1]
            );
        acase ASTROWARS:
            zprintf
            (   "%8s %s %8s %s %8s %s %8s %s\n",
                colourname[from_g_x_spr[(memory[pvibase + PVI_SPR01COLOURS] >> 3) & 7]],
                bits[0][1],
                colourname[from_g_x_spr[ memory[pvibase + PVI_SPR01COLOURS]       & 7]],
                bits[1][1],
                colourname[from_g_x_spr[(memory[pvibase + PVI_SPR23COLOURS] >> 3) & 7]],
                bits[2][1],
                colourname[from_g_x_spr[ memory[pvibase + PVI_SPR23COLOURS]       & 7]],
                bits[3][1]
            );
        acase GALAXIA:
            zprintf
            (   "%8s %s %8s %s %8s %s %8s %s\n",
                colourname[from_g_w_spr[(memory[pvibase + PVI_SPR01COLOURS] >> 3) & 7]],
                bits[0][1],
                colourname[from_g_w_spr[ memory[pvibase + PVI_SPR01COLOURS]       & 7]],
                bits[1][1],
                colourname[from_g_w_spr[(memory[pvibase + PVI_SPR23COLOURS] >> 3) & 7]],
                bits[2][1],
                colourname[from_g_w_spr[ memory[pvibase + PVI_SPR23COLOURS]       & 7]],
                bits[3][1]
            );
        acase MALZAK:
            zprintf
            (   "%8s %s %8s %s %8s %s %8s %s\n",
                colourname[from_m_spr[  (memory[pvibase + PVI_SPR01COLOURS] >> 3) & 7]],
                bits[0][1],
                colourname[from_m_spr[   memory[pvibase + PVI_SPR01COLOURS]       & 7]],
                bits[1][1],
                colourname[from_m_spr[  (memory[pvibase + PVI_SPR23COLOURS] >> 3) & 7]],
                bits[2][1],
                colourname[from_m_spr[   memory[pvibase + PVI_SPR23COLOURS]       & 7]],
                bits[3][1]
            );
        }

        zprintf
        (   " %s) %s  %s) %s  %s) %s  %s) %s\n",
            ie_sizename[ memory[pvibase + PVI_SIZES]       & 3],
            bits[0][2],
            ie_sizename[(memory[pvibase + PVI_SIZES] >> 2) & 3],
            bits[1][2],
            ie_sizename[(memory[pvibase + PVI_SIZES] >> 4) & 3],
            bits[2][2],
            ie_sizename[(memory[pvibase + PVI_SIZES] >> 6) & 3],
            bits[3][2]
        );
        for (i = 3; i < 10; i++)
        {   zprintf
            (   "         %s          %s          %s          %s\n",
                bits[0][i], bits[1][i], bits[2][i], bits[3][i]
            );
    }   }
    zprintf("\n");
    T2;
}

EXPORT void view_imagery(void)
{   int howmany,
        i, j, k;

    switch (memmap)
    {
    case MEMMAP_ASTROWARS:
    case MEMMAP_GALAXIA:
        howmany = 13 * KILOBYTE;
    acase MEMMAP_MALZAK1:
        howmany = 19 * KILOBYTE;
    acase MEMMAP_MALZAK2:
        howmany = 29 * KILOBYTE;  
    acase MEMMAP_PHUNSY:
        howmany = 32 * KILOBYTE;
    adefault:
        howmany = romsize;
    }

    bits[0][0][8] = 0;
    T1; textpen(TEXTPEN_VIEW);
    for (i = 0; i < howmany; i++)
    {   if (memmap == MEMMAP_D && i > 0x15FF)
        {   break;
        } // implied else
        if (machine == ARCADIA && i >= 4096)
        {   k = i + 4096; // skip $1xxx region
        } else
        {   k = i;
        }
        for (j = 0; j < 8; j++)
        {   if (memory[k] & (0x80 >> j))
            {   bits[0][0][j] = '#';
            } else
            {   bits[0][0][j] = '.';
        }   }

        zprintf
        (   "        db      $%02X ;%s $%04X\n",
            memory[k],
            bits[0][0],
            k
        );
    }
    zprintf("\n");
    T2;
}

EXPORT void view_controls_engine(void)
{   int i;

    for (i = 0; i < KEYS; i++)
    {   oldkeys[i] =
        newkeys[i] = 0;
    }

    sx[0] =
    sx[1] =
    sy[0] =
    sy[1] = 128; // so we don't have any permanently lit paddles

    mouseup();
}

EXPORT void view_bram(void)
{   int  i, y;

    T1; textpen(TEXTPEN_VIEW);

    switch (machine)
    {
    case ELEKTOR:
        if (memory[0x8B4] & PSU_S  ) psubits[0] = 'S'; else psubits[0] = 's';
        if (memory[0x8B4] & PSU_F  ) psubits[1] = 'F'; else psubits[1] = 'f';
        if (memory[0x8B4] & PSU_II ) psubits[2] = 'I'; else psubits[2] = 'i';
        psubits[3] = 0;
        if (memory[0x8B3] & PSL_IDC) pslbits[0] = 'D'; else pslbits[0] = 'd';
        if (memory[0x8B3] & PSL_RS ) pslbits[1] = 'R'; else pslbits[1] = 'r';
        if (memory[0x8B3] & PSL_WC ) pslbits[2] = 'W'; else pslbits[2] = 'w';
        if (memory[0x8B3] & PSL_OVF) pslbits[3] = 'O'; else pslbits[3] = 'o';
        if (memory[0x8B3] & PSL_COM) pslbits[4] = 'M'; else pslbits[4] = 'm';
        if (memory[0x8B3] & PSL_C  ) pslbits[5] = 'C'; else pslbits[5] = 'c';
        pslbits[6] = 0;

        for (y = 0; y < 6; y++) // 6 character rows
        {   if (y == 0)
            {   zprintf("MONOB:  \"");
            } else
            {   zprintf("        \"");
            }

            monob(y);
            zprintf("%s", letters);

            if (y == 0)
            {   zprintf
                (   "\" $8A0..$8AB: $%02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X %02X\n",
                    memory[0x8A0],
                    memory[0x8A1],
                    memory[0x8A2],
                    memory[0x8A3],
                    memory[0x8A4],
                    memory[0x8A5],
                    memory[0x8A6],
                    memory[0x8A7],
                    memory[0x8A8],
                    memory[0x8A9],
                    memory[0x8AA],
                    memory[0x8AB]
                );
            } elif (y == 1)
            {   zprintf
                (   "\" $8B9..$8BD:                            $%02X %02X %02X %02X %02X\n",
                    memory[0x8B9],
                    memory[0x8BA],
                    memory[0x8BB],
                    memory[0x8BC],
                    memory[0x8BD]
                ); // this area would be better shown as code rather than data
            } elif (y == 3)
            {   zprintf
                (   "\" LKBST:         $%02X RKBST:         $%02X MKBST:         $%02X\n",
                    memory[E_LKBST],
                    memory[E_RKBST],
                    memory[E_MKBST]
                );
            } elif (y == 4)
            {   zprintf
                (   "\" BK1:         $%04X BK2:         $%04X PC:          $%04X\n",
                    (256 * memory[E_BK1]) + memory[E_BK1 + 1],
                    (256 * memory[E_BK2]) + memory[E_BK2 + 1],
                    (256 * memory[E_PC ]) + memory[E_PC  + 1]
                );
            } elif (y == 5)
            {   zprintf
                (   "\" R0:            $%02X PSU:           %s PSL:        %s\n",
                    memory[0x8AC],
                    psubits,
                    pslbits
                );
            } else
            {   zprintf("\"\n");
        }   }

        zprintf("MLINE:  \"");
        for (i = 0; i < 8; i++)
        {   zprintf("%c", elektor_chars1[memory[0x890 + i]]);
        }
        zprintf
        (   "\" R1:            $%02X R2:            $%02X R3:            $%02X\n",
            memory[0x8AD],
            memory[0x8AE],
            memory[0x8AF]
        );
        zprintf
        (   "                   R4:            $%02X R5:            $%02X R6:            $%02X\n",
            memory[0x8B0],
            memory[0x8B1],
            memory[0x8B2]
        );
        zprintf
        (   "FSEQ:          $%02X ENTM:          $%02X MFUNC:         $%02X MSCR:        $%04X\n\n",
            memory[E_FSEQ], // FSEQ is aka SILENCE
            memory[E_ENTM],
            memory[E_MFUNC],
            (256 * memory[E_MSCR]) + memory[E_MSCR + 1]
        );
    acase PIPBUG:
        // zprintf("\n"); would take care of the PIPBUG prompt ("*").

        zprintf
        (   "COM  ($400..$408):  $%02X %02X %02X %02X %02X %02X %02X %02X %02X\n",
            memory[0x400],
            memory[0x401],
            memory[0x402],
            memory[0x403],
            memory[0x404],
            memory[0x405],
            memory[0x406],
            memory[0x407],
            memory[0x408]
        );
        zprintf
        (   "XGOT ($409..$40C):  $%02X %02X %02X %02X\n", // better to show this as code
            memory[0x409],
            memory[0x40A],
            memory[0x40B],
            memory[0x40C]
        );
        zprintf
        (   "BUFF ($413..$426):  $%02X %02X %02X %02X %02X %02X %02X %02X %02X %02X\n",
            memory[0x413],
            memory[0x414],
            memory[0x415],
            memory[0x416],
            memory[0x417],
            memory[0x418],
            memory[0x419],
            memory[0x41A],
            memory[0x41B],
            memory[0x41C]
        );
        zprintf
        (   "                    $%02X %02X %02X %02X %02X %02X %02X %02X %02X %02X\n",
            memory[0x41D],
            memory[0x41E],
            memory[0x41F],
            memory[0x420],
            memory[0x421],
            memory[0x422],
            memory[0x423],
            memory[0x424],
            memory[0x425],
            memory[0x426]
        );
        zprintf
        (   "TEMP: $%04X TEMQ: $%04X TEMR:   $%02X TEMS:   $%02X\n",
            (memory[0x40D] * 256) + memory[0x40E],
            (memory[0x40F] * 256) + memory[0x410],
            memory[T_TEMR],
            memory[T_TEMS]
        );
        zprintf
        (   "MARK: $%04X HDAT: $%04X LDAT: $%04X HADR: $%04X LADR: $%04X\n",
            (memory[T_MARK] * 256) + memory[T_MARK + 1],
            (memory[T_HDAT] * 256) + memory[T_HDAT + 1],
            (memory[T_LDAT] * 256) + memory[T_LDAT + 1],
            (memory[T_HADR] * 256) + memory[T_HADR + 1],
            (memory[T_LADR] * 256) + memory[T_LADR + 1]
        );
        zprintf
        (   "BPTR:   $%02X MCNT:   $%02X CNT:    $%02X CODE:   $%02X OKGO:   $%02X BCC:   $%02X\n\n",
            memory[T_BPTR],
            memory[T_MCNT],
            memory[T_CNT ],
            memory[T_CODE],
            memory[T_OKGO],
            memory[T_BCC ]
        );
    acase INSTRUCTOR:
        for (i = 0x17C0; i < 0x1800; i += 16)
        {   dumprow(i);
        }
        zprintf("\n");
    acase CD2650:
        zprintf
        (   "ADD:  $%04X BKP:  $%04X BPD:  $%04X CUR:  $%04X\n",
            (256 * memory[C_ADD1]) + memory[C_ADD2],
            (256 * memory[C_BKP1]) + memory[C_BKP2],
            (256 * memory[C_BPD1]) + memory[C_BPD2],
            (256 * memory[C_CUR1]) + memory[C_CUR2]
        );
        zprintf
        (   "LENT:   $%02X SUMC:   $%02X\n\n",
            memory[C_LENT],
            memory[C_SUMC]
        );
    acase PHUNSY:
        zprintf("MDCR RAM ($C00..$E40): \n");
        for (i = 0xC00; i <= 0xE40; i += 16) // it would be better to omit $E41..$E4F but dumprow() doesn't allow this
        {   dumprow(i);
        }
        zprintf("\nBIOS RAM (scratchpad) ($EE0..$F00):\n"); // localize this!
        zprintf
        (   "ZB1A:   $%04X ZB2A:   $%04X CMAD:   $%04X ERET:   $%04X\n",
            (256 * memory[0xEE0]) + memory[0xEE1], // ZB1A
            (256 * memory[0xEE2]) + memory[0xEE3], // ZB2A
            (256 * memory[0xEE4]) + memory[0xEE5], // CMAD
            (256 * memory[0xEE6]) + memory[0xEE7]  // ERET
        );
        zprintf
        (   "IOAD:   $%04X IIAD:   $%04X ICAD:   $%04X\n",
            (256 * memory[0xEE8]) + memory[0xEE9], // IOAD
            (256 * memory[0xEEA]) + memory[0xEEB], // IIAD
            (256 * memory[0xEEC]) + memory[0xEED]  // ICAD
        );
        zprintf
        (   "CURS:   $%04X TBSR:   $%04X\n",
            (256 * memory[0xEF4]) + memory[0xEF5], // CURS (cursor position)
            (256 * memory[0xEFE]) + memory[0xEFF]  // TBSR (two bytes register)
        );
        zprintf
        (   "STAD:   $%04X ENAD:   $%04X TOAD:   $%04X\n",
            (256 * memory[0xEF8]) + memory[0xEF9], // STAD (start address)
            (256 * memory[0xEFA]) + memory[0xEFB], // ENAD (end address)
            (256 * memory[0xEFC]) + memory[0xEFD]  // TOAD (target address)
        );
        zprintf
        (   "BRTRES:   $%02X CURRES:   $%02X COMRES:   $%02X CPORES:   $%02X\n",
            memory[0xEEE], // BRTRES (baud rate)
            memory[0xEEF], // CURRES (cursor position)
            memory[0xEF6], // COMRES (command compare)
            memory[0xEF7]  // CPORES (control port out storage)
        );
        zprintf
        (   "LESTRE:   $%02X RIENRE:   $%02X UPSTRE:   $%02X LOENRE:   $%02X\n",
            memory[0xEF1], // LESTRE (screen border left)
            memory[0xEF3], // RIENRE (screen border right)
            memory[0xEF0], // UPSTRE (screen border upper)
            memory[0xEF2]  // LESTRE (screen border lower)
        );
        zprintf
        (   "INPRAM:   $%02X\n",
            memory[0xF00]  // INPRAM (index)
        );
        zprintf("\nBIOS RAM (command input buffer) ($F01..$FFF):\n"); // localize this!
        for (i = 0xF01; i < 0x1000; i += 16) // $F01..$1000 (it would be better to omit $1000 but dumprow() doesn't allow this
        {   dumprow(i);
        }
        zprintf("\n");
    }

    T2;
}

EXPORT void update_monitor(FLAG force)
{   PERSIST int i, j, // all PERSISTent for speed
                sprites;

    update_controls();
    update_memory(force);

    if (!MonitorWindowPtr)
    {   return;
    }

 /* In some games, eg. Grand Slam Tennis, the monitor will always report
    PnPADDLE as being $FF; this is because after the game reads these
    registers, they are set to $FF (being read-once).

    TwoStr[0..6]    = RGad[0..6]   = r[0..6]
    TwoStr[7]       = CCGad        = CC
    OneStr[0]       = OneGad[0]    = S
    OneStr[1]       = OneGad[1]    = F
    OneStr[2]       = OneGad[2]    = II
    OneStr[3]       = OneGad[3]    = SP
    OneStr[4]       = OneGad[4]    = IDC
    OneStr[5]       = OneGad[5]    = RS
    OneStr[6]       = OneGad[6]    = WC
    OneStr[7]       = OneGad[7]    = OVF
    OneStr[8]       = OneGad[8]    = COM
    OneStr[9]       = OneGad[9]    = C
    FourStr[0..7]   = RASGad[0..7] = ras[0..7]
    FourStr[8]      = IARGad       = iar
    FourStr[9..12]  = various      = 1st PSG
    FourStr[13..16] = various      = 2nd PSG

    CPU chip... */

    for (i = 0; i < 7; i++)
    {   if (r[i] != monitorr[i] || force)
        {   monitorr[i] = r[i];
            hex1(TwoStr[i], monitorr[i]);
            setmonitor(RGad[i], TwoStr[i]);
    }   }

    if (psu != monitorpsu || force)
    {   if ((psu & PSU_S  ) != (monitorpsu & PSU_S  ) || force)
        {   if (psu & PSU_S  ) OneStr[0][0] = 'S'; else OneStr[0][0] = 's';
            setmonitor(OneGad[0], OneStr[0]);
        }
        if ((psu & PSU_F  ) != (monitorpsu & PSU_F  ) || force)
        {   if (psu & PSU_F  ) OneStr[1][0] = 'F'; else OneStr[1][0] = 'f';
            setmonitor(OneGad[1], OneStr[1]);
        }
        if ((psu & PSU_II ) != (monitorpsu & PSU_II ) || force)
        {   if (psu & PSU_II ) OneStr[2][0] = 'I'; else OneStr[2][0] = 'i';
            setmonitor(OneGad[2], OneStr[2]);
        }
        if ((psu & PSU_SP) != (monitorpsu & PSU_SP) || force)
        {   OneStr[3][0] = '0' + (psu & PSU_SP);
            setmonitor(OneGad[3], OneStr[3]);
        }
        monitorpsu = psu;
    }

    if (psl != monitorpsl || force)
    {   if ((psl & PSL_CC ) != (monitorpsl & PSL_CC ) || force)
        {   strcpy(TwoStr[7], (STRPTR) (ccstring[(psl & 0xC0) >> 6])); // %11000000 -> %00000011
            setmonitor(CCGad, TwoStr[7]);
        }
        if ((psl & PSL_IDC) != (monitorpsl & PSL_IDC) || force)
        {   if (psl & PSL_IDC) OneStr[4][0] = 'D'; else OneStr[4][0] = 'd';
            setmonitor(OneGad[4], OneStr[4]);
        }
        if ((psl & PSL_RS ) != (monitorpsl & PSL_RS ) || force)
        {   if (psl & PSL_RS ) OneStr[5][0] = 'R'; else OneStr[5][0] = 'r';
            setmonitor(OneGad[5], OneStr[5]);
        }
        if ((psl & PSL_WC ) != (monitorpsl & PSL_WC ) || force)
        {   if (psl & PSL_WC ) OneStr[6][0] = 'W'; else OneStr[6][0] = 'w';
            setmonitor(OneGad[6], OneStr[6]);
        }
        if ((psl & PSL_OVF) != (monitorpsl & PSL_OVF) || force)
        {   if (psl & PSL_OVF) OneStr[7][0] = 'O'; else OneStr[7][0] = 'o';
            setmonitor(OneGad[7], OneStr[7]);
        }
        if ((psl & PSL_COM) != (monitorpsl & PSL_COM) || force)
        {   if (psl & PSL_COM) OneStr[8][0] = 'M'; else OneStr[8][0] = 'm';
            setmonitor(OneGad[8], OneStr[8]);
        }
        if ((psl & PSL_C  ) != (monitorpsl & PSL_C  ) || force)
        {   if (psl & PSL_C  ) OneStr[9][0] = 'C'; else OneStr[9][0] = 'c';
            setmonitor(OneGad[9], OneStr[9]);
        }
        monitorpsl = psl;
    }

    for (i = 0; i < 8; i++)
    {   if (ras[i] != monitorras[i] || force)
        {   monitorras[i] = ras[i];
            hex2(FourStr[i], monitorras[i]);
            setmonitor(RASGad[i], FourStr[i]);
    }   }

    if (iar != monitoriar || force)
    {   monitoriar = iar;
        hex2(FourStr[8], monitoriar);
        setmonitor(IARGad, FourStr[8]);
    }

    if (monitorelapsed != elapsed + cycles || force)
    {   monitorelapsed = elapsed + cycles;
        longcomma(monitorelapsed, ClockStr);
        setmonitor(ClockGad, ClockStr);
    }

    if (frames != monitorframes || force)
    {   monitorframes = frames;
        longcomma(monitorframes, FrameStr);
        setmonitor(FrameGad, FrameStr);

        calcrunningtime();
        setmonitor(TimeGad, TimeStr);
    }

    if
    (   machine != ARCADIA
     && machine != INTERTON
     && machine != ELEKTOR
     && machine != ASTROWARS
     && machine != GALAXIA
     && machine != MALZAK
    ) // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS, DOLPHIN, PHUNSY
    {   return;
    }

    if (rast != monitorrast || force)
    {   monitorrast = rast;
        DISCARD stcul_d(RasterStr, (unsigned long) rast);
        setmonitor(RasterGad, RasterStr);
    }

    if (machine == ARCADIA)
    {   for (i = 73; i <= 177; i++)
        {   if
            (   memory[monitoraddr[i]] != monitorcontents[i]
             || force
            )
            {   monitorcontents[i] = memory[monitoraddr[i]];
                hex1(MonitorStr[i], monitorcontents[i]);
                setmonitor(monitorgad[i], MonitorStr[i]);
        }   }

        for (i = 0; i < 4; i++)
        {   if
            (   monitorSPRITEnAX[i] != memory[A_SPRITE0X + (i * 2)] // SPRITEnX
             || monitorSPRITEnAY[i] != memory[A_SPRITE0Y + (i * 2)] // SPRITEnY
             || force
            )
            {   monitorSPRITEnAX[i] =  memory[A_SPRITE0X + (i * 2)];
                monitorSPRITEnAY[i] =  memory[A_SPRITE0Y + (i * 2)];
                hex1(FiveStr[i], monitorSPRITEnAX[i]);
                // FiveStr[i][2] = ','; not needed
                hex1(&FiveStr[i][3], monitorSPRITEnAY[i]);
                setmonitor(SPRITEnAXYGad[i], FiveStr[i]);
    }   }   }
    else
    {   // assert(machine == INTERTON || machine == ELEKTOR || machine == ASTROWARS || machine == GALAXIA || machine == MALZAK);

        for (i = 0; i <= 11; i++)
        {   if (memory[pvibase + monitoraddr[i]] != monitorcontents[i] || force)
            {   monitorcontents[i] = memory[pvibase + monitoraddr[i]];
                hex1(MonitorStr[i], monitorcontents[i]);
                setmonitor(monitorgad[i], MonitorStr[i]);
        }   }

        if (machine == INTERTON || machine == ELEKTOR)
        {   for (i = 12; i <= 18; i++)
            {   if (memory[monitoraddr[i]] != monitorcontents[i] || force)
                {   monitorcontents[i] = memory[monitoraddr[i]];
                    hex1(MonitorStr[i], monitorcontents[i]);
                    setmonitor(monitorgad[i], MonitorStr[i]);
            }   }

            sprites = 4;

            if (machine == INTERTON)
            {   if (memory[monitoraddr[19]] != monitorcontents[19] || force)
                {   monitorcontents[19] = memory[pvibase + monitoraddr[19]];
                    hex1(MonitorStr[19], monitorcontents[19]);
                    setmonitor(monitorgad[19], MonitorStr[19]);
                }
                for (i = 36; i <= 72; i++)
                {   if (memory[monitoraddr[i]] != monitorcontents[i] || force)
                    {   monitorcontents[i] = memory[monitoraddr[i]];
                        hex1(MonitorStr[i], monitorcontents[i]);
                        setmonitor(monitorgad[i], MonitorStr[i]);
            }   }   }
            elif (memmap == MEMMAP_F)
            {   // assert(machine == ELEKTOR);
                for (i = 19; i <= 35; i++)
                {   if (memory[monitoraddr[i]] != monitorcontents[i] || force)
                    {   monitorcontents[i] = memory[monitoraddr[i]];
                        hex1(MonitorStr[i], monitorcontents[i]);
                        setmonitor(monitorgad[i], MonitorStr[i]);
                }   }

                for (i = j = 0; i <= 1; i++, j += 16)
                {   if (monitorPITCHA[i] != ((memory[E_PITCHA1_H + j] * 256) + memory[E_PITCHA1_L + j]) || force)
                    {   monitorPITCHA[i] = (memory[E_PITCHA1_H + j] * 256) + memory[E_PITCHA1_L + j];
                        hex2(FourStr[9 + (i * 4)], monitorPITCHA[i]);
                        setmonitor(PITCHAGad[i], FourStr[9 + (i * 4)]);
                    }
                    if (monitorPITCHB[i] != ((memory[E_PITCHB1_H + j] * 256) + memory[E_PITCHB1_L + j]) || force)
                    {   monitorPITCHB[i] = (memory[E_PITCHB1_H + j] * 256) + memory[E_PITCHB1_L + j];
                        hex2(FourStr[10 + (i * 4)], monitorPITCHB[i]);
                        setmonitor(PITCHBGad[i], FourStr[10 + (i * 4)]);
                    }
                    if (monitorPITCHC[i] != ((memory[E_PITCHC1_H + j] * 256) + memory[E_PITCHC1_L + j]) || force)
                    {   monitorPITCHC[i] = (memory[E_PITCHC1_H + j] * 256) + memory[E_PITCHC1_L + j];
                        hex2(FourStr[11 + (i * 4)], monitorPITCHC[i]);
                        setmonitor(PITCHCGad[i], FourStr[11 + (i * 4)]);
                    }
                    if (monitorPERIOD[i] != ((memory[E_PERIOD1_H + j] * 256) + memory[E_PERIOD1_L + j]) || force)
                    {   monitorPERIOD[i] = (memory[E_PERIOD1_H + j] * 256) + memory[E_PERIOD1_L + j];
                        hex2(FourStr[12 + (i * 4)], monitorPERIOD[i]);
                        setmonitor(PERIODGad[i], FourStr[12 + (i * 4)]);
        }   }   }   }
        else
        {   // assert(machine == MALZAK || machine == ASTROWARS || machine == GALAXIA);

            if (memory[pvibase + monitoraddr[15]] != monitorcontents[15] || force)
            {   monitorcontents[15] = memory[pvibase + monitoraddr[15]];
                hex1(MonitorStr[15], monitorcontents[15]);
                setmonitor(monitorgad[15], MonitorStr[15]);
            }

            switch (machine)
            {
            case ASTROWARS:
                sprites = 4;
            acase GALAXIA:
                for (i = 178; i <= 189; i++) // 2nd PVI
                {   if (memory[0x1600 + monitoraddr[i]] != monitorcontents[i] || force)
                    {   monitorcontents[i] = memory[0x1600 + monitoraddr[i]];
                        hex1(MonitorStr[i], monitorcontents[i]);
                        setmonitor(monitorgad[i], MonitorStr[i]);
                }   }
                for (i = 190; i <= 201; i++) // 3rd PVI
                {   if (memory[0x1700 + monitoraddr[i]] != monitorcontents[i] || force)
                    {   monitorcontents[i] = memory[0x1700 + monitoraddr[i]];
                        hex1(MonitorStr[i], monitorcontents[i]);
                        setmonitor(monitorgad[i], MonitorStr[i]);
                }   }

                sprites = 12;
            acase MALZAK:
                for (i = 178; i <= 189; i++) // 2nd PVI
                {   if (memory[0x1500 + monitoraddr[i]] != monitorcontents[i] || force)
                    {   monitorcontents[i] = memory[0x1500 + monitoraddr[i]];
                        hex1(MonitorStr[i], monitorcontents[i]);
                        setmonitor(monitorgad[i], MonitorStr[i]);
                }   }

                sprites = 8;
        }   }

        for (i = 0; i < sprites; i++)
        {   if
            (   monitorSPRITEnAX[i] != memory[pvibase + pvi_spritedata[i % 4] + 10] // SPRITEnAX
             || monitorSPRITEnAY[i] != memory[pvibase + pvi_spritedata[i % 4] + 12] // SPRITEnAY
             || force
            )
            {   monitorSPRITEnAX[i] =  memory[pvibase + pvi_spritedata[i % 4] + 10];
                monitorSPRITEnAY[i] =  memory[pvibase + pvi_spritedata[i % 4] + 12];
                hex1(FiveStr[i], monitorSPRITEnAX[i]);
                // FiveStr[i][2] = ',';
                hex1(&FiveStr[i][3], monitorSPRITEnAY[i]);
                setmonitor(SPRITEnAXYGad[i], FiveStr[i]);
            }

            if
            (   monitorSPRITEnBX[i] != memory[pvibase + pvi_spritedata[i % 4] + 11] // SPRITEnBX
             || monitorSPRITEnBY[i] != memory[pvibase + pvi_spritedata[i % 4] + 13] // SPRITEnBY
             || force
            )
            {   monitorSPRITEnBX[i] =  memory[pvibase + pvi_spritedata[i % 4] + 11];
                monitorSPRITEnBY[i] =  memory[pvibase + pvi_spritedata[i % 4] + 13];
                hex1(FiveStr[i + 12], monitorSPRITEnBX[i]);
                // FiveStr[i + 12][2] = ',';
                hex1(&FiveStr[i + 12][3], monitorSPRITEnBY[i]);
                setmonitor(SPRITEnBXYGad[i], FiveStr[i + 12]);
            }
            if (i == 3 || i == 7 || i == 11)
            {   pvibase += 0x100;
        }   }

        pvibase -= 0x100 * (sprites / 4);
    }

#ifdef WIN32
    if (machine == ARCADIA || machine == INTERTON || machine == ELEKTOR)
    {   doaxes();
    }
#endif
}

MODULE void configure(FLAG silent)
{   int i;

    if (!silent)
    {   /* This say()ing must be done before changing "machine", because the modal say()
           requester generates WM_PAINT messages, and therefore redrawscreen() gets
           called. */
        machine = oldmachine;
        if (known[whichgame].dump == DUMP_BAD)
        {   say((STRPTR) GetCatalogStr(CatalogPtr,
                MSG_ENGINE_BADDUMP,
                "This is a known bad dump."
            ),
            TRUE);
        } elif (known[whichgame].dump == DUMP_SUSPECT)
        {   say((STRPTR) GetCatalogStr(CatalogPtr,
                MSG_ENGINE_SUSPECTDUMP,
                "This is a suspect dump."
            ),
            TRUE);
    }   }

    machine        = (int)   known[whichgame].machine;
    analog         = (ULONG) known[whichgame].analog;
    if (machine == INSTRUCTOR)
    {   s_io          = (int) known[whichgame].sensitivity;
        s_intselector = (int) known[whichgame].swapped;
        s_intdir      = (int) known[whichgame].flagline;
    } else
    {   sensitivity = (int)  known[whichgame].sensitivity;
        swapped     = (ULONG) known[whichgame].swapped;
        flagline    = (ULONG) known[whichgame].flagline;
    }

    if (known[whichgame].demultiplex)
    {   demultiplex = TRUE;
        flips       = known[whichgame].demultiplex & 0x0F;
        protect[0]  = known[whichgame].demultiplex & 0x80;
        protect[1]  = known[whichgame].demultiplex & 0x40;
        protect[2]  = known[whichgame].demultiplex & 0x20;
        protect[3]  = known[whichgame].demultiplex & 0x10;
    } else
    {   demultiplex = FALSE;
        flips       = 1; // defaults to 1:1
        protect[0]  =
        protect[1]  =
        protect[2]  =
        protect[3]  = FALSE;
    }
    memmap         = (int)   known[whichgame].memmap;
    downframes     = (ULONG) known[whichgame].downframes;
    totalframes    = (ULONG) known[whichgame].totalframes;
    // upframes    = totalframes - downframes;
    key1           = (int)   known[whichgame].key1;
    key2           = (int)   known[whichgame].key2;
    key3           = (int)   known[whichgame].key3;
    key4           = (int)   known[whichgame].key4;
    for (i = 0; i < 4; i++)
    {   p1bgcol[i] = (int)   known[whichgame].p1bgcol[i];
        p2bgcol[i] = (int)   known[whichgame].p2bgcol[i];
    }
    for (i = 0; i < 6; i++)
    {   p1sprcol[i] = (int)  known[whichgame].p1sprcol[i];
        p2sprcol[i] = (int)  known[whichgame].p2sprcol[i];
    }
    paddleup       = (int)   known[whichgame].up;
    paddledown     = (int)   known[whichgame].down;
    paddleleft     = (int)   known[whichgame].left;
    paddleright    = (int)   known[whichgame].right;
    ggnode         =         known[whichgame].ggnode;
    starfield      = (int)   known[whichgame].starfield;
    firstmnemonic  = (int)   known[whichgame].firstmnemonic;
    lastmnemonic   = (int)   known[whichgame].lastmnemonic;
    make_stars();

    if (usehacks && (machine == INTERTON || machine == ELEKTOR))
    {   latestsprite   = (int) known[whichgame].latestsprite;
        cpl            = (int) known[whichgame].cpl;
        vblank         = (int) known[whichgame].vblank;
        forceinterrupt = (int) known[whichgame].forceinterrupt;
    }
    whichoverlay   = (int)   known[whichgame].whichoverlay;

    if (machine != INSTRUCTOR && machine != PIPBUG && machine != CD2650 && machine != PHUNSY)
    {   generate_overlaytext();
    }
    if (machine == PHUNSY && known[whichgame].kind == KIND_BIN) // not for PGMs!
    {   startaddr_h = (UBYTE) (known[whichgame].startaddr / 256); // needed!
        startaddr_l = (UBYTE) (known[whichgame].startaddr % 256); // needed!
    }

#ifdef VERBOSE
    T1; textpen(TEXTPEN_VERBOSE); zprintf("Autosensed game #%d.\n", whichgame); T2;
#endif

    if (whichgame == GOLFPOS)
    {   // assert(machine == ARCADIA);
        // assert(romsize == 6144);
        golf = TRUE;
    } else
    {   golf = FALSE;
    }
    if (undither && (whichgame == ESCAPEPOS || whichgame == ROBOTKILLERPOS))
    {   undithernow = TRUE;
    } else
    {   undithernow = FALSE;
    }
    if (whichgame == MUSIC1POS || whichgame == MUSIC2POS)
    {   music = TRUE;
    } else
    {   music = FALSE;
    }
    if (whichgame == PRINTERPOS)
    {   printer = TRUE;
    } else
    {   printer = FALSE;
    }
    if (whichgame == P_MORSE1POS || whichgame == P_MORSE2POS)
    {   p_morse = TRUE;
    } else
    {   p_morse = FALSE;
    }
    if (whichgame == LIFE1POS || whichgame == LIFE2POS)
    {   life = TRUE;
    } else
    {   life = FALSE;
    }
    if (whichgame == PIANOPOS)
    {   piano = TRUE;
    } else
    {   piano = FALSE;
    }
    if (whichgame == FIFTYTHEMEPOS)
    {   fiftytheme = TRUE;
    } else
    {   fiftytheme = FALSE;
    }
    if (whichgame == BELMACH1POS || whichgame == BELMACH2POS)
    {   belmach = TRUE;
    } else
    {   belmach = FALSE;
    }
    if (whichgame == BELMACH0POS)
    {   belmach0 = TRUE;
    } else
    {   belmach0 = FALSE;
    }

    generate_autotext();
}

EXPORT void generate_autotext(void)
{   int thesize;

    // assert(machine != CHAOS);

    charset = CHARSET_WIN;

    switch (machine)
    {
    case ARCADIA:
        strcpy(autotext[AUTOLINE_MACHINE], "Emerson Arcadia 2001, type \"");
        if (memmap == MEMMAP_G)
        {   strcat(autotext[AUTOLINE_MACHINE], "G\" (Emerson)");
        } elif (memmap == MEMMAP_H)
        {   strcat(autotext[AUTOLINE_MACHINE], "H\" (Tele-Fever)");
        } else
        {   // assert(memmap == MEMMAP_I);
            strcat(autotext[AUTOLINE_MACHINE], "I\" (Palladium)");
        }
    acase INTERTON:
        strcpy(autotext[AUTOLINE_MACHINE], "Interton VC 4000, type \"");
        if (memmap == MEMMAP_A)
        {   strcat(autotext[AUTOLINE_MACHINE], "A\" (2K ROM + 0");
        } elif (memmap == MEMMAP_B)
        {   strcat(autotext[AUTOLINE_MACHINE], "B\" (4K ROM + 0");
        } elif (memmap == MEMMAP_C)
        {   strcat(autotext[AUTOLINE_MACHINE], "C\" (4K ROM + 1");
        } else
        {   // assert(memmap == MEMMAP_D);
            strcat(autotext[AUTOLINE_MACHINE], "D\" (6K ROM + 1");
        }
        strcat(autotext[AUTOLINE_MACHINE], "K RAM)");
    acase ELEKTOR:
        strcpy(autotext[AUTOLINE_MACHINE], "Elektor TV Games Computer (");
        if (memmap == MEMMAP_E)
        {   strcat(autotext[AUTOLINE_MACHINE], "basic)");
        } else
        {   // assert(memmap == MEMMAP_F);
            strcat(autotext[AUTOLINE_MACHINE], "expanded)");
        }
    acase PIPBUG:
        strcpy(autotext[AUTOLINE_MACHINE], "PIPBUG-based machine (");
        if (memmap == MEMMAP_J)
        {   strcat(autotext[AUTOLINE_MACHINE], "EA 77up2)");
        } elif (memmap == MEMMAP_K)
        {   strcat(autotext[AUTOLINE_MACHINE], "Signetics ABC)");
        } elif (memmap == MEMMAP_L)
        {   strcat(autotext[AUTOLINE_MACHINE], "basic (1K) EA 78up5)");
        } elif (memmap == MEMMAP_M)
        {   strcat(autotext[AUTOLINE_MACHINE], "expanded (7K) EA 78up5)");
        } else
        {   // assert(memmap == MEMMAP_P);
            strcat(autotext[AUTOLINE_MACHINE], "11K EA 78up5 with EPROM)");
        }
    acase INSTRUCTOR:
        strcpy(autotext[AUTOLINE_MACHINE], "Signetics Instructor 50 (");
        if (memmap == MEMMAP_N)
        {   strcat(autotext[AUTOLINE_MACHINE], "basic)");
        } else
        {   // assert(memmap == MEMMAP_O);
            strcat(autotext[AUTOLINE_MACHINE], "expanded)");
        }
    acase CD2650:
        strcpy(autotext[AUTOLINE_MACHINE], "Central Data 2650");
    acase ASTROWARS:
    case GALAXIA:
    case MALZAK:
        strcpy
        (   autotext[AUTOLINE_MACHINE],
            GetCatalogStr(CatalogPtr,
                MSG_COINOP,
                "Coin-op"
        )   );
        generate_overlaytext();
    acase DOLPHIN:
        strcpy(autotext[AUTOLINE_MACHINE], "Dolphin");
    acase PHUNSY:
        strcpy(autotext[AUTOLINE_MACHINE], "PHUNSY");
    }

    switch (memmap)
    {
    case MEMMAP_ASTROWARS:
        thesize = (12 * KILOBYTE) + 512; // 12.5K
        strcpy(autotext[AUTOLINE_GAMENAME ], "Astro Wars");
        strcpy(autotext[AUTOLINE_CREDITS  ], "Zaccaria, 1980");
        strcpy(autotext[AUTOLINE_REFERENCE], "-");
        strcpy
        (   autotext[AUTOLINE_COMPLANG],
            GetCatalogStr(CatalogPtr,
                MSG_MACHINECODE,   "Machine code"
        )   );
    acase MEMMAP_GALAXIA:
        thesize = (12 * KILOBYTE) + 512; // 12.5K
        strcpy(autotext[AUTOLINE_GAMENAME ], "Galaxia");
        strcpy(autotext[AUTOLINE_CREDITS  ], "Zaccaria, 1979");
        strcpy(autotext[AUTOLINE_REFERENCE], "-");
        strcpy
        (   autotext[AUTOLINE_COMPLANG],
            GetCatalogStr(CatalogPtr,
                MSG_MACHINECODE,   "Machine code"
        )   );
    acase MEMMAP_MALZAK1:
        thesize = 8704; // 8.5K
        strcpy(autotext[AUTOLINE_GAMENAME ], "Malzak 1");
        strcpy(autotext[AUTOLINE_CREDITS  ], "Kitronix, 1980?");
        strcpy(autotext[AUTOLINE_REFERENCE], "-");
        strcpy
        (   autotext[AUTOLINE_COMPLANG],
            GetCatalogStr(CatalogPtr,
                MSG_MACHINECODE,   "Machine code"
        )   );
    acase MEMMAP_MALZAK2:
        thesize = 8192; // 8  K
        strcpy(autotext[AUTOLINE_GAMENAME ], "Malzak 2");
        strcpy(autotext[AUTOLINE_CREDITS  ], "Kitronix, 1980?");
        strcpy(autotext[AUTOLINE_REFERENCE], "-");
        strcpy
        (   autotext[AUTOLINE_COMPLANG],
            GetCatalogStr(CatalogPtr,
                MSG_MACHINECODE,   "Machine code"
        )   );
    adefault:
#ifdef WIN32
        if (language == LANGUAGE_RUS && whichgame == TETRISPOS)
        {   strcpy(autotext[AUTOLINE_GAMENAME], "");
        } else
#endif
            strcpy(autotext[AUTOLINE_GAMENAME], known[whichgame].name);

        strcpy(autotext[AUTOLINE_CREDITS  ], known[whichgame].credits);
        strcpy(autotext[AUTOLINE_REFERENCE], known[whichgame].reference);

        switch (known[whichgame].complang)
        {
        case CL_MISC:
            strcpy
            (   autotext[AUTOLINE_COMPLANG],
                GetCatalogStr(CatalogPtr,
                    MSG_MISCELLANEOUS, "Miscellaneous"
            )   );
        acase CL_MC:
            strcpy
            (   autotext[AUTOLINE_COMPLANG],
                GetCatalogStr(CatalogPtr,
                    MSG_MACHINECODE,   "Machine code"
            )   );
        acase CL_MWB:
            strcpy(autotext[AUTOLINE_COMPLANG], "MicroWorld BASIC");
        acase CL_FORTH:
            strcpy(autotext[AUTOLINE_COMPLANG], "FORTH");
        acase CL_12KB:
            strcpy(autotext[AUTOLINE_COMPLANG], "12K BASIC");
        acase CL_MB:
            strcpy(autotext[AUTOLINE_COMPLANG], "Micro BASIC");
        adefault: // should never happen
            strcpy(autotext[AUTOLINE_COMPLANG], "?");
        }

        thesize = known[whichgame].size;
    }

    switch (thesize % KILOBYTE)
    {
    case 0:
        sprintf
        (   autotext[AUTOLINE_SIZE],
            GetCatalogStr(CatalogPtr,
                MSG_ENGINE_KILOBYTES,
                "%dK"
            ),
            (long int) (thesize / KILOBYTE)
        );
    acase 256:
        if (thesize == 256)
        {   sprintf
            (   autotext[AUTOLINE_SIZE],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_KILOBYTES,
                    "K" // 1/4
            )   );
        } else
        {   sprintf
            (   autotext[AUTOLINE_SIZE],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_KILOBYTES,
                    "%dK" // 1/4
                ),
                (long int) (thesize / KILOBYTE)
            );
        }
    acase 512:
        if (thesize == 512)
        {   sprintf
            (   autotext[AUTOLINE_SIZE],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_KILOBYTES,
                    "K" // 1/2
            )   );
        } else
        {   sprintf
            (   autotext[AUTOLINE_SIZE],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_KILOBYTES,
                    "%dK" // 1/2
                ),
                (long int) (thesize / KILOBYTE)
            );
        }
    acase 768:
        if (thesize == 768)
        {   sprintf
            (   autotext[AUTOLINE_SIZE],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_KILOBYTES,
                    "K" // 3/4
            )   );
        } else
        {   sprintf
            (   autotext[AUTOLINE_SIZE],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_KILOBYTES,
                    "%dK" // 3/4
                ),
                (long int) (thesize / KILOBYTE)
            );
        }
    adefault:
        sprintf
        (   autotext[AUTOLINE_SIZE],
            GetCatalogStr(CatalogPtr,
                MSG_ENGINE_BYTES,
                "%d bytes"
            ),
            (long int) thesize
        );
    }

    if
    (   machine == ELEKTOR
     || machine == PIPBUG
     || machine == INSTRUCTOR
     || machine == CD2650
     || machine == PHUNSY
    )
    {   sprintf
        (   autotext[AUTOLINE_STARTADDR],
            "$%X",
            (unsigned int) ((startaddr_h * 256) + startaddr_l)
        );
    } else
    {   /* assert(
            machine == ARCADIA
         || machine == INTERTON
         || machine == ASTROWARS
         || machine == GALAXIA
         || machine == MALZAK
         || machine == DOLPHIN
        ); */
        strcpy(autotext[AUTOLINE_STARTADDR], "$0");
    }

    if
    (   machine == PIPBUG
     || machine == INSTRUCTOR
     || machine == CD2650
     || machine == DOLPHIN
     || machine == PHUNSY
    )
    {   strcpy
        (   autotext[AUTOLINE_AUTOFIRE],
            GetCatalogStr(CatalogPtr,
                MSG_ENGINE_NOTAPPLICABLE,
                "n/a"
        )   );
        strcpy
        (   autotext[AUTOLINE_CONTROLLERS],
            GetCatalogStr(CatalogPtr,
                MSG_ENGINE_NOTAPPLICABLE,
                "n/a"
        )   );
    } else
    {   /* assert
        (   machine == ARCADIA
         || machine == INTERTON
         || machine == ELEKTOR
         || machine == ASTROWARS
         || machine == GALAXIA
         || machine == MALZAK
        ); */

        sprintf
        (   autotext[AUTOLINE_AUTOFIRE],
            "%d:%d",
            (int) downframes,
            (int) (totalframes - downframes)
        );

        if (analog)
        {   strcpy
            (   autotext[AUTOLINE_CONTROLLERS],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_ANALOG,
                    "Analog"
            )   );
        } else
        {   strcpy
            (   autotext[AUTOLINE_CONTROLLERS],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_DIGITAL,
                    "Digital"
            )   );
        }
        strcat(autotext[AUTOLINE_CONTROLLERS], ", ");
        if (swapped)
        {   strcat
            (   autotext[AUTOLINE_CONTROLLERS],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_SWAPPED,
                    "swapped"
            )   );
        } else
        {   strcat
            (   autotext[AUTOLINE_CONTROLLERS],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_UNSWAPPED,
                    "unswapped"
            )   );
        }
        strcat(autotext[AUTOLINE_CONTROLLERS], " (");
        switch (sensitivity)
        {
        case SENSITIVITY_VLOW:
            strcat
            (   autotext[AUTOLINE_CONTROLLERS],
                GetCatalogStr(CatalogPtr,
                MSG_ENGINE_VERYLOW,
                "very low"
            )   );
        acase SENSITIVITY_LOW:
            strcat
            (   autotext[AUTOLINE_CONTROLLERS],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_LOW,
                    "low"
            )   );
        acase SENSITIVITY_MEDIUM:
            strcat
            (   autotext[AUTOLINE_CONTROLLERS],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_MEDIUM,
                    "medium"
            )   );
        acase SENSITIVITY_HIGH:
            strcat
            (   autotext[AUTOLINE_CONTROLLERS],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_HIGH,
                    "high"
            )   );
        acase SENSITIVITY_VHIGH:
            strcat
            (   autotext[AUTOLINE_CONTROLLERS],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_VERYHIGH,
                    "very high"
            )   );
        adefault:
            strcat(autotext[AUTOLINE_CONTROLLERS], "?"); // should never happen
        }
        strcat(autotext[AUTOLINE_CONTROLLERS], " ");
        strcat
        (   autotext[AUTOLINE_CONTROLLERS],
            GetCatalogStr(CatalogPtr,
                MSG_ENGINE_SENSITIVITY,
                "sensitivity"
        )   );
        strcat(autotext[AUTOLINE_CONTROLLERS], ")");
    }

    if (machine == ARCADIA)
    {   if (demultiplex)
        {   strcpy
            (   autotext[AUTOLINE_DEMULTIPLEX],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_YES,
                    "Yes"
            )   );
        } else
        {   strcpy
            (   autotext[AUTOLINE_DEMULTIPLEX],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_NO,
                    "No"
            )   );
        }
        strcat(autotext[AUTOLINE_DEMULTIPLEX], " ");
        sprintf
        (   &autotext[AUTOLINE_DEMULTIPLEX][strlen(autotext[AUTOLINE_DEMULTIPLEX])],
            GetCatalogStr(CatalogPtr,
                MSG_ENGINE_RETAIN,
                "(retain previous %d frame(s))"
            ),
            flips
        );
    } else
    {   strcpy
        (   autotext[AUTOLINE_DEMULTIPLEX],
            GetCatalogStr(CatalogPtr,
                MSG_ENGINE_NOTAPPLICABLE,
                "n/a"
        )   );
    }

    switch (machine)
    {
    case ARCADIA:
    case INTERTON:
    case ELEKTOR:
    case CD2650:
        if (flagline)
        {   strcpy
            (   autotext[AUTOLINE_FLAGLINE],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_ON2,
                    "On"
            )   );
        } else
        {   strcpy
            (   autotext[AUTOLINE_FLAGLINE],
                GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_OFF2,
                    "Off"
            )   );
        }
    adefault:
        /* assert
        (   machine == PIPBUG
         || machine == INSTRUCTOR
         || machine == ASTROWARS
         || machine == GALAXIA
         || machine == MALZAK
         || machine == DOLPHIN
         || machine == PHUNSY
        ); */
        strcpy
        (   autotext[AUTOLINE_FLAGLINE],
            GetCatalogStr(CatalogPtr,
               MSG_ENGINE_NOTAPPLICABLE,
               "n/a"
        )   );
    }

    charset = CHARSET_DOS;
}

MODULE void unautosense(void)
{   int i;

    downframes     = 4;
    totalframes    = 16;
    latestsprite   = PVI_LATESTSPRITE;
    if (machine != MALZAK && machine != ASTROWARS && machine != GALAXIA)
    {   cpl         = PVI_CPL;
        starfield   = 16;
        vblank      = PVI_VBLANK;
        key1        = 2;
        key2        = 1;
        key3        = 3;
        key4        = 0;
        for (i = 0; i < 6; i++)
        {   p1sprcol[i] =
            p2sprcol[i] = 0;
        }
        for (i = 0; i < 4; i++)
        {   p1bgcol[i]  =
            p2bgcol[i]  = 0;
    }   }
    forceinterrupt =  0;
    whichgame      = -1;
    sensitivity    = SENSITIVITY_DEFAULT;
    paddleup       =
    paddledown     =
    paddleleft     =
    paddleright    = -1;
    firstmnemonic  =
    lastmnemonic   = -1;
    flips          =  1;
    protect[0]     =
    protect[1]     =
    protect[2]     =
    protect[3]     = FALSE;
    // we deliberately leave region, flagline, s_io, s_intselector, s_intdir alone

    // found       = FALSE;

    golf           =
    undithernow    =
    music          =
    p_morse        =
    life           =
    printer        =
    piano          =
    fiftytheme     =
    belmach        =
    belmach0       = FALSE;

    // Generally we default to the most powerful memory map of a machine.
    if (machine != MALZAK)
    {   memmap = machines[machine].memmap;
    }

    ggnode         = "";
    whichoverlay   = machines[machine].overlay;
    for (i = 0; i < GAMEINFOLINES; i++)
    {   strcpy(autotext[i], "?");
}   }

EXPORT void project_reset(FLAG keepalive)
{   if (!keepalive)
    {   NetPlay(NET_OFF);
        updatemenu(MENUITEM_SERVER);
        updatemenu(MENUITEM_CLIENT);
    }

    macro_stop();
    engine_reset();
}

EXPORT void dumprow(int i)
{   PERSIST int  x;
    PERSIST TEXT rowchars[16 + 1];

    rowchars[16] = EOS;

    // this assumes that the console is already open,
    // and it doesn't close it either

    switch (machine)
    {
    case ARCADIA:
        for (x = 0; x < 16; x++)
        {   rowchars[x] = arcadia_chars[memory[i + x] & 0x3F];
        }
    acase PIPBUG:
    case CHAOS:
    case PHUNSY:
        for (x = 0; x < 16; x++)
        {   if (memory[i + x] <= 31) rowchars[x] = '.'; else rowchars[x] = memory[i + x];
        }
    acase CD2650:
        for (x = 0; x < 16; x++)
        {   rowchars[x] = cd2650_chars[memory[i + x] & 0x7F];
        }
    acase ASTROWARS:
        for (x = 0; x < 16; x++)
        {   rowchars[x] = astrowars_chars[memory[i + x]];
        }
    acase GALAXIA:
        for (x = 0; x < 16; x++)
        {   rowchars[x] = galaxia_chars[memory[i + x] & 0x7F];
        }
    acase MALZAK:
        for (x = 0; x < 16; x++)
        {   if ((memory[i + x] & 0x7F) <= 31) rowchars[x] = '.'; else rowchars[x] = memory[i + x] & 0x7F;
        }
    adefault: // eg. INTERTON, ELEKTOR, INSTRUCTOR, DOLPHIN
        rowchars[0] = EOS;
    }

    zprintf
    (   "$%04X: %02X %02X %02X %02X %02X %02X %02X %02X  %02X %02X %02X %02X %02X %02X %02X %02X %s\n",
        i,
        memory[i     ],
        memory[i +  1],
        memory[i +  2],
        memory[i +  3],
        memory[i +  4],
        memory[i +  5],
        memory[i +  6],
        memory[i +  7],
        memory[i +  8],
        memory[i +  9],
        memory[i + 10],
        memory[i + 11],
        memory[i + 12],
        memory[i + 13],
        memory[i + 14],
        memory[i + 15],
        rowchars
    );
}

EXPORT void debug_cli(FLAG showcpu)
{   PERSIST   TEXT  userinput[MAX_PATH + 80 + 1]; // PERSISTent so as not to overflow the stack
    TRANSIENT FLAG  done       = FALSE;
#ifdef WIN32
    TRANSIENT TEXT  ch;
    TRANSIENT FLAG  ignorenext = FALSE,
                    innerdone;
    TRANSIENT int   i;
#endif

    if (!cli)
    {   return;
    }

    if (traceverbose && showcpu)
    {   view_cpu(FALSE);
        view_next();
    }

    while (!done)
    {   update_monitor(TRUE); // must be forced because user may have changed memory viewer region

        T1; textpen(TEXTPEN_CLIOUTPUT); // some function may have closed it
        zprintf(">");
        textpen(TEXTPEN_CLIINPUT);
        fflush(stdout); // maybe unnecessary?
        fflush(stdin);  // maybe unnecessary?

#ifdef WIN32
        activateconsole();
#endif

        lockgui();

#ifdef WIN32
        capslock_off();

        innerdone = FALSE;
        i = 0;
        while (!innerdone)
        {   checkinput();
            if (_kbhit())
            {   ch = (TEXT) _getche();
                if (ignorenext)
                {   ignorenext = FALSE;
                    printf("\b \b"); // not zprintf()!
                } elif (ch == 8) // backspace
                {   if (i)
                    {   printf(" \b"); // not zprintf()! really "\b \b" because _kbit() prints the first "\b"
                        i--;
                    } else
                    {   printf(">"); // not zprintf()!
                }   }
                elif (ch == 224) // cursor
                {   ignorenext = TRUE;
                } else
                {   userinput[i++] = ch;
                    if (ch == CR)
                    {   userinput[i++] = 0;
                        innerdone = TRUE;
        }   }   }   }
        printf("%c", LF); // not zprintf()!

        if (LogfileHandle) // *not* if (logfile), because user might have recently turned on log-to-file via menu
        {   if (strlen(userinput) > 1)
            {   DISCARD fwrite(userinput, strlen(userinput) - 1, 1, LogfileHandle);
            }
            DISCARD fwrite("\n", strlen("\n"), 1, LogfileHandle);
        }
#endif
#ifdef AMIGA
        DISCARD gets(userinput);

        if (logfile)
        {   DISCARD fwrite(userinput, strlen(userinput), 1, LogfileHandle);
            DISCARD fwrite("\n", strlen("\n"), 1, LogfileHandle);
        }
#endif

        unlockgui();
        clearmsgs();
        clearkybd();
        update_monitor(FALSE);
        checkinput();

        arg1[0] =
        arg2[0] =
        arg3[0] =
        arg4[0] =
        arg5[0] = 0;
        DISCARD sscanf(userinput, "%s %s %s %s %s", arg1, arg2, arg3, arg4, arg5);

#ifdef VERBOSE
        zprintf("\"%s\", \"%s\", \"%s\", \"%s\", \"%s\", \"%s\"\n", userinput, arg1, arg2, arg3, arg4, arg5);
#endif

        textpen(TEXTPEN_CLIOUTPUT);
        done = debug_command();
    }

    T2;
    // We want to set focus to the Ami/WinArcadia window, but this doesn't
    // work (on IBM-PC at least).
    clearmsgs();
}

EXPORT FLAG debug_command(void)
{   TRANSIENT int   address1, // not UWORD!
                    address2, // not UWORD!
                    address3, // not UWORD!
                    localsize,
                    i, j;
    TRANSIENT FLAG  done = FALSE,
                    ok;
    TRANSIENT FILE* TheLocalHandle; // LocalHandle is a variable of winbase.h
    PERSIST   UBYTE trainval;

    if (!arg1[0] || !stricmp(arg1, "S"))
    {   docommand(MENUITEM_STEP);
        done = TRUE;
    } elif (!stricmp(arg1, "O"))
    {   done = TRUE;
        docommand(MENUITEM_STEPOVER);
    } elif (!stricmp(arg1, "ALL"))
    {   view_allmemory();
    } elif (!stricmp(arg1, "BASE"))
    {   if (arg3[0])
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_BASE, "Usage: BASE [!|@|$|%%]\n\n"));
        } elif (arg2[0])
        {   if (!strcmp(arg2, "!"))
            {   base = BASE_DECIMAL;
                zprintf
                (   "%s %s (!).\n\n",
                    LLL, MSG_WRITEBASE,   "Set default input base to"),
                    LLL, MSG_DECIMAL,     "decimal")
                );
            } elif (!strcmp(arg2, "$"))
            {   base = BASE_HEX;
                zprintf
                (   "%s %s ($).\n\n",
                    LLL, MSG_WRITEBASE,   "Set default input base to"),
                    LLL, MSG_HEXADECIMAL, "hexadecimal")
                );
            } elif (!strcmp(arg2, "%"))
            {   base = BASE_BINARY;
                zprintf
                (   "%s %s (%%).\n\n",
                    LLL, MSG_WRITEBASE,   "Set default input base to"),
                    LLL, MSG_BINARY,      "binary")
                );
            } elif (!strcmp(arg2, "@"))
            {   base = BASE_OCTAL;
                zprintf
                (   "%s %s (@).\n\n",
                    LLL, MSG_WRITEBASE,   "Set default input base to"),
                    LLL, MSG_OCTAL,       "octal")
                );
            } else
            {   zprintf(LLL, MSG_USAGE_BASE, "Usage: BASE [!|@|$|%%]\n\n"));
        }   }
        else
        {   zprintf
            (   "%s ",
                LLL, MSG_READBASE, "Current default input base is")
            );
            switch (base)
            {
            case BASE_DECIMAL:
                zprintf("%s (!).\n\n",  LLL, MSG_DECIMAL,     "decimal"    ));
            acase BASE_HEX:
                zprintf("%s ($).\n\n",  LLL, MSG_HEXADECIMAL, "hexadecimal"));
            acase BASE_BINARY:
                zprintf("%s (%%).\n\n", LLL, MSG_BINARY,      "binary"     ));
            acase BASE_OCTAL:
                zprintf("%s (@).\n\n",  LLL, MSG_OCTAL,       "octal"      ));
    }   }   }
    elif (!stricmp(arg1, "BC"))
    {   if (arg4[0])
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_BC, "Usage: BC [<address1> [<address2>]]\n\n"));
        } elif (arg3[0])
        {   address1 = friendly_to_number(arg2);
            address2 = friendly_to_number(arg3);
            if
            (   address1 != OUTOFRANGE
             && address2 != OUTOFRANGE
             && address1 <= address2
            )
            {   for (; address1 <= address2; address1++)
                {   memflags[address1] &= ~(BREAKPOINT);
                }
                for (address1 = 0; address1 < 32768; address1++)
                {   if (memflags[address1] & BREAKPOINT)
                    {   bps = FALSE;
                        break; // for speed
                }   }
                updatemenu(MENUITEM_CLEARBP );
                updatemenu(MENUITEM_CLEARBPS);
                update_memory(TRUE);
                zprintf("OK.\n\n"); // localize this!
        }   }
        elif (arg2[0])
        {   address1 = friendly_to_number(arg2);
            DISCARD bp_clear(address1);
        } else
        {   DISCARD bp_clearall();
    }   }
    elif (!stricmp(arg1, "BL"))
    {   ok = FALSE;
        for (i = 0; i <= 32767; i++)
        {   if (memflags[i] & BREAKPOINT)
            {   DISCARD getfriendly(i);
                zprintf("%s\n", friendly);
                ok = TRUE;
        }   }
        if (ok)
        {   zprintf("\n");
        } else
        {   zprintf
            (   "%s.\n\n",
                GetCatalogStr(CatalogPtr,
                    MSG_NONE,
                    "None"
            )   );
    }   }
    elif (!stricmp(arg1, "BP"))
    {   if (arg4[0])
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_BP, "Usage: BP [<address1> [<address2>]]\n\n"));
        } elif (arg3[0])
        {   address1 = friendly_to_number(arg2);
            address2 = friendly_to_number(arg3);
            if
            (   address1 != OUTOFRANGE
             && address2 != OUTOFRANGE
             && address1 <= address2
            )
            {   for (; address1 <= address2; address1++)
                {   memflags[address1] |= BREAKPOINT;
                    bps = TRUE;
                }
                updatemenu(MENUITEM_CLEARBP );
                updatemenu(MENUITEM_CLEARBPS);
                update_memory(TRUE);
                zprintf("OK.\n\n"); // localize this!
        }   }
        elif (arg2[0])
        {   address1 = friendly_to_number(arg2);
            DISCARD bp_add(address1);
        } else
        {   debug_set(KIND_BP);
    }   }
    elif (!stricmp(arg1, "BRAM"))
    {   switch (machine)
        {
        case ELEKTOR:
        case PIPBUG:
        case INSTRUCTOR:
        case CD2650:
        case PHUNSY:
            view_bram();
        adefault:
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_UNAVAILABLE,
                    "Command is unavailable.\n\n"
            )   );
    }   }
    elif (  !stricmp(arg1, "COLOURS"))
    {   if (machine != ASTROWARS && machine != GALAXIA)
        {   zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_UNAVAILABLE,
                    "Command is unavailable.\n\n"
            )   );
        } else
        {   g_viewcolours();
    }   }
    elif (!stricmp(arg1, "COVER"))
    {   view_coverage();
    } elif (!stricmp(arg1, "CPU"))
    {   view_cpu(TRUE);
    } elif (!stricmp(arg1, "CRAM"))
    {   if
        (   machine == INTERTON
         && (memmap == MEMMAP_C || memmap == MEMMAP_D)
        )
        {   view_cram();
        } else
        {   zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_UNAVAILABLE,
                    "Command is unavailable.\n\n"
            )   );
    }   }
    elif (!stricmp(arg1, "D") || !stricmp(arg1, "PEEK"))
    {   if (!arg2[0])
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_D, "Usage: D|PEEK <address1> [<address2>]\n\n"));
        } else
        {   address1 = friendly_to_number(arg2);
            if (arg3[0])
            {   address2 = friendly_to_number(arg3);
            } else
            {   address2 = address1 + 255;
            }
            if (address2 % 16 != 15) // round up
            {   address2 = ((address2 / 16) * 16) + 15;
            }
            if
            (   address1 >  address2
             || address1 == OUTOFRANGE
             || address2 == OUTOFRANGE
            )
            {   zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_INVALIDRANGE,
                        "Valid range is $0000..$7FFF!\n\n"
                )   );
            } else
            {   for (i = address1; i <= address2; i += 16)
                {   dumprow(i);
                }
                zprintf("\n");
    }   }   }
    elif (!stricmp(arg1, "DIS"))
    {   if (arg2[0])
        {   address1 = friendly_to_number(arg2);
        } else
        {   address1 = iar;
        }
        if (arg3[0])
        {   address2 = friendly_to_number(arg3);
        } else
        {   if (traceverbose)
            {   address2 = address1 + 11 - 1;
            } else
            {   address2 = address1 + 23 - 1;
// up to 23 lines of disassembly (to allow for blank line and prompt),
//  plus labels (up to 23 more lines)
// (IBM-PC console has 25 rows by 80 columns)
        }   }
        if
        (   address1 >  address2
         || address1 == OUTOFRANGE
         || address2 == OUTOFRANGE
        )
        {   zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_INVALIDRANGE,
                    "Valid range is $0000..$7FFF!\n\n"
            )   );
        } else
        {   disassemble(address1, address2);
            zprintf("\n");
    }   }
    elif (!stricmp(arg1, "E") || !stricmp(arg1, "POKE"))
    {   jump = FALSE;
        if (arg2[0] && arg3[0])
        {   change_read(arg2); // sets tolimit, fromnum, etc.
            tonum = readbase(arg3);
            change_write(arg2, FALSE);
        } elif (!arg2[0] && !arg3[0])
        {   debug_edit();
        } else
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_E, "Usage: E|POKE [<address> <value>]\n\n"));
    }   }
    elif (!stricmp(arg1, "FILL"))
    {   if (arg2[0])
        {   address1 = friendly_to_number(arg2);
            address2 = friendly_to_number(arg3);
            if
            (   address1 >  address2
             || address1 == OUTOFRANGE
             || address2 == OUTOFRANGE
            )
            {   zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_INVALIDRANGE,
                        "Valid range is $0000..$7FFF!\n\n"
                )   );
            } else
            {   tonum = readbase(arg4);
                if (tonum >= -128 && tonum <= 255)
                {   for (i = address1; i <= address2; i++)
                    {   memory[i] = (UBYTE) tonum;
                    }
                    zprintf
                    (   GetCatalogStr(CatalogPtr,
                            MSG_FILLED,
                            "Filled $%X..$%X with $%X.\n\n"
                        ),
                        address1,
                        address2,
                        tonum
                    );
                } else
                {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_F, "Usage: FILL <address1> <address2> <value>\n\n"));
        }   }   }
        else
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_F, "Usage: FILL <address1> <address2> <value>\n\n"));
    }   }
    elif
    (   !stricmp(arg1, "FIND")
     || !stricmp(arg1, "TROVA") // Italian for Find
    )
    {   if (arg4[0])
        {   address1 = friendly_to_number(arg2);
            address2 = friendly_to_number(arg3);
            if
            (   address1 >  address2
             || address1 == OUTOFRANGE
             || address2 == OUTOFRANGE
            )
            {   zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_INVALIDRANGE,
                        "Valid range is $0000..$7FFF!\n\n"
                )   );
            } else
            {   tonum = readbase(arg4);
                for (i = address1; i <= address2; i++)
                {   if (memory[i] == tonum)
                    {   DISCARD getfriendly(i);
                        zprintf("%s\n", friendly);
                }   }
                zprintf("\n");
        }   }
        elif (arg2[0] && !arg3[0])
        {   tonum = readbase(arg2);
            for (i = 0; i <= 0x7FFF; i++)
            {   if (memory[i] == tonum)
                {   DISCARD getfriendly(i);
                    zprintf("%s\n", friendly);
            }   }
            zprintf("\n");
        } else
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_FIND, "Usage: FIND [<address1> <address2>] <value>\n\n"));
    }   }
    elif (!stricmp(arg1, "G"))
    {   if (arg3[0])
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_G, "Usage: G [<address>]\n\n"));
        } elif (arg2[0])
        {   address1 = friendly_to_number(arg2);
            if (bp_add(address1))
            {   DISCARD getfriendly(address1);
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_ADDBPANDGOING,
                        "Adding breakpoint at %s and going.\n\n"
                    ),
                    friendly
                );
                done = TRUE;
                emu_unpause();
        }   }
        else
        {   zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_GOING,
                    "Going.\n\n"
            )   );
            done = TRUE;
            emu_unpause();
    }   }
    elif
    (   !stricmp(arg1, "H"    )
     || !stricmp(arg1, "HELP" )
     || !stricmp(arg1, "?"    )
     || !stricmp(arg1, "AIUTO") // Italian for Help
    )
    {   if (arg3[0])
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_H, "Usage: H|HELP|? [1|2|3]\n\n"));
        } elif (arg2[0])
        {   if (!strcmp(arg2, "1"))
            {   help_commands(0);
            } elif (!strcmp(arg2, "2"))
            {   help_commands(1);
            } elif (!strcmp(arg2, "3"))
            {   help_commands(2);
            } else
            {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_H, "Usage: H|HELP|? [1|2|3]\n\n"));
        }   }
        else
        {   help_commands(0);
    }   }
    elif (!stricmp(arg1, "I"))
    {   docommand(MENUITEM_IGNORE);
    } elif (!stricmp(arg1, "PSG"))
    {   if (memmap == MEMMAP_F)
        {   view_psgs();
        } else
        {   zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_UNAVAILABLE,
                    "Command is unavailable.\n\n"
            )   );
    }   }
    elif (!stricmp(arg1, "GI"))
    {   if (machine == INTERTON || machine == ELEKTOR)
        {   docommand(MENUITEM_INTERRUPT);
        } else
        {   zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_UNAVAILABLE,
                    "Command is unavailable.\n\n"
            )   );
    }   }
    elif (!stricmp(arg1, "L"))
    {   if (!arg2[0] || arg3[0])
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_L, "Usage: L A|E|I|M|P|S\n\n"));
        } elif (!stricmp(arg2, "A"))
        {   zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_LOGREADWRITES,
                    "Log illegal reads/writes"
            )   );
            zprintf(" ");
            if (logreadwrites)
            {   logreadwrites = FALSE;
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_OFF,
                        "off"
                )   );
            } else
            {   logreadwrites = TRUE;
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_ON,
                        "on"
                )   );
            }
            zprintf(".\n\n");
            updatemenu(MENUITEM_READWRITES);
        } elif (!stricmp(arg2, "E"))
        {   if (memmap != MEMMAP_F && memmap != MEMMAP_P && machine != PHUNSY)
            {   zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_UNAVAILABLE,
                        "Command is unavailable.\n\n"
                )   );
            } else
            {   zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_LOGEPROM,
                        "Log EPROM accesses"
                )   );
                zprintf(" ");
                if (logeprom)
                {   logeprom = FALSE;
                    zprintf
                    (   GetCatalogStr(CatalogPtr,
                            MSG_ENGINE_OFF,
                            "off"
                    )   );
                } else
                {   logeprom = TRUE;
                    zprintf
                    (   GetCatalogStr(CatalogPtr,
                            MSG_ENGINE_ON,
                            "on"
                    )   );
                }
                zprintf(".\n\n");
                updatemenu(MENUITEM_LOGEPROMS);
        }   }
        elif (!stricmp(arg2, "I"))
        {   zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_LOGINSTRUCTIONS,
                    "Log illegal instructions"
            )   );
            zprintf(" ");
            if (loginstructions)
            {   loginstructions = FALSE;
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_OFF,
                        "off"
                )   );
            } else
            {   loginstructions = TRUE;
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_ON,
                        "on"
                )   );
            }
            zprintf(".\n\n");
            updatemenu(MENUITEM_INSTRUCTIONS);
        } elif (!stricmp(arg2, "M"))
        {   if
            (   machine == ELEKTOR
             || machine == PIPBUG
             || machine == INSTRUCTOR
             || machine == CD2650
             || machine == PHUNSY
            )
            {   zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_LOGMONITOR,
                        "Log BIOS calls"
                )   );
                zprintf(" ");
                if (logbios)
                {   logbios = FALSE;
                    zprintf
                    (   GetCatalogStr(CatalogPtr,
                            MSG_ENGINE_OFF,
                            "off"
                    )   );
                } else
                {   logbios = TRUE;
                    zprintf
                    (   GetCatalogStr(CatalogPtr,
                            MSG_ENGINE_ON,
                            "on"
                    )   );
                }
                zprintf(".\n\n");
                updatemenu(MENUITEM_LOGBIOS);
            } else // eg. ARCADIA, INTERTON, coin-ops, CHAOS, DOLPHIN
            {   zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_UNAVAILABLE,
                        "Command is unavailable.\n\n"
                )   );
        }   }
        elif (!stricmp(arg2, "P"))
        {   zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_LOGIOPORTS,
                    "Log I/O port accesses"
            )   );
            zprintf(" ");
            if (logioports)
            {   logioports = FALSE;
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_OFF,
                        "off"
                )   );
            } else
            {   logioports = TRUE;
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_ON,
                        "on"
                )   );
            }
            zprintf(".\n\n");
            updatemenu(MENUITEM_IOPORTS);
        } elif (!stricmp(arg2, "S"))
        {   zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_LOGSUBROUTINES,
                    "Log subroutine calls/returns"
            )   );
            zprintf(" ");
            if (logsubroutines)
            {   logsubroutines = FALSE;
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_OFF,
                        "off"
                )   );
            } else
            {   logsubroutines = TRUE;
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_ON,
                        "on"
                )   );
            }
            zprintf(".\n\n");
            updatemenu(MENUITEM_SUBROUTINES);
        } else
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_L, "Usage: L A|E|I|M|P|S\n\n"));
    }   }
    elif
    (   !stricmp(arg1, "LOAD")
     || !stricmp(arg1, "CARICA") // Italian for Load
    )
    {   if (!arg2[0] || !arg3[0] || arg4[0])
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_LOAD, "Usage: LOAD <address> <filename>\n\n"));
        } else
        {   address1 = friendly_to_number(arg2);
            if (address1 == OUTOFRANGE)
            {   zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_INVALIDRANGE,
                        "Valid range is $0000..$7FFF!\n\n"
                )   );
            } else
            {   if (getsize(arg3) == 0)
                {   fixextension(".RAW", arg3, FALSE);
                }
                localsize = (int) getsize(arg3);
                if (localsize + address1 > 32768)
                {   zprintf
                    (   GetCatalogStr(CatalogPtr,
                            MSG_ENGINE_TOOLARGE2,
                            "File is too large!\n\n"
                    )   );
                } elif ((TheLocalHandle = fopen(arg3, "rb")))
                {   if (fread(&memory[address1], (size_t) localsize, 1, TheLocalHandle) == 1)
                    {   zprintf
                        (   "%s.\n\n",
                            GetCatalogStr(CatalogPtr,
                                MSG_OK,
                                "OK"
                        )   );
                    } else
                    {   zprintf(GetCatalogStr(CatalogPtr, MSG_FAILED, "Failed!\n\n"));
                    }
                    DISCARD fclose(TheLocalHandle);
                    // TheLocalHandle = NULL;
                } else
                {   zprintf(GetCatalogStr(CatalogPtr, MSG_FAILED, "Failed!\n\n"));
    }   }   }   }
    elif (!stricmp(arg1, "LOADER"))
    {   if (memmap == MEMMAP_F)
        {   docommand(COMMAND_LOADER);
        } else
        {   zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_UNAVAILABLE,
                    "Command is unavailable.\n\n"
            )   );
    }   }
    elif (!stricmp(arg1, "MRAM"))
    {   view_mram();
    } elif (!stricmp(arg1, "NVRAM"))
    {   if (memmap == MEMMAP_MALZAK2)
        {   view_nvram();
        } else
        {   zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_UNAVAILABLE,
                    "Command is unavailable.\n\n"
            )   );
    }   }
    elif (!stricmp(arg1, "PL"))
    {   if (pauselog)
        {   pauselog = FALSE;
            zprintf(GetCatalogStr(CatalogPtr, MSG_PL_OFF, "Pause after logging is now off.\n\n"));
        } else
        {   pauselog = TRUE;
            zprintf(GetCatalogStr(CatalogPtr, MSG_PL_ON,  "Pause after logging is now on.\n\n" ));
        }
        updatemenu(MENUITEM_PAUSELOG);
    } elif (  !stricmp(arg1, "PLAYFIELD"))
    {   if (machine != MALZAK)
        {   zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_UNAVAILABLE,
                    "Command is unavailable.\n\n"
            )   );
        } else
        {   m_viewplayfield();
    }   }
    elif (!stricmp(arg1, "PVI"))
    {   if
        (   machine == INTERTON
         || machine == ELEKTOR
         || machine == MALZAK
         || machine == ASTROWARS
         || machine == GALAXIA
        )
        {   view_xvi();
        } else
        {   zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_UNAVAILABLE,
                    "Command is unavailable.\n\n"
            )   );
    }   }
    elif
    (   !stricmp(arg1, "Q")
     || !stricmp(arg1, "QUIT")
    )
    {   done = TRUE;
        cli = FALSE;
        docommand(MENUITEM_CLI);
    } elif (!stricmp(arg1, "="))
    {   if (arg3[0])
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_EQUALS, "Usage: = [<address>]\n\n"));
        } elif (arg2[0])
        {   address1 = friendly_to_number(arg2);
            if (address1 == OUTOFRANGE)
            {   zprintf(LLL, MSG_INVALIDRANGE, "Valid range is $0000..$7FFF!\n\n"));
            } else
            {   DISCARD getfriendly(address1);
                zprintf("%s\n", friendly);
                if
                (   mirror_r[address1] != address1
                 || mirror_w[address1] != address1
                )
                {   DISCARD getfriendly(mirror_r[address1]);
                    zprintf(LLL, MSG_READMIRROR,  "Read mirror of %s." ), friendly);
                    zprintf(" ");
                    DISCARD getfriendly(mirror_w[address1]);
                    zprintf(LLL, MSG_WRITEMIRROR, "Write mirror of %s."), friendly);
                    zprintf("\n\n");
                } else
                {   zprintf("\n");
        }   }   }
        else
        {   j = 0;
            for (i = 0; i <= 0x7FFF; i++)
            {   if (number_to_friendly(i, friendly, FALSE))
                {   if (j != 0 && j % 3 == 0)
                    {   zprintf("\n");
                    }
                    j++;
                    zprintf("%17s = $%04X ", friendly, i);
            }   }
            zprintf("\n\n");
    }   }
    elif (!stricmp(arg1, "R"))
    {   if (arg3[0])
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_R, "Usage: R [D|F|I|L|R|S|<number>]\n\n"));
        } elif (!stricmp(arg2, "D"))
        {   if (machine == ARCADIA)
            {   docommand(MENUITEM_RUNTODMA);
                done = TRUE;
            } else
            {   zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_UNAVAILABLE,
                        "Command is unavailable.\n\n"
                )   );
        }   }
        elif (!stricmp(arg2, "F"))
        {   docommand(MENUITEM_RUNTOFRAME);
            done = TRUE;
        } elif (!stricmp(arg2, "I"))
        {   if
            (   machine == INTERTON
             || machine == ELEKTOR
             || machine == ASTROWARS
             || machine == GALAXIA
             || machine == MALZAK
            )
            {   docommand(MENUITEM_RUNTOINTERRUPT);
                done = TRUE;
            } else // eg. ARCADIA, PIPBUG, INSTRUCTOR, CD2650, CHAOS, DOLPHIN, PHUNSY
            {   zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_UNAVAILABLE,
                        "Command is unavailable.\n\n"
                )   );
        }   }
        elif (!stricmp(arg2, "L"))
        {   docommand(MENUITEM_RUNTOLOOPEND);
            done = TRUE;
        } elif (!stricmp(arg2, "R"))
        {   switch (machine)
            {
            case ARCADIA:
            case INTERTON:
            case ELEKTOR:
            case ASTROWARS:
            case GALAXIA:
            case MALZAK:
                docommand(MENUITEM_RUNTORAST);
                done = TRUE;
            adefault: // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS, DOLPHIN, PHUNSY
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_UNAVAILABLE,
                        "Command is unavailable.\n\n"
                )   );
        }   }
        elif (!stricmp(arg2, "S"))
        {   docommand(MENUITEM_RUNTOSUBEND);
            done = TRUE;
        } elif (arg2[0])
        {   switch (machine)
            {
            case ARCADIA:
            case INTERTON:
            case ELEKTOR:
            case ASTROWARS:
            case GALAXIA:
            case MALZAK:
                // this is similar but different to docommand(MENUITEM_RUNTORASTN);
                clearruntos();
                DISCARD stcd_l(arg2, (void *) &rastn); // string -> decimal number
                emu_unpause();
                done = TRUE;
            adefault: // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_UNAVAILABLE,
                        "Command is unavailable.\n\n"
                )   );
        }   }
        else
        {   switch (machine)
            {
            case ARCADIA:
            case INTERTON:
            case ELEKTOR:
            case ASTROWARS:
            case GALAXIA:
            case MALZAK:
                docommand(MENUITEM_RUNTORASTN);
                done = TRUE;
            adefault: // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS, DOLPHIN, PHUNSY
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_ENGINE_UNAVAILABLE,
                        "Command is unavailable.\n\n"
                )   );
    }   }   }
    elif (!stricmp(arg1, "RAM"))
    {   view_mram();
        if (memmap == MEMMAP_C || memmap == MEMMAP_D)
        {   view_cram();
    }   }
    elif (!stricmp(arg1, "READPORT"))
    {   if (!arg2[0] || arg3[0])
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_READPORT, "Usage: READPORT <port>|C|D\n\n"));
        } else
        {   if (arg2[0] == 'C' || arg2[0] == 'c')
            {   zprintf("%d\n\n", readport(PORTC));
            } elif (arg2[0] == 'D' || arg2[0] == 'd')
            {   zprintf("%d\n\n", readport(PORTD));
            } else
            {   DISCARD stcd_l(arg2, (void *) &address1); // string -> decimal number
                if (address1 >= 0 && address1 <= 255)
                {   zprintf("%d\n\n", readport(address1));
                } else
                {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_READPORT, "Usage: READPORT <port>|C|D\n\n"));
    }   }   }   }
    elif (!stricmp(arg1, "ROM"))
    {   if
        (   machine != ASTROWARS
         && machine != GALAXIA
         && machine != MALZAK
         && machine != PHUNSY
         && (   !game
             || (   machine != ARCADIA
                 && machine != INTERTON
                 && machine != ELEKTOR
        )   )   )
        {   zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_UNAVAILABLE,
                    "Command is unavailable.\n\n"
            )   );
        } else
        {   view_imagery();
    }   }
    elif
    (   !stricmp(arg1, "SAVE")
     || !stricmp(arg1, "SALVA") // Italian for Save
    )
    {   if (arg2[0] && arg3[0] && arg4[0])
        {   address1 = friendly_to_number(arg2);
            address2 = friendly_to_number(arg3);
            if
            (   address1 >  address2 // perhaps we might want a different error message for this case
             || address1 == OUTOFRANGE
             || address2 == OUTOFRANGE
            )
            {   zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_INVALIDRANGE,
                        "Valid range is $0000..$7FFF!\n\n"
                )   );
            } else
            {   fixextension(".RAW", arg4, TRUE);
                if ((TheLocalHandle = fopen(arg4, "wb")))
                {   DISCARD fwrite(&memory[address1], (size_t) (address2 - address1 + 1), 1, TheLocalHandle);
                    DISCARD fclose(TheLocalHandle);
                    // TheLocalHandle = NULL;
                    zprintf
                    (   "%s.\n\n",
                        GetCatalogStr(CatalogPtr,
                            MSG_OK,
                            "OK"
                    )   );
                } else
                {   zprintf(GetCatalogStr(CatalogPtr, MSG_FAILED, "Failed!\n\n"));
        }   }   }
        else
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_SAVE, "Usage: SAVE <address1> <address2> <filename>\n\n"));
    }   }
    elif (!stricmp(arg1, "SAVEAOF"))
    {   if (arg2[0] && arg3[0] && arg4[0])
        {   address1 = friendly_to_number(arg2);
            address2 = friendly_to_number(arg3);
            if (arg5[0])
            {   address3 = friendly_to_number(arg5);
            } else
            {   address3 = (startaddr_h * 256) + startaddr_l;
            }
            if
            (   address1 >  address2 // perhaps we might want a different error message for this case
             || address1 == OUTOFRANGE
             || address2 == OUTOFRANGE
             || address3 == OUTOFRANGE
            )
            {   zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_INVALIDRANGE,
                        "Valid range is $0000..$7FFF!\n\n"
                )   );
            } else
            {   fixextension(".AOF", arg4, TRUE);
                if (save_aof(arg4, address1, address2, address3))
                {   zprintf
                    (   "%s.\n\n",
                        GetCatalogStr(CatalogPtr,
                            MSG_OK,
                            "OK"
                    )   );
                } else
                {   zprintf(GetCatalogStr(CatalogPtr, MSG_FAILED, "Failed!\n\n"));
        }   }   }
        else
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_SAVEAOF, "Usage: SAVEAOF <address1> <address2> <filename> [<startaddr>]\n\n"));
    }   }
    elif (!stricmp(arg1, "SAVER"))
    {   if (memmap == MEMMAP_F)
        {   docommand(COMMAND_SAVER);
        } else
        {   zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_UNAVAILABLE,
                    "Command is unavailable.\n\n"
            )   );
    }   }
    elif (!stricmp(arg1, "SCRN"))
    {   switch (machine)
        {
        case ARCADIA:
        case INTERTON:
        case ELEKTOR:
        case INSTRUCTOR:
        case CD2650:
        case ASTROWARS:
        case GALAXIA:
        case MALZAK:
        case DOLPHIN:
        case PHUNSY:
            view_screen();
        adefault: // eg. PIPBUG, CHAOS
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_UNAVAILABLE,
                    "Command is unavailable.\n\n"
            )   );
    }   }
    elif (!stricmp(arg1, "SYS") || !stricmp(arg1, "SYSTEM"))
    {   cleanexit(EXIT_SUCCESS);
    } elif (!stricmp(arg1, "T"))
    {   zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_ENGINE_TRACECPU,
                "CPU tracing"
        )   );
        zprintf(" ");
        if (trace)
        {   trace = FALSE;
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_OFF,
                    "off"
            )   );
        } else
        {   trace = TRUE;
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_ON,
                    "on"
            )   );
        }
        zprintf(".\n\n");
        updatemenu(MENUITEM_TRACECPU);
    } elif (!stricmp(arg1, "TRAIN"))
    {   if (arg3[0])
        {   zprintf("Usage: TRAIN [<value>]\n\n");
        } elif (arg2[0])
        {   trainval = readbase(arg2);
            for (i = 0; i <= 0x7FFF; i++)
            {   if
                (   !(memflags[i] & (NOREAD | NOWRITE))
                 && memory[i] == trainval
                )
                {   DISCARD getfriendly(i);
                    zprintf("%s\n", friendly);
                    memflags[i] |= TRAINABLE;
                } else
                {   memflags[i] &= ~(TRAINABLE);
            }   }
            zprintf("\n");
        } else
        {   for (i = 0; i <= 0x7FFF; i++)
            {   if
                (   (memflags[i] & TRAINABLE)
                 && memory[i] == trainval - 1
                )
                {   DISCARD getfriendly(i);
                    zprintf("%s\n", friendly);
            }   }
            zprintf("\n");
    }   }
    elif (!stricmp(arg1, "UDC"))
    {   switch (machine)
        {
        case ARCADIA:
        case INTERTON:
        case ELEKTOR:
        case MALZAK:
        case ASTROWARS:
        case GALAXIA:
            view_udcs();
        adefault: // eg. PIPBUG, INSTRUCTOR, CD2650, CHAOS
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_UNAVAILABLE,
                    "Command is unavailable.\n\n"
            )   );
    }   }
    elif (!stricmp(arg1, "UVI"))
    {   if (machine == ARCADIA)
        {   view_xvi();
        } else
        {   zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_UNAVAILABLE,
                    "Command is unavailable.\n\n"
            )   );
    }   }
    elif (!stricmp(arg1, "V"))
    {   zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_ENGINE_VERBOSETRACING,
                "Verbose tracing"
        )   );
        zprintf(" ");
        if (traceverbose)
        {   traceverbose = FALSE;
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_OFF,
                    "off"
            )   );
        } else
        {   traceverbose = TRUE;
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_ON,
                    "on"
            )   );
        }
        zprintf(".\n\n");
        updatemenu(MENUITEM_TRACEVERBOSE);
    } elif (!stricmp(arg1, "WC"))
    {   if (arg4[0])
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_WC, "Usage: WC [<address1> [<address2>]]\n\n"));
        } elif (arg3[0])
        {   address1 = friendly_to_number(arg2);
            address2 = friendly_to_number(arg3);
            if
            (   address1 != OUTOFRANGE
             && address2 != OUTOFRANGE
             && address1 <= address2
            )
            {   for (; address1 <= address2; address1++)
                {   memflags[address1] &= ~(WATCHPOINT);
                }
                for (address1 = 0; address1 < 32768; address1++)
                {   if (memflags[address1] & WATCHPOINT)
                    {   wps = FALSE;
                        break; // for speed
                }   }
                updatemenu(MENUITEM_CLEARWP );
                updatemenu(MENUITEM_CLEARWPS);
                update_memory(TRUE);
                zprintf("OK.\n\n"); // localize this!
        }   }
        elif (arg2[0])
        {   address1 = friendly_to_number(arg2);
            DISCARD wp_clear(address1);
        } else
        {   DISCARD wp_clearall();
    }   }
    elif (!stricmp(arg1, "WL"))
    {   ok = FALSE;
        for (i = 0; i <= 32767; i++)
        {   if (memflags[i] & WATCHPOINT)
            {   DISCARD getfriendly(i);
                zprintf("%s\n", friendly);
                ok = TRUE;
        }   }
        if (ok)
        {   zprintf("\n");
        } else
        {   zprintf
            (   "%s.\n\n",
                GetCatalogStr(CatalogPtr,
                    MSG_NONE,
                    "None"
            )   );
    }   }
    elif (!stricmp(arg1, "WP"))
    {   if (arg4[0])
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_WP, "Usage: WP [<address1> [<address2>]]\n\n"));
        } elif (arg3[0])
        {   address1 = friendly_to_number(arg2);
            address2 = friendly_to_number(arg3);
            if
            (   address1 != OUTOFRANGE
             && address2 != OUTOFRANGE
             && address1 <= address2
            )
            {   for (; address1 <= address2; address1++)
                {   memflags[address1] |= WATCHPOINT;
                    wps = TRUE;
                }
                updatemenu(MENUITEM_CLEARWP );
                updatemenu(MENUITEM_CLEARWPS);
                update_memory(TRUE);
                zprintf("OK.\n\n"); // localize this!
        }   }
        elif (arg2[0])
        {   address1 = friendly_to_number(arg2);
            DISCARD wp_add(address1);
        } else
        {   debug_set(KIND_WP);
    }   }
    elif (!stricmp(arg1, "WR"))
    {   if (watchreads)
        {   watchreads = FALSE;
            zprintf(GetCatalogStr(CatalogPtr, MSG_WR_OFF, "Watch reads off.\n\n"));
        } else
        {   watchreads = TRUE;
            zprintf(GetCatalogStr(CatalogPtr, MSG_WR_ON,  "Watch reads on.\n\n" ));
        }
        updatemenu(MENUITEM_WATCHREADS);
    } elif (!stricmp(arg1, "WRITEPORT"))
    {   if (!arg3[0] || arg4[0])
        {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_WRITEPORT, "Usage: WRITEPORT <port>|C|D <value>\n\n"));
        } else
        {   address2 = readbase(arg3);
            if (arg2[0] == 'C' || arg2[0] == 'c')
            {   writeport(PORTC, (UBYTE) address2);
                zprintf("OK.\n\n"); // localize this!
            } elif (arg2[0] == 'D' || arg2[0] == 'd')
            {   writeport(PORTD, (UBYTE) address2);
                zprintf("OK.\n\n"); // localize this!
            } else
            {   address1 = readbase(arg2);
                if (address1 >= 0 && address1 <= 255)
                {   writeport(address1, (UBYTE) address2);
                    zprintf("OK.\n\n"); // localize this!
                } else
                {   zprintf(GetCatalogStr(CatalogPtr, MSG_USAGE_WRITEPORT, "Usage: WRITEPORT <port>|C|D <value>\n\n"));
    }   }   }   }
    else
    {   zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_ENGINE_UNKNOWNCOMMAND,
                "Unknown command \"%s\" (\"H\" for help).\n\n"
            ),
            strupr(arg1)
        );
    }

    return done;
}

MODULE void help_commands(int whichpage)
{   PERSIST const int    commandmsg[3]  =
    { MSG_ENGINE_DEBUGHELP,
      MSG_ENGINE_DEBUGHELP2,
      MSG_ENGINE_DEBUGHELP3
    };
    PERSIST const STRPTR commandhelp[3] = {
// HELP 1 (default)
"BL|WL                           list all breakpoints/watchpoints\n"\
"BP|BC|WP|WC [<addr1> [<addr2>]] add/clear bp/wp from addr1 to addr2\n"\
"D|PEEK <address1> [<address2>]  display address1 to address2 as data\n"\
"E|POKE [<address> <value>]      change address to value\n"\
"FILL <addr1> <addr2> <value>    fill addr1 to addr2 with value\n"\
"G [<address>]                   set breakpoint to address and go\n"\
"H|HELP|? [1|2|3]                view help page # (default is 1)\n"\
"I                               ignore next instruction\n"\
"O                               step over subroutine\n"\
"Q|QUIT                          quit debugger CLI\n"\
"R [<number>]                    run to rasterline number\n"\
"R D                             run to next DMA\n"\
"R F                             run to next frame\n"\
"R I                             run to next interrupt\n"\
"R L                             run to end of loop\n"\
"R R                             run to next rasterline\n"\
"R S                             run to end of subroutine\n"\
"[S]                             step next instruction\n"\
"SYS|SYSTEM                      quit emulator\n\n",
// HELP 2
"=                               list all known labels\n" \
"= <address>                     view address as label and hexadecimal\n"\
"BASE [!|@|$|%%]                  set default input base\n"\
"COVER                           view coverage report\n"\
"DIS [<address1> [<address2>]]   disassemble address1 to address2 as code\n"\
"FIND [<addr1> <addr2>] <value>  find value in address1 to address2\n"\
"GI                              generate an interrupt\n"\
"L A                             log illegal memory accesses on/off\n"\
"L E                             log EPROM accesses on/off\n"\
"L I                             log illegal instructions on/off\n"\
"L M                             log BIOS calls on/off\n"\
"L P                             log I/O port accesses on/off\n"\
"L S                             log subroutine calls on/off\n"\
"LOAD <address> <filename>       load from address from filename\n"\
"PL                              pause after logging /on/off\n"\
"SAVE <addr1> <addr2> <filename> save address1 to address2 as filename\n"\
"T                               trace on/off\n"\
"TRAIN [<value>]                 make trainer\n"\
"V                               verbose tracing on/off\n"\
"WR                              watch reads on/off\n\n",
// HELP 3
"ALL                             view all memory\n"\
"BRAM                            view BIOS RAM\n"\
"COLOURS                         view colour attribute RAM\n"\
"CPU                             view CPU status\n"\
"CRAM                            view cartridge RAM\n"\
"LOADER                          insert loader EPROM\n" \
"MRAM                            view motherboard RAM\n"\
"NVRAM                           view non-volatile RAM\n"\
"PLAYFIELD                       view playfield tiles\n"\
"PSG                             view PSGs status\n"\
"PVI                             view PVI status\n"\
"RAM                             view motherboard and cartridge RAM\n"\
"READPORT <port>|C|D             read from I/O port\n"\
"ROM                             view ROM as imagery\n"\
"SAVEAOF <address1> <address2> <filename> [<startaddress>]\n"\
"                                save address1 to address2 as filename\n"\
"SAVER                           insert saver EPROM\n" \
"SCRN                            view screen contents\n"\
"UDC                             view user-defined characters\n"\
"UVI                             view UVI status\n"\
"WRITEPORT <port>|C|D <value>    write to I/O port\n\n"\
};

    zprintf(GetCatalogStr(CatalogPtr, commandmsg[whichpage], commandhelp[whichpage]));
}

EXPORT UBYTE ReadByteAt(int where)
{   UBYTE result;

    offset = where;
    result = IOBuffer[offset++];
    return result;
}
EXPORT UWORD ReadWord(void)
{   UWORD result;

    // big endian 16-bit word
    result = (UWORD) (256 * IOBuffer[offset])
                          + IOBuffer[offset + 1];
    offset += 2;
    return result;
}
EXPORT ULONG ReadLong(void)
{   ULONG result;

    // big endian 32-bit longword
    result = (ULONG) (16777216 * IOBuffer[offset    ])
                   + (   65536 * IOBuffer[offset + 1])
                   + (     256 * IOBuffer[offset + 2])
                   +             IOBuffer[offset + 3];
    offset += 4;

    return result;
}
EXPORT UWORD ReadWordAt(int where)
{   UWORD result;

    offset = where;
    // big endian 16-bit word
    result = (UWORD) (256 * IOBuffer[offset])
                          + IOBuffer[offset + 1];
    offset += 2;
    return result;
}

// this routine reads but doesn't write offset
MODULE FLAG autosense(int thesize, FLAG silent)
{   ULONG crc; // crc32() is a function name
    int   i;

    crc = crc32(IOBuffer + offset, thesize);
#ifdef VERBOSE
    T1; textpen(TEXTPEN_VERBOSE); zprintf("ROM size: %d. CRC32: $%X.\n", romsize, crc); T2;
#endif

    for (i = 0; i < KNOWNGAMES; i++)
    {   if (thesize == known[i].size && crc == known[i].crc32)
        {   whichgame = i;
            configure(silent);
            return TRUE; // for speed
    }   }

#ifdef VERBOSE
    T1; textpen(TEXTPEN_VERBOSE); zprintf("Unknown game.\n"); T2;
#endif

    return FALSE;
}

MODULE void calcrunningtime(void)
{   PERSIST int hours,
                minutes,
                seconds,
                micros; // all PERSISTent for speed

    if (region == REGION_NTSC)
    {   seconds = (int)          (frames / 60) ; // how many seconds
        micros  = (int) (16667 * (frames % 60)); // microseconds
    } else
    {   // assert(region == REGION_PAL);
        seconds = (int)          (frames / 50) ; // how many seconds
        micros  = (int) (20000 * (frames % 50)); // microseconds
    }
    hours   = (seconds / 3600) % 100; // resets to 00 hours after 100 hours
    minutes = (seconds % 3600) / 60;
    seconds %= 60;

    TimeStr[0]  = (TEXT) ('0' + (hours   / 10)            );
    TimeStr[1]  = (TEXT) ('0' + (hours   % 10)            );
    TimeStr[3]  = (TEXT) ('0' + (minutes / 10)            );
    TimeStr[4]  = (TEXT) ('0' + (minutes % 10)            );
    TimeStr[6]  = (TEXT) ('0' + (seconds / 10)            );
    TimeStr[7]  = (TEXT) ('0' + (seconds % 10)            );
    TimeStr[9]  = (TEXT) ('0' + (micros  / 100000)        );
    TimeStr[10] = (TEXT) ('0' + (micros  % 100000) / 10000);
}

EXPORT void clearruntos(void)
{   runtoframe     =
    runtointerrupt =
    runtoloopend   =
    runtodma       = FALSE;
    rastn          = -1;
}

EXPORT FLAG bp_add(int address)
{   T1; textpen(TEXTPEN_DEBUG);

    if (address < 0 || address >= OUTOFRANGE)
    {   zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_INVALIDRANGE,
                "Valid range is $0000..$7FFF!\n\n"
        )   );
        T2; return FALSE;
    }

    if (memflags[address] & BREAKPOINT)
    {   zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_BPALREADYSET,
                "Breakpoint is already set!\n\n"
        )   );
        T2; return FALSE;
    }

    memflags[address] |= BREAKPOINT;
    bps = TRUE;
    updatemenu(MENUITEM_CLEARBP );
    updatemenu(MENUITEM_CLEARBPS);
    update_memory(TRUE);

    DISCARD getfriendly(address);
    zprintf
    (   GetCatalogStr
        (   CatalogPtr,
            MSG_ADDBP,
            "Added code breakpoint at %s.\n\n"
        ),
        friendly
    );
    T2; return TRUE;
}
EXPORT FLAG bp_clear(int address)
{   int i;

    T1; textpen(TEXTPEN_DEBUG);

    if (address < 0 || address >= OUTOFRANGE)
    {   zprintf
        (   GetCatalogStr
            (   CatalogPtr,
                MSG_INVALIDRANGE,
                "Valid range is $0000..$7FFF!\n\n"
        )   );
        T2; return FALSE;
    }

    DISCARD getfriendly(address);
    if (memflags[address] & BREAKPOINT)
    {   memflags[address] &= ~(BREAKPOINT);
        zprintf
        (   GetCatalogStr
            (   CatalogPtr,
                MSG_ENGINE_CLEAREDBP,
                "Cleared code breakpoint at %s.\n\n"
            ),
            friendly
        );
        T2;
        update_memory(TRUE);
        for (i = 0; i <= 32767; i++)
        {   if (memflags[i] & BREAKPOINT)
            {   return TRUE;
        }   }
        bps = FALSE;
        updatemenu(MENUITEM_CLEARBP );
        updatemenu(MENUITEM_CLEARBPS);
        return TRUE;
    } else
    {   zprintf
        (   GetCatalogStr
            (   CatalogPtr,
                MSG_ENGINE_BPNOTFOUND,
                "Breakpoint not found at %s!\n\n"
            ),
            friendly
        );
        T2; return FALSE;
}   }
EXPORT FLAG bp_clearall(void)
{   int address;

    T1; textpen(TEXTPEN_DEBUG);

    if (bps)
    {   for (address = 0; address <= 32767; address++)
        {   if (memflags[address] & BREAKPOINT)
            {   memflags[address] &= ~(BREAKPOINT);
        }   }
        bps = FALSE;
        updatemenu(MENUITEM_CLEARBP );
        updatemenu(MENUITEM_CLEARBPS);
        update_memory(TRUE);
        zprintf
        (   GetCatalogStr
            (   CatalogPtr,
                MSG_ENGINE_CLEAREDBPS,
                "Cleared all code breakpoints.\n\n"
        )   );
    } else
    {   zprintf
        (   GetCatalogStr
            (   CatalogPtr,
                MSG_NOBPS,
                "No breakpoints to clear!\n\n"
        )   );
    }
    T2; return TRUE;
}
EXPORT FLAG wp_add(int address)
{   T1; textpen(TEXTPEN_DEBUG);

    if (address < 0 || address >= OUTOFRANGE)
    {   zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_INVALIDRANGE,
                "Valid range is $0000..$7FFF!\n\n"
        )   );
        T2; return FALSE;
    }

    if (memflags[address] & WATCHPOINT)
    {   zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_WPALREADYSET,
                "Watchpoint is already set!\n\n"
        )   );
        T2; return FALSE;
    }

    memflags[address] |= WATCHPOINT;
    wps = TRUE;
    updatemenu(MENUITEM_CLEARWP );
    updatemenu(MENUITEM_CLEARWPS);
    update_memory(TRUE);

    DISCARD getfriendly(address);
    zprintf
    (   GetCatalogStr
        (   CatalogPtr,
            MSG_ADDWP,
            "Added data watchpoint at %s.\n\n"
        ),
        friendly
    );
    T2; return TRUE;
}
EXPORT FLAG wp_clear(int address)
{   int i;

    T1; textpen(TEXTPEN_DEBUG);

    if (address < 0 || address >= OUTOFRANGE)
    {   zprintf
        (   GetCatalogStr
            (   CatalogPtr,
                MSG_INVALIDRANGE,
                "Valid range is $0000..$7FFF!\n\n"
        )   );
        T2; return FALSE;
    }

    DISCARD getfriendly(address);
    if (memflags[address] & WATCHPOINT)
    {   memflags[address] &= ~(WATCHPOINT);
        zprintf
        (   GetCatalogStr
            (   CatalogPtr,
                MSG_ENGINE_CLEAREDWP,
                "Cleared data watchpoint at %s.\n\n"
            ),
            friendly
        );
        T2;
        update_memory(TRUE);
        for (i = 0; i <= 32767; i++)
        {   if (memflags[i] & WATCHPOINT)
            {   return TRUE;
        }   }
        wps = FALSE;
        updatemenu(MENUITEM_CLEARWP );
        updatemenu(MENUITEM_CLEARWPS);
        return TRUE;
    } else
    {   zprintf
        (   GetCatalogStr
            (   CatalogPtr,
                MSG_ENGINE_WPNOTFOUND,
                "Watchpoint not found at %s!\n\n"
            ),
            friendly
        );
        T2; return FALSE;
}   }

EXPORT FLAG wp_clearall(void)
{   int address;

    T1; textpen(TEXTPEN_DEBUG);

    if (wps)
    {   for (address = 0; address <= 32767; address++)
        {   if (memflags[address] & WATCHPOINT)
            {   memflags[address] &= ~(WATCHPOINT);
        }   }
        wps = FALSE;
        updatemenu(MENUITEM_CLEARWP );
        updatemenu(MENUITEM_CLEARWPS);
        update_memory(TRUE);
        zprintf
        (   GetCatalogStr
            (   CatalogPtr,
                MSG_ENGINE_CLEAREDWPS,
                "Cleared all data watchpoints.\n\n"
        )   );
    } else
    {   zprintf
        (   GetCatalogStr
            (   CatalogPtr,
                MSG_NOWPS,
                "No watchpoints to clear!\n\n"
        )   );
    }
    T2; return TRUE;
}

EXPORT void edit_copytext_engine(void)
{   int  i, j, x, y;
    TEXT tempstring[3];
    FLAG done;

    // assert(machine != INTERTON);

    ProjectBuffer[0] = 0;

    switch (machine)
    {
    case ARCADIA:
        for (i = 0x1800; i < 0x18D0; i += 16)
        {   tempstring[1] = 0;
            for (j = 0; j < 16; j++)
            {   tempstring[0] = arcadia_chars[memory[i + j] % 64];
                strcat(ProjectBuffer, tempstring);
            }
#ifdef WIN32
            tempstring[0] = CR;
            tempstring[1] = LF;
            tempstring[2] = 0;
#endif
#ifdef AMIGA
            tempstring[0] = LF;
            tempstring[1] = 0;
#endif
            strcat(ProjectBuffer, tempstring);
        }

        if
        (   (memory[A_RESOLUTION] & 0x40)
         && (memory[A_BGCOLOUR]   & 0x80)
        )
        {   for (i = 0x1A00; i < 0x1AD0; i += 16)
            {   tempstring[1] = 0;
                for (j = 0; j < 16; j++)
                {   tempstring[0] = arcadia_chars[memory[i + j] % 64];
                    strcat(ProjectBuffer, tempstring);
                }
#ifdef WIN32
                tempstring[0] = CR;
                tempstring[1] = LF;
                tempstring[2] = 0;
#endif
#ifdef AMIGA
                tempstring[0] = LF;
                tempstring[1] = 0;
#endif
                strcat(ProjectBuffer, tempstring);
        }   }
    acase ELEKTOR:
        for (y = 0; y < 6; y++)
        {   monob(y);
            strcat(ProjectBuffer, letters);
#ifdef WIN32
            tempstring[0] = CR;
            tempstring[1] = LF;
            tempstring[2] = 0;
#endif
#ifdef AMIGA
            tempstring[0] = LF;
            tempstring[1] = 0;
#endif
            strcat(ProjectBuffer, tempstring);
        }
    acase PIPBUG:
    case CHAOS:
        i = 0;
        for (y = 0; y < PH_VDUROWS; y++)
        {   for (x = 0; x < vducols; x++)
            {   if (vdu[x][y] != BLANK)
                {   ProjectBuffer[i++] = vdu[x][y] + 0x20;
            }   }
            if (y < PH_VDUROWS - 1)
            {
#ifdef WIN32
                ProjectBuffer[i++] = CR;
#endif
                ProjectBuffer[i++] = LF;
        }   }
        ProjectBuffer[i] = 0;
    acase INSTRUCTOR:
        ProjectBuffer[0] = instructor_chars[memory[0x17D1]];
        ProjectBuffer[1] = instructor_chars[memory[0x17D2]];
        ProjectBuffer[2] = instructor_chars[memory[0x17D3]];
        ProjectBuffer[3] = instructor_chars[memory[0x17D4]];
        ProjectBuffer[4] = instructor_chars[memory[0x17D5]];
        ProjectBuffer[5] = instructor_chars[memory[0x17D6]];
        ProjectBuffer[6] = instructor_chars[memory[0x17D7]];
        ProjectBuffer[7] = instructor_chars[memory[0x17D8]];
        ProjectBuffer[8] = 0;
    acase CD2650:
        i = 0;
        for (y = 0; y < 16; y++)
        {   done = FALSE;
            for (x = 0; x < 80; x++)
            {   if (memory[0x1000 + (x * 16) + y])
                {   ProjectBuffer[i++] = cd2650_chars[memory[0x1000 + (x * 16) + y] & 0x7F];
                    done = TRUE;
            }   }
            if (y < 15 && done)
            {
#ifdef WIN32
                ProjectBuffer[i++] = CR;
#endif
                ProjectBuffer[i++] = LF;
        }   }
        ProjectBuffer[i] = 0;
    acase PHUNSY:
        i = 0;
        for (y = 0; y < 32; y++)
        {   done = FALSE;
            for (x = 0; x < 64; x++)
            {   if (memory[0x1000 + (y * 64) + x])
                {   ProjectBuffer[i++] = memory[0x1000 + (y * 64) + x];
                    done = TRUE;
            }   }
            if (y < 31 && done)
            {
#ifdef WIN32
                ProjectBuffer[i++] = CR;
#endif
                ProjectBuffer[i++] = LF;
        }   }
        ProjectBuffer[i] = 0;
}   }

EXPORT int bin2int(STRPTR thestring)
{   int i, j,
        value = 0;

    // assert(strlen(thestring) <= 32);

    for (i = (int) strlen(thestring) - 1, j = 0; i >= 0; i--, j++)
    {   if (thestring[i] == '1' || thestring[i] == '#')
        {   value |= (1 << j);
        } elif (thestring[i] != '0' && thestring[i] != '.')
        {   value = OUTOFRANGE;
            return value;
    }   }
    return value;
}

EXPORT int oct2int(STRPTR thestring)
{   int i, j,
        value = 0;

    // this assumes the string is in range and only has valid characters (0..7)

    for (i = (int) strlen(thestring) - 1, j = 0; i >= 0; i--, j++)
    {   value += (thestring[i] - '0') * (1 << (j * 3));
    }
    return value;
}

EXPORT FLAG IsZip(STRPTR filename)
{   UBYTE buf[2];
    FILE* fp;

    if ((fp = fopen(filename, "rb")))
    {   DISCARD fread(buf, 2, 1, fp);
        fclose(fp);
        return (FLAG) (memcmp(buf, "PK", 2) == 0);
    }

    return FALSE;
}

MODULE FLAG loadzip(STRPTR passedname, FLAG auditing, FLAG quiet)
{   TRANSIENT unzFile       fp;
    TRANSIENT unz_file_info info;
    TRANSIENT TEXT          buf[3 + 1];
    TRANSIENT TEXT*         ptr;
    TRANSIENT FLAG          found;
    PERSIST   TEXT          filename_buf[MAX_PATH + 1];

    // assert(IsZip(passedname));

    info.uncompressed_size = 0; // to avoid spurious SAS/C compiler warnings

    if (!(fp = (unzFile) unzOpen(passedname)))
    {   if (!quiet) say("Can't open ZIP archive!", TRUE);
        return FALSE;
    } // implied else
    if (unzGoToFirstFile(fp) != 0) // UNZ_OK
    {   DISCARD unzClose(fp);
        if (!quiet) say("Corrupt ZIP archive!", TRUE);
        return FALSE;
    } // implied else

    found = FALSE;
    for (;;)
    {   // Get file descriptor and analyse
        if (unzGetCurrentFileInfo(fp, &info, filename_buf, 0x100, NULL, 0, NULL, 0) != 0) // UNZ_OK
        {   break; // implied else
        }
        ptr = strchr(filename_buf, '.');
        if (!ptr)
        {   break;
        }
        ptr++; buf[0] = (TEXT) toupper(*ptr);
        ptr++; buf[1] = (TEXT) toupper(*ptr);
        ptr++; buf[2] = (TEXT) toupper(*ptr);
        buf[3] = 0;

        if
        (   !stricmp(buf, "BIN")
         || !stricmp(buf, "PGM")
         || !stricmp(buf, "TVC")
         || (   !auditing
             && (   !stricmp(buf, "AOF")
                 || !stricmp(buf, "COS")
                 || !stricmp(buf, "COR")
        )   )   )
        {   found = TRUE;
            break;
        }

        // No match so lets try the next file
        if (unzGoToNextFile(fp) != 0) // UNZ_OK
        {   break;
    }   }

    if (!found)
    {   DISCARD unzClose(fp);
        if (!quiet) say("No Signetics files found in ZIP archive!", TRUE);
        return FALSE;
    } // implied else

    if (unzOpenCurrentFile(fp) != 0) // UNZ_OK
    {   DISCARD unzClose(fp);
        if (!quiet) say("Corrupt ZIP archive!", TRUE);
        return FALSE;
    } // implied else

    ZIPBufferSize = (int) (info.uncompressed_size);
    ZIPBufferPtr  = malloc((unsigned int) ZIPBufferSize);

    if (unzReadCurrentFile(fp, ZIPBufferPtr, (unsigned int) ZIPBufferSize) != (int) ZIPBufferSize)
    {   DISCARD unzCloseCurrentFile(fp);
        DISCARD unzClose(fp);
        if (!quiet)
        {   say("Corrupt ZIP archive!", TRUE);
        }
        return FALSE;
    } // implied else

    DISCARD unzCloseCurrentFile(fp);
    DISCARD unzClose(fp);
    return TRUE;
}

EXPORT void auditfile(STRPTR passedname, int passedsize, FLAG user)
{   TRANSIENT UBYTE* AuditBufferPtr;
    TRANSIENT FILE*  TheLocalHandle /* = NULL */ ; // LocalHandle is a variable of winbase.h
    TRANSIENT int    howmany,
                     i,
                     length;
    TRANSIENT ULONG  crc; // crc32() is a function name
    TRANSIENT FLAG   found             = FALSE;
#ifdef WIN32
    TRANSIENT STRPTR localfilepart;
    PERSIST   TEXT   tempstring[MAX_PATH + 1];
#endif

    if (foundgames >= FOUNDGAMES_MAX)
    {   return;
    }

    if (IsZip(passedname))
    {   if (!loadzip(passedname, TRUE, TRUE))
        {   return;
        }
        passedsize     = ZIPBufferSize;
        AuditBufferPtr = (UBYTE *) ZIPBufferPtr;
    } else
    {   if (!(TheLocalHandle = fopen((char *) passedname, "rb"))) // just cast for lint
        {   return;
        }
        AuditBufferPtr = malloc((unsigned int) passedsize);
#ifdef _lint
        if (!AuditBufferPtr)
        {   return;
        }
#endif
        if (passedsize > 32768)
        {   howmany = 32768;
        } else
        {   howmany = passedsize;
        }
        if (fread(AuditBufferPtr, (size_t) howmany, 1, TheLocalHandle) != 1)
        {   free(AuditBufferPtr);
            fclose(TheLocalHandle);
            // TheLocalHandle = NULL;
            return;
        }
        fclose(TheLocalHandle);
        // TheLocalHandle = NULL;
    }

    if (passedsize <= 32768)
    {   // check whether it is a known BIN
        crc = crc32(AuditBufferPtr, passedsize);
        for (i = 0; i < KNOWNGAMES; i++)
        {   if
            (   known[i].size  == passedsize
             && known[i].crc32 == crc
            )
            {   if (known[i].dump == DUMP_GOOD || known[i].dump == DUMP_OK)
                {   strcpy(auditlist[foundgames].filename, passedname);
                    strcpy(auditlist[foundgames].gamename, known[i].name);
                    auditlist[foundgames].whichgame = (SWORD) i;
                    auditlist[foundgames].glyph     =
                    auditlist[foundgames].machine   = known[i].machine;
                    auditlist[foundgames].memmap    = known[i].memmap;
                    sidebar[foundgames]             = foundgames;
                    foundgames++;
                    foundgood++;
                    found = TRUE;
                    break; // for speed
                } else
                {   foundbad++;
                    found = TRUE;
                    if (user)
                    {   T1; textpen(TEXTPEN_VERBOSE);
                        if (known[i].dump == DUMP_SUSPECT)
                        {   zprintf("%s is suspect.\n", passedname);
                        } else
                        {   // assert(known[i].dump == DUMP_BAD);
                            zprintf("%s is bad.\n", passedname);
                        }
                        T2;
                    }
                    break; // for speed
    }   }   }   }

    if (!found)
    {   if (AuditBufferPtr[0] == 0x3A) // AOF
        {   // file is an AOF
            foundaofs++;
        } elif (AuditBufferPtr[0] == 0xBF && AuditBufferPtr[3] <= 12) // COS/COR
        {   // file is a COS or COR
            length = strlen(passedname);
            for (i = 0; i < length; i++)
            {   if (passedname[i] >= 'a' && passedname[i] <= 'z')
                {   passedname[i] -= 32; // convert to uppercase
            }   }
            strcpy(auditlist[foundgames].filename, passedname);
#ifdef AMIGA
            strcpy(auditlist[foundgames].gamename, FilePart(passedname));
#endif
#ifdef WIN32
            DISCARD GetFullPathName
            (   passedname,
                MAX_PATH,
                tempstring,
                &localfilepart
            );
            strcpy(auditlist[foundgames].gamename, localfilepart);
#endif
            switch (AuditBufferPtr[3])
            {
            case 0:
                auditlist[foundgames].glyph   =
                auditlist[foundgames].machine = MALZAK;
                auditlist[foundgames].memmap  = MEMMAP_MALZAK1;
            acase 1:
                auditlist[foundgames].glyph   =
                auditlist[foundgames].machine = MALZAK;
                auditlist[foundgames].memmap  = MEMMAP_MALZAK2;
            acase 2:
                auditlist[foundgames].glyph   =
                auditlist[foundgames].machine = GALAXIA;
                auditlist[foundgames].memmap  = MEMMAP_GALAXIA;
            acase 3:
                auditlist[foundgames].glyph   =
                auditlist[foundgames].machine = ASTROWARS;
                auditlist[foundgames].memmap  = MEMMAP_ASTROWARS;
            acase 4:
                auditlist[foundgames].glyph   =
                auditlist[foundgames].machine = CHAOS;
                auditlist[foundgames].memmap  = MEMMAP_CHAOS;
            acase 5:
                auditlist[foundgames].glyph   =
                auditlist[foundgames].machine = DOLPHIN;
                auditlist[foundgames].memmap  = MEMMAP_DOLPHIN;
            acase 6:
                auditlist[foundgames].glyph   =
                auditlist[foundgames].machine = ARCADIA;
                auditlist[foundgames].memmap  = machines[ARCADIA].memmap; // not important for it to be exactly correct
            acase 7:
                auditlist[foundgames].glyph   =
                auditlist[foundgames].machine = INTERTON;
                auditlist[foundgames].memmap  = machines[INTERTON].memmap; // not important for it to be exactly correct
            acase 8:
                auditlist[foundgames].glyph   =
                auditlist[foundgames].machine = ELEKTOR;
                auditlist[foundgames].memmap  = machines[ELEKTOR].memmap; // not important for it to be exactly correct
            acase 9:
                auditlist[foundgames].glyph   =
                auditlist[foundgames].machine = PIPBUG;
                auditlist[foundgames].memmap  = machines[PIPBUG].memmap; // not important for it to be exactly correct
            acase 10:
                auditlist[foundgames].glyph   =
                auditlist[foundgames].machine = INSTRUCTOR;
                auditlist[foundgames].memmap  = machines[INSTRUCTOR].memmap; // not important for it to be exactly correct
            acase 11:
                auditlist[foundgames].glyph   =
                auditlist[foundgames].machine = CD2650;
                auditlist[foundgames].memmap  = machines[CD2650].memmap; // not important for it to be exactly correct
            acase 12:
                auditlist[foundgames].glyph   =
                auditlist[foundgames].machine = PHUNSY;
                auditlist[foundgames].memmap  = machines[PHUNSY].memmap; // not important for it to be exactly correct
            }
            // we should check it's a supported COS/COR version!
            auditlist[foundgames].whichgame = (AuditBufferPtr[33] * 256)
                                            +  AuditBufferPtr[34];
            sidebar[foundgames]             = foundgames;
            foundgames++;
            foundsnaps++;
        } else
        {   foundunknown++;
            if (user)
            {   T1; zprintf("%s is unknown.\n", passedname); T2;
    }   }   }

    free(AuditBufferPtr);
    // AuditBufferPtr = NULL;
}

EXPORT void alloc_iobuffer(ULONG thesize)
{   // If, for example, we were playing an EAR and then a BIN was dragged
    // over, we will need to free the old IOBuffer
    if (IOBuffer)
    {   free_iobuffer();
    }
    IOBuffer = (UBYTE *) malloc(thesize);
#ifdef VERBOSE
    T1; textpen(TEXTPEN_VERBOSE); zprintf("Allocated I/O buffer.\n"); T2;
#endif
}

EXPORT void WriteWord(UWORD what)
{   // big endian 16-bit word
    IOBuffer[offset]     = (UBYTE) (what / 256);
    IOBuffer[offset + 1] = (UBYTE) (what % 256);
    offset += 2;
}

EXPORT void endofframe(void)
{   if
    (   (frames % (ULONG) frameskip == 0)
     && (    machine == ARCADIA
          || machine == INTERTON
          || machine == ELEKTOR
          || machine == INSTRUCTOR
          || machine == MALZAK
          || machine == ASTROWARS
          || machine == GALAXIA
          || changed
#ifndef SOLIDCURSOR
          || (machine == PIPBUG && frames % 25 == 0)
#endif
          || (recmode == RECMODE_RECORD && anims)
    )   )
    {   switch (machine)
        {
        case PIPBUG:
        case CHAOS:
            draw_pipbug();
        acase INSTRUCTOR:
            s_drawglows();
        acase CD2650:
            draw_cd2650();
        acase DOLPHIN:
            d_drawglows();
        acase PHUNSY:
            draw_phunsy();
        }
        updatescreen();
    }

    frames++;

    inframe = FALSE;
    if (crippled)
    {   uncripple();
    }

    if
    (   machine == ARCADIA
     && collisions
     && game
     && whichgame != -1
     && known[whichgame].hiscore != -1
     && recmode != RECMODE_PLAY
     && !paused
    )
    {   updatehiscores();
    }

    if (recmode == RECMODE_PLAY && offset >= recsize)
    {   if (loop)
        {   macro_restartplayback();
        } else
        {   macro_stop();
    }   }

    if (runtoframe)
    {   clearruntos();
        T1; textpen(TEXTPEN_DEBUG);
        zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_XVI_REACHEDFRAME,
                "Reached next frame (%d).\n\n"
            ),
            (long int) frames
        );
        T2;
        emu_pause(TRUE);
    }
#ifdef AMIGA
    if (blockarexx)
    {   blockarexx--;
    }
#endif
}

EXPORT void calc_wait(void)
{   PERSIST   ULONG howlong = 0;
    TRANSIENT ULONG waitfor;

    if
    (
#ifdef WIN32
        ctrl()
     &&
#endif
        machine != PIPBUG
     && machine != CD2650
     && machine != CHAOS
     && KeyDown(SCAN_APOSTROPHE)
    )
    {   waitfor = 1000;
    } else
    {   if (region == REGION_NTSC)
        {   if ((howlong % 3) == 0)
            {   waitfor = 16;
            } else
            {   waitfor = 17;
        }   }
        else
        {   // assert(region == REGION_PAL);
            waitfor = 20;
        }
        if
        (
#ifdef WIN32
            ctrl()
         &&
#endif
            machine != PIPBUG
         && machine != CD2650
         && machine != CHAOS
         && KeyDown(SCAN_SEMICOLON)
        )
        {   waitfor *= 4;
        } elif (turbo)
        {   waitfor = 0;
    }   }
    // waitfor is how long the frame should nominally last for (in milliseconds)

    howlong++;
    newtime = thetime();
    if (hadpaused)
    {   lateness  = 0;
        oldtime   = newtime;
        oldframes = frames;
        hadpaused = FALSE;
    } else
    {   lateness = newtime - waittill;
        // assert(lateness >= 0);
        if (lateness > waitfor) // running more than a frame late
        {   lateness = waitfor;
    }   }
    waittill = newtime + waitfor - lateness;
}

/* Timing subsystem works as follows:

   1. Get the time (absolute clock value).
   2. Calculate how long the next frame should normally go.
   3. Subtract lateness.
   4. This value is waittill, it is an absolute clock value to wait until.
   5. Emulate the frame.
   6. Get the time (absolute clock value).
   7. If we are running early, go to sleep until it is time to wake up. */

EXPORT void do_wait(void)
{   newtime = thetime();
    if (newtime < waittill)
    {
#ifdef AMIGA
        thesleep(waittill - newtime);
#endif
#ifdef WIN32
        Sleep(waittill - newtime);
#endif
        newtime = thetime();
    }

    if (newtime - oldtime >= 1000)
    {   oldtime += 1000;
        if (!paused)
        {   fps = frames - oldframes; // overflow will EVENTUALLY happen but not very harmful
            oldframes = frames;
            updatefps();
}   }   }

EXPORT void view_nvram(void)
{   int i;

    // assert(machine == MALZAK);
    // assert(memmap == MEMMAP_V);

    T1; textpen(TEXTPEN_VIEW);

    for (i = 0x1700; i <= 0x17FF; i += 16)
    {   dumprow(i);
    }
    zprintf("Potentiometer: $%02X\n\n", memory[0x14CC]);

    T2;
}

EXPORT int readbase(STRPTR input)
{   int i,
        length,
        value;

    length = strlen(input);
    if (length == 0) return OUTOFRANGE;

    if (input[0] == '\'') // ASCII
    {   switch (machine)
        {
        case ARCADIA:
            for (i = 0; i < 64; i++)
            {   if (input[1] == arcadia_chars[i])
                {   return i;
            }   }
        acase ELEKTOR:
            for (i = 0; i < 256; i++)
            {   if (input[1] == elektor_chars1[i]) // what about elektor_chars2[]?
                {   return i;
            }   }
        acase PIPBUG:
        case CD2650:
        case MALZAK:
        case CHAOS:
        case DOLPHIN:
        case PHUNSY:
            return input[1]; // ASCII
        acase INSTRUCTOR:
            for (i = 0; i <= 0x1C; i++)
            {   if (input[1] == instructor_chars[i])
                {   return i;
            }   }
        acase ASTROWARS:
            for (i = 0; i < 256; i++)
            {   if (input[1] == astrowars_chars[i])
                {   return i;
            }   }
        acase GALAXIA:
            for (i = 0; i < 128; i++)
            {   if (input[1] == galaxia_chars[i])
                {   return i;
        }   }   }
        return OUTOFRANGE;
    } elif (input[0] == '$') // hex
    {   if (length == 1 || length > 5) return OUTOFRANGE;
        for (i = 1; i < length; i++)
        {   if
            (   (input[i] >= 'a' && input[i] <= 'f')
             || (input[i] >= 'A' && input[i] <= 'F')
             || (input[i] >= '0' && input[i] <= '9')
            )
            {   ;
            } else
            {   return OUTOFRANGE;
        }   }
        DISCARD stch_l(&input[1], (void *) &value); // hexstring -> number
    } elif (input[0] == '!') // decimal
    {   if (length == 1 || length > 6) return OUTOFRANGE;
        for (i = 1; i < length; i++)
        {   if (input[i] < '0' || input[i] > '9')
            {   return OUTOFRANGE;
        }   }
        DISCARD stcd_l(&input[1], (void *) &value); // decstring -> number
    } elif (input[0] == '%') // binary
    {   if (length == 1 || length > 17) return OUTOFRANGE;
        for (i = 1; i < length; i++)
        {   if
            (   input[i] != '0'
             && input[i] != '1'
             && input[i] != '.'
             && input[i] != '#'
            )
            {   return OUTOFRANGE;
        }   }
        value = bin2int(&input[1]);         // binstring -> number
    } elif (input[0] == '@') // octal
    {   if (length == 1 || length > 6) return OUTOFRANGE; // maximum is @77777 (=$7FFF)
        for (i = 1; i < length; i++)
        {   if (input[i] < '0' || input[i] > '7')
            {   return OUTOFRANGE;
        }   }
        value = oct2int(&input[1]);
    } elif (base == BASE_DECIMAL)
    {   if (length > 5) return OUTOFRANGE;
        for (i = 0; i < length; i++)
        {   if (input[i] < '0' || input[i] > '9')
            {   return OUTOFRANGE;
        }   }
        DISCARD stcd_l(input, (void *) &value); // decstring -> number
    } elif (base == BASE_HEX)
    {   if (length > 4) return OUTOFRANGE;
        for (i = 0; i < length; i++)
        {   if
            (   (input[i] >= 'a' && input[i] <= 'f')
             || (input[i] >= 'A' && input[i] <= 'F')
             || (input[i] >= '0' && input[i] <= '9')
            )
            {   ;
            } else
            {   return OUTOFRANGE;
        }   }
        DISCARD stch_l(input, (void *) &value); // hexstring -> number
    } elif (base == BASE_OCTAL)
    {   if (length > 5) return OUTOFRANGE; // maximum is @77777 (=$7FFF)
        for (i = 0; i < length; i++)
        {   if (input[i] < '0' || input[i] > '7')
            {   return OUTOFRANGE;
        }   }
        value = oct2int(input);
    } else
    {   // assert(base == BASE_BINARY);
        if (length > 16) return OUTOFRANGE;
        for (i = 0; i < length; i++)
        {   if
            (   input[i] != '0'
             && input[i] != '1'
             && input[i] != '.'
             && input[i] != '#'
            )
            {   return OUTOFRANGE;
        }   }
        value = bin2int(input);         // binstring -> number
    }

    return value;
}

EXPORT void zprintf(const char* format, ...)
{   TRANSIENT va_list list;
    TRANSIENT int     i;
    PERSIST   char    string[2048]; // 1K is not always enough! PERSISTent for speed and so as not to blow the stack

    // Be aware that this function isn't callable recursively
    // (due to the use of persistent variables).

#ifdef AMIGA
    if (quiet)
    {   return;
    }
#endif

    va_start(list, format);
    DISCARD vsprintf((char *) string, format, list);
    va_end(list);

    if (logfile && !LogfileHandle)
    {   logfile_open();
    }

    i = 0;
    // why the fuck does puts() always write newlines, it is shit!
    while (string[i])
    {   putchar(string[i++]);
    }
    if (LogfileHandle)
    {   DISCARD fwrite(string, strlen(string), 1, LogfileHandle);
}   }

EXPORT void logfile_open(void)
{   PERSIST char string[80 + 1]; // PERSISTent for speed
    // if we allowed the user to specify their own filename, string[]
    // would need to be enlarged (to MAX_PATH + enough for the rest of the
    // string).

    // assert(logfile);

    if ((LogfileHandle = fopen(LOGFILENAME, "a")))
    {   T1; textpen(TEXTPEN_LOG);
        // we can't call zprintf() because of string corruption issues
        sprintf
        (   string,
            GetCatalogStr(CatalogPtr, MSG_LOGOPENED, "Opened %s for appending.\n\n"),
            LOGFILENAME
        );
        printf(string); // not zprintf()!
        DISCARD fwrite(string, strlen(string), 1, LogfileHandle);
        T2;
    } else
    {   T1; textpen(TEXTPEN_LOG);
        printf("Can't open %s for appending!\n\n", LOGFILENAME); // not zprintf()!
        T2;

        logfile = FALSE;
        updatemenu(MENUITEM_LOGTOFILE);
}   }

EXPORT void logfile_close(void)
{   if (LogfileHandle)
    {   T1; textpen(TEXTPEN_LOG);
        zprintf(GetCatalogStr(CatalogPtr, MSG_LOGCLOSED, "Closed %s.\n\n"), LOGFILENAME);
        T2;
        // order-dependent!
        DISCARD fclose(LogfileHandle);
        LogfileHandle = NULL;
}   }

MODULE void generate_overlaytext(void)
{   int i,
        j = 0;

    for (i = 0; i <= (int) strlen(overlays[whichoverlay][32]); i++)
    {
#ifdef WIN32
        if (overlays[whichoverlay][32][i] == '\n')
        {   pgmtext[j++] = '\r';
        }
#endif
        pgmtext[j++] = overlays[whichoverlay][32][i];
}   }

MODULE void view_mram_pvi(void)
{   zprintf
    (   "$%04X:                                            %02X %02X\n",
        pvibase + 0x0E,
        memory[pvibase + 0x0E],
        memory[pvibase + 0x0F]
    );
    zprintf
    (   "$%04X:                                            %02X %02X\n",
        pvibase + 0x1E,
        memory[pvibase + 0x1E],
        memory[pvibase + 0x1F]
    );
    zprintf
    (   "$%04X:                                            %02X %02X\n",
        pvibase + 0x4E,
        memory[pvibase + 0x4E],
        memory[pvibase + 0x4F]
    );
    dumprow(pvibase + 0x50);
    dumprow(pvibase + 0x60);
    zprintf
    (   "$%04X:                                            %02X\n",
        pvibase + 0xAD,
        memory[pvibase + 0xAD]
    );
}

EXPORT void update_memory(FLAG force)
{   PERSIST int  i;
#ifdef WIN32
    PERSIST int  oldregionstart;     // all PERSISTent for speed
    PERSIST TEXT tempstring[8 + 1]; // enough for "$123x: $"
#endif

    update_spriteviewer(force);

    if (!MemoryWindowPtr)
    {   return;
    }

#ifdef AMIGA
    if (force)
    {   update_memory_system();
    }
#endif
#ifdef WIN32
    oldregionstart = regionstart;
    regionstart = SendMessage(hRegionListBox, LB_GETTOPINDEX, 0, 0);
    if (regionstart == LB_ERR)
    {   regionstart = 0;
    }
    regionstart *= MEMGADGETS;

    if (regionstart != oldregionstart || force)
    {   force = TRUE;

        sprintf(tempstring, "$%03Xx: $",  regionstart / 16     );
        DISCARD SetDlgItemText(MemoryWindowPtr, IDL_REGION0, tempstring);
        sprintf(tempstring, "$%03Xx: $", (regionstart / 16) + 1);
        DISCARD SetDlgItemText(MemoryWindowPtr, IDL_REGION1, tempstring);
        sprintf(tempstring, "$%03Xx: $", (regionstart / 16) + 2);
        DISCARD SetDlgItemText(MemoryWindowPtr, IDL_REGION2, tempstring);
        sprintf(tempstring, "$%03Xx: $", (regionstart / 16) + 3);
        DISCARD SetDlgItemText(MemoryWindowPtr, IDL_REGION3, tempstring);
        sprintf(tempstring, "$%03Xx: $", (regionstart / 16) + 4);
        DISCARD SetDlgItemText(MemoryWindowPtr, IDL_REGION4, tempstring);
        sprintf(tempstring, "$%03Xx: $", (regionstart / 16) + 5);
        DISCARD SetDlgItemText(MemoryWindowPtr, IDL_REGION5, tempstring);
        sprintf(tempstring, "$%03Xx: $", (regionstart / 16) + 6);
        DISCARD SetDlgItemText(MemoryWindowPtr, IDL_REGION6, tempstring);
        sprintf(tempstring, "$%03Xx: $", (regionstart / 16) + 7);
        DISCARD SetDlgItemText(MemoryWindowPtr, IDL_REGION7, tempstring);
    }
#endif

    for (i = 0; i < MEMGADGETS; i++)
    {   if
        (   memory[regionstart + i] != memorycontents[i]
         || force
        )
        {   memorycontents[i] = memory[regionstart + i];
            hex1(MemoryStr[i], memorycontents[i]);
            setmemory(i, MemoryStr[i]);
}   }   }

EXPORT void dosidebar(void)
{
#ifdef WIN32
    int done = FALSE,
        i,
        length;
#endif

    switch (auditlist[sidebar[currentgame]].memmap)
    {
    case -2:
#ifdef WIN32
        if (sidebar[currentgame] == 0)
        {   strcpy(pathpart, "\\");
        } elif (sidebar[currentgame] == 1)
        {   if (pathpart[strlen(pathpart) - 1] != '\\')
            {   strcat(pathpart, "\\");
            }
            length = strlen(pathpart);
            if (length >= 2)
            {   for (i = length - 2; i >= 0; i--)
                {   if (pathpart[i] == '\\')
                    {   pathpart[i + 1] = EOS;
                        done = TRUE;
                        break;
            }   }   }
            else
            {   done = TRUE;
            }
            if (!done)
            {   pathpart[0] = EOS;
        }   }
#endif
#ifdef AMIGA
        if (sidebar[currentgame] == 0) // was (!strcmp(sidebar[currentgame].gamename, ":"))
        {   strcpy(pathpart, ":");
        } elif (sidebar[currentgame] == 1) // was (!strcmp(sidebar[currentgame].gamename, "/"))
        {   DISCARD AddPart(pathpart, "/", MAX_PATH + 1);
        }
#endif
        audit(FALSE);
        setselection();
    acase -1:
        // T1; zprintf("You clicked directory \"%s\"!\n", sidebar[currentgame].gamename); T2;
#ifdef WIN32
        if (pathpart[strlen(pathpart) - 1] != '\\')
        {   strcat(pathpart, "\\");
        }
        strcat(pathpart, auditlist[sidebar[currentgame]].filename);
#endif
#ifdef AMIGA
        DISCARD AddPart(pathpart, auditlist[sidebar[currentgame]].filename, MAX_PATH + 1);
#endif
        audit(FALSE);
        setselection();
    acase MEMMAP_ASTROWARS:
        docommand(COMMAND_ASTROWARS);
    acase MEMMAP_GALAXIA:
        docommand(COMMAND_GALAXIA);
    acase MEMMAP_MALZAK1:
        docommand(COMMAND_MALZAK1);
    acase MEMMAP_MALZAK2:
        docommand(COMMAND_MALZAK2);
    adefault:
        DISCARD strcpy(fn_game, auditlist[sidebar[currentgame]].filename);
        breakparts();
        if (!(engine_load(FALSE)))
        {   beep();
}   }   }

MODULE int load_aof(int localsize)
{   FLAG foundeof = FALSE;
    int  rc       =     1;

    cursor = 0;
    while (cursor < localsize)
    {   rc = readblock(localsize);

        if (rc == -1)
        {   foundeof = TRUE;
            break;
        } elif (rc >= 0)
        {   break;
    }   }

    if (rc <= 0)
    {   T1; zprintf("Successfully loaded AOF file.\n\n"); T2;
        if (!foundeof)
        {   T1; zprintf("Warning: no EOF block found.\n"); T2;
            return 1;
        } else
        {   return 2;
    }   }
    else
    {   T1; zprintf("Failed to load AOF file!\n\n"); T2;
        return 0;
}   }

MODULE int readblock(int localsize)
{   UBYTE blocklength,
          ourbcc, // *MUST* be UBYTE!
          theirbcc;
    int   i,
          loadaddr;

/* For normal AOF files:
   0:                                       $3A (':') (block header)
   1                 .. 4:                  load/start address
   5                 .. 6:                  block length
   7                 .. 8:                  1st BCC
   9                 .. 8+(block length*2): data
   9+(block length*2)..10+(block length*2): 2nd BCC

   For no-BCC files:
   0:                                       $3A (':') (block header)
   1                 .. 4:                  load/start address
   5                 .. 6:                  block length
   7                 .. 6+(block length*2): data

    Return codes are:
       -1  = successful, EOF block
       -2  = successful, data block
        0  = end of file
        1+ = unsuccessful */

    while (cursor < localsize && IOBuffer[cursor] <= 32)
    {   cursor++;
    } // should we warn the user if no gap is found?
    if (cursor == localsize)
    {   return 0;
    }

    if (IOBuffer[cursor] != ':')
    {   T1; zprintf("Error: block header not found!\n"); T2;
        return 1;
    }
    if (!isok(cursor + 1, cursor + 6))
    {   T1; zprintf("Error: illegal character found!\n"); T2;
        return 1;
    }
    if (readaofword(cursor + 1) >= 0x8000)
    {   T1; zprintf("Error: address is out of range!\n"); T2;
        return 1;
    }

    blocklength = readaofbyte(cursor + 5);
    if (blocklength == 0)
    {   T1; zprintf("Found EOF block.\n"); T2;
        startaddr_h = readaofbyte(cursor + 1);
        startaddr_l = readaofbyte(cursor + 3);
        cursor += 7;
        T1; zprintf("Start address is $%02X%02X.\n", startaddr_h, startaddr_l); T2;
        return -1;
    } else // should we warn the user if no blocksize != 30? or perhaps if blocksize > 30?
    {   T1; zprintf("Found data block of length %d.\n", blocklength); T2;

#ifndef NOBCC
        if (cursor + 11 + (blocklength * 2) > localsize)
        {   T1; zprintf("Error: incorrect length!\n"); T2;
            return 1;
        }
        if (!isok(cursor + 7, cursor + 10 + (blocklength * 2)))
        {   T1; zprintf("Error: illegal character found!\n"); T2;
            return 1;
        }

        ourbcc = makebcc(cursor + 1, cursor + 6);
        T1; zprintf("Our 1st BCC is $%02X.\n", ourbcc); T2;
        theirbcc = readaofbyte(cursor + 7);
        T1; zprintf("Their 1st BCC is $%02X.\n", theirbcc); T2;
        if (ourbcc != theirbcc)
        {   T1; zprintf("Error: 1st BCC is incorrect!\n"); T2;
            return 1;
        }
        ourbcc = makebcc(cursor + 9, cursor + 8 + (blocklength * 2));
        T1; zprintf("Our 2nd BCC is $%02X.\n", ourbcc); T2;
        theirbcc = readaofbyte(cursor + 9 + (blocklength * 2));
        T1; zprintf("Their 2nd BCC is $%02X.\n", theirbcc); T2;
        if (ourbcc != theirbcc)
        {   T1; zprintf("Error: 2nd BCC is incorrect!\n"); T2;
            return 1;
        }
#endif

        loadaddr = readaofword(cursor + 1);
        T1; zprintf("Load address is $%04X..$%04X.\n", loadaddr, loadaddr + blocklength - 1); T2;
        for (i = 0; i < blocklength; i++)
        {
#ifdef NOBCC
            memory[loadaddr + i] = readaofbyte(cursor + 7 + (i * 2));
#else
            memory[loadaddr + i] = readaofbyte(cursor + 9 + (i * 2));
#endif
        }
#ifdef NOBCC
        cursor +=  7 + (blocklength * 2);
#else
        cursor += 11 + (blocklength * 2);
#endif
        return -2;
}   }

MODULE FLAG save_aof(STRPTR filename, int startaddr, int endaddr, int progstart)
{   FILE* aofhandle;
    int   blocksize,
          tempend;

    // assert(startaddr <= endaddr);

#ifdef VERBOSE
    T1; zprintf("Saving %s from $%X..$%X...\n", filename, startaddr, endaddr); T2;
#endif

    if (!(aofhandle = fopen(filename, "wb")))
    {   return FALSE;
    } // implied else

    tempend = endaddr;
    do
    {   blocksize = endaddr - startaddr + 1;
        if (blocksize >= 30)
        {   blocksize = 30;
            tempend = startaddr + 29;
        } else
        {   tempend += blocksize - 1;
        }

        writeblock(aofhandle, startaddr, tempend);
        startaddr += blocksize;
    } while (startaddr <= endaddr);

    // end block

    IOBuffer[0] = ':';
    write_aof_byte(&IOBuffer[1], (UBYTE) (progstart / 256)); // 1..2
    write_aof_byte(&IOBuffer[3], (UBYTE) (progstart % 256)); // 3..4
    write_aof_byte(&IOBuffer[5],          0               ); // 5..6
    DISCARD fwrite(IOBuffer, (size_t) 7, 1, aofhandle);

#ifdef VERBOSE
    IOBuffer[7] = EOS;
    T1; zprintf("%s\n", IOBuffer); T2;
#endif

    DISCARD fclose(aofhandle);
    // aofhandle = NULL;

    return TRUE;
}

MODULE void writeblock(FILE* aofhandle, int startaddr, int endaddr)
{   int i,
        localoffset;

    // assert(aofhandle);

    IOBuffer[0] = ':';
    write_aof_byte(&IOBuffer[1], (UBYTE) (startaddr / 256));         // 1..2
    write_aof_byte(&IOBuffer[3], (UBYTE) (startaddr % 256));         // 3..4
    write_aof_byte(&IOBuffer[5], (UBYTE) (endaddr - startaddr + 1)); // 5..6
    write_aof_byte(&IOBuffer[7], makebcc(1, 6));                     // 7..8
    localoffset = 9;
    for (i = startaddr; i <= endaddr; i++)
    {   write_aof_byte(&IOBuffer[localoffset], memory[i]);
        localoffset += 2;
    }
    write_aof_byte
    (   &IOBuffer[localoffset],
        makebcc(9, localoffset - 1)
    );

    localoffset += 2;
    IOBuffer[localoffset++] = CR;
    IOBuffer[localoffset++] = LF;
    DISCARD fwrite(IOBuffer, (size_t) localoffset, 1, aofhandle);
    IOBuffer[localoffset] = EOS;

#ifdef VERBOSE
    T1; zprintf("%s", IOBuffer); T2;
#endif
}

MODULE void write_aof_byte(UBYTE* output, UBYTE value)
{   if (value / 16 >= 10)
    {   *(output    ) = (UBYTE) ('A' + (value / 16) - 10);
    } else
    {   *(output    ) = (UBYTE) ('0' + (value / 16)     );
    }
    if (value % 16 >= 10)
    {   *(output + 1) = (UBYTE) ('A' + (value % 16) - 10);
    } else
    {   *(output + 1) = (UBYTE) ('0' + (value % 16)     );
}   }

MODULE FLAG isok(int start, int end)
{   int i;

    for (i = start; i <= end; i++)
    {   if
        (   (IOBuffer[i] < '0' || IOBuffer[i] > '9')
         && (IOBuffer[i] < 'A' || IOBuffer[i] > 'F')
         && (IOBuffer[i] < 'a' || IOBuffer[i] > 'f')
        )
        {   return FALSE;
    }   }

    return TRUE;
}

MODULE UBYTE getdigit(UBYTE data)
{   if (data >= '0' && data <= '9')
    {   return (UBYTE) (data - '0');
    } elif (data >= 'A' && data <= 'F')
    {   return (UBYTE) (data - 'A' + 10);
    } elif (data >= 'a' && data <= 'f')
    {   return (UBYTE) (data - 'a' + 10);
    } else // should never happen
    {   // assert(0);
        return 0; // to avoid spurious GCC compiler warnings
}   }

MODULE UBYTE makebcc(int start, int end)
{   UBYTE ourbcc = 0;
    int   i;

    for (i = start; i <= end; i += 2)
    {   ourbcc ^= readaofbyte(i);
        if (ourbcc >= 128)
        {   ourbcc = ((ourbcc & 127) * 2) + 1;
        } else ourbcc <<= 1;
    }

    return ourbcc;
}

MODULE UBYTE readaofbyte(int start)
{   return (UBYTE)
    (   (getdigit(IOBuffer[start    ]) *   16)
      +  getdigit(IOBuffer[start + 1])
    );
}

MODULE UWORD readaofword(int start)
{   return (UWORD)
    (   (getdigit(IOBuffer[start    ]) * 4096)
      + (getdigit(IOBuffer[start + 1]) *  256)
      + (getdigit(IOBuffer[start + 2]) *   16)
      +  getdigit(IOBuffer[start + 3])
    );
}

MODULE void serializecos(int version)
{   int i, j,
        x, y;

    offset = 4;
    serialize_word(&iar);                //  4.. 5
    serialize_byte(&psu);                //      6
    serialize_byte(&psl);                //      7
    for (i = 0; i < 8; i++)
    {   serialize_word(&ras[i]);         //  8..23
    }
    for (i = 0; i < 7; i++)
    {   serialize_byte(&r[i]);           // 24..30
    }
    serialize_byte_int(&overcalc);       //     31
    serialize_byte_int(&zinterrupt);     //     32

    switch (machine)
    {
    case ARCADIA:
        serialize_word_int(&whichgame);  // 33..34
        serialize_word_int(&romsize);    // 35..36
        serialize_byte_int(&ax[0]);      //     37
        serialize_byte_int(&ay[0]);      //     38
        serialize_byte_int(&ax[1]);      //     39
        serialize_byte_int(&ay[1]);      //     40

        // UVI/RAM dump
        for (i = 0x1000; i <= 0x10FF; i++) /* 256 bytes */ serialize_byte(&memory[i]);
        for (i = 0x1200; i <= 0x12FF; i++) /* 256 bytes */ serialize_byte(&memory[i]);
        for (i = 0x1800; i <= 0x1908; i++) /* 265 bytes */ serialize_byte(&memory[i]);
        for (i = 0x1980; i <= 0x19BF; i++) /*  64 bytes */ serialize_byte(&memory[i]);
        for (i = 0x19F8; i <= 0x1AFF; i++) /* 264 bytes */ serialize_byte(&memory[i]);
                        /* = 1105 bytes */
    acase INTERTON:
        serialize_word_int(&whichgame);  // 33..34
        serialize_word_int(&romsize);    // 35..36
        serialize_byte_int(&ax[0]);      //     37
        serialize_byte_int(&ay[0]);      //     38
        serialize_byte_int(&ax[1]);      //     39
        serialize_byte_int(&ay[1]);      //     40

        for (i = 0x1000; i <= 0x15FF; i++) /*   1.5K    */ serialize_byte(&memory[i]);
        for (i = 0x1800; i <= 0x1DFF; i++) /*   1.5K    */ serialize_byte(&memory[i]);
        for (i = 0x1E88; i <= 0x1E8E; i++) /*   7 bytes */ serialize_byte(&memory[i]);
        for (i = 0x1F00; i <= 0x1F2D; i++) /*  46 bytes */ serialize_byte(&memory[i]);
        for (i = 0x1F40; i <= 0x1F6D; i++) /*  46 bytes */ serialize_byte(&memory[i]);
        for (i = 0x1F80; i <= 0x1FAD; i++) /*  46 bytes */ serialize_byte(&memory[i]);
        for (i = 0x1FC0; i <= 0x1FC3; i++) /*   4 bytes */ serialize_byte(&memory[i]);
        for (i = 0x1FC6; i <= 0x1FCD; i++) /*   8 bytes */ serialize_byte(&memory[i]);
    acase ELEKTOR:
        serialize_word_int(&whichgame);  // 33..34
        serialize_word_int(&romsize);    // 35..36
        serialize_byte_int(&ax[0]);      //     37
        serialize_byte_int(&ay[0]);      //     38
        serialize_byte_int(&ax[1]);      //     39
        serialize_byte_int(&ay[1]);      //     40
        serialize_byte(&startaddr_h);    //     41
        serialize_byte(&startaddr_l);    //     42

        for (i = 0x800; i <= 0x1FFF; i++) /*    6K      */ serialize_byte(&memory[i]);
    acase PIPBUG:
        serialize_word_int(&whichgame);  // 33..34
        serialize_word_int(&romsize);    // 35..36
        serialize_byte(&startaddr_h);    //     37
        serialize_byte(&startaddr_l);    //     38
        serialize_byte_int(&pbx);        //     39
        for (y = 0; y < PH_VDUROWS; y++)
        {   for (x = 0; x < P_VDUCOLUMNS; x++)
            {   serialize_byte(&vdu[x][y]); // 40..1063
        }   }
        serialize_byte(&p_digit[0]);     //   1064
        serialize_byte(&p_digit[1]);     //   1065
        serialize_byte(&p_digit[2]);     //   1066
        serialize_byte(&p_digit[3]);     //   1067

        for (i = 0x400; i <= 0x7FFF; i++) /*  31K       */ serialize_byte(&memory[i]);
    acase INSTRUCTOR:
        serialize_word_int(&whichgame);  // 33..34
        serialize_word_int(&romsize);    // 35..36
        serialize_byte(&startaddr_h);    //     37
        serialize_byte(&startaddr_l);    //     38
        serialize_byte(&s_glow);         //     39
        serialize_byte(&s_toggles);      //     40
        serialize_byte_int(&s_io);       //     41
        serialize_byte_int(&s_intdir);   //     42
        serialize_byte_int(&s_intselector); //  43

        for (i =    0x0; i <= 0x17FF; i++) /*  6K       */ serialize_byte(&memory[i]);
        for (i = 0x2000; i <= 0x7FFF; i++) /* 24K       */ serialize_byte(&memory[i]);
    acase CD2650:
        serialize_word_int(&whichgame);  // 33..34
        serialize_word_int(&romsize);    // 35..36
        serialize_byte(&startaddr_h);    //     37
        serialize_byte(&startaddr_l);    //     38
        for (i =  0x3FF; i <= 0x7FFF; i++) /* 31K       */ serialize_byte(&memory[i]);
    acase MALZAK:
        serialize_byte_int(&ax[0]);      //     33
        serialize_byte_int(&ay[0]);      //     34
        serialize_byte_int(&ax[1]);      //     35
        serialize_byte_int(&ay[1]);      //     36
        serialize_byte(&mg_console);     //     37
        serialize_byte_int(&m_bank1);    //     38
        serialize_byte_int(&m_bank2);    //     39
        serialize_byte_int(&malzak_x);   //     40
        serialize_byte_int(&malzak_y);   //     41

        for (x = 0; x < 16; x++)
        {   for (y = 0; y < 16; y++)
            {   serialize_byte(&m_field[x][y]); // 16*16=256
        }   }

        for (i = 0x1000; i <= 0x1FFF; i++) /* 4K    */ serialize_byte(&memory[i]);
        /* We could probably get away with omitting some memory regions. */
    acase ASTROWARS:
    case GALAXIA:
        serialize_byte_int(&ax[0]);      //     33
        serialize_byte_int(&ay[0]);      //     34
        serialize_byte_int(&ax[1]);      //     35
        serialize_byte_int(&ay[1]);      //     36
        serialize_byte(&mg_console);     //     37
        serialize_byte(&g_paddles);      //     38
        serialize_byte(&g_dips);         //     39

        for (i = 0x1400; i <= 0x1FFF; i++) /* 3K     */ serialize_byte(&memory[i]);
        /* We could probably get away with omitting some memory regions. */

        if (version >= 2) // V12.4+
        {   for (i = 0; i <= 0x3FF; i++)   /* 1K     */ serialize_byte(&g_bank[i]);
        } /* else
        {   ; // assert(version == 1); // V12.3
        } */
    acase CHAOS:
        switch (version)
        {
        case 1: // V13.0
            for (i = 0x800; i <= 0x7FFF; i++) serialize_byte(&memory[i]);
        acase 2: // V13.01+
            for (i =  0x800; i <=  0xFFF; i++) /*  2K    */ serialize_byte(&memory[i]);
            for (i = 0x2000; i <= 0x7FFF; i++) /* 24K    */ serialize_byte(&memory[i]);
            // We could probably get away with omitting some memory regions.
        }

        serialize_byte_int(&pbx);
        for (y = 0; y < PH_VDUROWS; y++)
        {   for (x = 0; x < H_VDUCOLUMNS; x++) // 64*32=2048
            {   serialize_byte(&vdu[x][y]);
        }   }
    acase DOLPHIN:
        for (i =  0x200; i <=  0xBFF; i++) /* 2.5K   */ serialize_byte(&memory[i]);
        if (version >= 2)
        {   serialize_word_int(&whichgame);
            serialize_word_int(&romsize);
            serialize_byte(&dolphin_leds[0]);
            serialize_byte(&dolphin_leds[1]);
            serialize_byte(&dolphin_leds[2]);
            serialize_byte(&dolphin_leds[3]);
        }
    acase PHUNSY:
        serialize_byte(&banked);
        // 4K + (13 * 2K) + 8K + (15 * 16K) = 278K
        for (i =  0x800; i <= 0x17FF; i++)              serialize_byte(&memory[i]);
        for (i = 0; i <= 15; i++)
        {   if (i >= 1 && i <= 3)
            {   continue;
            }
            for (j = 0; j <=  0x7FF; j++)               serialize_byte(&u_bank[i][j]);
        }
        for (i = 1; i <= 15; i++)
        {   for (j = 0; j <= 0x3FFF; j++)               serialize_byte(&q_bank[i][j]);
        }

        if (version >= 2) // V15.0
        {   for (i = 0x2000; i <= 0x3FFF; i++) /* 8K */ serialize_byte(&memory[i]);
            serialize_word_int(&whichgame); // 284,706..284,707
            serialize_word_int(&romsize);   // 284,708..284,709
            if (version >= 3) // V15.1
            {   serialize_byte(&startaddr_h);        // 284,710
                serialize_byte(&startaddr_l);        // 284,711
                for (i = 1; i <= 3; i++)
                {   for (j = 0; j <= 0x7FF; j++)        serialize_byte(&u_bank[i][j]);
                }
                for (i = 0; i <= 0x3FFF; i++)           serialize_byte(&q_bank[0][i]);
}   }   }   }

EXPORT void serialize_byte(UBYTE* var)
{   if (serializemode == SERIALIZE_WRITE)
    {   WriteByte(*var);
    } else
    {   // assert(serializemode == SERIALIZE_READ);
        *var = ReadByte();
}   }
EXPORT void serialize_word(UWORD* var)
{   if (serializemode == SERIALIZE_WRITE)
    {   WriteWord(*var);
    } else
    {   // assert(serializemode == SERIALIZE_READ);
        *var = ReadWord();
}   }
EXPORT void serialize_byte_int(int* var)
{   if (serializemode == SERIALIZE_WRITE)
    {   WriteByte((UBYTE) (*var));
    } else
    {   // assert(serializemode == SERIALIZE_READ);
        *var = (int) ReadByte();
}   }
EXPORT void serialize_byte_ulong(ULONG* var)
{   if (serializemode == SERIALIZE_WRITE)
    {   WriteByte((UBYTE) (*var));
    } else
    {   // assert(serializemode == SERIALIZE_READ);
        *var = (ULONG) ReadByte();
}   }
EXPORT void serialize_word_int(int* var)
{   if (serializemode == SERIALIZE_WRITE)
    {   WriteWord((UWORD) (*var));
    } else
    {   // assert(serializemode == SERIALIZE_READ);
        *var = (int) ReadWord();
        if (*var == 0xFFFF)
        {   *var = -1;
}   }   }

EXPORT void makeaudittext(STRPTR outputstring)
{   sprintf
    (   outputstring,
        GetCatalogStr(CatalogPtr, MSG_AUDITTEXT,
        "In directory \"%s\":\n\n" \
        "%d good dump(s) found.\n" \
        "%d bad/suspect dump(s) found.\n" \
        "%d snapshot(s)/recording(s) found.\n" \
        "%d AOF(s) found.\n" \
        "%d icon(s) found.\n" \
        "%d directory/ies found.\n" \
        "%d unknown file(s) found.\n"),
        pathpart,
        foundgood,
        foundbad,
        foundsnaps,
        foundaofs,
        foundicons,
        founddirs,
        foundunknown
    );
}

EXPORT void calcspritecolour(void)
{   // viewingsprite and pvibase are already set appropriately

    switch (machine)
    {
    case ARCADIA:
        switch (viewingsprite)
        {
        case 0:
            sprviewcolour = (memory[A_SPRITES01CTRL] & 0x38) >> 3;
        acase 1:
            sprviewcolour =  memory[A_SPRITES01CTRL] & 0x07;
        acase 2:
            sprviewcolour = (memory[A_SPRITES23CTRL] & 0x38) >> 3;
        acase 3:
            sprviewcolour =  memory[A_SPRITES23CTRL] & 0x07;
        adefault: // 4..7
            sprviewcolour = WHITE; // we should look at the current foreground text colour and use that instead
        }
        if (flagline && (psu & PSU_F))
        {   sprviewcolour = from_a[7 - sprviewcolour];
        } else
        {   sprviewcolour = from_a[    sprviewcolour];
        }
    acase INTERTON:
    case ELEKTOR:
    case ASTROWARS:
    case GALAXIA:
    case MALZAK:
        switch (viewingsprite % 4)
        {
        case 0:
            sprviewcolour = (memory[pvibase + PVI_SPR01COLOURS] >> 3) & 7;
        acase 1:
            sprviewcolour =  memory[pvibase + PVI_SPR01COLOURS]       & 7;
        acase 2:
            sprviewcolour = (memory[pvibase + PVI_SPR23COLOURS] >> 3) & 7;
        acase 3:
            sprviewcolour =  memory[pvibase + PVI_SPR23COLOURS]       & 7;
        }
        switch (machine)
        {
        case ASTROWARS:
            sprviewcolour = from_g_x_spr[sprviewcolour];
        acase GALAXIA:
            sprviewcolour = from_g_w_spr[sprviewcolour];
        acase MALZAK:
            sprviewcolour = from_m_spr[sprviewcolour];
}   }   }

EXPORT void loadconfig(void)
{   ULONG thesize;
    FILE* TheLocalHandle /* = NULL */ ; // LocalHandle is a variable of winbase.h

    thesize = getsize(fn_config);
    if (thesize < SNPSIZE_MAX)
    {   alloc_iobuffer(thesize);
        if ((TheLocalHandle = fopen(fn_config, "rb")))
        {   if (fread(IOBuffer, thesize, 1, TheLocalHandle) == 1)
            {   if
                (   IOBuffer[0] == 0xC6
                 && IOBuffer[1] == 0x02
                 && IOBuffer[2] >= OLDESTCONFIGVERSION
                 && IOBuffer[2] <= NEWCONFIGVERSION
                )
                {   serializemode = SERIALIZE_READ;
                    serializeconfig();
                    system_loadedconfig();

                    // textpen(TEXTPEN_VERBOSE);
                    // zprintf("Read config!\n");
            }   }

            DISCARD fclose(TheLocalHandle);
            // TheLocalHandle = NULL;
        }
        free_iobuffer();
}   }

EXPORT void saveconfig(void)
{   FILE* TheLocalHandle /* = NULL */ ;

#ifdef AMIGA
    configmachine = CONFIGMACHINE_AMIGA;
#endif
#ifdef WIN32
    configmachine = CONFIGMACHINE_IBM;
#endif
    configversion = NEWCONFIGVERSION;
    serializemode = SERIALIZE_WRITE;
    alloc_iobuffer(1024 + MAX_PATH);
    serializeconfig();
    strcpy(&IOBuffer[offset], pathpart);
    offset += strlen(pathpart) + 1;

    if ((TheLocalHandle = fopen(fn_config, "wb")))
    {   DISCARD fwrite(IOBuffer, offset, 1, TheLocalHandle);
        DISCARD fclose(TheLocalHandle);
        // TheLocalHandle = NULL;

        // textpen(TEXTPEN_VERBOSE);
        // zprintf("Wrote config!\n");
    }
    free_iobuffer();
}

EXPORT void longcomma(ULONG value, STRPTR out)
{   ULONG number,
          thevalue = value,
          i;
    UBYTE calc;
    FLAG  started  = FALSE;

    /* With 32 bits the maximum possible is 4 gigabytes
       (about 4 billion bytes).

       " ,   ,   ,   "
        01234567890123 */

    number = ONE_BILLION;
    for (i = 0; i <= 12; i++)
    {   if (i == 1 || (i >= 5 && ((i - 1) % 4) == 0))
        {   if (started)
            {   *(out + i) = ',';
            } else
            {   *(out + i) = ' ';
        }   }
        else
        {   if (thevalue >= number)
            {   calc = (UBYTE) (thevalue / number); // calc = (UBYTE) thevalue / number; was wrong!
            } else
            {   calc = 0;
            }
            if (calc || started || i == 12)
            {   *(out + i) = (char) ('0' + calc);
                started = TRUE;
                thevalue %= number;
            } else
            {   *(out + i) = ' ';
            }
            number /= 10;
    }   }

    // assert(*(out + 13) == EOS);
}

