/* WinArcadia 15
   A multi-emulator of Signetics-based machines

Known WinArcadia issues:
 (a) Some supposedly random noise samples sound more like tones, especially
     after playing a tone. Eg. SHOOTGAL.BIN, HUNTING.BIN demonstrate this.
     This could be caused by oddities in the host randomization, Windows
     multimedia subsystem, or sound hardware.
 (b) Very short noises are not really audible.
 (c) Network play does not seem to work except when the server and client
     are both on the same local computer (timeout issues?). (This probably
     applies to AmiArcadia also.)
 (d) Zoom button glyph should change between "maximize" and "restore" as
     appropriate (but how?).
 (e) If we are playing a tone (at least) at loud host volume, and then
     turn off sound, and then change host volume to soft, and then turn
     the sound back on, we get a moment of loud sound before it goes soft.
     This is yet another Windows bug, as it also occurs with other
     software. */

#ifdef _lint
    #define _inline
    #define WIN32
#endif

#include "ibm.h"

#include <commctrl.h>
#include <mmreg.h>    // for eg. AVIFileOpen()
#include <vfw.h>      // for eg. PAVIFILE
#include "dinput.h"

#include "resource.h"

#include <stdio.h>
#include <fcntl.h>    // for _O_TEXT
#include <io.h>       // for _open_osfhandle()

#define EXEC_TYPES_H
#include "aa.h"
#define CATCOMP_NUMBERS
#define CATCOMP_BLOCK
#include "aa_strings.h"

#define WM_MOUSEWHEEL 0x20A
#define WHEELDELTA      120
#define SIDEBARWIDTH    118

#define BELOW_NORMAL_PRIORITY_CLASS 0x00004000
#define ABOVE_NORMAL_PRIORITY_CLASS 0x00008000

/* The emulators currently run at these speeds...?
   For NTSC Arcadia:          0.879840MHz ( 879.840 KHz).
   For PAL  Arcadia:          0.733200MHz ( 733.200 KHz).
   For PAL  Interton/Elektor: 0.873600MHz ( 873.600 KHz).
   For NTSC Interton/Elektor: 1.048320MHz (1048.320 KHz).
   "Clock Frequency:          3.546895Mhz +/- 35Hz" - Fountain
                                                      Service Manual

NTSC master frequency is 3,579,545Hz.
PAL  master frequency is 3,546,895Hz.
*/

EXPORT const char g_szClassName[] = "WinArcadia";

#define IDC_TOOLBAR       102
#define IDC_FPS           103 // actual FPS value
#define IDL_FPS           104 // "FPS:" label
// #define IDC_STATIC     105 // unused

// this #ifndef is for Visual Studio 9 compatibility
#ifndef _WINUSER_
    typedef struct
    {   DWORD  vkCode;
        DWORD  scanCode;
        DWORD  flags;
        DWORD  time;
        ULONG* dwExtraInfo;
    } KBDLLHOOKSTRUCT, *PKBDLLHOOKSTRUCT;
#endif

// EXPORTED VARIABLES ----------------------------------------------------

EXPORT UWORD             jf[2];
EXPORT RECT              therect;
EXPORT POINT             ThePoint;
EXPORT FLAG              capslock             = FALSE, // what if user starts emulator with Caps Lock on?
                         jump,
                         lame                 = FALSE,
                         lmb                  = FALSE,
                         mmb                  = FALSE,
                         opening              = FALSE,
                         SampleOpened[SAMPLES],
                         SamplePrepared[SAMPLES],
                         showingpointer,
                         upgrade              = FALSE;
                                               // AOF   BIN   COR   COS   PGM   TVC   ZIP
EXPORT int               assoc[7]             = { TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, FALSE },
                         charset              = CHARSET_DOS,
                         clientheight,
                         clientwidth,
                         currentgame          = -1,
                         destwidth,
                         destheight,
                         fastsize,
                         fastwide,
                         hostcontroller[2]    = { CONTROLLER_NONE, CONTROLLER_NONE },
                         hostvolume           = HOSTVOLUME_DEFAULT,
                         language             = LANGUAGE_ENG,
                         needaudit            = FALSE,
                         priority             = PRI_NORMAL,
                         toolbarheight,
                         useinternal          = FALSE,
                         rastwidth,
                         rastheight,
                         realsize,
                         realwide,
                         reassociate          = FALSE,
                         showlegend           = TRUE,
                         sidebarwidth         = SIDEBARWIDTH,
                         sprmakerexists       = FALSE,
                         sourcewidth,
                         sourceheight,
                         txtmakerexists       = FALSE,
                         useff                = TRUE,
                         winheight,
                         winwidth,
                         winleftx             = -1,
                         wintopy              = -1,
                         xmas,
                         xoffset,
                         yoffset;
EXPORT ULONG             keyframes[SCANCODES],
                         pens[2][PENS],
                         SampleLength[SAMPLES];
EXPORT ULONG            *ctrldisplay          = NULL,
                        *display              = NULL,
                        *stars                = NULL;
EXPORT UBYTE             button[2][8]         = { { 1,2,3,4,5,6,7,8 }, { 1,2,3,4,5,6,7,8 } },
                         jx[2], jy[2],
                         KeyMatrix[64]; // to allow keycodes up to 511
EXPORT ASCREEN           screen[BOXWIDTH][BOXHEIGHT],
                         ctrlscreen[MINIWIDTH][MINIHEIGHT];
EXPORT UBYTE*            SampleBuffer[SAMPLES];
EXPORT HWND              ConsoleWindowPtr     = NULL,
                         ControlsWindowPtr    = NULL,
                         DIPsWindowPtr        = NULL,
                         FPSPtr               = NULL,
                         GameInfoWindowPtr    = NULL,
                         ListeningWindowPtr   = NULL,
                         MainWindowPtr        = NULL,
                         MemoryWindowPtr      = NULL,
                         MonitorWindowPtr     = NULL,
                         PaletteWindowPtr     = NULL,
                         SpriteWindowPtr      = NULL,
                         hSideBar             = NULL,
                         hToolbar             = NULL;
EXPORT HWAVEOUT          hSample[SAMPLES];
EXPORT HCURSOR           hArrow,
                         hCross;
EXPORT HFONT             hRussianFont         = NULL;
EXPORT HMENU             MenuPtr              = NULL,
                         PopUpMenuPtr         = NULL;
EXPORT HBITMAP           OurhBitMap           = NULL;
EXPORT TEXT              globalstring[32 + 1 + 1],
                         hostname[256 + 1], // should hopefully be long enough
                         olddir[MAX_PATH + 1],
                         ProgDir[MAX_PATH + 1] = "",
                         ProgName[MAX_PATH + 1],
                         titlebartext[MAX_PATH + 1 + TITLEBARTEXTLENGTH + 6 + 80]; // " () ()" is 6 bytes
EXPORT HDC               OurhDC               = NULL;
EXPORT int               CatalogPtr, // never used
                         joys                 = 0,
                         unit[2]              = {-1, -1}; // must be signed!
EXPORT HINSTANCE         InstancePtr          = NULL;
EXPORT HANDLE            hConsole             = NULL;
#ifdef OLDWAY
EXPORT struct
{   BITMAPINFOHEADER Header;
    DWORD            Colours[3];
} TheBitMapInfo;
#endif
#ifdef NEWWAY
EXPORT BITMAPINFO        TheBitMapInfo;
#endif
EXPORT const struct HostMachineStruct hostmachines[MACHINES] =
{ {    "WinArcadia " SHORTVERSION, IDI_ARCADIA    },
  {   "WinInterton " SHORTVERSION, IDI_INTERTON   },
  {    "WinElektor " SHORTVERSION, IDI_ELEKTOR    },
  {     "WinPIPBUG " SHORTVERSION, IDI_PIPBUG     },
  { "WinInstructor " SHORTVERSION, IDI_INSTRUCTOR },
  {     "WinCD2650 " SHORTVERSION, IDI_CD2650     },
  {  "WinAstroWars " SHORTVERSION, IDI_ASTROWARS  },
  {    "WinGalaxia " SHORTVERSION, IDI_GALAXIA    },
  {     "WinMalzak " SHORTVERSION, IDI_MALZAK     },
  {      "WinChaos " SHORTVERSION, IDI_CHAOS      },
  {    "WinDolphin " SHORTVERSION, IDI_DOLPHIN    },
  {     "WinPHUNSY " SHORTVERSION, IDI_PHUNSY     }
};

EXPORT const DWORD       joyfires[8] = { JOYFIRE1, JOYFIRE2, JOYFIRE3, JOYFIRE4, JOYA, JOYB, JOYAUTOFIRE, JOYPAUSE };

// MODULE VARIABLES ------------------------------------------------------

MODULE FLAG              needcontrols      = FALSE,
                         needdips          = FALSE,
                         needmemory        = FALSE,
                         needmonitor       = FALSE,
                         needpalette       = FALSE,
                         needgameinfo      = FALSE,
                         needsprite        = FALSE,
                         ready             = FALSE;
MODULE TEXT              coinstring[2][40 + 1],
                         fpsstring[10 + 1],
                         SysInfoName[MAX_PATH + 1];
MODULE HANDLE            ProcessPtr        = NULL,
                         ThreadPtr         = NULL;
MODULE HIMAGELIST        himl;
MODULE HMENU             SystemMenuPtr     = NULL;
MODULE HWND              LabelPtr          = NULL;
MODULE int               configversion,
                         DisplayWidth,
                         DisplayHeight,
                         dlgtype,
                         highestid         = 0,
                         menuheight,
                         minwinwidth,
                         side,
                         sidebarheight,
                         sidebartopy,
                         titleheight,
                         whichkey;
MODULE ULONG             animframe         = (ULONG) -1,
                         oldfps            = ~0;
MODULE UWORD             temp_console[4],
                         temp_keypads[2][NUMKEYS];
MODULE HFONT             hBoldFont         = NULL,
                         hFont             = NULL;
MODULE struct
{   BITMAPINFOHEADER Header;
    DWORD            Colours[3];
} CtrlBitMapInfo;

#define MENUBITMAPS 33 // counting from 1
MODULE struct
{   const DWORD   menuitem,
                  resource;
    const FLAG    checkmark;
// uninitialized
          HBITMAP bitmap;
} mib[MENUBITMAPS] =
{ { ID_FILE_RESET,          IDB_M_RESET      , FALSE },
  { ID_FILE_OPEN,           IDB_M_OPEN       , FALSE },
  { ID_FILE_SAVESNP,        IDB_M_SAVESNP    , FALSE },
  { ID_FILE_SAVEILBM,       IDB_M_SAVEILBM   , FALSE },
  { ID_FILE_EXIT,           IDB_M_EXIT       , FALSE },
  { ID_EDIT_COPY,           IDB_M_COPY       , FALSE },
  { ID_EDIT_PASTETEXT,      IDB_M_PASTE      , FALSE },
  { ID_DEBUG_STEP,          IDB_M_STEP       , FALSE },
  { ID_DEBUG_STEPOVER,      IDB_M_STEPOVER   , FALSE },
  { ID_DEBUG_SETBP,         IDB_M_SETBP      , FALSE },
  { ID_DEBUG_CLEARBPS,      IDB_M_CLEARBPS   , FALSE },
  { ID_LANGUAGE_ENG,        IDB_M_ENG        , TRUE  },
  { ID_LANGUAGE_HOL,        IDB_M_HOL        , TRUE  },
  { ID_LANGUAGE_FRA,        IDB_M_FRA        , TRUE  },
  { ID_LANGUAGE_GER,        IDB_M_GER        , TRUE  },
  { ID_LANGUAGE_ITA,        IDB_M_ITA        , TRUE  },
  { ID_LANGUAGE_RUS,        IDB_M_RUS        , TRUE  },
  { ID_HELP_MANUAL,         IDB_M_MANUAL     , FALSE },
  { ID_HELP_ABOUT,          IDB_M_ABOUT      , FALSE },
  { ID_MACHINE_ARCADIA,     IDB_M_A          , TRUE  },
  { ID_MACHINE_INTERTON,    IDB_M_I          , TRUE  },
  { ID_MACHINE_ELEKTOR,     IDB_M_E          , TRUE  },
  { ID_MACHINE_PIPBUG,      IDB_M_P          , TRUE  },
  { ID_MACHINE_INSTRUCTOR,  IDB_M_S          , TRUE  },
  { ID_MACHINE_CD2650,      IDB_M_C          , TRUE  },
  { ID_MACHINE_ASTROWARS,   IDB_M_ASTROWARS  , TRUE  },
  { ID_MACHINE_GALAXIA,     IDB_M_GALAXIA    , TRUE  },
  { ID_MACHINE_M1,          IDB_M_M          , TRUE  },
  { ID_MACHINE_M2,          IDB_M_M          , TRUE  },
  { ID_MACHINE_CHAOS,       IDB_M_H          , TRUE  },
  { ID_MACHINE_DOLPHIN,     IDB_M_D          , TRUE  },
  { ID_LANGUAGE_SPA,        IDB_M_SPA        , TRUE  },
  { ID_MACHINE_PHUNSY,      IDB_M_PHUNSY     , TRUE  }
};
EXPORT const struct RearrangeStruct rearrange[48] = {
{ { 1,2,3,4 }, { IDC_1STRA1234, IDC_2NDRA1234 } },
{ { 1,2,4,3 }, { IDC_1STRA1243, IDC_2NDRA1243 } },
{ { 1,3,2,4 }, { IDC_1STRA1324, IDC_2NDRA1324 } },
{ { 1,3,4,2 }, { IDC_1STRA1342, IDC_2NDRA1342 } },
{ { 1,4,2,3 }, { IDC_1STRA1423, IDC_2NDRA1423 } },
{ { 1,4,3,2 }, { IDC_1STRA1432, IDC_2NDRA1432 } },
{ { 2,1,3,4 }, { IDC_1STRA2134, IDC_2NDRA2134 } },
{ { 2,1,4,3 }, { IDC_1STRA2143, IDC_2NDRA2143 } },
{ { 2,3,1,4 }, { IDC_1STRA2314, IDC_2NDRA2314 } },
{ { 2,3,4,1 }, { IDC_1STRA2341, IDC_2NDRA2341 } },
{ { 2,4,1,3 }, { IDC_1STRA2413, IDC_2NDRA2413 } },
{ { 2,4,3,1 }, { IDC_1STRA2431, IDC_2NDRA2431 } },
{ { 3,1,2,4 }, { IDC_1STRA3124, IDC_2NDRA3124 } },
{ { 3,1,4,2 }, { IDC_1STRA3142, IDC_2NDRA3142 } },
{ { 3,2,1,4 }, { IDC_1STRA3214, IDC_2NDRA3214 } },
{ { 3,2,4,1 }, { IDC_1STRA3241, IDC_2NDRA3241 } },
{ { 3,4,1,2 }, { IDC_1STRA3412, IDC_2NDRA3412 } },
{ { 3,4,2,1 }, { IDC_1STRA3421, IDC_2NDRA3421 } },
{ { 4,1,2,3 }, { IDC_1STRA4123, IDC_2NDRA4123 } },
{ { 4,1,3,2 }, { IDC_1STRA4132, IDC_2NDRA4132 } },
{ { 4,2,1,3 }, { IDC_1STRA4213, IDC_2NDRA4213 } },
{ { 4,2,3,1 }, { IDC_1STRA4231, IDC_2NDRA4231 } },
{ { 4,3,1,2 }, { IDC_1STRA4312, IDC_2NDRA4312 } },
{ { 4,3,2,1 }, { IDC_1STRA4321, IDC_2NDRA4321 } },
{ { 5,6,7,8 }, { IDC_1STRA5678, IDC_2NDRA5678 } },
{ { 5,6,8,7 }, { IDC_1STRA5687, IDC_2NDRA5687 } },
{ { 5,7,6,8 }, { IDC_1STRA5768, IDC_2NDRA5768 } },
{ { 5,7,8,6 }, { IDC_1STRA5786, IDC_2NDRA5786 } },
{ { 5,8,6,7 }, { IDC_1STRA5867, IDC_2NDRA5867 } },
{ { 5,8,7,6 }, { IDC_1STRA5876, IDC_2NDRA5876 } },
{ { 6,5,7,8 }, { IDC_1STRA6578, IDC_2NDRA6578 } },
{ { 6,5,8,7 }, { IDC_1STRA6587, IDC_2NDRA6587 } },
{ { 6,7,5,8 }, { IDC_1STRA6758, IDC_2NDRA6758 } },
{ { 6,7,8,5 }, { IDC_1STRA6785, IDC_2NDRA6785 } },
{ { 6,8,5,7 }, { IDC_1STRA6857, IDC_2NDRA6857 } },
{ { 6,8,7,5 }, { IDC_1STRA6875, IDC_2NDRA6875 } },
{ { 7,5,6,8 }, { IDC_1STRA7568, IDC_2NDRA7568 } },
{ { 7,5,8,6 }, { IDC_1STRA7586, IDC_2NDRA7586 } },
{ { 7,6,5,8 }, { IDC_1STRA7658, IDC_2NDRA7658 } },
{ { 7,6,8,5 }, { IDC_1STRA7685, IDC_2NDRA7685 } },
{ { 7,8,5,6 }, { IDC_1STRA7856, IDC_2NDRA7856 } },
{ { 7,8,6,5 }, { IDC_1STRA7865, IDC_2NDRA7865 } },
{ { 8,5,6,7 }, { IDC_1STRA8567, IDC_2NDRA8567 } },
{ { 8,5,7,6 }, { IDC_1STRA8576, IDC_2NDRA8576 } },
{ { 8,6,5,7 }, { IDC_1STRA8657, IDC_2NDRA8657 } },
{ { 8,6,7,5 }, { IDC_1STRA8675, IDC_2NDRA8675 } },
{ { 8,7,5,6 }, { IDC_1STRA8756, IDC_2NDRA8756 } },
{ { 8,7,6,5 }, { IDC_1STRA8765, IDC_2NDRA8765 } }
};

// IMPORTED VARIABLES ----------------------------------------------------

// ReAction gadgets
IMPORT       ULONG  analog,
                    autofire,
                    collisions,
                    demultiplex,
                    flagline,
                    fps,
                    paused,
                    sound,
                    swapped,
                    turbo;
// other ULONGs here
IMPORT       ULONG  frames,
                    memflags[32768],
                    region;
IMPORT       UWORD  default_console[4],
                    default_keypads[2][NUMKEYS],
                    console[4],
                    keypads[2][NUMKEYS],
                    mirror_r[32768],
                    mirror_w[32768];
IMPORT       UBYTE  g_dips,
                    memory[32768],
                    mg_console,
                    psu,
                    s_toggles,
                    sx[2],
                    sy[2];
IMPORT       int    anims,
                    autopause,
                    autosave,
                    base,
                    cli,
                    colourset,
                    confine,
                    confirmable,
                    connected,
                    console_start,
                    console_a,
                    console_b,
                    console_reset,
                    consoleopen,
                    crippled,
                    dotspace,
                    drawcorners,
                    drawunlit,
                    echoconsole,
                    echoprinter,
                    enhancestars,
                    fastcpu,
                    foundgames,
                    frameskip,
                    fullscreen,
                    game,
                    highres,
                    ignorepaint,
                    inverse,
                    logeprom,
                    logfile,
                    loginstructions,
                    logioports,
                    logbios,
                    logreadwrites,
                    logsubroutines,
                    loop,
                    lowercase,
                    machine,
                    memmap,
                    offset,
                    pauselog,
                    rastlines,
                    rastn,
                    recmode,
                    regionstart,
                    requirebutton,
                    retune,
                    rotate,
                    s_io,
                    s_intselector,
                    s_intdir,
                    scale2x,
                    sensegame,
                    sensitivity,
                    shadowing,
                    showmenubar,
                    showpointer,
                    showsidebar,
                    showtitlebar,
                    showtoolbar,
                    size,
                    sortby,
                    sparex,
                    sparey,
                    speakoffset,
                    spritesvisible,
                    stretch,
                    tonum,
                    tolimit,
                    trace,
                    traceverbose,
                    undither,
                    useansi,
                    useguidepixel,
                    usehacks,
                    usespeech,
                    watchreads,
                    wsc,
                    whichgame,
                    wide;
IMPORT       TEXT   filepart[MAX_PATH + 1], // file only
                    friendly[21 + 1],
                    fromstr[5 + 1 + 1],
                    joyname[2][MAX_PATH],
                    tostr[5 + 1 + 1],
                    lastchange[32 + 1 + 1],
                    pathpart[MAX_PATH + 1],
                    fn_config[MAX_PATH + 1],
                    fn_game[MAX_PATH + 1]; // the entire pathname (path and file)
IMPORT       FLAG   changed;
IMPORT       STRPTR colourname[8];
IMPORT       RECT   axesrect;
IMPORT       UBYTE* IOBuffer;
#ifdef USEDIRECTSOUND
    IMPORT LPDIRECTSOUNDBUFFER     inst[7];
    IMPORT struct IDirectSound*    lpds;
#endif
IMPORT struct KindStruct           filekind[KINDS];
IMPORT const struct MachineStruct  machines[MACHINES];
IMPORT struct HiScoreStruct        hiscore[HISCORES];
IMPORT struct RTCStruct            rtc;
IMPORT struct IDirectInput*        lpdi;
IMPORT struct IDirectInputDevice2* ffjoy2[2];
IMPORT const int                   gadgetimage[9][2],
                                   menucode[MENUITEMS],
                                   menucode2[MENUOPTS];
IMPORT const ULONG                 defpencolours[2][3][GUESTCOLOURS];

// MODULE ARRAYS ---------------------------------------------------------

// (none)

// EXPORTED STRUCTURES----------------------------------------------------

/* Keys we don't hear: both Windows keys, both Alt keys, F10.
Also Power, Sleep, Wake, Fn.
Esc and PrtSc both use scancode 1. */
EXPORT struct KeyNameStruct keyname[SCANCODES] = {
{   "(0)",     0  ,  0  ,  0  , FALSE}, // $00
{   "PrtSc",  27  , 27  ,  0  , FALSE}, // $01
{   "a1",     '1' , '1' , '!' , FALSE}, // $02
{   "a2",     '2' , '2' , '@' , FALSE}, // $03
{   "a3",     '3' , '3' , '#' , FALSE}, // $04
{   "a4",     '4' , '4' , '$' , FALSE}, // $05
{   "a5",     '5' , '5' , '%' , FALSE}, // $06
{   "a6",     '6' , '6' , '^' , FALSE}, // $07
{   "a7",     '7' , '7' , '&' , FALSE}, // $08
{   "a8",     '8' , '8' , '*' , FALSE}, // $09
{   "a9",     '9' , '9' , '(' , FALSE}, // $0A
{   "a0",     '0' , '0' , ')' , FALSE}, // $0B
{   "a-",     '-' , '-' , '_' , FALSE}, // $0C
{   "=",      '=' , '=' , '+' , FALSE}, // $0D
{   "BkSp",   '\b', '\b', '\b', FALSE}, // $0E
{   "Tab",    '\t', '\t', '\t', FALSE}, // $0F
{   "Q",      'q' , 'Q' , 'Q' , FALSE}, // $10
{   "W",      'w' , 'W' , 'W' , FALSE}, // $11
{   "E",      'e' , 'E' , 'E' , FALSE}, // $12
{   "R",      'r' , 'R' , 'R' , FALSE}, // $13
{   "T",      't' , 'T' , 'T' , FALSE}, // $14
{   "Y",      'y' , 'Y' , 'Y' , FALSE}, // $15
{   "U",      'u' , 'U' , 'U' , FALSE}, // $16
{   "I",      'i' , 'I' , 'I' , FALSE}, // $17
{   "O",      'o' , 'O' , 'O' , FALSE}, // $18
{   "P",      'p' , 'P' , 'P' , FALSE}, // $19
{   "[",      '[' , '[' , '{' , FALSE}, // $1A
{   "]",      ']' , ']' , '}' , FALSE}, // $1B
{   "aE",     13  , 13  , 13  , FALSE}, // $1C was \r
{   "LCtrl",   0  ,  0  ,  0  , TRUE }, // $1D
{   "A",      'a' , 'A' , 'A' , FALSE}, // $1E
{   "S",      's' , 'S' , 'S' , FALSE}, // $1F
{   "D",      'd' , 'D' , 'D' , FALSE}, // $20
{   "F",      'f' , 'F' , 'F' , FALSE}, // $21
{   "G",      'g' , 'G' , 'G' , FALSE}, // $22
{   "H",      'h' , 'H' , 'H' , FALSE}, // $23
{   "J",      'j' , 'J' , 'J' , FALSE}, // $24
{   "K",      'k' , 'K' , 'K' , FALSE}, // $25
{   "L",      'l' , 'L' , 'L' , FALSE}, // $26
{   ";",      ';' , ';' , ':' , FALSE}, // $27
{   "'",      '\'', '\'', '"' , FALSE}, // $28
{   "`",      '`' , '`' , '~' , FALSE}, // $29
{   "LShift",  0  ,  0  ,  0  , FALSE}, // $2A
{   "\\",     '\\', '\\', '|' , FALSE}, // $2B
{   "Z",      'z' , 'Z' , 'Z' , FALSE}, // $2C
{   "X",      'x' , 'X' , 'X' , FALSE}, // $2D
{   "C",      'c' , 'C' , 'C' , FALSE}, // $2E
{   "V",      'v' , 'V' , 'V' , FALSE}, // $2F
{   "B",      'b' , 'B' , 'B' , FALSE}, // $30
{   "N",      'n' , 'N' , 'N' , FALSE}, // $31
{   "M",      'm' , 'M' , 'M' , FALSE}, // $32
{   ",",      ',' , ',' , '<' , FALSE}, // $33
{   "a.",     '.' , '.' , '>' , FALSE}, // $34
{   "a/",     '/' , '/' , '?' , FALSE}, // $35
{   "RShift",  0  ,  0  ,  0  , FALSE}, // $36
{   "n*",     '*' , '*' , '*' , FALSE}, // $37
{   "Alt",     0  ,  0  ,  0  , FALSE}, // $38
{   "Space",  ' ' , ' ' , ' ' , FALSE}, // $39
{   "CL",      0  ,  0  ,  0  , FALSE}, // $3A
{   "F1",      0  ,  0  ,  0  , FALSE}, // $3B
{   "F2",      0  ,  0  ,  0  , FALSE}, // $3C
{   "F3",      0  ,  0  ,  0  , FALSE}, // $3D
{   "F4",      0  ,  0  ,  0  , FALSE}, // $3E
{   "F5",      0  ,  0  ,  0  , TRUE }, // $3F
{   "F6",      0  ,  0  ,  0  , FALSE}, // $40
{   "F7",      0  ,  0  ,  0  , TRUE }, // $41
{   "F8",      0  ,  0  ,  0  , TRUE }, // $42
{   "F9",      0  ,  0  ,  0  , TRUE }, // $43
{   "F10",     0  ,  0  ,  0  , TRUE }, // $44
{   "Pause",   0  ,  0  ,  0  , TRUE }, // $45
{   "SL",      0  ,  0  ,  0  , FALSE}, // $46
{   "n7",     '7' , '7' , '7' , FALSE}, // $47
{   "n8",     '8' , '8' , '8' , FALSE}, // $48
{   "n9",     '9' , '9' , '9' , FALSE}, // $49
{   "n-",     '-' , '-' , '-' , FALSE}, // $4A
{   "n4",     '4' , '4' , '4' , FALSE}, // $4B
{   "n5",     '5' , '5' , '5' , FALSE}, // $4C
{   "n6",     '6' , '6' , '6' , FALSE}, // $4D
{   "n+",     '+' , '+' , '+' , FALSE}, // $4E
{   "n1",     '1' , '1' , '1' , FALSE}, // $4F
{   "n2",     '2' , '2' , '2' , FALSE}, // $50
{   "n3",     '3' , '3' , '3' , FALSE}, // $51
{   "n0",     '0' , '0' , '0' , FALSE}, // $52
{   "n.",     '.' , '.' , '.' , FALSE}, // $53
{   "(84)",    0  ,  0  ,  0  , FALSE}, // $54
{   "(85)",    0  ,  0  ,  0  , FALSE}, // $55
{   "(86)",    0  ,  0  ,  0  , FALSE}, // $56 Backslash
{   "F11",     0  ,  0  ,  0  , TRUE }, // $57
{   "F12",     0  ,  0  ,  0  , TRUE }, // $58
{   "(89)",    0  ,  0  ,  0  , FALSE}, // $59
{   "(90)",    0  ,  0  ,  0  , FALSE}, // $5A
{   "(91)",    0  ,  0  ,  0  , FALSE}, // $5B
{   "(92)",    0  ,  0  ,  0  , FALSE}, // $5C
{   "(93)",    0  ,  0  ,  0  , FALSE}, // $5D
{   "(94)",    0  ,  0  ,  0  , FALSE}, // $5E
{   "(95)",    0  ,  0  ,  0  , FALSE}, // $5F
{   "(96)",    0  ,  0  ,  0  , FALSE}, // $60
{   "(97)",    0  ,  0  ,  0  , FALSE}, // $61
{   "(98)",    0  ,  0  ,  0  , FALSE}, // $62
{   "(99)",    0  ,  0  ,  0  , FALSE}, // $63
{   "(100)",   0  ,  0  ,  0  , FALSE}, // $64
{   "(101)",   0  ,  0  ,  0  , FALSE}, // $65
{   "(102)",   0  ,  0  ,  0  , FALSE}, // $66
{   "(103)",   0  ,  0  ,  0  , FALSE}, // $67
{   "(104)",   0  ,  0  ,  0  , FALSE}, // $68
{   "(105)",   0  ,  0  ,  0  , FALSE}, // $69
{   "(106)",   0  ,  0  ,  0  , FALSE}, // $6A
{   "(107)",   0  ,  0  ,  0  , FALSE}, // $6B
{   "(108)",   0  ,  0  ,  0  , FALSE}, // $6C
{   "(109)",   0  ,  0  ,  0  , FALSE}, // $6D
{   "(110)",   0  ,  0  ,  0  , FALSE}, // $6E
{   "(111)",   0  ,  0  ,  0  , FALSE}, // $6F
{   "(112)",   0  ,  0  ,  0  , FALSE}, // $70
{   "(113)",   0  ,  0  ,  0  , FALSE}, // $71
{   "(114)",   0  ,  0  ,  0  , FALSE}, // $72
{   "(115)",   0  ,  0  ,  0  , FALSE}, // $73
{   "(116)",   0  ,  0  ,  0  , FALSE}, // $74
{   "(117)",   0  ,  0  ,  0  , FALSE}, // $75
{   "(118)",   0  ,  0  ,  0  , FALSE}, // $76
{   "(119)",   0  ,  0  ,  0  , FALSE}, // $77
{   "(120)",   0  ,  0  ,  0  , FALSE}, // $78
{   "(121)",   0  ,  0  ,  0  , FALSE}, // $79
{   "(122)",   0  ,  0  ,  0  , FALSE}, // $7A
{   "(123)",   0  ,  0  ,  0  , FALSE}, // $7B
{   "(124)",   0  ,  0  ,  0  , FALSE}, // $7C
{   "(125)",   0  ,  0  ,  0  , FALSE}, // $7D
{   "(126)",   0  ,  0  ,  0  , FALSE}, // $7E
{   "(127)",   0  ,  0  ,  0  , FALSE}, // $7F
{   "(128)",   0  ,  0  ,  0  , FALSE}, // $80
{   "(129)",   0  ,  0  ,  0  , FALSE}, // $81
{   "(130)",   0  ,  0  ,  0  , FALSE}, // $82
{   "(131)",   0  ,  0  ,  0  , FALSE}, // $83
{   "(132)",   0  ,  0  ,  0  , FALSE}, // $84
{   "(133)",   0  ,  0  ,  0  , FALSE}, // $85
{   "(134)",   0  ,  0  ,  0  , FALSE}, // $86
{   "(135)",   0  ,  0  ,  0  , FALSE}, // $87
{   "(136)",   0  ,  0  ,  0  , FALSE}, // $88
{   "(137)",   0  ,  0  ,  0  , FALSE}, // $89
{   "(138)",   0  ,  0  ,  0  , FALSE}, // $8A
{   "(139)",   0  ,  0  ,  0  , FALSE}, // $8B
{   "(140)",   0  ,  0  ,  0  , FALSE}, // $8C
{   "(141)",   0  ,  0  ,  0  , FALSE}, // $8D
{   "(142)",   0  ,  0  ,  0  , FALSE}, // $8E
{   "(143)",   0  ,  0  ,  0  , FALSE}, // $8F
{   "PrvTrk",  0  ,  0  ,  0  , TRUE }, // $90 Prev Track
{   "(145)",   0  ,  0  ,  0  , FALSE}, // $91
{   "(146)",   0  ,  0  ,  0  , FALSE}, // $92
{   "(147)",   0  ,  0  ,  0  , FALSE}, // $93
{   "(148)",   0  ,  0  ,  0  , FALSE}, // $94
{   "(149)",   0  ,  0  ,  0  , FALSE}, // $95
{   "(150)",   0  ,  0  ,  0  , FALSE}, // $96
{   "(151)",   0  ,  0  ,  0  , FALSE}, // $97
{   "(152)",   0  ,  0  ,  0  , FALSE}, // $98
{   "NxtTrk",  0  ,  0  ,  0  , TRUE }, // $99 Next Track
{   "(154)",   0  ,  0  ,  0  , FALSE}, // $9A
{   "(155)",   0  ,  0  ,  0  , FALSE}, // $9B
{   "(156)",   0  ,  0  ,  0  , FALSE}, // $9C Numeric ENTER
{   "(157)",   0  ,  0  ,  0  , FALSE}, // $9D Right Ctrl
{   "(158)",   0  ,  0  ,  0  , FALSE}, // $9E
{   "(159)",   0  ,  0  ,  0  , FALSE}, // $9F
{   "Mute",    0  ,  0  ,  0  , TRUE }, // $A0
{   "Calc",    0  ,  0  ,  0  , TRUE }, // $A1 Calculator
{   "Play",    0  ,  0  ,  0  , TRUE }, // $A2 Play/Pause
{   "(163)",   0  ,  0  ,  0  , FALSE}, // $A3
{   "MedStp",  0  ,  0  ,  0  , TRUE }, // $A4 Media Stop
{   "(165)",   0  ,  0  ,  0  , FALSE}, // $A5
{   "(166)",   0  ,  0  ,  0  , FALSE}, // $A6
{   "(167)",   0  ,  0  ,  0  , FALSE}, // $A7
{   "(168)",   0  ,  0  ,  0  , FALSE}, // $A8
{   "(169)",   0  ,  0  ,  0  , FALSE}, // $A9
{   "(170)",   0  ,  0  ,  0  , FALSE}, // $AA
{   "(171)",   0  ,  0  ,  0  , FALSE}, // $AB
{   "(172)",   0  ,  0  ,  0  , FALSE}, // $AC
{   "(173)",   0  ,  0  ,  0  , FALSE}, // $AD
{   "Vol-",    0  ,  0  ,  0  , TRUE }, // $AE Volume -
{   "(175)",   0  ,  0  ,  0  , FALSE}, // $AF
{   "Vol+",    0  ,  0  ,  0  , TRUE }, // $B0 Volume +
{   "(177)",   0  ,  0  ,  0  , FALSE}, // $B1
{   "Web",     0  ,  0  ,  0  , TRUE }, // $B2 Web/Home
{   "(179)",   0  ,  0  ,  0  , FALSE}, // $B3
{   "(180)",   0  ,  0  ,  0  , FALSE}, // $B4
{   "(181)",   0  ,  0  ,  0  , FALSE}, // $B5 Num /
{   "(182)",   0  ,  0  ,  0  , FALSE}, // $B6
{   "(183)",   0  ,  0  ,  0  , FALSE}, // $B7 SysRq
{   "(184)",   0  ,  0  ,  0  , FALSE}, // $B8 Right Alt
{   "(185)",   0  ,  0  ,  0  , FALSE}, // $B9
{   "(186)",   0  ,  0  ,  0  , FALSE}, // $BA
{   "(187)",   0  ,  0  ,  0  , FALSE}, // $BB
{   "(188)",   0  ,  0  ,  0  , FALSE}, // $BC
{   "(189)",   0  ,  0  ,  0  , FALSE}, // $BD
{   "(190)",   0  ,  0  ,  0  , FALSE}, // $BE
{   "(191)",   0  ,  0  ,  0  , FALSE}, // $BF
{   "(192)",   0  ,  0  ,  0  , FALSE}, // $C0
{   "(193)",   0  ,  0  ,  0  , FALSE}, // $C1
{   "(194)",   0  ,  0  ,  0  , FALSE}, // $C2
{   "(195)",   0  ,  0  ,  0  , FALSE}, // $C3
{   "(196)",   0  ,  0  ,  0  , FALSE}, // $C4
{   "(197)",   0  ,  0  ,  0  , FALSE}, // $C5 Pause
{   "(198)",   0  ,  0  ,  0  , FALSE}, // $C6
{   "(199)",   0  ,  0  ,  0  , FALSE}, // $C7 Home
{   "(200)",   0  ,  0  ,  0  , FALSE}, // $C8 Up
{   "(201)",   0  ,  0  ,  0  , FALSE}, // $C9 Page Up
{   "(202)",   0  ,  0  ,  0  , FALSE}, // $CA
{   "(203)",   0  ,  0  ,  0  , FALSE}, // $CB Left
{   "(204)",   0  ,  0  ,  0  , FALSE}, // $CC
{   "(205)",   0  ,  0  ,  0  , FALSE}, // $CD Right
{   "(206)",   0  ,  0  ,  0  , FALSE}, // $CE
{   "(207)",   0  ,  0  ,  0  , FALSE}, // $CF End
{   "(208)",   0  ,  0  ,  0  , FALSE}, // $D0 Down
{   "(209)",   0  ,  0  ,  0  , FALSE}, // $D1 Page Down
{   "(210)",   0  ,  0  ,  0  , FALSE}, // $D2 Insert
{   "(211)",   0  ,  0  ,  0  , FALSE}, // $D3 Delete
{   "(212)",   0  ,  0  ,  0  , FALSE}, // $D4
{   "(213)",   0  ,  0  ,  0  , FALSE}, // $D5
{   "(214)",   0  ,  0  ,  0  , FALSE}, // $D6
{   "(215)",   0  ,  0  ,  0  , FALSE}, // $D7
{   "(216)",   0  ,  0  ,  0  , FALSE}, // $D8
{   "(217)",   0  ,  0  ,  0  , FALSE}, // $D9
{   "(218)",   0  ,  0  ,  0  , FALSE}, // $DA
{   "(219)",   0  ,  0  ,  0  , FALSE}, // $DB Left Windows
{   "(220)",   0  ,  0  ,  0  , FALSE}, // $DC Right Windows
{   "App",     0  ,  0  ,  0  , FALSE}, // $DD Application
{   "(222)",   0  ,  0  ,  0  , FALSE}, // $DE Power
{   "(223)",   0  ,  0  ,  0  , FALSE}, // $DF Sleep
{   "(224)",   0  ,  0  ,  0  , FALSE}, // $E0
{   "(225)",   0  ,  0  ,  0  , FALSE}, // $E1
{   "(226)",   0  ,  0  ,  0  , FALSE}, // $E2
{   "(227)",   0  ,  0  ,  0  , FALSE}, // $E3 Wake
{   "(228)",   0  ,  0  ,  0  , FALSE}, // $E4
{   "Search",  0  ,  0  ,  0  , FALSE}, // $E5
{   "Faves",   0  ,  0  ,  0  , FALSE}, // $E6 Favo[u]rites
{   "Rfrsh",   0  ,  0  ,  0  , FALSE}, // $E7 Refresh
{   "WebStp",  0  ,  0  ,  0  , FALSE}, // $E8 Web Stop
{   "Frwds",   0  ,  0  ,  0  , TRUE }, // $E9 Forward[s]
{   "Back",    0  ,  0  ,  0  , FALSE}, // $EA
{   "Comp",    0  ,  0  ,  0  , FALSE}, // $EB My Computer
{   "Mail",    0  ,  0  ,  0  , FALSE}, // $EC
{   "Media",   0  ,  0  ,  0  , FALSE}, // $ED
{   "(238)",   0  ,  0  ,  0  , FALSE}, // $EE
{   "(239)",   0  ,  0  ,  0  , FALSE}, // $EF
{   "(240)",   0  ,  0  ,  0  , FALSE},
{   "(241)",   0  ,  0  ,  0  , FALSE},
{   "(242)",   0  ,  0  ,  0  , FALSE},
{   "(243)",   0  ,  0  ,  0  , FALSE},
{   "(244)",   0  ,  0  ,  0  , FALSE},
{   "(245)",   0  ,  0  ,  0  , FALSE},
{   "(246)",   0  ,  0  ,  0  , FALSE},
{   "(247)",   0  ,  0  ,  0  , FALSE},
{   "(248)",   0  ,  0  ,  0  , FALSE},
{   "(249)",   0  ,  0  ,  0  , FALSE},
{   "(250)",   0  ,  0  ,  0  , FALSE},
{   "(251)",   0  ,  0  ,  0  , FALSE},
{   "(252)",   0  ,  0  ,  0  , FALSE},
{   "(253)",   0  ,  0  ,  0  , FALSE},
{   "(254)",   0  ,  0  ,  0  , FALSE},
{   "(255)",   0  ,  0  ,  0  , FALSE},
{   "(256)",   0  ,  0  ,  0  , FALSE},
{   "(257)",   0  ,  0  ,  0  , FALSE},
{   "(258)",   0  ,  0  ,  0  , FALSE},
{   "(259)",   0  ,  0  ,  0  , FALSE},
{   "(260)",   0  ,  0  ,  0  , FALSE},
{   "(261)",   0  ,  0  ,  0  , FALSE},
{   "(262)",   0  ,  0  ,  0  , FALSE},
{   "(263)",   0  ,  0  ,  0  , FALSE},
{   "(264)",   0  ,  0  ,  0  , FALSE},
{   "(265)",   0  ,  0  ,  0  , FALSE},
{   "(266)",   0  ,  0  ,  0  , FALSE},
{   "(267)",   0  ,  0  ,  0  , FALSE},
{   "(268)",   0  ,  0  ,  0  , FALSE},
{   "(269)",   0  ,  0  ,  0  , FALSE},
{   "(270)",   0  ,  0  ,  0  , FALSE},
{   "(271)",   0  ,  0  ,  0  , FALSE},
{   "(272)",   0  ,  0  ,  0  , FALSE},
{   "(273)",   0  ,  0  ,  0  , FALSE},
{   "(274)",   0  ,  0  ,  0  , FALSE},
{   "(275)",   0  ,  0  ,  0  , FALSE},
{   "(276)",   0  ,  0  ,  0  , FALSE},
{   "(277)",   0  ,  0  ,  0  , FALSE},
{   "(278)",   0  ,  0  ,  0  , FALSE},
{   "(279)",   0  ,  0  ,  0  , FALSE},
{   "(280)",   0  ,  0  ,  0  , FALSE},
{   "(281)",   0  ,  0  ,  0  , FALSE},
{   "(282)",   0  ,  0  ,  0  , FALSE},
{   "(283)",   0  ,  0  ,  0  , FALSE},
{   "nE",     13  , 13  , 13  , FALSE},
{   "RCtrl",   0  ,  0  ,  0  , TRUE },
{   "(286)",   0  ,  0  ,  0  , FALSE},
{   "(287)",   0  ,  0  ,  0  , FALSE},
{   "(288)",   0  ,  0  ,  0  , FALSE},
{   "(289)",   0  ,  0  ,  0  , FALSE},
{   "(290)",   0  ,  0  ,  0  , FALSE},
{   "(291)",   0  ,  0  ,  0  , FALSE},
{   "(292)",   0  ,  0  ,  0  , FALSE},
{   "(293)",   0  ,  0  ,  0  , FALSE},
{   "(294)",   0  ,  0  ,  0  , FALSE},
{   "(295)",   0  ,  0  ,  0  , FALSE},
{   "(296)",   0  ,  0  ,  0  , FALSE},
{   "(297)",   0  ,  0  ,  0  , FALSE},
{   "(298)",   0  ,  0  ,  0  , FALSE},
{   "(299)",   0  ,  0  ,  0  , FALSE},
{   "(300)",   0  ,  0  ,  0  , FALSE},
{   "(301)",   0  ,  0  ,  0  , FALSE},
{   "(302)",   0  ,  0  ,  0  , FALSE},
{   "(303)",   0  ,  0  ,  0  , FALSE},
{   "(304)",   0  ,  0  ,  0  , FALSE},
{   "(305)",   0  ,  0  ,  0  , FALSE},
{   "(306)",   0  ,  0  ,  0  , FALSE},
{   "(307)",   0  ,  0  ,  0  , FALSE},
{   "(308)",   0  ,  0  ,  0  , FALSE},
{   "n/",     '/' , '/' , '/' , FALSE},
{   "(310)",   0  ,  0  ,  0  , FALSE},
{   "(311)",   0  ,  0  ,  0  , FALSE},
{   "(312)",   0  ,  0  ,  0  , FALSE},
{   "(313)",   0  ,  0  ,  0  , FALSE},
{   "(314)",   0  ,  0  ,  0  , FALSE},
{   "(315)",   0  ,  0  ,  0  , FALSE},
{   "(316)",   0  ,  0  ,  0  , FALSE},
{   "(317)",   0  ,  0  ,  0  , FALSE},
{   "(318)",   0  ,  0  ,  0  , FALSE},
{   "(319)",   0  ,  0  ,  0  , FALSE},
{   "(320)",   0  ,  0  ,  0  , FALSE},
{   "(321)",   0  ,  0  ,  0  , FALSE},
{   "(322)",   0  ,  0  ,  0  , FALSE},
{   "(323)",   0  ,  0  ,  0  , FALSE},
{   "(324)",   0  ,  0  ,  0  , FALSE},
{   "NL",      0  ,  0  ,  0  , FALSE},
{   "(326)",   0  ,  0  ,  0  , FALSE},
{   "Home",    0  ,  0  ,  0  , FALSE},
{   "Up",      0  ,  0  ,  0  , FALSE},
{   "PgUp",    0  ,  0  ,  0  , FALSE},
{   "(330)",   0  ,  0  ,  0  , FALSE},
{   "Lt",     '\b', '\b', '\b', FALSE},
{   "(332)",   0  ,  0  ,  0  , FALSE},
{   "Rt",     ' ' , ' ' , ' ' , FALSE},
{   "(334)",   0  ,  0  ,  0  , FALSE},
{   "End",     0  ,  0  ,  0  , FALSE},
{   "Dn",     10  , 10 ,  10  , FALSE},
{   "PgDn",    0  ,  0  ,  0  , FALSE},
{   "Ins",     0  ,  0  ,  0  , FALSE},
{   "Del",    127 , 127 , 127 , FALSE},
{   "(340)",   0  ,  0  ,  0  , FALSE},
{   "(341)",   0  ,  0  ,  0  , FALSE},
{   "(342)",   0  ,  0  ,  0  , FALSE},
{   "(343)",   0  ,  0  ,  0  , FALSE},
{   "(344)",   0  ,  0  ,  0  , FALSE},
{   "(345)",   0  ,  0  ,  0  , FALSE},
{   "(346)",   0  ,  0  ,  0  , FALSE},
{   "(347)",   0  ,  0  ,  0  , FALSE},
{   "(348)",   0  ,  0  ,  0  , FALSE},
{   "Menu",    0  ,  0  ,  0  , FALSE},
{   "(350)",   0  ,  0  ,  0  , FALSE},
{   "(351)",   0  ,  0  ,  0  , FALSE},
{   "(352)",   0  ,  0  ,  0  , FALSE},
{   "(353)",   0  ,  0  ,  0  , FALSE},
{   "(354)",   0  ,  0  ,  0  , FALSE},
{   "(355)",   0  ,  0  ,  0  , FALSE},
{   "(356)",   0  ,  0  ,  0  , FALSE},
{   "(357)",   0  ,  0  ,  0  , FALSE},
{   "(358)",   0  ,  0  ,  0  , FALSE},
{   "(359)",   0  ,  0  ,  0  , FALSE},
{   "(360)",   0  ,  0  ,  0  , FALSE},
{   "(361)",   0  ,  0  ,  0  , FALSE},
{   "(362)",   0  ,  0  ,  0  , FALSE},
{   "(363)",   0  ,  0  ,  0  , FALSE},
{   "(364)",   0  ,  0  ,  0  , FALSE},
{   "(365)",   0  ,  0  ,  0  , FALSE},
{   "(366)",   0  ,  0  ,  0  , FALSE},
{   "(367)",   0  ,  0  ,  0  , FALSE},
{   "(368)",   0  ,  0  ,  0  , FALSE},
{   "(369)",   0  ,  0  ,  0  , FALSE},
{   "(370)",   0  ,  0  ,  0  , FALSE},
{   "(371)",   0  ,  0  ,  0  , FALSE},
{   "(372)",   0  ,  0  ,  0  , FALSE},
{   "(373)",   0  ,  0  ,  0  , FALSE},
{   "(374)",   0  ,  0  ,  0  , FALSE},
{   "(375)",   0  ,  0  ,  0  , FALSE},
{   "(376)",   0  ,  0  ,  0  , FALSE},
{   "(377)",   0  ,  0  ,  0  , FALSE},
{   "(378)",   0  ,  0  ,  0  , FALSE},
{   "(379)",   0  ,  0  ,  0  , FALSE},
{   "(380)",   0  ,  0  ,  0  , FALSE},
{   "(381)",   0  ,  0  ,  0  , FALSE},
{   "(382)",   0  ,  0  ,  0  , FALSE},
{   "(383)",   0  ,  0  ,  0  , FALSE},
{   "(384)",   0  ,  0  ,  0  , FALSE},
{   "(385)",   0  ,  0  ,  0  , FALSE},
{   "(386)",   0  ,  0  ,  0  , FALSE},
{   "(387)",   0  ,  0  ,  0  , FALSE},
{   "(388)",   0  ,  0  ,  0  , FALSE},
{   "(389)",   0  ,  0  ,  0  , FALSE},
{   "(390)",   0  ,  0  ,  0  , FALSE},
{   "(391)",   0  ,  0  ,  0  , FALSE},
{   "(392)",   0  ,  0  ,  0  , FALSE},
{   "(393)",   0  ,  0  ,  0  , FALSE},
{   "(394)",   0  ,  0  ,  0  , FALSE},
{   "(395)",   0  ,  0  ,  0  , FALSE},
{   "(396)",   0  ,  0  ,  0  , FALSE},
{   "(397)",   0  ,  0  ,  0  , FALSE},
{   "(398)",   0  ,  0  ,  0  , FALSE},
{   "(399)",   0  ,  0  ,  0  , FALSE},
{   "(400)",   0  ,  0  ,  0  , FALSE},
{   "(401)",   0  ,  0  ,  0  , FALSE},
{   "(402)",   0  ,  0  ,  0  , FALSE},
{   "(403)",   0  ,  0  ,  0  , FALSE},
{   "(404)",   0  ,  0  ,  0  , FALSE},
{   "(405)",   0  ,  0  ,  0  , FALSE},
{   "(406)",   0  ,  0  ,  0  , FALSE},
{   "(407)",   0  ,  0  ,  0  , FALSE},
{   "(408)",   0  ,  0  ,  0  , FALSE},
{   "(409)",   0  ,  0  ,  0  , FALSE},
{   "(410)",   0  ,  0  ,  0  , FALSE},
{   "(411)",   0  ,  0  ,  0  , FALSE},
{   "(412)",   0  ,  0  ,  0  , FALSE},
{   "(413)",   0  ,  0  ,  0  , FALSE},
{   "(414)",   0  ,  0  ,  0  , FALSE},
{   "(415)",   0  ,  0  ,  0  , FALSE},
{   "(416)",   0  ,  0  ,  0  , FALSE},
{   "(417)",   0  ,  0  ,  0  , FALSE},
{   "(418)",   0  ,  0  ,  0  , FALSE},
{   "(419)",   0  ,  0  ,  0  , FALSE},
{   "(420)",   0  ,  0  ,  0  , FALSE},
{   "(421)",   0  ,  0  ,  0  , FALSE},
{   "(422)",   0  ,  0  ,  0  , FALSE},
{   "(423)",   0  ,  0  ,  0  , FALSE},
{   "(424)",   0  ,  0  ,  0  , FALSE},
{   "(425)",   0  ,  0  ,  0  , FALSE},
{   "(426)",   0  ,  0  ,  0  , FALSE},
{   "(427)",   0  ,  0  ,  0  , FALSE},
{   "(428)",   0  ,  0  ,  0  , FALSE},
{   "(429)",   0  ,  0  ,  0  , FALSE},
{   "(430)",   0  ,  0  ,  0  , FALSE},
{   "(431)",   0  ,  0  ,  0  , FALSE},
{   "(432)",   0  ,  0  ,  0  , FALSE},
{   "(433)",   0  ,  0  ,  0  , FALSE},
{   "(434)",   0  ,  0  ,  0  , FALSE},
{   "(435)",   0  ,  0  ,  0  , FALSE},
{   "(436)",   0  ,  0  ,  0  , FALSE},
{   "(437)",   0  ,  0  ,  0  , FALSE},
{   "(438)",   0  ,  0  ,  0  , FALSE},
{   "(439)",   0  ,  0  ,  0  , FALSE},
{   "(440)",   0  ,  0  ,  0  , FALSE},
{   "(441)",   0  ,  0  ,  0  , FALSE},
{   "(442)",   0  ,  0  ,  0  , FALSE},
{   "(443)",   0  ,  0  ,  0  , FALSE},
{   "(444)",   0  ,  0  ,  0  , FALSE},
{   "(445)",   0  ,  0  ,  0  , FALSE},
{   "(446)",   0  ,  0  ,  0  , FALSE},
{   "(447)",   0  ,  0  ,  0  , FALSE},
{   "(448)",   0  ,  0  ,  0  , FALSE},
{   "(449)",   0  ,  0  ,  0  , FALSE},
{   "(450)",   0  ,  0  ,  0  , FALSE},
{   "(451)",   0  ,  0  ,  0  , FALSE},
{   "(452)",   0  ,  0  ,  0  , FALSE},
{   "(453)",   0  ,  0  ,  0  , FALSE},
{   "(454)",   0  ,  0  ,  0  , FALSE},
{   "(455)",   0  ,  0  ,  0  , FALSE},
{   "(456)",   0  ,  0  ,  0  , FALSE},
{   "(457)",   0  ,  0  ,  0  , FALSE},
{   "(458)",   0  ,  0  ,  0  , FALSE},
{   "(459)",   0  ,  0  ,  0  , FALSE},
{   "(460)",   0  ,  0  ,  0  , FALSE},
{   "(461)",   0  ,  0  ,  0  , FALSE},
{   "(462)",   0  ,  0  ,  0  , FALSE},
{   "(463)",   0  ,  0  ,  0  , FALSE},
{   "(464)",   0  ,  0  ,  0  , FALSE},
{   "(465)",   0  ,  0  ,  0  , FALSE},
{   "(466)",   0  ,  0  ,  0  , FALSE},
{   "(467)",   0  ,  0  ,  0  , FALSE},
{   "(468)",   0  ,  0  ,  0  , FALSE},
{   "(469)",   0  ,  0  ,  0  , FALSE},
{   "(470)",   0  ,  0  ,  0  , FALSE},
{   "(471)",   0  ,  0  ,  0  , FALSE},
{   "(472)",   0  ,  0  ,  0  , FALSE},
{   "(473)",   0  ,  0  ,  0  , FALSE},
{   "(474)",   0  ,  0  ,  0  , FALSE},
{   "(475)",   0  ,  0  ,  0  , FALSE},
{   "(476)",   0  ,  0  ,  0  , FALSE},
{   "(477)",   0  ,  0  ,  0  , FALSE},
{   "(478)",   0  ,  0  ,  0  , FALSE},
{   "(479)",   0  ,  0  ,  0  , FALSE},
{   "(480)",   0  ,  0  ,  0  , FALSE},
{   "(481)",   0  ,  0  ,  0  , FALSE},
{   "(482)",   0  ,  0  ,  0  , FALSE},
{   "(483)",   0  ,  0  ,  0  , FALSE},
{   "(484)",   0  ,  0  ,  0  , FALSE},
{   "(485)",   0  ,  0  ,  0  , FALSE},
{   "(486)",   0  ,  0  ,  0  , FALSE},
{   "(487)",   0  ,  0  ,  0  , FALSE},
{   "(488)",   0  ,  0  ,  0  , FALSE},
{   "(489)",   0  ,  0  ,  0  , FALSE},
{   "(490)",   0  ,  0  ,  0  , FALSE},
{   "(491)",   0  ,  0  ,  0  , FALSE},
{   "(492)",   0  ,  0  ,  0  , FALSE},
{   "(493)",   0  ,  0  ,  0  , FALSE},
{   "(494)",   0  ,  0  ,  0  , FALSE},
{   "(495)",   0  ,  0  ,  0  , FALSE},
{   "(496)",   0  ,  0  ,  0  , FALSE},
{   "(497)",   0  ,  0  ,  0  , FALSE},
{   "(498)",   0  ,  0  ,  0  , FALSE},
{   "(499)",   0  ,  0  ,  0  , FALSE},
{   "(500)",   0  ,  0  ,  0  , FALSE},
{   "(501)",   0  ,  0  ,  0  , FALSE},
{   "(502)",   0  ,  0  ,  0  , FALSE},
{   "(503)",   0  ,  0  ,  0  , FALSE},
{   "(504)",   0  ,  0  ,  0  , FALSE},
{   "(505)",   0  ,  0  ,  0  , FALSE},
{   "(506)",   0  ,  0  ,  0  , FALSE},
{   "(507)",   0  ,  0  ,  0  , FALSE},
{   "(508)",   0  ,  0  ,  0  , FALSE},
{   "(509)",   0  ,  0  ,  0  , FALSE},
{   "(510)",   0  ,  0  ,  0  , FALSE},
{   "(511)",   0  ,  0  ,  0  , FALSE}
};

EXPORT struct LangStruct langs[LANGUAGES] = {
{ IDR_MAINMENU_ENG,
  IDR_POPUPMENU_ENG,
  IDR_TRAYMENU_ENG,
  338, 190, 133,
  NULL,
  { "Reset to game (F5, Alt+R)",
    "Open (Ctrl+O)",
    "Save snapshot (Ctrl+S)",
    "Save screenshot as IFF ILBM (Ctrl+_)",
    "Copy screenshot (Ctrl+C)",
    "Quickload snapshot (Ctrl+Shift+9, Alt+L)",
    "Quicksave snapshot (Ctrl+Shift+0, Alt+S)",
    "Flag line emulation? (Ctrl+P)",
    "Autofire? (Ctrl+B)",
    "Controllers swapped? (F7, Alt+J)",
    "Paused? (Pause, Play/Pause)",
    "Turbo speed? (F8, Alt+W, Forwards)",
    "Collisions? (Ctrl+L)",
    "Sound? (Ctrl+U, Mute)",
    "Play/stop (Ctrl+Plus (+)/Ctrl+T, Stop)",
    "Record/stop (Ctrl+R/Ctrl+T, Stop)",
    "START/START/START/SENS/P1 (F1)",
    "A/SELECT/UC/INT/P2 (F2)",
    "B/-/LC/MON/Coin A (F3)",
    "-/-/RESET/RST/Coin B (F4)"
} },
{ IDR_MAINMENU_HOL,
  IDR_POPUPMENU_HOL,
  IDR_TRAYMENU_HOL,
  448, 238, 181,
  ".\\Catalogs\\Nederlands.iff",
  { "Reset om spel (F5, Alt+R)",
    "Open (Ctrl+O)",
    "Opslaan momentopname (Ctrl+S)",
    "Opslaan screenshot als IFF ILBM (Ctrl+_)",
    "Copy screenshot (Ctrl+C)",
    "Snel laden momentopname (Ctrl+Shift+9, Alt+L)",
    "Snel opslaan momentopname (Ctrl+Shift+0, Alt+S)",
    "Flag lijn emulatie? (Ctrl+P)",
    "Autofire? (Ctrl+B)",
    "Controllers verwisseld? (F7, Alt+J)",
    "Onderbroken? (Pause, Play/Pause)",
    "Turbo snelheid? (F8, Alt+W, Forwards)",
    "Aanvaringen? (Ctrl+L)",
    "Geluid? (Ctrl+U, Mute)",
    "Afspelen/stop (Ctrl+Plus (+)/Ctrl+T, Stop)",
    "Opname/stop (Ctrl+R/Ctrl+T, Stop)",
    "START/START/START/SENS/P1 (F1)",
    "A/SELECT/UC/INT/P2 (F2)",
    "B/-/LC/MON/Munt A (F3)",
    "-/-/RESET/RST/Munt B (F4)"
} },
{ IDR_MAINMENU_FRA,
  IDR_POPUPMENU_FRA,
  IDR_TRAYMENU_FRA,
  509, 282, 184,
  ".\\Catalogs\\Franais.iff",
  { "Rinitialiser au jeu (F5, Alt+R)",
    "Ouvrir (Ctrl+O)",
    "Enregistrer instantann (Ctrl+S)",
    "Enregistrer une capture d'cran en IFF ILBM (Ctrl+_)",
    "Copier capture d'cran (Ctrl+C)",
    "Chargement d'instantann rapide (Ctrl+Shift+9, Alt+L)",
    "Enregistrement d'instantann rapide (Ctrl+Shift+0, Alt+S)",
    "Emuler Flag line? (Ctrl+P)",
    "Autofire? (Ctrl+B)",
    "Inverser les contrleurs? (F7, Alt+J)",
    "Pause? (Pause, Play/Pause)",
    "Vitesse turbo (F8, Alt+W, Forwards)",
    "Collisions? (Ctrl+L)",
    "Audio? (Ctrl+U, Mute)",
    "Play/stop (Ctrl+Plus (+)/Ctrl+T, Stop)",
    "Record/stop (Ctrl+R/Ctrl+T, Stop)",
    "START/START/START/SENS/P1 (F1)",
    "A/SELECT/UC/INT/P2 (F2)",
    "B/-/LC/MON/Coin A (F3)",
    "-/-/RESET/RST/Coin B (F4)"
} },
{ IDR_MAINMENU_GER,
  IDR_POPUPMENU_GER,
  IDR_TRAYMENU_GER,
  620, 334, 224,
  ".\\Catalogs\\Deutsch.iff",
  { "Rcksetzen auf Spiel (F5, Alt+R)",
    "ffnen (Ctrl+O)",
    "Speichern Momentaufnahme (Ctrl+S)",
    "Speichern als IFF ILBM-Bildschirmfoto (Ctrl+_)",
    "Kopieren Bildschirmfoto (Ctrl+C)",
    "Schnellladen Momentaufnahme (Ctrl+Shift+9, Alt+L)",
    "Schnellspeichern Momentaufnahme (Ctrl+Shift+0, Alt+S",
    "Flag-Emulation? (Ctrl+P)",
    "Selbstschuss? (Ctrl+B)",
    "Steuerung vertauschen? (F7, Alt+J)",
    "Pausiert? (Pause, Play/Pause)",
    "Hchstgeschwindigkeit? (F8, Alt+W, Forwards)",
    "Kollisionen? (Ctrl+L)",
    "Ton? (Ctrl+U, Mute)",
    "Play/stop (Ctrl+Plus (+)/Ctrl+T, Stoppa)",
    "Record/stop (Ctrl+R/Ctrl+T, Stoppa)",
    "START/START/START/SENS/P1 (F1)",
    "A/SELECT/UC/INT/P2 (F2)",
    "B/-/LC/MON/Mnze A (F3)",
    "-/-/RESET/RST/Mnze B (F4)"
} },
{ IDR_MAINMENU_ITA,
  IDR_POPUPMENU_ITA,
  IDR_TRAYMENU_ITA,
  543, 294, 210,
  ".\\Catalogs\\Italiano.iff",
  { "Ritorna al gioco (F5, Alt+R)",
    "Apri (Ctrl+O)",
    "Salva snapshot (Ctrl+S)",
    "Salva screenshot come IFF ILBM (Ctrl+_)",
    "Copia screenshot (Ctrl+C)",
    "Caricamento rapido snapshot (Ctrl+Shift+9, Alt+L)",
    "Salvataggio rapido snapshot (Ctrl+Shift+0, Alt+S)",
    "Emulazione linea flag? (Ctrl+P)",
    "Autofire? (Ctrl+B)",
    "Scambia controlli? (F7, Alt+J)",
    "Pausa? (Pausa, Play/Pausa)",
    "Velocit Turbo? (F8, Alt+W, Avanzamento)",
    "Collisioni? (Ctrl+L)",
    "Audio? (Ctrl+U, Mute)",
    "Play/stoppa (Ctrl+Plus (+)/Ctrl+T, Stoppa)",
    "Record/stoppa (Ctrl+R/Ctrl+T, Stoppa)",
    "AVVIO/AVVIO/AVVIO/SENS/P1 (F1)",
    "A/SELEZIONA/UC/INT/P2 (F2)",
    "B/-/LC/MON/Coin A (F3)",
    "-/-/RESET/RST/Coin B (F4)"
} },
{ IDR_MAINMENU_RUS,
  IDR_POPUPMENU_RUS,
  IDR_TRAYMENU_RUS,
  629, 343, 228,
  ".\\Catalogs\\Russian.iff",
  { "  (F5, Alt+R)",
    " (Ctrl+O)",
    "  (Ctrl+S)",
    "   IFF ILBM (Ctrl+_)",
    "  (Ctrl+C)",
    "   (Ctrl+Shift+9, Alt+L)",
    "   (Ctrl+Shift+0, Alt+S)",
    "  ? (Ctrl+P)",
    "? (Ctrl+B)",
    " ? (F7, Alt+J)",
    "? (Pause, Play/Pause)",
    "? (F8, Alt+W, Forwards)",
    "? (Ctrl+L)",
    "? (Ctrl+U, Mute)",
    "/ (Ctrl+Plus (+)/Ctrl+T, Stop)",
    "/ (Ctrl+R/Ctrl+T, Stop)",
    "START/START/START/SENS/P1 (F1)",
    "A/SELECT/UC/INT/P2 (F2)",
    "B/-/LC/MON/Coin A (F3)",
    "-/-/RESET/RST/Coin B (F4)"
// All these (at least) are used for Russian:
//          
} },
{ IDR_MAINMENU_SPA,
  IDR_POPUPMENU_SPA,
  IDR_TRAYMENU_SPA,
  503, 259, 205,
  ".\\Catalogs\\Espaol.iff",
  { "Reiniciar el juego (F5, Alt+R)",
    "Abrir (Ctrl+O)",
    "Guardar el estado actual (Ctrl+S)",
    "Guardar la imagen de la pantalla en formato IFF ILBM (Ctrl+_)",
    "Copiar la imagen de la pantalla (Ctrl+C)",
    "Carga rpida del estado (Ctrl+Shift+9, Alt+L)",
    "Grabacin rpida de estado actual (Ctrl+Shift+0, Alt+S)",
    "Emulacin de la lnea de estado FLAG? (Ctrl+P)",
    "Activar el disparo automtico? (Ctrl+B)",
    "Intercambiar los controladores? (F7, Alt+J)",
    "En modo pausa? (Pausa, Reproducir/pausa)",
    "Velocidad turbo? (F8, Alt+W, Forwards)",
    "Activar las colisiones? (Ctrl+L)",
    "Activar el sonido? (Ctrl+U, Mute)",
    "Reproduccin/detener (Ctrl+Plus (+)/Ctrl+T, Detener)",
    "Grabacin/detener (Ctrl+R/Ctrl+T, Detener)",
    "START/START/START/SENS/P1 (F1)",
    "A/SELECT/UC/INT/P2 (F2)",
    "B/-/LC/MON/Ranura A (F3)",
    "-/-/RESET/RST/Ranura B (F4)"
} } };

MODULE const UBYTE translation[2][96] = {
{ // Code page Amiga-1252 (Western European) -> DOS-437 (Western European)
  255, // 160  nonbreaking space
  173, // 161 
  162, // 162  cents
  156, // 163  pounds
  207, // 164 
  190, // 165  yen
  221, // 166 
  245, // 167  page break
  249, // 168 
  184, // 169  copyright
  166, // 170 
  174, // 171  much less than
  170, // 172 
  240, // 173 
  169, // 174  registered trademark
  238, // 175 
  248, // 176  degrees
  241, // 177  plus/minus
  253, // 178  squared
  252, // 179  cubed
  239, // 180 
  230, // 181 
  244, // 182  paragraph break
  250, // 183 
  247, // 184 
  251, // 185 
  167, // 186 
  175, // 187  much greater than
  172, // 188  one quarter
  171, // 189  one half
  243, // 190  three quarters
  168, // 191 
  183, // 192 
  181, // 193 
  182, // 194 
  199, // 195 
  142, // 196 
  143, // 197 
  146, // 198 
  128, // 199 
  212, // 200 
  144, // 201 
  210, // 202 
  211, // 203 
  222, // 204 
  214, // 205 
  215, // 206 
  216, // 207 
  209, // 208 
  165, // 209 
  227, // 210 
  224, // 211 
  226, // 212 
  229, // 213 
  153, // 214 
  158, // 215 
  157, // 216 
  235, // 217 
  233, // 218 
  234, // 219 
  154, // 220 
  237, // 221 
  232, // 222 
  225, // 223 
  133, // 224 
  160, // 225 
  131, // 226 
  198, // 227 
  132, // 228 
  134, // 229 
  145, // 230 
  135, // 231 
  138, // 232 
  130, // 233 
  136, // 234 
  137, // 235 
  141, // 236 
  161, // 237 
  140, // 238 
  139, // 239 
  208, // 240 
  164, // 241 
  149, // 242 
  162, // 243 
  147, // 244 
  228, // 245 
  148, // 246 
  246, // 247  divisor
  155, // 248 
  151, // 249 
  163, // 250 
  150, // 251 
  129, // 252 
  236, // 253 
  231, // 254 
  152  // 255 
},
{ // Code page Amiga-1251 (Russian) -> DOS-866 (Russian)
  255, // 160  nonbreaking space
  '?', // 161 
  '?', // 162  cents
  '?', // 163  pounds
  253, // 164 
  '?', // 165  yen
  '?', // 166 
  '?', // 167  page break
  '?', // 168 
  '?', // 169  copyright
  '?', // 170 
  '?', // 171  much less than
  '?', // 172 
  '?', // 173 
  '?', // 174  registered trademark
  '?', // 175 
  248, // 176  degrees
  '?', // 177  plus/minus
  '?', // 178  squared
  '?', // 179  cubed
  '?', // 180 
  '?', // 181 
  '?', // 182  paragraph break
  250, // 183 
  '?', // 184 
  '?', // 185 
  '?', // 186 
  '?', // 187  much greater than
  '?', // 188  one quarter
  '?', // 189  one half
  '?', // 190  three quarters
  '?', // 191 
  128, // 192
  129, // 193
  130, // 194
  131, // 195
  132, // 196
  133, // 197
  134, // 198
  135, // 199
  136, // 200
  137, // 201
  138, // 202
  139, // 203
  140, // 204
  141, // 205
  142, // 206
  143, // 207
  144, // 208
  145, // 209
  146, // 210
  147, // 211
  148, // 212
  149, // 213
  150, // 214
  151, // 215
  152, // 216
  153, // 217
  154, // 218
  155, // 219
  156, // 220
  157, // 221
  158, // 222
  159, // 223
  160, // 224
  161, // 225
  162, // 226
  163, // 227
  164, // 228
  165, // 229
  166, // 230
  167, // 231
  168, // 232
  169, // 233
  170, // 234
  171, // 235
  172, // 236
  173, // 237
  174, // 238
  175, // 239
  224, // 240
  225, // 241
  226, // 242
  227, // 243
  228, // 244
  229, // 245
  230, // 246
  231, // 247
  232, // 248
  233, // 249
  234, // 250
  235, // 251
  236, // 252
  237, // 253
  238, // 254
  239  // 255
} };

// MODULE FUNCTIONS-------------------------------------------------------

BOOL CALLBACK         SetDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK       RastNDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK        EditDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK     PaletteDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);

MODULE void maketoolbar(void);
MODULE void printusage(STRPTR progname);
MODULE void upgradeconfigs(void);
MODULE void subreq(HWND hwnd);

// CODE-------------------------------------------------------------------

LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{   PAINTSTRUCT   ps;
    LPTOOLTIPTEXT lpttt;
    int           i;
    signed int    oldcurrentgame,
                  movement;
    LPNMLISTVIEW  lpnmlv;

    /* "When a window procedure processes a message, it should return 0 from the window
       procedure. All messages that a window procedure chooses not to process must be
       passed to a Windows function named DefWindowProc(). The value returned from
       DefWindowProc() must be returned from the window procedure. */

    switch (msg)
    {
    case WM_SYSCOMMAND:
        if (wParam == SC_MAXIMIZE)
        {   if (fullscreen) fullscreen = FALSE; else fullscreen = TRUE;
            docommand(MENUITEM_FULLSCREEN);
        } elif (LOWORD(wParam) == ID_FILE_ICONIFY)
        {   iconify();
        } else
        {   return DefWindowProc(hwnd, msg, wParam, lParam);
        }
    acase WM_NCLBUTTONDBLCLK:
        if (fullscreen) fullscreen = FALSE; else fullscreen = TRUE;
        docommand(MENUITEM_FULLSCREEN);
    acase WM_DROPFILES:
        DISCARD DragQueryFile((HDROP) wParam, 0, fn_game, MAX_PATH);
        DragFinish((HDROP) wParam);
        // runtime_// assert(strlen(fn_game) <= MAX_PATH);
        DISCARD engine_load(FALSE);
    acase WM_PAINT:
        BeginPaint(hwnd, &ps);
        if (fullscreen && !stretch)
        {   clearscreen(); // so that the grey border is repainted
        }
        redrawscreen();
        DISCARD EndPaint(hwnd, &ps);
    acase USR_NOTIFYICON:
        // assert(iconified);
        switch (lParam)
        {
        case WM_LBUTTONDBLCLK:
        uniconify(FALSE);
        acase WM_RBUTTONUP:
        traymenu();
        }
    acase WM_CLOSE:
        if (confirm())
        {   cleanexit(EXIT_SUCCESS);
        }
    acase WM_DESTROY:
        PostQuitMessage(0);
    acase WM_MOUSEMOVE:
        updatepointer(FALSE);
    acase WM_MOVE:
        if (fullscreen)
        {   DISCARD SetWindowPos
        (   MainWindowPtr,
            HWND_TOP,
            0, 0,
            winwidth,
            winheight,
            SWP_NOSIZE | SWP_NOCOPYBITS
        ); // move it back!
        } elif (ready)
        {   winleftx = (int) LOWORD(lParam);
        wintopy  = (int) HIWORD(lParam);
        }
    acase WM_LBUTTONDBLCLK:
    case WM_LBUTTONDOWN:
        if (hostcontroller[0] == CONTROLLER_TRACKBALL)
        {   // assuming it is within in guest screen boundaries
        lmb = TRUE;
        }
    acase WM_LBUTTONUP:
        lmb = FALSE;
    acase WM_MBUTTONDBLCLK:
    case WM_MBUTTONDOWN:
        mmb = TRUE;
    acase WM_MBUTTONUP:
        mmb = FALSE;
    acase WM_MOUSEWHEEL:
        if
        (    showsidebar
          && foundgames
          && !crippled
          && ctrl()
        )
        {   movement = ((short) HIWORD(wParam)) / WHEELDELTA;
            oldcurrentgame = currentgame;
            if (movement >= 1) // forwards/up/away
            {   if (currentgame == -1)
                {   currentgame = foundgames; // move to bottom
                } else
                {   for (i = 0; i < movement; i++)
                    {   if (currentgame > 0)
                        {   currentgame--; // move up by 1
            }   }   }   }
            elif (movement <= -1) // back/down/towards
            {   if (currentgame == -1)
                {   currentgame = 0; // move to top
                } else
                {   for (i = 0; i < abs(movement); i++)
                    {   if (currentgame < foundgames - 1)
                        {   currentgame++; // move down by 1
            }   }   }   }
            if (currentgame != oldcurrentgame)
            {   ListView_SetItemState(hSideBar, oldcurrentgame, 0, 0x000F);
                ListView_SetItemState(hSideBar, currentgame, (LVIS_FOCUSED | LVIS_SELECTED), 0x000F);
                if (currentgame != foundgames - 1)
                {   DISCARD ListView_EnsureVisible(hSideBar, currentgame, TRUE); // doesn't seem to work properly for items in the rightmost column
                }
                macro_stop();
                dosidebar(); // but window might be about to be resized!
        }   }
    acase WM_INITMENU:
        // sound_off(FALSE); we would need to turn sound back on too at the appropriate point
        ff_off(0);
        ff_off(1);
    acase WM_ACTIVATE:
        if ((wParam == WA_ACTIVE || wParam == WA_CLICKACTIVE) && hSideBar && currentgame != -1)
        {   SetFocus(hSideBar);
        }
        do_autopause(wParam, lParam);
        return 0;
    case WM_SIZE: // no need for acase
        if (LOWORD(lParam) != winwidth && showsidebar && !fullscreen && ready)
        {   if (LOWORD(lParam) > minwinwidth)
        {   winwidth = LOWORD(lParam);
        } else
        {   winwidth = minwinwidth;
        }
        sidebarwidth = winwidth - rastwidth;
        
        // The preferred way to do it is:
        resize(size);

        /* The alternative way is:
        closewindow();
        openwindow(); */
        }
    acase WM_COMMAND:
        if (!ListeningWindowPtr)
        {   for (i = 0; i < MENUITEMS; i++)
        {   if (menucode[i] == LOWORD(wParam))
            {   handle_menu(i);
                return 0; // for speed
        }   }
        for (i = 0; i < MENUOPTS; i++)
        {   if (menucode2[i] == LOWORD(wParam))
            {   handle_menu2(i);
                return 0; // for speed
        }   }
        switch (LOWORD(wParam))
        {
        case ID_MACRO_TOGGLERECORDING:
            if (recmode == RECMODE_NORMAL)
            {   project_save(KIND_COR);
            } elif (recmode == RECMODE_RECORD)
            {   macro_stop();
            } // else assert(recmode == RECMODE_PLAY);
        acase ID_MACRO_TOGGLEPLAYING:
            macro_stop();
        acase ID_CONSOLE_START:
            if (machine == ELEKTOR)
            {   console_start = 24; // needed for eg. Penalty
            } else
            {   console_start = 5; // needed for eg. Super Bug
            }
        acase ID_CONSOLE_A:
            if (machine == ELEKTOR)
            {   console_a     = 24; // needed for eg. Penalty (Elektor)
            } else
            {   console_a     = 5; // 24 is too much for eg. Hunting (Interton)
            }
        acase ID_CONSOLE_B:
            console_b     = 5;
        acase ID_CONSOLE_RESET:
            console_reset = 5;
        }   }
    acase WM_NOTIFY:
        switch (((LPNMHDR) lParam)->code)
        {
        case TTN_NEEDTEXT:
            lpttt = (LPTOOLTIPTEXT) lParam; // button info
            lpttt->hinst = InstancePtr;

            switch (lpttt->hdr.idFrom) // Check which button we hover on
            {
            case ID_FILE_RESET:
                lpttt->lpszText = langs[language].tooltip[0];
            acase ID_FILE_OPEN:
                lpttt->lpszText = langs[language].tooltip[1];
            acase ID_FILE_SAVESNP:
                lpttt->lpszText = langs[language].tooltip[2];
            acase ID_FILE_SAVEILBM:
                lpttt->lpszText = langs[language].tooltip[3];
            acase ID_EDIT_COPY:
                lpttt->lpszText = langs[language].tooltip[4];
            acase ID_FILE_QUICKLOAD:
                lpttt->lpszText = langs[language].tooltip[5];
            acase ID_FILE_QUICKSAVE:
                lpttt->lpszText = langs[language].tooltip[6];
            acase ID_GRAPHICS_FLAGLINE:
                lpttt->lpszText = langs[language].tooltip[7];
            acase ID_CONTROLLERS_AUTOFIRE:
                lpttt->lpszText = langs[language].tooltip[8];
            acase ID_CONTROLLERS_SWAPPED:
                lpttt->lpszText = langs[language].tooltip[9];
            acase ID_SPEED_PAUSED:
                lpttt->lpszText = langs[language].tooltip[10];
            acase ID_SPEED_TURBO:
                lpttt->lpszText = langs[language].tooltip[11];
            acase ID_SPRITES_COLLISIONS:
                lpttt->lpszText = langs[language].tooltip[12];
            acase ID_SOUND_ENABLED:
                lpttt->lpszText = langs[language].tooltip[13];
            acase ID_MACRO_TOGGLEPLAYING:
                lpttt->lpszText = langs[language].tooltip[14];
            acase ID_MACRO_TOGGLERECORDING:
                lpttt->lpszText = langs[language].tooltip[15];
            acase ID_CONSOLE_START:
                switch (machine)
                {
                case ARCADIA:
                case INTERTON:
                case ELEKTOR:
                    lpttt->lpszText = "START (F1)";
                acase INSTRUCTOR:
                    lpttt->lpszText = "SENS (F1)";
                acase ASTROWARS:
                case GALAXIA:
                case MALZAK:
                    lpttt->lpszText = "P1 (F1)";
                adefault: // eg. PIPBUG, CD2650, CHAOS, DOLPHIN
                    lpttt->lpszText = langs[language].tooltip[16];
                }
            acase ID_CONSOLE_A:
                switch (machine)
                {
                case ARCADIA:
                    lpttt->lpszText = "A (F2)";
                acase INTERTON:
                    lpttt->lpszText = "SELECT (F2)";
                acase ELEKTOR:
                    lpttt->lpszText = "UC (F2)";
                acase INSTRUCTOR:
                    lpttt->lpszText = "INT (F2)";
                acase ASTROWARS:
                case GALAXIA:
                case MALZAK:
                    lpttt->lpszText = "P2 (F2)";
                adefault: // eg. PIPBUG, CD2650, CHAOS, DOLPHIN
                    lpttt->lpszText = langs[language].tooltip[17];
                }
            acase ID_CONSOLE_B:
                switch (machine)
                {
                case ARCADIA:
                    lpttt->lpszText = "B (F3)";
                acase ELEKTOR:
                    lpttt->lpszText = "LC (F3)";
                acase INSTRUCTOR:
                    lpttt->lpszText = "MON (F3)";
                acase ASTROWARS:
                case GALAXIA:
                case MALZAK:
                    charset = CHARSET_WIN;
                    strcpy(coinstring[0], GetCatalogStr(CatalogPtr, MSG_COIN_A, "Coin A"));
                    strcat(coinstring[0], " (F3)");
                    lpttt->lpszText = coinstring[0];
                    charset = CHARSET_DOS;
                adefault: // eg. INTERTON, PIPBUG, CD2650, CHAOS, DOLPHIN
                lpttt->lpszText = langs[language].tooltip[18];
            }
        acase ID_CONSOLE_RESET:
            switch (machine)
            {
            case ELEKTOR:
                lpttt->lpszText = "RESET (F4)";
            acase INSTRUCTOR:
                lpttt->lpszText = "RST (F4)";
            acase ASTROWARS:
            case GALAXIA:
                charset = CHARSET_WIN;
                strcpy(coinstring[1], GetCatalogStr(CatalogPtr, MSG_COIN_B, "Coin B"));
                strcat(coinstring[1], " (F4)");
                lpttt->lpszText = coinstring[1];
                charset = CHARSET_DOS;
            acase MALZAK:
                if (memmap == MEMMAP_MALZAK1)
                {   lpttt->lpszText = langs[language].tooltip[19];
                } else
                {   // assert(memmap == MEMMAP_MALZAK2);
                    charset = CHARSET_WIN;
                    strcpy(coinstring[1], GetCatalogStr(CatalogPtr, MSG_CHANGE_DIP, "Change DIP switch"));
                    strcat(coinstring[1], " (F4)");
                    lpttt->lpszText = coinstring[1];
                    charset = CHARSET_DOS;
                }
            adefault: // eg. ARCADIA, INTERTON, PIPBUG, CD2650, CHAOS, DOLPHIN
                lpttt->lpszText = langs[language].tooltip[19];
        }   }
        acase LVN_BEGINLABELEDIT:
            return 1; // this disallows label editing
        acase NM_CLICK:
        case NM_DBLCLK:
        case NM_RCLICK:
            lpnmlv = (LPNMLISTVIEW) lParam;
            if (lpnmlv->hdr.hwndFrom == hSideBar)
            {   if (lpnmlv->iItem == -1)
                {   if (currentgame != -1) 
                    {   ListView_SetItemState(hSideBar, currentgame, (LVIS_FOCUSED | LVIS_SELECTED), 0x000F);
                }   }
                else
                {   // assert(!crippled);
                    macro_stop();
                    if (currentgame != -1 && lpnmlv->iItem != currentgame) // to handle the user holding down Ctrl and clicking
                    {   ListView_SetItemState(hSideBar, currentgame, 0, 0x000F);
                    }
                    currentgame = lpnmlv->iItem;
                    dosidebar();
        }   }   }
    adefault:
        return DefWindowProc(hwnd, msg, wParam, lParam);
    }

    return 0;
}

#ifdef GUI
    int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
#endif
#ifdef CLI
    int main(void)
#endif
{   FLAG      settings    = FALSE;
    WNDCLASSEX    wc;
    int           argc,
                  gotfilename = FALSE,
                  i, j,
                  oldlanguage,
                  x, y;
    char**        argv;
    WSADATA       WSAInfo;
    JOYINFOEX     joyInfoEx;
    OSVERSIONINFO Vers;
    LANGID        langid;
    SYSTEMTIME    TheTime;

    // Start of program.

    InitCommonControls();

    argc = __argc;
    argv = __argv;

    for (x = 0; x < BOXWIDTH; x++)
    {   for (y = 0; y < BOXHEIGHT; y++)
        {   screen[x][y]   = 8;
    }   }
    for (i = 0; i < SAMPLES; i++)
    {   SampleLength[i]    = 0;
        SampleOpened[i]    =
        SamplePrepared[i]  = FALSE;
        SampleBuffer[i]    = NULL;
        hSample[i]         = NULL;
    }
    for (i = 0; i < MENUBITMAPS; i++)
    {   mib[i].bitmap      = NULL;
    }
    for (i = 0; i < LANGUAGES; i++)
    {   langs[i].dosbuffer =
        langs[i].winbuffer = NULL;
        langs[i].offset    = NULL;
    }
    joyname[0][0] =
    joyname[1][0] = 0;

    for (i = 0; i < 2; i++)
    {   for (j = 0; j < GUESTCOLOURS; j++)
        {   resetcolour(i, j, FALSE);
    }   }

    engine_setup();

    InstancePtr = GetModuleHandle(NULL);

    DisplayWidth  = GetSystemMetrics(SM_CXSCREEN); // these 2 lines must be done before loadconfig()
    DisplayHeight = GetSystemMetrics(SM_CYSCREEN);

    stars = malloc(MAXBOXWIDTH * MAXBOXHEIGHT * sizeof(ULONG));

    // major    minor
    //   3        51     Windows NT version 3.51
    //   4         0     Windows NT version 4.0
    //   4        10     Windows 98SE

    Vers.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    DISCARD GetVersionEx((LPOSVERSIONINFO) &Vers);
    if (Vers.dwPlatformId == VER_PLATFORM_WIN32_NT)
    {   // T1; zprintf("Lame OS detected!\n"); T2;
        lame = TRUE;
    } elif (Vers.dwPlatformId == VER_PLATFORM_WIN32s)
    {   ; // T1; zprintf("Please upgrade to Windows 9x!\n"); T2;
    } elif (Vers.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)
    {   ; // T1; zprintf("Windows 9x detected.\n"); T2;
    } else
    {   // T1; zprintf("Unknown OS!\n"); T2;
        lame = TRUE;
    }
    // T1;
    // zprintf("%s\n", Vers.szCSDVersion);
    // zprintf("%d, %d\n", Vers.dwMajorVersion, Vers.dwMinorVersion);
    // T2;

    speech_init();

    hArrow = LoadCursor(NULL, IDC_ARROW);
    hCross = LoadCursor(NULL, IDC_CROSS);

    DISCARD GetCurrentDirectory(MAX_PATH, ProgDir);
    filepart[0] =
    hostname[0] = 0;
    // no need to zero fn_screenshot[0]

    if (GetFileAttributes("SpriteMakerEA6.exe") != 0xFFFFFFFF)
    {   sprmakerexists = TRUE;
    }
    if (GetFileAttributes("TextMaker1.exe") != 0xFFFFFFFF)
    {   txtmakerexists = TRUE;
    }

    strcpy(fn_game, ProgDir);
    if (fn_game[0] != EOS)
    {   strcat(fn_game, "\\");
    }
    strcat(fn_game, "Games\\");
    breakparts();
    if (GetFileAttributes(fn_game) != 0xFFFFFFFF)
    {   ;
#ifdef VERBOSE
        T1; zprintf("Found %s.\n", fn_game); T2;
#endif
    } else
    {
#ifdef VERBOSE
        T1; zprintf("Didn't find %s.\n", fn_game); T2;
#endif
        strcpy(fn_game, ProgDir);
        if (fn_game[0] != EOS)
        {   strcat(fn_game, "\\");
        }
        breakparts();
#ifdef VERBOSE
        T1; zprintf("So using %s.\n", fn_game); T2;
#endif
    }

    strcpy(fn_config, ProgDir);
    if (fn_config[0] != EOS)
    {   strcat(fn_config, "\\");
    }
    strcat(fn_config, "Configs\\WA.CFG");
#ifdef VERBOSE
    T1; zprintf("Configuration file is %s!\n", fn_config); T2;
#endif

    // Initialize networking
    DISCARD WSAStartup(MAKEWORD(2, 2), &WSAInfo);

    langid = GetUserDefaultLangID() & 0xFF;

#ifdef VERBOSE
    T1; zprintf("User language ID is $%2X.\n", langid); T2;
#endif

    switch (langid)
    {
    case 0x07:
        language = LANGUAGE_GER;
    acase 0x0A:
        language = LANGUAGE_SPA;
    acase 0x0C:
        language = LANGUAGE_FRA;
    acase 0x10:
        language = LANGUAGE_ITA;
    acase 0x13:
        language = LANGUAGE_HOL;
    acase 0x19:
        language = LANGUAGE_RUS;
    adefault:
        language = LANGUAGE_ENG;
    }

#ifdef VERBOSE
    T1; zprintf("Emulator language is %d.\n", language); T2;
#endif

    if (DisplayWidth >= 1280 && DisplayHeight >= 1024) // 1280*1024
    {   realsize = 3;
    } elif (DisplayWidth >= 1024 && DisplayHeight >= 768) // 1024*768
    {   realsize = 2;
    } else // 800*600, 640*480
    {   realsize = 1;
    }
    size = realsize;

    loadconfig(); // load default configuration file

    strcpy(ProgName, argv[0]);
    if (argc >= 2)
    {   for (i = 1; i < argc; i++)
        {   if (!strcmp(argv[i], "?"))
            {   printusage(argv[0]);
            } elif (!strnicmp(argv[i], "AUTOSAVE", 8))
            {   if (!stricmp(&argv[i][8], "=ON"))
                {   autosave = TRUE;
                } elif (!stricmp(&argv[i][8], "=OFF"))
                {   autosave = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "COLLISIONS", 10))
            {   if (!stricmp(&argv[i][10], "=ON"))
                {   collisions = TRUE;
                } elif (!stricmp(&argv[i][10], "=OFF"))
                {   collisions = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "FRAMESKIP", 9))
            {   if (!strcmp(&argv[i][9], "=1"))
                {   frameskip = 1;
                } elif (!strcmp(&argv[i][9], "=2"))
                {   frameskip = 2;
                } elif (!strcmp(&argv[i][9], "=3"))
                {   frameskip = 3;
                } elif (!strcmp(&argv[i][9], "=4"))
                {   frameskip = 4;
                } elif (!strcmp(&argv[i][9], "=5"))
                {   frameskip = 5;
                } elif (!strcmp(&argv[i][9], "=6"))
                {   frameskip = 6;
                } elif (!strcmp(&argv[i][9], "=7"))
                {   frameskip = 7;
                } elif (!strcmp(&argv[i][9], "=8"))
                {   frameskip = 8;
                } elif (!strcmp(&argv[i][9], "=9"))
                {   frameskip = 9;
                } elif (!strcmp(&argv[i][9], "=10"))
                {   frameskip = 10;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "FULLSCREEN", 10))
            {   if (!stricmp(&argv[i][10], "=ON"))
                {   fullscreen = TRUE;
                } elif (!stricmp(&argv[i][10], "=OFF"))
                {   fullscreen = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "MACHINE", 7))
            {   machine = MACHINES;
                if (argv[i][7] == '=')
                {   for (j = 0; j < MACHINES; j++)
                    {   if (!stricmp(&argv[i][8], machines[j].cli))
                        {   machine = j;
                            break;
                }   }   }

                if (machine == MACHINES)
                {   machine = ARCADIA;
                    printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "MENUBAR", 7))
            {   if (!stricmp(&argv[i][7], "=ON"))
                {   showmenubar = TRUE;
                } elif (!stricmp(&argv[i][7], "=OFF"))
                {   showmenubar = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "PAUSED", 6))
            {   if (!stricmp(&argv[i][6], "=ON"))
                {   paused = TRUE;
                } elif (!stricmp(&argv[i][6], "=OFF"))
                {   paused = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "SETTINGS", 8))
            {   if (argv[i][8] != '=')
                {   printusage(argv[0]);
                }
                strcpy(fn_config, ".\\Configs\\");
                strcat(fn_config, (STRPTR) &argv[i][9]);

                // what if the user wants to give a full pathname? We don't support that.
                settings = TRUE;
                // we don't load it yet
            } elif (!strnicmp(argv[i], "SIZE", 4))
            {   if (!strcmp(&argv[i][4], "=1"))
                {   size = 1;
                } elif (!strcmp(&argv[i][4], "=2"))
                {   size = 2;
                } elif (!strcmp(&argv[i][4], "=3"))
                {   size = 3;
                } elif (!strcmp(&argv[i][4], "=4"))
                {   size = 4;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "SOUND", 5))
            {   if (!stricmp(&argv[i][5], "=ON"))
                {   sound = TRUE;
                } elif (!stricmp(&argv[i][5], "=OFF"))
                {   sound = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "STRETCH", 7))
            {   if (!stricmp(&argv[i][7], "=ON"))
                {   stretch = TRUE;
                } elif (!stricmp(&argv[i][7], "=OFF"))
                {   stretch = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "TITLEBAR", 8))
            {   if (!stricmp(&argv[i][8], "=ON"))
                {   showtitlebar = TRUE;
                } elif (!stricmp(&argv[i][8], "=OFF"))
                {   showtitlebar = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "TOOLBAR", 7))
            {   if (!stricmp(&argv[i][7], "=ON"))
                {   showtoolbar = TRUE;
                } elif (!stricmp(&argv[i][7], "=OFF"))
                {   showtoolbar = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!stricmp(argv[i], "UPGRADE"))
            {   upgrade = TRUE;
            } elif (!strnicmp(argv[i], "TURBO", 5))
            {   if (!stricmp(&argv[i][5], "=ON"))
                {   turbo = TRUE;
                } elif (!stricmp(&argv[i][5], "=OFF"))
                {   turbo = FALSE;
                } else
                {   printusage(argv[0]);
            }   }
            elif (!strnicmp(argv[i], "FILE", 4))
            {   if (argv[i][4] != '=')
                {   printusage(argv[0]);
                }
                strcpy(fn_game, (STRPTR) &argv[i][5]);
                gotfilename = TRUE;
                // we don't load it yet
            } else
            {   strcpy(fn_game, argv[i]);
                gotfilename = TRUE;
                // we don't load it yet
    }   }   }

    if (upgrade)
    {   upgradeconfigs();
    } else
    {   if (settings)
        {   oldlanguage = language;
            loadconfig();
            if (language != oldlanguage)
            {   changelanguage();
        }   }
        loadhiscores();
    }

    if (reassociate)
    {   if (assoc[0]) associate(KIND_AOF);
        if (assoc[1]) associate(KIND_BIN);
        if (assoc[2]) associate(KIND_COR);
        if (assoc[3]) associate(KIND_COS);
        if (assoc[4]) associate(KIND_PGM);
        if (assoc[5]) associate(KIND_TVC);
        if (assoc[6]) associate(KIND_ZIP);
    }

    menuheight    = GetSystemMetrics(SM_CYMENU);    // normally 19 (Win9x) or 20 (WinXP)
    titleheight   = GetSystemMetrics(SM_CYCAPTION); // normally 19 (Win9x) or 26 (WinXP)
    toolbarheight = 30; // 12 + bitmap height
    // we get an exact value later

    hFont = CreateFont
    (   8,                      // int nHeight
        0,                      // int nWidth
        0,                      // int nEscapement
        0,                      // int nOrientation
        FW_MEDIUM,              // int nWeight
        FALSE,                  // BYTE bItalic
        FALSE,                  // BYTE bUnderline
        0,                      // BYTE cStrikeOut
        ANSI_CHARSET,           // BYTE nCharSet
        OUT_DEFAULT_PRECIS,     // BYTE nOutPrecision
        CLIP_CHARACTER_PRECIS,  // BYTE nClipPrecision
        DEFAULT_QUALITY,        // BYTE nQuality
        FF_DECORATIVE,          // BYTE nPitchAndFamily
        "MS Sans Serif"         // LPCTSTR lpszFacename
    );
    hBoldFont = CreateFont
    (   8,                      // int nHeight
        0,                      // int nWidth
        0,                      // int nEscapement
        0,                      // int nOrientation
        FW_EXTRABOLD,           // int nWeight
        FALSE,                  // BYTE bItalic
        FALSE,                  // BYTE bUnderline
        0,                      // BYTE cStrikeOut
        ANSI_CHARSET,           // BYTE nCharSet
        OUT_DEFAULT_PRECIS,     // BYTE nOutPrecision
        CLIP_CHARACTER_PRECIS,  // BYTE nClipPrecision
        DEFAULT_QUALITY,        // BYTE nQuality
        FF_DECORATIVE,          // BYTE nPitchAndFamily
        "MS Sans Serif"         // LPCTSTR lpszFacename
    );
    hRussianFont = CreateFont
    (   15,                     // int nHeight
        0,                      // int nWidth
        0,                      // int nEscapement
        0,                      // int nOrientation
        FW_MEDIUM,              // int nWeight
        FALSE,                  // BYTE bItalic
        FALSE,                  // BYTE bUnderline
        0,                      // BYTE cStrikeOut
        RUSSIAN_CHARSET,        // BYTE nCharSet
        OUT_DEFAULT_PRECIS,     // BYTE nOutPrecision
        CLIP_CHARACTER_PRECIS,  // BYTE nClipPrecision
        DEFAULT_QUALITY,        // BYTE nQuality
        FF_DECORATIVE,          // BYTE nPitchAndFamily
        "Comic Sans MS (Cyrillic)" // LPCTSTR lpszFacename
    );
    // should we check these return codes?
    // should we destroy these fonts later?

    wc.cbSize        = sizeof(WNDCLASSEX);
    wc.style         = 0;
    wc.lpfnWndProc   = WndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = InstancePtr;
    wc.hIcon         = LoadIcon(InstancePtr, MAKEINTRESOURCE(IDI_ICON));
    wc.hCursor       = NULL; // was LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)(COLOR_GRAYTEXT+1); // was (HBRUSH)(COLOR_WINDOW+1);
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = g_szClassName;
    wc.hIconSm       = LoadImage(InstancePtr, MAKEINTRESOURCE(IDI_ICON), IMAGE_ICON, 16, 16, 0);

    if (!RegisterClassEx(&wc))
    {   MessageBox(NULL, "Window registration failed!", "WinArcadia: Error", MB_ICONEXCLAMATION | MB_OK);
        cleanexit(EXIT_FAILURE);
    }

    memset(&CtrlBitMapInfo, 0, sizeof(CtrlBitMapInfo));
    CtrlBitMapInfo.Header.biSize        = sizeof(BITMAPINFOHEADER);
    CtrlBitMapInfo.Header.biWidth       = MINIWIDTH;
    CtrlBitMapInfo.Header.biHeight      = -MINIHEIGHT;
    CtrlBitMapInfo.Header.biBitCount    = 32;
    CtrlBitMapInfo.Header.biPlanes      = 1;
    CtrlBitMapInfo.Header.biCompression = BI_RGB;
    CtrlBitMapInfo.Colours[0]           = 0x00FF0000;
    CtrlBitMapInfo.Colours[1]           = 0x0000FF00;
    CtrlBitMapInfo.Colours[2]           = 0x000000FF;

    for (i = 0; i < MENUBITMAPS; i++)
    {   mib[i].bitmap = LoadBitmap(InstancePtr, MAKEINTRESOURCE(mib[i].resource));
    }
    display     = malloc(MAXBOXWIDTH * MAXBOXHEIGHT * sizeof(ULONG)); // (480*8)*(304*4)*4=18,677,760
    ctrldisplay = malloc(MINIWIDTH   * MINIHEIGHT   * sizeof(ULONG));

    // The preferred way of doing it is:
    changemachine(machine, FALSE);
    load_catalog();
    translate();
    // Unfortunately we can't do load_catalog() earlier because it
    // (via generate_autotext()) relies on having coherence between the
    // machine and memmap variables, and changemachine() makes them
    // coherent.

    needaudit = TRUE;
 
    // this pointer setup needs to be done as early as possible (it won't work
    // properly if done much later)
    if (!showpointer)
    {   ShowCursor(FALSE);
    } elif
    (   machine != PIPBUG
     && machine != INSTRUCTOR
     && machine != CD2650
     && machine != CHAOS
     && machine != DOLPHIN
     && (   hostcontroller[0] == CONTROLLER_TRACKBALL
         || hostcontroller[1] == CONTROLLER_TRACKBALL
    )   )
    {   SetCursor(hCross);
    }
    showingpointer = (FLAG) showpointer;

    if (gotfilename)
    {   if (!engine_load(FALSE))
        {   // don't beep, the user may have just wanted to set the directory.
            project_reset(FALSE);
    }   }
    else
    {   strcpy(fn_game, pathpart);
        if (autosave)
        {   changefilepart("AUTOSAVE.COS");
            if (!engine_load(TRUE))
            {   // don't beep
                project_reset(FALSE);
        }   }
        else
        {   project_reset(FALSE);
    }   }

    updatespeech();
    updateprinter();

    GetLocalTime(&TheTime);
    if (TheTime.wMonth == 12 && TheTime.wDay >= 24 && TheTime.wDay <= 26)
    {   xmas = TRUE;
    } else
    {   xmas = FALSE;
    }

    ff_init();
    openwindow();

    load_samples(); // must not be done until after opening the window
    OpenChannels(); // we keep the channels open permanently
    if (!paused) // must not be done until after loading the game of course
    {   sound_on();
    }

    if (lpdi)
    {   for (i = 0; i < joys; i++)
        {   ff_initjoy(i);
        }
        if (ffjoy2[0] && hostcontroller[0] == CONTROLLER_NONE)
        {   hostcontroller[0] = CONTROLLER_1STDJOY;
            updatemenu(MENUITEM_LEFT);
        }
        if (ffjoy2[1] && hostcontroller[1] == CONTROLLER_NONE)
        {   hostcontroller[1] = CONTROLLER_2NDDJOY;
            updatemenu(MENUITEM_RIGHT);
    }   }
    else
    {   // assert(JOYSTICKID1 == 0);
        joys = (int) joyGetNumDevs();
        // T1; zprintf("Maximum supportable joysticks is is %d.\n", joys); T2;
        j = 0;
        for (i = 0; i < joys; i++)
        {   joyInfoEx.dwSize  = sizeof(joyInfoEx);
            joyInfoEx.dwFlags = JOY_RETURNALL;
            if (joyGetPosEx((UINT) i, &joyInfoEx) == JOYERR_NOERROR)
            {   // T1; zprintf("Found joystick unit %d!\n", i); T2;
                if (j == 0)
                {   hostcontroller[0] = CONTROLLER_1STDJOY;
                    unit[0] = i;
                    j = 1;
                } elif (j == 1)
                {   hostcontroller[1] = CONTROLLER_2NDDJOY;
                    unit[1] = i;
                    j = 2;
    }   }   }   }
    // wNumButtons is usually 12
    // szName      is usually "Microsoft PC-joystick driver"
    // szRegKey    is usually ""(FixedKey)
    // szOEMVxD    is usually "joyhid.vxd"
    // manuf. ID   is usually "Microfilth Corporation" (MM_MICROFILTH ) ( 1)
    // prod.  ID:  is usually "Joystick adapter"       (MM_PC_JOYSTICK) (12)

    ProcessPtr = GetCurrentProcess();
    ThreadPtr  = GetCurrentThread();
    setpri();

    for (;;)
    {   if (!paused)
        {   calc_wait();
            emulate();
            update_monitor(FALSE);
            checkinput();
            do_wait();
        } else
        {   update_monitor(FALSE);
            checkinput();
    }   }

    // control can never reach here
    // "The return value is only really useful if your program is designed
    // to be called by another program and you want to return a specific
    // value."

#ifndef _lint
    return 0; // Visual C gets upset if this return is missing, Lint gets upset if it is present :-(
#endif
}

BOOL CALLBACK AboutDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   HKEY hKey;
    int  thesize = MAX_PATH;
 // TEXT portstr[80 + 1];

    switch (Message)
    {
    case WM_INITDIALOG:
        charset = CHARSET_WIN;
        DISCARD SetWindowText( hwnd,                 LLL, MSG_HAIL_ABOUT2,        "About WinArcadia"));
        
        if (language == LANGUAGE_RUS)
        {   DISCARD SendMessage(GetDlgItem(hwnd, IDC_ABOUT         ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_DESC          ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_TRANSLATOR    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_SYSINFO       ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDOK              ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
        }
        DISCARD SetDlgItemText(hwnd, IDC_ABOUT,      TITLEBARTEXT);
        DISCARD SetDlgItemText(hwnd, IDC_DESC,       LLL, MSG_DESC,               "An emulator of Signetics-based machines"));
        DISCARD SetDlgItemText(hwnd, IDC_TRANSLATOR, LLL, MSG_TRANSLATOR,         "English translation by James Jacobs"));
/*      sprintf
        (   portstr,
            GetCatalogStr(CatalogPtr, MSG_PORTER, "%s port by %s"),
            "Windows",
            "James Jacobs"
        );
        DISCARD SetDlgItemText(hwnd, IDC_PORTER,     LLL, MSG_PORTER,             portstr)); */
        DISCARD SetDlgItemText(hwnd, IDC_SYSINFO,    LLL, MSG_SYSTEM_INFORMATION, "System information..."));
        DISCARD SetDlgItemText(hwnd, IDOK,           LLL, MSG_OK2,                "&OK"));
        charset = CHARSET_DOS;

        if (RegOpenKeyEx
        (   HKEY_LOCAL_MACHINE,
            "SOFTWARE\\Microsoft\\Shared Tools\\MSInfo",
            0,
            KEY_READ,
            &hKey
        ) == ERROR_SUCCESS)
        {   if (RegQueryValueEx
            (   hKey,
                "Path",
                0,
                NULL,
                SysInfoName,
                (LPDWORD) &thesize
            ) == ERROR_SUCCESS && SysInfoName[0])
            {   EnableWindow(GetDlgItem(hwnd, IDC_SYSINFO), TRUE);
            }
            DISCARD RegCloseKey(hKey);
        }

        if (sound)
        {   DISCARD PlaySound("sounds\\amigan.wav", NULL, SND_ASYNC | SND_FILENAME | SND_NODEFAULT);
        }
        // play_sample(SAMPLE_AMIGAN); has clicks before and after sample for some reason

        DISCARD SendMessage(GetDlgItem(hwnd, IDC_ABOUT), WM_SETFONT, (WPARAM) hBoldFont, MAKELPARAM(TRUE, 0));

        return TRUE;
    case WM_CLOSE: // no need for acase
        clearkybd();
        DISCARD EndDialog(hwnd, 0);
        return TRUE;
    case WM_COMMAND: // no need for acase
        if
        (   (LOWORD(wParam) == IDC_LINK1)
         || (LOWORD(wParam) == IDC_AMIGANSOFTWARE)
        )
        {   openurl("http://amigan.1emu.net/releases/");
        } elif
        (   (LOWORD(wParam) == IDC_LINK2)
         || (LOWORD(wParam) == IDC_EA2001CENTRAL)
        )
        {   openurl("http://amigan.yatho.com/");
        } elif
        (   (LOWORD(wParam) == IDC_SYSINFO)
        )
        {   openurl(SysInfoName);
        } elif (LOWORD (wParam) == IDOK)
        {   clearkybd();
            DISCARD EndDialog(hwnd, 0);
        }
        return TRUE;
    default: // no need for adefault
        return FALSE;
}   }

BOOL CALLBACK RedefineDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   TRANSIENT int    i, j,
                     mousex, mousey;
    TRANSIENT LONG   gid;
    PERSIST   HBRUSH hBrush[4];

    switch (Message)
    {
    case WM_INITDIALOG:
        charset = CHARSET_WIN;
        DISCARD SetWindowText( hwnd,                      LLL, MSG_HAIL_REDEFINEKEYS, "Redefine Keys"             ));

        if (language == LANGUAGE_RUS)
        {   DISCARD SendMessage(GetDlgItem(hwnd, IDC_KEY_START), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_KEY_A    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_KEY_B    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_KEY_RESET), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));

            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_2        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_X1       ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_X2       ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_X3       ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_1        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_X4       ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_3        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_4        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_5        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_6        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_7        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_8        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_9        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_C        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_0        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_E        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_UPLEFT   ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_UP       ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_UPRIGHT  ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_LEFT     ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_RIGHT    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_DOWNLEFT ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_DOWN     ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_DOWNRIGHT), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_1ST      ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_2ND      ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_3RD      ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P1_4TH      ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));

            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_2        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_X1       ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_X2       ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_X3       ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_1        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_X4       ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_3        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_4        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_5        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_6        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_7        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_8        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_9        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_C        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_0        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_E        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_UPLEFT   ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_UP       ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_UPRIGHT  ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_LEFT     ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_RIGHT    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_DOWNLEFT ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_DOWN     ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_DOWNRIGHT), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_1ST      ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_2ND      ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_3RD      ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_P2_4TH      ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));

            DISCARD SendMessage(GetDlgItem(hwnd, IDC_RESETTODEFAULTS), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDOK               ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDCANCEL           ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_LABEL_LEFT     ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_LABEL_RIGHT    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_F1             ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_F2             ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_F3             ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_F4             ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
        }
        DISCARD SetDlgItemText(hwnd, IDC_RESETTODEFAULTS, LLL, MSG_RESETTODEFAULTS  , "Reset to defaults"         ));
        DISCARD SetDlgItemText(hwnd, IDOK               , LLL, MSG_OK2              , "&OK"                       ));
        DISCARD SetDlgItemText(hwnd, IDCANCEL           , LLL, MSG_CANCEL2          , "&Cancel"                   ));
        DISCARD SetDlgItemText(hwnd, IDC_LABEL_LEFT     , LLL, MSG_LEFTCONTROLLER   , "Left Controller"           ));
        DISCARD SetDlgItemText(hwnd, IDC_LABEL_RIGHT    , LLL, MSG_RIGHTCONTROLLER  , "Right Controller"          ));
        DISCARD SetDlgItemText(hwnd, IDC_F1             , LLL, MSG_CONSOLE_START    , "START/START/START/SENS/P1:"));
        DISCARD SetDlgItemText(hwnd, IDC_F2             , LLL, MSG_CONSOLE_A        , "A/SELECT/UC/INT/P2:"       ));
        DISCARD SetDlgItemText(hwnd, IDC_F3             , LLL, MSG_CONSOLE_B        , "B/-/LC/MON/Coin A:"        ));
        DISCARD SetDlgItemText(hwnd, IDC_F4             , LLL, MSG_CONSOLE_RESET    , "-/-/RESET/RST/Coin B:"     ));
        charset = CHARSET_DOS;

        for (i = 0; i < 2; i++)
        {   for (j = 0; j < NUMKEYS; j++)
            {   temp_keypads[i][j] = keypads[i][j];
        }   }
        for (i = 0; i < 4; i++)
        {   temp_console[i] = console[i];
        }

        if (machine != ARCADIA) // eg. INTERTON, ELEKTOR, INSTRUCTOR, COINOP, DOLPHIN
        {   EnableWindow(GetDlgItem(hwnd, IDC_P1_X1), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_X2), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_X3), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_X4), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_X1), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_X2), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_X3), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_X4), FALSE);
        }
        if
        (   machine != ARCADIA
         && machine != INTERTON
         && machine != ELEKTOR
         && machine != INSTRUCTOR
         && machine != ASTROWARS
         && machine != GALAXIA
         && machine != MALZAK
        ) // eg. DOLPHIN
        {   EnableWindow(GetDlgItem(hwnd, IDC_KEY_START), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_KEY_A    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_4     ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_5     ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_7     ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_8     ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_C     ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_0     ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_2ND  ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_3RD  ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_4TH  ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_2ND  ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_3RD  ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_4TH  ), FALSE);
        }
        if
        (   machine != ARCADIA
         && machine != INTERTON
         && machine != ELEKTOR
         && machine != ASTROWARS
         && machine != GALAXIA
         && machine != MALZAK
        ) // eg. INSTRUCTOR, DOLPHIN
        {   EnableWindow(GetDlgItem(hwnd, IDC_P1_1ST  ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_1ST  ), FALSE);
        }
        if
        (   machine != ARCADIA
         && machine != ELEKTOR
         && machine != INSTRUCTOR
         && machine != ASTROWARS
         && machine != GALAXIA
         && machine != MALZAK
        ) // eg. INTERTON, DOLPHIN
        {   EnableWindow(GetDlgItem(hwnd, IDC_KEY_B), FALSE);
        }
        if
        (   machine != ELEKTOR
         && machine != INSTRUCTOR
         && machine != ASTROWARS
         && machine != GALAXIA
        ) // eg. ARCADIA, INTERTON, Malzak, DOLPHIN
        {   EnableWindow(GetDlgItem(hwnd, IDC_KEY_RESET), FALSE);
        }
        if
        (   machine != ARCADIA
         && machine != INTERTON
         && machine != ELEKTOR
         && machine != MALZAK
         && machine != ASTROWARS
         && machine != GALAXIA
        ) // eg. INSTRUCTOR, DOLPHIN
        {   EnableWindow(GetDlgItem(hwnd, IDC_P1_LEFT ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_RIGHT), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_LEFT ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_RIGHT), FALSE);
        }
        if
        (   machine != ARCADIA
         && machine != INTERTON
         && machine != ELEKTOR
         && machine != MALZAK
        ) // eg. INSTRUCTOR, Astro Wars, Galaxia, DOLPHIN
        {   EnableWindow(GetDlgItem(hwnd, IDC_P1_UPLEFT   ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_UP       ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_UPRIGHT  ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_DOWNLEFT ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_DOWN     ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_DOWNRIGHT), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_UPLEFT   ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_UP       ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_UPRIGHT  ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_DOWNLEFT ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_DOWN     ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_DOWNRIGHT), FALSE);
        }
        if
        (   machine != ARCADIA
         && machine != INTERTON
         && machine != ELEKTOR
         && machine != INSTRUCTOR
         && machine != DOLPHIN
        ) // eg. coin-ops
        {   EnableWindow(GetDlgItem(hwnd, IDC_P1_1    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_3    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_4    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_5    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_6    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_7    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_8    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_9    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_C    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_0    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_E    ), FALSE);

            EnableWindow(GetDlgItem(hwnd, IDC_P2_1    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_3    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_4    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_5    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_6    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_7    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_8    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_9    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_C    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_0    ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_E    ), FALSE);

            EnableWindow(GetDlgItem(hwnd, IDC_P1_2ND  ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_3RD  ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P1_4TH  ), FALSE);

            EnableWindow(GetDlgItem(hwnd, IDC_P2_2ND  ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_3RD  ), FALSE);
            EnableWindow(GetDlgItem(hwnd, IDC_P2_4TH  ), FALSE);
        }
        
        refreshkeys(hwnd);                    

        hBrush[0] = CreateSolidBrush(EMUPEN_RED);
        hBrush[1] = CreateSolidBrush(EMUPEN_BLUE);
        hBrush[2] = CreateSolidBrush(EMUPEN_GREEN);
        hBrush[3] = CreateSolidBrush(EMUPEN_YELLOW);

        return TRUE;
    case WM_LBUTTONDBLCLK: // no need for acase
    case WM_LBUTTONDOWN:
        if (machine == INSTRUCTOR)
        {   break;
        }
        mousex = (int) (LOWORD(lParam)) / 2; // pixels -> dialog units
        mousey = (int) (HIWORD(lParam) - 50) / 2; // pixels -> dialog units
        if (mousex >= 12 && mousex <= 66)
        {   if     (mousey >= 168 && mousey <= 176)
            {   if
                (   machine == ARCADIA
                 || machine == INTERTON
                 || machine == ELEKTOR
                 || machine == ASTROWARS
                 || machine == GALAXIA
                 || machine == MALZAK
                ) // excludes INSTRUCTOR, DOLPHIN
                {   side = 0; whichkey =  0; // IDC_P1_1ST
                    subreq(hwnd);
            }   }
            elif (mousey >= 178 && mousey <= 186)
            {   if
                (   machine == ARCADIA
                 || machine == INTERTON
                 || machine == ELEKTOR
                ) // excludes INSTRUCTOR, ASTROWARS, GALAXIA, MALZAK, DOLPHIN
                {   side = 0; whichkey = 21; // IDC_P1_2ND
                    subreq(hwnd);
            }   }
            elif (mousey >= 188 && mousey <= 196)
            {   if
                (   machine == ARCADIA
                 || machine == INTERTON
                 || machine == ELEKTOR
                ) // excludes INSTRUCTOR, ASTROWARS, GALAXIA, MALZAK, DOLPHIN
                {   side = 0; whichkey = 22; // IDC_P1_3RD
                    subreq(hwnd);
            }   }
            elif (mousey >= 198 && mousey <= 206)
            {   if
                (   machine == ARCADIA
                 || machine == INTERTON
                 || machine == ELEKTOR
                ) // excludes INSTRUCTOR, ASTROWARS, GALAXIA, MALZAK, DOLPHIN
                {   side = 0; whichkey = 23; // IDC_P1_4TH
                    subreq(hwnd);
        }   }   }
        elif (mousex >= 100 && mousex <= 154)
        {   if     (mousey >= 168 && mousey <= 176)
            {   if
                (   machine == ARCADIA
                 || machine == INTERTON
                 || machine == ELEKTOR
                 || machine == ASTROWARS
                 || machine == GALAXIA
                 || machine == MALZAK
                ) // excludes INSTRUCTOR, DOLPHIN
                {   side = 1; whichkey =  0; // IDC_P2_1ST
                    subreq(hwnd);
            }   }
            elif (mousey >= 178 && mousey <= 186)
            {   if
                (   machine == ARCADIA
                 || machine == INTERTON
                 || machine == ELEKTOR
                ) // excludes INSTRUCTOR, ASTROWARS, GALAXIA, MALZAK, DOLPHIN
                {   side = 1; whichkey = 21; // IDC_P2_2ND
                    subreq(hwnd);
            }   }
            elif (mousey >= 188 && mousey <= 196)
            {   if
                (   machine == ARCADIA
                 || machine == INTERTON
                 || machine == ELEKTOR
                ) // excludes INSTRUCTOR, ASTROWARS, GALAXIA, MALZAK, DOLPHIN
                {   side = 1; whichkey = 22; // IDC_P2_3RD
                    subreq(hwnd);
            }   }
            elif (mousey >= 198 && mousey <= 206)
            {   if
                (   machine == ARCADIA
                 || machine == INTERTON
                 || machine == ELEKTOR
                ) // excludes INSTRUCTOR, ASTROWARS, GALAXIA, MALZAK, DOLPHIN
                {   side = 1; whichkey = 23; // IDC_P2_4TH
                    subreq(hwnd);
        }   }   }
    acase WM_CTLCOLORSTATIC:
        gid = GetWindowLong((HWND) lParam, GWL_ID);
        
        if (    gid == IDC_P1_1ST || gid == IDC_P2_1ST)
        {   SetBkColor((HDC) wParam, EMUPEN_RED);
            return (LRESULT) hBrush[0];
        } elif (gid == IDC_P1_2ND || gid == IDC_P2_2ND)
        {   SetBkColor((HDC) wParam, EMUPEN_BLUE);
            return (LRESULT) hBrush[1];
        } elif (gid == IDC_P1_3RD || gid == IDC_P2_3RD)
        {   SetBkColor((HDC) wParam, EMUPEN_GREEN);
            return (LRESULT) hBrush[2];
        } elif (gid == IDC_P1_4TH || gid == IDC_P2_4TH)
        {   SetBkColor((HDC) wParam, EMUPEN_YELLOW);
            return (LRESULT) hBrush[3];
        } else return TRUE;
    case WM_CLOSE:
        for (i = 0; i < 2; i++)
        {   for (j = 0; j < NUMKEYS; j++)
            {   keypads[i][j] = temp_keypads[i][j];
        }   }
        for (i = 0; i < 4; i++)
        {   console[i] = temp_console[i];
        }
        clearkybd();
        DISCARD EndDialog(hwnd, 0);
        return TRUE;
    case WM_DESTROY:
        DeleteObject(hBrush[0]);
        DeleteObject(hBrush[1]);
        DeleteObject(hBrush[2]);
        DeleteObject(hBrush[3]);
        return TRUE;
    case WM_COMMAND:
        switch (LOWORD(wParam))
        {
        case IDOK:
            for (i = 0; i < 2; i++)
            {   for (j = 0; j < NUMKEYS; j++)
                {   keypads[i][j] = temp_keypads[i][j];
            }   }
            for (i = 0; i < 4; i++)
            {   console[i] = temp_console[i];
            }
        //lint -fallthrough
        case IDCANCEL:
            clearkybd();
            DISCARD EndDialog(hwnd, 0);
            return TRUE;
        case IDC_RESETTODEFAULTS: // no need for acase
            for (i = 0; i < 2; i++)
            {   for (j = 0; j < NUMKEYS; j++)
                {   temp_keypads[i][j] = default_keypads[i][j];
            }   }
            for (i = 0; i < 4; i++)
            {   temp_console[i] = default_console[i];
            }
            refreshkeys(hwnd);
        acase IDC_P1_1:
            side = 0; whichkey = 1;
            subreq(hwnd);
        acase IDC_P1_2:
            side = 0; whichkey = 2;
            subreq(hwnd);
        acase IDC_P1_3:
            side = 0; whichkey = 3;
            subreq(hwnd);
        acase IDC_P1_4:
            side = 0; whichkey = 4;
            subreq(hwnd);
        acase IDC_P1_5:
            side = 0; whichkey = 5;
            subreq(hwnd);
        acase IDC_P1_6:
            side = 0; whichkey = 6;
            subreq(hwnd);
        acase IDC_P1_7:
            side = 0; whichkey = 7;
            subreq(hwnd);
        acase IDC_P1_8:
            side = 0; whichkey = 8;
            subreq(hwnd);
        acase IDC_P1_9:
            side = 0; whichkey = 9;
            subreq(hwnd);
        acase IDC_P1_C:
            side = 0; whichkey = 10;
            subreq(hwnd);
        acase IDC_P1_0:
            side = 0; whichkey = 11;
            subreq(hwnd);
        acase IDC_P1_E:
            side = 0; whichkey = 12;
            subreq(hwnd);
        acase IDC_P1_X1:
            side = 0; whichkey = 13;
            subreq(hwnd);
        acase IDC_P1_X2:
            side = 0; whichkey = 14;
            subreq(hwnd);
        acase IDC_P1_X3:
            side = 0; whichkey = 15;
            subreq(hwnd);
        acase IDC_P1_X4:
            side = 0; whichkey = 16;
            subreq(hwnd);
        acase IDC_P1_UPLEFT:
            side = 0; whichkey = 24;
            subreq(hwnd);
        acase IDC_P1_UP:
            side = 0; whichkey = 17;
            subreq(hwnd);
        acase IDC_P1_UPRIGHT:
            side = 0; whichkey = 25;
            subreq(hwnd);
        acase IDC_P1_LEFT:
            side = 0; whichkey = 19;
            subreq(hwnd);
        acase IDC_P1_RIGHT:
            side = 0; whichkey = 20;
            subreq(hwnd);
        acase IDC_P1_DOWNLEFT:
            side = 0; whichkey = 26;
            subreq(hwnd);
        acase IDC_P1_DOWN:
            side = 0; whichkey = 18;
            subreq(hwnd);
        acase IDC_P1_DOWNRIGHT:
            side = 0; whichkey = 27;
            subreq(hwnd);
        acase IDC_P2_1:
            side = 1; whichkey = 1;
            subreq(hwnd);
        acase IDC_P2_2:
            side = 1; whichkey = 2;
            subreq(hwnd);
        acase IDC_P2_3:
            side = 1; whichkey = 3;
            subreq(hwnd);
        acase IDC_P2_4:
            side = 1; whichkey = 4;
            subreq(hwnd);
        acase IDC_P2_5:
            side = 1; whichkey = 5;
            subreq(hwnd);
        acase IDC_P2_6:
            side = 1; whichkey = 6;
            subreq(hwnd);
        acase IDC_P2_7:
            side = 1; whichkey = 7;
            subreq(hwnd);
        acase IDC_P2_8:
            side = 1; whichkey = 8;
            subreq(hwnd);
        acase IDC_P2_9:
            side = 1; whichkey = 9;
            subreq(hwnd);
        acase IDC_P2_C:
            side = 1; whichkey = 10;
            subreq(hwnd);
        acase IDC_P2_0:
            side = 1; whichkey = 11;
            subreq(hwnd);
        acase IDC_P2_E:
            side = 1; whichkey = 12;
            subreq(hwnd);
        acase IDC_P2_X1:
            side = 1; whichkey = 13;
            subreq(hwnd);
        acase IDC_P2_X2:
            side = 1; whichkey = 14;
            subreq(hwnd);
        acase IDC_P2_X3:
            side = 1; whichkey = 15;
            subreq(hwnd);
        acase IDC_P2_X4:
            side = 1; whichkey = 16;
            subreq(hwnd);
        acase IDC_P2_UPLEFT:
            side = 1; whichkey = 24;
            subreq(hwnd);
        acase IDC_P2_UP:
            side = 1; whichkey = 17;
            subreq(hwnd);
        acase IDC_P2_UPRIGHT:
            side = 1; whichkey = 25;
            subreq(hwnd);
        acase IDC_P2_LEFT:
            side = 1; whichkey = 19;
            subreq(hwnd);
        acase IDC_P2_RIGHT:
            side = 1; whichkey = 20;
            subreq(hwnd);
        acase IDC_P2_DOWNLEFT:
            side = 1; whichkey = 26;
            subreq(hwnd);
        acase IDC_P2_DOWN:
            side = 1; whichkey = 18;
            subreq(hwnd);
        acase IDC_P2_DOWNRIGHT:
            side = 1; whichkey = 27;
            subreq(hwnd);
        acase IDC_KEY_START:
            side = 2; whichkey = 0;
            subreq(hwnd);
        acase IDC_KEY_A:
            side = 2; whichkey = 1;
            subreq(hwnd);
        acase IDC_KEY_B:
            side = 2; whichkey = 2;
            subreq(hwnd);
        acase IDC_KEY_RESET:
            side = 2; whichkey = 3;
            subreq(hwnd);
    }   }
    
    return FALSE;
}

EXPORT void refreshkeys(HWND hwnd)
{   FLAG ok;
    int  i,
         j;

    DISCARD SetDlgItemText(hwnd, IDC_KEY_START, keyname[temp_console[0]].name);
    DISCARD SetDlgItemText(hwnd, IDC_KEY_A    , keyname[temp_console[1]].name);
    DISCARD SetDlgItemText(hwnd, IDC_KEY_B    , keyname[temp_console[2]].name);
    DISCARD SetDlgItemText(hwnd, IDC_KEY_RESET, keyname[temp_console[3]].name);

    DISCARD SetDlgItemText(hwnd, IDC_P1_2     , keyname[temp_keypads[0][ 2]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_X1    , keyname[temp_keypads[0][13]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_X2    , keyname[temp_keypads[0][14]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_X3    , keyname[temp_keypads[0][15]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_1     , keyname[temp_keypads[0][ 1]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_X4    , keyname[temp_keypads[0][16]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_3     , keyname[temp_keypads[0][ 3]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_4     , keyname[temp_keypads[0][ 4]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_5     , keyname[temp_keypads[0][ 5]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_6     , keyname[temp_keypads[0][ 6]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_7     , keyname[temp_keypads[0][ 7]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_8     , keyname[temp_keypads[0][ 8]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_9     , keyname[temp_keypads[0][ 9]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_C     , keyname[temp_keypads[0][10]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_0     , keyname[temp_keypads[0][11]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_E     , keyname[temp_keypads[0][12]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_UPLEFT  , keyname[temp_keypads[0][24]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_UP      , keyname[temp_keypads[0][17]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_UPRIGHT , keyname[temp_keypads[0][25]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_LEFT    , keyname[temp_keypads[0][19]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_RIGHT   , keyname[temp_keypads[0][20]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_DOWNLEFT, keyname[temp_keypads[0][26]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_DOWN    , keyname[temp_keypads[0][18]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_DOWNRIGHT,keyname[temp_keypads[0][27]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_1ST     , keyname[temp_keypads[0][ 0]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_2ND     , keyname[temp_keypads[0][21]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_3RD     , keyname[temp_keypads[0][22]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P1_4TH     , keyname[temp_keypads[0][23]].name);

    DISCARD SetDlgItemText(hwnd, IDC_P2_2     , keyname[temp_keypads[1][ 2]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_X1    , keyname[temp_keypads[1][13]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_X2    , keyname[temp_keypads[1][14]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_X3    , keyname[temp_keypads[1][15]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_1     , keyname[temp_keypads[1][ 1]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_X4    , keyname[temp_keypads[1][16]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_3     , keyname[temp_keypads[1][ 3]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_4     , keyname[temp_keypads[1][ 4]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_5     , keyname[temp_keypads[1][ 5]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_6     , keyname[temp_keypads[1][ 6]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_7     , keyname[temp_keypads[1][ 7]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_8     , keyname[temp_keypads[1][ 8]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_9     , keyname[temp_keypads[1][ 9]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_C     , keyname[temp_keypads[1][10]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_0     , keyname[temp_keypads[1][11]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_E     , keyname[temp_keypads[1][12]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_UPLEFT  , keyname[temp_keypads[1][24]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_UP      , keyname[temp_keypads[1][17]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_UPRIGHT , keyname[temp_keypads[1][25]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_LEFT    , keyname[temp_keypads[1][19]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_RIGHT   , keyname[temp_keypads[1][20]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_DOWNLEFT, keyname[temp_keypads[1][26]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_DOWN    , keyname[temp_keypads[1][18]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_DOWNRIGHT,keyname[temp_keypads[1][27]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_1ST     , keyname[temp_keypads[1][ 0]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_2ND     , keyname[temp_keypads[1][21]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_3RD     , keyname[temp_keypads[1][22]].name);
    DISCARD SetDlgItemText(hwnd, IDC_P2_4TH     , keyname[temp_keypads[1][23]].name);

    ok = TRUE;
    for (i = 0; i < 2; i++)
    {   for (j = 0; j < NUMKEYS; j++)
        {   if (temp_keypads[i][j] != default_keypads[i][j])
            {   ok = FALSE;
                break; // for speed
    }   }   }
    if (ok)
    {   for (i = 0; i < 4; i++)
        {   if (temp_console[i] != default_console[i])
            {   ok = FALSE;
                break; // for speed
    }   }   }

    if (ok) // all are defaults
    {   EnableWindow(GetDlgItem(hwnd, IDC_RESETTODEFAULTS), FALSE);
    } else // not all are defaults
    {   EnableWindow(GetDlgItem(hwnd, IDC_RESETTODEFAULTS), TRUE );
}   }

BOOL CALLBACK RedefineSubDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   ULONG code;

    switch (Message)
    {
    case WM_INITDIALOG:
        charset = CHARSET_WIN;
        DISCARD SetWindowText(hwnd, LLL, MSG_HAIL_REDEFINEKEY, "Redefine key..."));
        charset = CHARSET_DOS;

        clearkybd();
        return TRUE;
    case WM_KEYUP: // WM_KEYDOWN causes problems with cursor down and with cursor right
    // no need for acase
        code = (lParam & 33488896) >> 16;

        if (!keyname[code].reserved)
        {   if (side == 2)
            {   temp_console[whichkey] = (UWORD) code;
            } else
            {   temp_keypads[side][whichkey] = (UWORD) code;
            }
            clearkybd();
            DISCARD EndDialog(hwnd, 0);
        }

        return FALSE;
    case WM_CLOSE: // no need for acase
        clearkybd();
        DISCARD EndDialog(hwnd, 0);
        return TRUE;
    default: // no need for adefault
        return FALSE;
}   }

BOOL CALLBACK SetDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   int outcome = 0;

    switch (Message)
    {
    case WM_INITDIALOG:
        charset = CHARSET_WIN;
        switch (dlgtype)
        {
        case KIND_BP:
            DISCARD SetWindowText(hwnd, LLL, MSG_HAIL_ADDBP,   "Add Breakpoint"  ));
        acase KIND_WP:
            DISCARD SetWindowText(hwnd, LLL, MSG_HAIL_ADDWP,   "Add Watchpoint"  ));
        acase KIND_BC:
            DISCARD SetWindowText(hwnd, LLL, MSG_HAIL_CLEARBP, "Clear Breakpoint"));
        acase KIND_WC:
            DISCARD SetWindowText(hwnd, LLL, MSG_HAIL_CLEARWP, "Clear Watchpoint"));
        }
        DISCARD SetDlgItemText(hwnd, IDC_POINT, "");
        charset = CHARSET_DOS;

        // DISCARD SetForegroundWindow(hwnd); doesn't work

        return TRUE;
    case WM_CLOSE: // no need for acase
        outcome = 2;
    acase WM_COMMAND:
        if (LOWORD (wParam) == IDOK)
        {   outcome = 1;
        } else
        {   return FALSE;
        }
    adefault:
        return FALSE;
    }

    DISCARD GetWindowText(GetDlgItem(hwnd, IDC_POINT), friendly, sizeof(friendly));

    if (outcome == 1)
    {   switch (dlgtype)
        {
        case KIND_BP:
            DISCARD bp_add(friendly_to_number(friendly));
        acase KIND_WP:
            DISCARD wp_add(friendly_to_number(friendly));
        acase KIND_BC:
            DISCARD bp_clear(friendly_to_number(friendly));
        acase KIND_WC:
            DISCARD wp_clear(friendly_to_number(friendly));
    }   }

    clearkybd();
    DISCARD EndDialog(hwnd, 0);

    return TRUE;
}

BOOL CALLBACK RastNDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   switch (Message)
    {
    case WM_INITDIALOG:
        charset = CHARSET_WIN;
        DISCARD SetWindowText(hwnd, LLL, MSG_HAIL_RUNTORASTERLINEN, "Raster line <n>"));
        charset = CHARSET_DOS;

        DISCARD stcl_d(globalstring, rastn); // number -> decimal string
        DISCARD SetDlgItemText(hwnd, IDC_POINT, globalstring);
        // DISCARD SetForegroundWindow(hwnd); doesn't work

        return TRUE;
    case WM_CLOSE: // no need for acase
        ;
    acase WM_COMMAND:
        if (LOWORD (wParam) == IDOK)
        {   break;
        } // implied else
        if (LOWORD (wParam) == IDCANCEL)
        {   clearkybd();
            DISCARD EndDialog(hwnd, 0);

            return TRUE;
        } // implied else
        return FALSE;
    default: // no need for adefault
        return FALSE;
    }

    DISCARD GetWindowText(GetDlgItem(hwnd, IDC_POINT), globalstring, sizeof(globalstring));
    DISCARD stcd_l(globalstring, &rastn); // string -> decimal number
    // we should really bounds check this number

    clearkybd();
    DISCARD EndDialog(hwnd, 0);

    return TRUE;
}

BOOL CALLBACK MonitorDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   TRANSIENT PAINTSTRUCT ps;
    PERSIST   HDC         MonitorRastPtr; // PERSISTent for speed
    PERSIST   FLAG        zoom           = FALSE;
    PERSIST   HICON       hicon,
                          zoomicon;

    switch (Message)
    {
    case WM_INITDIALOG:
        charset = CHARSET_WIN;
        DISCARD SetWindowText(hwnd, GetCatalogStr(CatalogPtr, machines[machine].monitor1, machines[machine].monitor2));        

        switch (machine)
        {
        case ARCADIA:
            if (language == LANGUAGE_RUS)
            {   DISCARD SendMessage(GetDlgItem(hwnd, IDC_UVI_GROUP     ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
                DISCARD SendMessage(GetDlgItem(hwnd, IDC_RAM_GROUP     ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
                DISCARD SendMessage(GetDlgItem(hwnd, IDC_GAD_RASTER    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            }
            DISCARD SetDlgItemText(hwnd, IDC_UVI_GROUP, LLL, MSG_UVI_GAD, "2637 UVI Status"));
            DISCARD SetDlgItemText(hwnd, IDC_RAM_GROUP, LLL, MSG_UVIRAM_GAD, "UVI RAM Contents"));
            DISCARD SetDlgItemText(hwnd, IDC_GAD_RASTER, LLL, MSG_RASTER_LONG, "Raster:"));

            if (xmas)
            {   hicon = LoadIcon(InstancePtr, MAKEINTRESOURCE(IDI_XMAS));
            } else
            {   hicon = LoadIcon(InstancePtr, MAKEINTRESOURCE(hostmachines[ARCADIA].icon));
            }
            DISCARD SendMessage(GetDlgItem(hwnd, IDL_AMONITORGLYPH), STM_SETICON, (WPARAM) hicon, (LPARAM) 0);
        acase INTERTON:
            if (language == LANGUAGE_RUS)
            {   DISCARD SendMessage(GetDlgItem(hwnd, IDC_PVI1_GROUP    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
                DISCARD SendMessage(GetDlgItem(hwnd, IDC_RAM_GROUP     ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
                DISCARD SendMessage(GetDlgItem(hwnd, IDC_GAD_RASTER    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            }
            DISCARD SetDlgItemText(hwnd, IDC_PVI1_GROUP, LLL, MSG_PVI_GAD,     "2636 PVI Status"));
            DISCARD SetDlgItemText(hwnd, IDC_RAM_GROUP,  LLL, MSG_PVIRAM_GAD,  "PVI RAM Contents"));
            DISCARD SetDlgItemText(hwnd, IDC_GAD_RASTER, LLL, MSG_RASTER_LONG, "Raster:"));
        acase ELEKTOR:
            if (memmap == MEMMAP_E)
            {   DISCARD SetWindowText( hwnd, LLL, MSG_HAIL_MONITOR_E_E, "Elektor TV Games Computer (Basic) Monitor"   ));
            } else
            {   DISCARD SetWindowText( hwnd, LLL, MSG_HAIL_MONITOR_E_F, "Elektor TV Games Computer (Expanded) Monitor"));

                if (language == LANGUAGE_RUS)
                {   DISCARD SendMessage(GetDlgItem(hwnd, IDC_PSG1_GROUP    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
                    DISCARD SendMessage(GetDlgItem(hwnd, IDC_PSG2_GROUP    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
                }
                DISCARD SetDlgItemText(hwnd, IDC_PSG1_GROUP, LLL, MSG_PSG1_GAD, "AY-3-8910 PSG #1 Status"));
                DISCARD SetDlgItemText(hwnd, IDC_PSG2_GROUP, LLL, MSG_PSG2_GAD, "AY-3-8910 PSG #2 Status"));
            }
            if (language == LANGUAGE_RUS)
            {   DISCARD SendMessage(GetDlgItem(hwnd, IDC_PVI1_GROUP    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
                DISCARD SendMessage(GetDlgItem(hwnd, IDC_GAD_RASTER    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            }
            DISCARD SetDlgItemText(hwnd, IDC_PVI1_GROUP, LLL, MSG_PVI_GAD,     "2636 PVI Status"));
            DISCARD SetDlgItemText(hwnd, IDC_GAD_RASTER, LLL, MSG_RASTER_LONG, "Raster:"        ));
        acase ASTROWARS:
            if (language == LANGUAGE_RUS)
            {   DISCARD SendMessage(GetDlgItem(hwnd, IDC_GAD_RASTER    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
                DISCARD SendMessage(GetDlgItem(hwnd, IDC_PVI1_GROUP    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            }
            DISCARD SetDlgItemText(hwnd, IDC_GAD_RASTER, LLL, MSG_RASTER_LONG, "Raster:"        ));
            DISCARD SetDlgItemText(hwnd, IDC_PVI1_GROUP, LLL, MSG_PVI_GAD,     "2636 PVI Status"));
        acase GALAXIA:
            if (language == LANGUAGE_RUS)
            {   DISCARD SendMessage(GetDlgItem(hwnd, IDC_GAD_RASTER    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
                DISCARD SendMessage(GetDlgItem(hwnd, IDC_PVI1_GROUP    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
                DISCARD SendMessage(GetDlgItem(hwnd, IDC_PVI2_GROUP    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
                DISCARD SendMessage(GetDlgItem(hwnd, IDC_PVI3_GROUP    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            }
            DISCARD SetDlgItemText(hwnd, IDC_GAD_RASTER, LLL, MSG_RASTER_LONG, "Raster:"));
            DISCARD SetDlgItemText(hwnd, IDC_PVI1_GROUP, LLL, MSG_PVI1_GAD, "2636 PVI #1 Status"));
            DISCARD SetDlgItemText(hwnd, IDC_PVI2_GROUP, LLL, MSG_PVI2_GAD, "2636 PVI #2 Status"));
            DISCARD SetDlgItemText(hwnd, IDC_PVI3_GROUP, LLL, MSG_PVI3_GAD, "2636 PVI #3 Status"));
        acase MALZAK:          
            if (language == LANGUAGE_RUS)
            {   DISCARD SendMessage(GetDlgItem(hwnd, IDC_PVI1_GROUP    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
                DISCARD SendMessage(GetDlgItem(hwnd, IDC_PVI2_GROUP    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
                DISCARD SendMessage(GetDlgItem(hwnd, IDC_GAD_RASTER    ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            }
            DISCARD SetDlgItemText(hwnd, IDC_PVI1_GROUP, LLL, MSG_PVI1_GAD, "2636 PVI #1 Status"));
            DISCARD SetDlgItemText(hwnd, IDC_PVI2_GROUP, LLL, MSG_PVI2_GAD, "2636 PVI #2 Status"));
            DISCARD SetDlgItemText(hwnd, IDC_GAD_RASTER, LLL, MSG_RASTER_LONG, "Raster:"));
        }

        if (language == LANGUAGE_RUS)
        {   DISCARD SendMessage(GetDlgItem(hwnd, IDC_CPU_GROUP     ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_GAD_FRAME     ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_GAD_TIME      ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_GAD_CLOCK     ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
        }
        DISCARD SetDlgItemText(hwnd, IDC_CPU_GROUP, LLL, MSG_CPU_GAD,    "2650 CPU Status"));
        DISCARD SetDlgItemText(hwnd, IDC_GAD_FRAME, LLL, MSG_FRAME_LONG, "Frame:"));
        DISCARD SetDlgItemText(hwnd, IDC_GAD_TIME,  LLL, MSG_TIME_LONG,  "Time:"));
        DISCARD SetDlgItemText(hwnd, IDC_GAD_CLOCK, LLL, MSG_CLOCK_LONG, "Clock:"));
        charset = CHARSET_DOS;

        return FALSE;
    acase WM_ACTIVATE:
        do_autopause(wParam, lParam);
    acase WM_CLOSE:
        DestroyWindow(hwnd);
        MonitorWindowPtr = NULL;
        updatemenu(MENUITEM_VIEWMONITOR);
    acase WM_PAINT:
        BeginPaint(hwnd, &ps);
        DISCARD EndPaint(hwnd, &ps);
        if (machine == ARCADIA || machine == INTERTON || machine == ELEKTOR)
        {   if (ignorepaint)
            {   ignorepaint = FALSE;
            } else
            {   ignorepaint = TRUE;
                UpdateWindow(GetDlgItem(MonitorWindowPtr, IDC_AXES));
                MonitorRastPtr = GetDC(GetDlgItem(MonitorWindowPtr, IDC_AXES));
                FillRect(MonitorRastPtr, &axesrect, GetStockObject(WHITE_BRUSH));
                ReleaseDC(MonitorWindowPtr, MonitorRastPtr);
        }   }
        return FALSE; // important!
    acase WM_COMMAND:
        switch (LOWORD(wParam))
        {
        case IDC_CANDY2:
            if (zoom)
            {   zoom = FALSE;
                zoomicon = LoadBitmap(InstancePtr, MAKEINTRESOURCE(IDB_CANDY2 ));
            } else
            {   zoom = TRUE;
                zoomicon = LoadBitmap(InstancePtr, MAKEINTRESOURCE(IDB_CANDY2B));
            }
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_CANDY2), STM_SETIMAGE, IMAGE_BITMAP, (LPARAM) zoomicon);
        acase IDC_CANDY5:
            if (zoom)
            {   zoom = FALSE;
                zoomicon = LoadBitmap(InstancePtr, MAKEINTRESOURCE(IDB_CANDY5 ));
            } else
            {   zoom = TRUE;
                zoomicon = LoadBitmap(InstancePtr, MAKEINTRESOURCE(IDB_CANDY5B));
            }
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_CANDY5), STM_SETIMAGE, IMAGE_BITMAP, (LPARAM) zoomicon);
        }
        return TRUE;
    adefault:
        return FALSE;
    }

    return TRUE;
}

BOOL CALLBACK MemoryDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   TRANSIENT LONG   gid;
    PERSIST   HBRUSH hBrush[6];
    PERSIST   int    i; // PERSISTent for speed

    switch (Message)
    {
    case WM_INITDIALOG:
        charset = CHARSET_WIN;
        DISCARD SetWindowText( hwnd,              LLL, MSG_HAIL_MEMORY, "Memory Viewer"        ));

        if (language == LANGUAGE_RUS)
        {   DISCARD SendMessage(GetDlgItem(hwnd, IDL_REGION        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDC_LEGEND        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDL_LEGEND1       ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDL_LEGEND2       ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDL_LEGEND3       ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDL_LEGEND4       ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDL_LEGEND5       ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDL_LEGEND6       ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
        }
        DISCARD SetDlgItemText(hwnd, IDL_REGION,  LLL, MSG_REGION,      "Region: $"            ));
        DISCARD SetDlgItemText(hwnd, IDC_LEGEND,  LLL, MSG_SHOWLEGEND,  "Show legend?"         ));
        DISCARD SetDlgItemText(hwnd, IDL_LEGEND1, LLL, MSG_LEGEND1,     "Breakpoint/watchpoint"));
        DISCARD SetDlgItemText(hwnd, IDL_LEGEND2, LLL, MSG_LEGEND2,     "Mirror"               ));
        DISCARD SetDlgItemText(hwnd, IDL_LEGEND3, LLL, MSG_LEGEND3,     "Unmapped"             ));
        DISCARD SetDlgItemText(hwnd, IDL_LEGEND4, LLL, MSG_LEGEND4,     "Write-only"           ));
        DISCARD SetDlgItemText(hwnd, IDL_LEGEND5, LLL, MSG_LEGEND5,     "Read-only"            ));
        DISCARD SetDlgItemText(hwnd, IDL_LEGEND6, LLL, MSG_LEGEND6,     "Read/write"           ));
        charset = CHARSET_DOS;

        hBrush[0] = CreateSolidBrush(EMUPEN_RED);
        hBrush[1] = CreateSolidBrush(EMUPEN_BLUE);
        hBrush[2] = CreateSolidBrush(EMUPEN_GREEN);
        hBrush[3] = CreateSolidBrush(EMUPEN_YELLOW);
        hBrush[4] = CreateSolidBrush(EMUPEN_PURPLE);
        hBrush[5] = CreateSolidBrush(EMUPEN_ORANGE);
        
        if (showlegend)
        {   DISCARD SendMessage(GetDlgItem(hwnd, IDC_LEGEND), BM_SETCHECK, BST_CHECKED, 0);
        }

        return FALSE;
    case WM_CTLCOLORSTATIC: // no need for acase
        gid = GetWindowLong((HWND) lParam, GWL_ID);

        if (gid >= ID_MEM_0 && gid <= ID_MEM_127)
        {   if
            (   (memflags[gid - ID_MEM_0 + regionstart] & BREAKPOINT)
             || (memflags[gid - ID_MEM_0 + regionstart] & WATCHPOINT)
            )
            {   SetBkColor((HDC) wParam, EMUPEN_ORANGE); // orange (bp/wp)
                return (LRESULT) hBrush[5];
            } elif
            (   mirror_r[gid - ID_MEM_0 + regionstart] != gid - ID_MEM_0 + regionstart
             || mirror_w[gid - ID_MEM_0 + regionstart] != gid - ID_MEM_0 + regionstart
            )
            {   SetBkColor((HDC) wParam, EMUPEN_PURPLE); // purple (mirror)
                return (LRESULT) hBrush[4];
            } elif (memflags[gid - ID_MEM_0 + regionstart] & NOWRITE)
            {   if (memflags[gid - ID_MEM_0 + regionstart] & NOREAD)
                {   SetBkColor((HDC) wParam, EMUPEN_RED); // red (unmapped)
                    return (LRESULT) hBrush[0];
                } else
                {   SetBkColor((HDC) wParam, EMUPEN_BLUE); // blue (read-only)
                    return (LRESULT) hBrush[1];
            }   }
            elif (memflags[gid - ID_MEM_0 + regionstart] & NOREAD)
            {   SetBkColor((HDC) wParam, EMUPEN_YELLOW); // yellow (write-only)
                return (LRESULT) hBrush[3];
            } else
            {   SetBkColor((HDC) wParam, EMUPEN_GREEN); // green (read/write)
                return (LRESULT) hBrush[2];
        }   }
        elif (showlegend)
        {   switch (gid)
            {
            case IDL_MEM_1ST:
                SetBkColor((HDC) wParam, EMUPEN_ORANGE); // orange (bp/wp)
                return (LRESULT) hBrush[5];
            acase IDL_MEM_2ND:
                SetBkColor((HDC) wParam, EMUPEN_PURPLE); // purple (mirror)
                return (LRESULT) hBrush[4];
            acase IDL_MEM_3RD:
                SetBkColor((HDC) wParam, EMUPEN_RED);    // red (unmapped)
                return (LRESULT) hBrush[0];
            acase IDL_MEM_4TH:
                SetBkColor((HDC) wParam, EMUPEN_YELLOW); // yellow (write-only)
                return (LRESULT) hBrush[3];
            acase IDL_MEM_5TH:
                SetBkColor((HDC) wParam, EMUPEN_BLUE);   // blue (read-only)
                return (LRESULT) hBrush[1];
            acase IDL_MEM_6TH:
                SetBkColor((HDC) wParam, EMUPEN_GREEN);  // green (read/write)
                return (LRESULT) hBrush[2];
          }   }
    acase WM_ACTIVATE:
        do_autopause(wParam, lParam);
    acase WM_COMMAND:
        switch (LOWORD(wParam))
        {
        case IDC_LEGEND:
            if (SendMessage(GetDlgItem(hwnd, IDC_LEGEND), BM_GETCHECK, 0, 0) == BST_CHECKED)
            {   showlegend = TRUE;
            } else
            {   showlegend = FALSE;
            }
            DestroyWindow(hwnd);
            MemoryWindowPtr = NULL;
            view_memory();
        }
    acase WM_CLOSE:
        DestroyWindow(hwnd);
        MemoryWindowPtr = NULL;
        updatemenu(MENUITEM_VIEWMEMORY);
    acase WM_DESTROY:
        for (i = 0; i < 6; i++)
        {   DeleteObject(hBrush[i]);
        }
    adefault:
        return FALSE;
    }
    return TRUE;
}

BOOL CALLBACK ControlsDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   PAINTSTRUCT ps;

    switch (Message)
    {
    case WM_INITDIALOG:
        charset = CHARSET_WIN;
        DISCARD SetWindowText(hwnd, LLL, MSG_HAIL_CONTROLS, "Controls"));
        charset = CHARSET_DOS;

        // this is so that paddle axes which are never read by the
        // game are reported to be centred.
        sx[0] =
        sy[0] =
        sx[1] =
        sy[1] = 128;
    acase WM_ACTIVATE:
        do_autopause(wParam, lParam);
    acase WM_LBUTTONDBLCLK:
    case WM_LBUTTONDOWN:
        mousedown((int) (LOWORD(lParam)), (int) (HIWORD(lParam)));
    acase WM_LBUTTONUP:
        mouseup();
    acase WM_CLOSE:
        DestroyWindow(hwnd);
        ControlsWindowPtr = NULL;
        updatemenu(MENUITEM_CONTROLS);
    acase WM_PAINT:
        if (machine == INSTRUCTOR)
        {   BeginPaint(hwnd, &ps);
            updatescreen();
            DISCARD EndPaint(hwnd, &ps);
        }
        return FALSE; // important!
    default: // no need for adefault
        return FALSE;
    }
    return TRUE;
}

EXPORT ULONG getsize(const STRPTR filename)
{   HANDLE hFile /* = NULL */ ;
    ULONG  localsize;

    hFile = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
    if (hFile == INVALID_HANDLE_VALUE)
    {   return 0;
    }
    localsize = GetFileSize(hFile, NULL);
    DISCARD CloseHandle(hFile);
    // hFile = NULL;

    if (localsize == (ULONG) -1)
    {   return 0;
    }

    return localsize;
}

EXPORT void checkinput(void)
{   TRANSIENT MSG   Msg;
    TRANSIENT ULONG scancode;
    PERSIST   FLAG  fresh[2] = { TRUE, TRUE };

    Sleep(0);
    /* We need to call this regularly even while paused, to check whether the
       user is trying to eg. unpause via host gamepad.
    if (paused && !cli)
    {   DISCARD WaitMessage(); // to give better multitasking
    } */

    while (PeekMessage(&Msg, NULL, 0, 0, PM_REMOVE))
    {   switch (Msg.message)
        {
        case WM_KEYDOWN:
            if (cli && paused)
            {   break;
            }
            if (machine == PIPBUG)
            {   psu &= ~(PSU_S);
            } // if it's a key for the emulator GUI, it would be better if it didn't affect the state of the Sense bit
            scancode = (Msg.lParam & 33488896) >> 16;
            if (scancode == SCAN_CAPSLOCK)
            {   if (capslock)
                {   capslock = FALSE;
                } else
                {   capslock = TRUE;
            }   }
            handlekybd(scancode);
        acase WM_SYSKEYDOWN:
            if (cli && paused)
            {   break;
            }
            scancode = (Msg.lParam & 33488896) >> 16;
            handlealtkybd(scancode);
            TranslateMessage(&Msg);
            DISCARD DispatchMessage(&Msg);
        acase WM_KEYUP:
            if (cli && paused)
            {   break;
            }
            if (machine == PIPBUG)
            {   psu |= PSU_S;
            }
            scancode = (Msg.lParam & 33488896) >> 16;
            KeyMatrix[scancode / 8] &= (255 - (1 << (scancode % 8)));
            keyframes[scancode] = 0;
        acase WM_RBUTTONUP:
            SetCursor(hArrow);
            if (!showingpointer)
            {   showingpointer = TRUE;
                ShowCursor(TRUE);
            }
            DISCARD ClipCursor(NULL);

            ThePoint.x = LOWORD(Msg.lParam);
            ThePoint.y = HIWORD(Msg.lParam);
            DISCARD ClientToScreen(MainWindowPtr, &ThePoint);
            TrackPopupMenu
            (   PopUpMenuPtr,
                TPM_RIGHTBUTTON,
                ThePoint.x,
                ThePoint.y,
                0,
                MainWindowPtr,
                NULL
            );

            updatepointer(TRUE);

            // TranslateMessage(&Msg);
            // DispatchMessage(&Msg);
        adefault:
            TranslateMessage(&Msg);
            DISCARD DispatchMessage(&Msg);
    }   }

    ReadJoystick(0);
    ReadJoystick(1);

    if (!crippled && (jf[0] & JOYRESET) || (jf[1] & JOYRESET))
    {   project_reset(FALSE);
    }

    // we should probably keep track of fresh[] separately for each player
    if ((jf[0] & JOYPAUSE) || (jf[1] & JOYPAUSE))
    {   if (fresh[0])
        {   fresh[0] = FALSE;
            if (paused)
            {   emu_unpause();
            } else
            {   emu_pause(TRUE);
    }   }   }
    else
    {   fresh[0] = TRUE;
    }
    if ((jf[0] & JOYAUTOFIRE) || (jf[1] & JOYAUTOFIRE))
    {   if (fresh[1])
        {   fresh[1] = FALSE;
            if (autofire) autofire = FALSE; else autofire = TRUE; // foo = ~foo; doesn't always work!
            docommand(MENUITEM_AUTOFIRE);
    }   }
    else
    {   fresh[1] = TRUE;
    }

    return; // return Msg.wParam;
}

EXPORT UBYTE KeyDown(UWORD scancode) // scancode must be at least 16 bits wide!
{   if (KeyMatrix[scancode / 8] & (1 << (scancode % 8)))
    {   return(1);
    }
    return(0);
}

EXPORT void debug_set(int kind)
{   sound_off(FALSE);
    dlgtype = kind;
    opening = TRUE;
    switch (kind)
    {
    case KIND_BP:
        DISCARD DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_SETBP),   MainWindowPtr, SetDlgProc);
    acase KIND_WP:
        DISCARD DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_SETWP),   MainWindowPtr, SetDlgProc);
    acase KIND_BC:
        DISCARD DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_CLEARBP), MainWindowPtr, SetDlgProc);
    acase KIND_WC:
        DISCARD DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_CLEARWP), MainWindowPtr, SetDlgProc);
    }
    opening = FALSE;
    updatemenus();
    sound_on();
}
EXPORT void debug_setrastn(void)
{   sound_off(FALSE);
    opening = TRUE;
    DISCARD DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_RUNTORASTN), MainWindowPtr, RastNDlgProc);
    opening = FALSE;
    sound_on();
}

EXPORT void redrawscreen(void)
{   PERSIST int x, y; // PERSISTent for speed

    if (machine == INSTRUCTOR)
    {   for (y = 0; y < MINIHEIGHT; y++)
        {   for (x = 0; x < MINIWIDTH; x++)
            {    ctrldisplay[x + (y * MINIWIDTH)] = pens[colourset][ctrlscreen[x][y]];
    }   }   }

    for (y = 0; y < machines[machine].height; y++)
    {   for (x = 0; x < machines[machine].width; x++)
        {   drawpixel(x, y, screen[x][y]);
    }   }

    updatescreen();
}

EXPORT void updatescreen(void)
{   PERSIST HDC     OnScreenRastPort;
#ifdef NEWWAY
    PERSIST HDC     hdcMem;
    PERSIST HBITMAP hbmOld;
#endif

    if (!MainWindowPtr)
    {   return; // important!
    }

    // caching the DC would be faster but we are not supposed to do that
    OnScreenRastPort = GetDC(MainWindowPtr);

    if (stretch && fullscreen)
    {   DISCARD StretchDIBits
        (   OnScreenRastPort,                              // dest rastport
            xoffset,                                       // dest leftx
            yoffset,                                       // dest topy
            destwidth,                                     // dest width
            destheight,                                    // dest height
            0,                                             // source leftx
            0,                                             // source topy
            sourcewidth,                                   // source width
            sourceheight,                                  // source height
            display,                                       // pointer to the bits
            (const struct tagBITMAPINFO *) &TheBitMapInfo, // pointer to BITMAPINFO structure
            DIB_RGB_COLORS,                                // format of data
            SRCCOPY                                        // blit mode
        );
    } else
    {   if (scale2x)
        {   if (realsize == 2)
            {   apply_scale2x();
            } elif (realsize == 3)
            {   apply_scale3x();
            } elif (realsize == 4)
            {   apply_scale4x();
        }   }
        apply_rastlines();

#ifdef OLDWAY
        DISCARD StretchDIBits
        (   OnScreenRastPort,
            xoffset,        // dest leftx
            yoffset,        // dest topy
            rastwidth,      // dest width
            rastheight,     // dest height
            0,              // source leftx
            0,              // source topy
            rastwidth,      // source width
            rastheight,     // source height
            display,        // pointer to the bits
            (const struct tagBITMAPINFO *) &TheBitMapInfo, // pointer to BITMAPINFO structure
            DIB_RGB_COLORS, // format of data
            SRCCOPY         // blit mode
        );
#endif
#ifdef NEWWAY
        hdcMem = CreateCompatibleDC(OnScreenRastPort);
        hbmOld = (HBITMAP) SelectObject(hdcMem, OurhBitMap);
        BitBlt
        (   OnScreenRastPort,
            xoffset,
            yoffset,
            rastwidth,
            rastheight,
            hdcMem,
            0, 0,
            SRCCOPY
        );
        SelectObject(hdcMem, hbmOld);
#endif
    }

    DISCARD ReleaseDC(MainWindowPtr, OnScreenRastPort);

    if (machine == INSTRUCTOR && ControlsWindowPtr)
    {   OnScreenRastPort = GetDC(ControlsWindowPtr);
        DISCARD StretchDIBits   
        (   OnScreenRastPort,
            248,            // dest leftx
             87,            // dest topy
            MINIWIDTH,      // dest width
            MINIHEIGHT,     // dest height
              0,            // source leftx             
              0,            // source topy
            MINIWIDTH,      // source width
            MINIHEIGHT,     // source height
            ctrldisplay,    // pointer to the bits
            (const struct tagBITMAPINFO *) &CtrlBitMapInfo, // pointer to BITMAPINFO structure
            DIB_RGB_COLORS, // format of data
            SRCCOPY         // blit mode
        );
        DISCARD ReleaseDC(ControlsWindowPtr, OnScreenRastPort);
    }

    if (recmode == RECMODE_RECORD && anims)
    {   if (animframe != frames)
        {   AddAnim();
            animframe = frames;
}   }   }

EXPORT void freeall(void)
{   int  i;
    BOOL pvParam;

    opening = TRUE;
// T1; zprintf("1...\n"); T2;
    macro_stop();

    // should we call close_monitor(), close_memory(), etc.?
    // we don't, currently.

// T1; zprintf("2...\n"); T2;
    sound_off(FALSE);
// T1; zprintf("3...\n"); T2;
    ff_die();
// T1; zprintf("4...\n"); T2;
    for (i = 0; i < 14; i++)
    {   SoundClose(i);
    }
// T1; zprintf("5...\n"); T2;

#ifdef USEDIRECTSOUND
    for (i = 0; i <= 6; i++)
    {   if (inst[i])
        {   inst[i]->lpVtbl->Release(inst[i]);
            // inst[i] = NULL;
    }   }
// T1; zprintf("6...\n"); T2;
    if (lpds)
    {   lpds->lpVtbl->Release(lpds);
        // lpds = NULL;
    }
#endif
// T1; zprintf("7...\n"); T2;

    NetPlay(NET_OFF);
// T1; zprintf("8...\n"); T2;

    DISCARD ClipCursor(NULL);
// T1; zprintf("9...\n"); T2;

    if (ProgDir[0])
    {   DISCARD SetCurrentDirectory(ProgDir);
    }
// T1; zprintf("10...\n"); T2;

    engine_quit();
// T1; zprintf("11...\n"); T2;

    /* This code causes crashes when exiting under Windows Vista and
       Windows 7.
    if (display)
    {   // "If I open the console for logging and I press the Esc key,
        // the program fails [here] and shows a Windows requester."
        // - Damaso Dominguez
        free(display);
        free(ctrldisplay);
    } */

    if (OurhBitMap)
    {   DeleteObject(OurhBitMap);
        OurhBitMap = NULL;
    }
// T1; zprintf("14...\n"); T2;
    if (OurhDC)
    {   DeleteDC(OurhDC);
        OurhDC = NULL;
    }
// T1; zprintf("15...\n"); T2;

    for (i = 0; i < SAMPLES; i++)
    {   stop_sample(i);
        if (SampleOpened[i])
        {   DISCARD waveOutClose(hSample[i]);
            // SampleOpened[i] = FALSE;
        }
        if (SampleBuffer[i])
        {   free(SampleBuffer[i]);
            // SampleBuffer[i] = NULL;
    }   }
// T1; zprintf("16...\n"); T2;

    for (i = 0; i < MENUBITMAPS; i++)
    {   if (mib[i].bitmap)
        {   DeleteObject(mib[i].bitmap);
            mib[i].bitmap = NULL;
    }   }
// T1; zprintf("17...\n"); T2;

    if (!upgrade)
    {   saveconfig();
// T1; zprintf("18...\n"); T2;
        savehiscores();
    }
// T1; zprintf("19...\n"); T2;

    echoprinter = FALSE; // must save config first
    updateprinter();
// T1; zprintf("20...\n"); T2;

    if (stars)
    {   free(stars);
// T1; zprintf("21...\n"); T2;
        stars = NULL;
    }
// T1; zprintf("22...\n"); T2;

    uniconify(TRUE);
// T1; zprintf("23...\n"); T2;

    freelanguage();
// T1; zprintf("24...\n"); T2;

    speech_die();
// T1; zprintf("25...\n"); T2;

    // Do we need to free listview nodes? It seems not.

    if (lame)
    {   SystemParametersInfo(SPI_GETSCREENSAVEACTIVE, 0, &pvParam, 0);
        if (pvParam == FALSE)
        {   SystemParametersInfo(SPI_SETSCREENSAVEACTIVE, TRUE, 0, SPIF_SENDWININICHANGE);
    }   }
// T1; zprintf("26...\n"); T2;

#ifdef GUI
    // we do this last so we can continue to zprintf() during exit
    if (consoleopen)
    {   FreeConsole();
        ConsoleWindowPtr = NULL;
    }
#endif
}

MODULE void upgradeconfigs(void)
{   WIN32_FIND_DATA FindData;
    HANDLE          FindHandle;
    int             thelength;

    DISCARD SetCurrentDirectory(".\\Configs");

    // T1;

    FindHandle = FindFirstFile("*", &FindData);
    if (FindHandle == INVALID_HANDLE_VALUE)
    {   DISCARD SetCurrentDirectory(olddir);
        // printf("No files found!\n"); // not zprintf()!
        // T2;
        cleanexit(EXIT_FAILURE);
    }

    do
    {   if (FindData.dwFileAttributes != FILE_ATTRIBUTE_DIRECTORY)
        {   thelength = (int) strlen(FindData.cFileName);
            if
            (   thelength > 4
             && !stricmp(&FindData.cFileName[thelength - 4], ".CFG")
            )
            {   strcpy(fn_config, FindData.cFileName);

                // printf("Upgrading %s...", FindData.cFileName); // not zprintf()!
                loadconfig();
                saveconfig();
                // printf("done.\n"); // not zprintf()!
            } else
            {   ; // printf("Ignoring %s.\n", FindData.cFileName); // not zprintf()!
    }   }   }
    while (FindNextFile(FindHandle, &FindData));
    DISCARD FindClose(FindHandle);

    // zprintf("All done.\n");
    // T2;

 /* MessageBox
    (   NULL,
        "All done.",
        "WinArcadia: Message",
        MB_ICONINFORMATION | MB_OK
    ); */
    cleanexit(EXIT_SUCCESS);
}

EXPORT void cleanexit(SBYTE rc)
{   if (consoleopen)
    {   textpen(TEXTPEN_DEFAULT);
        zprintf("%s.\n", LLL, MSG_QUITTING, "Quitting"));
    }

    freeall();
// T1; zprintf("25...\n"); T2;

    // PostMessage(MainWindowPtr, WM_CLOSE, 0, 0);
    DestroyWindow(MainWindowPtr); // this will also destroy all child windows
    // MainWindowPtr = NULL;
// T1; zprintf("26...\n"); T2;

    exit(rc);
}

EXPORT void mm_ReadJoystick(UWORD joynum)
{   JOYINFOEX joyInfoEx;
    HWND      activewindow;

    // joynum will be '0' if they want the PRIMARY port (the FIRST detected joystick)
    // joynum will be '1' if they want the SECONDARY port (the SECOND detected joystick)

    if (unit[joynum] == -1)
    {   jx[joynum] =
        jy[joynum] = 128;
        jf[joynum] = 0;
        return;
    }

    activewindow = GetActiveWindow();
    if
    (   activewindow != MainWindowPtr
     && (!ControlsWindowPtr || activewindow != ControlsWindowPtr)
     && (    !DIPsWindowPtr || activewindow !=     DIPsWindowPtr)
     && (!GameInfoWindowPtr || activewindow != GameInfoWindowPtr)
     && (  !MemoryWindowPtr || activewindow !=   MemoryWindowPtr)
     && ( !MonitorWindowPtr || activewindow !=  MonitorWindowPtr)
     && ( !PaletteWindowPtr || activewindow !=  PaletteWindowPtr)
     && (  !SpriteWindowPtr || activewindow !=   SpriteWindowPtr)
    )
    {   jx[joynum] =
        jy[joynum] = 128;
        jf[joynum] = 0;
        return;
    }

    ZeroMemory(&joyInfoEx, sizeof(joyInfoEx));
    joyInfoEx.dwSize  = sizeof(joyInfoEx);
    joyInfoEx.dwFlags = JOY_RETURNALL;

    // the upper byte of dwRpos is the vertical axis of the right analog knob (0..255).

    if (joyGetPosEx((UINT) unit[joynum], &joyInfoEx) == JOYERR_NOERROR)
    {   jx[joynum] = (UBYTE) (joyInfoEx.dwXpos >> 8);
        jy[joynum] = (UBYTE) (joyInfoEx.dwYpos >> 8);

        if (machine == ARCADIA && jx[joynum] >= 0x7E && jx[joynum] <= 0x81) // && game == ALIENINV
        {   jx[joynum] = 0x7D;
        }

        jf[joynum] = 0;
        if (joyInfoEx.dwButtons &  1             ) jf[joynum] |= joyfires[button[joynum][0] - 1];
        if (joyInfoEx.dwButtons & (1 << ( 2 - 1))) jf[joynum] |= joyfires[button[joynum][1] - 1];
        if (joyInfoEx.dwButtons & (1 << ( 3 - 1))) jf[joynum] |= joyfires[button[joynum][2] - 1];
        if (joyInfoEx.dwButtons & (1 << ( 4 - 1))) jf[joynum] |= joyfires[button[joynum][3] - 1];
        if (joyInfoEx.dwButtons & (1 << ( 5 - 1))) jf[joynum] |= joyfires[button[joynum][4] - 1];
        if (joyInfoEx.dwButtons & (1 << ( 6 - 1))) jf[joynum] |= joyfires[button[joynum][5] - 1];
        if (joyInfoEx.dwButtons & (1 << ( 7 - 1))) jf[joynum] |= joyfires[button[joynum][6] - 1];
        if (joyInfoEx.dwButtons & (1 << ( 8 - 1))) jf[joynum] |= joyfires[button[joynum][7] - 1];
        if (joyInfoEx.dwButtons & (1 << ( 9 - 1))) jf[joynum] |= JOYRESET;
        if (joyInfoEx.dwButtons & (1 << (10 - 1))) jf[joynum] |= JOYSTART;
        if (joyInfoEx.dwButtons & (1 << (11 - 1))) jf[joynum] |= JOYFIRE1;
        if (joyInfoEx.dwButtons & (1 << (12 - 1))) jf[joynum] |= JOYFIRE1;
}   }

EXPORT void sound_off(FLAG force)
{   int i;

    if (!force)
    {   ff_off(0);
        ff_off(1);
        if (!sound)
        {   return; // important!
    }   }

    for (i = 0; i < 14; i++)
    {   SoundStop(i);
}   }

EXPORT void sound_on(void)
{   if (!sound)
    {   return; // important!
    }

    /* if (!waveOutGetNumDevs()) // if no sound output devices found
    {   MessageBox(MainWindowPtr, "waveOutGetNumDevs() failed!", "WinArcadia: Error", MB_OK | MB_ICONERROR);
    } */

    if (!paused)
    {   playsound(TRUE);
}   }

EXPORT void calcsize(void)
{   int realsizex,
        realsizey;

/* xoffset      = distance from left edge of window to left edge of guest screen
   yoffset      = distance from top  edge of window to top  edge of guest screen
   winwidth     = overall window width
   winheight    = overall window height
   clientwidth  = width  of window excluding eg. gadgets (only used in full-screen mode)
   clientheight = height of window excluding eg. gadgets (only used in full-screen mode)
   rastwidth    = width  of guest screen
   rastheight   = height of guest screen
   winleftx     = screen coordinate of left edge of window
   wintopy      = screen coordinate of top  edge of window

   All quantities are in host pixels.
*/

    if (fullscreen)
    {   winwidth  = clientwidth  = DisplayWidth;
        winheight = clientheight = DisplayHeight;

        if (showtitlebar)
        {   clientheight -= titleheight;
        }
        if (showmenubar)
        {   clientheight -= menuheight;
        }
        if (showtoolbar)
        {   clientheight -= toolbarheight;
        }
        if (showsidebar)
        {   sidebarwidth = SIDEBARWIDTH; // in case the user has made the sidebar very wide
            // Ideally, we would remember the old value, and restore it when going back to windowed mode
            clientwidth  -= sidebarwidth + 4;
            if (!showtitlebar)
            {   sidebarwidth += 4;
        }   }

        realwide  = machines[machine].wide;
        clientwidth -= 2; // so that sidebar divider pane is visible

        realsizex = clientwidth  / (machines[machine].width * realwide);
        if (realsizex > 4)
        {   realsizex = 4; // because fullscreen unstretched si50 could be 5x or more graphics scaling!
        }
        realsizey = clientheight /  machines[machine].height;
        if (realsizey > 4)
        {   realsizey = 4;
        }
        if (realsizey < realsizex)
        {   realsize = realsizey;
        } else
        {   realsize = realsizex;
        }
        fastwide     =          realsize                 * realwide;
        rastwidth    = (SWORD) (machines[machine].width  * fastwide);
        rastheight   = (SWORD) (machines[machine].height * realsize);

        if (stretch)
        {   xoffset = 0;
        } else
        {   xoffset = (clientwidth - rastwidth ) / 2;
    }   }
    else
    {   winwidth  = rastwidth  = machines[machine].width  * size * wide;
        winheight = rastheight = machines[machine].height * size;

        xoffset   = 0;
        winwidth  += 8; // was 2 until V14.2
        winheight += 8; // was 6 until V12.0
        if (showtitlebar)
        {   winheight += titleheight;
        }
        if (showsidebar)
        {   winwidth += sidebarwidth;
        }
        if (showmenubar)
        {   winheight += menuheight;
            if (winwidth < langs[language].minwidth1)
            {   winheight += menuheight; // because menu will be 2 rows in height
                if (winwidth < langs[language].minwidth2)
                {   winheight += menuheight; // because menu will be 3 rows in height
                    if (winwidth < langs[language].minwidth3)
                    {   winheight += menuheight; // because menu will be 4 (or more) rows in height
        }   }   }   }
        if (showtoolbar)
        {   winheight += toolbarheight;
        }
        clientheight = winheight;
        if (showtitlebar)
        {   clientheight -= titleheight;
        }

        realwide = wide;
        realsize = size;
    }

    /* SM_CXSIZEFRAME, SM_CYSIZEFRAME are normally 4.
       SM_CXBORDER,    SM_CXBORDER    are normally 1. */

    if (showtoolbar)
    {   yoffset = toolbarheight;
    } else
    {   yoffset = 0;
    }
    sidebartopy = yoffset;
    if (fullscreen && !stretch)
    {   xoffset += GetSystemMetrics(SM_CXSIZEFRAME);
        yoffset += (clientheight - rastheight) / 2;
    }

    if (winleftx == -1)
    {   winleftx = (DisplayWidth  / 2) - (winwidth  / 2);
    }
    if (wintopy == -1)
    {   wintopy  = (DisplayHeight / 2) - (winheight / 2);
    }

    if (winleftx < 0 || winleftx >= DisplayWidth)
    {   winleftx = 0;
    }
    if (wintopy < 0 || wintopy >= DisplayHeight)
    {   wintopy = 0;
    }

    minwinwidth = winwidth - sidebarwidth + SIDEBARWIDTH;

    if
    (    (fullscreen    && stretch      )
     ||  (realsize == 1 && realwide == 1)
    )                                     fastsize = 0;
    elif (realsize == 1 && realwide == 2) fastsize = 1;
    elif (realsize == 2 && realwide == 1) fastsize = 2;
    elif (realsize == 2 && realwide == 2) fastsize = 3;
    elif (realsize == 3 && realwide == 1) fastsize = 4;
    elif (realsize == 3 && realwide == 2) fastsize = 5;
    elif (realsize == 4 && realwide == 1) fastsize = 6;
    elif (realsize == 4 && realwide == 2) fastsize = 7;
    // else assert(0);

    if (fullscreen)
    {   sidebarheight = clientheight;
        if (showtitlebar)
        {   sidebarheight -= 4; // presumably this depends on frame thickness
        }
        if (!showtitlebar)
        {   clientwidth += 2;
    }   }
    else
    {   sidebarheight = rastheight - 1;
    }

    if (fullscreen && stretch)
    {   sourcewidth  = machines[machine].width;
        sourceheight = machines[machine].height;
        destwidth    = clientwidth;
        destheight   = clientheight;
        if (showtitlebar)
        {   destheight -= 6;
    }   }
    else
    {   sourcewidth  =
        destwidth    = rastwidth;
        sourceheight =
        destheight   = rastheight;
    }

    fastwide = realsize * realwide;

#ifdef VERBOSE
    T1; zprintf("Position is %d,%d. Size is %d,%d.\n", winleftx, wintopy, winwidth, winheight); T2;
#endif

#ifdef OLDWAY
    makedisplay();
#endif
}

EXPORT void resize(int newsize)
{   int oldwidth  = winwidth,
        oldheight = winheight;

    size = realsize = newsize;
    updatemenu(MENUITEM_SIZE);
    calcsize();
#ifdef NEWWAY
    makedisplay();
#endif

    if (winwidth == oldwidth && winheight == oldheight)
    {   return;
    }

    ready = FALSE;

    DISCARD SetWindowPos
    (   MainWindowPtr,
        HWND_TOP,
        winleftx,
        wintopy,
        winwidth,
        winheight,
        SWP_NOCOPYBITS
    );
    maketoolbar();
    if (showsidebar)
    {   if (fullscreen)
        {   DISCARD SetWindowPos
            (   hSideBar,
                HWND_TOP,
                showtitlebar ? (winwidth - sidebarwidth - 4) : (winwidth - sidebarwidth + 2),
                sidebartopy,
                sidebarwidth,
                sidebarheight,
                SWP_NOCOPYBITS
            );
        } else
        {   DISCARD SetWindowPos
            (   hSideBar,
                HWND_TOP,
                winwidth - sidebarwidth - 6,
                sidebartopy,
                sidebarwidth,
                sidebarheight,
                SWP_NOCOPYBITS
            );
        }
        DISCARD ListView_EnsureVisible(hSideBar, currentgame, TRUE); // doesn't seem to work properly for items in the rightmost column
    }

    redrawscreen();
    updatepointer(TRUE);

    ready = TRUE;
}

EXPORT void beep(void)      { ; }
EXPORT void readkybd(void)  { ; }

MODULE void maketoolbar(void)
{   TRANSIENT TBBUTTON tbb[BIGGADGETS + 11 + 5];
    TRANSIENT int      next,
                       sparewidth,
                       widegadgets;

    // We can't use MainWindowPtr, as it is not set up yet!
    // (This function gets called when the window is 'half-open'.)

    if (hToolbar)
    {   DestroyWindow(hToolbar);
        hToolbar = NULL;
    }
    if (LabelPtr)
    {   DestroyWindow(LabelPtr);
        LabelPtr = NULL;
    }
    if (FPSPtr)
    {   DestroyWindow(FPSPtr);
        FPSPtr = NULL;
    }

    if (!showtoolbar)
    {   return;
    }

    ZeroMemory(tbb, sizeof(tbb));

/* Left window border and margin is 2.
   Each bland button is 25x24.
   Gap (separator) is 8.
   Each colourful button is 25x24.
   Any spare room is here.
   An aesthetic gap of 8 here.
   FPS label is 36.
   FPS gadget is 52.
   Right window border and margin is 10. */
   
    sparewidth = winwidth - 22 - 52; // 22 is for margins/gap, 52 is width of FPS gadget

    if (sparewidth >= (9 * 25) + (1 * 8))
    {   sparewidth  -= (9 * 25) + (1 * 8);
        widegadgets =  2;
    } elif (sparewidth >= (7 * 25) + (1 * 8))
    {   sparewidth  -= (7 * 25) + (1 * 8);
        widegadgets =  1;
    } else
    {   widegadgets =  0;
    }

    if (sparewidth >= (7 * 25) + (1 * 8))
    {   tbb[0].iBitmap   = 0;
        tbb[0].fsState   = 0;
        tbb[0].fsStyle   = TBSTYLE_BUTTON;
        tbb[0].idCommand = ID_FILE_RESET;
        tbb[1].iBitmap   = 1;
        tbb[1].fsState   = TBSTATE_ENABLED;
        tbb[1].fsStyle   = TBSTYLE_BUTTON;
        tbb[1].idCommand = ID_FILE_OPEN;
        tbb[2].iBitmap   = 2;
        tbb[2].fsState   = 0;
        tbb[2].fsStyle   = TBSTYLE_BUTTON;
        tbb[2].idCommand = ID_FILE_SAVESNP;
        tbb[3].iBitmap   = 3;
        tbb[3].fsState   = TBSTATE_ENABLED;
        tbb[3].fsStyle   = TBSTYLE_BUTTON;
        tbb[3].idCommand = ID_FILE_SAVEILBM;
        tbb[4].iBitmap   = 4;
        tbb[4].fsState   = TBSTATE_ENABLED;
        tbb[4].fsStyle   = TBSTYLE_BUTTON;
        tbb[4].idCommand = ID_EDIT_COPY;
        tbb[5].iBitmap   = 5;
        tbb[5].fsState   = TBSTATE_ENABLED;
        tbb[5].fsStyle   = TBSTYLE_BUTTON;
        tbb[5].idCommand = ID_FILE_QUICKLOAD;
        tbb[6].iBitmap   = 6;
        tbb[6].fsState   = TBSTATE_ENABLED;
        tbb[6].fsStyle   = TBSTYLE_BUTTON;
        tbb[6].idCommand = ID_FILE_QUICKSAVE;

        tbb[7].iBitmap   = 0;
        tbb[7].fsState   = 0;
        tbb[7].fsStyle   = TBSTYLE_SEP;
        tbb[7].idCommand = 0;

        next             = 8;
        sparewidth       -= (7 * 25) + (1 * 8);
    } elif (sparewidth >= (3 * 25) + (1 * 8))
    {   tbb[0].iBitmap   = 0;
        tbb[0].fsState   = 0;
        tbb[0].fsStyle   = TBSTYLE_BUTTON;
        tbb[0].idCommand = ID_FILE_RESET;
        tbb[1].iBitmap   = 1;
        tbb[1].fsState   = TBSTATE_ENABLED;
        tbb[1].fsStyle   = TBSTYLE_BUTTON;
        tbb[1].idCommand = ID_FILE_OPEN;
        tbb[2].iBitmap   = 2;
        tbb[2].fsState   = 0;
        tbb[2].fsStyle   = TBSTYLE_BUTTON;
        tbb[2].idCommand = ID_FILE_SAVESNP;

        tbb[3].iBitmap   = 0;
        tbb[3].fsState   = 0;
        tbb[3].fsStyle   = TBSTYLE_SEP;
        tbb[3].idCommand = 0;

        next             = 4;
        sparewidth       -= (3 * 25) + (1 * 8);
    } else
    {   next             = 0;
    }

    if (widegadgets == 2)
    {   tbb[next    ].iBitmap   = gadgetimage[GADPOS_FLAGLINE][0];
        tbb[next    ].fsStyle   = TBSTYLE_CHECK;
        tbb[next    ].idCommand = ID_GRAPHICS_FLAGLINE;
        tbb[next    ].fsState   = TBSTATE_ENABLED;  
        tbb[next + 1].iBitmap   = gadgetimage[GADPOS_AUTOFIRE][0];
        tbb[next + 1].fsState   = TBSTATE_ENABLED;
        tbb[next + 1].fsStyle   = TBSTYLE_CHECK;
        tbb[next + 1].idCommand = ID_CONTROLLERS_AUTOFIRE;
        tbb[next + 2].iBitmap   = gadgetimage[GADPOS_SWAPPED][0];
        tbb[next + 2].fsState   = TBSTATE_ENABLED;
        tbb[next + 2].fsStyle   = TBSTYLE_CHECK;
        tbb[next + 2].idCommand = ID_CONTROLLERS_SWAPPED;
        tbb[next + 3].iBitmap   = gadgetimage[GADPOS_PAUSED][0];
        tbb[next + 3].fsState   = TBSTATE_ENABLED;
        tbb[next + 3].fsStyle   = TBSTYLE_CHECK;
        tbb[next + 3].idCommand = ID_SPEED_PAUSED;
        tbb[next + 4].iBitmap   = gadgetimage[GADPOS_TURBO][0];
        tbb[next + 4].fsState   = TBSTATE_ENABLED;
        tbb[next + 4].fsStyle   = TBSTYLE_CHECK;
        tbb[next + 4].idCommand = ID_SPEED_TURBO;
        tbb[next + 5].iBitmap   = gadgetimage[GADPOS_COLLISIONS][0];
        tbb[next + 5].fsState   = TBSTATE_ENABLED;
        tbb[next + 5].fsStyle   = TBSTYLE_CHECK;
        tbb[next + 5].idCommand = ID_SPRITES_COLLISIONS;
        tbb[next + 6].iBitmap   = gadgetimage[GADPOS_SOUND][0];
        tbb[next + 6].fsState   = TBSTATE_ENABLED;
        tbb[next + 6].fsStyle   = TBSTYLE_CHECK;
        tbb[next + 6].idCommand = ID_SOUND_ENABLED;

        tbb[next + 7].iBitmap   = 0;
        tbb[next + 7].fsState   = 0;
        tbb[next + 7].fsStyle   = TBSTYLE_SEP;
        tbb[next + 7].idCommand = 0;

        tbb[next + 8].iBitmap    = BIGGADGETS + 7;
        tbb[next + 8].fsState    = 0;
        tbb[next + 8].fsStyle    = TBSTYLE_CHECK;
        tbb[next + 8].idCommand  = ID_MACRO_TOGGLEPLAYING;
        tbb[next + 9].iBitmap   = BIGGADGETS + 8;
        tbb[next + 9].fsState   = 0;
        tbb[next + 9].fsStyle   = TBSTYLE_CHECK;
        tbb[next + 9].idCommand = ID_MACRO_TOGGLERECORDING;

        next += 10;
    } elif (widegadgets == 1)
    {   tbb[next    ].iBitmap   = gadgetimage[GADPOS_AUTOFIRE][0];
        tbb[next    ].fsState   = TBSTATE_ENABLED;
        tbb[next    ].fsStyle   = TBSTYLE_CHECK;
        tbb[next    ].idCommand = ID_CONTROLLERS_AUTOFIRE;
        tbb[next + 1].iBitmap   = gadgetimage[GADPOS_PAUSED][0];
        tbb[next + 1].fsState   = TBSTATE_ENABLED;
        tbb[next + 1].fsStyle   = TBSTYLE_CHECK;
        tbb[next + 1].idCommand = ID_SPEED_PAUSED;
        tbb[next + 2].iBitmap   = gadgetimage[GADPOS_TURBO][0];
        tbb[next + 2].fsState   = TBSTATE_ENABLED;
        tbb[next + 2].fsStyle   = TBSTYLE_CHECK;
        tbb[next + 2].idCommand = ID_SPEED_TURBO;
        tbb[next + 3].iBitmap   = gadgetimage[GADPOS_COLLISIONS][0];
        tbb[next + 3].fsState   = TBSTATE_ENABLED;
        tbb[next + 3].fsStyle   = TBSTYLE_CHECK;
        tbb[next + 3].idCommand = ID_SPRITES_COLLISIONS;
        tbb[next + 4].iBitmap   = gadgetimage[GADPOS_SOUND][0];
        tbb[next + 4].fsState   = TBSTATE_ENABLED;
        tbb[next + 4].fsStyle   = TBSTYLE_CHECK;
        tbb[next + 4].idCommand = ID_SOUND_ENABLED;

        tbb[next + 5].iBitmap   = 0;
        tbb[next + 5].fsState   = 0;
        tbb[next + 5].fsStyle   = TBSTYLE_SEP;
        tbb[next + 5].idCommand = 0;

        tbb[next + 6].iBitmap   = BIGGADGETS + 7;
        tbb[next + 6].fsState   = 0;
        tbb[next + 6].fsStyle   = TBSTYLE_CHECK;
        tbb[next + 6].idCommand = ID_MACRO_TOGGLEPLAYING;
        tbb[next + 7].iBitmap   = BIGGADGETS + 8;
        tbb[next + 7].fsState   = 0;
        tbb[next + 7].fsStyle   = TBSTYLE_CHECK;
        tbb[next + 7].idCommand = ID_MACRO_TOGGLERECORDING;

        next += 8;
    }

    if (sparewidth >= (4 * 25) + (1 * 8))
    {   sparewidth -= (4 * 25) + (1 * 8);

        tbb[next    ].iBitmap   = 0;
        tbb[next    ].fsState   = 0;
        tbb[next    ].fsStyle   = TBSTYLE_SEP;
        tbb[next    ].idCommand = 0;

        tbb[next + 1].iBitmap   = 23;
        tbb[next + 1].fsState   = TBSTATE_ENABLED;
        tbb[next + 1].fsStyle   = TBSTYLE_BUTTON;
        tbb[next + 1].idCommand = ID_CONSOLE_START;
        tbb[next + 2].iBitmap   = 24;
        tbb[next + 2].fsState   = TBSTATE_ENABLED;
        tbb[next + 2].fsStyle   = TBSTYLE_BUTTON;
        tbb[next + 2].idCommand = ID_CONSOLE_A;
        tbb[next + 3].iBitmap   = 25;
        tbb[next + 3].fsState   = TBSTATE_ENABLED;
        tbb[next + 3].fsStyle   = TBSTYLE_BUTTON;
        tbb[next + 3].idCommand = ID_CONSOLE_B;
        tbb[next + 4].iBitmap   = 26;
        tbb[next + 4].fsState   = TBSTATE_ENABLED;
        tbb[next + 4].fsStyle   = TBSTYLE_BUTTON;
        tbb[next + 4].idCommand = ID_CONSOLE_RESET;

        next += 5;
    }

    hToolbar = CreateToolbarEx
    (   MainWindowPtr,
        WS_CHILD | WS_VISIBLE | TBSTYLE_TOOLTIPS,
        IDC_TOOLBAR,
        27, // number of images in IDB_TOOLBAR
        GetModuleHandle(NULL),
        (UINT) MAKEINTRESOURCE(IDB_TOOLBAR),
        tbb,
        (int) next,
        18,
        18,
        18,
        18,
        sizeof(TBBUTTON)
    );
    if (hToolbar == NULL)
    {   MessageBox(MainWindowPtr, "Can't create toolbar!", "WinArcadia: Error", MB_OK | MB_ICONERROR);
        cleanexit(EXIT_FAILURE);
    }

    // Send the TB_BUTTONSTRUCTSIZE message, which is required
    DISCARD SendMessage(hToolbar, TB_BUTTONSTRUCTSIZE, (WPARAM) sizeof(TBBUTTON), 0);
    DISCARD SendMessage(hToolbar, TB_AUTOSIZE, 0, 0);
    updatebiggads();
    updatesmlgads();

#define LABELWIDTH  22 // "FPS:"
#define NUMBERWIDTH 27 // "####"
    if (sparewidth >= LABELWIDTH + 5)
    {   if (!(LabelPtr = CreateWindowEx
        (   0, "STATIC",
            GetCatalogStr(CatalogPtr, MSG_FPS_GAD, "FPS:"),
            WS_CHILD | WS_VISIBLE | SS_RIGHT,
            (int) (winwidth - 15 - NUMBERWIDTH - 5 - LABELWIDTH), 9, LABELWIDTH, 18,
            MainWindowPtr, (HMENU) IDL_FPS, GetModuleHandle(NULL), NULL
        )))
        {   MessageBox(MainWindowPtr, "Can't create gadget!", "WinArcadia: Error", MB_OK | MB_ICONERROR);
            cleanexit(EXIT_FAILURE);
        }
        sparewidth -= LABELWIDTH + 5;
        DISCARD SendMessage(LabelPtr, WM_SETFONT, (WPARAM) hFont, MAKELPARAM(FALSE, 0));
    }
    oldfps = ~0;

    if (!(FPSPtr = CreateWindowEx
    (   0, "STATIC",
        "-",
        WS_CHILD | WS_VISIBLE | SS_CENTER,
        (int) (winwidth - 15 - NUMBERWIDTH), 9, NUMBERWIDTH, 18,
        MainWindowPtr, (HMENU) IDC_FPS, GetModuleHandle(NULL), NULL
    )))
    {   MessageBox(MainWindowPtr, "Can't create gadget!", "WinArcadia: Error", MB_OK | MB_ICONERROR);
        cleanexit(EXIT_FAILURE);
    }
    DISCARD SendMessage(FPSPtr, WM_SETFONT, (WPARAM) hFont, MAKELPARAM(FALSE, 0));

    DISCARD GetWindowRect(hToolbar, &therect);
    toolbarheight = therect.bottom - therect.top + 1;
}

EXPORT void settitle(void)
{   TRANSIENT STRPTR localfilepart;
    PERSIST   TEXT   tempstring[MAX_PATH + 1]; // PERSISTent so as not to overflow stack

    strcpy(titlebartext, hostmachines[machine].titlebartext);
    if (game)
    {   DISCARD GetFullPathName
        (   fn_game,
            MAX_PATH,
            tempstring,
            &localfilepart
        );
        if (localfilepart && localfilepart[0])
        {   strcat(titlebartext, ": ");
            strcat(titlebartext, localfilepart);
    }   }
    if (connected == NET_SERVER)
    {   strcat(titlebartext, " (");
        strcat(titlebartext, GetCatalogStr(CatalogPtr, MSG_SERVER, "server"));
        strcat(titlebartext, ")");
    } elif (connected == NET_CLIENT)
    {   strcat(titlebartext, " (");
        strcat(titlebartext, GetCatalogStr(CatalogPtr, MSG_CLIENT, "client"));
        strcat(titlebartext, ")");
    }
    if (paused)
    {   strcat(titlebartext, " (");
        strcat(titlebartext, GetCatalogStr(CatalogPtr, MSG_PAUSED, "paused"));
        strcat(titlebartext, ")");
    }

    /* T1; textpen(TEXTPEN_VERBOSE);
    zprintf("Setting title to %s!\n", titlebartext);
    T2; */

    DISCARD SetWindowText(MainWindowPtr, titlebartext);
}

EXPORT FLAG ctrl(void)
{   if
    (   (KeyMatrix[SCAN_LCTRL  / 8] & (1 << (SCAN_LCTRL  % 8)))
     || (KeyMatrix[SCAN_RCTRL  / 8] & (1 << (SCAN_RCTRL  % 8)))
    )
    {   return TRUE;
    } else
    {   return FALSE;
}   }
EXPORT FLAG shift(void)
{   if
    (   (KeyMatrix[SCAN_LSHIFT / 8] & (1 << (SCAN_LSHIFT % 8)))
     || (KeyMatrix[SCAN_RSHIFT / 8] & (1 << (SCAN_RSHIFT % 8)))
    )
    {   return TRUE;
    } else
    {   return FALSE;
}   }

EXPORT void openwindow(void)
{   int   i;
    DWORD style;
    HICON hicon;
    BOOL  pvParam;

    calcsize();

/* When measuring the localized menu widths (eg. when adding support for a new language),
   we give winwidth = <number>; here and look at how many rows the menu requires. */

    if (fullscreen)
    {   style = WS_POPUP | WS_MINIMIZEBOX | WS_MAXIMIZEBOX;
        // still seems to have a border anyway! :-(
    } else
    {   style = WS_POPUP | WS_MINIMIZEBOX | WS_MAXIMIZEBOX | WS_BORDER | WS_THICKFRAME;
        // Ideally, we would lock the vertical axis of the window.
        // However, Windows seems too lame to allow this; you can only
        //  have all borders draggable or none.
    }

    if (showmenubar)
    {   if (!(MenuPtr = LoadMenu(InstancePtr, MAKEINTRESOURCE(langs[language].mainmenu))))
        {   MessageBox // for some reason this doesn't get displayed
            (   NULL,
                "LoadMenu() failed!",
                "WinArcadia: Error",
                MB_ICONEXCLAMATION | MB_OK | MB_TASKMODAL
            );
            cleanexit(EXIT_FAILURE);
        }
        if (joyname[0][0])
        {   ModifyMenu(MenuPtr, ID_LEFT_1STDJOY,  MF_BYCOMMAND | MF_STRING, ID_LEFT_1STDJOY,  joyname[0]);
            ModifyMenu(MenuPtr, ID_RIGHT_1STDJOY, MF_BYCOMMAND | MF_STRING, ID_RIGHT_1STDJOY, joyname[0]);
        }
        if (joyname[1][0])
        {   ModifyMenu(MenuPtr, ID_LEFT_2NDDJOY,  MF_BYCOMMAND | MF_STRING, ID_LEFT_2NDDJOY,  joyname[1]);
            ModifyMenu(MenuPtr, ID_RIGHT_2NDDJOY, MF_BYCOMMAND | MF_STRING, ID_RIGHT_2NDDJOY, joyname[1]);
    }   }
    else
    {   MenuPtr = NULL;
    }
    ready = FALSE;
    if (showtitlebar)
    {   MainWindowPtr = CreateWindowEx(
            0,
            g_szClassName,
            hostmachines[machine].titlebartext,
            style | WS_CAPTION | WS_SYSMENU,
            winleftx,
            wintopy,
            winwidth,
            winheight,
            NULL,
            MenuPtr,
            InstancePtr, NULL);
    } else
    {   MainWindowPtr = CreateWindowEx(
            0,
            g_szClassName,
            NULL,
            style,
            winleftx,
            wintopy,
            winwidth,
            winheight,    
            NULL,
            MenuPtr,
            InstancePtr, NULL);
    }

    if (!MainWindowPtr)
    {   MessageBox(NULL, "Window creation failed!", "WinArcadia: Error", MB_ICONEXCLAMATION | MB_OK);
        cleanexit(EXIT_FAILURE);
    }

#ifdef NEWWAY
    makedisplay();
#endif

    maketoolbar();

    if (showsidebar)
    {   if (needaudit)
        {   audit(FALSE);
        }
        
        himl = ImageList_Create(16, 16, ILC_COLOR, ICONS, 0);
        // sidebar listview control automatically destroys this image list for us when it is itself destroyed

        // This has to be done in machine order.
        for (i = 0; i < ICONS - 2; i++)
        {   if (i == ARCADIA && xmas)
            {   hicon = LoadIcon(InstancePtr, MAKEINTRESOURCE(IDI_XMAS            ));
            } else
            {   hicon = LoadIcon(InstancePtr, MAKEINTRESOURCE(hostmachines[i].icon));
            }
            DISCARD ImageList_AddIcon(himl, hicon); 
            DeleteObject(hicon); 
        }
        hicon = LoadIcon(InstancePtr, MAKEINTRESOURCE(IDI_FOLDER));
        DISCARD ImageList_AddIcon(himl, hicon); 
        DeleteObject(hicon); 
        hicon = LoadIcon(InstancePtr, MAKEINTRESOURCE(IDI_SYSFOLDER));
        DISCARD ImageList_AddIcon(himl, hicon); 
        DeleteObject(hicon); 
        
        if (fullscreen)
        {   hSideBar = CreateWindowEx
            (   0,
                WC_LISTVIEW,
                "",
                WS_CHILD
              | WS_BORDER
              | WS_HSCROLL
              | WS_TABSTOP
              | WS_VSCROLL
              | WS_VISIBLE
              | LVS_LIST
              | LVS_SINGLESEL
              | LVS_SHOWSELALWAYS
              | WS_EX_CLIENTEDGE,
                showtitlebar ? (winwidth - sidebarwidth - 4) : (winwidth - sidebarwidth + 2),
                sidebartopy,
                sidebarwidth,
                sidebarheight,
                MainWindowPtr,
                NULL,
                InstancePtr,
                NULL
            );
        } else
        {   hSideBar = CreateWindowEx
            (   0,
                WC_LISTVIEW,
                "",
                WS_CHILD
              | WS_BORDER
              | WS_HSCROLL
              | WS_TABSTOP
              | WS_VSCROLL
              | WS_VISIBLE
              | LVS_LIST
              | LVS_SINGLESEL
              | LVS_SHOWSELALWAYS
              | WS_EX_CLIENTEDGE,
                winwidth - sidebarwidth - 6,
                sidebartopy,
                sidebarwidth,
                sidebarheight,
                MainWindowPtr,
                NULL,
                InstancePtr,
                NULL
            );
        }

        if (hSideBar == NULL)
        {   MessageBox(MainWindowPtr, "Can't create listview!", "WinArcadia: Error", MB_OK | MB_ICONERROR);
            cleanexit(EXIT_FAILURE);
        }
        DISCARD SendMessage(hSideBar, WM_SETFONT, (WPARAM) hFont, MAKELPARAM(TRUE, 0));
        ListView_SetImageList(hSideBar, himl, LVSIL_SMALL);
    } else
    {   hSideBar = NULL;
    }

    if (!(PopUpMenuPtr = LoadMenu(InstancePtr, MAKEINTRESOURCE(langs[language].popupmenu))))
    {   MessageBox // for some reason this doesn't get displayed
        (   NULL,
            "LoadMenu() failed!",
            "WinArcadia: Error",
             MB_ICONEXCLAMATION | MB_OK | MB_TASKMODAL
        );
        cleanexit(EXIT_FAILURE);
    }
    PopUpMenuPtr = GetSubMenu(PopUpMenuPtr, 0);
    if (joyname[0][0])
    {   ModifyMenu(MenuPtr, ID_LEFT_1STDJOY,  MF_BYCOMMAND | MF_STRING, ID_LEFT_1STDJOY,  joyname[0]);
        ModifyMenu(MenuPtr, ID_RIGHT_1STDJOY, MF_BYCOMMAND | MF_STRING, ID_RIGHT_1STDJOY, joyname[0]);
    }
    if (joyname[1][0])
    {   ModifyMenu(MenuPtr, ID_LEFT_2NDDJOY,  MF_BYCOMMAND | MF_STRING, ID_LEFT_2NDDJOY,  joyname[1]);
        ModifyMenu(MenuPtr, ID_RIGHT_2NDDJOY, MF_BYCOMMAND | MF_STRING, ID_RIGHT_2NDDJOY, joyname[1]);
    }

    if (MenuPtr)
    {   // assert(showmenubar);
        for (i = 0; i < MENUBITMAPS; i++)
        {   if (mib[i].checkmark)
            {   DISCARD SetMenuItemBitmaps
                (   MenuPtr,         // handle of menu
                    mib[i].menuitem, // menu item to receive new bitmaps
                    MF_BYCOMMAND,    // menu item flags
                    NULL,            // handle of checked bitmap
                    mib[i].bitmap    // handle of unchecked bitmap
                );
            } else
            {   DISCARD SetMenuItemBitmaps
                (   MenuPtr,         // handle of menu
                    mib[i].menuitem, // menu item to receive new bitmaps
                    MF_BYCOMMAND,    // menu item flags
                    mib[i].bitmap,   // handle of unchecked bitmap
                    NULL             // handle of checked bitmap
                );
    }   }   }
    if (PopUpMenuPtr)
    {   for (i = 0; i < MENUBITMAPS; i++)
        {   DISCARD SetMenuItemBitmaps
            (   PopUpMenuPtr,     // handle of menu
                mib[i].menuitem,  // menu item to receive new bitmaps
                MF_BYCOMMAND,     // menu item flags
                mib[i].checkmark ? NULL : mib[i].bitmap, // handle of unchecked bitmap
                mib[i].checkmark ? mib[i].bitmap : NULL  // handle of checked bitmap
            );
    }   }

    SystemMenuPtr = GetSystemMenu(MainWindowPtr, FALSE);
    if (SystemMenuPtr)
    {   DISCARD InsertMenu(SystemMenuPtr, 6, MF_BYPOSITION | MF_STRING,    ID_FILE_ICONIFY, "Iconify\tCtrl+F12");
        DISCARD InsertMenu(SystemMenuPtr, 7, MF_BYPOSITION | MF_SEPARATOR, 0,               NULL               );
    }

    if (!lpdi)
    {   DISCARD joySetCapture(MainWindowPtr, JOYSTICKID1, 0, FALSE);
        DISCARD joySetCapture(MainWindowPtr, JOYSTICKID2, 0, FALSE);
        // are these still necessary and relevant?
        // captures are automatically released when window is destroyed.
    }

    if (needcontrols)
    {   view_controls();
        needcontrols = FALSE;
    }
    if (needdips)
    {   edit_dips();
        needdips     = FALSE;
    }
    if (needmemory)
    {   view_memory();
        needmemory   = FALSE;
    }
    if (needmonitor)
    {   view_monitor();
        needmonitor  = FALSE;
    }
    if (needpalette)
    {   edit_palette();
        needpalette  = FALSE;
    }
    if (needsprite)
    {   open_spriteviewer();
        needsprite   = FALSE;
    }
    if (needgameinfo)
    {   help_gameinfo();
        needgameinfo = FALSE;
    }

    updatebiggads();
    updatesmlgads();
    updatemenus();
    redrawscreen();
    settitle();

    if (fps <= 9999)
    {   DISCARD stcul_d(fpsstring, fps);
    } else
    {   strcpy(fpsstring, "####");
    }
    DISCARD SetWindowText(FPSPtr, fpsstring);

    buildlistbrowser();
    setselection();

    DragAcceptFiles(MainWindowPtr, TRUE);
    ShowWindow(MainWindowPtr, SW_SHOW);
    DISCARD UpdateWindow(MainWindowPtr);

    if (lame)
    {   SystemParametersInfo(SPI_GETSCREENSAVEACTIVE, 0, &pvParam, 0);
        if (pvParam == TRUE)
        {   SystemParametersInfo(SPI_SETSCREENSAVEACTIVE, FALSE, 0, SPIF_SENDWININICHANGE);
    }   }

    ready = TRUE;

/*  T1; textpen(TEXTPEN_VERBOSE);
    GetWindowRect(MainWindowPtr, &therect);
    zprintf
    (   "Frame:  %d,%d-%d,%d! (%d*%d).\n",
        therect.left,
        therect.top,
        therect.right,
        therect.bottom,
        therect.right - therect.left + 1,
        therect.bottom - therect.top + 1
    );

    GetClientRect(MainWindowPtr, &therect);
    zprintf
    (   "Client: %d,%d-%d,%d! (%d*%d).\n",
        therect.left,
        therect.top,
        therect.right,
        therect.bottom,
        therect.right - therect.left + 1,
        therect.bottom - therect.top + 1
    );
    T2;

    Frame is  8 pixels wider than client.
    Frame is 46 pixels wider than client.
    The toolbar IS part of the client area, it is normally 30 pixels tall. */
}

EXPORT void closewindow(void)
{   if (ControlsWindowPtr)
    {   needcontrols = TRUE;
    }
    if (DIPsWindowPtr)
    {   needdips     = TRUE;
    }
    if (MemoryWindowPtr)
    {   needmemory   = TRUE;
    }
    if (MonitorWindowPtr)
    {   needmonitor  = TRUE;
    }
    if (PaletteWindowPtr)
    {   needpalette  = TRUE;
    }
    if (SpriteWindowPtr)
    {   needsprite   = TRUE;
    }
    if (GameInfoWindowPtr)
    {   needgameinfo = TRUE;
    }

    if (MenuPtr)
    {   SetMenu(MainWindowPtr, MenuPtr); // so it is (supposedly) automatically freed for us
    }
    if (PopUpMenuPtr)
    {   DISCARD DestroyMenu(PopUpMenuPtr);
        PopUpMenuPtr = NULL;
    }
    // PostMessage(MainWindowPtr, WM_CLOSE, 0, 0);
    DestroyWindow(MainWindowPtr);
    MainWindowPtr = NULL;
}

EXPORT void listen_close(void)
{   if (!ListeningWindowPtr) // important!
    {   return;
    }

    DestroyWindow(ListeningWindowPtr);
    ListeningWindowPtr = NULL;

    sound_on();
}

MODULE void printusage(STRPTR progname)
{   T1; // textpen(TEXTPEN_DEFAULT);
    printf(               "%s: %s\n" // not zprintf()!
                 "[[FILE]=<file>]\n"
               "[AUTOSAVE=ON|OFF]\n"
             "[COLLISIONS=ON|OFF]\n"
"[FRAMESKIP=1|2|3|4|5|6|7|8|9|10]\n"
             "[FULLSCREEN=ON|OFF]\n"
                "[MACHINE=ARCADIA|INTERTON|ELEKTOR|PIPBUG|INSTRUCTOR|CD2650\n"
                         "|ASTROWARS|GALAXIA|MALZAK|CHAOS|DOLPHIN]\n"
                "[MENUBAR=ON|OFF]\n"
                 "[PAUSED=ON|OFF]\n"
               "[SETTINGS=<file>]\n"
                  "[SIZE=1|2|3|4]\n"
                  "[SOUND=ON|OFF]\n"
                "[STRETCH=ON|OFF]\n"
               "[TITLEBAR=ON|OFF]\n"
                "[TOOLBAR=ON|OFF]\n"
                  "[TURBO=ON|OFF]\n"
                       "[UPGRADE]\n"
                 , LLL, MSG_USAGE, "Usage")
                 , progname
    );

    cleanexit(EXIT_FAILURE);
}

EXPORT void debug_edit()
{   sound_off(FALSE);
    opening = TRUE;
    DISCARD DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_CHANGE), MainWindowPtr, EditDlgProc);
    opening = FALSE;
    sound_on();
}

BOOL CALLBACK EditDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   PERSIST FLAG dontaddr,
                 dontto;

    // IAR and RAS0..RAS7 are allowed to be up to 32767.
    // All others are limited to 255.

    switch (Message)
    {
    case WM_INITDIALOG:
        charset = CHARSET_WIN;
        tolimit = 255;
        dontto = dontaddr = FALSE;
        if (jump)
        {   DISCARD SetWindowText(hwnd, LLL, MSG_HAIL_JUMP, "Jump to address"));
            DISCARD SetDlgItemText(hwnd, IDC_CHANGE_ADDRESS, "IAR");
            EnableWindow(GetDlgItem(hwnd, IDC_CHANGE_ADDRESS), FALSE);
            change_read("IAR");
        } else
        {   DISCARD SetWindowText(hwnd, LLL, MSG_HAIL_EDIT, "Edit Register/Memory"));
            DISCARD SetDlgItemText(hwnd, IDC_CHANGE_ADDRESS, lastchange);
            change_read(lastchange);
        }
        DISCARD SetDlgItemText(hwnd, IDC_CHANGE_FROM,                    fromstr  );
        DISCARD SetDlgItemText(hwnd, IDC_CHANGE_TO,                      tostr    );
        if (language == LANGUAGE_RUS)
        {   DISCARD SendMessage(GetDlgItem(hwnd, IDL_CHANGE        ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDL_FROM          ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDL_TO            ), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
        }
        DISCARD SetDlgItemText(hwnd, IDL_CHANGE,        LLL, MSG_CHANGE, "Change"));
        DISCARD SetDlgItemText(hwnd, IDL_FROM,          LLL, MSG_FROM  , "from"  ));
        DISCARD SetDlgItemText(hwnd, IDL_TO,            LLL, MSG_TO    , "to"    ));
        charset = CHARSET_DOS;

        // DISCARD SetForegroundWindow(hwnd); doesn't work

        return TRUE;
    case WM_CLOSE: // no need for acase
        change_write(globalstring, FALSE);
        clearkybd();
        DISCARD EndDialog(hwnd, 0);

        return TRUE;
    case WM_COMMAND: // no need for acase
        if (LOWORD (wParam) == IDOK)
        {   // assert(tonum >= 0 && tonum <= tolimit);
            change_write(globalstring, FALSE);
            clearkybd();
            DISCARD EndDialog(hwnd, 0);

            return TRUE;
        }
        if (LOWORD (wParam) == IDCANCEL) // eg Esc key
        {   T1; textpen(TEXTPEN_DEBUG); zprintf("Cancelled.\n\n"); T2;
            clearkybd();
            DISCARD EndDialog(hwnd, 0);

            return TRUE;
        }
        if (HIWORD (wParam) == EN_UPDATE)
        {   if (LOWORD (wParam) == IDC_CHANGE_TO)
            {   if (dontto)
                {   dontto = FALSE;
                } else
                {   DISCARD GetWindowText(GetDlgItem(hwnd, IDC_CHANGE_TO), tostr, sizeof(tostr));
                    if (strlen(tostr) > 5) // "$7FFF" or "32767"
                    {   tostr[5] = 0;
                        dontto = TRUE;
                        DISCARD SetDlgItemText(hwnd, IDC_CHANGE_TO, tostr);
                    }
                    // maybe we should check for illegal chars?
                    tonum = readbase(tostr);
                    if (tonum < 0 || tonum > tolimit)
                    {   tonum = 0;
                        if (tolimit == 255)
                        {   strcpy(tostr, "$00");
                        } else
                        {   // assert(tolimit == 32767);
                            strcpy(tostr, "$0000");
                        }
                        dontto = TRUE;
                        DISCARD SetDlgItemText(hwnd, IDC_CHANGE_TO, tostr);
            }   }   }
            elif (LOWORD (wParam) == IDC_CHANGE_ADDRESS)
            {   if (dontaddr)
                {   dontaddr = FALSE;
                } else
                {   DISCARD GetWindowText(GetDlgItem(hwnd, IDC_CHANGE_ADDRESS), globalstring, sizeof(globalstring));
                    DISCARD strupr(globalstring);
                    if (strlen(globalstring) > 32)
                    {   globalstring[32] = 0;
                        dontaddr = TRUE;
                        DISCARD SetDlgItemText(hwnd, IDC_CHANGE_ADDRESS, globalstring);
                    }

                    change_read(globalstring);
                    DISCARD SetDlgItemText(hwnd, IDC_CHANGE_FROM, fromstr);
                    DISCARD SetDlgItemText(hwnd, IDC_CHANGE_TO  ,   tostr);
        }   }   }

        return FALSE;
    default: // no need for adefault
        return FALSE;
}   }

EXPORT void openconsole(void)
{
#ifdef GUI
    int   hKybd, hScrn;
    FILE* hf;

    // assert(!consoleopen);
#endif

    consoleopen = TRUE;
    if (lame && language == LANGUAGE_RUS)
// Is codepage 866 going to be available in the registry? or only 1251?
// Indeed, 1251 would be better; we wouldn't have to do the translation to codepage 866.
    {   DISCARD SetConsoleCP(866);
        DISCARD SetConsoleOutputCP(866);
        // Error 120: "This function is not supported on this system." under non-lame versions
    }

#ifdef CLI
    // shouldn't we initialize ConsoleWindowPtr?
    return;
#endif
#ifdef GUI
    DISCARD AllocConsole();
    hConsole = GetStdHandle(STD_OUTPUT_HANDLE);
    hScrn = _open_osfhandle((long) hConsole,                        _O_TEXT); // might need freeing at program exit
    hKybd = _open_osfhandle((long) GetStdHandle(STD_INPUT_HANDLE ), _O_TEXT); // might need freeing at program exit
    hf = _fdopen(hScrn, "w"); // might need freeing at program exit
    *stdout = *hf;
    hf = _fdopen(hKybd, "r"); // might need freeing at program exit
    *stdin = *hf;
    DISCARD setvbuf(stdout, NULL, _IONBF, 0);
    DISCARD setvbuf(stdin , NULL, _IONBF, 0);

    DISCARD SetConsoleTitle("WinArcadia Console");
    Sleep(40); // to ensure the window title was updated
    ConsoleWindowPtr = FindWindow(NULL, "WinArcadia Console");
#endif
}

EXPORT void activateconsole(void)
{   if (ConsoleWindowPtr)
    {   DISCARD SetForegroundWindow(ConsoleWindowPtr);
}   }

EXPORT void breakparts(void)
{   TRANSIENT signed int i,
                         length,
                         where;
    TRANSIENT STRPTR     localfilepart;
    PERSIST   TEXT       tempstring[MAX_PATH + 1]; // PERSISTent so as not to overflow stack

    // Splits a complete pathname (in fn_game) into pathpart and filepart.

    DISCARD GetFullPathName
    (   fn_game,
        MAX_PATH,
        tempstring,
        &localfilepart
    );
    strcpy(fn_game, tempstring);

    strcpy(pathpart, fn_game);
    where = -1;
    length = (int) strlen(fn_game);
    for (i = 0; i < length; i++)
    {   if
        (   fn_game[i] == ':'
         || fn_game[i] == '\\'
        )
        {   where = i;
            // don't break, we want the LAST component, not the FIRST
    }   }
    strcpy(filepart, &fn_game[where + 1]);
    pathpart[where + 1] = 0;
}

EXPORT void do_autopause(WPARAM wParam, LPARAM lParam)
{   // For WA_INACTIVE: "(HWND) lParam is the handle of the window being activated."

    if
    (   LOWORD(wParam) == WA_INACTIVE
     && autopause
     && !crippled
     && !paused
     && !opening
     && (       (HWND) lParam == NULL
         || (   (HWND) lParam !=      MainWindowPtr
             && (HWND) lParam !=   MonitorWindowPtr
             && (HWND) lParam !=  ControlsWindowPtr
             && (HWND) lParam !=      DIPsWindowPtr
             && (HWND) lParam != ListeningWindowPtr
             && (HWND) lParam !=    MemoryWindowPtr
             && (HWND) lParam !=   PaletteWindowPtr
             && (HWND) lParam !=    SpriteWindowPtr
             && (HWND) lParam !=  GameInfoWindowPtr
    )   )   )
    {   emu_pause(TRUE);
}   }

EXPORT void updatefps(void)
{   if (!MainWindowPtr)
    {   return; // unimportant?
    }

    if (fps != oldfps)
    {   oldfps = fps;
        if (fps <= 9999)
        {   DISCARD stcul_d(fpsstring, fps);
        } else
        {   strcpy(fpsstring, "####");
        }
        DISCARD SetWindowText(FPSPtr, fpsstring);
}   }

MODULE void subreq(HWND hwnd)
{   opening = TRUE;
    DISCARD DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REDEFINESUB), hwnd, RedefineSubDlgProc);
    opening = FALSE;
    refreshkeys(hwnd); // this is overkill
}

EXPORT void project_register(void)
{   sound_off(FALSE);
    opening = TRUE;
    DISCARD DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_REGISTER), MainWindowPtr, RegisterDlgProc);
    opening = FALSE;
    sound_on();
}

EXPORT void datestamp(int which)
{   SYSTEMTIME TheTime;

    GetLocalTime(&TheTime);
    sprintf(hiscore[which].thetime, "%d:%02d", (int) TheTime.wHour, (int) TheTime.wMinute);
    sprintf(hiscore[which].thedate, "%d/%d/%02d", (int) TheTime.wDay, (int) TheTime.wMonth, (int) (TheTime.wYear % 100));
}

/* Some relevant locale IDs are:
   $4 07 = German
   $4 09 = American
   $4 0C = French
   $4 10 = Italian
   $8 09 = English
   $C 09 = Australian
   $? 13 = Dutch
   $? 19 = Russian */

EXPORT void freelanguage(void)
{   if (langs[language].dosbuffer)
    {   free(langs[language].dosbuffer);
        langs[language].dosbuffer = NULL;
    }
    if (langs[language].winbuffer)
    {   free(langs[language].winbuffer);
        langs[language].winbuffer = NULL;
    }
    if (langs[language].offset)
    {   free(langs[language].offset);
        langs[language].offset = NULL;
}   }

EXPORT void changelanguage(void)
{   closewindow();
    load_catalog();
    translate();
    openwindow();
}

EXPORT void load_catalog(void)
{   UBYTE        *p8data,
                 *fbase;
    UBYTE         LocalIOBuffer[8];
    UBYTE*        dataptr = NULL;
    int           code;
    struct Chunk* p8Chunk;
    ULONG         fsize,
                  thesize;
    FILE*         TheLocalHandle; // LocalHandle is a variable of winbase.h

/* Open the file.
   Read the first 8 bytes.
   Assure it is a FORM.
   Allocate enough memory to read the rest of the FORM chunk.
   Read the rest of the FORM chunk.
   Close the file. */

    if (language == LANGUAGE_ENG)
    {   return;
    }

    if (!(TheLocalHandle = fopen(langs[language].filename, "rb")))
    {   code = 1;                               /* can't open file */
        goto DONE;
    }
    if (fread(LocalIOBuffer, 8, 1, TheLocalHandle) != 1)
    {   code = 2;                               /* can't read file (2), or too short (3) */
        DISCARD fclose(TheLocalHandle);
        // TheLocalHandle = NULL;
        goto DONE;
    }

    p8Chunk = (struct Chunk *) &LocalIOBuffer[0];
    if (p8Chunk->ckID != (ULONG) ID_FORM)
    {   code = 3;
        DISCARD fclose(TheLocalHandle);
        // TheLocalHandle = NULL;
        goto DONE;
    }
    fsize = (LocalIOBuffer[4] * 16777216)
          + (LocalIOBuffer[5] *    65536)
          + (LocalIOBuffer[6] *      256)
          +  LocalIOBuffer[7];
    if (!(fbase = (UBYTE *) malloc(fsize)))
    {   code = 4;
        DISCARD fclose(TheLocalHandle);
        // TheLocalHandle = NULL;
        goto DONE;
    }
    p8data = fbase;
    if (fread(p8data, fsize, 1, TheLocalHandle) != 1)
    {   code = 5;
        DISCARD fclose(TheLocalHandle);
        // TheLocalHandle = NULL;
        goto DONE;
    }
    DISCARD fclose(TheLocalHandle);
    // TheLocalHandle = NULL;

/* Assure it is a CTLG.
   Look at each chunk: if it is a STRS, remember offset and length.
   Allocate enough memory to store a copy of the STRS chunk.
   Copy the STRS chunk. */

    if (MAKE_ID(*p8data, *(p8data + 1), *(p8data + 2), *(p8data + 3)) != ID_CTLG)
    {   code = 6;
        goto DONE;
    }
    p8data += 4;

    while (p8data < fbase + fsize)
    {   thesize = (p8data[4] * 16777216)
                + (p8data[5] *    65536)
                + (p8data[6] *      256)
                +  p8data[7];
        p8Chunk = (struct Chunk *) p8data;
        switch (p8Chunk->ckID)
        {
        case ID_STRS:
            langs[language].size = (int) thesize;
            dataptr = (UBYTE *) (p8data + 8);
        }
        p8data += 8 + thesize;
        if ((thesize & 1) == 1)
        {   p8data++;
    }   }

    if (dataptr == NULL) // no STRS chunk found
    {   code = 7;
        goto DONE;
    }

    langs[language].winbuffer = (UBYTE *) malloc((ULONG) langs[language].size);
    if (langs[language].winbuffer == NULL)
    {   code = 8;
        goto DONE;
    }
    langs[language].dosbuffer = (UBYTE *) malloc((ULONG) langs[language].size);
    if (langs[language].dosbuffer == NULL)
    {   free(langs[language].winbuffer);
        langs[language].winbuffer = NULL;
        code = 9;
        goto DONE;
    }

    code = 0; // success
    memcpy(langs[language].winbuffer, dataptr, (unsigned int) langs[language].size);

    free(fbase);
    // fbase = NULL;

DONE:
    if (code)
    {   T1; textpen(TEXTPEN_ERROR);
        zprintf("Can't load \"%s\" (error %d)!\n\n", langs[language].filename, code); T2;
    } else
    {   // T1; zprintf("Loaded \"%s\" (%d bytes at $%X).\n\n", langs[language].filename, langs[language].size, dataptr); T2;
        parse_catalog();
}   }

EXPORT STRPTR GetCatalogStr(int catalog, LONG stringNum, STRPTR defaultString)
{   if
    (   language                          == LANGUAGE_ENG
     || langs[language].dosbuffer         == NULL
     || langs[language].winbuffer         == NULL
     || langs[language].offset            == NULL
     || stringNum                         >  highestid
     || langs[language].offset[stringNum] == 0
    )
    {   return defaultString;
    }
    if (charset == CHARSET_DOS)
    {   return &langs[language].dosbuffer[langs[language].offset[stringNum]];
    } else
    {   // assert(charset == CHARSET_WIN);
        return &langs[language].winbuffer[langs[language].offset[stringNum]];
}   }

EXPORT void parse_catalog(void)
{   int i, j,
        id,
        length;
    
    // assert(langs[language].winbuffer);

    // find the highest ID, and convert strings to MS-DOS format
    i = 0;
    highestid = 0;
    while (i < langs[language].size)
    {   id     = (langs[language].winbuffer[i    ] * 16777216)
               + (langs[language].winbuffer[i + 1] *    65536)
               + (langs[language].winbuffer[i + 2] *      256)
               + (langs[language].winbuffer[i + 3]);
        if (id > highestid)
        {   highestid = id;
        }
        length = (langs[language].winbuffer[i + 4] * 16777216)
               + (langs[language].winbuffer[i + 5] *    65536)
               + (langs[language].winbuffer[i + 6] *      256)
               + (langs[language].winbuffer[i + 7]);

        i += 8;
        for (j = 0; j <= length; j++)
        {   if (langs[language].winbuffer[i + j] >= 160)
            {   langs[language].dosbuffer[i + j] = translation[language == LANGUAGE_RUS ? 1 : 0][langs[language].winbuffer[i + j] - 160];
            } else
            {   langs[language].dosbuffer[i + j] = langs[language].winbuffer[i + j];
        }   }
        i += length + 1;
        if   ((i % 4) == 1) i += 3;
        elif ((i % 4) == 2) i += 2;
        elif ((i % 4) == 3) i++;
    }

    langs[language].offset = (ULONG*) calloc((size_t) (4 * (highestid + 1)), 1);

    // fill langs[].offset[] array
    i = 0;
    while (i < langs[language].size)
    {   id     = (langs[language].winbuffer[i    ] * 16777216)
               + (langs[language].winbuffer[i + 1] *    65536)
               + (langs[language].winbuffer[i + 2] *      256)
               + (langs[language].winbuffer[i + 3]);
        length = (langs[language].winbuffer[i + 4] * 16777216)
               + (langs[language].winbuffer[i + 5] *    65536)
               + (langs[language].winbuffer[i + 6] *      256)
               + (langs[language].winbuffer[i + 7]);
        langs[language].offset[id] = (ULONG) (i + 8);

        i += 8 + length + 1;
        if   ((i % 4) == 1) i += 3;
        elif ((i % 4) == 2) i += 2;
        elif ((i % 4) == 3) i++;
}   }

MODULE void translate(void)
{   charset = CHARSET_WIN;

    keyname[SCAN_SPACEBAR].name   = LLL, MSG_KEY_SPACE                 , "Space"                               );
//  keyname[SCAN_BKSP    ].name   = LLL, MSG_KEY_BKSP                  , "BkSp"                                );
    keyname[SCAN_TAB     ].name   = LLL, MSG_KEY_TAB                   , "Tab"                                 );
//  keyname[SCAN_ESC     ].name   = LLL, MSG_KEY_ESC                   , "Esc"                                 );
    keyname[SCAN_DEL     ].name   = LLL, MSG_KEY_DEL                   , "Del"                                 );
    keyname[SCAN_UP      ].name   = LLL, MSG_KEY_UP                    , "Up"                                  );
    keyname[SCAN_DOWN    ].name   = LLL, MSG_KEY_DN                    , "Dn"                                  );
    keyname[SCAN_RIGHT   ].name   = LLL, MSG_KEY_RT                    , "Rt"                                  );
    keyname[SCAN_LEFT    ].name   = LLL, MSG_KEY_LT                    , "Lt"                                  );
//  keyname[SCAN_HELP    ].name   = LLL, MSG_KEY_HELP                  , "Help"                                );
    keyname[SCAN_LSHIFT  ].name   = LLL, MSG_KEY_LSHFT                 , "LShft"                               );
    keyname[SCAN_RSHIFT  ].name   = LLL, MSG_KEY_RSHFT                 , "RShft"                               );
//  keyname[SCAN_CL      ].name   = LLL, MSG_KEY_CL                    , "CL"                                  );
//  keyname[SCAN_LCTRL   ].name   = LLL, MSG_KEY_LCTRL                 , "LCtrl"                               );
//  keyname[SCAN_RCTRL   ].name   = LLL, MSG_KEY_RCTRL                 , "RCtrl"                               );
    keyname[SCAN_LALT    ].name   = LLL, MSG_KEY_LALT                  , "LAlt"                                );
    keyname[SCAN_RALT    ].name   = LLL, MSG_KEY_RALT                  , "RAlt"                                );
//  keyname[SCAN_LAMIGA  ].name   = LLL, MSG_KEY_LAMI                  , "LAmi"                                );
//  keyname[SCAN_RAMIGA  ].name   = LLL, MSG_KEY_RAMI                  , "RAmi"                                );
//  keyname[SCAN_STOP    ].name   = LLL, MSG_KEY_STOP                  , "Stop"                                );
//  keyname[SCAN_PLAY    ].name   = LLL, MSG_KEY_PLAY                  , "Play"                                );
//  keyname[SCAN_PREV    ].name   = LLL, MSG_KEY_PREV                  , "Prev"                                );
//  keyname[SCAN_NEXT    ].name   = LLL, MSG_KEY_NEXT                  , "Next"                                );
//  keyname[SCAN_RWND    ].name   = LLL, MSG_KEY_RWND                  , "Rwnd"                                );
//  keyname[SCAN_FFWD    ].name   = LLL, MSG_KEY_FFWD                  , "FFwd"                                );
//  keyname[SCAN_RESET   ].name   = LLL, MSG_KEY_RESET                 , "Reset"                               );
//  keyname[SCAN_WHLDN   ].name   = LLL, MSG_KEY_WHLDN                 , "WhlDn"                               );
//  keyname[SCAN_WHLUP   ].name   = LLL, MSG_KEY_WHLUP                 , "WhlUp"                               );

    filekind[0].hail              = LLL, MSG_HAIL_SAVEILBM             , "Save Screenshot as IFF ILBM"   );
    filekind[1].hail              = LLL, MSG_HAIL_SAVEACBM             , "Save Screenshot as IFF ACBM"   );
    filekind[2].hail              = LLL, MSG_HAIL_SAVEBMP              , "Save Screenshot as BMP"        );
    filekind[3].hail              = LLL, MSG_HAIL_SAVEPCX              , "Save Screenshot as PCX"        );
    filekind[4].hail              = LLL, MSG_HAIL_SAVEBIN              , "Save Game ROM"                 );
    filekind[5].hail              = LLL, MSG_HAIL_SAVEPGM              , "Save Program (PGM)"            );
    filekind[6].hail              = LLL, MSG_HAIL_SAVETVC              , "Save Elektor Program (TVC)"    );
    filekind[7].hail              = LLL, MSG_HAIL_SAVECOS              , "Save Snapshot"                 );
    filekind[8].hail              = LLL, MSG_HAIL_SAVECOR              , "Start Recording"               );
    filekind[9].hail              = LLL, MSG_HAIL_SAVEEAP              , "Save SpriteMakerEA Project"    );
    filekind[10].hail             = LLL, MSG_HAIL_SAVEEAL              , "Save TextMakerEA Project"      );
 // filekind[11].hail             = LLL, MSG_HAIL_SAVEAVI              , "Start Recording (RIFF AVI)"    );
 // filekind[12].hail             = NULL; // ZIP
 // filekind[13].hail             = NULL; // RAW
 // filekind[14].hail             = NULL; // AOF

    charset = CHARSET_DOS;

    colourname[0] = (STRPTR) GetCatalogStr(CatalogPtr, MSG_COLOUR_WHITE,  "white" );
    colourname[1] = (STRPTR) GetCatalogStr(CatalogPtr, MSG_COLOUR_YELLOW, "yellow");
    colourname[2] = (STRPTR) GetCatalogStr(CatalogPtr, MSG_COLOUR_PURPLE, "purple");
    colourname[3] = (STRPTR) GetCatalogStr(CatalogPtr, MSG_COLOUR_RED,    "red"   );
    colourname[4] = (STRPTR) GetCatalogStr(CatalogPtr, MSG_COLOUR_CYAN,   "cyan"  );
    colourname[5] = (STRPTR) GetCatalogStr(CatalogPtr, MSG_COLOUR_GREEN,  "green" );
    colourname[6] = (STRPTR) GetCatalogStr(CatalogPtr, MSG_COLOUR_BLUE,   "blue"  );
    colourname[7] = (STRPTR) GetCatalogStr(CatalogPtr, MSG_COLOUR_BLACK,  "black" );

    if
    (   machine == ASTROWARS
     || machine == GALAXIA
     || machine == MALZAK
     || (game && whichgame != -1)
    )
    {   generate_autotext();
}   }

EXPORT void capslock_off(void)
{   UBYTE keys[256];

    GetKeyboardState(keys);
    if (keys[VK_CAPITAL])
    {   if (lame)
        {   // simulate keypress
            keybd_event(VK_CAPITAL, 0x45, KEYEVENTF_EXTENDEDKEY | 0, 0);
            // simulate key release
            keybd_event(VK_CAPITAL, 0x45, KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP, 0);
        } else
        {   keys[VK_CAPITAL] = 1;
            SetKeyboardState(keys);
}   }   }

EXPORT void ReadJoystick(int joy)
{   if (lpdi)
    {   ff_ReadJoystick(0);
        ff_ReadJoystick(1);
    } else
    {   mm_ReadJoystick(0);
        mm_ReadJoystick(1);
}   }

EXPORT void setpri(void)
{   // assert(ProcessPtr);
    // assert(ThreadPtr);

    switch (priority)
    {
	case PRI_HIGH:
        SetPriorityClass(ProcessPtr, HIGH_PRIORITY_CLASS);
        SetThreadPriority(ThreadPtr, THREAD_PRIORITY_NORMAL      ); //  0
     // SetThreadPriority(ThreadPtr, THREAD_PRIORITY_ABOVE_NORMAL); // +1
#ifdef VERBOSE
        T1; zprintf("High priority.\n"); T2;
#endif
    acase PRI_NORMAL:
        SetPriorityClass(ProcessPtr, NORMAL_PRIORITY_CLASS);
        SetThreadPriority(ThreadPtr, THREAD_PRIORITY_NORMAL      ); //  0
#ifdef VERBOSE
        T1; zprintf("Normal priority.\n"); T2;
#endif
    acase PRI_LOW:
        if (lame)
        {   SetPriorityClass(ProcessPtr, BELOW_NORMAL_PRIORITY_CLASS);
        } else
        {   SetPriorityClass(ProcessPtr, NORMAL_PRIORITY_CLASS);
        }
        SetThreadPriority(ThreadPtr, THREAD_PRIORITY_BELOW_NORMAL); // -1
#ifdef VERBOSE
        T1; zprintf("Low priority.\n"); T2;
#endif
}   }

EXPORT void read_rtc(void)
{   PERSIST SYSTEMTIME TheTime; // PERSISTent for speed

    GetLocalTime(&TheTime);

    rtc.wkday = (int) TheTime.wDayOfWeek;
    rtc.month = (int) TheTime.wMonth;
    rtc.day   = (int) TheTime.wDay;
    rtc.hour  = (int) TheTime.wHour;
    rtc.min   = (int) TheTime.wMinute;
    rtc.sec   = (int) TheTime.wSecond;
}

EXPORT FLAG confirm(void)
{   if (!confirmable)
    {   return TRUE;
    } // implied else

    sound_off(FALSE);
    if (MessageBox
    (    NULL,
         GetCatalogStr(CatalogPtr, MSG_CONFIRM, "Really quit?"),
         "WinArcadia",
         MB_ICONQUESTION | MB_YESNO
    ) == IDYES)
    {   sound_on();
        return TRUE;
    } // implied else
    sound_on();
    return FALSE;
}

BOOL CALLBACK RearrangeDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   int i;

    switch (Message)
    {
    case WM_INITDIALOG:
        charset = CHARSET_WIN;
        DISCARD SetWindowText( hwnd,             LLL, MSG_HAIL_REARRANGE, "Rearrange Gamepad Buttons" ));

        if (language == LANGUAGE_RUS)
        {   DISCARD SendMessage(GetDlgItem(hwnd, IDL_1STPAD), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
            DISCARD SendMessage(GetDlgItem(hwnd, IDL_2NDPAD), WM_SETFONT, (WPARAM) hRussianFont, MAKELPARAM(FALSE, 0));
        }       
        if (joyname[0][0])
        {   DISCARD SetDlgItemText(hwnd, IDL_1STPAD, joyname[0]                     );
        } else
        {   DISCARD SetDlgItemText(hwnd, IDL_1STPAD, LLL, MSG_1STDPAD, "1st gamepad"));
        }
        // these strings ("1st|2nd gamepad") aren't actually in the catalogs
        if (joyname[1][0])
        {   DISCARD SetDlgItemText(hwnd, IDL_2NDPAD, joyname[1]                     );
        } else
        {   DISCARD SetDlgItemText(hwnd, IDL_2NDPAD, LLL, MSG_2NDDPAD, "2nd gamepad"));
        }
        charset = CHARSET_DOS;

        for (i = 0; i < 24; i++)
        {   if
            (   button[0][0] == rearrange[i].rearr[0]
             && button[0][1] == rearrange[i].rearr[1]
             && button[0][2] == rearrange[i].rearr[2]
            )
            {   // assert(button[0][3] == rearrange[i].rearr[3]);
                DISCARD CheckRadioButton
                (   hwnd,
                    IDC_1STRA1234,
                    IDC_1STRA4321,
                    rearrange[i].hGad[0]
                );
            }
            if
            (   button[1][0] == rearrange[i].rearr[0]
             && button[1][1] == rearrange[i].rearr[1]
             && button[1][2] == rearrange[i].rearr[2]
            )
            {   // assert(button[1][3] == rearrange[i].rearr[3]);
                DISCARD CheckRadioButton
                (   hwnd,
                    IDC_2NDRA1234,
                    IDC_2NDRA4321,
                    rearrange[i].hGad[1]
                );
        }   }
        for (i = 24; i < 48; i++)
        {   if
            (   button[0][4] == rearrange[i].rearr[0]
             && button[0][5] == rearrange[i].rearr[1]
             && button[0][6] == rearrange[i].rearr[2]
            )
            {   // assert(button[0][7] == rearrange[i].rearr[3]);
                DISCARD CheckRadioButton
                (   hwnd,
                    IDC_1STRA5678,
                    IDC_1STRA8765,
                    rearrange[i].hGad[0]
                );
            }
            if
            (   button[1][4] == rearrange[i].rearr[0]
             && button[1][5] == rearrange[i].rearr[1]
             && button[1][6] == rearrange[i].rearr[2]
            )
            {   // assert(button[1][7] == rearrange[i].rearr[3]);
                DISCARD CheckRadioButton
                (   hwnd,
                    IDC_2NDRA5678,
                    IDC_2NDRA8765,
                    rearrange[i].hGad[1]
                );
        }   }

        return TRUE;
    case WM_CLOSE:
        clearkybd();
        DISCARD EndDialog(hwnd, 0);
        return TRUE;
    case WM_DESTROY:
        return FALSE;
    case WM_COMMAND:
        for (i = 0; i < 24; i++)
        {   if (LOWORD(wParam) == rearrange[i].hGad[0])
            {   button[0][0] = rearrange[i].rearr[0];
                button[0][1] = rearrange[i].rearr[1];
                button[0][2] = rearrange[i].rearr[2];
                button[0][3] = rearrange[i].rearr[3];
            }
            if (LOWORD(wParam) == rearrange[i].hGad[1])
            {   button[1][0] = rearrange[i].rearr[0];
                button[1][1] = rearrange[i].rearr[1];
                button[1][2] = rearrange[i].rearr[2];
                button[1][3] = rearrange[i].rearr[3];
        }   }
        for (i = 24; i < 48; i++)
        {   if (LOWORD(wParam) == rearrange[i].hGad[0])
            {   button[0][4] = rearrange[i].rearr[0];
                button[0][5] = rearrange[i].rearr[1];
                button[0][6] = rearrange[i].rearr[2];
                button[0][7] = rearrange[i].rearr[3];
            }
            if (LOWORD(wParam) == rearrange[i].hGad[1])
            {   button[1][4] = rearrange[i].rearr[0];
                button[1][5] = rearrange[i].rearr[1];
                button[1][6] = rearrange[i].rearr[2];
                button[1][7] = rearrange[i].rearr[3];
        }   }
        return FALSE;
    default: // no need for adefault
        return FALSE;
}   }

EXPORT void updatespeech(void)
{   speakoffset = 0;
}

EXPORT void flipbool(int* thevar, int command)
{   if (*thevar) *thevar = FALSE; else *thevar = TRUE;
    updatemenu(command);
}

EXPORT void help_manual(void)
{   PERSIST TEXT localolddir[MAX_PATH + 1]; // PERSISTent so as not to overflow the stack

    DISCARD GetCurrentDirectory(MAX_PATH + 1, localolddir);
    DISCARD SetCurrentDirectory(ProgDir);
    openurl("WA.HTM");
    DISCARD SetCurrentDirectory(localolddir);
}

EXPORT void calibrate(void)
{   // assert(lpdi);
    IDirectInput_RunControlPanel(lpdi, MainWindowPtr, 0);
}

EXPORT void openurl(STRPTR command)
{   DISCARD ShellExecute(0, NULL, command, NULL, NULL, SW_SHOW);
}

EXPORT void system_loadedconfig(void)
{   int i;

    for (i = 0; i < GUESTCOLOURS; i++)
    {   pens[COLOURFUL][GUESTCOLOURS + i] = (pens[COLOURFUL][i] & 0x00FEFEFE) >> 1;
        pens[GREYSCALE][GUESTCOLOURS + i] = (pens[GREYSCALE][i] & 0x00FEFEFE) >> 1;
    }
    if (lame)
    {   useinternal = FALSE;
}   }

EXPORT void resetcolour(int thecolourset, int i, FLAG update)
{   // assert(!update);

    pens[thecolourset][               i] = defpencolours[thecolourset][wsc][i];
    pens[thecolourset][GUESTCOLOURS + i] = (pens[thecolourset][i] & 0x00FEFEFE) >> 1;
}

EXPORT void fixshadowcolours(void)
{   int i, j;

    for (i = 0; i < 2; i++)
    {   for (j = 0; j < GUESTCOLOURS; j++)
        {   pens[i][GUESTCOLOURS + j] = (pens[i][j] & 0x00FEFEFE) >> 1;
}   }   }
